#include "CDAS2ADST.h"

#include <RecEvent.h>
#include <SDEvent.h>
#include <Traces.h>
#include <Er.h>
#include <TTimeStamp.h>
#include <TMath.h>
#include <IoSdData.h>
#include <SdRecLpnhe.h>
#include <STClibrary.h>

#include <Es.h>
#include <Ec.h>

#include <iostream>
#include <sstream>

using namespace std;
using namespace cdas;

CDAS2ADST::CDAS2ADST():
  fMinEnergyToStoreTraces(0.),
  fStoreVEMTraces(false),
  fStoreFADCTraces(false)
{

}

CDAS2ADST::~CDAS2ADST(){
}



void   
CDAS2ADST::FillADSTEvent(RecEvent& adst, const TErEvent& cdas){
  
  stringstream augerId;
  augerId << cdas.AugerId;
  adst.SetEventId(augerId.str());
  
  TTimeStamp evTime(cdas.UTCTime);
  const unsigned int time= evTime.GetTime();
  const unsigned int date= evTime.GetDate();
  adst.SetHHMMSS(time);
  adst.SetYYMMDD( date - 20000000);
  
  TBits stationBits(1600); // TBits enlarges automatically of needed!
  SDEvent& sd = adst.GetSDEvent();
  FillSDEvent(sd, cdas);
  FillStationsInfo(sd, stationBits, cdas.fSelectedStations, eCandidate, cdas);
  FillStationsInfo(sd, stationBits, cdas.fRandomStations, eAccidental, cdas);
  FillStationsInfo(sd, stationBits, cdas.fAmbiguousStations, eAccidental, cdas);
  FillStationsInfo(sd, stationBits, cdas.fSilentStations, eSilent, cdas);
  FillStationsInfo(sd, stationBits, cdas.fBadStations, eUnknown, cdas);
  
  stationBits.Compact();

  adst.GetDetector().SetActiveStations(stationBits);
  


}


void CDAS2ADST::FillSDEvent(SDEvent& sd, const TErEvent& cdas){
  
  sd.SetId(cdas.Id);
  sd.SetNumberOfCandidates( cdas.fSelectedStations.size() );
  sd.SetEventTime(cdas.fRec.fGPSSecond, 0);
  
  double julianDay = 0.;
  double refJD = 0;
  gps2jd(cdas.fRec.fGPSSecond, &refJD, &julianDay);
  double lst = 0;
  const double E0  = 477256.0;
  const double N0  = 6099203.0;
  
  double longitude = 0;
  double latitude = 0;
  UTMtoLL(kSTC::WGS84,N0,E0,kSTC::AugerUTMZone, latitude, longitude);

  ct2lst(longitude, julianDay, &lst, refJD);
  
#warning fix the rec level and trigger 
  if (cdas.fRec.fE > 0){
    sd.SetRecLevel(eHasLDF);
  } else 
    sd.SetRecLevel(eHasTriggeredStations);

  sd.SetT4Trigger(cdas.GetT4Level());
  
  int t5Trigger= 0;
  
  if (cdas.Is6T5())
    t5Trigger = 1;
  
  if (cdas.Is5T5Pos())
    t5Trigger = 2;
  
  if (cdas.Is5T5Pos() && cdas.Is6T5())
    t5Trigger = 3;
  
  sd.SetT5Trigger(t5Trigger);
  
  SdRecShower& sdRec = sd.GetSdRecShower();
  
  sdRec.SetEnergy(cdas.fRec.fE*1e18); 
  sdRec.SetEnergyError(cdas.fRec.fdE*1e18); 
  // sdRec.SetLocalSiderealTime(lst);
  if (cdas.fRec.dR() > 0 ){
    sdRec.SetCurvature(1./cdas.fRec.fR, 1./pow(cdas.fRec.fR,2)*cdas.fRec.dR() );
    if (cdas.fRec.fE > 0 )
      sd.SetRecLevel(eHasCurvature);
  }
  
#warning fZCore is barycenter
  TVector3 core(cdas.fRec.fXCore, cdas.fRec.fYCore, cdas.fRec.fZCore); 
  sdRec.SetCoreSiteCS(core);
  
  
  
  TVector3 coreUTM (cdas.fRec.Easting(),cdas.fRec.Northing(),  cdas.fRec.fZCore);
  sdRec.SetCoreUTMCS(coreUTM);
  sdRec.SetCoreNorthingError(cdas.fRec.dNorthing());
  sdRec.SetCoreEastingError(cdas.fRec.dEasting());
#warning check matrix for the logic
  sdRec.SetCoreNorthingEastingCorrelation(cdas.fRec.dXYCoreCorrelation());
  
#warning ask what is T0  
  sdRec.SetCoreTime(cdas.fRec.fT0, 0);
  sdRec.SetGalacticLongitude(cdas.fRec.L()*TMath::Pi()/180.);
  sdRec.SetGalacticLatitude(cdas.fRec.B()*TMath::Pi()/180.);
  sdRec.SetDeclination(cdas.fRec.Dec()*TMath::Pi()/180.);
  sdRec.SetRightAscension(cdas.fRec.Ra()*TMath::Pi()/180.);
  
  TVector3 axisCore(0.,0.,1.);
  axisCore.SetTheta(cdas.fRec.Theta("rad"));
  axisCore.SetPhi(cdas.fRec.Phi("rad"));
  sdRec.SetAxisSiteCS(axisCore);
  
  TVector3 axisLocal(0.,0.,1.);
  axisLocal.SetTheta(cdas.fRec.LocalTheta("rad"));
  axisLocal.SetPhi(cdas.fRec.LocalPhi("rad"));
  sdRec.SetAxisCoreCS(axisLocal);
  sdRec.SetZenithError(cdas.fRec.dTheta("rad"));
  sdRec.SetAzimuthError(cdas.fRec.dPhi("rad"));
  
  if (cdas.fRec.fRefDist == 450)
    sdRec.GetLDF().SetShowerSizeLabel("S450");
  if (cdas.fRec.fRefDist == 1000)
    sdRec.GetLDF().SetShowerSizeLabel("S1000");
 
  sdRec.GetLDF().SetReferenceDistance(cdas.fRec.fRefDist);
  sdRec.GetLDF().SetShowerSize(cdas.fRec.fSRef, cdas.fRec.dSRef());
  sdRec.GetLDF().SetBeta(cdas.fRec.fBeta, cdas.fRec.fFitBeta?1:0);
  sdRec.GetLDF().SetGamma(cdas.fRec.fGamma, 0);

  
}

void CDAS2ADST::FillStationsInfo(SDEvent& theSd, TBits& stationBits, 
                                 const vector<TCalibStation*>& stations,
                                 const EStationStatus stat, const TErEvent& cdas){
  

  
  for (vector<TCalibStation*>::const_iterator stIter= stations.begin();
       stIter!= stations.end(); ++stIter){
    const TCalibStation& station = *(*stIter);
    const int id = station.id();
    stationBits.SetBitNumber(id,true);
    
    const DetectorGeometry::StationPosMap& adstStations
      = fDetectorGeometry.GetStations();
    
    if (adstStations.find(id) == adstStations.end()){
      const double x = station.fX;
      const double y = station.fY;
      const double z = station.fZ;
      fDetectorGeometry.SetStation(id, x, y, z, 1);
      fDetectorGeometry.SetStationName(id, station.name());
    }
              
    if (!station.fadc()){
      continue;
    }
    
    SdRecStation recStation;  
    recStation.SetId(id);
    
    if (stat != eSilent){
      
      recStation.SetTotalSignal(station.fSigInVEM, sqrt(station.fSigInVEM));
      
      if (station.IsTOT())
        recStation.SetToT();
      else 
        if (station.trigger().IsT1Threshold())
          recStation.SetT1Threshold();
        else 
          if (station.trigger().IsT2Threshold())
            recStation.SetT2Threshold();
      
      if (station.IsLowGainSaturated())
        recStation.SetLowGainSaturated();
      if (station.IsHighGainSaturated())
        recStation.SetHighGainSaturated();
      
      recStation.SetTime(station.Gps->Second, station.Gps->NanoSecond); 
      
      if (stat == eCandidate){
        recStation.SetCandidate();
        recStation.SetSPDistance(cdas.Dist(station), 10.);
        recStation.SetAzimuthSP(acos(cdas.CosZeta(station)));
      
        if (cdas.fRec.fR)
          recStation.SetCurvatureTimeResidual(cdas.Dt(station));
        else
          recStation.SetPlaneTimeResidual(cdas.Dt(station));
        recStation.SetRiseTime(station.T50()-station.T10(), 0);
        recStation.SetFallTime(station.T90()-station.T50(), 0);
        const double startIntegrationTime = station.fStartIntegration*kEc::TimeSlot;
        recStation.SetTime50(station.T50()-startIntegrationTime, 0);
        recStation.SetLDFResidual(0.5);
        
      } else
        recStation.SetAccidental();
      
      
    
      for (unsigned int i = 0; i<3; ++i){
        //      cout << "Number of tubes: " << station.calib()->ntubesok() << endl;
        Traces recTraces;
        recTraces.SetPMTId(i+1);
        recTraces.SetCharge(station.fPmt[i].fVemCharge, 0);
        recTraces.SetPeak(station.fPmt[i].fVemPeak,0);
        recTraces.SetDynodeAnodeRatio(station.fPmt[i].fDA);
        recStation.SetSignalStartAndEnd(station.fStartIntegration, station.fEndIntegration);
        if (cdas.fRec.fE*1.e18 >= fMinEnergyToStoreTraces){
          vector<float> vemTrace;
          vector<short unsigned int> lowGain;
          vector<short unsigned int> highGain;
          
          for (int k = 0; k<kIoSd::MAXSAMPLE; ++k){
            if (fStoreVEMTraces)
              vemTrace.push_back(station.fPmt[i].fTrace[k]);
            if (fStoreFADCTraces){
              highGain.push_back(station.fadc()->Trace[i][1][k]);
              lowGain.push_back(station.fadc()->Trace[i][0][k]);
            }
          }
          
          recTraces.SetVEMComponent(vemTrace);
          recTraces.SetLowGainComponent(lowGain);
          recTraces.SetHighGainComponent(highGain);
        }
        recStation.AddPMTTraces(recTraces);
        recTraces.Clear();
      }
    
    
      theSd.AddStation(recStation);
    }
  }

}


void CDAS2ADST::FillFileInfo(){

  fFileInfo.SetLDFType(ePL);
  char host[80];
  gethostname(host, 80); //defined in unistd.h

  fFileInfo.SetOfflineVersion("HERALD"); // from Offline utl/config.h
  fFileInfo.SetHost(host);
  fFileInfo.SetUser(getenv("USER") ? getenv("USER") : "unknown");

  ostringstream RecEventVersion;
  RecEventVersion << RecEvent::GetReleaseName() << " " << RecEvent::GetVersion()
                  << " (SVN: " << RecEvent::GetSVNVersion() << " rev "
                  << RecEvent::GetSVNRevision() << ")";
 
  fFileInfo.SetRecEventVersion(RecEventVersion.str());
  
}

void CDAS2ADST::FillRecInfoADSTEvent(RecEvent& adst, const SdRecLpnhe& reco){

  SdRecShower& sdRec = adst.GetSDEvent().GetSdRecShower();
  #warning is there any way to get the ndof for the time fit?
  sdRec.SetAngleChi2(reco.angularChi2, 0);
  sdRec.GetLDF().SetLDFChi2(reco.globalChi2, reco.ndof);
  sdRec.GetLDF().SetBeta(reco.fBeta, reco.fdBeta);
  sdRec.GetLDF().SetGamma(reco.fGamma, 0.);
  

}
