#include <math.h>
#include <Er.h>
#include <IoAuger.h>
#include <TCdasUtil.h>
#include <SdRecLpnhe.h>
#include <ToolsLib.h>
#include <TTree.h>

#include <CDAS2ADST.h>
#include <RecEvent.h>
#include <RecEventFile.h>
#include <DetectorGeometry.h>

using namespace std;
string gOutputFileName = "ADST.root";
bool gUseInfillStations = false;
bool gUseInfillReconstruction = false;


void Usage(){
  
  cout << " usage: cdas2adstAllInfo [<options>] <fileNames> \n"
       << endl;
  cout << " Options: -o: name of the output file         (default: ADST.root) \n"
       << "          -i: use infill stations (0/1)       (default: 0) \n"
       << "          -r: reconstruction (0 -standard/ 1- infilll)  (default: 0) \n"
       << endl;
}


/*========================================================

   reads command line options

==========================================================*/
int getOptions(int argc, char** argv){
  
  int c;
  while ((c = getopt (argc, argv, "c:r:o:i:")) != -1) {
    switch (c) {
    case 'o': 
      gOutputFileName = string(optarg);
      cout << "            Output file name: " << gOutputFileName << endl;
      break;
    case 'i':
      gUseInfillStations = atoi(optarg);
      cout << "       Using infill stations: " << gUseInfillStations << endl;
      break;
    case 'r':
      gUseInfillReconstruction = atoi(optarg);
      cout << " Using infill reconstruction: " << gUseInfillReconstruction <<  endl;
      break;
    case 'h': 
      return -1;
    default:
      return -1;
    }
  }
  return optind;
}






int main (int argc, char *argv[]) {
  
  int nOptions = getOptions(argc,argv);
  if ( nOptions < 0 || argc-nOptions < 1 ) {
    Usage();
    return 1;
  }

  TTree::SetMaxTreeSize(200000000000ll);

  
  bool useInfillStations = gUseInfillStations; 
  const bool useInfillRec = gUseInfillReconstruction; 
  if (useInfillRec && (!useInfillStations)) {
    cout << "Changing the option <useInfillStations>  as you requested infill reconstruction" << endl;
    useInfillStations = 1; 
  }

  double coslim = cos( 15.*kCdasUtil::DEG2RAD );
  vector<string> dataFileNames(argc-nOptions);
  std::copy(argv + nOptions, argv + argc, dataFileNames.begin());
  AugerIoSd input (dataFileNames);
  
  cdas::CDAS2ADST transformer;
  transformer.SetMinEnergyToStoreTraces(0);
  transformer.SetStoreVEMTraces(true);
  transformer.SetStoreFADCTraces(true);

  RecEventFile* outputFile=  new RecEventFile(gOutputFileName, 
                                              RecEventFile::eWrite);
  

  RecEvent* recEvent= new RecEvent();
  outputFile->SetBranchStatus("event.fFDEvents", 0);
  outputFile->SetBranchStatus("event.fGenShower.*", 0);
  outputFile->SetBuffers(&recEvent);
  
  
  const int numberOfEvents = input.LastEvent();
  bool firstIteration = true;
  
  for ( EventPos pos = input.FirstEvent(); 
        pos < input.LastEvent(); pos = input.NextEvent()) {
    
    if (!(pos%100))
      cout << pos << "/" << numberOfEvents << endl;
    TErEvent event(pos);
    
    if (firstIteration && useInfillStations){
      kEs::UseInfillStations(useInfillStations);
      firstIteration = false;
    }
    //clean up the recEvent
    *recEvent= RecEvent();
    

    
    event.UseInfillRec = useInfillRec;
    
    //  if (!event.IsT4()) 
    //  continue;
    
    event.OfficialSelection();
    
    theLpnheRec().SetDefault();
    theLpnheRec().SetVerbosity(0);
    theLpnheRec().SetRec(eGeomOptimal, eLDFOptimal);
    
    const int ret = event.Reconstruct( &theLpnheRec() );
    if (ret){
      
      // these have to be called to set the flags
      event.IsT5();
      event.IsICRC2005Posterior(event.fRec.fXCore, event.fRec.fYCore);
      
      // Compute rec/estimate difference
      double dcos=event.fEstimated.fU*event.fRec.fU+event.fEstimated.fV*event.fRec.fV
        +sqrt(1-event.fEstimated.fU*event.fEstimated.fU-event.fEstimated.fV*event.fEstimated.fV)
        *sqrt(1-event.fRec.fU*event.fRec.fU-event.fRec.fV*event.fRec.fV);
    
      double dgrnd=sqrt(pow(event.fEstimated.fXCore-event.fRec.fXCore,2)
                        +pow(event.fEstimated.fYCore-event.fRec.fYCore,2));
      int ok = (int) ! ((dcos<coslim) || (dgrnd>800) || (dgrnd>600 && event.fSelectedStations.size()==4) 
                        || (dgrnd>400 && event.fSelectedStations.size()==3));
    
    
      if ( (! (event.fRec.Theta()>=0 && event.fRec.Theta()<=M_PI) ) 
           || (!(event.fRec.fE>0)) || (!(event.fRec.fSRef>=0))){
        cout << "Bad event: " << event.AugerId  << endl; 
        //  continue;
      }
      
      double energy = compute_energy("NEWFD", event.fRec.fSRef, event.fRec.Theta(), 1,
                                     event.fRec.dSRef(), &event.fRec.fdE);
      event.fRec.fE = energy;
    }
    

    event.SortStationsBySignal();
    transformer.FillADSTEvent(*recEvent, event);
    transformer.FillRecInfoADSTEvent(*recEvent, theLpnheRec());
    outputFile->WriteEvent();

  }

  transformer.FillFileInfo();
  const DetectorGeometry& detectorGeometry = transformer.GetDetectorGeometry();
  outputFile->WriteDetectorGeometry(detectorGeometry);
  outputFile->WriteFileInfo(transformer.GetFileInfo());
  outputFile->Close();
  delete outputFile;
  delete recEvent;
}
