#include <math.h>
#include <Er.h>
#include <IoAuger.h>
#include <TCdasUtil.h>
#include <SdRecLpnhe.h>
#include <ToolsLib.h>
#include <TTree.h>

#include <CDAS2ADST.h>
#include <RecEvent.h>
#include <RecEventFile.h>
#include <DetectorGeometry.h>

using namespace std;
string gOutputFileName = "ADST.root";
bool gUseInfillStations = false;
bool gUseInfillReconstruction = false;
bool gSaveNotRecEvents = false;
string gOutputFileBadEvents = "badEvents.root";
bool gVerbose = false; 

void Usage(){
  
  cout << " usage: cdas2adstExample [<options>] <fileNames> \n"
       << endl;
  cout << " Options: -o: name of the output file         (default: ADST.root) \n"
       << "          -i: use infill stations (0/1)       (default: 0) \n"
       << "          -r: reconstruction (0 -standard/ 1- infilll)  (default: 0) \n"
       << "          -s: name of the file with events not reconstructed \n "<<  endl;
}


/*========================================================

   reads command line options

==========================================================*/
int getOptions(int argc, char** argv){
  
  int c;
  while ((c = getopt (argc, argv, "c:r:o:i:s:v:")) != -1) {
    switch (c) {
    case 'o': 
      gOutputFileName = string(optarg);
      cout << "            Output file name: " << gOutputFileName << endl;
      break;
    case 'i':
      gUseInfillStations = atoi(optarg);
      cout << "       Using infill stations: " << gUseInfillStations << endl;
      break;
    case 'r':
      gUseInfillReconstruction = atoi(optarg);
      cout << " Using infill reconstruction: " 
	   << gUseInfillReconstruction <<  endl;
      break;
    case 's':
      gSaveNotRecEvents = true;
      gOutputFileBadEvents = string(optarg);
      cout << "   Saving bad events in file: " 
	   << gOutputFileBadEvents << endl;
      break;
    case 'v':
      gVerbose = true; 
      break;
    case 'h': 
      return -1;
    default:
      return -1;
    }
  }
  return optind;
}






int main (int argc, char *argv[]) {
  
  int nOptions = getOptions(argc,argv);

  if ( nOptions < 0 || argc-nOptions < 1 ) {
    Usage();
    return 1;
  }

  TTree::SetMaxTreeSize(200000000000ll);
  
  bool useInfillStations = gUseInfillStations; 
  const bool useInfillRec = gUseInfillReconstruction; 
  if (useInfillRec && (!useInfillStations)) {
    cout << " Changing the option <useInfillStations>  as you requested " 
      " infill reconstruction" << endl;
    useInfillStations = 1; 
  }

  
  double coslim = cos( 20.*kCdasUtil::DEG2RAD );
  vector<string> dataFileNames(argc-nOptions);
  std::copy(argv + nOptions, argv + argc, dataFileNames.begin());
  AugerIoSd input(dataFileNames);
  
  cdas::CDAS2ADST transformer;
  transformer.SetMinEnergyToStoreTraces(0);
  transformer.SetStoreVEMTraces(true);
  transformer.SetStoreFADCTraces(true);

  RecEventFile* outputFile=  new RecEventFile(gOutputFileName, 
                                              RecEventFile::eWrite);
  

  RecEvent* recEvent= new RecEvent();
  outputFile->SetBuffers(&recEvent);
  outputFile->SetBranchStatus("event.fFDEvents", 0);
  outputFile->SetBranchStatus("event.fGenShower.*", 0);
  
  
  const int numberOfEvents = input.LastEvent();
  bool firstIteration = true;
  
  
  AugerFile *ioAdFile = NULL; 
  

  if (gSaveNotRecEvents){
    ioAdFile = new AugerFile(gOutputFileBadEvents.c_str(),
			     AugerFile::eWrite);
  }
  
  for ( EventPos pos = input.FirstEvent(); 
        pos < input.LastEvent(); pos = input.NextEvent()) {
    
    if (!(pos%100))
      cout << pos << "/" << numberOfEvents << endl;
    
    TErEvent event(pos);
    
    if (firstIteration && useInfillStations){
      kEs::UseInfillStations(useInfillStations);
      firstIteration = false;
    }
    //clean up the recEvent
    *recEvent= RecEvent();
    
    event.UseInfillRec = useInfillRec;
   
    if (!event.IsT4()) 
      continue;
    
    event.OfficialSelection();
    
    theLpnheRec().SetDefault();
    theLpnheRec().SetVerbosity(0);
    theLpnheRec().SetRec(eGeomOptimal, eLDFOptimal);
    
    int ret = event.Reconstruct( &theLpnheRec() );
#warning ask about the continueeeesss    
    if (!ret){
      if (gVerbose){
	cout << "Not reconstructed event: " << event.AugerId  << endl; 
	cout << endl << "Event not rec: " <<  event.Id << endl
	     << "======================" << endl;
      }
      continue;
    }
    
    // these have to be called to set the flags
    event.IsT5();
    
    event.IsICRC2005Posterior(event.fRec.fXCore, event.fRec.fYCore);
    
    // Compute rec/estimate difference
    const double dcos=event.fEstimated.fU*event.fRec.fU
      +event.fEstimated.fV*event.fRec.fV
      +sqrt(1-event.fEstimated.fU*event.fEstimated.fU
	    -event.fEstimated.fV*event.fEstimated.fV)
      *sqrt(1-event.fRec.fU*event.fRec.fU-event.fRec.fV*event.fRec.fV);
    
    const double dgrnd=sqrt(pow(event.fEstimated.fXCore-event.fRec.fXCore,2)
			    +pow(event.fEstimated.fYCore-event.fRec.fYCore,2));
    const int ok = (int) ! ((dcos<coslim) || (dgrnd>800) || (dgrnd>600 && event.fSelectedStations.size()==4) 
			    || (dgrnd>400 && event.fSelectedStations.size()==3));
    
    if (gVerbose && !ok){
      cout << endl << "Event: " <<  event.Id 
	   << "======================" << endl;
      if (dcos < coslim)
	cout << "Event axis changed too much: " << dcos << " > " 
	     << coslim << endl; 
      if (dgrnd > 800)
	cout << " Core location changed by more than 800 m: " 
	     << dgrnd << endl;  
      if ( dgrnd > 600 && event.fSelectedStations.size() == 4)
	cout << " Event has 4 stations and core changed more than 600 m " 
	     << dgrnd << endl; 
      if ( dgrnd > 400 && event.fSelectedStations.size() == 3)
	cout << " Event has 3 stations and core changed more than 400 m " 
	     << dgrnd << endl; 
    }

    
    
    if ( (! (event.fRec.Theta()>=0 && event.fRec.Theta()<=M_PI) ) 
         || (!(event.fRec.fE>0)) || (!(event.fRec.fSRef>=0))){
      
      if (gVerbose){
	cout << endl << "Event: " <<  event.Id << endl
	     << "======================" << endl;
	cout << " E =  " << event.fRec.fE  << " eV " 
	     << " Sref = " << event.fRec.fSRef << " [VEM] " 
	     << " Theta = " << event.fRec.Theta()*180./M_PI << " [degree] " 
	     << endl;
      }
      ret = 0;
      
    }
    
    if (!ok)
      ret = 0;
    
    const double energy = compute_energy(useInfillRec?"CICINFILL":"NEWFD",
					 event.fRec.fSRef, 
					 event.fRec.Theta(), 1,
					 event.fRec.dSRef(), &event.fRec.fdE);
    event.fRec.fE = (!ret)?0:energy;
    
    event.SortStationsBySignal();
    transformer.FillADSTEvent(*recEvent, event);
    transformer.FillRecInfoADSTEvent(*recEvent, theLpnheRec());
    outputFile->WriteEvent();
    
    if (gSaveNotRecEvents && !ret && event.fRec.Theta()*180./3.14 < 60)
      ioAdFile->Write(*(event.RawAugerEvent()), "VERBOSE");
    
  }

  transformer.FillFileInfo();
  const DetectorGeometry& detectorGeometry = transformer.GetDetectorGeometry();
  outputFile->WriteDetectorGeometry(detectorGeometry);
  outputFile->WriteFileInfo(transformer.GetFileInfo());
  outputFile->Close();
  delete outputFile;
  delete recEvent;

  if (ioAdFile){
    ioAdFile->Close(); 
    delete ioAdFile;
  }
}
