################################################################################
# Module to find FDEventLib                                                    #
#                                                                              #
# The following variables are defined:                                         #
#   - FDEVENTLIB_FOUND                                                         #
#   - FDEVENTLIB_HOME                                                          #
#   - FDEVENTLIB_LIBRARIES                                                     #
#   - FDEVENTLIB_INCLUDE_DIR                                                   #
#   - FDEVENTLIB_LIB_DIR                                                       #
#   - FDEVENTLIB_CPPFLAGS                                                      #
#   - FDEVENTLIB_LDFLAGS                                                       #
#   - FDEVENTLIB_DEFINITIONS                                                   #
#   - FDLIB_V3R3 true if version >= 3.3                                        #
#   - FDLIB_V3R4 true if version >= 3.4                                        #
################################################################################

FIND_PACKAGE(Root QUIET)
IF(NOT Root_FOUND)
  MESSAGE(STATUS "Root is required to use FDEventLib")
ENDIF(NOT Root_FOUND)

# Search directories for FDEventLib
SET (_DIRECTORIES
  $ENV{FD_ROOT}
  $ENV{AUGER_BASE}/FDEventLib
  $ENV{AUGER_BASE}/FDEventLib/pro
  $ENV{AUGER_BASE}/FDEventLib/ROOT-${ROOT_VERSION}/pro
  /usr/local/FDEventLib
)

INCLUDE (CheckCXXSourceCompiles)
INCLUDE (AugerCheckCxxSourceRuns)
INCLUDE (CMakeCompareVersionStrings)

MACRO(CHECK_FDEVENTLIB_VERSION _RESULT _VERSION_MAJOR _VERSION_MINOR _VERSION_PATCH)
  CHECK_CXX_SOURCE_COMPILES ("
    #include<FDEventLibVersion.hh>
    int main() {
    #if (! defined FDEVENTLIB_VERSION_CODE) || (FDEVENTLIB_VERSION_CODE<ModuleVersionCode(${_VERSION_MAJOR},${_VERSION_MINOR},${_VERSION_PATCH}))
    #error
    #endif
    return 0;}"
    ${_RESULT}
  )
ENDMACRO(CHECK_FDEVENTLIB_VERSION)

MACRO(GET_FDEVENTLIB_VERSION)
  SET (CMAKE_REQUIRED_INCLUDES
    ${FDEVENTLIB_INCLUDE_DIR}
    ${ROOT_INCLUDE_DIR})
  SET (CMAKE_REQUIRED_LIBRARIES
    ${FDEVENTLIB_LIBRARIES}
    ${ROOT_LIBRARIES})
  SET (CMAKE_REQUIRED_DEFINITIONS
    ${FDEVENTLIB_DEFINITIONS})

  IF(NOT CMAKE_CROSSCOMPILING)
    AUGER_CHECK_CXX_SOURCE_RUNS (
      FDEventLib_VERSION
      "#include<FDEventLibVersion.hh>
      #include<iostream>
      int main() {
        std::cout << kFDEventLibVersion << \".\" << kFDEventLibSubVersion << \".\" << kFDEventLibSubSubVersion << std::endl;
        return 0;
      }"
      FD_VERSION
    )
    IF(FD_VERSION)
      STRING(REGEX REPLACE "\n" "" FDEventLib_VERSION "${FDEventLib_VERSION}")
      IF (NOT FDEventLib_FIND_QUIETLY)
        MESSAGE(STATUS "FDEventLib version: '${FDEventLib_VERSION}'")
      ENDIF (NOT FDEventLib_FIND_QUIETLY)
      SET(FDEventLib_VERSION ${FDEventLib_VERSION} CACHE STRING "FDEventLib version")
    ENDIF(FD_VERSION)
  ENDIF(NOT CMAKE_CROSSCOMPILING)

  CHECK_FDEVENTLIB_VERSION(FDLIB_V4R0 4 0 0)
  MARK_AS_ADVANCED (FDLIB_V4R0)
  CHECK_FDEVENTLIB_VERSION(FDLIB_V3R6 3 6 0)
  MARK_AS_ADVANCED (FDLIB_V3R6)
  CHECK_FDEVENTLIB_VERSION(FDLIB_V3R5 3 5 0)
  MARK_AS_ADVANCED (FDLIB_V3R5)
  CHECK_FDEVENTLIB_VERSION(FDLIB_V3R4 3 4 0)
  MARK_AS_ADVANCED (FDLIB_V3R4)
  CHECK_FDEVENTLIB_VERSION(FDLIB_V3R3 3 3 0)
  MARK_AS_ADVANCED (FDLIB_V3R3)
ENDMACRO(GET_FDEVENTLIB_VERSION)





# First try to use the new config script

FIND_PATH (_FDEVENTLIB_BIN_DIR FDEventLib-config
  PATHS $ENV{FD_ROOT}/bin
  NO_DEFAULT_PATH
)
FIND_PATH (_FDEVENTLIB_BIN_DIR FDEventLib-config)

IF (_FDEVENTLIB_BIN_DIR)

  SET (FDEVENTLIB_BIN_DIR ${_FDEVENTLIB_BIN_DIR})

  # FDEVENTLIB_VERSION
  SET (_CONFIG_SCRIPT "${FDEVENTLIB_BIN_DIR}/FDEventLib-config")
  EXECUTE_PROCESS (COMMAND ${_CONFIG_SCRIPT} --version
    OUTPUT_VARIABLE FDEVENTLIB_VERSION
    ERROR_VARIABLE _ERROR
  )
  STRING (REGEX REPLACE "[ \t\r\n]+" "" FDEVENTLIB_VERSION "${FDEVENTLIB_VERSION}")

  IF (NOT FDEVENTLIB_VERSION STREQUAL "")
    SET (FDEVENTLIB_FOUND TRUE)
  ELSE (NOT FDEVENTLIB_VERSION STREQUAL "")
    MESSAGE (STATUS "Error setting FDEventLib version: ${_CONFIG_SCRIPT} --version")
    IF (NOT _ERROR STREQUAL "")
      MESSAGE (STATUS "Error setting FDEventLib version: ${_ERROR}")
    ENDIF (NOT _ERROR STREQUAL "")
  ENDIF (NOT FDEVENTLIB_VERSION STREQUAL "")

  # FDEVENTLIB_ROOT
  EXECUTE_PROCESS (COMMAND ${_CONFIG_SCRIPT} --prefix
    OUTPUT_VARIABLE FDEVENTLIB_HOME
    ERROR_VARIABLE _ERROR
    )
  STRING (REGEX REPLACE "[ \t\r\n]+" "" FDEVENTLIB_HOME "${FDEVENTLIB_HOME}")

  # FDEVENTLIB_LIB_DIR
  EXECUTE_PROCESS (COMMAND ${_CONFIG_SCRIPT} --libdir
    OUTPUT_VARIABLE FDEVENTLIB_LIB_DIR
    ERROR_VARIABLE _ERROR
    )
  STRING (REGEX REPLACE "[ \t\r\n]+" "" FDEVENTLIB_LIB_DIR "${FDEVENTLIB_LIB_DIR}")

  # FDEVENTLIB_INCLUDE_DIR
  EXECUTE_PROCESS (COMMAND ${_CONFIG_SCRIPT} --incdir
    OUTPUT_VARIABLE FDEVENTLIB_INCLUDE_DIR
    ERROR_VARIABLE _ERROR
    )
  STRING (REGEX REPLACE "[ \t\r\n]+" "" FDEVENTLIB_INCLUDE_DIR "${FDEVENTLIB_INCLUDE_DIR}")

  # FDEVENTLIB_LIBRARIES
  EXECUTE_PROCESS (COMMAND ${_CONFIG_SCRIPT} --libnames
    OUTPUT_VARIABLE _FDEVENTLIB_LIBS
    ERROR_VARIABLE _ERROR
    )
  STRING (REGEX REPLACE "[ \t\r\n]+" " " _FDEVENTLIB_LIBS "${_FDEVENTLIB_LIBS}")
  SEPARATE_ARGUMENTS (_FDEVENTLIB_LIBS)
  FOREACH (_file ${_FDEVENTLIB_LIBS})
    FIND_LIBRARY (_FDLIB_${_file} ${_file} ${FDEVENTLIB_LIB_DIR} NO_DEFAULT_PATH)
    IF (_FDLIB_${_file})
      LIST (APPEND FDEVENTLIB_LIBRARIES ${_FDLIB_${_file}})
    ELSE (_FDLIB_${_file})
      IF (NOT FDEventLib_FIND_QUIETLY)
        MESSAGE (STATUS "Warning! FDEventLib library 'lib${_file}' not found in ${FDEVENTLIB_LIB_DIR}")
      ENDIF (NOT FDEventLib_FIND_QUIETLY)
      SET (FDEVENTLIB_FOUND FALSE)
    ENDIF (_FDLIB_${_file})
    SET (_FDLIB_${_file} ${_FDLIB_${_file}} CACHE INTERNAL "" FORCE)
  ENDFOREACH (_file ${_FDEVENTLIB_LIBS})

  # FDEVENTLIB_CPPFLAGS
  EXECUTE_PROCESS (COMMAND ${_CONFIG_SCRIPT} --cflags
    OUTPUT_VARIABLE _FDEVENTLIB_CPPFLAGS
    ERROR_VARIABLE _ERROR
    )
  STRING (REGEX REPLACE "[ \t\r\n]+" " " FDEVENTLIB_CPPFLAGS "${_FDEVENTLIB_CPPFLAGS}")

  # FDEVENTLIB_LDFLAGS
  EXECUTE_PROCESS (COMMAND ${_CONFIG_SCRIPT} --libs
    OUTPUT_VARIABLE _FDEVENTLIB_LDFLAGS
    ERROR_VARIABLE _ERROR
    )
  STRING (REGEX REPLACE "[ \t\r\n]+" " " FDEVENTLIB_LDFLAGS "${_FDEVENTLIB_LDFLAGS}")

  # FDEVENTLIB_DEFINITIONS
  SET (FDEVENTLIB_DEFINITIONS -DDPA)

  # some flags that seem to be used (trivial since the config script is added later):
  SET(FDLIB_V2R6 1)
  SET(FDLIB_V3R3 1)
  SET(FDLIB_V3R4 1)
  SET(FDLIB_V3R5 1)
  SET(FDLIB_V4R0 1)

ELSE (_FDEVENTLIB_BIN_DIR)
#try to determine FDEVENTLIB_HOME by finding the header file the rest goes from there
FIND_PATH (FDEVENTLIB_HOME include/FdConfig.hh PATHS ${_DIRECTORIES} NO_DEFAULT_PATH)

SET (FDEVENTLIB_INCLUDE_DIR
  ${FDEVENTLIB_HOME}/include)
SET (FDEVENTLIB_LIB_DIR
  ${FDEVENTLIB_HOME}/lib)

FIND_LIBRARY (FDEventLib_Utillib_LIBRARY
  Utillib
  ${FDEVENTLIB_LIB_DIR}
  NO_DEFAULT_PATH)
FIND_LIBRARY (FDEventLib_RootUtillib_LIBRARY
  RootUtillib
  ${FDEVENTLIB_LIB_DIR}
  NO_DEFAULT_PATH)
FIND_LIBRARY (FDEventLib_MiEvent_LIBRARY
  MiEvent
  ${FDEVENTLIB_LIB_DIR}
  NO_DEFAULT_PATH)
FIND_LIBRARY (FDEventLib_EyeEvent_LIBRARY
  EyeEvent
  ${FDEVENTLIB_LIB_DIR}
  NO_DEFAULT_PATH)

SET (FDEVENTLIB_DEFINITIONS -DDPA)
SET (FDEVENTLIB_LIBRARIES
  ${FDEventLib_Utillib_LIBRARY}
  ${FDEventLib_RootUtillib_LIBRARY}
  ${FDEventLib_MiEvent_LIBRARY}
  ${FDEventLib_EyeEvent_LIBRARY}
  CACHE STRING "All FDEventLib libraries")
MARK_AS_ADVANCED (FDEVENTLIB_LIBRARIES)

IF (FDEventLib_Utillib_LIBRARY AND FDEVENTLIB_HOME AND Root_FOUND)

  SET (FDEVENTLIB_BIN_DIR "${FDEVENTLIB_HOME}/bin")
  IF (NOT FDEventLib_FIND_QUIETLY)
    MESSAGE (STATUS "FDEventLib ${FDEventLib_Utillib_LIBRARY}")
  ENDIF (NOT FDEventLib_FIND_QUIETLY)

  # Basic libraries for all versions of FDEventLib.
  SET (FDEVENTLIB_LDFLAGS
    "${FDEventLib_Utillib_LIBRARY} ${FDEventLib_RootUtillib_LIBRARY} ${FDEventLib_MiEvent_LIBRARY} ${FDEventLib_EyeEvent_LIBRARY}")

  GET_FDEVENTLIB_VERSION()

  # V3.3: add the Tlt, MiReadout, and EyeT3 libraries.
  IF (FDLIB_V3R3)
    FIND_LIBRARY (FDEventLib_Tlt_LIBRARY
      Tlt
      ${FDEVENTLIB_LIB_DIR}
      NO_DEFAULT_PATH)
    FIND_LIBRARY (FDEventLib_MiReadout_LIBRARY
      MiReadout
      ${FDEVENTLIB_LIB_DIR}
      NO_DEFAULT_PATH)
    FIND_LIBRARY (FDEventLib_EyeT3_LIBRARY
      EyeT3
      ${FDEVENTLIB_LIB_DIR}
      NO_DEFAULT_PATH)

    IF(NOT FDEventLib_Tlt_LIBRARY AND NOT FDEventLib_FIND_QUIETLY)
      MESSAGE(STATUS "FDEventLib's TLT library not found!")
    ENDIF(NOT FDEventLib_Tlt_LIBRARY AND NOT FDEventLib_FIND_QUIETLY)
    IF(NOT FDEventLib_MiReadout_LIBRARY AND NOT FDEventLib_FIND_QUIETLY)
      MESSAGE(STATUS "FDEventLib's MiReadout library not found!")
    ENDIF(NOT FDEventLib_MiReadout_LIBRARY AND NOT FDEventLib_FIND_QUIETLY)
    IF(NOT FDEventLib_EyeT3_LIBRARY AND NOT FDEventLib_FIND_QUIETLY)
      MESSAGE(STATUS "FDEventLib's EyeT3 library not found!")
    ENDIF(NOT FDEventLib_EyeT3_LIBRARY AND NOT FDEventLib_FIND_QUIETLY)

    LIST (APPEND FDEVENTLIB_LIBRARIES
      ${FDEventLib_Tlt_LIBRARY}
      ${FDEventLib_MiReadout_LIBRARY}
      ${FDEventLib_EyeT3_LIBRARY})

    SET (FDEVENTLIB_LDFLAGS
      "${FDEVENTLIB_LDFLAGS} ${FDEventLib_Tlt_LIBRARY} ${FDEventLib_MiReadout_LIBRARY} ${FDEventLib_EyeT3_LIBRARY}")

    LIST (APPEND FDEVENTLIB_DEFINITIONS -DUSE_FEMIEVENT)
  ENDIF (FDLIB_V3R3)

  # V3.4: add the MiT3 library (removed in V3R5).
  IF (FDLIB_V3R4 AND NOT FDLIB_V3R5)
    FIND_LIBRARY (FDEventLib_MiT3_LIBRARY
      MiT3
      ${FDEVENTLIB_LIB_DIR}
      NO_DEFAULT_PATH)

    # Additional check that MiT3 is actually available...
    IF (FDEventLib_MiT3_LIBRARY)
      LIST (APPEND FDEVENTLIB_LIBRARIES
        ${FDEventLib_MiT3_LIBRARY})
      SET (FDEVENTLIB_LDFLAGS
        "${FDEVENTLIB_LDFLAGS} ${FDEventLib_MiT3_LIBRARY}")
    ENDIF (FDEventLib_MiT3_LIBRARY)
    IF(NOT FDEventLib_EyeT3_LIBRARY AND NOT FDEventLib_FIND_QUIETLY)
      MESSAGE(STATUS "FDEventLib's MiT3 library not found!")
    ENDIF(NOT FDEventLib_EyeT3_LIBRARY AND NOT FDEventLib_FIND_QUIETLY)
  ENDIF (FDLIB_V3R4 AND NOT FDLIB_V3R5)


  STRING (REGEX REPLACE ";" " " FDEVENTLIB_CPPFLAGS "-I${FDEVENTLIB_INCLUDE_DIR} ${FDEVENTLIB_DEFINITIONS}")

  IF(FDEVENTLIB_LIBRARIES MATCHES "LIBRARY-NOTFOUND")
    SET (FDEVENTLIB_FOUND FALSE)
    SET (HAVE_FDEVENTLIB 0)
  ELSE(FDEVENTLIB_LIBRARIES MATCHES "LIBRARY-NOTFOUND")
    SET (FDEVENTLIB_FOUND TRUE)
    SET (HAVE_FDEVENTLIB 1)
  ENDIF(FDEVENTLIB_LIBRARIES MATCHES "LIBRARY-NOTFOUND")

ENDIF (FDEventLib_Utillib_LIBRARY AND FDEVENTLIB_HOME AND Root_FOUND)
ENDIF (_FDEVENTLIB_BIN_DIR)


IF (NOT FDEVENTLIB_FOUND)
  if (FDEventLib_FIND_REQUIRED)
    MESSAGE (FATAL_ERROR "FDEventLib not found")
  else (FDEventLib_FIND_REQUIRED)
    IF (NOT FDEventLib_FIND_QUIETLY)
      MESSAGE (STATUS "FDEventLib not found")
    ENDIF (NOT FDEventLib_FIND_QUIETLY)
  endif (FDEventLib_FIND_REQUIRED)
ELSE (NOT FDEVENTLIB_FOUND)
  IF (NOT FDEventLib_FIND_QUIETLY)
    MESSAGE (STATUS "FDEventLib found")
    MESSAGE (STATUS "FDEVENTLIB_HOME: ${FDEVENTLIB_HOME}")
    MESSAGE (STATUS "FDEVENTLIB_INCLUDE_DIR: ${FDEVENTLIB_INCLUDE_DIR}")
    MESSAGE (STATUS "FDEVENTLIB_LIB_DIR: ${FDEVENTLIB_LIB_DIR}")
    MESSAGE (STATUS "FDEVENTLIB_LIBRARIES: ${FDEVENTLIB_LIBRARIES}")
    MESSAGE (STATUS "FDEVENTLIB_CPPFLAGS: ${FDEVENTLIB_CPPFLAGS}")
    MESSAGE (STATUS "FDEVENTLIB_LDFLAGS: ${FDEVENTLIB_LDFLAGS}")
    MESSAGE (STATUS "FDEVENTLIB_DEFINITIONS: ${FDEVENTLIB_DEFINITIONS}")
  ENDIF (NOT FDEventLib_FIND_QUIETLY)
ENDIF (NOT FDEVENTLIB_FOUND)

