#include "PBRec.h"
#include "TCdasUtil.h"
#include "ToolsLib.h"
#include "THorReconstruct.h"

double Poisson(double a, int n) {
  // using Stirling
  double val = exp(-a);
  if(n>10) return val * pow(a*M_E/n,n) / sqrt(2*M_PI*n);
  for (int i=1; i<=n; i++) val *= a/i;
  return val;
}

double LDFfunc(double dist, double beta, double gamma) {
  // log-log parabola (not valid close the core)
  if(dist>DIST_MIN_LDF)
    return 1/pow(dist/thePBRec().LDF_DISTREF,beta+gamma*log(dist/thePBRec().LDF_DISTREF));
  // power law at short distance (with smooth connection to the log-log parabola)
  else {
    double fact = 1/pow(DIST_MIN_LDF/thePBRec().LDF_DISTREF,beta+gamma*log(DIST_MIN_LDF/thePBRec().LDF_DISTREF));
    double slope = beta+2*gamma*log(DIST_MIN_LDF/thePBRec().LDF_DISTREF);
    return fact/pow(dist/DIST_MIN_LDF,slope);
  }
}

double PARfunc(double dist, double x) {
  double xx = x/1000;
  double b = -3.9  - .297*xx + 3.99*xx*xx - 2.49*xx*xx*xx;
  double c = -.358 + .180*xx +.0151*xx*xx - .256*xx*xx*xx;

  double logr = log(dist/1000);
  //printf("b %f c %f logr %f\n",b,c,logr);
  return exp(b*logr+c*logr*logr);
}


TPBRec * TPBRec::_instance = 0;

/// returns the TPBRec singleton: ensures that there is only one instance of this object
TPBRec *TPBRec::Instance() {
  if (_instance == 0) _instance = new TPBRec();
  return _instance;
}

TPBRec & thePBRec() {
  return *TPBRec::Instance();
}

/*-------------------------------------------------------------------*/
TPBRec::TPBRec()
/*-------------------------------------------------------------------*/
{
  // easting and northing of Los Leones
  /* EASTING_ORIGIN = kIoSd_FdX(1);
     NORTHING_ORIGIN = kIoSd_FdY(1);
     ALTITUDE_ORIGIN = kIoSd_FdZ(1); */
  EASTING_ORIGIN = kIoSd::E_LL;
  NORTHING_ORIGIN = kIoSd::N_LL;
  ALTITUDE_ORIGIN = kIoSd::A_LL;

  InitializeRec();
  OpenOutputFiles();
  if(FitHoriz) ReadHorizTables();
  n_user_file = 0;

}

/*-------------------------------------------------------------------*/
TPBRec::~TPBRec()
/*-------------------------------------------------------------------*/
{
  CloseOutputFiles();
}

/*-------------------------------------------------------------------*/
void TPBRec::OpenOutputFiles()
/*-------------------------------------------------------------------*/
{
  if(Write1_result && !res1) res1 = fopen("prov1.res","w");
  if(Write2_result && !res2) res2 = fopen("prov2.res","w");
  if(Write3_result && !res3) res3 = fopen("prov3.res","w");
  if(Auxil1_output && !aux1) aux1 = fopen("prov1.aux","w");
  if(Auxil2_output && !aux2) aux2 = fopen("prov2.aux","w");
  if(Auxil3_output && !aux3) aux3 = fopen("prov3.aux","w");
  if(Auxil4_output && !aux4) aux4 = fopen("prov4.aux","w");
  if(Diagno_output && !diag) diag = fopen("prov.diag","w");
  if(Display_output && !disp) disp = fopen("prov.disp","w");
}

/*-------------------------------------------------------------------*/
void TPBRec::CloseOutputFiles()
/*-------------------------------------------------------------------*/
{
  if(Write1_result && res1) fclose(res1);
  if(Write2_result && res2) fclose(res2);
  if(Write3_result && res3) fclose(res3);
  if(Auxil1_output && aux1) fclose(aux1);
  if(Auxil2_output && aux2) fclose(aux2);
  if(Auxil3_output && aux3) fclose(aux3);
  if(Auxil4_output && aux4) fclose(aux4);
  if(Diagno_output && diag) fclose(diag);
  if(Display_output && disp) fclose(disp);
}

/*-------------------------------------------------------------------*/
void TPBRec::InitializeRec()
/*-------------------------------------------------------------------*/
{
  NevMax        = 1000000;
  Simulation          = 0;
  ReadTruth           = 0;
  Input_sel_evts      = 0;
  Write_sel_evts      = 0;
  Process_3st         = 1;
  Process_gen         = 1;
  Reconstruct         = 1;
  FitHoriz            = 1;
  RemoveInfillStation = 1;
  RemoveDoublet       = 1;
  RemoveEA            = 1;
  SelectContained     = 1;
  NMIN_AROUND_HIGHEST_SIGNAL = 5;
  CorrectAsym         = 1;
  PrintLevel          = 0;
  Write1_result       = 0;
  Write2_result       = 0;
  Write3_result       = 0;
  Auxil1_output       = 0;
  Auxil2_output       = 0;
  Auxil3_output       = 0;
  Auxil4_output       = 0;
  Diagno_output       = 0;
  Display_output      = 0;
  ROOT_output         = 0;
  List_event          = 0;
  sprintf(HorizModel,"qgsjet_prot");
  //sprintf(truth_name,"../files/simul/Fnal/Iron_0_90/cumul.txt");
  truth_root          = 0;
  ISOLATION_DISTANCE     = 3200;
  ISOLATION_DISTANCE2    = 3800;
  MAGNETIC_NORTH         = 86.4;
  SELECT_TOLRES          =  80.;
  SELECT_TOLRESM         = 100.;
  SELECT_CURV0           = .00014;
  SELECT_DISTMAX_PROJ_ST = 1300;
  ANGREC_TOLRESC         =  60.;
  TIME_ERR               = 100.;
  TIME_ERR0              =  25.;
  TIME_ERR_FACTD2        = .00005;
  TIME_ERR_CURV          =  50.;
  TIME_START             =    0;
  TIME_START_SELECT      =    0;
  TIME_START_PLAN        =    0;
  TIME_START_CURV        =    0;
  TIME_START_GLOB        =    0;
  LDF_DISTREF            = 1000;
  LDF_SINTHMAX           =  .94;
  LDF_CHI2MAX            =   10;
  LDF_DCHI2MAX           =   10;
  LDF_DISTMAX_BARY_CORE  =  500;
  LDF_DISTMAX_SILENT     = 4000;
  LDF_SIGTHRESH          =    3;
  BARY_POWER             = .333;
  NSTAT_IN_BARY          =    3;
  GLOB_CHI2MAX           =   10;
  GLOB_DCHI2MAX          =   10;
  HORIZ_MUON             =    0;
  HORIZ_SIGTHRESH        =  1.5;
  ANA3_RATMIN            =  1.6;
  ANA3_AREAMAX           = 1200000;
  ANA3_AREAMIN           = 200000;
  ANA3_DISTMAX           = 2800;
  ANA3_CHI2_LDF_MAX      =    5;
  ANA3_CHI2_LDF_SILENT_MAX = 10;
  ANA3_DISTMAX_BARY_CORE =  500;
  list_file = res1 = res2 = res3 = aux1 = aux2
                                          = aux3 = aux4 = diag = disp = user_file = NULL;

}

/*-------------------------------------------------------------------*/
void TPBRec::FitPlan(int corrt, int errt, double& t0, double& u, double& v,
                     double *res_t, double& res, double& chi2, double *cov, int& errflag)/// Fitting a plane on the times of signals in the stations\n
/// (possibly including corrections for altitude and front curvature)
/*-------------------------------------------------------------------*/
{
  int is,ns;
  double wgt,dx,dy,dist2,dt,sumw,sumx,sumy,sumx2,sumy2,sumxy,sumt,sumxt,sumyt;
  double rxx,ryy,rxy,rx,ry,rr,deter,up,vp,sum_res2;

  ns=0;
  sumw=sumx=sumy=sumx2=sumy2=sumxy=sumt=sumxt=sumyt=0;
  for(is=0; is<nActiveStation; is++) if(InFit[is]) {
      dx = X[is] - XBarycenter;
      dy = Y[is] - YBarycenter;
      dt = T[is] + corrt*TCorr[is] - TBarycenter;
      if(errt) {
        dist2 = sqr(dx) + sqr(dy) - sqr(u*dx + v*dy);
        double costh2 = maxval(1-u*u-v*v,0);
        wgt=1./(sqr(TIME_ERR0)+sqr(TIME_ERR_FACTD2*dist2)*costh2);
      } else  wgt=1./sqr(TIME_ERR);

      sumw  += wgt;
      sumx  += wgt*dx;
      sumy  += wgt*dy;
      sumx2 += wgt*sqr(dx);
      sumy2 += wgt*sqr(dy);
      sumxy += wgt*dx*dy;
      sumt  += wgt*dt;
      sumxt += wgt*dt*dx;
      sumyt += wgt*dt*dy;
      ns++;
    }
  rxx = sumw*sumy2-sqr(sumy);
  ryy = sumw*sumx2-sqr(sumx);
  rxy = sumx*sumy-sumw*sumxy;
  rx  = sumxy*sumy-sumx*sumy2;
  ry  = sumxy*sumx-sumy*sumx2;
  rr  = sumx2*sumy2-sqr(sumxy);
  deter = sumx2*rxx + sumxy*rxy + sumx*rx;
  if(deter==0) {
    errflag=1; return;
  }
  up = (rxx*sumxt+rxy*sumyt+rx*sumt)/deter;
  vp = (rxy*sumxt+ryy*sumyt+ry*sumt)/deter;
  t0= (rx *sumxt+ry *sumyt+rr*sumt)/deter;
  cov[0] = rr/deter;
  cov[1] = rx/deter;
  cov[2] = rxx/deter;
  cov[3] = ry/deter;
  cov[4] = rxy/deter;
  cov[5] = ryy/deter;

  chi2 = sum_res2 = 0;
  for(is=0; is<nActiveStation; is++) {
    dt = T[is] + corrt*TCorr[is] - TBarycenter;
    dx = X[is] -XBarycenter ; dy = Y[is] - YBarycenter;
    res_t[is] = dt - (t0 + up*dx + vp*dy);
    if(InFit[is]) {
      sum_res2 += sqr(res_t[is]);
      if(errt) {
        dist2 = sqr(dx) + sqr(dy) - sqr(u*dx + v*dy);
        double costh2 = maxval(1-u*u-v*v,0);
        chi2 += sqr(res_t[is])/(sqr(TIME_ERR0)+sqr(TIME_ERR_FACTD2*dist2)*costh2);
      } else chi2 += sqr(res_t[is]/TIME_ERR);
    }
  }

  // going to direction cosines
  u = -up*kCdasUtil::CSPEED;
  v = -vp*kCdasUtil::CSPEED;
  cov[1] *= (-1.0*kCdasUtil::CSPEED);
  cov[3] *= (-1.0*kCdasUtil::CSPEED);
  cov[2] *= kCdasUtil::CSPEED*kCdasUtil::CSPEED;
  cov[4] *= kCdasUtil::CSPEED*kCdasUtil::CSPEED;
  cov[5] *= kCdasUtil::CSPEED*kCdasUtil::CSPEED;

  fVarCov[TSdParams::eT0][TSdParams::eT0] = cov[0];
  fVarCov[TSdParams::eU][TSdParams::eU] = cov[2];
  fVarCov[TSdParams::eU][TSdParams::eV] = fVarCov[TSdParams::eV][TSdParams::eU] = cov[4];
  fVarCov[TSdParams::eV][TSdParams::eV] = cov[5];

  if(ns>3) res = sqrt(sum_res2/(ns-3)); else res = sqrt(sum_res2);
  errflag = 0;

  if(PrintLevel>2) printf("FitPlan :   t0 %5.0f  u,v %5.3f %5.3f  chi2 %5.2f\n",t0,u,v,chi2);
}

/*-------------------------------------------------------------------*/
void TPBRec::FitCurved(int corrt, int errt, double& t0, double& u, double& v,
                       double& curv, double *res_t, double& res, double& chi2, double *cov, int& errflag)/// Fitting a front with a free curvature parameter
/*-------------------------------------------------------------------*/
{
  int i,is,ns;
  double wgt,dx,dy,dist2,dt,s[10],t[4];
  double t0p,up,vp,curvp,sum_res2;

  ns=0;
  for(i=0; i<10; i++) s[i]=0;
  for(i=0; i<4; i++) t[i]=0;

  for(is=0; is<nActiveStation; is++) if(InFit[is]) {
      ns++;
      dx = X[is] - XBarycenter;
      dy = Y[is] - YBarycenter;
      dt = T[is] + corrt*TCorr[is] - TBarycenter;

      if(errt) {
        dist2 = sqr(dx) + sqr(dy) - sqr(u*dx + v*dy);
        double costh2 = maxval(1-u*u-v*v,0);
        wgt=1./(sqr(TIME_ERR0)+sqr(TIME_ERR_FACTD2*dist2)*costh2);
      } else wgt=1./sqr(TIME_ERR_CURV);
      dist2 = sqr(dx) + sqr(dy) - sqr(u*dx + v*dy);

      s[0] += wgt;
      s[1] += wgt*dx;
      s[2] += wgt*dx*dx;
      s[3] += wgt*dy;
      s[4] += wgt*dy*dx;
      s[5] += wgt*dy*dy;
      s[6] += wgt*dist2;
      s[7] += wgt*dist2*dx;
      s[8] += wgt*dist2*dy;
      s[9] += wgt*dist2*dist2;
      t[0] += wgt*dt;
      t[1] += wgt*dt*dx;
      t[2] += wgt*dt*dy;
      t[3] += wgt*dt*dist2;
    }
  errflag = !matinv4(s,cov);
  if(errflag) return;

  t0p   = cov[0]*t[0]+cov[1]*t[1]+cov[3]*t[2]+cov[6]*t[3];
  up    = cov[1]*t[0]+cov[2]*t[1]+cov[4]*t[2]+cov[7]*t[3];
  vp    = cov[3]*t[0]+cov[4]*t[1]+cov[5]*t[2]+cov[8]*t[3];
  curvp = cov[6]*t[0]+cov[7]*t[1]+cov[8]*t[2]+cov[9]*t[3];

  chi2 = sum_res2 = 0;
  for(is=0; is<nActiveStation; is++) {
    dx = X[is] - XBarycenter;
    dy = Y[is] - YBarycenter;
    dist2 = sqr(dx)+sqr(dy)-sqr(u*dx + v*dy);
    dt = T[is] + corrt*TCorr[is] - TBarycenter;
    res_t[is] = dt - (t0p + up*dx + vp*dy + curvp*dist2);
    if(InFit[is]) {
      sum_res2 += sqr(res_t[is]);
      if(errt) {
        dist2 = sqr(dx) + sqr(dy) - sqr(u*dx + v*dy);
        double costh2 = maxval(1-u*u-v*v,0);
        chi2 += sqr(res_t[is])/(sqr(TIME_ERR0)+sqr(TIME_ERR_FACTD2*dist2)*costh2);
      } else chi2 += sqr(res_t[is]/TIME_ERR_CURV);
    }
  }

  if(ns>4) res = sqrt(sum_res2/(ns-4)); else res = sqrt(sum_res2);

  // going to direction cosines and 1/R
  t0 = t0p;
  u = -up*kCdasUtil::CSPEED;
  v = -vp*kCdasUtil::CSPEED;
  curv = 2*curvp*kCdasUtil::CSPEED;
  cov[1] *= (-1.0*kCdasUtil::CSPEED);
  cov[3] *= (-1.0*kCdasUtil::CSPEED);
  cov[2] *= kCdasUtil::CSPEED*kCdasUtil::CSPEED;
  cov[4] *= kCdasUtil::CSPEED*kCdasUtil::CSPEED;
  cov[5] *= kCdasUtil::CSPEED*kCdasUtil::CSPEED;
  cov[6] *= 2*kCdasUtil::CSPEED;
  cov[7] *= -2*kCdasUtil::CSPEED*kCdasUtil::CSPEED;
  cov[8] *= -2*kCdasUtil::CSPEED*kCdasUtil::CSPEED;
  cov[9] *= 4*kCdasUtil::CSPEED*kCdasUtil::CSPEED;

  fVarCov[TSdParams::eT0][TSdParams::eT0] = cov[0];
  fVarCov[TSdParams::eU][TSdParams::eU] = cov[2];
  fVarCov[TSdParams::eU][TSdParams::eV] = fVarCov[TSdParams::eV][TSdParams::eU] = cov[4];
  fVarCov[TSdParams::eV][TSdParams::eV] = cov[5];
  errflag = 0;

  if(PrintLevel>2) printf("FitCurved :   t0 %5.0f  u,v %5.3f %5.3f  1/R %7.5f  chi2 %5.2f\n",t0,u,v,curv,chi2);
}

static void fcnLDF(int &npar, double *gin, double &f, double *par, int iflag) {
  double signal,signal_th,dx,dy,dist,chi2;
  double beta = 0., gamma = 0., xmax = 0.;

  if(0) printf("%d %f %d",npar,gin[0],iflag); // to avoid compiler warning

  double sig1000 = exp(par[2]);
  if(thePBRec().FitLDFSlope<0) xmax = par[3];
  else {
    beta = par[3];
    gamma = par[4];
  }

  chi2=0;
  for (int i=0; i<thePBRec().nActiveStation; i++) if(thePBRec().InFit[i]) {
      signal = thePBRec().Signal[i];
      dx = thePBRec().X[i] - thePBRec().XBarycenter - par[0];
      dy = thePBRec().Y[i] - thePBRec().YBarycenter - par[1];
      dist = sqrt(sqr(dx)+sqr(dy)-sqr(thePBRec().uCurrent*dx + thePBRec().vCurrent*dy));
      if(thePBRec().FitLDFSlope<0) {
        signal_th = sig1000*PARfunc(dist,870/sqrt(1-sqr(thePBRec().uCurrent)-sqr(thePBRec().vCurrent))-xmax);
      } else signal_th = sig1000*LDFfunc(dist,beta,gamma);

      if(thePBRec().CorrectAsym) signal_th *= thePBRec().AsymFact(dx,dy);
      // lower bound on saturated stations
      if((thePBRec().Satur1[i]<1000 && thePBRec().Satur2[i]<1000 && thePBRec().Satur3[i]<1000) || signal>signal_th)
        chi2 += sqr(signal-signal_th)/signal_th;
    }
  if(thePBRec().SilentLDF) {
    for (int i=0; i<thePBRec().nSilentStation; i++) if(thePBRec().SilentInFit[i]) {
        dx = thePBRec().XSilent[i] - thePBRec().XBarycenter - par[0];
        dy = thePBRec().YSilent[i] - thePBRec().YBarycenter - par[1];
        dist = sqrt(sqr(dx)+sqr(dy)-sqr(thePBRec().uCurrent*dx + thePBRec().vCurrent*dy));
        if(thePBRec().FitLDFSlope<0)
          signal_th = sig1000*PARfunc(dist,870/sqrt(1-sqr(thePBRec().uCurrent)-sqr(thePBRec().vCurrent))-xmax);
        else
          signal_th = sig1000*LDFfunc(dist,beta,gamma);
        if(thePBRec().CorrectAsym) signal_th *= thePBRec().AsymFact(dx,dy);
        chi2 += signal_th / thePBRec().LDF_SIGTHRESH;
      }
  }
  f = chi2;
}

/*-------------------------------------------------------------------*/
void TPBRec::FitLDFMinuit(double& xc, double& dxc, double& yc, double& dyc,
                          double& s1000, double& ds1000, double *distc, double& chi2,
                          double& beta, double& dbeta, double& gamma, int& errflag)/// Fitting the LDF with Minuit
/*-------------------------------------------------------------------*/
{
  Int_t ierflg;
  double dgamma;
  double xmax,dxmax;

  TMinuit* gMin = new TMinuit(5);
  gMin->SetPrintLevel(-1);
  gMin->SetFCN(fcnLDF);
  gMin->mnparm(0,"dx",xc,100,0,0,ierflg);
  gMin->mnparm(1,"dy",yc,100,0,0,ierflg);
  gMin->mnparm(2,"S1000",log(s1000),1,0,0,ierflg);
  if(FitLDFSlope>=0) {
    gMin->mnparm(3,"beta",beta,.1,0,0,ierflg);
    gMin->mnparm(4,"gamma",gamma,.1,0,0,ierflg);
    if(FitLDFSlope<1) gMin->FixParameter(3);
    if(FitLDFSlope<2) gMin->FixParameter(4);
  } else {
    gMin->mnparm(3,"xmax",800,50,0,0,ierflg);
  }

  gMin->Command("MIGRAD 500");
  Double_t amin, edm, errdef;
  Int_t npari, nparx;
  gMin->mnstat(amin,edm,errdef,npari,nparx,ierflg);
  errflag = (ierflg<3);

  gMin->GetParameter(0,xc,dxc);
  gMin->GetParameter(1,yc,dyc);
  gMin->GetParameter(2,s1000,ds1000);
  s1000 = exp(s1000);
  ds1000 *= s1000;
  if(FitLDFSlope>0) gMin->GetParameter(3,beta,dbeta);
  if(FitLDFSlope>1) gMin->GetParameter(4,gamma,dgamma);
  if(FitLDFSlope<0) gMin->GetParameter(3,xmax,dxmax);
  chi2 = gMin->fAmin;

  fVarCov[TSdParams::eXCore][TSdParams::eXCore] = sqr(dxc);
  fVarCov[TSdParams::eYCore][TSdParams::eYCore] = sqr(dyc);
  fVarCov[TSdParams::eSRef][TSdParams::eSRef] = sqr(ds1000);

  for(int i=0; i<nActiveStation; i++) {
    double dx = X[i] - XBarycenter - xc;
    double dy = Y[i] - YBarycenter - yc;
    distc[i] = sqrt(sqr(dx)+sqr(dy)-sqr(uCurrent*dx + vCurrent*dy));
    if(FitLDFSlope<0) {
      sigFit[i] = s1000*PARfunc(distc[i],870/cos(Theta)-xmax);
    } else
      sigFit[i] = s1000*LDFfunc(distc[i],beta,gamma);
    if(CorrectAsym) sigFit[i] *= AsymFact(dx,dy);
  }

  if(PrintLevel>1) {
    printf(" dCore");
    for(int i=0; i<nActiveStation; i++)
      printf("%6.0f ",distc[i]);
    printf("\n");
    printf(" sFit ");
    for(int i=0; i<nActiveStation; i++)
      printf("%6.1f ",sigFit[i]);
    printf("  LDF");
    if(FitLDFSlope) printf("b");
    if(SilentLDF) printf("s");
    printf("   ");
    printf("xc %2.0f+-%2.0f  yc %2.0f+-%2.0f  s1000 %5.2f+-%5.2f  chi2 %4.2f",
           xc,dxc,yc,dyc,s1000,ds1000,chi2);
    printf("  beta %6.3f",beta);
    if(FitLDFSlope) printf("+-%5.3f",dbeta);
    printf("  gamma %6.3f",gamma);
    if(FitLDFSlope>1) printf("+-%5.3f",dgamma);
    printf("\n");
  }

  if(SilentLDF && PrintLevel>2) {
    printf(" silent : mxId  dist  sigf\n");
    for(int i=0; i<nSilentStation; i++) if(SilentInFit[i]) {
        double dx = XSilent[i] - XBarycenter - xc;
        double dy = YSilent[i] - YBarycenter - yc;
        double dist = sqrt(sqr(dx)+sqr(dy)-sqr(uCurrent*dx + vCurrent*dy));
        double sigf = s1000;
        sigf *= LDFfunc(dist,beta,gamma);
        if(CorrectAsym) sigf *= AsymFact(dx,dy);
        printf("        %5d %5.0f  %4.2f\n",mxIdSilent[i],dist,sigf);
      }
  }
  delete gMin;

}

static void fcnGlob(int &npar, double *gin, double &f, double *par, int iflag) {
  double sinth2,costh,beta,gamma,dx,dy,tplane,dist2,dist,dtcurv,wgt;
  double zsh,s1000,signal,signal_th,chi2;

  if(0) printf("%d %f %d",npar,gin[0],iflag); // to avoid compiler warning

  sinth2 = minval(sqr(par[1])+sqr(par[2]),1);
  costh = sqrt(1-sinth2);
  beta = thePBRec().fBeta;
  gamma = thePBRec().fGamma;

  s1000 = exp(par[6]);

  chi2=0;
  for (int i=0; i<thePBRec().nActiveStation; i++) if(thePBRec().InFit[i]) {
      dx = thePBRec().X[i] - thePBRec().XBarycenter - par[4];
      dy = thePBRec().Y[i] - thePBRec().YBarycenter - par[5];
      // time
      zsh = par[1]*dx + par[2]*dy;
      tplane = par[0] - zsh/kCdasUtil::CSPEED;
      dist2 = sqr(dx)+sqr(dy)-sqr(zsh);
      dtcurv = par[3]*(1+par[3]*zsh)*dist2/kCdasUtil::CSPEED/2;
      wgt=1./(sqr(thePBRec().TIME_ERR0)+sqr(thePBRec().TIME_ERR_FACTD2*dist2*costh));
      chi2 += wgt * sqr(thePBRec().T[i] - thePBRec().TBarycenter - dtcurv - tplane);
      // amplitude
      signal = thePBRec().Signal[i];
      dist = sqrt(dist2);
      signal_th =  s1000*LDFfunc(dist,beta,gamma);
      if(thePBRec().CorrectAsym) signal_th *= thePBRec().AsymFact(dx,dy);
      // lower bound on saturated stations
      if((thePBRec().Satur1[i]<1000 && thePBRec().Satur2[i]<1000 && thePBRec().Satur3[i]<1000) || signal>signal_th)
        chi2 += sqr(signal-signal_th)/signal_th;
    }
  if(thePBRec().SilentLDF) {
    for (int i=0; i<thePBRec().nSilentStation; i++) if(thePBRec().SilentInFit[i]) {
        dx = thePBRec().XSilent[i] - thePBRec().XBarycenter - par[4];
        dy = thePBRec().YSilent[i] - thePBRec().YBarycenter - par[5];
        dist = sqrt(sqr(dx)+sqr(dy)-sqr(par[1]*dx + par[2]*dy));
        signal_th =  s1000*LDFfunc(dist,beta,gamma);
        if(thePBRec().CorrectAsym) signal_th *= thePBRec().AsymFact(dx,dy);
        chi2 += signal_th / thePBRec().LDF_SIGTHRESH;
      }
  }
  f = chi2;
}

/*-------------------------------------------------------------------*/
void TPBRec::FitGlobMinuit(double& t0, double& u, double& v, double& curv,
                           double& xc, double& yc, double& s1000, double *distc, double *cov,
                           double& chi2, int& errflag)/// Global fit with Minuit
/*-------------------------------------------------------------------*/
{
  Int_t ierflg;
  double beta,gamma,du,dv,dcurv,dxc,dyc,ds1000;

  TMinuit* gMin = new TMinuit(7);
  gMin->SetPrintLevel(-1);
  gMin->SetFCN(fcnGlob);
  gMin->mnparm(0,"t0",t0,100,0,0,ierflg);
  gMin->mnparm(1,"u",u,.01,0,0,ierflg);
  gMin->mnparm(2,"v",v,.01,0,0,ierflg);
  gMin->mnparm(3,"curv",curv,.0001,0,0,ierflg);
  gMin->mnparm(4,"xc",xc,100,0,0,ierflg);
  gMin->mnparm(5,"yc",yc,100,0,0,ierflg);
  gMin->mnparm(6,"logs",log(s1000),1,0,0,ierflg);

  gMin->Command("MIGRAD 500");
  Double_t amin, edm, errdef;
  Int_t npari, nparx;
  gMin->mnstat(amin,edm,errdef,npari,nparx,ierflg);
  errflag = (ierflg<3);

  gMin->GetParameter(1,u,du);
  gMin->GetParameter(2,v,dv);
  gMin->GetParameter(3,curv,dcurv);
  gMin->GetParameter(4,xc,dxc);
  gMin->GetParameter(5,yc,dyc);
  gMin->GetParameter(6,s1000,ds1000);
  gMin->mnemat(cov,7);
  s1000 = exp(s1000);
  ds1000 *= s1000;
  chi2 = gMin->fAmin;

  fVarCov[TSdParams::eU][TSdParams::eU] = sqr(du) ;
  fVarCov[TSdParams::eV][TSdParams::eV] = sqr(dv) ;
  fVarCov[TSdParams::eU][TSdParams::eV] =
    fVarCov[TSdParams::eV][TSdParams::eU] = cov[9];
  fVarCov[TSdParams::eXCore][TSdParams::eXCore] = sqr(dxc);
  fVarCov[TSdParams::eYCore][TSdParams::eYCore] = sqr(dyc);
  fVarCov[TSdParams::eSRef][TSdParams::eSRef] = sqr(ds1000);
  fVarCov[TSdParams::eR][TSdParams::eR] = sqr(dcurv/curv/curv);

  for(int i=0; i<nActiveStation; i++) {
    double dx = X[i] - XBarycenter - xc;
    double dy = Y[i] - YBarycenter - yc;
    distc[i] = sqrt(sqr(dx)+sqr(dy)-sqr(u*dx + v*dy));
  }

  if(PrintLevel>1) {
    printf(" dCore");
    for(int i=0; i<nActiveStation; i++)
      printf("%6.0f ",distc[i]);
    printf("\n");
    beta = fBeta;
    gamma = fGamma;
    printf(" sFit ");
    for(int i=0; i<nActiveStation; i++) {
      double signal_th =  s1000*LDFfunc(distc[i],beta,gamma);
      if(CorrectAsym) signal_th *= AsymFact(X[i]-XBarycenter-xc,Y[i]-YBarycenter-yc);
      printf("%6.1f ",signal_th);
    }
    if(SilentLDF) printf("  GLOBs  ");
    else printf("  GLOB   ");
    printf("xc %2.0f+-%2.0f  yc %2.0f+-%2.0f  s1000 %5.2f+-%5.2f  chi2 %4.2f  th %5.2f  ph %5.2f\n",
           xc,dxc,yc,dyc,s1000,ds1000,chi2,asin(sqrt(u*u+v*v))*180/M_PI,atan2(v,u)*180/M_PI);
  }

  delete gMin;
}

static void fcnGlobFixCurv(int &npar, double *gin, double &f, double *par, int iflag) {
  double sinth2,costh,beta,gamma,dx,dy,tplane,dist2,dist,curv,dtcurv,wgt;
  double zsh,s1000,signal,signal_th,chi2;

  if(0) printf("%d %f %d",npar,gin[0],iflag); // to avoid compiler warning

  sinth2 = minval(sqr(par[1])+sqr(par[2]),1);
  costh = sqrt(1-sinth2);
  curv = thePBRec().SELECT_CURV0*costh;
  beta = thePBRec().fBeta;
  gamma = thePBRec().fGamma;
  s1000 = exp(par[5]);

  chi2=0;
  for (int i=0; i<thePBRec().nActiveStation; i++) if(thePBRec().InFit[i]) {
      dx = thePBRec().X[i] - thePBRec().XBarycenter - par[3];
      dy = thePBRec().Y[i] - thePBRec().YBarycenter - par[4];
      // time
      zsh = par[1]*dx + par[2]*dy;
      tplane = par[0] - zsh/kCdasUtil::CSPEED;
      dist2 = sqr(dx)+sqr(dy)-sqr(zsh);
      dtcurv = curv*(1+curv*zsh)*dist2/kCdasUtil::CSPEED/2;
      wgt=1./(sqr(thePBRec().TIME_ERR0)+sqr(thePBRec().TIME_ERR_FACTD2*dist2*costh));
      chi2 += wgt * sqr(thePBRec().T[i] - thePBRec().TBarycenter - dtcurv - tplane);
      // amplitude
      signal = thePBRec().Signal[i];
      dist = sqrt(dist2);
      signal_th = s1000*LDFfunc(dist,beta,gamma);
      if(thePBRec().CorrectAsym) signal_th *= thePBRec().AsymFact(dx,dy);
      // lower bound on saturated stations
      if((thePBRec().Satur1[i]<1000 && thePBRec().Satur2[i]<1000 && thePBRec().Satur3[i]<1000) || signal>signal_th)
        chi2 += sqr(signal-signal_th)/signal_th;
    }
  if(thePBRec().SilentLDF) {
    for (int i=0; i<thePBRec().nSilentStation; i++) if(thePBRec().SilentInFit[i]) {
        dx = thePBRec().XSilent[i] - thePBRec().XBarycenter - par[3];
        dy = thePBRec().YSilent[i] - thePBRec().YBarycenter - par[4];
        dist = sqrt(sqr(dx)+sqr(dy)-sqr(par[1]*dx + par[2]*dy));
        signal_th = s1000*LDFfunc(dist,beta,gamma);
        if(thePBRec().CorrectAsym) signal_th *= thePBRec().AsymFact(dx,dy);
        chi2 += signal_th / thePBRec().LDF_SIGTHRESH;
      }
  }
  f = chi2;
}

/*-------------------------------------------------------------------*/
void TPBRec::FitGlobFixCurvMinuit(double& t0, double& u, double& v, double& xc, double& yc,
                                  double& s1000, double *distc, double *cov, double& chi2, int& errflag)/// global fit with Minuit (fixed curvature)
/*-------------------------------------------------------------------*/
{
  Int_t ierflg;
  double beta,gamma,du,dv,dxc,dyc,ds1000;

  TMinuit* gMin = new TMinuit(6);
  gMin->SetPrintLevel(-1);
  gMin->SetFCN(fcnGlobFixCurv);
  gMin->mnparm(0,"t0",t0,100,0,0,ierflg);
  gMin->mnparm(1,"u",u,.01,0,0,ierflg);
  gMin->mnparm(2,"v",v,.01,0,0,ierflg);
  gMin->mnparm(3,"xc",xc,100,0,0,ierflg);
  gMin->mnparm(4,"yc",yc,100,0,0,ierflg);
  gMin->mnparm(5,"logs",log(s1000),1,0,0,ierflg);

  gMin->Command("MIGRAD 500");
  Double_t amin, edm, errdef;
  Int_t npari, nparx;
  gMin->mnstat(amin,edm,errdef,npari,nparx,ierflg);
  errflag = (ierflg<3);

  gMin->GetParameter(1,u,du);
  gMin->GetParameter(2,v,dv);
  gMin->GetParameter(3,xc,dxc);
  gMin->GetParameter(4,yc,dyc);
  gMin->GetParameter(5,s1000,ds1000);
  gMin->mnemat(cov,6);
  s1000 = exp(s1000);
  ds1000 *= s1000;
  chi2 = gMin->fAmin;

  fVarCov[TSdParams::eU][TSdParams::eU] = sqr(du) ;
  fVarCov[TSdParams::eV][TSdParams::eV] = sqr(dv) ;
  fVarCov[TSdParams::eU][TSdParams::eV] =
    fVarCov[TSdParams::eV][TSdParams::eU] = cov[8];
  fVarCov[TSdParams::eXCore][TSdParams::eXCore] = sqr(dxc);
  fVarCov[TSdParams::eYCore][TSdParams::eYCore] = sqr(dyc);
  fVarCov[TSdParams::eSRef][TSdParams::eSRef] = sqr(ds1000);

  for(int i=0; i<nActiveStation; i++) {
    double dx = X[i] - XBarycenter - xc;
    double dy = Y[i] - YBarycenter - yc;
    distc[i] = sqrt(sqr(dx)+sqr(dy)-sqr(u*dx + v*dy));
  }

  if(PrintLevel>1) {
    printf(" dCore");
    for(int i=0; i<nActiveStation; i++)
      printf("%6.0f ",distc[i]);
    printf("\n");
    double costh = sqrt(maxval(1-u*u-v*v,.1));
    if(costh>.5)  beta = 4.735 - 1.236/costh;
    else          beta = 2.672 - 0.161/costh;
    if(costh>.45)  gamma = .26 + .27*sqr(1/costh-1.4);
    else           gamma = .43;
    printf(" sFit ");
    for(int i=0; i<nActiveStation; i++) printf("%6.1f ",s1000*LDFfunc(distc[i],beta,gamma));
    if(SilentLDF) printf("  GLOBs  ");
    else printf("  GLOB   ");
    printf("xc %2.0f+-%2.0f  yc %2.0f+-%2.0f  s1000 %5.2f+-%5.2f  chi2 %4.2f  th %5.2f  ph %5.2f\n",
           xc,dxc,yc,dyc,s1000,ds1000,chi2,asin(sqrt(u*u+v*v))*180/M_PI,atan2(v,u)*180/M_PI);
  }

  delete gMin;
}

/*-------------------------------------------------------------------*/
double TPBRec::AsymFact(double dx, double dy)
/*-------------------------------------------------------------------*/
{
  double sinth,costh,xs,ys,dist,psi,fact;
  sinth = sqrt(sqr(uCurrent)+sqr(vCurrent));
  costh = sqrt(1-sinth*sinth);
  xs = (dx*uCurrent + dy*vCurrent) * costh/sinth;
  ys = (dy*uCurrent - dx*vCurrent) / sinth;
  dist = sqrt(xs*xs+ys*ys);
  psi = atan2(ys,xs);
  if(costh<.3) fact = .05;
  else fact = .22*(1-costh)*(costh-.25)/sqr(.375);
  fact *= atan(dist/300.);
  return 1+fact*cos(psi);
}

/*-------------------------------------------------------------------*/
void TPBRec::ConvertS1000ToEnergy(double costh,double s1000,double ds1000)
/*-------------------------------------------------------------------*/
{
  double Xmax,X,fact;

  // first approximation
  fact = 7.8/sqrt(1+11.8*sqr(1/costh-1));
  fE = pow(s1000/fact,1.05);
  // parametrization as a function of Xground-Xmax
  Xmax = 720 + 60*log10(fE); // proton average
  //Xmax = 640 + 60*log10(fE); // iron average
  X = 870/costh-Xmax;
  fact = .8 + 4/(1+sqr((X-150)/420));  // new param for proton
  //fact = 1 + 5/(1+sqr((X-150)/420));  // new param for iron
  fE = pow(s1000/fact,1.05);

  //iteration
  Xmax = 720 + 60*log10(fE); // proton average
  //Xmax = 640 + 60*log10(fE); // iron average
  X = 870/costh-Xmax;
  fact = .8 + 4/(1+sqr((X-150)/420));  // new param for proton
  //fact = 1 + 5/(1+sqr((X-150)/420));  // new param for iron

  fE = pow(s1000/fact,1.05);
  fdE = (ds1000/s1000)*fE*1.05;

}

#define EASTING_CENTER  477256.66   //  UTM coord. of the reference point
#define NORTHING_CENTER 6099203.68  //
/*-------------------------------------------------------------------*/
void TPBRec::SideralCoord()
/*-------------------------------------------------------------------*/
{
  //double longitude = -69.25 + 7.35e-6*(fXCore+EASTING_ORIGIN-EASTING_CENTER);
  //double latitude = -35.25+9e-6*(fYCore+NORTHING_ORIGIN-NORTHING_CENTER);

  RightAscension = fErEvent->fRec.Ra(); if(RightAscension<0) RightAscension += 360;
  Declination = fErEvent->fRec.Dec();
  GalacticL = fErEvent->fRec.L(); if(GalacticL<0) GalacticL += 360;
  GalacticB = fErEvent->fRec.B();
}

/*-------------------------------------------------------------------*/
int TPBRec::Process()
/*-------------------------------------------------------------------*/
{
  memset((TShowerParams *)this, 0, sizeof(TShowerParams));
  PlanFitOK = CurvFitOK = LDFFitOK = GlobFitOK = Fit3stOK = 0;
  uPlan = vPlan = uCurv = vCurv = uCurrent = vCurrent = Curv = fR = dCurv = 0;
  Theta = dTheta = Phi = dPhi = 0;
  fXCore = dXCore = fYCore = dYCore = fSRef = dS1000 = fE = fdE = 0;
  chi2Plan = chi2Curv = residPlan = residCurv = 0;
  chi2LDF = chi2LDFwithSilent = chi2Glob = chi2GlobwithSilent = 0;
  fBeta = 0;

  // define the active stations
  if(!ActiveStations()) return 0;
  if(PrintLevel) {
    printf("==== event %d ==== %d stations (%d window)  ",NumEvent,nActiveStation,nWindow);
    cout<<"trigger "<<fErEvent->Trigger.Algo<<" (T4 = "<<T4Flag<<") ==== "<<fErEvent->UTCDate<<endl;
    if(Simulation && ReadTruth)
      printf(" --- simulated values:  prim %d  energy %6.3f  theta %5.2f  phi %5.2f  x %5.0f  y %6.0f  Xmax %5.1f\n",
             true_prim,true_energy/1e9,true_theta*180/M_PI,true_phi*180/M_PI,true_east-EASTING_ORIGIN,true_north-NORTHING_ORIGIN,true_xmax);
  }

  // flag for containment inside array
  ContainedOK = ContainedEvent();
  if(PrintLevel && !ContainedOK) printf("--- event %d not contained\n",NumEvent);
  if(SelectContained && !ContainedOK) return 0;

  // stop here if "reconstruction" is not requested
  nStationInFit=0;
  if(!Reconstruct) return 0;

  // build the list of "silent" stations (error=2 and window not 0)
  SilentStations();
  if(PrintLevel>1) PrintStations();

  // event analysis (either 3-fold or "general")

  // reconstruct event as "3-fold" (excluding CM + another one)
  if(Process_3st && nActiveStation==3 &&
      nCarmenMiranda<2 && nDiaNoche<2 && nMoulinRouge<2) Anal3st(0,1,2);

  // reconstruct event as "general" (at least 4 stations, counting a doublet as one)
  if(Process_gen &&
      //((nActiveStation>3 && nCarmenMiranda<2 && nDiaNoche<2 && nMoulinRouge<2)
      //|| nActiveStation>4)) {
      nActiveStation>3) {

    if(Select()) {
      AnalGen();
    } else if(Process_3st) {
      // trying to recover rejected events as 3-fold ones
      if(PrintLevel>1) printf("---- Trying to recover a 3-fold configuration\n");
      Recover3st();
    }

  }

  AvgRiseTime = AsymRiseTime = AvgFallTime = AsymFallTime = AvgShape = AsymShape = 0;
  if((nStationInFit>3 && LDFFitOK) || (nStationInFit==3 && Fit3stOK)) {
    for(int is=0; is<nActiveStation; is++) {
      double tstart = (stat[is]->fStartBin-1)*kEc::TimeSlot;
      Tfrac10[is] = stat[is]->T10()-tstart;
      Tfrac50[is] = stat[is]->T50()-tstart;
      Tfrac90[is] = stat[is]->T90()-tstart;
    }
    CleanTrace();
    RiseFallTime();
    if(LDFFitOK || Fit3stOK) FindPeak();
    if(LDFFitOK || Fit3stOK) MuonCount();

  }
  // compute equatorial and galactic coordinates of shower direction
  SideralCoord();

  // write the results on specific streams
  if(Write1_result || Write2_result) {
    GlobalTriggerType = 0;
    if(!strncmp(fErEvent->Trigger.Algo.c_str(),"2C1&3C2",7) ||
        !strncmp(fErEvent->Trigger.Algo.c_str(),"3C2&",4)) GlobalTriggerType = 1;
    if(!strncmp(fErEvent->Trigger.Algo.c_str(),"TOT",3)) GlobalTriggerType = 2;
    if(!strncmp(fErEvent->Trigger.Algo.c_str(),"3C1H",4)) GlobalTriggerType = 5;
    if(!strncmp(fErEvent->Trigger.Algo.c_str(),"RAN",3)) GlobalTriggerType = 6;
    if(!strncmp(fErEvent->Trigger.Algo.c_str(),"FD L",4)) GlobalTriggerType = 11;
    if(!strncmp(fErEvent->Trigger.Algo.c_str(),"FD C",4)) GlobalTriggerType = 12;
    if(!strncmp(fErEvent->Trigger.Algo.c_str(),"FD M",4)) GlobalTriggerType = 13;
  }

  UTCTimeEvent = (int)fErEvent->UTCTime;
  GeomFlag = PlanFitOK+10*CurvFitOK+100*ContainedOK;
  if(nStationInFit>3) LDFFlag = LDFFitOK;
  else LDFFlag = Fit3stOK;
#ifdef __DARWIN__
  if(isnan(chi2LDF)) {
    chi2LDF = 0; LDFFitOK=0;
  }
  if(isnan(chi2Glob)) {
    chi2Glob = 0; GlobFitOK=0;
  }
#else
  if(isinf(chi2LDF) || isnan(chi2LDF)) {
    chi2LDF = 0; LDFFitOK=0;
  }
  if(isinf(chi2Glob) || isnan(chi2Glob)) {
    chi2Glob = 0; GlobFitOK=0;
  }
#endif
  for(int is=0; is<nActiveStation; is++) {
    selId[is] = stId[is];
    if(!InFit[is])  selId[is] *= -1;
    int ttyp = stat[is]->Trigger.Type;
    LocalTriggerType[is] = (ttyp&1 || (ttyp>>8)&1) + 2*(ttyp&2 || (ttyp>>8)&2);
    Tshift[is] = T[is] - fT0 - TBarycenter;
  }

  if(Write1_result) {
    //event data (short)
    fprintf(res1,"%6u %5d %5d %2d %2d  %5.2f %7.2f %6.3f %2d %2d %2d ",
            NumEvent,MonitDay,TimeInDay,GlobalTriggerType,T4Flag,
            Theta*180/M_PI,Phi*180/M_PI,fE,nActiveStation,nWindow,nStationInFit);
    //station list
    for(int is=0; is<nActiveStation; is++) fprintf(res1," %3d",selId[is]);
    fprintf(res1,"\n");
  }

  if(Write2_result && nStationInFit>0) {
    //event data
    fprintf(res2,"%6u %d  %5d %5d %2d %2d  %2d %2d %2d  %3d %4d %d  ",
            NumEvent,UTCTimeEvent,MonitDay,TimeInDay,GlobalTriggerType,T4Flag,
            nActiveStation,nWindow,nStationInFit,GeomFlag,LDFFlag,GlobFitOK);
    fprintf(res2,"%5.2f %4.2f %7.2f %4.2f %7.4f %5.4f ",
            Theta*180/M_PI,dTheta*180/M_PI,Phi*180/M_PI,dPhi*180/M_PI,1000*Curv,1000*dCurv);
    fprintf(res2,"%6.2f %5.2f %6.3f %5.3f  %6.0f %3.0f %6.0f %3.0f %5.3f %5.3f ",
            fSRef,dS1000,fE,fdE,fXCore,dXCore,fYCore,dYCore,fBeta,dbetaLDF);
    fprintf(res2,"%5.2f %5.2f %5.2f %5.2f",chi2Plan,chi2Curv,chi2LDF,chi2Glob);
    fprintf(res2," %6.3f %6.3f %6.3f %6.3f  %6.4f %6.4f ",
            AvgRiseTime,AsymRiseTime,AvgFallTime,AsymFallTime,AvgShape,AsymShape);
    fprintf(res2," %6.2f %6.2f %6.2f %6.2f ",
            RightAscension,Declination,GalacticL,GalacticB);
    fprintf(res2,"\n");
  }

  if(Write3_result && nStationInFit>0) {
    //event data
    fprintf(res3,"%6u %d  %5d %5d %2d %2d  %2d %2d %2d  %3d %4d %d  ",
            NumEvent,UTCTimeEvent,MonitDay,TimeInDay,GlobalTriggerType,T4Flag,
            nActiveStation,nWindow,nStationInFit,GeomFlag,LDFFlag,GlobFitOK);
    fprintf(res3,"%5.2f %4.2f %7.2f %4.2f %7.4f %5.4f ",
            Theta*180/M_PI,dTheta*180/M_PI,Phi*180/M_PI,dPhi*180/M_PI,1000*Curv,1000*dCurv);
    fprintf(res3,"%6.2f %5.2f %6.3f %5.3f  %6.0f %3.0f %6.0f %3.0f %5.3f %5.3f ",
            fSRef,dS1000,fE,fdE,fXCore,dXCore,fYCore,dYCore,fBeta,dbetaLDF);
    fprintf(res3,"%5.2f %5.2f %5.2f %5.2f",chi2Plan,chi2Curv,chi2LDF,chi2Glob);
    fprintf(res3," %6.3f %6.3f %6.3f %6.3f  %6.4f %6.4f ",
            AvgRiseTime,AsymRiseTime,AvgFallTime,AsymFallTime,AvgShape,AsymShape);
    fprintf(res3," %6.2f %6.2f %6.2f %6.2f ",
            RightAscension,Declination,GalacticL,GalacticB);
    fprintf(res3,"\n");

    //station data
    for(int is=0; is<nActiveStation; is++) {
      if(InFit[is]) fprintf(res3,"+%d %4d ",stId[is],mxId[is]);
      else          fprintf(res3,"-%d %4d ",stId[is],mxId[is]);
      fprintf(res3,"%d %2d %6.0f %6.0f %3.0f %6.0f %8.3f %8.3f  %4.0f %6.3f  %4.0f %4.0f %4.0f  %6.4f ",
              LocalTriggerType[is],Anom[is],X[is],Y[is],Z[is],Tshift[is],Signal[is],sigFit[is],distCore[is],azimCore[is],
              Tclean10[is],Tclean50[is],Tclean90[is],Shape[is]);
      fprintf(res3," %4d %4d %4d ",Satur1[is],Satur2[is],Satur3[is]);
      fprintf(res3," %3d",nMuPeak[is]);
      fprintf(res3," %6.4f",AsymPMT[is]);
      fprintf(res3,"\n");
    }
  }

  // printout of the results of the reconstruction
  if(PrintLevel) {
    if(PrintLevel>1) printf("\n");
    printf("%d stat. kept - Plan %d Curv %d LDF %d Glob %d 3st %d    barycenter %4.0f %4.0f\n",
           nStationInFit,PlanFitOK,CurvFitOK,LDFFitOK,GlobFitOK,Fit3stOK,XBarycenter,YBarycenter);
    printf("th %5.2f+-%4.2f  ph %7.2f+-%4.2f  Rcurv %6.0f+-%4.0f  ",
           Theta*180/M_PI,dTheta*180/M_PI,Phi*180/M_PI,dPhi*180/M_PI,1/Curv,dCurv/Curv/Curv);
    printf("S %5.2f+-%4.2f  E %4.2f+-%4.2f  ",
           fSRef,dS1000,fE,fdE);
    printf("X,Y %4.0f+-%2.0f %4.0f+-%2.0f\n",
           fXCore,dXCore,fYCore,dYCore);
    printf("chi2  P %4.2f  C %4.2f  L %4.2f  Ls %4.2f  G %4.2f  Gs %4.2f\n",
           chi2Plan, chi2Curv, chi2LDF, chi2LDFwithSilent, chi2Glob, chi2GlobwithSilent);
  }

  // special output for graphical animated display
  if(Display_output && nActiveStation>3) WriteDisplay();

  // Filling TShowerParams parameters
  fBeta = -fBeta;
  fGamma = -fGamma;
  fT0 += TBarycenter;

  return 1;
}

/*-------------------------------------------------------------------*/
int TPBRec::ActiveStations()
/*-------------------------------------------------------------------*/
{
  AnomEvent = AnomalyEvent();

  // loop over stations to extract the "active" ones (those with error=0)
  nActiveStation = 0;
  unsigned int size = fErEvent->fSelectedStations.size();
  for(unsigned int i=0; i<size; i++) {
    int ids = fErEvent->fSelectedStations[i]->Id;
    fErEvent->fRandomStations.push_back(fErEvent->fSelectedStations[i]);
    // if requested: remove Enginneering Array
    if(!Simulation && RemoveEA && ids<100) continue;
    // if requested: remove second member of doublets
    if(!Simulation && RemoveDoublet && (ids==64 || ids==185 || ids==186)) continue;
    //if(!Simulation && RemoveDoublet && (ids==64 || ids==139 || ids==140)) continue;
    // if requested: remove infill stations
    if(!Simulation && RemoveInfillStation && (ids==36 || ids==67)) continue;

    // reject isolated stations
    int neighbour = 0;
    int neighbour2 = 0;
    for(unsigned int j=0; j<size; j++) if(j!=i) {
        double dist2 = sqr(fErEvent->fSelectedStations[j]->fX-fErEvent->fSelectedStations[i]->fX)
                       + sqr(fErEvent->fSelectedStations[j]->fY-fErEvent->fSelectedStations[i]->fY);
        neighbour += dist2<sqr(ISOLATION_DISTANCE);
        neighbour2 += dist2<sqr(ISOLATION_DISTANCE2);
      }
    // reject if no neighbour within ISOLATION_DISTANCE or less than 2 within ISOLATION_DISTANCE2
    if(!neighbour || neighbour2<2) continue;

    // Look for anomalies in stations
    int anomaly_level = fErEvent->fSelectedStations[i]->Anomaly();
    if(anomaly_level%10>2 || anomaly_level >= 100) continue;

    // create a new "active" station
    stat[nActiveStation] = fErEvent->fSelectedStations[i];
    Anom[nActiveStation] =  anomaly_level;

    // filter the FADC trace recompute the integrated signal from RecStation
    int start,stop;
    if(!CleanFilter(fErEvent->fSelectedStations[i],start,stop)) continue;

    CleanStart[nActiveStation] = start;
    CleanStop[nActiveStation] = stop;
    if(CleanSignal(stat[nActiveStation],start,stop,Signal[nActiveStation],Peak[nActiveStation])) {
      nActiveStation++;
      fErEvent->fRandomStations.erase(fErEvent->fRandomStations.end()-1);
    }
  }

  for(unsigned int i=0; i<fErEvent->fRandomStations.size(); i++) {
    for(unsigned int j=0; j<fErEvent->fSelectedStations.size(); j++)
      if(fErEvent->fSelectedStations[j]->Id==
          fErEvent->fRandomStations[i]->Id) {
        fErEvent->fSelectedStations.erase(fErEvent->fSelectedStations.begin()+j);
        break;
      }
  }

  if(nActiveStation<3) {
    if(PrintLevel) printf("!!!! event %d rejected (%d active stations)\n",NumEvent,nActiveStation);
    if(Diagno_output) fprintf(diag,"event %d rejected (%d active stations)\n",NumEvent,nActiveStation);
    AnomEvent += 10;
    return 0;
  }

  // sort the stations by decreasing value of the integrated signal

  for(int i=1; i<nActiveStation; i++)
    for (unsigned int j=i; j>0; j--)
      if(Signal[j] > Signal[j-1]) {
        TCalibStation* temp = (TCalibStation*)stat[j]; stat[j] = stat[j-1]; stat[j-1] = temp;
        int itemp = Anom[j]; Anom[j] = Anom[j-1]; Anom[j-1] = itemp;
        itemp = CleanStart[j]; CleanStart[j] = CleanStart[j-1]; CleanStart[j-1] = itemp;
        itemp = CleanStop[j]; CleanStop[j] = CleanStop[j-1]; CleanStop[j-1] = itemp;
        double dtemp = Signal[j]; Signal[j] = Signal[j-1]; Signal[j-1] = dtemp;
      }

  // loop over active stations
  nCarmenMiranda = nDiaNoche = nMoulinRouge = 0;
  nWindow=0;
  for(int i=0; i<nActiveStation; i++) {
    stId[i] = stat[i]->Id;
    int ix = (int)((stat[i]->Easting-MXID_EASTING_ORIGIN)/MXID_EASTING_STEP);
    int iy = (int)((stat[i]->Northing-MXID_NORTHING_ORIGIN)/MXID_NORTHING_STEP);
    mxId[i] = ix + 100*iy;

    // counting doublets
    if(stat[i]->Id == 49 || stat[i]->Id == 64) nCarmenMiranda++;
    if(stat[i]->Id == 139 || stat[i]->Id == 186) nDiaNoche++;
    if(stat[i]->Id == 140 || stat[i]->Id == 185) nMoulinRouge++;
    // counting "windows" (T1 stations recovered by global trigger)
    if(stat[i]->Trigger.Window) nWindow++;
    //X[i] = stat[i]->Easting - EASTING_ORIGIN;
    //Y[i] = stat[i]->Northing - NORTHING_ORIGIN;
    //Z[i] = stat[i]->Altitude - ALTITUDE_ORIGIN;
    X[i] = stat[i]->fX;
    Y[i] = stat[i]->fY;
    Z[i] = stat[i]->fZ;
    Tclean10[i] = Tclean50[i] = Tclean90[i] = 0;
    TCorr[i] = 0;

    // Saturation flag: 1000 * Nlow + Nhigh
    // Nlow (Nhigh) = number of slots at 1023 in low (high) gain
    Satur1[i] = Satur2[i] = Satur3[i] = 0 ;
    for(unsigned int j=0; j<kIoSd::MAXSAMPLE; j++) {
      if(stat[i]->Fadc->Trace[0][1][j] >= 1023) Satur1[i]++;
      if(stat[i]->Fadc->Trace[0][0][j] >= 1023) Satur1[i] += 1000;
      if(stat[i]->Fadc->Trace[1][1][j] >= 1023) Satur2[i]++;
      if(stat[i]->Fadc->Trace[1][0][j] >= 1023) Satur2[i] += 1000;
      if(stat[i]->Fadc->Trace[2][1][j] >= 1023) Satur3[i]++;
      if(stat[i]->Fadc->Trace[2][0][j] >= 1023) Satur3[i] += 1000;
    }
    AsymPMT[i] = 0;
    nMuPeak[i] = 0;
  }
  SetTime(TIME_START);

  return 1;
}

/*-------------------------------------------------------------------*/
int TPBRec::ContainedEvent()
/*-------------------------------------------------------------------*/
{
  // counting awake stations around the first one (highest integrated signal)
  nNeighboursOfMax = 0;

  for(unsigned int i=0; i<fErEvent->Trigger.NumberOfStation; i++) {
    IoSdStation *st = &(fErEvent->Stations[i]);
    if(RemoveEA && st->Id<100) continue;
    if(RemoveInfillStation && (st->Id==36 || st->Id==67)) continue;
    if(RemoveDoublet && (st->Id==185 || st->Id==186)) continue;
    nNeighboursOfMax +=
      st->Id!=stat[0]->Id && (st->Error==IoSdEvent::eNoError || st->Error==IoSdEvent::eT3NotFound)
      && sqr(st->Easting-stat[0]->Easting)+sqr(st->Northing-stat[0]->Northing)<sqr(1800);
  }
  if(nActiveStation==3) return nNeighboursOfMax>=6;
  else                  return nNeighboursOfMax>=NMIN_AROUND_HIGHEST_SIGNAL;
}

#define DTMAX_SD 60000
#define DTMAX_FD 240000
/*-------------------------------------------------------------------*/
int TPBRec::AnomalyEvent()
/*-------------------------------------------------------------------*/
{
  int anomaly_level = 0;
  unsigned int tmax=0, tmin=1000000000;
  unsigned int size = fErEvent->fSelectedStations.size();
  //  search for anomalies in time differences between stations
  for(unsigned int i=0; i<size; i++) {
    tmin = minval(tmin,(unsigned int)fErEvent->fSelectedStations[i]->Gps->NanoSecond);
    tmax = maxval(tmax,(unsigned int)fErEvent->fSelectedStations[i]->Gps->NanoSecond);
  }
  if((strncmp(fErEvent->Trigger.Algo.c_str(),"FD",2) && tmax-tmin > DTMAX_SD) || tmax-tmin > DTMAX_FD) {
    if(PrintLevel) {
      printf("!!!! event %llu %s  %s\n",fErEvent->AugerId,fErEvent->UTCDate.c_str(),fErEvent->Trigger.Algo.c_str());
      for(unsigned int i=0; i<fErEvent->NumberOfErrorZeroStation; i++)
        printf("    %3d %2d %3d  %9d\n",fErEvent->fSelectedStations[i]->Id,fErEvent->fSelectedStations[i]->Trigger.Type,
               fErEvent->fSelectedStations[i]->Trigger.Window,fErEvent->fSelectedStations[i]->Gps->NanoSecond);
    }
    if(Diagno_output) fprintf(diag,"event %llu %s  %s : time anomaly\n",
                                fErEvent->AugerId,fErEvent->UTCDate.c_str(),
                                fErEvent->Trigger.Algo.c_str()
                               );
    anomaly_level++;
  }
  return anomaly_level;
}

/*-------------------------------------------------------------------*/
void TPBRec::SilentStations()
/*-------------------------------------------------------------------*/
{
  nSilentStation = 0;
  unsigned int size = fErEvent->fSilentStations.size();
  for(unsigned int i=0; i<size; i++) {
    TCalibStation * st = fErEvent->fSilentStations[i];
    int ids = st->Id;
    if(ids<=0) continue;
    if(RemoveInfillStation && (ids==36 || ids==67)) continue;
    if(RemoveDoublet && (ids==64 || ids==185 || ids==186)) continue;
    //XSilent[nSilentStation] = st->Easting - EASTING_ORIGIN;
    //YSilent[nSilentStation] = st->Northing - NORTHING_ORIGIN;
    //ZSilent[nSilentStation] = st->Altitude - ALTITUDE_ORIGIN;
    XSilent[nSilentStation] = st->fX;
    YSilent[nSilentStation] = st->fY;
    ZSilent[nSilentStation] = st->fZ;
    int ix = (int)((st->Easting-MXID_EASTING_ORIGIN)/MXID_EASTING_STEP);
    int iy = (int)((st->Northing-MXID_NORTHING_ORIGIN)/MXID_NORTHING_STEP);
    mxIdSilent[nSilentStation] = ix + 100*iy;
    nSilentStation++;
  }
}

/*-------------------------------------------------------------------*/
void TPBRec::WriteDisplay()
/*-------------------------------------------------------------------*/
{
  int nstat = 0;
  if(Simulation)
    for(unsigned int i=0; i<fErEvent->Trigger.NumberOfStation; i++)
      nstat += (fErEvent->Stations[i].Error==IoSdEvent::eNoError);
  else nstat = fErEvent->Trigger.NumberOfStation;
  fprintf(disp,"%u %u %u %d %6.0f %6.4f %6.4f\n",
          NumEvent,nstat,nActiveStation,PlanFitOK,t0Plan,uPlan,vPlan);

  for(unsigned int i=0; i<fErEvent->Trigger.NumberOfStation; i++) {
    if(Simulation) {
      if(fErEvent->Stations[i].Error==IoSdEvent::eNoError) fprintf(disp,"%u %u %6.0f %6.0f\n",
            fErEvent->Stations[i].Id,fErEvent->Stations[i].Error,
            fErEvent->Stations[i].Easting - fErEvent->Stations[0].Easting,
            fErEvent->Stations[i].Northing - fErEvent->Stations[0].Northing);
    } else
      fprintf(disp,"%u %u %6.0f %6.0f\n",
              fErEvent->Stations[i].Id,fErEvent->Stations[i].Error,
              fErEvent->Stations[i].Easting - EASTING_ORIGIN,
              fErEvent->Stations[i].Northing - NORTHING_ORIGIN);
  }
  for(int i=0; i<nActiveStation; i++) {
    if(Simulation)
      fprintf(disp,"%u %6.0f %6.0f %6.0f %6.0f %d %d %d\n",
              stat[i]->Id, X[i]-X[0], Y[i]-Y[0], Z[i], T[i]-T[0],
              stat[i]->Gps->NanoSecond-stat[0]->Gps->NanoSecond, stat[i]->Trigger.Window,
              InFit[i]);
    else
      fprintf(disp,"%u %6.0f %6.0f %6.0f %6.0f %d %d %d\n",
              stat[i]->Id, X[i], Y[i], Z[i], T[i]-T[0],
              stat[i]->Gps->NanoSecond-stat[0]->Gps->NanoSecond, stat[i]->Trigger.Window,       InFit[i]);
    for(int k=0; k<3; k++) {
      for(unsigned int j=0; j<kIoSd::MAXSAMPLE; j++)
        fprintf(disp,"%d ",(int)(100*stat[i]->fPmt[k].fTrace[j]));
      fprintf(disp,"\n");
    }
  }
}

// colum-wise printout of the characteristics of the stations
/*-------------------------------------------------------------------*/
void TPBRec::PrintStations()
/*-------------------------------------------------------------------*/
{
  printf("Id    ");
  for(int i=0; i<nActiveStation; i++) printf("%6d ",stat[i]->Id);
  printf("\nMxId  ");
  for(int i=0; i<nActiveStation; i++) printf("%6d ",mxId[i]);
  printf("\nwindow");
  for(int i=0; i<nActiveStation; i++)
    printf("   %3d ",stat[i]->Trigger.Window);
  printf("\ntrig ");
  for(int i=0; i<nActiveStation; i++)
    printf("  %2d/%2d",(stat[i]->Trigger.Type)&255,(stat[i]->Trigger.Type)/256);
  printf("\nsatur ");
  for(int i=0; i<nActiveStation; i++) {
    int sat1 = Satur1[i]%1000 > 0; if(Satur1[i] > 1000) sat1=2;
    int sat2 = Satur2[i]%1000 > 0; if(Satur2[i] > 1000) sat2=2;
    int sat3 = Satur3[i]%1000 > 0; if(Satur3[i] > 1000) sat3=2;
    printf(" %d/%d/%d ",sat1,sat2,sat3);
  }
  printf("\n  X   ");
  for(int i=0; i<nActiveStation; i++) printf("%6.0f ",X[i]);
  printf("\n  Y   ");
  for(int i=0; i<nActiveStation; i++) printf("%6.0f ",Y[i]);
  printf("\n  Z   ");
  for(int i=0; i<nActiveStation; i++) printf("%6.0f ",Z[i]);
  printf("\n Time ");
  for(int i=0; i<nActiveStation; i++) printf("%6.0f ",T[i]-T[0]);
  printf("\n strt ");
  for(int i=0; i<nActiveStation; i++) printf("%6d ",CleanStart[i]);
  printf("\n stop ");
  for(int i=0; i<nActiveStation; i++) printf("%6d ",CleanStop[i]);
  printf("\n VEM  ");
  for(int i=0; i<nActiveStation; i++) printf("%6.2f ",Signal[i]);
  printf("\n S/P  ");
  for(int i=0; i<nActiveStation; i++) printf("%6.2f ",Signal[i]/Peak[i]);

  printf("\n");
}

/*-------------------------------------------------------------------*/
int TPBRec::Anal3st(int is1, int is2,int is3)/// Analysis of 3-fold events
/*-------------------------------------------------------------------*/
{
  double deter,u,v,sinth,costh,t0,xc,dxc,yc,dyc,s1000,ds1000,beta,dbeta,gamma,chi2,chi2w;
  double res,res_t[NSTATMAX],cov[6],covGlob[6][6],distc[NSTATMAX];
  int errLDF, errGlob, errflag;

  //--- Fit with stations is1,is2,is3 in the list of active stations
  //    (should be the only ones with InFit=1)

  PlanFitOK = Fit3stOK = GlobFitOK = 0;
  errLDF = 0;

  if(stat[is1]->Id==stat[is2]->Id || stat[is1]->Id==stat[is3]->Id || stat[is2]->Id==stat[is3]->Id ) {
    printf("!!! %d %d %d not a triplet\n",is1,is2,is3);
    return 0;
  }

  nStationInFit = 3;
  for(int is=0; is<nActiveStation; is++) InFit[is] = 0;
  InFit[is1] = InFit[is2] = InFit[is3] = 1;

  ComputeBarycenter();

  //--- selection of events
  // shape of signals : at least 2 stations with ToT trigger or with "wide" signals
  int nstsel=0;
  double ratio[3];
  double ratio_avg = 0;
  ratio[0] = Signal[is1]/Peak[is1];
  ratio[1] = Signal[is2]/Peak[is2];
  ratio[2] = Signal[is3]/Peak[is3];
  nstsel = (stat[is1]->IsTOT() || ratio[0]>ANA3_RATMIN)
           + (stat[is2]->IsTOT() || ratio[1]>ANA3_RATMIN)
           + (stat[is3]->IsTOT() || ratio[2]>ANA3_RATMIN);
  ratio_avg = (ratio[0]+ratio[1]+ratio[2])/3;

  // area and maximal distance
  deter = X[is1]*Y[is2]-Y[is1]*X[is2]
          +X[is2]*Y[is3]-Y[is2]*X[is3]
          +X[is3]*Y[is1]-Y[is3]*X[is1];
  double dist2 = sqr(X[is1]-X[is2]) + sqr(Y[is1]-Y[is2]);
  dist2 = maxval(dist2, sqr(X[is2]-X[is3])+sqr(Y[is2]-Y[is3]));
  dist2 = maxval(dist2, sqr(X[is3]-X[is1])+sqr(Y[is3]-Y[is1]));

  //--- computation of direction
  double dt1 = T[is2]-T[is1];
  double dt2 = T[is3]-T[is1];
  u = kCdasUtil::CSPEED*((Y[is1]-Y[is3])*dt1-(Y[is1]-Y[is2])*dt2)/deter;
  v = kCdasUtil::CSPEED*((X[is3]-X[is1])*dt1-(X[is2]-X[is1])*dt2)/deter;

  if(nstsel<2) {
    if(PrintLevel>1) printf("Anal3st %d %d %d : reject (nstsel = %d)\n",
                              stat[is1]->Id,stat[is2]->Id,stat[is3]->Id,nstsel);
    nStationInFit = 0; InFit[is1] = InFit[is2] = InFit[is3] = 0;
    return 0;
  }

  // geometrical extension : cut on area of the triangle
  // non activated option: modulation by the signal/peak ratio, averaged over the stations
  //if(fabs(deter/2) > ratio_avg*ANA3_AREAMAX || fabs(deter/2) < ANA3_AREAMIN ) {   //opt_modul

  if(fabs(deter/2) > ANA3_AREAMAX || fabs(deter/2) < ANA3_AREAMIN ) {
    if(PrintLevel>2)
      printf("Anal3st %d %d %d : reject (area = %7.0f)\n",
             stat[is1]->Id,stat[is2]->Id,stat[is3]->Id,fabs(deter/2));
    nStationInFit = 0; InFit[is1] = InFit[is2] = InFit[is3] = 0;
    return 0;
  }
  // cut on distance between stations (modulated as before)
  //if(dist2 > sqr(ratio_avg*ANA3_DISTMAX)) {   \\opt_modul
  if(dist2 > sqr(ANA3_DISTMAX)) {
    if(PrintLevel>1) printf("Anal3st %d %d %d : reject (dist = %4.0f)\n",
                              stat[is1]->Id,stat[is2]->Id,stat[is3]->Id,sqrt(dist2));
    nStationInFit = 0; InFit[is1] = InFit[is2] = InFit[is3] = 0;
    return 0;
  }

  // refit with corrections and get error matrix
  CorrectTimeAltitude(u,v);
  FitPlan(1,1,t0Plan,u,v,res_t,res,chi2,cov,errflag);

  // refit after front curvature corrections on times (depending on theta)
  CorrectTimeAltitudeCurvature(u,v,0,0);
  FitPlan(1,1,t0Plan,u,v,res_t,res,chi2,cov,errflag);
  // iterate
  CorrectTimeAltitudeCurvature(u,v,0,0);
  FitPlan(1,1,t0Plan,u,v,res_t,res,chi2,cov,errflag);

  if(u*u+v*v<1) {
    PlanFitOK = 1;
    uCurrent = u; vCurrent = v;
    sinth = sqrt(u*u+v*v);
    costh = sqrt(1-sinth*sinth);
    Theta = asin(sinth);
    dTheta = sqrt(u*u*cov[2] + v*v*cov[5] + 2*u*v*cov[4])/sinth/costh;
    Phi = atan2(v,u);
    dPhi = sqrt(v*v*cov[2] + u*u*cov[5] - 2*u*v*cov[4])/sinth/sinth;
    Curv = SELECT_CURV0*costh;
    dCurv = 0;
    fT0 = t0Plan;
    fU = uCurrent;
    fV = vCurrent;
    fR = 0;
  }
  //--- fit of LDF if acceptable direction
  if (u*u+v*v>sqr(LDF_SINTHMAX)) {
    if(PrintLevel>1) printf("Anal3st %d %d %d : no LDF fit (sinth = %4.3f)\n",
                              stat[is1]->Id,stat[is2]->Id,stat[is3]->Id,sqrt(u*u+v*v));
    nStationInFit = 0; InFit[is1] = InFit[is2] = InFit[is3] = 0;
    return 0;
  }

  if(PrintLevel>1) {
    printf("Trying LDF fit with stations %d %d %d  th %5.2f  ph %7.2f  bary %4.0f %4.0f\n",
           stat[is1]->Id,stat[is2]->Id,stat[is3]->Id,Theta*180/M_PI,Phi*180/M_PI,XBarycenter,YBarycenter);
  }
  //--- fit on the 3 active stations
  // evaluation of the parameters of the LDF from zenith angle
  costh = sqrt(1-uCurrent*uCurrent-vCurrent*vCurrent);
  beta = 3 + .7*atan(6*(costh-.65));
  gamma = .5 - .05*sin(8*(costh-.6));

  //--- First approximation of Signal(1000) from third station
  double sigc = Signal[is3];
  double dx = X[is3] - XBarycenter;
  double dy = Y[is3] - YBarycenter;
  double dist = sqrt(sqr(dx)+sqr(dy)-sqr(uCurrent*dx + vCurrent*dy));
  s1000 = sigc*pow(dist/LDF_DISTREF,beta);
  xc = yc = 0;
  FitLDFSlope = 0;
  SilentLDF = 0;

  FitLDFMinuit(xc, dxc, yc, dyc, s1000, ds1000, distc, chi2, beta, dbeta, gamma, errflag);
  if(errflag>0) {
    errLDF++;
    if(PrintLevel>1) printf("Anal3st %d %d %d : LDF errflag>0\n",
                              stat[is1]->Id,stat[is2]->Id,stat[is3]->Id);
    return 0;
  }
  //--- quality of the fit
  // distance (at ground) from barycenter to fitted core
  if(xc*xc+yc*yc > sqr(ANA3_DISTMAX_BARY_CORE)) {
    if(PrintLevel>1) printf("Anal3st %d %d %d : reject (dist. core-bary = %4.0f)\n",
                              stat[is1]->Id,stat[is2]->Id,stat[is3]->Id,sqrt(xc*xc+yc*yc));
    //nStationInFit = 0; InFit[is1] = InFit[is2] = InFit[is3] = 0;
    errLDF += 10;
    return 0;
  }
  // chi2 of the LDF fit
  if(chi2 > ANA3_CHI2_LDF_MAX) {
    if(PrintLevel>2) printf("Anal3st %d %d %d : reject (LDF chi2 = %6.2f)\n",
                              stat[is1]->Id,stat[is2]->Id,stat[is3]->Id,chi2);
    errLDF += 100;
    return 0;
  }
  Fit3stOK += 1;
  SaveLDFParam(xc, dxc, yc, dyc, s1000, ds1000, distc, beta, dbeta, gamma);
  chi2LDF = chi2;
  if(PrintLevel>1)
    printf("Anal3st %d %d %d  w/o silent : xc yc %6.0f %6.0f  s1000 %6.2f  energy %6.2f  chi2 %6.4f\n",
           stat[is1]->Id,stat[is2]->Id,stat[is3]->Id, xc, yc, s1000, fE,chi2);

  // fit including the silent stations
  // search for silent stations within LDF_DIST_SILENT_MAX from estimated axis
  for(int i=0; i<nSilentStation; i++) {
    double dx = XSilent[i] - XBarycenter;
    double dy = YSilent[i] - YBarycenter;
    double dist = sqrt(sqr(dx)+sqr(dy)-sqr(uCurrent*dx + vCurrent*dy));
    SilentInFit[i] = (!Simulation || dist > LDF_DISTREF) && dist < LDF_DISTMAX_SILENT;
  }
  SilentLDF = 1;
  FitLDFMinuit(xc, dxc, yc, dyc, s1000, ds1000, distc, chi2w, beta, dbeta, gamma, errflag);
  errLDF = errLDF || (errflag>0);

  // final cuts
  if(errLDF || chi2w > ANA3_CHI2_LDF_SILENT_MAX) {
    if(PrintLevel>2)
      printf("Anal3st %d %d %d with silent : reject (errflag = %d  chi2 = %6.2f)\n",
             stat[is1]->Id,stat[is2]->Id,stat[is3]->Id,errflag,chi2w);
    //nStationInFit = 0; for (int i=0; i<nStationInFit; i++) InFit[i]=0;
    return 0;
  }
  Fit3stOK += 100;
  SaveLDFParam(xc, dxc, yc, dyc, s1000, ds1000, distc, beta, dbeta, gamma);
  chi2LDFwithSilent = chi2w;
  if(PrintLevel>2)
    printf("Anal3st %d %d %d  with silent : xc yc %6.0f %6.0f  s1000 %6.2f  energy %6.2f  chi2 %6.4f\n",
           stat[is1]->Id,stat[is2]->Id,stat[is3]->Id, xc, yc, s1000,fE,chi2w);

  //--- Global fit of angles, core position and energy (with fixed curvature)
  t0 = t0Plan;
  u = uCurrent;
  v = vCurrent;
  xc = 0;
  yc = 0;

  // Fit with active stations + silent
  SilentLDF = 1;
  errGlob = 0;
  FitGlobFixCurvMinuit(t0, u, v, xc, yc, s1000, distc, covGlob[0], chi2, errflag);
  if(errflag>0) {
    errGlob ++;
    if(PrintLevel>1) printf("Anal3st Glob : errflag>0\n");
    return 0;
  }
  if(errflag) chi2Glob = chi2;
  if(chi2Glob > ANA3_CHI2_LDF_SILENT_MAX) {
    errGlob += 10;
    if(PrintLevel>1) printf("Anal3st Glob : chi2 %6.2f > %6.2f\n",chi2Glob,ANA3_CHI2_LDF_MAX);
    return 0;
  }
  if(sqrt(u*u+v*v) > LDF_SINTHMAX) {
    errGlob += 100;
    if(PrintLevel>1) printf("Anal3st Glob : sinth %5.3f\n",sqrt(u*u+v*v));
    return 0;
  }
  u3st = u; v3st = v;
  sinth = sqrt(u*u+v*v);
  costh = sqrt(1-sinth*sinth);
  Theta = asin(sinth);
  dTheta = sqrt(u*u*covGlob[1][1] + v*v*covGlob[2][2] + 2*u*v*covGlob[1][2])/sinth/costh;
  Phi = atan2(v,u);
  dPhi = sqrt(v*v*covGlob[1][1] + u*u*covGlob[2][2] - 2*u*v*covGlob[1][2])/sinth/sinth;
  GlobFitOK += 1;
  SaveLDFParam(xc, dxc, yc, dyc, s1000, ds1000, distc, beta, dbeta, gamma);
  fT0 = t0;
  fU = u3st;
  fV = v3st;
  fR = 0;

  return 1;
}

/*-------------------------------------------------------------------*/
int TPBRec::Select()/// Selection of "general" events
/*-------------------------------------------------------------------*/
{
  int is,j, ord[NSTATMAX];

  // initialize the configuration
  SetTime(TIME_START_SELECT);
  for(is=0; is<nActiveStation; is++) {
    InFit[is] = 1;
    ord[is] = is;
  }

  // Sort the stations by increasing "quality"
  // (combination : amplitude * signal/peak)
  for(is=1; is<nActiveStation; is++)
    for (j=is; j>0; j--)
      if(sqr(Signal[ord[j-1]])/Peak[ord[j-1]] >
          sqr(Signal[ord[j]])/Peak[ord[j]]) {
        int temp = ord[j]; ord[j] = ord[j-1]; ord[j-1] = temp;
      }

  // barycenter of 'InFit' stations
  ComputeBarycenter();

  // first trial with all stations
  PlanFitOK = TryWithAllStations();

  //--- if rejected with more than 4 stations, try removing one
  //    (by decreasing "quality")
  if(!PlanFitOK && nActiveStation>4) PlanFitOK = TryRemoveOneStation(ord);

  //--- if rejected with more than 5 stations, retry removing two
  if(!PlanFitOK && nActiveStation>5) PlanFitOK = TryRemoveTwoStations(ord);

  //--- if rejected with more than 6 stations, retry removing three
  if(!PlanFitOK && nActiveStation>6) PlanFitOK = TryRemoveThreeStations(ord);

  //--- if rejected with more than 7 stations, retry removing four
  if(!PlanFitOK && nActiveStation>7) PlanFitOK = TryRemoveFourStations(ord);

  //--- if rejected with more than 8 stations, retry removing four
  if(!PlanFitOK && nActiveStation>8) PlanFitOK = TryRemoveFiveStations(ord);

  //--- if rejected with more than 9 stations, retry removing four
  if(!PlanFitOK && nActiveStation>9) PlanFitOK = TryRemoveSixStations(ord);

  //--- if rejected with more than 10 stations, retry removing four
  if(!PlanFitOK && nActiveStation>10) PlanFitOK = TryRemoveSevenStations(ord);

  //--- if rejected with more than 11 stations, retry removing four
  if(!PlanFitOK && nActiveStation>11) PlanFitOK = TryRemoveEightStations(ord);

  if(PlanFitOK) {
    nStationInFit = 0;
    for(is=0; is<nActiveStation; is++) {
      nStationInFit += InFit[is];
      if(InFit[is]==0)
        for(unsigned int i=0; i<fErEvent->fSelectedStations.size(); i++)
          if(fErEvent->fSelectedStations[i]->Id==stat[is]->Id) {
            fErEvent->fRandomStations.push_back(fErEvent->fSelectedStations[i]);
            fErEvent->fSelectedStations.erase(fErEvent->fSelectedStations.begin()+i);
            break;
          }
    }
  } else {
    if(PrintLevel>1) printf("PlanFit : rejected\n");
    nStationInFit = 0;
    for(is=0; is<nActiveStation; is++) InFit[is]=0;
  }

  return PlanFitOK;
}

/*-------------------------------------------------------------------*/
int TPBRec::TryWithAllStations()
/*-------------------------------------------------------------------*/
{
  int errflag;
  double t0,u,v,res,max_res,chi2,res_t[NSTATMAX],cov[10];
  double tolr,tolrm,tolres,tolresm;

  tolr = nActiveStation*SELECT_TOLRES;
  tolrm = nActiveStation*SELECT_TOLRESM;

  // fit without corrections
  FitPlan(0,0,t0,u,v,res_t,res,chi2,cov,errflag);

  // refit after corrections on times (depending on altitude and direction)
  CorrectTimeAltitude(u,v);
  FitPlan(1,0,t0,u,v,res_t,res,chi2,cov,errflag);

  // refit after front curvature corrections on times (depending on theta)
  CorrectTimeAltitudeCurvature(u,v,0,0);
  FitPlan(1,0,t0,u,v,res_t,res,chi2,cov,errflag);

  // compute the largest residual
  max_res = 0;
  for(int is=0; is<nActiveStation; is++) if(InFit[is])
      max_res = maxval(max_res,fabs(res_t[is]));

  // criteria to accept the configuration
  tolres  = tolr *maxval(.2,sqrt(maxval(0,1-u*u-v*v)));
  tolresm = tolrm*maxval(.2,sqrt(maxval(0,1-u*u-v*v)));
  RejectCode = 100*(res>tolres || max_res>tolresm) + 10*(!compact_config(u,v))
               + (u*u+v*v>1);

  if(!errflag && !RejectCode) return 1;

  if(PrintLevel>1) {
    printf("initial configuration rejected :");
    if(!compact_config(u,v)) printf("  not compact");
    if(errflag) printf("  errflag");
    if(res>=tolres) printf("  resid %3.0f > %3.0f",res,tolres);
    if(max_res>=tolresm) printf("  resmax %3.0f > %3.0f",max_res,tolresm);
    if(u*u+v*v>=1) printf("  u2+v2 %6.4f",u*u+v*v);
    printf("\n");
  }
  return 0;
}

/*-------------------------------------------------------------------*/
int TPBRec::_TryRemove1Station(int *ord)
/*-------------------------------------------------------------------*/
{
  int is,errflag;
  double t0,u,v,res,max_res,chi2,res_t[NSTATMAX],cov[10];
  double tolr,tolrm,tolres,tolresm;

  tolr = (nActiveStation-_indexes.size())*SELECT_TOLRES;
  tolrm = (nActiveStation-_indexes.size())*SELECT_TOLRESM;

  // loop over all stations according to 'ord'
  for(is=0; is<_indexes[0]; is++) if(InFit[ord[is]]) {
      InFit[ord[is]] = 0;
      // try to refit if the configuration fulfills the "compacity" criterion
      ComputeBarycenter();
      FitPlan(0,0,t0,u,v,res_t,res,chi2,cov,errflag);
      if(compact_config(u,v)) {
        CorrectTimeAltitude(u,v);
        FitPlan(1,0,t0,u,v,res_t,res,chi2,cov,errflag);
        CorrectTimeAltitudeCurvature(u,v,0,0);
        FitPlan(1,0,t0,u,v,res_t,res,chi2,cov,errflag);
        max_res = 0;
        for(int ist=0; ist<nActiveStation; ist++) if(InFit[ist])
            max_res = maxval(max_res,fabs(res_t[ist]));
        tolres  = tolr *maxval(.2,sqrt(maxval(0,1-u*u-v*v)));
        tolresm = tolrm*maxval(.2,sqrt(maxval(0,1-u*u-v*v)));
        if(!errflag && res<tolres && max_res<tolresm && u*u+v*v<1) {
          if(PrintLevel>1) {
            cout<<"acceptable removing "<<stat[ord[is]]->Id;
            if(_indexes.size()>1) for (unsigned int i=0; i<_indexes.size()-1; i++)
                cout<<" "<<stat[_indexes[i]]->Id;
            cout<<endl;
          }
          return 1;
        } else if(PrintLevel>2) {
          cout<<"  without "<<stat[ord[is]]->Id;
          if(_indexes.size()>1) for (unsigned int i=0; i<_indexes.size(); i++)
              cout<<" "<<stat[_indexes[i]]->Id;
          cout<<endl;
          if(errflag) printf("  errflag");
          if(res>=tolres) printf("  resid %6.2f",res);
          if(max_res>=tolresm) printf("  resmax %6.2f",max_res);
          if(u*u+v*v>1) printf("  u2+v2 %5.3f",u*u+v*v);
          printf("\n");
        }
      } else if(PrintLevel>2) {
        cout<<"  without "<<stat[ord[is]]->Id;
        if(_indexes.size()>1) for (unsigned int i=0; i<_indexes.size(); i++)
            cout<<" "<<stat[_indexes[i]]->Id;
        cout<<endl;
      }
      InFit[ord[is]] = 1;
    }
  return 0;

}

/*-------------------------------------------------------------------*/
int TPBRec::_TryRemove2Stations(int *ord)
/*-------------------------------------------------------------------*/
{
  for(int is=1; is<_indexes[1] && !PlanFitOK; is++) if(InFit[ord[is]]) {
      InFit[ord[is]] = 0;
      _indexes[0] = is;
      if(_TryRemove1Station(ord)) return 1;
      InFit[ord[is]] = 1;
    }
  return 0;
}

/*-------------------------------------------------------------------*/
int TPBRec::_TryRemove3Stations(int *ord)
/*-------------------------------------------------------------------*/
{
  for(int is=2; is<_indexes[2] && !PlanFitOK; is++) if(InFit[ord[is]]) {
      InFit[ord[is]] = 0;
      _indexes[1] = is;
      if(_TryRemove2Stations(ord)) return 1;
      InFit[ord[is]] = 1;
    }
  return 0;
}

/*-------------------------------------------------------------------*/
int TPBRec::_TryRemove4Stations(int *ord)
/*-------------------------------------------------------------------*/
{
  for(int is=3; is<_indexes[3] && !PlanFitOK; is++) if(InFit[ord[is]]) {
      InFit[ord[is]] = 0;
      _indexes[2] = is;
      if(_TryRemove3Stations(ord)) return 1;
      InFit[ord[is]] = 1;
    }
  return 0;
}

/*-------------------------------------------------------------------*/
int TPBRec::_TryRemove5Stations(int *ord)
/*-------------------------------------------------------------------*/
{
  for(int is=4; is<_indexes[4] && !PlanFitOK; is++) if(InFit[ord[is]]) {
      InFit[ord[is]] = 0;
      _indexes[3] = is;
      if(_TryRemove4Stations(ord)) return 1;
      InFit[ord[is]] = 1;
    }
  return 0;
}

/*-------------------------------------------------------------------*/
int TPBRec::_TryRemove6Stations(int *ord)
/*-------------------------------------------------------------------*/
{
  for(int is=5; is<_indexes[5] && !PlanFitOK; is++) if(InFit[ord[is]]) {
      InFit[ord[is]] = 0;
      _indexes[4] = is;
      if(_TryRemove5Stations(ord)) return 1;
      InFit[ord[is]] = 1;
    }
  return 0;
}

/*-------------------------------------------------------------------*/
int TPBRec::_TryRemove7Stations(int *ord)
/*-------------------------------------------------------------------*/
{
  for(int is=6; is<_indexes[6] && !PlanFitOK; is++) if(InFit[ord[is]]) {
      InFit[ord[is]] = 0;
      _indexes[5] = is;
      if(_TryRemove6Stations(ord)) return 1;
      InFit[ord[is]] = 1;
    }
  return 0;
}

/*-------------------------------------------------------------------*/
int TPBRec::_TryRemove8Stations(int *ord)
/*-------------------------------------------------------------------*/
{
  for(int is=7; is<_indexes[7] && !PlanFitOK; is++) if(InFit[ord[is]]) {
      InFit[ord[is]] = 0;
      _indexes[6] = is;
      if(_TryRemove7Stations(ord)) return 1;
      InFit[ord[is]] = 1;
    }
  return 0;
}

// Analysis of "general" events
/*-------------------------------------------------------------------*/
int TPBRec::AnalGen()
/*-------------------------------------------------------------------*/
{
  int is;

  // fit of the direction (and curvature)

  AngularFit();

  if(PrintLevel>1) {
    printf(" InFit");
    for(is=0; is<nActiveStation; is++) printf("%6d ",InFit[is]); printf("\n");
    printf(" dBary");
    for(is=0; is<nActiveStation; is++) printf("%6.0f ",distBary[is]); printf("\n");
    printf(" resP ");
    for(int i=0; i<nActiveStation; i++) printf("%6.0f ",resTPlan[i]); printf("\n");
    printf("--- PlanFit %d  resid %6.2f  chi2 %4.2f  u %6.4f+-%6.4f  v %6.4f+-%6.4f  ",
           PlanFitOK,residPlan,chi2Plan,uPlan,sqrt(covPlan[2]),vPlan,sqrt(covPlan[5]));
    printf("theta %6.2f  phi %6.2f\n",
           asin(sqrt(uPlan*uPlan+vPlan*vPlan))*180/M_PI,atan2(vPlan,uPlan)*180/M_PI);
    if(PlanFitOK && nPositionInFit>4) {
      printf(" resC ");
      for(int i=0; i<nActiveStation; i++) printf("%6.0f ",resTCurv[i]); printf("\n");
      printf("--- CurvFit %d  resid %6.2f  chi2 %4.2f  u %6.4f+-%6.4f  v %6.4f+-%6.4f  ",
             CurvFitOK,residCurv,chi2Curv,uCurv,sqrt(covCurv[2]),vCurv,sqrt(covCurv[5]));
      printf("theta %6.2f  phi %6.2f  ",
             asin(sqrt(uCurv*uCurv+vCurv*vCurv))*180/M_PI,atan2(vCurv,uCurv)*180/M_PI);
      printf("1/R(km) %6.4f+-%6.4f\n",1000*Curv,1000*sqrt(covCurv[9]));
    }
  }

  LDFFitOK = 0;
  // special for horizontal event: "normal" procedure followed by specific one
  // old param: "horizontal event" means "there a table for it"
  // if(FitHoriz && (HorizFit=ReadTableHoriz(NumEvent,HorizModel)))
  // new param
  if(FitHoriz && ((PlanFitOK && nPositionInFit<=4) || (CurvFitOK && nPositionInFit>4)) &&
      sqr(uCurrent)+sqr(vCurrent)>sqr(LDF_SINTHMAX) && sqr(uCurrent)+sqr(vCurrent)<.998) {
    double ener1=0, ener2=0;
    double xc1=0, yc1=0, xc2=0, yc2=0;
    LDFFit();
    if(LDFFitOK) {
      ener1 = fE; xc1=fXCore; yc1=fYCore;
    }
    LDFhorFit();
    if(LDFFitOK) {
      ener2 = fE; xc2=fXCore; yc2=fYCore;
    }
  }
  // standard reconstruction of core and S1000, after cuts
  else {
    if(((nPositionInFit>4 && CurvFitOK) || (nPositionInFit<=4 && PlanFitOK))
        && sqr(uCurrent)+sqr(vCurrent)<sqr(LDF_SINTHMAX))
      LDFFit();

    // combined reconstruction
    if(LDFFitOK && sqr(uCurrent)+sqr(vCurrent)<sqr(LDF_SINTHMAX)) {
      // recomputing the direction using the fitted core position
      if(CurvFitOK) {
        double uu = fR*uCurrent+XBarycenter-fXCore;
        double vv = fR*vCurrent+YBarycenter-fYCore;
        double ww = fR*sqrt(1-uCurrent*uCurrent-vCurrent*vCurrent);
        if(ww<0) {
          uu = -uu; vv = -vv; ww = -ww;
        }
        Theta = atan(sqrt(uu*uu+vv*vv)/ww);
        Phi = atan2(vv,uu);
      }
      // global fit of direction, core and S1000
      SetTime(TIME_START_GLOB);
      ComputeBarycenter();
      GlobalFit();
    }
  }

  return 1;
}

// main function for geometrical reconstruction of the front
/*-------------------------------------------------------------------*/
int TPBRec::AngularFit()
/*-------------------------------------------------------------------*/
{
  int i,is,errflag;
  double u,v,costh,sinth;

  // refit a front plane with tuned errors on time  ------------------------
  if(PlanFitOK) {
    SetTime(TIME_START_PLAN);
    // new barycenter
    ComputeBarycenter();
    // fit without corrections
    FitPlan(0,0,t0Plan,uPlan,vPlan,resTPlan,residPlan,chi2Plan,covPlan,errflag);
    // refit with corrections (with iteration)
    CorrectTimeAltitudeCurvature(uPlan,vPlan,0,0);
    FitPlan(1,1,t0Plan,uPlan,vPlan,resTPlan,residPlan,chi2Plan,covPlan,errflag);
    CorrectTimeAltitudeCurvature(uPlan,vPlan,0,0);
    FitPlan(1,1,t0Plan,uPlan,vPlan,resTPlan,residPlan,chi2Plan,covPlan,errflag);

    // save the results of plane fit
    fT0 = t0Plan;
    uCurrent = uPlan; vCurrent = vPlan;
    fU = uCurrent; fV = vCurrent;
    sinth = sqrt(sqr(uCurrent)+sqr(vCurrent));
    costh = sqrt(1-sinth*sinth);
    Theta = asin(sinth);
    for(i=0; i<6; i++) covCurrent[i]=covPlan[i];
    for(i=6; i<10; i++) covCurrent[i]=0;
    dTheta = sqrt( sqr(uCurrent)*covCurrent[2] + sqr(vCurrent)*covCurrent[5]
                   + 2*uCurrent*vCurrent*covCurrent[4] )/sinth/costh;
    Phi = atan2(vCurrent,uCurrent);
    dPhi = sqrt( sqr(vCurrent)*covCurrent[2] + sqr(uCurrent)*covCurrent[5]
                 - 2*uCurrent*vCurrent*covCurrent[4] )/sinth/sinth;
    for(is=0; is<nActiveStation; is++) {
      double dx = X[is] - XBarycenter;
      double dy = Y[is] - YBarycenter;
      distBary[is] = sqrt(sqr(dx) + sqr(dy) - sqr(uCurrent*dx + vCurrent*dy));
    }
  }

  // counting Carmen/Miranda in selected stations
  nDoubletInFit = 0;
  int nCM = 0;
  int nMR = 0;
  int nDN = 0;
  for(is=0; is<nActiveStation; is++) if(InFit[is]) {
      if(stat[is]->Id==49 || stat[is]->Id==64) nCM++;
      if(stat[is]->Id==140 || stat[is]->Id==185) nMR++;
      if(stat[is]->Id==139 || stat[is]->Id==186) nDN++;
    }
  nDoubletInFit  = (nCM>1)+(nMR>1)+(nDN>1);
  nPositionInFit = nStationInFit-nDoubletInFit;
  if(PrintLevel>1 && nStationInFit>4 && nPositionInFit<5)
    printf("event with doublet(s), less than 5 positions : no fit with curvature\n");

  //--- Fit with curvature as a free parameter
  //     (if at least 5 stations, counting one for a doublet)

  CurvFitOK = 0;
  if(PlanFitOK) {
    // default value for the curvature
    if(nPositionInFit<5) {
      Curv = SELECT_CURV0*cos(Theta);
      fR = 1/Curv;
    } else {
      SetTime(TIME_START_CURV);
      ComputeBarycenter();
      u = uPlan; v = vPlan;
      // iterations are needed because of the altitude corrections
      // and the evaluation of distance to axis depend on the direction
      FitCurved(0,1,t0Curv,u,v,Curv,resTCurv,residCurv,chi2Curv,covCurv,errflag);
      CorrectTimeAltitude(u,v);
      FitCurved(1,1,t0Curv,u,v,Curv,resTCurv,residCurv,chi2Curv,covCurv,errflag);
      CorrectTimeAltitude(u,v);
      FitCurved(1,1,t0Curv,u,v,Curv,resTCurv,residCurv,chi2Curv,covCurv,errflag);
      if(PrintLevel>1 && errflag) printf("CurvFit :  error !\n");

      // save the results
      fT0 = t0Curv;
      uCurv = u; vCurv = v;
      fU = uCurv; fV = vCurv;
      fR = 1/Curv;
      CurvFitOK = !errflag && residCurv<nStationInFit*ANGREC_TOLRESC
                  && uCurv*uCurv+vCurv*vCurv<1;
      if(CurvFitOK) {
        uCurrent = uCurv; vCurrent = vCurv;
        dCurv = sqrt(covCurv[9]);
        for(i=0; i<10; i++) covCurrent[i]=covCurv[i];
        for(is=0; is<nActiveStation; is++) {
          double dx = X[is] - XBarycenter;
          double dy = Y[is] - YBarycenter;
          distBary[is] = sqrt(sqr(dx) + sqr(dy) - sqr(uCurv*dx + vCurv*dy));
        }
      }
    }
  }
  sinth = sqrt(sqr(uCurrent)+sqr(vCurrent));
  costh = sqrt(1-sinth*sinth);
  Theta = asin(sinth);
  dTheta = sqrt( sqr(uCurrent)*covCurrent[2] + sqr(vCurrent)*covCurrent[5]
                 + 2*uCurrent*vCurrent*covCurrent[4] )/sinth/costh;
  Phi = atan2(vCurrent,uCurrent);
  dPhi = sqrt( sqr(vCurrent)*covCurrent[2] + sqr(uCurrent)*covCurrent[5]
               - 2*uCurrent*vCurrent*covCurrent[4] )/sinth/sinth;

  return 1;
}

/*-------------------------------------------------------------------*/
void TPBRec::ComputeBarycenter()
/*-------------------------------------------------------------------*/
{
  // barycenter in space and time
  double wgt, sumw;
  sumw = XBarycenter = YBarycenter = ZBarycenter = TBarycenter = 0;
  int ns = 0;
  for(int i=0; i<nActiveStation; i++) if(InFit[i]) {
      wgt = pow(Signal[i],BARY_POWER);
      sumw += wgt;
      XBarycenter += wgt*X[i];
      YBarycenter += wgt*Y[i];
      ZBarycenter += wgt*Z[i];
      TBarycenter += wgt*T[i];
      ns++;
      if(ns==NSTAT_IN_BARY) break;
    }
  XBarycenter /= sumw; YBarycenter /= sumw; ZBarycenter /= sumw;
  TBarycenter /= sumw;
  fZCore = ZBarycenter; //for further compatibility

}

/*-------------------------------------------------------------------*/
void TPBRec::CorrectTimeAltitude(double u, double v)
/*-------------------------------------------------------------------*/
{
  double costh = sqrt(maxval(0,1-u*u-v*v));
  for(int i=0; i<nActiveStation; i++) TCorr[i] = costh*(Z[i]-ZBarycenter)/kCdasUtil::CSPEED;
}

/*-------------------------------------------------------------------*/
void TPBRec::CorrectTimeAltitudeCurvature(double u, double v, double xc, double yc)
/*-------------------------------------------------------------------*/
{
  double costh = sqrt(maxval(0,1-u*u-v*v));
  // rough estimation of curvature 1/R
  double curv = SELECT_CURV0*costh;
  for(int i=0; i<nActiveStation; i++) {
    double dx = X[i] - XBarycenter - xc;
    double dy = Y[i] - YBarycenter - yc;
    double dist2 = sqr(dx)+sqr(dy)-sqr(u*dx + v*dy);
    TCorr[i] = (costh*(Z[i]-ZBarycenter) - curv*dist2/2) / kCdasUtil::CSPEED;
  }
}

/*-------------------------------------------------------------------*/
int TPBRec::compact_config(double u, double v)
/*-------------------------------------------------------------------*/
{
  double xb, yb, dx, dy, wgt, sumw, dist2_max_proj, dist2;
  int ns,nCM,nMR,nDN;

  nCM = nMR = nDN = 0;
  for(int i=0; i<nActiveStation; i++) if(InFit[i]) {
      if(stat[i]->Id==49 || stat[i]->Id==64) nCM++;
      if(stat[i]->Id==140 || stat[i]->Id==185) nMR++;
      if(stat[i]->Id==139 || stat[i]->Id==186) nDN++;
    }

  //barycenter
  xb = yb = sumw = 0;
  for(int i=0; i<nActiveStation; i++) if(InFit[i]) {
      wgt = pow(Signal[i],BARY_POWER);
      xb += X[i]*wgt;
      yb += Y[i]*wgt;
      sumw += wgt;
    }
  xb /= sumw;
  yb /= sumw;

  // counting stations "in fit" (doublets count for one)
  ns = 0;
  if(nCM==2) ns--;
  if(nMR==2) ns--;
  if(nDN==2) ns--;
  for(int i=0; i<nActiveStation; i++) if(InFit[i]) ns++;

  dist2_max_proj = sqr(SELECT_DISTMAX_PROJ_ST) * (ns-2);
  for(int i=nActiveStation-1; i>=0; i--) if(InFit[i]) {
      dx = X[i] - xb;
      dy = Y[i] - yb;
      dist2 = sqr(dx)+sqr(dy) - sqr(u*dx+v*dy);
      if(dist2 > dist2_max_proj) return 0;
    }
  return 1;
}

// main function for reconstruction of the core and S1000
/*-------------------------------------------------------------------*/
int TPBRec::LDFFit()
/*-------------------------------------------------------------------*/
{
  int errflag;
  double costh,xc,dxc,yc,dyc,s1000,ds1000,distc[NSTATMAX],chi2,beta,dbeta,gamma;
  double s1000_start,beta_start,gamma_start;
  int nstatldf, ngooddist, fitSlopeOK;
  double r_inf, r_sup;

  //--- evaluation of the parameters of the LDF from zenith angle
  costh = sqrt(1-uCurrent*uCurrent-vCurrent*vCurrent);
  beta_start = 3 + .7*atan(6*(costh-.65));
  gamma_start = .5 - .05*sin(8*(costh-.6));

  //--- First approximation of Signal(1000) from station with distance
  //    to barycenter closest to 1000 m
  double distmin = 99999;
  double dist_refstat = LDF_DISTREF;
  double sig_refstat = Signal[2];
  for(int i=0; i<nActiveStation; i++) if(InFit[i]) {
      double dx = X[i] - XBarycenter;
      double dy = Y[i] - YBarycenter;
      double dist = sqrt(sqr(dx)+sqr(dy)-sqr(uCurv*dx + vCurv*dy));
      if(fabs(dist-LDF_DISTREF)<distmin) {
        distmin = fabs(dist-LDF_DISTREF);
        sig_refstat = Signal[i];
        dist_refstat = dist;
      }
    }
  s1000_start = sig_refstat*pow(dist_refstat/LDF_DISTREF,beta_start);

  //--- Fit with active stations only, fixed slope
  FitLDFSlope = fitSlopeOK = 0;
  SilentLDF = 0;
  beta = beta_start;
  gamma = gamma_start;
  s1000 = s1000_start;
  xc = yc = 0;
  // counting non-saturated stations
  nstatldf = 0;
  for (int i=0; i<nActiveStation; i++)
    nstatldf += (InFit[i] && Satur1[i]<1000 && Satur2[i]<1000 && Satur3[i]<1000);

  FitLDFMinuit(xc, dxc, yc, dyc, s1000, ds1000, distc, chi2, beta, dbeta, gamma, errflag);
  if (errflag==0 && chi2<LDF_CHI2MAX*maxval(nstatldf-3,1) &&
      sqr(xc)+sqr(yc)-sqr(uCurrent*xc+vCurrent*yc) < sqr(LDF_DISTMAX_BARY_CORE)) {
    LDFFitOK += 1;
    SaveLDFParam(xc, dxc, yc, dyc, s1000, ds1000, distc, beta, dbeta, gamma);
    chi2LDF = chi2;
  } else if(PrintLevel>1)
    PrintLDFfail("LDF",errflag,chi2,maxval(nstatldf-3,1),sqrt(sqr(xc)+sqr(yc)-sqr(uCurrent*xc+vCurrent*yc)));

  // try fitting with variable beta if enough stations in the fit
  // and enough "lever arm" around 1000 m
  //   if fit with fixed slope failed, redefine starting parameters
  if(LDFFitOK==0) {
    for (int i=0; i<nActiveStation; i++) distc[i] = distBary[i];
    beta = beta_start;
    gamma = gamma_start;
    s1000 = s1000_start;
    xc = yc = 0;
  }
  r_inf = 2*LDF_DISTREF;
  r_sup = 0;
  ngooddist = 0;
  for (int i=0; i<nActiveStation; i++) {
    if(InFit[i] && Satur1[i]<1000 && Satur2[i]<1000 && Satur3[i]<1000) {
      if(distc[i] < LDF_DISTREF && distc[i] > 0.4*LDF_DISTREF) {
        r_inf = minval(r_inf,distc[i]);
        ngooddist++;
      }
      if(distc[i] > LDF_DISTREF && distc[i] < 1.6*LDF_DISTREF) {
        r_sup = maxval(r_sup,distc[i]);
        ngooddist++;
      }
    }
  }
  if( nstatldf>4 && (
        (ngooddist>1 && r_sup-r_inf>.9*LDF_DISTREF) ||
        (ngooddist>2 && r_sup-r_inf>.8*LDF_DISTREF) ||
        (ngooddist>3 && r_sup-r_inf>.7*LDF_DISTREF)
      )) {
    FitLDFSlope = 1;

    FitLDFMinuit(xc, dxc, yc, dyc, s1000, ds1000, distc, chi2, beta, dbeta, gamma, errflag);
    if (errflag==0 && chi2<LDF_CHI2MAX*maxval(nstatldf-4,1) &&
        sqr(xc)+sqr(yc)-sqr(uCurrent*xc+vCurrent*yc) < sqr(LDF_DISTMAX_BARY_CORE)) {
      LDFFitOK += 10;
      SaveLDFParam(xc, dxc, yc, dyc, s1000, ds1000, distc, beta, dbeta, gamma);
      chi2LDF = chi2;
    } else if(PrintLevel>1)
      PrintLDFfail("LDFb",errflag,chi2,maxval(nstatldf-4,1),sqrt(sqr(xc)+sqr(yc)-sqr(uCurrent*xc+vCurrent*yc)));
  }

  //--- Fit with active+silent stations, fixed parameters

  // search for silent stations within LDF_DIST_SILENT_MAX from estimated axis
  for(int i=0; i<nSilentStation; i++) {
    double dx = XSilent[i] - XBarycenter;
    double dy = YSilent[i] - YBarycenter;
    double dist = sqrt(sqr(dx)+sqr(dy)-sqr(uCurrent*dx + vCurrent*dy));
    SilentInFit[i] = (!Simulation || dist > LDF_DISTREF) && dist < LDF_DISTMAX_SILENT;
  }

  // if the fit without silent stations failed, redefine the starting values
  if(LDFFitOK==0) {
    s1000 = s1000_start;
    beta = beta_start;
    gamma = gamma_start;
    xc = yc = 0;
  }
  SilentLDF = 1;
  FitLDFSlope = 0;
  FitLDFMinuit(xc, dxc, yc, dyc, s1000, ds1000, distc, chi2, beta, dbeta, gamma, errflag);
  if (errflag==0 && chi2 < (LDF_CHI2MAX+LDF_DCHI2MAX)*(nstatldf-3) &&
      sqr(xc)+sqr(yc)-sqr(uCurrent*xc+vCurrent*yc) < sqr(LDF_DISTMAX_BARY_CORE)) {
    LDFFitOK += 100;
    SaveLDFParam(xc, dxc, yc, dyc, s1000, ds1000, distc, beta, dbeta, gamma);
    chi2LDFwithSilent = chi2;
  } else if(PrintLevel>1)
    PrintLDFfail("LDFs",errflag,chi2,maxval(nstatldf-3,1),sqrt(sqr(xc)+sqr(yc)-sqr(uCurrent*xc+vCurrent*yc)));

  // if enough stations, retry fit with adjustable slope
  if( nstatldf>4 && (
        (ngooddist>1 && r_sup-r_inf>.5*LDF_DISTREF) ||
        (ngooddist>2 && r_sup-r_inf>.4*LDF_DISTREF) ||
        (ngooddist>3 && r_sup-r_inf>.3*LDF_DISTREF)
      )) {
    if(LDFFitOK/100==0) {
      s1000 = s1000_start;
      beta = beta_start;
      gamma = gamma_start;
      xc = yc = 0;
    }
    FitLDFSlope = 1;
    FitLDFMinuit(xc, dxc, yc, dyc, s1000, ds1000, distc, chi2, beta, dbeta, gamma, errflag);

    if (errflag==0 && chi2 < (LDF_CHI2MAX+LDF_DCHI2MAX)*maxval(nstatldf-4,1) &&
        sqr(xc)+sqr(yc)-sqr(uCurrent*xc+vCurrent*yc) < sqr(LDF_DISTMAX_BARY_CORE)) {
      LDFFitOK += 1000;
      SaveLDFParam(xc, dxc, yc, dyc, s1000, ds1000, distc, beta, dbeta, gamma);
      chi2LDFwithSilent = chi2;
    } else if(PrintLevel>1)
      PrintLDFfail("LDFbs",errflag,chi2,maxval(nstatldf-4,1),sqrt(sqr(xc)+sqr(yc)-sqr(uCurrent*xc+vCurrent*yc)));
  }

  if(PrintLevel>1) {
    if(LDFFitOK) {
      printf(" Shape");
      for(int i=0; i<nActiveStation; i++) printf("%6.2f ",Shape[i]);
      printf("\n");
    } else printf("LDF fit failed\n");
  }

  return 1;
}

/*-------------------------------------------------------------------*/
void TPBRec::PrintLDFfail(const char *stage, int errflag, double chi2, int ndof, double dist)
/*-------------------------------------------------------------------*/
{
  printf("!!! %s : ",stage);

  if(errflag!=0) printf("errflag %d",errflag);

  if((stage[3]!='s' && stage[4]!='s') && chi2>LDF_CHI2MAX*ndof)
    printf("  chi2 %5.2f > %4.1f*%d",chi2,LDF_CHI2MAX,ndof);

  if((stage[3]=='s'||stage[4]=='s') && chi2>(LDF_CHI2MAX+LDF_DCHI2MAX)*ndof)
    printf("  chi2 %5.2f > %4.1f*%d",chi2,LDF_CHI2MAX+LDF_DCHI2MAX,ndof);

  if(dist > LDF_DISTMAX_BARY_CORE)
    printf("  dist %5.0f > % 5.0f",dist,LDF_DISTMAX_BARY_CORE);

  printf("\n");
}

/*-------------------------------------------------------------------*/
void TPBRec::SaveLDFParam(double xc, double dxc, double yc, double dyc, double s1000, double ds1000,
                          double *distc, double beta, double dbeta, double gamma)
/*-------------------------------------------------------------------*/
{
  fBeta = beta;
  dbetaLDF = dbeta;
  fGamma = gamma;
  fXCore = XBarycenter + xc;
  fYCore = YBarycenter + yc;
  dXCore = dxc;
  dYCore = dyc;
  fSRef = s1000;
  dS1000 = ds1000;
  ConvertS1000ToEnergy(sqrt(1-uCurrent*uCurrent-vCurrent*vCurrent),s1000,ds1000);

  for(int i=0; i<nActiveStation; i++) {
    distCore[i] = distc[i];
    double dxs = ((X[i]-fXCore)*cos(Phi)+(Y[i]-fYCore)*sin(Phi))*cos(Theta);
    double dys =  (Y[i]-fYCore)*cos(Phi)-(X[i]-fXCore)*sin(Phi);
    azimCore[i] = atan2(dys,dxs);
    Shape[i] = stat[i]->ShapeParameter(distCore[i]);
  }
}

// main function for reconstruction of the core and S1000
/*-------------------------------------------------------------------*/
int TPBRec::GlobalFit()
/*-------------------------------------------------------------------*/
{
  int errGlob, errflag, ndegfit;
  double t0,u,v,curv,xc,yc,s1000,distc[NSTATMAX],cov6[6][6],cov7[7][7],chi2;

  //--- Fit with active stations only
  t0 = fT0;
  u = sin(Theta)*cos(Phi);
  v = sin(Theta)*sin(Phi);
  curv = 1/fR;
  s1000 = fSRef;
  xc = fXCore - XBarycenter;
  yc = fYCore - YBarycenter;

  SilentLDF = 0;
  if(nPositionInFit>4) FitGlobMinuit(t0, u, v, curv, xc, yc, s1000, distc, cov7[0], chi2, errflag);
  else                 FitGlobFixCurvMinuit(t0, u, v, xc, yc, s1000, distc, cov6[0], chi2, errflag);
  errGlob = (errflag>0);
  chi2Glob = chi2;

  //--- Fit with active+silent stations, fixed parameters
  // search for silent stations within LDF_DISTMAX_SILENT from estimated axis
  for(int i=0; i<nSilentStation; i++) {
    double dx = XSilent[i] - fXCore;
    double dy = YSilent[i] - fYCore;
    double dist = sqrt(sqr(dx)+sqr(dy)-sqr(uCurrent*dx + vCurrent*dy));
    SilentInFit[i] = (!Simulation || dist > LDF_DISTREF) && dist < LDF_DISTMAX_SILENT;
  }
  t0 = fT0;
  u = sin(Theta)*cos(Phi);
  v = sin(Theta)*sin(Phi);
  curv = 1/fR;
  s1000 = fSRef;
  xc = fXCore - XBarycenter;
  yc = fYCore - YBarycenter;

  SilentLDF = 1;
  if(nPositionInFit>4) FitGlobMinuit(t0, u, v, curv, xc, yc, s1000, distc, cov7[0], chi2, errflag);
  else                 FitGlobFixCurvMinuit(t0, u, v, xc, yc, s1000, distc, cov6[0], chi2, errflag);
  errGlob = errGlob || (errflag>0);
  chi2GlobwithSilent = chi2;

  //--- quality criteria
  if(nPositionInFit>4) ndegfit = 2*nStationInFit-7;
  else                 ndegfit = 2*nStationInFit-6;
  GlobFitOK = (errGlob == 0 &&
               // acceptable direction cosines
               u*u+v*v < 1 &&
               // chi2 of global fit (active stations only)
               chi2Glob < GLOB_CHI2MAX*ndegfit &&
               // increase of chi2 of Glob fit due to silent stations
               chi2GlobwithSilent-chi2Glob < GLOB_DCHI2MAX*nStationInFit &&
               // distance (in space) from barycenter to fitted axis
               sqr(xc)+sqr(yc) - sqr(u*xc+v*yc) < sqr(LDF_DISTMAX_BARY_CORE));

  if(GlobFitOK) {
    // save results
    fT0 = t0;
    uCurrent = u; vCurrent = v;
    fU = uCurrent;
    fV = vCurrent;
    fR = 1/curv;
    fSRef = s1000;
    fXCore = XBarycenter + xc;
    fYCore = YBarycenter + yc;

    if(nPositionInFit>4) {
      dXCore = sqrt(cov7[4][4]);
      dYCore = sqrt(cov7[5][5]);
      dS1000 = sqrt(cov7[6][6])*s1000;
    } else {
      dXCore = sqrt(cov6[3][3]);
      dYCore = sqrt(cov6[4][4]);
      dS1000 = sqrt(cov6[5][5])*s1000;
    }
    ConvertS1000ToEnergy(sqrt(1-u*u-v*v),fSRef,dS1000);

    for(int i=0; i<nActiveStation; i++) {
      distCore[i] = distc[i];
      double dxc = X[i]-fXCore;
      double dyc = Y[i]-fYCore;
      double dxs = (dxc*cos(Phi)+dyc*sin(Phi))*cos(Theta);
      double dys =  dyc*cos(Phi)-dxc*sin(Phi);
      azimCore[i] = atan2(dys,dxs);
      Shape[i] = stat[i]->ShapeParameter(distCore[i]);
    }

    double sinth = sqrt(u*u+v*v);
    double costh = sqrt(1-sinth*sinth);
    Theta = asin(sinth);
    Phi = atan2(v,u);
    if(nPositionInFit>4) {
      dTheta = sqrt(u*u*cov7[1][1] + v*v*cov7[2][2] + 2*u*v*cov7[1][2])/sinth/costh;
      dPhi = sqrt(v*v*cov7[1][1] + u*u*cov7[2][2] - 2*u*v*cov7[1][2])/sinth/sinth;
    } else {
      dTheta = sqrt(u*u*cov6[1][1] + v*v*cov6[2][2] + 2*u*v*cov6[1][2])/sinth/costh;
      dPhi = sqrt(v*v*cov6[1][1] + u*u*cov6[2][2] - 2*u*v*cov6[1][2])/sinth/sinth;
    }
  } else if(PrintLevel>1)
    printf("Glob fit failed - errflag %d u2+v2 %5.3f chi2 %6.2f dchi2 %6.2f dist %6.0f \n",
           errflag, u*u+v*v, chi2Glob, chi2GlobwithSilent-chi2Glob,
           sqrt(sqr(xc)+sqr(yc)-sqr(u*xc+v*yc)));

  return 1;
}

/*-------------------------------------------------------------------*/
int TPBRec::Recover3st()
/*-------------------------------------------------------------------*/
{
  int i,j,k;

  // loop over all triplets
  for(i=0; i<nActiveStation-2; i++) {
    for(j=i+1; j<nActiveStation-1; j++) {
      for(k=j+1; k<nActiveStation; k++) {
        // analyze this triplet as a 3-fold event
        Anal3st(i,j,k);
        if(Fit3stOK) {
          InFit[i] = InFit[j] = InFit[k] = 1;
          break;
        }
      }
      if(Fit3stOK) break;
    }
    if(Fit3stOK) break;
  }

  return 1;
}

/*-------------------------------------------------------------------*/
void TPBRec::SetTime(int time_start)
/*-------------------------------------------------------------------*/
{
  for(int i=0; i<nActiveStation; i++) {
    switch(time_start) {
    case 1 : T[i] = stat[i]->Gps->NanoSecond + stat[i]->T10(); break;  // a possible definition
    case 0 :
    default : T[i] = stat[i]->Gps->NanoSecond - (kIoSd::MAXSAMPLE-CleanStart[i])*kEc::TimeSlot;
    }
  }
}

#define RISEFALL_RMIN 300.
#define RISEFALL_RMAX 2500.
#define DXMIN 1500.
/*-------------------------------------------------------------------*/
void TPBRec::RiseFallTime()
/*-------------------------------------------------------------------*/
{
  int fit_asym;
  double asym;

  if(PrintLevel>2) printf("  Id   dist  cosa    t10  t50  t90  shape  weight\n");

  // criterion to introduce the asymmetry as an adjustable parameter : longitudinal
  double xinf = 10000;
  double xsup = -10000;
  int nst = 0;
  for(int is=0; is<nActiveStation; is++) if(InFit[is]) {
      double dx_fr = cos(Theta)*(cos(Phi)*(X[is]-fXCore)+sin(Phi)*(Y[is]-fYCore));
      double dy_fr = cos(Phi)*(Y[is]-fYCore)-sin(Phi)*(X[is]-fXCore);
      double dist = sqrt(dx_fr*dx_fr+dy_fr*dy_fr);
      if(dist>RISEFALL_RMIN && dist<RISEFALL_RMAX) {
        xinf = minval(dx_fr,xinf);
        xsup = maxval(dx_fr,xsup);
        nst++;
      }
    }
  if(nst<3) return;
  fit_asym = xsup-xinf>DXMIN;

  double sumw=0, sumw_r2=0, sumw_cos=0, sumw_cos2=0, sumw_r2cos=0, sumw_r2cos2=0,
                                     sumw_r_rt=0, sumw_rcos_rt=0, sumw_r_ft=0, sumw_rcos_ft=0, sumw_sh=0, sumw_cos_sh=0;

  for(int is=0; is<nActiveStation; is++) if(InFit[is]) {
      double dx_fr = cos(Theta)*(cos(Phi)*(X[is]-fXCore)+sin(Phi)*(Y[is]-fYCore));
      double dy_fr = cos(Phi)*(Y[is]-fYCore)-sin(Phi)*(X[is]-fXCore);
      double dist2 = dx_fr*dx_fr+dy_fr*dy_fr;
      double dist = sqrt(dist2);
      if(PrintLevel>2) printf("%4d %6.0f %6.3f  %4.0f %4.0f %4.0f  %5.3f  %6.2f\n",
                                stat[is]->Id,dist,dx_fr/dist,Tclean10[is],Tclean50[is],Tclean90[is],Shape[is],Signal[is]);

      if(dist>RISEFALL_RMIN && dist<RISEFALL_RMAX) {
        double cosa = dx_fr/dist;
        double tau_rt = Tclean50[is]-Tclean10[is];
        double tau_ft = Tclean90[is]-Tclean50[is];
        double weight = Signal[is];
        if(!fit_asym) {
          double costh = cos(Theta);
          if(costh>.26) {
            asym = .2*(1-costh)*(costh-.26)*4/sqr(1-.26);
            tau_rt /= (1+asym*cosa);
            asym = .3*(1-costh)*(costh-.26)*4/sqr(1-.26);
            tau_ft /= (1+asym*cosa);
          }
        }
        sumw += weight;
        sumw_sh += weight*Shape[is];
        sumw_r2 += weight*dist2;
        sumw_r_rt  += weight*dist*tau_rt;
        sumw_r_ft  += weight*dist*tau_ft;
        if(fit_asym) {
          sumw_cos += weight*cosa;
          sumw_cos2 += weight*cosa*cosa;
          sumw_cos_sh  += weight*cosa*Shape[is];
          sumw_rcos_rt  += weight*dist*cosa*tau_rt;
          sumw_rcos_ft  += weight*dist*cosa*tau_ft;
          sumw_r2cos += weight*dist2*cosa;
          sumw_r2cos2 += weight*dist2*cosa*cosa;
        }
      }
    }
  if(fit_asym) {
    double deter_t = sumw_r2*sumw_r2cos2-sumw_r2cos*sumw_r2cos;
    double deter_s = sumw*sumw_cos2-sumw_cos*sumw_cos;
    AvgRiseTime = (sumw_r2cos2*sumw_r_rt-sumw_r2cos*sumw_rcos_rt)/deter_t;
    AsymRiseTime = (sumw_r2*sumw_rcos_rt-sumw_r2cos*sumw_r_rt)/deter_t/AvgRiseTime;
    AvgFallTime = (sumw_r2cos2*sumw_r_ft-sumw_r2cos*sumw_rcos_ft)/deter_t;
    AsymFallTime = (sumw_r2*sumw_rcos_ft-sumw_r2cos*sumw_r_ft)/deter_t/AvgFallTime;
    AvgShape = (sumw_cos2*sumw_sh-sumw_cos*sumw_cos_sh)/deter_s;
    AsymShape = (sumw*sumw_cos_sh-sumw_cos*sumw_sh)/deter_s/AvgShape;
  } else {
    AvgRiseTime = sumw_r_rt/sumw_r2;
    AvgFallTime = sumw_r_ft/sumw_r2;
    AvgShape = sumw_sh/sumw;
  }

  if(PrintLevel>2) {
    printf("%d stations kept - Rise avg/asym %3.0f %5.3f Fall avg/asym %3.0f %5.3f  Shape avg/asym %5.3f %5.3f\n",
           nst,1000*AvgRiseTime,AsymRiseTime,1000*AvgFallTime,AsymFallTime,AvgShape,AsymShape);
    printf("\n");
  }
}

#define NCOEF 2
#define DIST_MIN_PEAK 500
#define FIRST_SLOT_PEAK 0
#define LAST_SLOT_PEAK 100
#define THRESHOLD_PEAK .3
#define MIN_PEAK1 .7
#define MIN_PEAK2 .8
#define MAX_PEAK2 1.5
#define SLP_MIN_PEAK2 .3
#define SLP_MAX_PEAK2 1.5
#define ALTIT_MU_ORIGIN 7000.
#define SECTH_MAX 3.
/*-------------------------------------------------------------------*/
int TPBRec::FindPeak()
/*-------------------------------------------------------------------*/
{
  double trace[kIoSd::MAXSAMPLE], ftrace[kIoSd::MAXSAMPLE];
  double coef[NCOEF] = {3.1,-2.1};

  int i, ntubeOK, n_mu_1slot, n_mu_2slot;
  double fact[3], sum_in, sum_out, sum_mu_1slot, sum_mu_2slot ;

  // loop over stations beyond DIST_MIN_PEAK
  for(i=0; i<nActiveStation; i++) {
    nMuPeak[i] = 0;
    if(InFit[i] && distCore[i] > DIST_MIN_PEAK) {
      n_mu_1slot = n_mu_2slot = 0;
      sum_mu_1slot = sum_mu_2slot = 0;

      for (unsigned int j=0; j<kIoSd::MAXSAMPLE; j++) trace[j] = ftrace[j] = 0;
      ntubeOK = 0;
      for(unsigned int k=0; k<3; k++) {
        if(stat[i]->Calib->TubeOk[k]) {
          ntubeOK++;
          fact[k] = stat[i]->fPmt[k].VemPeak() / stat[i]->fPmt[k].VemCharge();
        }
      }

      double sumPMT[3] = {0,0,0}; //test
      for(unsigned int j=0; j<kIoSd::MAXSAMPLE; j++) for(unsigned int k=0; k<3; k++) if(stat[i]->Calib->TubeOk[k]) {
            double val =  stat[i]->fPmt[k].fTrace[j];
            sumPMT[k] += val; //test
            if(j>0 && j<kIoSd::MAXSAMPLE-1) {
              double valp =  stat[i]->fPmt[k].fTrace[j-1];
              double valf =  stat[i]->fPmt[k].fTrace[j+1];
              // suppress isolated PMT pulse (1 slot well above the previous and the next one)
              if (val>3*valp && val>3*valf) val = (valp+valf)/2;
            }
            trace[j] += maxval(0,val) * fact[k] / ntubeOK;
          }

      // build the filtered trace
      for(unsigned int j=NCOEF-1; j<kIoSd::MAXSAMPLE; j++) for(unsigned int k=0; k<NCOEF; k++)
          ftrace[j] += coef[k]*trace[j-k];

      int istart = stat[i]->fStartBin;
      for (int j=istart+FIRST_SLOT_PEAK; j<istart+LAST_SLOT_PEAK && j<(int)kIoSd::MAXSAMPLE-1; j++) {
        // search for a local maximum...
        if(ftrace[j]>ftrace[j-1] && ftrace[j]>ftrace[j+1]) {
          sum_out = ftrace[j-1]+ftrace[j+1];
          sum_in  = ftrace[j];
          // condition defining a "peak"
          if(sum_in>THRESHOLD_PEAK && sum_in-sum_out/2>THRESHOLD_PEAK
              && sum_in>2*sum_out) {
            if(sum_in > MIN_PEAK1) {
              n_mu_1slot++;
              sum_mu_1slot += sum_in;
            }
          }
          // essai sur deux slots
          if(ftrace[j-1]>ftrace[j+1]) {
            sum_in = ftrace[j]+ftrace[j-1];
            sum_out = ftrace[j-2]+ftrace[j+1];
          } else {
            sum_in = ftrace[j]+ftrace[j+1];
            sum_out = ftrace[j-1]+ftrace[j+2];
          }
          // counting "muonlike" peaks
          if(sum_in>THRESHOLD_PEAK && sum_in-sum_out/2>THRESHOLD_PEAK
              && sum_in>3*sum_out) {
            double costh = cos(Theta);
            double sinth = sqrt(1-costh*costh);
            double xx = (ALTIT_MU_ORIGIN*sinth-distCore[i]*cos(azimCore[i]))/costh;
            double yy = -distCore[i]*sin(azimCore[i]);
            double secth_mu = minval(SECTH_MAX,sqrt(sqr(xx)+sqr(yy)+sqr(ALTIT_MU_ORIGIN))/ALTIT_MU_ORIGIN);
            if(sum_in > MIN_PEAK2+(secth_mu-1)*SLP_MIN_PEAK2 && sum_in < MAX_PEAK2+(secth_mu-1)*SLP_MAX_PEAK2)
              nMuPeak[i]++;
          }
        }
      }
    }
  }
  return 0;
}

#define DIST_MIN_MUON 300
#define DIST_MAX_MUON 3000
#define START_BIN_MUON 150
/*-------------------------------------------------------------------*/
int TPBRec::MuonCount()
/*-------------------------------------------------------------------*/
{
  float trace[3], sum_in, sum_out, asym_in, asym_out, asym_dir_in, asym_dir_out;

  unsigned int j, k; int i,  ntubeOK;

  for(i=0; i<nActiveStation; i++) if(InFit[i] && distCore[i] > DIST_MIN_MUON && distCore[i] < DIST_MAX_MUON) {
      ntubeOK = 0;
      for(k=0; k<3; k++) if(stat[i]->Calib->TubeOk[k]) ntubeOK++;
      if(ntubeOK<3) continue;

      sum_in = sum_out = asym_in = asym_out = asym_dir_in = asym_dir_out = 0;
      for(j=0; j<kIoSd::MAXSAMPLE; j++) {
        float trmax = 0;
        float sumtr = 0;
        float sumtr_dir = 0;
        for(k=0; k<3; k++) {
          trace[k] =  stat[i]->fPmt[k].fTrace[j];
          sumtr += trace[k];
          if(trace[k]>trmax) trmax=trace[k];
          sumtr_dir += trace[k]*cos(Phi+M_PI/2-k*2*M_PI/3);
        }
        if(j>=START_BIN_MUON && j<START_BIN_MUON+kIoSd::MAXSAMPLE/2) {
          sum_in += sumtr;
          asym_in += trmax-sumtr/3;
          asym_dir_in += sumtr_dir*2/3;
        } else {
          sum_out += sumtr;
          asym_out += trmax-sumtr/3;
          asym_dir_out += sumtr_dir*2/3;
        }
      }
      AsymPMT[i] = asym_in/sum_in;
    }
  return 1;
}

#define NSIGMA_BEFORE 3
#define NSIGMA_AFTER 5
/*-------------------------------------------------------------------*/
int TPBRec::CleanTrace()
/*-------------------------------------------------------------------*/
{
  double trace[kIoSd::MAXSAMPLE], fact[3];
  double sum, sum1, sum2, average, sigma, dsum, tc10, tc50, tc90;
  int ntubeOK,j1,j2,j1n,j2n;

  for(int i=0; i<nActiveStation; i++) if(InFit[i]) {
      for (unsigned int j=0; j<kIoSd::MAXSAMPLE; j++) trace[j] = 0;
      ntubeOK = 0;
      for(unsigned int k=0; k<3; k++) {
        if(stat[i]->Calib->TubeOk[k]) {
          ntubeOK++;
          fact[k] = stat[i]->fPmt[k].VemPeak() / stat[i]->fPmt[k].VemCharge();
        }
      }

      for(unsigned int j=0; j<kIoSd::MAXSAMPLE; j++) for(unsigned int k=0; k<3; k++) if(stat[i]->Calib->TubeOk[k]) {
            double val =  stat[i]->fPmt[k].fTrace[j];

            // suppress isolated PMT pulse (1 slot well above the previous and the next one)
            if(j>0 && j<kIoSd::MAXSAMPLE-1) {
              double valp = stat[i]->fPmt[k].fTrace[j-1];
              double valf = stat[i]->fPmt[k].fTrace[j+1];
              if (val>3*valp && val>3*valf) val = (valp+valf)/2;
            }
            trace[j] += maxval(0,val) * fact[k] / ntubeOK;
          }

      j1n = CleanStart[i];
      j2n = CleanStop[i];
      do {
        j1 = j1n; j2 = j2n;
        sum = sum1 = sum2 = 0;
        for(int j=j1; j<j2; j++) {
          sum += trace[j];
          sum1 += j*trace[j];
          sum2 += j*j*trace[j];
        }
        average = sum1/sum;
        sigma = sqrt(sum2/sum-average*average);
        j1n = maxval(CleanStart[i],(int)(average-NSIGMA_BEFORE*sigma));
        j2n = minval(CleanStop[i],(int)(average+NSIGMA_AFTER*sigma));
      } while(j1n<j1 || j2n>j2);

      sum = 0;
      for(int j=j1n; j<j2n; j++) sum += trace[j];
      sum1 = tc10 = tc50 = tc90 = 0;
      for(int j=j1n; j<j2n; j++) {
        sum1 += trace[j];
        if(!tc10) if((dsum=sum1-.1*sum)>0) tc10 = j-dsum/trace[j];
        if(!tc50) if((dsum=sum1-.5*sum)>0) tc50 = j-dsum/trace[j];
        if(!tc90) if((dsum=sum1-.9*sum)>0) tc90 = j-dsum/trace[j];
      }
      Tclean10[i] = (tc10-CleanStart[i]+1)*kEc::TimeSlot;
      Tclean50[i] = (tc50-CleanStart[i]+1)*kEc::TimeSlot;
      Tclean90[i] = (tc90-CleanStart[i]+1)*kEc::TimeSlot;
    }
  return 1;
}

#define NCOEF 2
#define START_SLOT 100
#define END_SLOT 500
#define NSLOT_WINDOW_UP 30
#define NSLOT_WINDOW_DOWN 50
#define THRESHOLD_CLEAN_UP 0.1
#define THRESHOLD_CLEAN_DOWN 0.05
#define MARGIN_STOP 5
#define THRESHOLD_SUM 3.
/*-------------------------------------------------------------------*/
int TPBRec::CleanFilter(TCalibStation *st, int& start, int& stop)
/*-------------------------------------------------------------------*/
{
  double trace[kIoSd::MAXSAMPLE], ftrace[kIoSd::MAXSAMPLE];
  double coef[NCOEF] = {3.1,-2.1}; // corresponds to a decay time of 64 ns for the light
  unsigned int j, k; int ntubeOK, above, nabove, npeak, first, last;
  double fact[3], sum;

  for (j=0; j<kIoSd::MAXSAMPLE; j++) trace[j] = ftrace[j] = 0;
  ntubeOK = 0;
  for(k=0; k<3; k++) {
    if(st->Calib->TubeOk[k]) {
      ntubeOK++;
      fact[k] = st->fPmt[k].VemPeak() / st->fPmt[k].VemCharge();
    }
  }

  for(j=0; j<kIoSd::MAXSAMPLE; j++) for(k=0; k<3; k++) if(st->Calib->TubeOk[k]) {
        double val =  st->fPmt[k].fTrace[j];
        if(j>0 && j<kIoSd::MAXSAMPLE-1) {
          double valp =  st->fPmt[k].fTrace[j-1];
          double valf =  st->fPmt[k].fTrace[j+1];
          // suppress isolated PMT pulse (1 slot well above the previous and the next one)
          if (val>3*valp && val>3*valf) val = (valp+valf)/2;
        }
        trace[j] += maxval(0,val) * fact[k] / ntubeOK;
      }

  // build the filtered trace
  for(j=NCOEF-1; j<kIoSd::MAXSAMPLE; j++) for(k=0; k<NCOEF; k++)
      ftrace[j] += coef[k]*trace[j-k];

  // search for the "start" point -------------------------------------

  //int start_peak = END_SLOT;

  start = stop = -1;
  // ascending search for start using a sliding window
  for(j=START_SLOT; j<END_SLOT-NSLOT_WINDOW_UP; j++) {
    first = -1;
    above = 0;
    nabove = 0;
    npeak = 0;
    sum = 0;
    // scan a window of length NSLOT_WINDOW_UP
    for(k=0; k<NSLOT_WINDOW_UP; k++) {
      if(ftrace[j+k]>THRESHOLD_CLEAN_UP) {
        if(first<0) first = j+k;
        if(!above) nabove = 0;
        above = 1;
        nabove++;
        //if(start_peak==END_SLOT) start_peak = j+k;
        sum += ftrace[j+k];
        if(nabove>2 || sum>THRESHOLD_SUM) {
          start = first; break;
        }
      } else {
        if(above) npeak++;
        if(npeak>1) {
          start = first; break;
        }
        above = 0;
        sum = 0;
      }
    }
    if(start>=0) break;
  }
  if(start<0) start = END_SLOT;

  // search for the "stop" point -------------------------------------

  //int stop_peak = 0;
  // descending search for stop using a sliding window
  for(j=END_SLOT; j>START_SLOT+NSLOT_WINDOW_DOWN; j--) {
    last = -1;
    above = 0;
    nabove = 0;
    npeak = 0;
    sum = 0;
    // scan a window of length NSLOT_WINDOW_DOWN
    for(k=0; k<NSLOT_WINDOW_DOWN; k++) {
      if(ftrace[j-k]>THRESHOLD_CLEAN_DOWN) {
        if(last<0) last = j-k;
        if(!above) nabove = 0;
        above = 1;
        nabove++;
        //if(stop_peak==0) stop_peak = j-k;
        sum += ftrace[j+k];
        if(nabove>2 || sum>THRESHOLD_SUM) {
          stop = last; break;
        }
      } else {
        if(above) npeak++;
        if(npeak>1) {
          stop = last; break;
        }
        above = 0;
        sum = 0;
      }
    }
    if(stop>=0) break;
  }
  if(stop<start) return 0;

  stop += MARGIN_STOP;
  return 1;
}

/*-------------------------------------------------------------------*/
int TPBRec::CleanSignal(TCalibStation *st, int start, int stop, double& signal, double& peak)
/*-------------------------------------------------------------------*/
{
  int j, k, nPMT;
  double fact[3];

  nPMT = 0;
  for(k=0; k<3; k++) if(st->Calib->TubeOk[k]) {
      fact[k] = st->fPmt[k].VemPeak() / st->fPmt[k].VemCharge();
      nPMT++;
    }

  signal = 0;
  peak = 0;
  for(j=start; j<stop; j++) {
    double vem_peak = 0;
    double vem_charge = 0;
    for(k=0; k<3; k++) if(st->Calib->TubeOk[k]) {
        vem_peak += st->fPmt[k].fTrace[j] / nPMT;
        vem_charge += st->fPmt[k].fTrace[j] * fact[k]/nPMT;
      }
    peak = maxval(peak,vem_peak);
    signal += vem_charge;
  }
  return (signal>0 && peak>0);
}

/*-------------------------------------------------------------------*/
int TPBRec::accept_station(IoSdStation& st, int window_condition)
/*-------------------------------------------------------------------*/
{
  if(st.Error || (st.Trigger.Window && window_condition)) return 0;
  if(!Simulation) {
    if(RemoveEA && st.Id<100) return 0;
    if(RemoveInfillStation && (st.Id==36 || st.Id==67)) return 0;
    if(RemoveDoublet && (st.Id==64 || st.Id==185 || st.Id==186)) return 0;
    //if(RemoveDoublet && (st.Id==64 || st.Id==139 || st.Id==140)) return 0;
  }
  return 1;
}

/*-------------------------------------------------------------------*/
int TPBRec::select_input()
/*-------------------------------------------------------------------*/
{
  int nst, topo_4C1, topo_4C1_strict, TOT_3C1, FD_3C1, doublet;
  int window_condition, *accept_st;
  IoSdStation *st1,*st2;

  nst = (int)fErEvent->Trigger.NumberOfStation;

  // build list of accepted stations (including T1)
  accept_st = (int *)calloc(sizeof(int),nst);
  for(int i=0; i<nst; i++) accept_st[i] = accept_station(fErEvent->Stations[i],0);

  topo_4C1 = 0;
  // topology "4C1"
  for(int i=0; i<nst; i++) if(accept_st[i]) {
      st1 = &(fErEvent->Stations[i]);
      int neighbour = 0;
      for(int j=0; j<nst; j++) if(j!=i && accept_st[j]) {
          st2 = &(fErEvent->Stations[j]);
          float dist2 = sqr(st2->Easting-st1->Easting) + sqr(st2->Northing-st1->Northing);
          neighbour += (sqrt(dist2)<1800 && abs(st2->Gps->NanoSecond-st1->Gps->NanoSecond)<6000);
        }
      topo_4C1 += (neighbour>2);
    }

  // rebuild list of accepted stations (if not FD trigger: keep T2 stations only)
  window_condition = strncmp(fErEvent->Trigger.Algo.c_str(),"FD",2);
  for(int i=0; i<nst; i++) accept_st[i] = accept_station(fErEvent->Stations[i],window_condition);

  topo_4C1_strict = 0;
  // topology "4C1" (restricted)
  for(int i=0; i<nst; i++) if(accept_st[i]) {
      st1 = &(fErEvent->Stations[i]);
      int neighbour = 0;
      for(int j=0; j<nst; j++) if(j!=i && accept_st[j]) {
          st2 = &(fErEvent->Stations[j]);
          float dist2 = sqr(st2->Easting-st1->Easting) + sqr(st2->Northing-st1->Northing);
          neighbour += (sqrt(dist2)<1800 && abs(st2->Gps->NanoSecond-st1->Gps->NanoSecond)<6000);
        }
      topo_4C1_strict += (neighbour>2);
    }

  TOT_3C1 = 0;
  // trigger TOT with a topology "3C1" (discarding doublets and infill stations)
  if (!strncmp(fErEvent->Trigger.Algo.c_str(),"TOT",3)) {
    for(unsigned int i=0; i<fErEvent->Trigger.NumberOfStation; i++)
      if(fErEvent->Stations[i].Error==0 && fErEvent->Stations[i].Trigger.Window==0) {
        int neighbour = 0;
        for(unsigned int j=0; j<fErEvent->Trigger.NumberOfStation; j++)
          if(j!=i && fErEvent->Stations[j].Error==0 && fErEvent->Stations[j].Trigger.Window==0) {
            float dist = sqrt(sqr(fErEvent->Stations[j].Easting-fErEvent->Stations[i].Easting) +
                              sqr(fErEvent->Stations[j].Northing-fErEvent->Stations[i].Northing));
            neighbour += (dist>1200 && dist<1800 &&
                          abs(fErEvent->Stations[j].Gps->NanoSecond-fErEvent->Stations[i].Gps->NanoSecond)<6000);
          }
        TOT_3C1 += (neighbour>1);
      }
  }

  FD_3C1 = 0;
  // trigger FD with condition "3C1" (discarding doublets and infill stations)
  if(!strncmp(fErEvent->Trigger.Algo.c_str(),"FD",2)) {
    for(unsigned int i=0; i<fErEvent->Trigger.NumberOfStation; i++) if(fErEvent->Stations[i].Error==0) {
        int neighbour = 0;
        for(unsigned int j=0; j<fErEvent->Trigger.NumberOfStation; j++) if(j!=i && fErEvent->Stations[j].Error==0) {
            float dist = sqrt(sqr(fErEvent->Stations[j].Easting-fErEvent->Stations[i].Easting) +
                              sqr(fErEvent->Stations[j].Northing-fErEvent->Stations[i].Northing));
            neighbour += (dist>1200 && dist<1800 &&
                          abs(fErEvent->Stations[j].Gps->NanoSecond-fErEvent->Stations[i].Gps->NanoSecond)<6000);
          }
        FD_3C1 += (neighbour>1);
      }
  }

  // event with a doublet (and at least 2 other stations) in a 3C1 topology
  doublet = 0;
  int ns=0, nDN=0, nMR = 0;
  for(unsigned int i=0; i<fErEvent->Trigger.NumberOfStation; i++) {
    if(fErEvent->Stations[i].Error==0) {
      ns++;
      nDN += (fErEvent->Stations[i].Id==139 || fErEvent->Stations[i].Id==186);
      nMR += (fErEvent->Stations[i].Id==140 || fErEvent->Stations[i].Id==185);
    }
  }
  if(nDN==2) ns--;
  if(nMR==2) ns--;
  if(ns>2 && (nDN>1 || nMR>1)) {       //there is an active doublet
    for(unsigned int i=0; i<fErEvent->Trigger.NumberOfStation; i++) if(fErEvent->Stations[i].Error==0) {
        // fErEvent->Stations[i] is the "current" station
        // neighbour counters : neighbDN for Dia or Noche, neighbMR for Moulin or Rouge
        int neighbour = 0, neighbDN = 0, neighbMR = 0;
        // nstatd = 1 if the current station is a member of an active doublet
        int nstatd = ((fErEvent->Stations[i].Id==139 || fErEvent->Stations[i].Id==186) && nDN==2)
                     || ((fErEvent->Stations[i].Id==140 || fErEvent->Stations[i].Id==185) && nMR==2);
        // loop over the other active stations
        for(unsigned int j=0; j<fErEvent->Trigger.NumberOfStation; j++) if(j!=i && fErEvent->Stations[j].Error==0) {
            float dist = sqrt(sqr(fErEvent->Stations[j].Easting-fErEvent->Stations[i].Easting) +
                              sqr(fErEvent->Stations[j].Northing-fErEvent->Stations[i].Northing));
            // selection on distance (in space and time) between this station and the current one
            if(dist>1200 && dist<1800 &&
                abs(fErEvent->Stations[j].Gps->NanoSecond-fErEvent->Stations[i].Gps->NanoSecond)<6000) {
              int isDN = (fErEvent->Stations[j].Id==139 || fErEvent->Stations[j].Id==186);
              int isMR = (fErEvent->Stations[j].Id==140 || fErEvent->Stations[j].Id==185);
              // increment the number of neighbours of the current station
              // (counting only one neighbour for Dia+Noche or Moulin+Rouge)
              if((!isDN && !isMR) || (isDN && !neighbDN) || (isMR && !neighbMR)) neighbour++;
              neighbDN += isDN;
              neighbMR += isMR;
              // increment the number of active doublets in the current topology
              nstatd += (isDN && nDN==2) || (isMR && nMR==2);
            }
          }
        // implement the condition : the current station has at least two active neighbours
        // and there is a doublet amongst the current station and its active neighbours
        doublet += (neighbour>1 && nstatd>0);
      }
  }

  free(accept_st);

  switch(Input_sel_evts) {
  case 0:  return 1;         // accept all
  case 1:  return topo_4C1;  // topology "4C1" only
  case 2:  return TOT_3C1;   // trigger TOT with a topology "3C1" only
  case 3:  return FD_3C1;    // trigger FD with a topology "3C1" only
  case 4:  return doublet;   // events with a doublet
  case 5:  return topo_4C1_strict;
  case 10: return topo_4C1+TOT_3C1+FD_3C1+doublet;  // all events with a "physical" configuration
  default: return 0;
  }
}


/*-------------------------------------------------------------------*/
int TPBRec::LDFhorFit()/// Fitting horizontal events to a table
/*-------------------------------------------------------------------*/
{
  //Using the same horizontal reconstruction as the others

  ///The horizontal reconstruction is performed by the THorReconstruct class.
  THorReconstruct horrec;
  *((TReconstruct *) & horrec) = *((TReconstruct *) this);

  horrec.FillInputs(TShowerParams::Theta("deg"),TShowerParams::Phi("deg"), fXCore, fYCore);
  //horrec.Verbose = _verbose;
  horrec.SetEDPlot(fErEvent->fHorPlot);
  if ((horrec.Reconstruct(*fErEvent)) == false ) return 0;
  *((TReconstruct *) this) = *((TReconstruct *) & horrec);

  // MatCurvedFit(1);

  if (TShowerParams::Theta("deg") > 84) {
    horrec.FillInputs(TShowerParams::Theta("deg"), TShowerParams::Phi("deg"), fXCore, fYCore);
    if ( (horrec.Reconstruct(*fErEvent)) == false ) return 0;
    *((TReconstruct *) this) = *((TReconstruct *) & horrec);
  }

  fVarCov[TSdParams::eXCore][TSdParams::eXCore] = sqr(horrec.dDXCore);
  fVarCov[TSdParams::eYCore][TSdParams::eYCore] = sqr(horrec.dDYCore);
  fVarCov[TSdParams::eSRef][TSdParams::eSRef] = sqr(fErEvent->fRec.fdE);

  for (int iy = 0; iy < 52; iy++) {
    for (int ix = 0; ix < 52; ix++) {
      fErEvent->fEDMapMu[ix][iy] = horrec.EDMapMu[ix][iy];
      fErEvent->fEDMapMuBis[ix][iy] = horrec.EDMapE[ix][iy];
    }
  }
  fErEvent->fEDdmax = (horrec.EDdxmax>horrec.EDdymax) ? horrec.EDdxmax: horrec.EDdymax;
  fErEvent->fEDdmax += 200;

  return 1;

  //   int errflag;
  //     double theta_deg, dth, phi, xc, dxc, yc, dyc, fact, dfact, chi2, alpha[5][7][5], beta[5][7][5];
  //     int ith,k,l,m;

  //     // build table for values of theta and phi in a given event

  //     theta_deg = asin(sqrt(uCurv*uCurv+vCurv*vCurv))*180/M_PI;
  //     ith = (int) ((theta_deg-70.)/2);
  //     if(ith<0) ith = 0;
  //     if(ith>=9) ith = 8;
  //     dth = (theta_deg - (70+2*ith))/2;
  //     phi = atan2(vCurv,uCurv) - MAGNETIC_NORTH*M_PI/180;

  //     // linear interpolation between upper and lower values of theta in the tables
  //     for(k=0; k<5; k++) for(l=0; l<7; l++) for(m=0; m<5; m++) {
  //     alpha[k][l][m] = (1-dth) * Horiz_alpha_tab[ith][k][l][m] +
  //     dth * Horiz_alpha_tab[ith+1][k][l][m];
  //     beta[k][l][m]  = (1-dth) * Horiz_beta_tab[ith][k][l][m]
  //     + dth * Horiz_beta_tab[ith+1][k][l][m];
  //     }

  //     // coefficients of the function of (r,psi) for this value of phi
  //     for(k=0; k<5; k++) for(l=0; l<7; l++) {
  //     Horiz_a[k][l] = Horiz_b[k][l] = 0;
  //     for(m=0; m<4; m++)  Horiz_a[k][l] += alpha[k][l][m] * cos(m*phi);
  //     if(l>0) {for(m=1; m<4; m++)  Horiz_b[k][l] += beta[k][l][m] * sin(m*phi);}
  //     }

  //     // minimization procedure
  //     SilentLDF = 0;
  //     FitLDFhorMinuit(xc, dxc, yc, dyc, fact, dfact, chi2, errflag);
  //     LDFFitOK = !errflag && chi2<(nActiveStation-3)*LDF_CHI2MAX
  //     && sqr(xc)+sqr(yc)-sqr(uCurrent*xc+vCurrent*yc) < sqr(LDF_DISTMAX_BARY_CORE);
  //     if(LDFFitOK) {
  //     chi2LDFwithSilent = chi2;
  //     fE = HORIZ_ENER*pow(fact,HORIZ_EXPONENT);
  //     fdE = HORIZ_EXPONENT*fE*(dfact/fact);
  //     fXCore = XBarycenter + xc;
  //     dXCore = dxc;
  //     fYCore = YBarycenter + yc;
  //     dYCore = dyc;
  //     }
  //     else {
  //     if(PrintLevel) printf("LDFhor fit failed - errflag %d chi2 %6.2f dist %3.0f\n",
  //     errflag,chi2LDF,sqrt(sqr(xc)+sqr(yc)-sqr(uCurrent*xc+vCurrent*yc)));
  //     return 1;
  //     }
  //     //SilentLDF = 1;
  //     //FitLDFhorMinuit(xc, dxc, yc, dyc, fact, dfact, chi2, errflag);
  //     //chi2LDFwithSilent = chi2;

  //     return 1;
}

/*-------------------------------------------------------------------*/
int TPBRec::ReadHorizTables()
/*-------------------------------------------------------------------*/
{
  FILE *tab;
  char filename[50];
  int ith, k, l, m;

  for(ith=0; ith<10; ith++) {
    sprintf(filename,"../files/horiz_prot_%d_%d.par",HORIZ_ENER,70+2*ith);
    tab = fopen(filename,"r");
    if(!tab) {
      if (PrintLevel) printf("!!!! tables for horizontal maps not found !\n");
      return 0;
    }
    for(k=0; k<5; k++) for(m=0; m<5; m++) fscanf(tab,"%lf",&Horiz_alpha_tab[ith][k][0][m]);
    for(k=0; k<5; k++) for(m=0; m<5; m++) Horiz_beta_tab[ith][k][0][m]=0;
    for(l=1; l<7; l++) {
      for(k=0; k<5; k++) for(m=0; m<5; m++) fscanf(tab,"%lf",&Horiz_alpha_tab[ith][k][l][m]);
      for(k=0; k<5; k++) {
        Horiz_beta_tab[ith][k][l][0] = 0;
        for(m=1; m<5; m++) fscanf(tab,"%lf",&Horiz_beta_tab[ith][k][l][m]);
      }
    }
  }
  return 1;
}

/*-------------------------------------------------------------------*/
int TPBRec::ReadTableHoriz(unsigned int event_id, char *model)
/*-------------------------------------------------------------------*/
{
  FILE *tab;
  char filename[50];

  sprintf(filename,"evt_hor_%u_%s.tab",event_id,model);
  tab = fopen(filename,"r");
  if(!tab) {
    if(PrintLevel) printf("file '%s' not found\n",filename);
    return 0;
  }
  fscanf(tab,"%lf %lf %lf %lf",&energ_hor,&phi_hor,&dx_grid,&dy_grid);
  phi_hor *= M_PI/180;
  cosphi_hor = cos(phi_hor);
  sinphi_hor = sin(phi_hor);

  for(int i=0; i<51; i++) for(int j=0; j<51; j++)
      fscanf(tab,"%lf",&nmu_hor[i][j]);
  for(int i=0; i<51; i++) for(int j=0; j<51; j++)
      fscanf(tab,"%lf",&signal_hor[i][j]);

  fclose(tab);
  return 1;
}


static void fcnLDFhor(int &npar, double *gin, double &f, double *par, int iflag) {
  double horiz_muon;

  if(0) printf("%d %f %d",npar,gin[0],iflag); // to avoid compiler warning

  // empirical approximation of the averaged signal of one muon
  if(thePBRec().HORIZ_MUON != 0) horiz_muon = thePBRec().HORIZ_MUON;            // fixed value
  else horiz_muon = 12/(10.2*cos(thePBRec().Theta)+4.3*sin(thePBRec().Theta));  // computed from angle

  double fact = exp(par[2]);

  f = 0;
  for (int i=0; i<thePBRec().nActiveStation; i++)
    if(thePBRec().InFit[i] && thePBRec().Satur1[i]<1000 && thePBRec().Satur2[i]<1000 && thePBRec().Satur3[i]<1000) {
      // "signal" is expressed here in units of averaged horizontal muons
      double signal = thePBRec().Signal[i] / horiz_muon;
      double dx = thePBRec().X[i] - thePBRec().XBarycenter - par[0];
      double dy = thePBRec().Y[i] - thePBRec().YBarycenter - par[1];
      double signal_th = fact*thePBRec().ldf_hor(dx,dy);
      //f += sqr(signal-signal_th)/signal_th;
      f -= log(Poisson(signal_th,(int)(signal+.5)));
    }
  if(thePBRec().SilentLDF) {
    for (int i=0; i<thePBRec().nSilentStation; i++) if(thePBRec().SilentInFit[i]) {
        double dx = thePBRec().XSilent[i] - thePBRec().XBarycenter - par[0];
        double dy = thePBRec().YSilent[i] - thePBRec().YBarycenter - par[1];
        //f += fact*ldf_hor(dx,dy) / HORIZ_SIGTHRESH;
        double signal_th = fact*thePBRec().ldf_hor(dx,dy);
        // simplified assumption: the probability of T1 with 1 muon is negligible - to be revised
        f -= log(Poisson(signal_th,0)+Poisson(signal_th,1));
      }
  }
}

// fitting the LDF of an horizontal event with Minuit
/*-------------------------------------------------------------------*/
void TPBRec::FitLDFhorMinuit(double& xc,double& dxc,double& yc,double& dyc,
                             double& fact,double& dfact, double& chi2, int& errflag)
/*-------------------------------------------------------------------*/
{
  Int_t ierflg;
  double horiz_muon;

  // empirical approximation of the averaged signal of one muon
  if(HORIZ_MUON != 0) horiz_muon = HORIZ_MUON;            // fixed value
  else horiz_muon = 12/(10.2*cos(Theta)+4.3*sin(Theta));  // computed from angle

  // first approximation for the scaling factor w.r.t. the reference energy of the table
  int n = 0; fact = 0;
  for (int i=0; i<nActiveStation; i++) if(InFit[i]) {
      fact += Signal[i]/horiz_muon/
              ldf_hor(X[i]-XBarycenter,Y[i]-YBarycenter);
      n++;
      if(n==3) break;
    }
  fact = log(fact/3);

  TMinuit* gMin = new TMinuit(3);
  gMin->SetPrintLevel(-1);
  gMin->SetFCN(fcnLDFhor);
  gMin->mnparm(0,"dx",0,100,0,0,ierflg);
  gMin->mnparm(1,"dy",0,100,0,0,ierflg);
  gMin->mnparm(2,"fact",fact,1,0,0,ierflg);

  gMin->Command("MIGRAD 500");
  Double_t amin, edm, errdef;
  Int_t npari, nparx;
  gMin->mnstat(amin,edm,errdef,npari,nparx,ierflg);
  errflag = (ierflg<3);

  gMin->GetParameter(0,xc,dxc);
  gMin->GetParameter(1,yc,dyc);
  gMin->GetParameter(2,fact,dfact);
  fact = exp(fact);
  dfact *= fact;

  //chi2 = gMin->fAmin;
  chi2=0;
  for (int i=0; i<nActiveStation; i++)
    if(InFit[i] && Satur1[i]<1000 && Satur2[i]<1000 && Satur3[i]<1000) {
      double signal = Signal[i] / horiz_muon;
      double signal_th = fact*ldf_hor(X[i]-XBarycenter-xc,Y[i]-YBarycenter-yc);
      chi2 += sqr(signal-signal_th)/signal_th;
    }

  if(PrintLevel>1) {
    printf(" sFit ");
    for (int i=0; i<nActiveStation; i++)  printf("%6.1f ",
          fact*horiz_muon*ldf_hor(X[i]-XBarycenter-xc,Y[i]-YBarycenter-yc));
    if(SilentLDF) printf("  LDFhors ");
    else printf("  LDFhor ");
    printf("xc %4.0f+-%3.0f  yc %4.0f+-%3.0f  fact %5.3f+-%4.3f  chi2 %6.2f\n",
           xc,dxc,yc,dyc,fact,dfact,chi2);
  }
  delete gMin;
}

#define RMAX 5000.
#define N_SLICE 40
/*-------------------------------------------------------------------*/
double TPBRec::ldf_hor(double x,double y)
/*-------------------------------------------------------------------*/
{
  int k,l;
  double xs,ys,r,xr,xx,psi,val;

  double sinth = sqrt(uCurv*uCurv+vCurv*vCurv);
  xs = ( x*uCurv + y*vCurv)/sinth*sqrt(1-sinth*sinth);
  ys = (-x*vCurv + y*uCurv)/sinth;

  r = sqrt(xs*xs+ys*ys);
  if(r>5000) return 1e-6;
  xr = 2*(sqrt(r/RMAX)-1./N_SLICE)-1;
  psi = atan2(ys,xs);
  val = 0; xx = 1;
  for(k=0; k<5; k++) {
    for(l=0; l<7; l++) {
      val += xx * Horiz_a[k][l] * cos(l*psi);
      if(l>0) val += xx * Horiz_b[k][l] * sin(l*psi);
    }
    xx *= xr;
  }
  return exp(val);
}
