#include <fstream>
#include <sstream>
#include "TCdasUtil.h"
#include "SdRecLpnhe.h"
#include <TMath.h> //JG

using namespace TSdParams;
char *gErOutputFormat = (char *) "RAW";

/*---------------------------------------------------------*/
///< Global variables that drives the reconstruction behavior. Can be modified using
//  their corresponding Set functions
/// By default, variable curvature is not used, the asymmetries of the signals are corrected,
/// the LDF uses the proton parametrization, and the LDF slope is not fitted.
/// The reconstruction is an iterative process which allows the rejection of stations with bad time residuals
/*---------------------------------------------------------*/
static bool gForceRec           = false;        ///<to force or not reconstruction; set by SetForceRec(bool)
static bool gOfficial           = false;        ///< Call OfficialSelection befroe reconstruction; set by SetOfficial(bool)
static bool gOptimize           = true;         ///<to optimize SelectedStations vector before reconstruction; set by SetOptimizedRec(bool)
static bool gCorrectAsymmetry   = true;         ///<to correct or not asymmetries; set by SetAsymmetry(bool)
static bool gUseSilent          = true;         ///< Use silent stations in fit; Set by SetSilent(bool,double);
static double gRCutSilent       = 5000.0;       ///< Use silent stations in fit up to RCutSilent meters. ; Set by SetSilent(bool,double);
static TPrimary gPrimary        = eProton;      ///< Primary to use in LDF parametrization; set by SetPrimary(Primaries)
static string gEnergyConverter = "LOG";         ///< To convert SRef <-> E; set by SetEnergyConverter() //S1000 ou S450
static bool gFixedCore          = false;        ///< Fix core during fit; Set by SetCore(double,double,double), unset with SetCore();
static int gVerbosity           = 1;            ///< Verbose Level. Set by SetVerbosity(int)
static TGeometry gRec1          = eGeomOptimal; ///< Geometry reconstruction type; Set by SetRec(TGeometry, TLDF)
static TLDF gRec2               = eLDFOptimal;  /// LDF reconstruction type. Set by SetRec...
static bool gGlobal             = true;         /// Fit all parameter at once. Set byt SetGlobal(bool)
static TVarianceTs gVariance     = eT50;        /// Variance model for Ts
static double gXCore = 0.0, gYCore = 0.0, gZCore = 1400.0;


static double _trackLength(double cosTheta) {
static double R = 1.8, h = 1.2;
double sinTheta = ((1-cosTheta*cosTheta)>0?sqrt(1-cosTheta*cosTheta):0);
return M_PI*R/(M_PI*R*cosTheta+2*h*sinTheta);
}

/*---------------------------------------------------------*/
// forward declarations of functions for minuit
/*---------------------------------------------------------*/
static void risetimefitfcn(int &, double *, double &f, double *par, int);
static void falltimefitfcn(int &, double *, double &f, double *par, int);
static void geomFit(int &, double *, double &f, double *par, int);
static void ldfFit(int &, double *, double &f, double *par, int);
static void globalFit(int &, double *, double &f, double *par, int);


static double _LDFfuncLogParabola(double dist, double beta, double gamma, double distref, double distcut);
static double _LDFfuncNKG(double dist, double beta, double gamma, double distref, double distcut);
static double (*_LDFfunc)(double, double, double, double, double) = &_LDFfuncLogParabola;
static int countDOF(TGeometry geom);
static int countDOF(TLDF ldf);

/*---------------------------------------------------------*/
// theLpnheRec() singleton constructor, destructor and access functions
/*---------------------------------------------------------*/
SdRecLpnhe * SdRecLpnhe::_instance = 0;
SdRecLpnhe::SdRecLpnhe():
  XmaxfromRiseTime(0),
  XmaxfromFallTime(0),
  fn_sRef(0),
  ftCore(0),
  fdBeta(0),
  ftau(0), 
  fdtau(0),
  falpha(0), 
  fdalpha(0),
  angularChi2(0),
  ldfChi2(0),
  globalChi2(0),
  residu(0),
  LDFFitOK(0),
  GlobalFitOK(0),
  ndof(0)
{
  SetDefault();
}
SdRecLpnhe::~SdRecLpnhe() {}
SdRecLpnhe *SdRecLpnhe::Instance() {
  if (_instance == 0) _instance = new SdRecLpnhe();
  return _instance;
}

SdRecLpnhe & theLpnheRec() /// returns the SdRecLpnhe singleton.
{
  return *SdRecLpnhe::Instance();
}

static TErEvent & event() {
  return *(theLpnheRec().fErEvent);
} /// for internal use.

/*---------------------------------------------------------*/
// Set functions...

/*---------------------------------------------------------*/
void SdRecLpnhe::SetGlobal(bool global)/// Set to true if one wsnts to fit all parameters jointly.
/*---------------------------------------------------------*/
{
  gGlobal = global;
}
/*---------------------------------------------------------*/
void SdRecLpnhe::SetOfficial(bool off)///Set to true if we want to finish reconstruction whatever are the time residuals
/*---------------------------------------------------------*/
{
  gOfficial = off;
}


/*---------------------------------------------------------*/
void SdRecLpnhe::SetForce(bool force)///Set to true if we want to finish reconstruction whatever are the time residuals
/*---------------------------------------------------------*/
{
  gForceRec = force;
}

/*---------------------------------------------------------*/
void SdRecLpnhe::SetOptimize(bool optimize)///Set to true to optimize SelectedStation vector before reconstruction
/*---------------------------------------------------------*/
{
  gOptimize = optimize;
}

/*---------------------------------------------------------*/
void SdRecLpnhe::SetAsymmetry(bool asym)///Do we want to correct the asymmetries or not ?
/*---------------------------------------------------------*/
{
  gCorrectAsymmetry = asym;
}

/*---------------------------------------------------------*/
void SdRecLpnhe::SetSilent(bool silent, double rCut)///Do we want to use silent stations or not ?
/*---------------------------------------------------------*/
{
  gUseSilent = silent;
  gRCutSilent = rCut;
}

/*---------------------------------------------------------*/
void SdRecLpnhe::SetVarianceModel(TVarianceTs var)///Set the variance model for Ts
/*---------------------------------------------------------*/
{
  gVariance = var;
}

/*---------------------------------------------------------*/
void SdRecLpnhe::SetPrimary(TPrimary prim)///Set the primary type for the LDF and energy parametrizations (eProton = 0, eIron = 1)
/*---------------------------------------------------------*/
{
  gPrimary = prim;
}

/*---------------------------------------------------------*/
void SdRecLpnhe::SetEnergyConverter(string e)///Set the primary type for the LDF and energy parametrizations (eProton = 0, eIron = 1)
/*---------------------------------------------------------*/
{
  gEnergyConverter = e;
}

/*---------------------------------------------------------*/
void SdRecLpnhe::SetVerbosity(int verb) /// Set the verbose level, 0 none, 1 reconstruction, 2 reconstruction and minuit
/*---------------------------------------------------------*/
{
  gVerbosity = 0;
  if (verb == 1) gVerbosity = 1;
  if (verb == 2) gVerbosity = 2;;
}

/*---------------------------------------------------------*/
void SdRecLpnhe::SetCore(double xc, double yc, double zc)///Impose core coordinates during fit. Put xc=yc=zc=0 to unset. Automatically unset on reading a new event.
/*---------------------------------------------------------*/
{
  gXCore = xc;
  gYCore = yc;
  gZCore = zc;
  if (xc == 0.0 && yc == 0.0 && zc == 1400.0) gFixedCore = false; else gFixedCore = true;
}

/*---------------------------------------------------------*/
void SdRecLpnhe::SetRec(TGeometry geom,///< is related to angular fits
                        TLDF ldf///< is related to energy fits
                       )///To set first (rec1) and second (rec2) types of reconstruction
/*---------------------------------------------------------*/
{
  /**
   * geom can be set to:
   * - SdLpnhe::ePlane: angular reconstruction without fitting the curvature, but a mean value is used to correct the stations time.
   * - SdLpnhe::eCurve: angular reconstruction including the curvature fit.
   * - SdLpnhe::eVCurve: angular reconstruction including the (variable) curvature fit.
   */

  /**
   * ldf can be set to : 
   * - SdLpnhe::eLog: the reconstruction of the core coordinates and the energy using a log like ldf
   * - SdLpnhe::eLogXmax: same as eLog but including the fit of the LDF slope, which finally leads to an estimation of the Xmax.
   * - SdLpnhe::eNKG: NKG-like parameterisation of the LDF.
   */
  gRec1 = geom;
  gRec2 = ldf;
}

/*---------------------------------------------------------*/
void SdRecLpnhe::SetErEvent(TErEvent* ev)
/*---------------------------------------------------------*/
{// this function is called by the TErEvent constructors
  SetCore(); // Unset core if it was fixed on previous event.
  Associate(ev);
}


/*---------------------------------------------------------*/
void SdRecLpnhe::SetDefault() /// Puts back everything to original default value.
/*---------------------------------------------------------*/
{
  SetGlobal(); SetForce(); SetOptimize(); SetAsymmetry(); SetOfficial();
  SetSilent(); SetPrimary(); SetVerbosity(); SetCore(); SetRec(); SetVarianceModel(); SetEnergyConverter();
}


/*---------------------------------------------------------*/
double SdRecLpnhe::Wgt(TCalibStation *st) /// Weight of station in time fit
/*---------------------------------------------------------*/
{ int i = 0;
while (&(event().fCalibStations[i]) != st) ++i;
return fWgt[i];
}


/*---------------------------------------------------------*/
void SdRecLpnhe::SetWgt(TCalibStation *st, double w)
/*---------------------------------------------------------*/
{ int i = 0;
while (&(event().fCalibStations[i]) != st) ++i;
fWgt[i] = w;
//cerr<<"In Er "<<i<<" "<<fWgt<<endl;
}

/*---------------------------------------------------------*/
int SdRecLpnhe::InitReconstruction()///Called before any independent reconstruction process
/*---------------------------------------------------------*/
{
  memset((TShowerParams *)this, 0, sizeof(TShowerParams));
  fWgt.clear(); fWgt.resize(event().fCalibStations.size(), 1.0);
  ftCore = 0; angularChi2 = ldfChi2 = globalChi2 = 0; ndof = 0;
  residu = LDFFitOK = GlobalFitOK = 0; XmaxfromRiseTime = XmaxfromFallTime = 0;
  for (unsigned int i = 0; i < event().fCalibStations.size(); i++)
    if (!event().fCalibStations[i].Error) {
      //event().fCalibStations[i].CorrectTraceTime(); // From PB reconstruction
      //fWgt[i] = 1.0; // Time fit default weight;
    }
  if (gOfficial) fErEvent->OfficialSelection();
  fLDFRcut = 300.;
  if(fErEvent->UseInfillRec){ //infill
    fRefDist = DISTREFINFILL;
    SetSilent(true,2500.0); 
    fLDFRcut/=2.;
  }
  else{ //normal
    fRefDist = DISTREF;
    SetSilent();
  }
  fGPSSecond=fErEvent->RefSecond();// fill GPS second
  _EstimateUV(); _EstimateXYZTCore();   _EstimateLDFPar(); 
  
  //T50(0, true); calculate event average T50
  //if (gOptimize && !(_OptimizeStationSelection() && _OptimizeStationSelection2())) return 0;
  if (gOptimize) _OptimizeStationSelection();
  if (gUseSilent) { // clean up a bit the silent vector to use only silent station at most gRCutSilent meters  away from core
    double dx, dy, d;
    for (unsigned int j = 0; j < event().fSilentStations.size(); j++) {
      dx = event().fSilentStations[j]->fX-theLpnheRec().fXCore;
      dy = event().fSilentStations[j]->fY-theLpnheRec().fYCore;
      d = distance2Core(dx, dy, fU, fV);
      if (d > gRCutSilent) {
        event().fBadStations.push_back(fErEvent->fSilentStations[j]);
        event().fSilentStations.erase(fErEvent->fSilentStations.begin() + j);
        --j;
      }
    }
  }
  //T50(0, true); calculate event average T50
  return 1;
}

/*---------------------------------------------------------*/
int SdRecLpnhe::XmaxOK()///Is Xmax estimation valid ?
//will surely change...(used by ED!)
/*---------------------------------------------------------*/
{
  unsigned int i, j, size = fErEvent->fSelectedStations.size();
  if (size < 6) return 0;
  int checked = 0;
  for (i = 0; i < size - 1; i++) {
    double r1 = fErEvent->Dist(*(fErEvent->fSelectedStations[i]));
    if (r1 < 500 || r1 > 1500) continue;
    for (j = i + 1; j < size; j++) {
      double r2 = fErEvent->Dist(*(fErEvent->fSelectedStations[j]));
      if (r2 < 500 || r2 > 1500) continue;
      if (fabs(r1 - r2) > 500) {
        checked++;
        break;
      }
    }
  }
  if (!checked) return 0;
  if (Theta("deg") < 60) return 1;
  else if (checked > 1) return 1;
  else return 0;
}

/*---------------------------------------------------------*/
static double variableRadius(double meanradius,///< mean radius of curvature
                             double dx,///< x distance to the core
                             double dy,///< y distance to the core
                             double u,///< direction vector of the shower axis
                             double v///< direction vector of the shower axis
                            )///Returns the value of the radius of curvature depending on the position of the station
/*---------------------------------------------------------*/
{
  double d2 = dx*dx+dy*dy;
  double d = sqrt(d2);
  double sintheta = sqrt(u*u + v*v);
  double theta = asin(sintheta);
  double alpha = M_PI_2+theta*(dx*u+dy*v)/(d*sintheta);
  return sqrt(sqr(meanradius) + d2 + 2*meanradius*d*cos(alpha));
}

/*---------------------------------------------------------*/
int SdRecLpnhe::_CountDoublets()///Checks if there are doublets in the event
/*---------------------------------------------------------*/
{
  int  cMInFit = 0, dNInFit = 0, mRInFit = 0;
  unsigned int size = fErEvent->fSelectedStations.size();
  for (unsigned int is = 0; is < size; is++) {
    if (fErEvent->fSelectedStations[is]->Id == kIoSd::CARMEN
        || fErEvent->fSelectedStations[is]->Id == kIoSd::MIRANDA)
      cMInFit++;
    if (fErEvent->fSelectedStations[is]->Id == kIoSd::MOULIN
        || fErEvent->fSelectedStations[is]->Id == kIoSd::ROUGE)
      mRInFit++;
    if (fErEvent->fSelectedStations[is]->Id == kIoSd::DIA
        || fErEvent->fSelectedStations[is]->Id == kIoSd::NOCHE)
      dNInFit++;
  }
  return (cMInFit > 1) + (mRInFit > 1) +(dNInFit > 1);
}

/*---------------------------------------------------------*/
double SdRecLpnhe::_Stat2Core(int statpos)///Returns the perpendicular distance from the station to the shower axis.
/*---------------------------------------------------------*/
{
  return distance2Core(event().fSelectedStations[statpos]->fX - fXCore, event().fSelectedStations[statpos]->fY - fYCore, fU, fV);
}

/*---------------------------------------------------------*/
double SdRecLpnhe::_StatDt(int statpos, int corrt /* Altitude correction */, TGeometry geom)
/*---------------------------------------------------------*/
{
  double dx = fErEvent->fSelectedStations[statpos]->fX - fXCore;
  double dy = fErEvent->fSelectedStations[statpos]->fY - fYCore;
  double r2 = sqr(dx) + sqr(dy) - sqr(fU * dx + fV * dy);
  double dt = fErEvent->fSelectedStations[statpos]->StartTime() * 1e+9 + corrt * CorrectTime(statpos) - ftCore;
  if (geom == ePlane && fR == 0.0)  return fT0 - (dx * fU + dy * fV) / kCdasUtil::CSPEED - dt;
  if (geom == ePlane || geom == eCurve)  return fT0 - (dx * fU + dy * fV - r2/(2.0*fR)) / kCdasUtil::CSPEED - dt;
  if (geom == eVCurve) return fT0 - (dx * fU + dy * fV - r2/(2*variableRadius(fR, dx, dy, fU, fV))) / kCdasUtil::CSPEED - dt;
  return 0.0;
}

/*---------------------------------------------------------*/
double SdRecLpnhe::_EstimateCurvature(double signal,double costheta,bool infillrec) /* Curvature radius estimation */
/*---------------------------------------------------------*/
{
  if (!infillrec)
    return 7500./costheta;

  //****** parametrization for theta < 55 deg. ***********
  const double a = -1.08; 
  const double b = 0.495;
  const double c = 0.193; 
  const double d = -0.0136;
  const double x = (1./costheta-1.);
  const double f1 = 1. + a*x + b*x*x;
  // not the current infill CIC, but it should not be changed 
  // as the paramtrisation is done for this attenuation curve 
  const double s35 = signal*exp(1.9*(x+1.-1./cos(35.*3.1415927/180.))); 
  const double f2 =  c+d*log(s35);
  //*********************************************************

  return (1000./(f1*f2)); // Radius of curvature in m.
}


//#define INITIAL_R (7500.0/W())
//#define INITIAL_R_INFILL (-3160.+9960./W())

/*---------------------------------------------------------*/
double SdRecLpnhe::CorrectTime(int i)/// Altitude correction to the time of the station
/*---------------------------------------------------------*/
{
  double costh = sqrt(maxval(0, 1 - sqr(fU) - sqr(fV)));
  return (costh * (fErEvent->fSelectedStations[i]->fZ - fZCore)) / kCdasUtil::CSPEED;
}

#define  DIST_MAX_ST 2500
#define  DIST_MAX_PROJ_ST 1500
/*---------------------------------------------------------*/
int SdRecLpnhe::_CompactConfig()///Is the configuration of selected stations compact enough ?
/*---------------------------------------------------------*/
{
  /**
   * Stations must be less than 1500*sqrt(n-2) perpendicularly from the shower axis, n being the number of selected stations, each present doublet counting for one station.
  */
  unsigned int i, size = fErEvent->fSelectedStations.size();
  // each doublet counts for one station :
  int ns = size - _CountDoublets();
  double dx, dy, r2 = 0;

  for (i = 0; i < size; i++) {
    dx = fErEvent->fSelectedStations[i]->fX - fXCore;
    dy = fErEvent->fSelectedStations[i]->fY - fYCore;
    r2 = maxval(r2, sqr(dx) + sqr(dy) - sqr(fU * dx + fV * dy)); //station to core axis
  }
  double dist2_max_proj = sqr(DIST_MAX_PROJ_ST) * (ns - 2);
  return (r2 < dist2_max_proj);
}

/*---------------------------------------------------------*/
double SdRecLpnhe::_EstimateTCore() /// Estimate time of ground impact using sqrt(sig) as stations weight
/*---------------------------------------------------------*/
{
  // Barycenter in space and time
  double wgt, sumw, tCore;
  unsigned int i, size = fErEvent->fSelectedStations.size();
  sumw = tCore = 0.;

  for (i = 0; i < size; i++) if (Wgt(event().fSelectedStations[i]) != 0.0) {
    wgt = sqrt(fErEvent->fSelectedStations[i]->fSigInVEM);
    sumw += wgt;
    tCore += wgt * fErEvent->fSelectedStations[i]->StartTime();
  }
  tCore = tCore * 1e+9 / sumw;
  return tCore;
}

/*---------------------------------------------------------*/
void SdRecLpnhe::_EstimateXYZTCore()///Estimates the core coordinates in the first steps of the reconstruction
/*---------------------------------------------------------*/
{
  ///The barycenter is the average of the coordinates of the stations,
  ///each coordinate being weighted by the root mean square of the corresponding signal.

  double wgt, sumw = 0;
  unsigned int i, size = fErEvent->fSelectedStations.size();
  ftCore = _EstimateTCore();
  if (gFixedCore) {
    fXCore = gXCore; fYCore = gYCore; fZCore = gZCore;
    return;
  }

  fXCore = fYCore = fZCore = 0.;
  for (i = 0; i < size; i++) if (Wgt(event().fSelectedStations[i]) != 0.0) {
    wgt = sqrt(fErEvent->fSelectedStations[i]->fSigInVEM);
    sumw += wgt;
    fXCore += wgt * fErEvent->fSelectedStations[i]->fX;
    fYCore += wgt * fErEvent->fSelectedStations[i]->fY;
    fZCore += wgt * fErEvent->fSelectedStations[i]->fZ;
  }
  fXCore /= sumw; fYCore /= sumw; fZCore /= sumw;
}

/*---------------------------------------------------------*/
int SdRecLpnhe::_EstimateUV()
/*---------------------------------------------------------*/
{
  int s0 = 0, s1 = 0, s2 = 0;
  double x0 = fErEvent->fSelectedStations[0]->fX;
  double y0 = fErEvent->fSelectedStations[0]->fY;
  double x1 = 0, x2 = 0, y1 = 0, y2 = 0;
  unsigned int i, j, size = fErEvent->fSelectedStations.size();

  for (i = 1; i < size; i++) if (Wgt(event().fSelectedStations[i]) != 0.0) {
    x1 = fErEvent->fSelectedStations[i]->fX - x0;
    y1 = fErEvent->fSelectedStations[i]->fY - y0;
    if (x1 * x1 + y1 * y1 > 250000) {
      s1 = i;
      break;
    }                       //no doublets
  }
  for (j = s1 + 1; j < size; j++) if (Wgt(event().fSelectedStations[i]) != 0.0) {
    x2 = fErEvent->fSelectedStations[j]->fX - x0;
    y2 = fErEvent->fSelectedStations[j]->fY - y0;
    if ((fabs(x1 * y2 - x2 * y1) / 2) > 100000) {
      s2 = j;
      break;
    }
  }
  if (s2 == 0) return 0;               //stations are aligned

  double t1 =
    fErEvent->fSelectedStations[s1]->StartTime() * 1e+9 -
    fErEvent->fSelectedStations[s0]->StartTime() * 1e+9;
  double t2 =
    fErEvent->fSelectedStations[s2]->StartTime() * 1e+9 -
    fErEvent->fSelectedStations[s0]->StartTime() * 1e+9;
  fU = kCdasUtil::CSPEED * ((t1 * y2 - t2 * y1) / (x2 * y1 - x1 * y2));
  fV = kCdasUtil::CSPEED * ((t1 * x2 - t2 * x1) / (x1 * y2 - x2 * y1));

  return 1;
}

/*---------------------------------------------------------*/
void SdRecLpnhe::_EstimateSRef()
/*---------------------------------------------------------*/
{
  double rmin = 99999;
  double distc = fRefDist;
  double sigc = 1;
  unsigned int i, size = fErEvent->fSelectedStations.size();

  for (i = 0; i < size; i++) {
    double r = _Stat2Core(i);
    if (fabs(r - fRefDist) < rmin) {
      rmin = fabs(r - fRefDist);
      sigc = fErEvent->fSelectedStations[i]->fSigInVEM;
      distc = r;
    }
  }
  fSRef = sigc / pow(distc / fRefDist, fBeta + fGamma * log(distc / fRefDist));

}


        
/*---------------------------------------------------------*/
double SdRecLpnhe::TimeVariance(int stat,
                                double dist,///< perpendicular distance from the station to the shower axis
                                double cth, ///< cos(theta)
                                TGeometry geom)///Time uncertainty for angular fits
/*---------------------------------------------------------*/
{
  switch (gVariance) {
  case eT50 : {
    TCalibStation *st = fErEvent->fSelectedStations[stat];
    double dt = st->T50() - st->fStartIntegration*kEc::TimeSlot;
    if (dt < 0 ) dt = 0;
    //double dt = T50(stat); This one use an average T50 for the whole event T50(r) = r*T50.
    double npart = st->SigInVEM()/_trackLength(cth);
    if (npart < 2) npart = 2;
        
    double var = 0;
    // correction a la Xavier -- March 2009
    if (dt < 0.4*(0.53*cth-0.11)*dist ) dt = 0.4*(0.53*cth-0.11)*dist; 
    double a2=0, b2=0;
    if (fErEvent->UseInfillRec){ //infill
      //a2 = 0.22; b2 = 148;
      a2 = 0.9506 - 1.685*cth + 0.9559*cth*cth;
      b2 = 89.54 + 94.8*cth - 7.4*cth*cth;
    }
    else {
    // New parameterization for a and b -- March 2009
    //a2 = 0.67612 + 0.16106*cth - 0.47641*cth*cth; 
    //b2 = 128 - 184*cth + 413*cth*cth;
    // Parameterization for a2 and b2 (paper 2011)
      a2 = 0.64871 + 0.22365*cth - 0.49971*cth*cth;
      b2 = 141.24 - 208.90*cth +412.19*cth*cth;
    }
    var = a2*sqr(2.*dt / npart)*((npart-1.0) / (npart + 1.0)) + b2;
    return var;
  }
  case eT70 : {
    TCalibStation *st = fErEvent->fSelectedStations[stat];
    double dt = st->T70() - st->fStartBin*kEc::TimeSlot - 50;
    if (dt < 0 ) dt = 0;
    double var = sqr(1.0*dt/st->SigInVEM()) + 400;
    return var;
  }
  case eT90 : {
    TCalibStation *st = fErEvent->fSelectedStations[stat];
    double dt = st->T90() - st->fStartBin*kEc::TimeSlot - 60;
    if (dt < 0 ) dt = 0;
    double var = sqr(0.8*dt/st->SigInVEM()) + 250;
    return var;
  }
  case eR4 :
  default : {
    double d = (dist<5000.0)? dist : 5000.0;
    double var = (sqr(25.) + sqr(5e-5 * sqr(d) * cth));
    // if (geom == ePlane) var *= 1.153; else var /= 4.2; !! not in PB code!
    return var;
  }
  }
}

class _TfitData {public: double chi2, residu, pointChi2, vFit;};
/*---------------------------------------------------------*/
int SdRecLpnhe::_OptimizeStationSelection(int loop) {
/*---------------------------------------------------------*/
  unsigned int size = fErEvent->fSelectedStations.size(); TGeometry geom = ePlane; int farStation, bad, badb;
  double chi2, pointChi2 = 0, pointChi2b = 0;
  vector<_TfitData> fitInfo;
  _TfitData fitData;
  if (loop == 1) {
    fU = fErEvent->fEstimated.fU; fV = fErEvent->fEstimated.fV; 
    fR = _EstimateCurvature(fSRef,W(),fErEvent->UseInfillRec);
  }
  if (loop > 2) return 1;
  int zero = 0;
  for (unsigned int i = 0; i < size; ++i) if (Wgt(event().fSelectedStations[i]) == 0.0) zero++;
  size -= zero;
  _EstimateXYZTCore();
  if (size == 3) {
    fR = _EstimateCurvature(fSRef,W(),fErEvent->UseInfillRec);
    _MatPlanFit(1, farStation);
    return 1;
  }
  if (size == 4) {
    fR = 0.0; geom = ePlane; _MatPlanFit(0, farStation); 
    fR = _EstimateCurvature(fSRef,W(),fErEvent->UseInfillRec);
    chi2 = _MatPlanFit(1, farStation);
    if (chi2 > 5) {
      for (int trial = 0; trial < 4; ++trial) {
        double dist = _Stat2Core(trial);
        double pChi2 = sqr(_StatDt(trial, 1, geom)) / TimeVariance(trial, dist, W(), geom);
        if (pChi2> pointChi2) {pointChi2 = pChi2; bad = trial;}
        else if (pChi2 > pointChi2b) {pointChi2b = pChi2; badb = trial;}
      }
      if (pointChi2 > 5 && fErEvent->fSelectedStations[bad]->SigInVEM() < 10.0) {
        // test for the second hihest if also above 5 but with smaller signal remove this tank (badb)
        if (pointChi2b > 5 && fErEvent->fSelectedStations[badb]->SigInVEM() < fErEvent->fSelectedStations[bad]->SigInVEM()) 
	  bad = badb;
	SetWgt(fErEvent->fSelectedStations[bad], 0.0);
        if (gVerbosity>1) cerr << fErEvent->AugerId <<" ++++++++++++++++++++++++++++ Removed station " << fErEvent->fSelectedStations[bad]->Id << " pChi2 =" << pointChi2 << " signal = " << fErEvent->fSelectedStations[bad]->SigInVEM() << endl;
      }
    }
    return 1;
  }
  geom = eCurve; _MatCurvedFit(0, farStation); chi2 = _MatCurvedFit(1, farStation);
  int ndf = countDOF(geom);
  if (ndf == 0 || TMath::Prob(chi2, ndf) > 0.15) return 1;
  for (unsigned trial = 0; trial < size; ++trial) {
    // remove station [i] calculate shower param with the rest and then residual of station [i].
    SetWgt(fErEvent->fSelectedStations[trial], 0.0); 
    if (size - 1 > 4) {
      geom = eCurve; _MatCurvedFit(0, farStation); fitData.chi2 = _MatCurvedFit(1, farStation);
    } 
    else {
      fR = 0.0; geom = ePlane; _MatPlanFit(0, farStation); 
      fR = _EstimateCurvature(fSRef,W(),fErEvent->UseInfillRec);
      fitData.chi2 = _MatPlanFit(1, farStation);
    }
    fitData.residu = _StatDt(trial, 1, geom); 
    fitData.pointChi2 = chi2 - fitData.chi2;
    fitInfo.push_back(fitData);
    SetWgt(fErEvent->fSelectedStations[trial], 1.0); // put station back 
    // We set back the Estimated values... We do not see a better way to do this
    // OBB & CBB
    fU = fErEvent->fEstimated.fU;
    fV = fErEvent->fEstimated.fV;
    fR = _EstimateCurvature(fSRef,W(),fErEvent->UseInfillRec);
    fT0 = fErEvent->fEstimated.fT0;
    //
    if (fitData.pointChi2 > pointChi2) {pointChi2=fitData.pointChi2; bad = trial;}
    else if (fitData.pointChi2 > pointChi2b) {pointChi2b=fitData.pointChi2; badb = trial;}
  }
//  cerr << "++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++" << endl;
//  cerr << " ndf = " << ndf << " chi2=" << chi2 << " prob=" << TMath::Prob(chi2, ndf) << endl;
//  for (unsigned trial = 0; trial < size; ++trial) {
//    cerr << fErEvent->fSelectedStations[trial]->Id << " S=" << fErEvent->fSelectedStations[trial]->SigInVEM() << " chi=";
//    cerr << fitInfo[trial].chi2 << " pchi=" << fitInfo[trial].pointChi2 << " res=" << fitInfo[trial].residu << endl;
//    }
  if (pointChi2 > 5 && fErEvent->fSelectedStations[bad]->SigInVEM() < 10.0) {
    // test for the second hihest if also above 5 but with smaller signal remove this tank (badb)
    if (pointChi2b > 5 && fErEvent->fSelectedStations[badb]->SigInVEM() < fErEvent->fSelectedStations[bad]->SigInVEM()) bad = badb;
    SetWgt(fErEvent->fSelectedStations[bad], 0.0);
    if (gVerbosity>1) cerr << fErEvent->AugerId <<" ++++++++++++++++++++++++++++ Removed station " << fErEvent->fSelectedStations[bad]->Id << " pChi2 =" << pointChi2 << " signal = " << fErEvent->fSelectedStations[bad]->SigInVEM() << endl;
    }
//  cout << "++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++" << endl;
  return _OptimizeStationSelection(++loop);
}

/*---------------------------------------------------------*/
int SdRecLpnhe::_TimeRec(TTimeRec rec)///Fits the rise/fall time at 1000 m, taking into account time asymmetries
/*---------------------------------------------------------*/
{
  ///Allows different determinations of the value of Xmax (cf. CalculateXmax() function), not correlated to the one given by the slope
  _errflag = 0;
  TMinuit *minuit = new TMinuit(2);
  double arglist[10];
  if(gVerbosity == 2) arglist[0] = 0;
  else arglist[0] = -1;            // Minuit not verbose
  minuit->mnexcm("SET PRI", arglist, 1, _errflag);
  minuit->mnexcm("SET ERR", arglist, 1, _errflag);
  minuit->mnexcm("SET NOW", arglist, 1, _errflag);

  if (rec == eRise) minuit->fFCN = risetimefitfcn; else minuit->fFCN = falltimefitfcn;
  ftau = 0.3;
  falpha = 0.4;
  minuit->mnparm(0, "Tau", ftau, 0.01, 0, 0, _errflag);
  minuit->mnparm(1, "Alpha", falpha, 0.01, 0, 0, _errflag);

  arglist[0] = 500;
  arglist[1] = 1.;
  //to cut remaining outputs of Minuit
  int fd = 0;
  if(gVerbosity < 2) {
    fd = dup(1); close(1);
  }
  minuit->mnexcm("MIGRAD", arglist, 2, _errflag);
//  minuit->mnexcm("MINOS", arglist, 0, _errflag);
  if(gVerbosity < 2) {
    dup2(fd, 1); close(fd);
  }
  //minuit->fAmin;
  if (_errflag != 0 && _errflag != 4) {
    if (gVerbosity)
      cout << "Time fit failed during Minuit." << endl;
    delete minuit;
    return 0;
  }

  minuit->GetParameter(0, ftau, fdtau);
  minuit->GetParameter(1, falpha, fdalpha);
  delete minuit;
  return 1;
}

// LDF paper + Markus e-mail of june 15th 4h10
static double _NKG_EFromSRef(double sRef, double cost) {
double a = 0.37-0.51/cost+0.30/cost/cost;
double b = 1.27-0.27/cost+0.08/cost/cost;
return a*pow(sRef,b);
}
static double _NKG_dEFromSRef(double sRef, double dsRef, double cost){
double b = 1.27-0.27/cost+0.08/cost/cost;
return _NKG_EFromSRef(sRef,cost)*b*dsRef/sRef;
}
static double _NKG_BetaFromE(double e, double cost) {
return  (-3.24 - 0.195*log10(e) + 0.98/cost);
}
static double _NKG_BetaFromSRef(double sRef, double cost) {
double e = _NKG_EFromSRef(sRef, cost);
return  _NKG_BetaFromE(e,cost);
}

/*---------------------------------------------------------*/
void SdRecLpnhe::_EstimateLDFPar(TLDF ldf)///Called before the fit of the LDF
/*---------------------------------------------------------*/
{ double cost = W(1.0);
  ///Calculates the values of fBeta, the LDF slope, and fGamma, the correction to the power law. Function ends by estimating SRef via _EstimateSRef() (necessary input of the fit).
  if (gPrimary == eProton || gPrimary == eIron) {
    double a = fU*fU + fV*fV;
    double theta = (a < 1) ? asin(sqrt(a)) : asin(1.0);
    GetBetaGammaFromTheta((int) gPrimary, theta, fBeta, fGamma);
  }
  else {                // Old one
    if (Theta("deg") < 58) {
      fBeta = -4.80 + 1.27 / cost - 0.717E-01 / pow(cost, 2);
      fGamma = -0.536 + 0.559 / cost - 0.231 / pow(cost, 2);
    } else {
      fBeta = -3.38 + 0.412 / cost;
      fGamma = -0.283 - 0.0287 / cost;
    }
  }
  if (fErEvent->UseInfillRec>0){
    double sectta=1./cos(Theta("rad"));
    //******* New parametrisation of beta, gamma
    //**** beta,gamma are function of x=1/sectta-1.27 and y=log(s450) -5

    const double b0 =2.74;    const double b1=-1.12;     const double b2 = -0.15;    
    const double b3=1.06;     const double b4 = 0.032; 
    const double g0 = 0.224;   const double g1 = 0.024;   
    
    double x=sectta-1.27;    double y=0; 

    fBeta=-1*(b0+b1*x+b2*x*x)*(b3+b4*y);    fGamma=-1*(g0+g1*y);
    
    fdBeta=0; fdGamma=0; // see later for errors...

    //******* Iterations to determine beta,gamma***
    for (unsigned int i=0;i<5;++i){
      _EstimateSRef(); //*** estimation of s450
      y=log(fSRef)-5.;
      fBeta=-1*(b0+b1*x+b2*x*x)*(b3+b4*y);    fGamma=-1*(g0+g1*y);
    }
    
  }
  _EstimateSRef(); // this estimate use log-parabola ldf according to initial values of beta and gamma computed above.
  if (ldf == eLog || ldf == eLogXmax) {
    fE = compute_energy(gEnergyConverter, fSRef, Theta(), gPrimary);
  }
  else if (ldf == eNKG) {
    // ICRC 05 LDF paper, properly corrected...
    // use energy dependent beta. beta = -2.26 - 0.195*log(e) + 0.98*(sect -1).
    // i.e beta = -3.24 - 0.195*log(e) + 0.98/cost;
    fE = _NKG_EFromSRef(fSRef, cost);
    fBeta = _NKG_BetaFromE(fE, cost);  
  }
  if (ldf == eLogXmax && (!fXmax || isnan(fXmax))) 
    fXmax = 56 * log10(fE * 1E18) - 307;
}


float xxtab[] =
  { 0, 50, 100, 150, 200, 250, 300, 350, 400, 450, 500, 550, 600, 650,
    700, 750, 800, 850, 900, 950, 1000, 1050, 1100
  };
float betatab[] =
  { -4.072, -4.07147, -4.05518, -4.02376, -3.97884, -3.92206, -3.85506,
    -3.77947, -3.69692, -3.60905, -3.5175, -3.4239, -3.32988, -3.23708,
    -3.14714,
    -3.06169, -2.98236, -2.91079, -2.84862, -2.79748, -2.759, -2.73482,
    -2.72658
  };

/*---------------------------------------------------------*/
void SdRecLpnhe::_CalculateXmax()///Estimates the value of Xmax at the end of the reconstruction
/*---------------------------------------------------------*/
{
  ///After the fit of all the parameters, estimates the value of Xmax with the value of the LDF slope, and with the values of the rise and fall times at r = 1000 m.
  double a, b, c, d;
  d = -4.071;
  c = -0.188E-03;
  b = 0.368E-05;
  a = -0.218E-08;
  float x;
  unsigned int i;
  unsigned int size = (unsigned int) (sizeof(xxtab) / sizeof(float));
  if (fBeta < betatab[0])
    x = interpol(betatab[0], betatab[1], xxtab[0], xxtab[1], fBeta);
  else {
    for (i = 0; i < size - 1; i++) {
      if (fBeta > betatab[i] && fBeta < betatab[i + 1]) break;
    }
    if (i == size)
      x = interpol(betatab[size - 2], betatab[size - 1],
                   xxtab[size - 2], xxtab[size - 1], fBeta);
    else
      x = interpol(betatab[i], betatab[i + 1], xxtab[i],
                   xxtab[i + 1], fBeta);
  }
  fXmax = XVGD / W() - x;
  fXmax *= W();
  if (isnan(fXmax)) fXmax = 0;

  // Be careful, the followoing is
  //only tested at 10^20 eV and theta=36, 45 and 53 °
  // with risetimes :
  _TimeRec(eRise);
  a = 0.49423 - ftau / 1000;
  b = -0.67354E-03;
  c = 0.25580E-06;
  double delta = sqr(b) - 4 * a * c;
  double num = -b + sqrt(delta);
  double x1 = num / (2 * a);
  XmaxfromRiseTime = XVGD / W() - 1 / x1;
  XmaxfromRiseTime *= W();
  if (isnan(XmaxfromRiseTime)) XmaxfromRiseTime = 0;
  //... and with falltimes :
  _TimeRec(eFall);
  a = 1.0459 - ftau / 1000;
  b = -0.59940E-03;
  c = -0.34316E-06;
  delta = sqr(b) - 4 * a * c;
  num = -b + sqrt(delta);
  x1 = num / (2 * a);
  XmaxfromFallTime = XVGD / W() - 1 / x1;
  XmaxfromFallTime *= W();
  if (isnan(XmaxfromFallTime)) XmaxfromFallTime = 0;
}

static void minuitSetFcn(TMinuit* minuit, TFit fit, TGeometry geom, TLDF ldf) {
  double d1,d2,d3; int n = -1; int flag = 0;
  if (fit == eGeom || fit == eGlobal) {
    switch (geom) {
    case eGeomNone : flag = -100; break;
    case ePlane : flag = -101; break;
    case eCurve: flag = -102; break;
    case eVCurve: flag = -103; break;
    case eGeomOptimal: cerr << "Problem in minuit set function" << endl; exit(1); break;
    }
    geomFit(n, &d1, d2, &d3, flag);
    if (fit == eGeom ) {
      minuit->fFCN = geomFit; return;
    }
  }
  if (fit == eLDF || fit == eGlobal) {
    switch (ldf) {
    case eLDFNone : flag = -100; break;
    case eLog : flag = -101; break;
    case eLogXmax : flag = -102; break;
    case eNKG : flag = -103; break;
    case eLDFOptimal: cerr << "Problem in minuit set function" << endl; exit(1); break;
    }
    ldfFit(n, &d1, d2, &d3, flag);
    if (fit == eLDF ) {
      minuit->fFCN = ldfFit; return;
    }
  }
  minuit->fFCN = globalFit;
  return;
}

static int countDOF(TGeometry geom) {
  int size = event().fSelectedStations.size();
  int zero = 0;
  for (int i = 0; i < size; i++) if (theLpnheRec().Wgt(event().fSelectedStations[i]) == 0.0) zero++;
  return (geom == ePlane)? size - 3 - zero: size - 4 - zero;
}

static int countDOF(TLDF ldf) {
  int ndf = event().fSelectedStations.size();
  if (!gFixedCore) ndf -= 2;
  return (ldf == eLogXmax)? ndf - 2: ndf - 1;
}

TMinuit* SdRecLpnhe::_NewMinuit(TFit fit, TGeometry geom, TLDF ldf) {
  double arglist[10];
  TMinuit *minuit = new TMinuit(8);
  ndof = 0;
  if(gVerbosity == 2) arglist[0] = 0; else arglist[0] = -1;    // Minuit not verbose
  minuit->mnexcm("SET PRI", arglist, 1, _errflag);
  minuit->mnexcm("SET ERR", arglist, 1, _errflag);
  minuit->mnexcm("SET NOW", arglist, 1, _errflag);

  double ka = (fR == 0.0) ? 0.0 : 1.0/(2*fR);
  minuit->mnparm(0, "T0", fT0, 0.0, 0, 0, _errflag);
  minuit->mnparm(1, "u", fU, 0.0, 0, 0, _errflag);
  minuit->mnparm(2, "v", fV, 0.0, 0, 0, _errflag);
  minuit->mnparm(3, "k", ka, 0.0, 0, 0, _errflag);
  minuit->mnparm(4, "xc", 0.0, 0.0, 0, 0, _errflag);
  minuit->mnparm(5, "yc", 0.0, 0.0, 0, 0, _errflag);
  minuit->mnparm(6, "lsRef", log(fSRef), 0.0, 0, 0, _errflag);
  minuit->mnparm(7, "beta", fBeta, 0.0, 0, 0, _errflag);

  if (fit == eGeom || fit == eGlobal) {
    minuit->mnparm(0, "T0", fT0, 1, 0, 0, _errflag);
    minuit->mnparm(1, "u", fU, 0.05, 0, 0, _errflag);
    minuit->mnparm(2, "v", fV, 0.05, 0, 0, _errflag);
    //ndof +=  fErEvent->fSelectedStations.size(); ndof -= 3;
    ndof +=  countDOF(geom);
    if (geom != ePlane)  { /*ndof -= 1;*/ minuit->mnparm(3, "k", ka, 0.01, 0, 0, _errflag);
    }
    if (fit == eGeom) {
      minuitSetFcn(minuit, fit, geom, ldf); return minuit;
    }
  }
  if (fit == eLDF || fit == eGlobal) {
    //ndof +=  fErEvent->fSelectedStations.size();
    ndof +=  countDOF(ldf);
    if (!gFixedCore) {
      minuit->mnparm(4, "xc", 0.0, 10.0, 0, 0, _errflag);
      minuit->mnparm(5, "yc", 0.0, 10.0, 0, 0, _errflag);
      //      ndof -= 2;
    }
    minuit->mnparm(6, "lsRef", log(fSRef), 0.1, 0, 0, _errflag);
    //    ndof -= 1;
    if (ldf == eLogXmax) {/*ndof -= 1;*/ minuit->mnparm(7, "beta", fBeta, 0.01, 0, 0, _errflag);
    }
    if (fit == eLDF) {
      minuitSetFcn(minuit, fit, geom, ldf); return minuit;
    }
  }
  minuitSetFcn(minuit, fit, geom, ldf);
  return minuit;
}

bool SdRecLpnhe::_Minimize(TMinuit *minuit) {
  double arglist[10]; int _errflag;
  int fd = 0;
  arglist[0] = 500; arglist[1] = 1.;
  if(gVerbosity < 2) {
    fd = dup(1); close(1);
  }
  minuit->mnexcm("MINIMIZE", arglist, 2, _errflag);
//  minuit->mnexcm("MINOS", arglist, 0, _errflag);
  if(gVerbosity < 2) {
    dup2(fd, 1); close(fd);
  }
  if (_errflag != 0 && _errflag != 4) {
    if (gVerbosity)
      cerr << "Minimize  failed during Minuit." << endl;
    return false;               //To change ?
  }
  return true;
}

double SdRecLpnhe::_GetResults(TMinuit *minuit, TFit fit, TGeometry geom, TLDF ldf) {
  Double_t cova[8][8]; // order t, u, v, r, xc,  yc, sRef, beta
  minuit->mnemat(&cova[0][0], 8);
  unsigned int offset = 0; // Minuit does not return in cova the fixed parameters so...
  if (fit == eGlobal) offset = 4;
  if (fit == eGlobal  && geom == ePlane) offset = 3;
  double lsRef, dlsRef, xc, yc, ka, dka, dK2dR;
  if (fit == eGeom || fit == eGlobal) {
    minuit->GetParameter(0, fT0, fVarCov[eT0][eT0]);
    minuit->GetParameter(1, fU, fVarCov[eU][eU]);
    minuit->GetParameter(2, fV, fVarCov[eV][eV]);
    minuit->GetParameter(3, ka, dka);
   
    if (ka != 0.0) fR = 1.0 / (2.0 * ka); else fR = 0.0;
    if (geom != ePlane) dK2dR = -2.0*fR*fR; else dK2dR = 0.0;
    fVarCov[eT0][eT0] = cova[0][0];
    fVarCov[eT0][eU] = fVarCov[eU][eT0] = cova[0][1];
    fVarCov[eT0][eV] = fVarCov[eV][eT0] = cova[0][2];
    fVarCov[eT0][eR] = fVarCov[eR][eT0] = cova[0][3]*dK2dR;
    fVarCov[eU][eU] = cova[1][1];
    fVarCov[eU][eV] = fVarCov[eV][eU] = cova[1][2];
    fVarCov[eU][eR] = fVarCov[eR][eU] = cova[1][3]*dK2dR;
    fVarCov[eV][eV] = cova[2][2];
    fVarCov[eV][eR] = fVarCov[eR][eV] = cova[2][3]*dK2dR;
    fVarCov[eR][eR] = cova[3][3]*dK2dR*dK2dR;
  }
  if (fit == eLDF || fit == eGlobal) {
    minuit->GetParameter(4, xc, fVarCov[eXCore][eXCore]);
    fXCore += xc;
    minuit->GetParameter(5, yc, fVarCov[eYCore][eYCore]);
    fYCore += yc;
    minuit->GetParameter(6, lsRef, dlsRef);
    fSRef = exp(lsRef);
    if (ldf == eLog || ldf == eLogXmax) fE = compute_energy(gEnergyConverter, fSRef,Theta(), gPrimary, dlsRef*fSRef, &fdE);
    else if (ldf == eNKG) {
      fE = _NKG_EFromSRef(fSRef,cos(Theta()));
      fdE = _NKG_dEFromSRef(fSRef, dlsRef*fSRef, cos(Theta()));
      fBeta = _NKG_BetaFromE(fE, cos(Theta()));
    }
    if (ldf == eLogXmax)
      minuit->GetParameter(7, fBeta, fdBeta);
    
    if (ldf == eLogXmax) _CalculateXmax();
    fVarCov[eXCore][eXCore] = cova[offset][offset];
    fVarCov[eXCore][eYCore] = fVarCov[eYCore][eXCore] = cova[offset][offset+1];
    fVarCov[eXCore][eSRef] = fVarCov[eSRef][eXCore] = cova[offset][offset+2]*fSRef;
    //  fVarCov[eXCore][eXmax] = fVarCov[eXmax][eXcore]  FIXME = f(cova[offset][offset+3]);
    fVarCov[eYCore][eYCore] = cova[offset+1][offset+1];
    fVarCov[eYCore][eSRef] = fVarCov[eSRef][eYCore] = cova[offset+1][offset+2]*fSRef;
    //  fVarCov[eYCore][eXmax] = fVarCov[eXmax][eYCore] FIXME = f(cova[offset+1][offset+3]);
    fVarCov[eSRef][eSRef] = cova[offset+2][offset+2]*fSRef*fSRef;
    //  fVarCov[eSRef][eXmax] = fVarCov[eXmax][eSRef]  FIXME = f(cova[offset+2][offset+3]);
    //  fVarCov[eXmax][eXmax]  FIXME = f(cova[offset+3][offset+3]);
    if (gPrimary != eProton && gPrimary != eIron) {
      double cost = cos(Theta()); double n_sRef;
      if (Theta("deg") < 58) n_sRef = -1.07 + 25.8 / cost - 23.5 / pow(cost, 2) + 5.7 / pow(cost, 3); else n_sRef = 4.77 - 1.30 / cost;
      fE = pow(fSRef/n_sRef,1./0.95);
      fdE = fE*dSRef()/fSRef/0.95;
    }
    // if (fit == eGlobal) { } FIXME (il manques une partie de VarCov....)
    // Recomputing an average T0 if we just fitted the ldf
    if (fit == eLDF) {
      fT0 = 0; unsigned int size = fErEvent->fSelectedStations.size();
      for (unsigned int i = 0; i < size; i++) {
        double d = _Stat2Core(i);
        fT0 += ((fErEvent->fSelectedStations[i]->fX - fXCore) * fU + (fErEvent->fSelectedStations[i]->fY - fYCore) * fV)
               / kCdasUtil::CSPEED + (fErEvent->fSelectedStations[i]->StartTime() * 1e+9 + CorrectTime(i) - ftCore);
        if (fR) fT0 += (-sqr(d) / (2 * fR)) / kCdasUtil::CSPEED;
      }
      fT0 /= size;
    }
  }
  fT0 += ftCore; ftCore = 0.0;
  fChi2 = minuit->fAmin;
  return fChi2;
}

/*---------------------------------------------------------*/
double SdRecLpnhe::_TimeResiduals(TGeometry geom, bool & changed)
/*---------------------------------------------------------*/
{
  bool allowChanges = changed; changed = false;
  double residu = 0.; double chi2 = 0.0, pChi2, dist, cth = W();
  int max = -1, ndf, size; double smin = 5.0*_trackLength(cth), pChi2max = 9.0; // 3 sigmas
  size = fErEvent->fSelectedStations.size();
  ndf = countDOF(geom);
  for (unsigned int i = 0; i < fErEvent->fSelectedStations.size(); i++) if (Wgt(fErEvent->fSelectedStations[i]) != 0.0) {
      pChi2= sqr(_StatDt(i, 1, geom));
      residu += pChi2;
      dist = _Stat2Core(i);
      pChi2 /= TimeVariance(i, dist, cth, geom);
      chi2 += pChi2;
      if (pChi2 > pChi2max  && fErEvent->fSelectedStations[i]->SigInVEM() < smin) {
        smin = fErEvent->fSelectedStations[i]->SigInVEM();
        max = i;
      }
    }
  residu = (ndf > 0)? sqrt(residu) / ndf : sqrt(residu);
  double prob = (ndf > 0) ? TMath::Prob(chi2, ndf) : 1.0;
  if (prob < 0.05 && max != -1 && gOptimize && allowChanges) {
  //if (gVerbosity)
  cout << "++++ **** Ignoring Station " << fErEvent->fSelectedStations[max]->Id << ", at " << _StatDt(max, 1, geom) << "±"<<sqrt(TimeVariance(max, _Stat2Core(max), cth, geom)) << " ns, during time fit. **** ++++"<<endl;
    SetWgt(fErEvent->fSelectedStations[max], 0.0);
    //    fErEvent->fRandomStations.push_back(fErEvent->fSelectedStations[max]);
    //    fErEvent->fSelectedStations.erase(fErEvent->fSelectedStations.begin() + max);
    changed = true;
  }
  return residu;
}

TGeometry SdRecLpnhe::_Best(TGeometry g) {
  int ndf = countDOF(g);
  if ((g == eCurve || g == eVCurve) && ndf < 0) {
    if (gVerbosity) cerr << "Not enough stations to fit a curved front. Will use fixed Radius of "<< (int) fR <<" meters" << endl;
    return ePlane;
  }
  if (g == eGeomOptimal) {
    if (ndf < 1) return ePlane;
    /* if (ndf < 4) */ return eCurve; // eVCuvre does not improve it seems FIXME ????
    // return eVCurve;
  }
  return g;
}

// THIS IS OLD
// From Pierre Billoir Thoughts....
// - if there are enough stations (especially at distances around 1000 m), a fit
//with a variable LDF slope is attempted (B is set as adjustable in (3)); the
//condition is:
//. at least 5 stations, with one of the following requirements:
//. at least 2 within (500,1500), with a difference at least 500 in between
//. at least 3 within (500,1500), with a maximal difference at least 400
//. at least 4 within (500,1500), with a maximal difference at least 300
// END THIS IS OLD
// New version much stricter
bool SdRecLpnhe::_FitBeta() {

 if(fErEvent->UseInfillRec){ //infill 
   return 0; // now we use the fixed beta,gamma parametrization

   int n450 = 0; double d, dmax = 0, c1,c2; int size = event().fSelectedStations.size();
   if (size < 5) return false;
   for (int i = 0; i < size; i++) if (((c1 = _Stat2Core(i)) > 300) && (c1 < 900)) {
       if (event().fSelectedStations[i]->fSat) continue;
       n450++;
       for (int j = i+1; j < size; j++) if (((c2 = _Stat2Core(j)) > 300) && (c2 < 900)) {
           if (event().fSelectedStations[j]->fSat) continue;
           d =  (c2 > c1 ? c2 - c1 : c1 - c2);
           if (d > dmax) dmax = d;
         }
     }
   fFitBeta = ((n450 > 1 && dmax > 300) || (n450 > 2 && dmax > 240) || (n450 > 3 && dmax > 180));
   return fFitBeta;
   
 }
 else{ //normal
   
  int nRef = 0; double d, dmax = 0, c1,c2; int size = event().fSelectedStations.size();
  if (size < 5) return false;
  for (int i = 0; i < size; i++) if (((c1 = _Stat2Core(i)) > 400) && (c1 < 1600)) {
      if (event().fSelectedStations[i]->fSat) continue;
      nRef++;
      for (int j = i+1; j < size; j++) if (((c2 = _Stat2Core(j)) > 400) && (c2 < 1600)) {
          //d =  sqrt(sqr(event().fSelectedStations[i]->fX - event().fSelectedStations[j]->fX) + sqr(event().fSelectedStations[i]->fY - event().fSelectedStations[j]->fY));
          if (event().fSelectedStations[j]->fSat) continue;
          d =  (c2 > c1 ? c2 - c1 : c1 - c2);
          if (d > dmax) dmax = d;
        }
    }
  fFitBeta= ((nRef > 1 && dmax > 900) || (nRef > 2 && dmax > 800) || (nRef > 3 && dmax > 700));
  return fFitBeta;
 }
}

TLDF SdRecLpnhe::_Best(TLDF ldf) {
  int ndf = countDOF(ldf); TLDF ret = ldf;
  if(ldf == eLogXmax && ndf < 0) {
    if (gVerbosity) cerr << "Not enough stations to fit the LDF slope" << endl;
    ret =  eLog;
  }
  if (ldf == eLDFOptimal) {
    if (_FitBeta()) ret = eLogXmax; else ret =  eLog;
  }
  return ret;
}

/*---------------------------------------------------------*/
int SdRecLpnhe::GeometricalRec(TGeometry rec)///User-callable angular reconstruction. options ePlane, eCurve, eVCurve.
/*---------------------------------------------------------*/
{ int res;
  ///Barycenter is used for the core coordinates unless gFixedCore is "true"
  if (!InitReconstruction()) return 0;
  fR = _EstimateCurvature(fSRef,W(),fErEvent->UseInfillRec);
  if (rec == ePlane) fR = 0.0;
  TGeometry geom = _Best(rec);
  res = _GeometricalRec(geom);
  fErEvent->fRec = *this;
  if (res && gVerbosity) AngularDump(geom, eLDFNone);
  return res;
}

/*---------------------------------------------------------*/
int SdRecLpnhe::_GeometricalRec(TGeometry  Geom, bool allow)
/*---------------------------------------------------------*/
{ TMinuit *minuit = 0; _errflag = 0;
 // fR = INITIAL_R; if (kind == ePlane) fR = 0.0;
 // TGeometry geom = _Best(kind);
  minuit = _NewMinuit(eGeom, Geom, eLDFNone);
  if (!_Minimize(minuit)) {
    delete minuit; return 0;
  }
  angularChi2 = _GetResults(minuit, eGeom, Geom, eLDFNone);
  delete minuit;
  if (isnan(Theta()) || isnan(Phi())) return 0;
  // We are done with geometry update estimated LDF par...
  //_EstimateLDFPar();
  //Calculate time residuals
  bool allowChanges = allow;
  residu = _TimeResiduals(Geom, allowChanges);
  if (allowChanges) return 2;
  if (gForceRec) return 1;
  return 1; // Always return 1... FIXME??  (ndof == 0 || TMath::Prob(angularChi2, ndof) > 1e-5);
}

/*---------------------------------------------------------*/
int SdRecLpnhe::LDFRec(TLDF LdfKind ////< LDF reconstruction type  eLog or eXmax or eNKG
                      )///Determines the core coordinates, the energy (via SRef) and the slope of the LDF (if possible and asked)
/*---------------------------------------------------------*/
{ int res;
  if (!InitReconstruction()) return 0;
  TLDF ldf = _Best(LdfKind);
  _EstimateLDFPar(ldf); 
  res = _LDFRec(ldf);
  fErEvent->fRec = *this;
  if (res && gVerbosity) LDFDump(ePlane, _Best(ldf));
  return res;
}

/*---------------------------------------------------------*/
int SdRecLpnhe::_LDFRec(TLDF ldf)
/*---------------------------------------------------------*/
{ TMinuit *minuit = 0; _errflag = 0;
 // TLDF ldf = _Best(kind);
 // _EstimateLDFPar(ldf);
  minuit = _NewMinuit(eLDF, eGeomNone, ldf);
  if (!_Minimize(minuit)) {
    delete minuit; minuit = NULL;
    if (ldf != eLogXmax) return 0;
  }
  if (minuit == NULL) { // we try just with eLog
    ldf = eLog;
    minuit = _NewMinuit(eLDF, eGeomNone, ldf);
    if (!_Minimize(minuit)) {delete minuit; return 0;}
  }
  ldfChi2 = _GetResults(minuit, eLDF, eGeomNone, ldf);
  delete minuit;
#if 0 // FIXME
  //--- quality criteria
  LDFFitOK = (!_errflag &&
              // chi2 of LDF fit (active stations only)
              ldfChi2 < CHI2LDFMAX &&
              // increase of chi2 of LDF fit due to silent stations
              rLdfChi2 - ldfChi2 < DCHI2LDFMAX &&
              // distance (in space) from barycenter to fitted axis
              sqr(xc) + sqr(yc) - sqr(fU * xc + fV * yc) <
              sqr(DIST_BARY_CORE_MAX));

  if (_verbose && !LDFFitOK)
    cerr << "LDF fit failed - errflag: " << _errflag << " chi2: " <<
    ldfChi2 << " dchi2: " << rLdfChi2 -
    ldfChi2 << " dist: " << sqrt(sqr(xc) + sqr(yc) -
                                 sqr(fU * xc + fV * yc)) << endl;

  return LDFFitOK;
#endif
  if (gForceRec) return 1;
  if (ldfChi2 > CHI2LDFMAX) {
    if (gVerbosity) {
      cerr << "LDF fit failed, chi2 too great with ";
      if (gUseSilent)  cerr << "silent stations" << endl; else cerr << "no silent stations" << endl;
    }
    return 0;
  }
  return 1;
}

/*---------------------------------------------------------*/
int SdRecLpnhe::CompleteRec(TGeometry Geom, TLDF Ldf)///Fits all the shower parameters at once
/*---------------------------------------------------------*/
{int res;
  if (!InitReconstruction()) return 0;
  fR = _EstimateCurvature(fSRef,W(),fErEvent->UseInfillRec); 
  if (Geom == ePlane) fR = 0.0;
  TGeometry geom = _Best(Geom); TLDF ldf = _Best(Ldf); _EstimateLDFPar(ldf); 
  res = _CompleteRec(geom, ldf);
  fErEvent->fRec = *this;
  if (res  && gVerbosity) CompleteDump(geom, ldf); else if (!res) cerr<< "Complete fit failed...."<<endl;
  return res;
}


/*---------------------------------------------------------*/
int SdRecLpnhe::_CompleteRec(TGeometry geom, TLDF ldf)
/*---------------------------------------------------------*/
{ TMinuit* minuit = 0;
  GlobalFitOK = 0; _errflag = 0; globalChi2 = 0;

  //fR = INITIAL_R; if (Geom == ePlane) fR = 0.0;
  //TGeometry geom = _Best(Geom);
  //TLDF ldf = _Best(LDF);
  //_EstimateLDFPar(ldf); //FIXME do we need this always? (second pass?)
  minuit = _NewMinuit(eGlobal, geom, ldf);
  if (!_Minimize(minuit)) {
	delete minuit; minuit = NULL;
	if (ldf != eLogXmax) return 0;
  }
  if (minuit == NULL) { // we have eLogXmax so try just with eLog
    ldf = eLog;
    minuit = _NewMinuit(eGlobal, geom, ldf);
    if (!_Minimize(minuit)) {delete minuit; return 0;}
  }
  globalChi2 = _GetResults(minuit, eGlobal, geom, ldf);
  delete minuit;
  if (isnan(Theta()) || isnan(Phi())) return 0;

  //  if (gForceRec) return 1;
  //  Calculate time residuals
  //  bool allowChange = false;
  //  residu = _TimeResiduals(geom, allowChange);
  //  if (TMath::Prob(globalChi2, ndof)<1e-5) {
  //    if (gVerbosity) cerr << "Complete fit failed, bad residu: " << residu << endl;
  //    return 0;
  //  }
  GlobalFitOK = 1;
  return 1;
}

bool SdRecLpnhe::_Reconstruct()///Reconstruct function inherited from TReconstruct
{
  ///This private function is called by TReconstruct::Reconstruct()
  //The reconstruction types are by default  geom = eGeomOptimal and ldf = eLDFOptimal),
  //but they can be reset by a call to SetRec(...).
  if (!InitReconstruction()) return false;
  fR = _EstimateCurvature(fSRef,W(),fErEvent->UseInfillRec);
  if (gRec1 == ePlane) fR = 0.0;
  TGeometry geom = _Best(gRec1); TLDF ldf = _Best(gRec2); _EstimateLDFPar(ldf); 
  bool step1 = (bool) _Reconstruct(geom, ldf, false);
  fErEvent->fRec = *this;
  if (!gGlobal || Theta("deg") >= 70 ) 
    return step1;
  bool step2  = (bool) _Reconstruct(geom, ldf, true);
  fErEvent->fRec = *this;
  return step2;
}


/*---------------------------------------------------------*/
int SdRecLpnhe::_Reconstruct(TGeometry geom, ///< Geometrical reconstruction (ePlane, eCurve or eVCurve)
                             TLDF ldf,          ///< LDF reconstruction (eLog or eXmax or eNKG). To use Silent station call SetSilent(true).
                             bool global)           ///< Fit goemetry  then LDF or everything globally
/*---------------------------------------------------------*/
{
  if (fErEvent->fSelectedStations.size() < 3) {
    if (gVerbosity) cerr << "Not enough valid stations for the reconstruction..."  << endl;
    return 0;
  }
  //  int stat;
  //  _MatPlanFit(1, stat); // to check if we will do Horizontal or normal LDF fit
  if (Theta("deg") >= 70 && fErEvent->fDoHorizontalRec == true 
      && _GeometricalRec(eCurve) && Theta("deg") >= 70) {
    int res = HorizontalReconstruction();
    if ((res || gForceRec) && gVerbosity)
      HorizontalReconstructionDump(geom);
    return res || gForceRec;
  }
  if (global) {/*# global reconstruction # */
    GlobalFitOK = _CompleteRec(geom, ldf);
    if (GlobalFitOK  && gVerbosity) 
      CompleteDump(geom, ldf);
    else 
      if (!GlobalFitOK) 
        cerr<< "Complete fit failed...."<<endl;
    return GlobalFitOK;
  }
  int res =  _GeometricalRec(geom);
  // now filling the shower params
  fChi2 = angularChi2;
  if (!res) return 0;
  if (gVerbosity) AngularDump(geom, eLDFNone);
  res = _LDFRec(ldf);
  globalChi2 = angularChi2 + ldfChi2;
  fChi2 = globalChi2;
  if (res && gVerbosity) 
    LDFDump(geom, ldf);
  return 1;
}

/*---------------------------------------------------------*/
double SdRecLpnhe::_MatPlanFit(int corrt, int &farthestStation)///Simple matrix inversion to calculate T0 and the angles of the shower
/*---------------------------------------------------------*/
{
  unsigned int is;
  double wgt, dx, dy, dt, sumw, sumx, sumy, sumx2, sumy2, sumxy, sumt,
  sumxt, sumyt;
  double rxx, ryy, rxy, rx, ry, rr, deter, chi2, r2, dr;
  unsigned int size = fErEvent->fSelectedStations.size();
  sumw = sumx = sumy = sumx2 = sumy2 = sumxy = sumt = sumxt = sumyt = 0;
  for (is = 0; is < size; is++) 
    if (Wgt(fErEvent->fSelectedStations[is]) != 0.0){
      double dist = _Stat2Core(is);
      double sigma2 = TimeVariance(is, dist, W(), ePlane);
      wgt = 1 / sigma2;
      dx = fErEvent->fSelectedStations[is]->fX - fXCore;
      dy = fErEvent->fSelectedStations[is]->fY - fYCore;
      r2 = sqr(dx) + sqr(dy) - sqr(fU * dx + fV * dy);
      if (fR !=  0.0) 
        dr = (- r2/(2.0*fR)) / kCdasUtil::CSPEED; 
      else 
        dr = 0.0;
      dt = fErEvent->fSelectedStations[is]->StartTime() * 1e+9 + corrt * CorrectTime(is) + dr - ftCore;
      sumw += wgt;
      sumx += wgt * dx;
      sumy += wgt * dy;
      sumx2 += wgt * sqr(dx);
      sumy2 += wgt * sqr(dy);
      sumxy += wgt * dx * dy;
      sumt += wgt * dt;
      sumxt += wgt * dt * dx;
      sumyt += wgt * dt * dy;
    }
  rxx = sumw * sumy2 - sqr(sumy);
  ryy = sumw * sumx2 - sqr(sumx);
  rxy = sumx * sumy - sumw * sumxy;
  rx = sumxy * sumy - sumx * sumy2;
  ry = sumxy * sumx - sumy * sumx2;
  rr = sumx2 * sumy2 - sqr(sumxy);
  deter = sumx2 * rxx + sumxy * rxy + sumx * rx;
  if (deter == 0) {
    _errflag = 1; return -1.0;
  }
  fU = (rxx * sumxt + rxy * sumyt + rx * sumt) / deter;
  fV = (rxy * sumxt + ryy * sumyt + ry * sumt) / deter;
  fT0 = (rx * sumxt + ry * sumyt + rr * sumt) / deter;
  fU *= (-1.0*kCdasUtil::CSPEED);
  fV *= (-1.0*kCdasUtil::CSPEED);

  fVarCov[eT0][eT0] = rr / deter;
  fVarCov[eT0][eU] = fVarCov[eU][eT0] = sqr(kCdasUtil::CSPEED) * rx / deter;
  fVarCov[eT0][eV] = fVarCov[eV][eT0] = sqr(kCdasUtil::CSPEED) * ry / deter;
  fVarCov[eU][eU] = sqr(kCdasUtil::CSPEED) * rxx / deter;
  fVarCov[eU][eV] = fVarCov[eV][eU] = sqr(kCdasUtil::CSPEED) * rxy / deter;
  fVarCov[eV][eV] = sqr(kCdasUtil::CSPEED) * ryy / deter;

  _errflag = 0; chi2 = residu = 0;
  double cth = W();
  if (isnan(cth)) {
    _errflag = 1; return -1.0;
  }
  double pointChi2, pointChi2Max = 0;
  farthestStation = -1;
  for (is = 0; is < size; is++) if (Wgt(fErEvent->fSelectedStations[is]) != 0.0){
    double dist = _Stat2Core(is);
    pointChi2 = sqr(_StatDt(is, corrt, ePlane)) / TimeVariance(is, dist, cth, ePlane);
    if ( pointChi2 > pointChi2Max && fErEvent->fSelectedStations[is]->SigInVEM() < 5) {
      pointChi2Max = pointChi2; farthestStation = is;
    }
    chi2 += pointChi2;
  }
  return chi2;
}

/*---------------------------------------------------------*/
double SdRecLpnhe::_MatCurvedFit(int corrt, int &farthestStation )///Simple matrix inversion to calculate T0, the angles of the shower and the curvature
/*---------------------------------------------------------*/
{
  ///_MatPlanFit(int) and this fucntion are alternatives to the use of Minuit. They are used in OptimizedGeometricalRec(), the function which rejects bad time residual stations.
  int i;
  unsigned int is;
  double wgt, dx, dy, r2, dt, s[10], t[4], c[10];
  double up, vp, curvp, chi2, parab, dparab;
  unsigned int size = fErEvent->fSelectedStations.size();
  up = -fU / kCdasUtil::CSPEED;
  vp = -fV / kCdasUtil::CSPEED;
  for (i = 0; i < 10; i++) s[i] = 0;
  for (i = 0; i < 4; i++) t[i] = 0;
  for (is = 0; is < size; is++) if (Wgt(fErEvent->fSelectedStations[is]) != 0.0) {
    double r = _Stat2Core(is);
    double sigma2 = TimeVariance(is, r, W(), eCurve);
    wgt = 1 / sigma2;
    dx = fErEvent->fSelectedStations[is]->fX - fXCore;
    dy = fErEvent->fSelectedStations[is]->fY - fYCore;
    r2 = sqr(dx) + sqr(dy) - sqr(fU * dx + fV * dy);
    dt = fErEvent->fSelectedStations[is]->StartTime() * 1e+9 + corrt * CorrectTime(is) - ftCore;
    s[0] += wgt;
    s[1] += wgt * dx;
    s[2] += wgt * dx * dx;
    s[3] += wgt * dy;
    s[4] += wgt * dy * dx;
    s[5] += wgt * dy * dy;
    s[6] += wgt * r2;
    s[7] += wgt * r2 * dx;
    s[8] += wgt * r2 * dy;
    s[9] += wgt * r2 * r2;
    t[0] += wgt * dt;
    t[1] += wgt * dt * dx;
    t[2] += wgt * dt * dy;
    t[3] += wgt * dt * r2;
  }
  _errflag = !matinv4(s, c);
  if (_errflag) return -1.0;

  fT0 = c[0] * t[0] + c[1] * t[1] + c[3] * t[2] + c[6] * t[3];
  up = c[1] * t[0] + c[2] * t[1] + c[4] * t[2] + c[7] * t[3];
  vp = c[3] * t[0] + c[4] * t[1] + c[5] * t[2] + c[8] * t[3];
  curvp = c[6] * t[0] + c[7] * t[1] + c[8] * t[2] + c[9] * t[3];

  fU = -up * kCdasUtil::CSPEED;
  fV = -vp * kCdasUtil::CSPEED;
  parab = curvp * kCdasUtil::CSPEED;
  _errflag = 0;
  dparab = c[9] * pow(kCdasUtil::CSPEED, 4);

  fVarCov[eT0][eT0] = c[0];
  fVarCov[eT0][eU] = fVarCov[eU][eT0] = c[1] * sqr(kCdasUtil::CSPEED);
  fVarCov[eT0][eV] = fVarCov[eV][eT0] = c[3] * sqr(kCdasUtil::CSPEED);
  fVarCov[eU][eU] = c[2] * sqr(kCdasUtil::CSPEED);
  fVarCov[eU][eV] = fVarCov[eV][eU] = c[4] * sqr(kCdasUtil::CSPEED);
  fVarCov[eV][eV] = c[5] * sqr(kCdasUtil::CSPEED);

  fR = 1. / (2 * parab);
  fVarCov[eR][eR]= dparab / sqr(2 * parab * parab);

  _errflag = 0;
  chi2 = 0, residu = 0;
  double cth = W();
  if (isnan(cth)) {
    _errflag = 1; return -1.0;
  }
  double pointChi2, pointChi2Max = 0;
  farthestStation = -1;
  for (is = 0; is < size; is++) if (Wgt(fErEvent->fSelectedStations[is]) != 0.0){
    double dist = _Stat2Core(is);
    pointChi2 = sqr(_StatDt(is, corrt, eCurve)) / TimeVariance(is, dist, cth, eCurve);
    if ( pointChi2 > pointChi2Max && fErEvent->fSelectedStations[is]->SigInVEM() < 5) {
      pointChi2Max = pointChi2; farthestStation = is;
    }
    chi2 += pointChi2;
  }
  return chi2;
}


#include <THorReconstruct.h>
/*---------------------------------------------------------*/
int SdRecLpnhe::HorizontalReconstruction()///Performs an horizontal reconstruction when theta > 70 deg
/*---------------------------------------------------------*/
{ TShowerParams par;
  ///The horizontal reconstruction is performed by the THorReconstruct class.
  THorReconstruct horrec;
//  if (!_GeometricalRec(eCurve)) return 0;
  par = *this;
  *((TShowerParams *) &horrec) = par;
  horrec.FillInputs(Theta("deg"), Phi("deg"), fXCore, fYCore);
  horrec.Verbose = gVerbosity;
  horrec.SetEDPlot(fErEvent->fHorPlot);
  if ((horrec.Reconstruct(*fErEvent)) == false ) return 0;
  par = horrec;
  *((TShowerParams *) this) = par;
  if (!_GeometricalRec(eCurve)) return 0;
  //int stat;
  //_MatCurvedFit(1, stat);

  if (Theta("deg") > 84) {
    horrec.FillInputs(Theta("deg"), Phi("deg"), fXCore, fYCore);
    if ( (horrec.Reconstruct(*fErEvent)) == false ) return 0;
    par = horrec;
    *((TShowerParams *) this) = par;
  }


  fVarCov[eXCore][eXCore] = sqr(horrec.dDXCore);
  fVarCov[eYCore][eYCore] = sqr(horrec.dDYCore);
  //fPhi = horrec.Phi*kCdasUtil::DEG2RAD;
  fVarCov[eSRef][eSRef] = sqr(fErEvent->fRec.fdE);

  for (int iy = 0; iy < 52; iy++) {
    for (int ix = 0; ix < 52; ix++) {
      fErEvent->fEDMapMu[ix][iy] = horrec.EDMapMu[ix][iy];
      fErEvent->fEDMapMuBis[ix][iy] = horrec.EDMapE[ix][iy];
    }
  }
  fErEvent->fEDdmax = (horrec.EDdxmax>horrec.EDdymax) ? horrec.EDdxmax: horrec.EDdymax;
  fErEvent->fEDdmax += 200;
  fChi2 = angularChi2;
  return 1;
}



/*######################################################*/
/*####################### DUMPS ########################*/
/*######################################################*/

/*---------------------------------------------------------*/
ostream & SdRecLpnhe::put(ostream & s, char *option)
/*---------------------------------------------------------*/
{
  s.setf(std::ios::fixed, std::ios::floatfield)
    ;
  int number = 0;
  for (unsigned int i = 0; i < fErEvent->Trigger.NumberOfStation; i++)
    if (!fErEvent->Stations[i].Error) number++;
  if (strncmp(option, "SUM", 3) == 0)
    s << std::setw(8) << std::setfill('0') << fErEvent->Id
    << " " << fErEvent->UTCTime << " " << number << endl;
  else
    s << std::setw(8) << std::setfill('0') << fErEvent->Id
    << " " << fErEvent->UTCTime << " " << fErEvent->Stations.size() << endl;
  //for (unsigned int i=0;i<fErEvent->Stations.size();i++)   FIXME   Was it used ???
  //s<< (RecSdStation)fErEvent->Stations[i];
  return s;
}

ostream & operator<<(ostream & s, SdRecLpnhe & r) {
  return r.put(s, gErOutputFormat);
}

/*---------------------------------------------------------*/
void SdRecLpnhe::StationDump(int pos, TGeometry geom)
/*---------------------------------------------------------*/
{ double dist = _Stat2Core(pos), sig = sqrt(TimeVariance(pos, dist, W(), geom)), dt = _StatDt(pos, 1, geom) - ftCore;
  cout.setf(std::ios::fixed, std::ios::floatfield);
  cout << std::setprecision(2);
  cout << "  Station " << fErEvent->fSelectedStations[pos]->Id<< ", " <<
  (int) dist << " m from core, " <<
  fErEvent->fSelectedStations[pos]->fSigInVEM << " VEM, expected " <<
    fSRef*_LDFfunc(dist,theLpnheRec().fBeta,theLpnheRec().fGamma, theLpnheRec().fRefDist,theLpnheRec().fLDFRcut) << " at time " <<
  (int) (fErEvent->fSelectedStations[pos]->StartTime()*1e+9) << " ns, dt = " <<  dt << " ± " << sig;
  if (Wgt(fErEvent->fSelectedStations[pos]) != 0.0) cout << " (" << dt/sig << ")" << endl;
  else cout << " (- 0 -)" << endl;
}

/*---------------------------------------------------------*/
void SdRecLpnhe::HeaderDump()
/*---------------------------------------------------------*/
{
  cout <<
  "#########################################################################################"
  << endl;
  //cout<< "##  Event "<< id<<" is at "<< (int)fYCore<< " fY, "<< (int)fXCore<< " fX and second "<< fErEvent->fSelectedStations[0]->gps.second<< ". ##"<< endl;
  cout << "##  Event " << fErEvent->AugerId << " (" <<  fErEvent->UTCDate.substr(0, 24) << ") is at X="
  << (int) fXCore << " and Y=" << (int) fYCore << "##" << endl;
  cout << "## Options in use : ";
  if (gUseSilent) cout << " Silent Stations / ";
  if (gFixedCore) cout << "Fixed Core / ";
  if (gOfficial) cout << "Official Selection / ";
  if (gForceRec) cout << "Forced reconstruction / ";
  if (gOptimize) cout <<"Optimized Selection / ";
  if (gCorrectAsymmetry) cout <<"Correct Asymmetries / ";
  if (gGlobal) cout << " All parameter fit / ";
  if (gPrimary == eProton) cout <<" Proton primaries." << endl;;
  if (gPrimary == eIron) cout <<" Iron primaries." << endl;;
  if (gPrimary != eIron && gPrimary != eProton) cout <<" Uknown primaries!" << endl;;
  cout << "#########################################################################################" << endl << endl;
}

/*---------------------------------------------------------*/
void SdRecLpnhe::RecDump(TLDF ldf)
/*---------------------------------------------------------*/
{
  cout.setf(std::ios::fixed, std::ios::floatfield);
  cout << std::setprecision(2);
  cout << " T0 = " << (int) (fT0) << " ns" << endl;
  cout << " theta = " << Theta() << " rad = " << Theta("deg")
  << " ± " << dTheta("deg") << " deg" << endl;
  cout << " phi = " << Phi() << " rad = " << Phi("deg")
  << " ± " << dPhi("deg") << " /sin(theta) deg" << endl;
  cout << " Ra = " << Ra() << " deg" << endl << " Dec = " << Dec() << " deg" << endl;
  if (fR) cout << " R = " << (int) fR << " ± " << (int) dR() << " m" << endl;
  cout << " X = " << (int) fXCore << " ± " << (int) dXCore() << " m" << endl;
  cout << " Y = " << (int) fYCore << " ± " << (int) dYCore() << " m" << endl;

  if (Theta("deg") < 70) {
  if (ldf != eNKG)  cout << " LDF: " << fSRef << "* r(km)**(" << fBeta << "+" << fGamma
    << "*log(r))" << endl;
  if (ldf == eLogXmax)  cout << " Xmax (preliminary) = " << std::setprecision(0) <<
      fXmax << " ± " << dXmax() << " g/cm²" << endl;
  if (ldf == eNKG) cout << "LDF: " << fSRef << " * {(r/Ref)((r+700)/1700)}^"<<fBeta<<endl;
    cout << std::setprecision(2);
    if(fErEvent->UseInfillRec) {// infill
      cout << " S(450) = " << fSRef << " ± " << dSRef() << " VEM" << endl;
    } 
    else {
      cout << " S(1000) = " << fSRef << " ± " << dSRef() << " VEM" << endl;
    }
  }
  if (fE && !isnan(fE))
    cout << " E = " << fE << " EeV ± " << 100 * fdE / fE << "%" << endl;
  cout <<
  "#########################################################################################"
  << endl << endl;
}

/*---------------------------------------------------------*/
void SdRecLpnhe::HorizontalResultsDump()
/*---------------------------------------------------------*/
{
  cout.setf(std::ios::fixed, std::ios::floatfield);
  cout << std::setprecision(2);
  cout << " theta = " << Theta() << " rad = " << Theta("deg")
  << " ± " << dTheta("deg") << " deg" << endl;
  cout << " phi = " << Phi() << " rad = " << Phi("deg")
  << " ± " << dPhi("deg") << " /sin(theta) deg" << endl;
  if (fR) cout << " R = " << (int) fR << " ± " << (int) dR() << " m" << endl;
  cout << " X = " << (int) fXCore << " ± " << (int) dXCore() << " m" << endl;
  cout << " Y = " << (int) fYCore << " ± " << (int) dYCore() << " m" << endl;
  cout << " E = " << fE << " EeV ± " << 100 * fdE / fE << "%" << endl;
  cout <<
  "#########################################################################################"
  << endl << endl;

}

/*---------------------------------------------------------*/
void SdRecLpnhe::AngularDump(TGeometry geom, TLDF ldf)
/*---------------------------------------------------------*/
{
  HeaderDump();
  cout.setf(std::ios::fixed, std::ios::floatfield);
  cout << std::setprecision(2);
  cout << "After angular reconstruction, Chi2  " <<
  angularChi2 << " for " << ndof << " degrees of freedom," << endl;
  unsigned int i, size = fErEvent->fSelectedStations.size();
  for (i = 0; i < size; i++) StationDump(i, geom);
  RecDump(ldf);
  cout.unsetf(std::ios::floatfield);
  cout << std::setprecision(6);
}

/*---------------------------------------------------------*/
void SdRecLpnhe::LDFDump(TGeometry geom, TLDF ldf)
/*---------------------------------------------------------*/
{
  cout.setf(std::ios::fixed, std::ios::floatfield);
  cout << std::setprecision(2);
  cout << "After energy reconstruction, Chi2 " <<
  ldfChi2 << " for " << ndof << " degrees of freedom." << endl;
  unsigned int i, size = fErEvent->fSelectedStations.size();
  for (i = 0; i < size; i++) StationDump(i, geom);
  RecDump(ldf);
  cout.unsetf(std::ios::floatfield);
  cout << std::setprecision(6);
}

/*---------------------------------------------------------*/
void SdRecLpnhe::HorizontalReconstructionDump(TGeometry geom)
/*---------------------------------------------------------*/
{ HeaderDump();
  cout.setf(std::ios::fixed, std::ios::floatfield);
  cout << std::setprecision(2);
  cout << "After horizontal reconstruction ";
  cout << " (angular Chi2  " << angularChi2 << " for " << ndof << " degrees of freedom)." << endl;
  unsigned int i, size = fErEvent->fSelectedStations.size();
  for (i = 0; i < size; i++) StationDump(i, geom);
  HorizontalResultsDump();
  cout.unsetf(std::ios::floatfield);
  cout << std::setprecision(6);
}

/*---------------------------------------------------------*/
void SdRecLpnhe::CompleteDump(TGeometry geom, TLDF ldf)
/*---------------------------------------------------------*/
{
  HeaderDump();
  cout.setf(std::ios::fixed, std::ios::floatfield);
  cout << std::setprecision(2);
  if (!GlobalFitOK && !gForceRec) {
    cout << "Reconstruction failed." << endl;
    return;
  }                           //including nan cases...
  cout << "After complete reconstruction, with Chi2 "<< globalChi2<< " for " << ndof << " degrees of freedom." << endl;
  unsigned int i, size = fErEvent->fSelectedStations.size();
  for (i = 0; i < size; i++) StationDump(i, geom);
  RecDump(ldf);
  cout.unsetf(std::ios::floatfield);
  cout << std::setprecision(6);
}

//------------------------------------------------------------------------------------------------------
// Auxiliary function for Minuit

static void risetimefitfcn(int &, double *, double &f, double *par, int) {
  f = 0;
  unsigned int i, size = event().fSelectedStations.size();

  for (i = 0; i < size; i++) {
    float dx = event().fSelectedStations[i]->fX - theLpnheRec().fXCore;
    float dy = event().fSelectedStations[i]->fY - theLpnheRec().fYCore;
    float r = distance2Core(dx, dy, theLpnheRec().fU, theLpnheRec().fV);
    if (r < 500 || r > 2500) continue;
    float Y = cos(theLpnheRec().Phi()) * dy - sin(theLpnheRec().Phi()) * dx;
    float X = (sin(theLpnheRec().Phi()) * dy + cos(theLpnheRec().Phi()) * dx) * cos(theLpnheRec().Theta());
    float azimut = atan2(Y, X);
    float rt = event().fSelectedStations[i]->T50() -
               event().fSelectedStations[i]->T10();
    f += sqr(rt - par[0] * (r / 1000) * (1 + par[1] * cos(azimut))) *
         event().fSelectedStations[i]->fSigInVEM;
  }
}

static void falltimefitfcn(int &, double *, double &f, double *par, int) {
  f = 0;
  unsigned int i, size = event().fSelectedStations.size();

  for (i = 0; i < size; i++) {
    float dx = event().fSelectedStations[i]->fX - theLpnheRec().fXCore;
    float dy = event().fSelectedStations[i]->fY - theLpnheRec().fYCore;
    float r = distance2Core(dx, dy, theLpnheRec().fU, theLpnheRec().fV);
    if (r < 500 || r > 2500) continue;
    float Y = cos(theLpnheRec().Phi()) * dy - sin(theLpnheRec().Phi()) * dx;
    float X = (sin(theLpnheRec().Phi()) * dy + cos(theLpnheRec().Phi()) * dx) * cos(theLpnheRec().Theta());
    float azimut = atan2(Y, X);
    float ft = event().fSelectedStations[i]->T90() -
               event().fSelectedStations[i]->T50();
    f += sqr(ft - par[0] * (r / 1000) * (1 + par[1] * cos(azimut))) *
         event().fSelectedStations[i]->fSigInVEM;
  }
}


static void geomFit(int &n, double *, double &f, double *par, int flag) {
  static TGeometry geom = eGeomNone;
  f = 0.0;
  if (n == -1 && flag  == -100) {
    if (gVerbosity) cout << "-------------------------------------------> Setting Fit geometry to eGeomNone  "<< endl; geom = eGeomNone; return;
  }
  if (n == -1 && flag  == -101) {
    if (gVerbosity) cout << "-------------------------------------------> Setting Fit geometry to ePlane "<< endl; geom = ePlane; return;
  }
  if (n == -1 && flag  == -102) {
    if (gVerbosity) cout << "-------------------------------------------> Setting Fit geometry to eCurve "<< endl; geom = eCurve; return;
  }
  if (n == -1 && flag  == -103) {
    if (gVerbosity) cout << "-------------------------------------------> Setting Fit geometry to eVCurve"<< endl; geom = eVCurve; return;
  }
  unsigned int i, size = event().fSelectedStations.size();
  double dx, dy, dist, dt, sigma2;
  double cth = maxval(sqrt(1 - par[1] * par[1] - par[2] * par[2]), 0);
  for (i = 0; i < size; i++) if (theLpnheRec().Wgt(event().fSelectedStations[i]) != 0.0) {
      dx = event().fSelectedStations[i]->fX - theLpnheRec().fXCore - par[4];
      dy = event().fSelectedStations[i]->fY - theLpnheRec().fYCore - par[5];
      dist = distance2Core(dx, dy, par[1], par[2]);
      dt = event().fSelectedStations[i]->StartTime()*1e+9 + theLpnheRec().CorrectTime(i) - theLpnheRec().ftCore;
      sigma2 = theLpnheRec().TimeVariance(i, dist, cth, geom);
      if (par[3] == 0.0) {
        f += sqr(par[0] - (dx*par[1] + dy*par[2])/kCdasUtil::CSPEED - dt)/ sigma2; continue;
      }
      double r = 1.0 / (2.0 * par[3]);
      if (geom == eVCurve)  r = variableRadius(r, dx, dy, par[1], par[2]);
      f += sqr(par[0] - (dx*par[1] + dy*par[2])/kCdasUtil::CSPEED + pow(dist,2)/(2*r*kCdasUtil::CSPEED) - dt)/ sigma2;
    }
}

static double _LDFfuncLogParabola(double dist, double beta, double gamma, double distref, double distcut) { // dist in m.
 // log-log parabola (not valid close the core)
  if (dist>distcut) // distcut=1500 m
    return pow(dist/distref,beta+gamma*log(dist/distref));
  // power law at short distance (with smooth connection to the log-log parabola)
  else {
    double fact = 1/pow(distcut/distref,beta+gamma*log(distcut/distref));
    double slope = beta+2*gamma*log(distcut/distref);
    return pow(dist/distcut,slope)/fact;
  }
}

// ICRC 05 LDF paper...
static double _LDFfuncNKG(double dist, double beta, double gamma, double distref, double distcut) { // dist in m.
	static double rs = 700.0, r0 = 1000.0; 
	return pow((dist/r0)*(dist+rs)/(r0+rs), beta);
}

static void ldfFit(int &n, double *, double &f, double *par, int flag) {
  static TLDF ldf = eLDFNone;
  f = 0.0;
  if (n == -1 && flag  == -100) {
    if (gVerbosity) cout << "-------------------------------------------> Setting LDF Fit type to eLDFNone  "<< endl; ldf = eLDFNone; return;
  }
  if (n == -1 && flag  == -101) {
    if (gVerbosity) cout << "-------------------------------------------> Setting LDF Fit type to eLog  "<< endl; ldf = eLog; _LDFfunc = &_LDFfuncLogParabola; return;
  }
  if (n == -1 && flag  == -102) {
    if (gVerbosity) cout << "-------------------------------------------> Setting LDF Fit type to eLogXmax  "<< endl; ldf = eLogXmax; _LDFfunc = &_LDFfuncLogParabola; return;
  }
  if (n == -1 && flag  == -103) {
    if (gVerbosity) cout << "-------------------------------------------> Setting LDF Fit type to eNKG  "<< endl; ldf = eNKG; _LDFfunc = &_LDFfuncNKG; return;
  }
  f = 0;
  double funcval;
  double d, dx, dy, sig,df=0;
  unsigned int i, size = event().fSelectedStations.size(); 
  double sRef = exp(par[6]);
  double gamma = theLpnheRec().fGamma;
  double sin2t = par[1]*par[1]+par[2]*par[2];
  double cost = (sin2t <= 1.0)? sqrt(1-sin2t):0;
  double beta = par[7];
  if (ldf == eNKG) {beta = _NKG_BetaFromSRef(sRef, cost); theLpnheRec().fBeta = beta; gamma = 0.0;}

  for (i = 0; i < size; i++) {
    sig = event().fSelectedStations[i]->fSigInVEM;
    dx = event().fSelectedStations[i]->fX-theLpnheRec().fXCore - par[4];
    dy = event().fSelectedStations[i]->fY-theLpnheRec().fYCore - par[5];
    d = distance2Core(dx, dy, par[1], par[2]);
    funcval = sRef* _LDFfunc(d, beta, gamma, theLpnheRec().fRefDist, theLpnheRec().fLDFRcut);
#if 0
    if ((event().fSelectedStations[i]->fPmt[0].fLowGainSat
         || event().fSelectedStations[i]->fPmt[1].fLowGainSat
         || event().fSelectedStations[i]->fPmt[2].fLowGainSat)
        && funcval > sig) continue;
    if (gCorrectAsymmetry) f += sqr(funcval - sig*compute_asymetry(dx,dy,d,theLpnheRec().fU,theLpnheRec().fV)) / funcval;
    else f += sqr(funcval - sig) / funcval;
#else
    if (gCorrectAsymmetry) sig *= compute_asymetry(dx, dy, d, par[1], par[2]);
    if (event().fSelectedStations[i]->fSat) {
      // Using lower bound at 2000 VEM or more -- XB 19/06/2012
      // else error is not in sqrt(s) but is proportionnal to s...
      if (sig<2000) df = sqr((funcval - sig) / (funcval*event().fSelectedStations[i]->fSat*0.01));
      else if (funcval > sig) df = 0;
      else df = sqr(funcval - sig) / (1.06*1.06*funcval);
    } else {
      // using GAP 2004-047, without angular dependance
      df = sqr(funcval - sig) / (1.06*1.06*funcval);
    }
    //cerr << event().fSelectedStations[i]->Id << " " << sig << " " << funcval << " " << df << endl;
    f+=df;
#endif

  }

  /// if required we add silent stations
  if (gUseSilent) {
    size = event().fSilentStations.size();
    for (unsigned int j = 0; j < size; j++) {
      dx = event().fSilentStations[j]->fX-theLpnheRec().fXCore - par[4];
      dy = event().fSilentStations[j]->fY-theLpnheRec().fYCore - par[5];
      d = distance2Core(dx, dy, par[1], par[2]);
      funcval = sRef * _LDFfunc(d, beta, gamma, theLpnheRec().fRefDist, theLpnheRec().fLDFRcut);
      f += funcval / SILSIGTHRESH;
    }
  }
}

static void globalFit(int &n, double *d1, double &f, double *par, int flag) {
  f = 0;
  double geom, ldf;
  geomFit(n, d1, geom, par, flag);
  ldfFit (n, d1, ldf, par, flag);
  f = geom + ldf;
}

#if 0
  if (chi2 < 0) return 0;
  if (farStation == -1) return 1;
  pointChi2 = sqr(_StatDt(farStation, 1, geom)) / TimeVariance(farStation, dist, W(), geom);
  if (pointChi2 < nSig*nSig) return 1;
  if (gVerbosity) cerr << "Station " << fErEvent->fSelectedStations[farStation]->Id  << " removed with time residuals: " <<
    _StatDt(farStation, 1, geom)  << " (" << sqrt(pointChi2) << " sigmas)" << endl;
  fErEvent->fRandomStations.push_back(fErEvent->fSelectedStations[farStation]);
  fErEvent->fSelectedStations.erase(fErEvent->fSelectedStations.begin() + farStation);
  return _OptimizeStationSelection(++loop);
#endif

#if 0
#define MAX_SIGMA 5.0
/*---------------------------------------------------------*/
int SdRecLpnhe::_OptimizeStationSelection(int loop) {
  /*---------------------------------------------------------*/
  // Optimize the station selection list until the PlaneFit or CurveFil chi2 falls below a certain threshold.
  // Up to 3 stations can be removed in the process.
  unsigned int size = fErEvent->fSelectedStations.size(); TGeometry geom = ePlane; int farStation;
  double nSig = MAX_SIGMA;
  double chi2, pointChi2;
  if (loop == 1) {
    //if (!gOfficial) fErEvent->OfficialSelection(); // keep station list as it is!
    fU = fErEvent->fEstimated.fU; fV = fErEvent->fEstimated.fV; 
    fR = _EstimateCurvature(fSRef,W(),fErEvent->UseInfillRec);
  }
  _EstimateXYZTCore();
  if (size == 3) {
    fR = _EstimateCurvature(fSRef,W(),fErEvent->UseInfillRec);
    _MatPlanFit(1, farStation);
    return 1;
  }
  if (loop > 5) return 0;
  if (size > 4) {
    geom = eCurve; _MatCurvedFit(0, farStation); chi2 = _MatCurvedFit(1, farStation);
  } else {
    fR = 0.0; geom = ePlane; _MatPlanFit(0, farStation); 
    fR = _EstimateCurvature(fSRef,W(),fErEvent->UseInfillRec); 
    chi2 = _MatPlanFit(1, farStation);
  }
  if (chi2 < 0) return 0;
  if (farStation == -1) return 1;
  double dist = _Stat2Core(farStation);
  pointChi2 = sqr(_StatDt(farStation, 1, geom)) / TimeVariance(farStation, dist, W(), geom);
  if (pointChi2 < nSig*nSig) return 1;
  if (gVerbosity) cerr << "Station " << fErEvent->fSelectedStations[farStation]->Id  << " removed with time residuals: " <<
    _StatDt(farStation, 1, geom)  << " (" << sqrt(pointChi2) << " sigmas)" << endl;
  fErEvent->fRandomStations.push_back(fErEvent->fSelectedStations[farStation]);
  fErEvent->fSelectedStations.erase(fErEvent->fSelectedStations.begin() + farStation);
  return _OptimizeStationSelection(++loop);
}

double SdRecLpnhe::T50(int stat, bool doIt) {
static double t50Value = 0.0;
if (doIt) {
  double sum=0.0, w = 0.0;
  for (unsigned int i = 0; i < fErEvent->fSelectedStations.size(); ++i) {
    if (fErEvent->fSelectedStations[i]->SigInVEM()<500) {
      double dx = fErEvent->fSelectedStations[i]->fX-fXCore;
      double dy = fErEvent->fSelectedStations[i]->fY-fYCore;
      double d = distance2Core(dx, dy, fU, fV);
      sum += fErEvent->fSelectedStations[i]->SigInVEM() * 
	     (fErEvent->fSelectedStations[i]->T50() - fErEvent->fSelectedStations[i]->fStartIntegration*kEc::TimeSlot)/d; 
      w += fErEvent->fSelectedStations[i]->SigInVEM();
      }
    }
  t50Value = sum/w;
  if (stat == -1) {
    for (unsigned int i = 0; i < fErEvent->fSelectedStations.size(); ++i) {
      double dx = fErEvent->fSelectedStations[i]->fX-fXCore;
      double dy = fErEvent->fSelectedStations[i]->fY-fYCore;
      double d = distance2Core(dx, dy, fU, fV);
      cout << fErEvent->fSelectedStations[i]->SigInVEM() << " " << t50Value * d << " ";
      cout << fErEvent->fSelectedStations[i]->T50() - fErEvent->fSelectedStations[i]->fStartIntegration*kEc::TimeSlot << endl;
      }
    return 0;
    }
  }
double dx = fErEvent->fSelectedStations[stat]->fX-fXCore;
double dy = fErEvent->fSelectedStations[stat]->fY-fYCore;
double d = distance2Core(dx, dy, fU, fV);                 
return t50Value * d;
}

int SdRecLpnhe::_OptimizeStationSelection2() {
  int res = 0;
  if (event().fSelectedStations.size() < 4) return 1;
  do {
    int ndf = countDOF(eCurve);
    if (ndf > 0)  res = _GeometricalRec(eCurve, true); else  {
      fR = _EstimateCurvature(fSRef,W(),fErEvent->UseInfillRec);
      res = _GeometricalRec(ePlane, true);
    }
  }
  while (res == 2);
  return res;
}


#endif
