#include "TDataToFit.h"

/**
   @file TDataToFit.cc
*/

/*--------------------------------------------------------------------*/
TDataToFit::TDataToFit(const vector<int>& ListIds,
                       const vector<double>& ListTimes,
                       const vector< vector<double> >& ListCoords)
    : fListIds(ListIds), fListTimes(ListTimes), fListCoords(ListCoords)
    /*---------------------------------------------------------------------*/
{
  fNTanks = fListTimes.size();
  ComputeConstTimesErrors(35e-9);
}

/*------------------------------------------------------------------*/
TDataToFit::TDataToFit(const vector<int>& ListIds,
                       const vector< vector<double> >& ListCoords,
                       const vector<double>& ListSignals)
    : fListIds(ListIds), fListCoords(ListCoords),
    fListSignals(ListSignals)
    /*------------------------------------------------------------------*/
{
  fNTanks = fListSignals.size();
  ComputeSignalsErrors();
}

/*-----------------------------------------------------------------*/
TDataToFit::TDataToFit(const vector<int>& ListIds,
                       const vector<double>& ListTimes,
                       const vector< vector<double> >& ListCoords,
                       const vector<double>& ListSignals)
    : fListIds(ListIds), fListTimes(ListTimes),
    fListCoords(ListCoords), fListSignals(ListSignals)
    /*-----------------------------------------------------------------*/
{
  fNTanks = fListTimes.size();
  ComputeConstTimesErrors(35e-9);
  ComputeSignalsErrors();
}

/*-----------------------------------------------------------------*/
TDataToFit::TDataToFit(const vector<int>& ListIds,
                       const vector<double>& ListTimes,
                       const vector< vector<double> >& ListCoords,
                       const vector< vector<double> >& ListCoordsSilent,
                       const vector<double>& ListSignals)
    : fListIds(ListIds), fListTimes(ListTimes),
    fListCoords(ListCoords), fListCoordsSilent(ListCoordsSilent),
    fListSignals(ListSignals)
    /*-----------------------------------------------------------------*/
{
  fNTanks = fListTimes.size();
  ComputeConstTimesErrors(35e-9);
  ComputeSignalsErrors();
}

/*----------------------------------------------------*/
void TDataToFit::SetLsIds(const vector<int>& ListIds)
/*----------------------------------------------------*/
{
  fListIds = ListIds;
}

/*-----------------------------------------------------------*/
void TDataToFit::SetTimes(const vector<double>& ListTimes)
/*-----------------------------------------------------------*/
{
  fListTimes = ListTimes;
}

/*----------------------------------------------------------------------*/
void TDataToFit::SetTimesErrors(const vector<double>& ListTimesErrors)
/*----------------------------------------------------------------------*/
{
  fListTimesErrors = ListTimesErrors;
}

/*---------------------------------------------------------------------*/
void TDataToFit::SetCoords(const vector< vector<double> >& ListCoords)
/*---------------------------------------------------------------------*/
{
  fListCoords = ListCoords;
}

/*---------------------------------------------------------------*/
void TDataToFit::SetSignals(const vector<double>& ListSignals)
/*---------------------------------------------------------------*/
{
  fListSignals = ListSignals;
}

/*---------------------------------------------------------------------------*/
void TDataToFit::SetSignalsErrors(const vector<double>& ListSignalsErrors)
/*---------------------------------------------------------------------------*/
{
  fListSignalsErrors = ListSignalsErrors;
}
/*---------------------------------------------------------------------------*/
void TDataToFit::SetSaturated(const vector<bool>& ListSaturated)
/*---------------------------------------------------------------------------*/
{
  fListSaturated = ListSaturated;
}

/*----------------------------------------------------------*/
void TDataToFit::ComputeConstTimesErrors(double sigma)
/*----------------------------------------------------------*/
{
  fListTimesErrors.resize(fNTanks);
  // constant value
  for( unsigned int i=0;i<fNTanks;i++ )
    fListTimesErrors[i] = sigma;
}

/*-----------------------------------------------------------------------*/
void TDataToFit::ComputeVarTimeError(int i,double r, double costheta2)
/*-----------------------------------------------------------------------*/
{
  fListTimesErrors.resize(fNTanks);
  // distance from axis depending value
  fListTimesErrors[i] = sqrt(25e-9*25e-9+5e-14*5e-14*pow(r,4.)*costheta2);
}

/*------------------------------------*/
void TDataToFit::ComputeSignalsErrors()
/*------------------------------------*/
{
  fListSignalsErrors.resize(fNTanks);
  // error is the sqrt of the signal
  for( unsigned int i=0;i<fNTanks;i++ )
    fListSignalsErrors[i] = sqrt(fListSignals[i]);
}

/*------------------------------*/
void TDataToFit::Remove(int lidx)
/*------------------------------*/
{
  if( lidx >= 0 && lidx < (int)fNTanks ) {
    fNTanks--;
    fListIds.erase(fListIds.begin()+lidx);
    cout << "Remaining LS : ";
    for( unsigned int i=0;i<fListIds.size();i++ )
      cout << fListIds[i] << " ";
    cout << endl;
    fListCoords.erase(fListCoords.begin()+lidx);
    fListTimes.erase(fListTimes.begin()+lidx);
    fListSignals.erase(fListSignals.begin()+lidx);
  }
}

/*-----------------------------*/
void TDataToFit::InitDataToFit()
/*-----------------------------*/
{
  fNTanks = 0;
  fListIds.clear();
  fListTimes.clear();
  fListTimesErrors.clear();
  fListTimesResiduals.clear();
  fListFittedTimes.clear();
  fListCoords.clear();
  fAxisDistances.clear();
  fCoreDistances.clear();
  fListSignals.clear();
  fListSigOverPeak.clear();
  fListSignalsErrors.clear();
  fLDF.clear();
  fListCoordsSilent.clear();
}

/*------------------------------*/
void TDataToFit::InitFitResults()
/*------------------------------*/
{
  fListTimesResiduals.clear();
  fListFittedTimes.clear();
  fAxisDistances.clear();
  fCoreDistances.clear();
  fLDF.clear();
}

/*----------------------------------------------------------*/
void CopyInstance(TDataToFit * dest, const TDataToFit * src)
/*----------------------------------------------------------*/
{
  *dest = *src;
}
