#include "TFitting.h"

/**
   @file TFitting.cc
*/

/*--------------------------------*/
vector<string> MakeParameterName()
/*--------------------------------*/
{
  vector<string> tmp(kFitNumber);
  tmp[kFitX0]   = "X0";
  tmp[kFitY0]   = "Y0";
  tmp[kFitZ0]   = "Z0";
  tmp[kFitT0]   = "T0";
  tmp[kFitU]    = "U";
  tmp[kFitV]    = "V";
  tmp[kFitC]    = "C";
  tmp[kFitS]    = "S";
  tmp[kFitE]    = "E";
  tmp[kFitBeta] = "Beta";
  tmp[kFitXMAX] = "XMAX";
  return tmp;
}

static TFitting * gFitter;
vector<string> gParameterName = MakeParameterName();

#define fpSmspEstim_INIT do { \
  if ( ! fpSmspEstim && fLDFType == "SMSP" ) \
		fpSmspEstim = new TSMSPEstim(SMSPPATH + (string)(SMSPDATA)); \
  } while(0)

TSMSPEstim* TFitting::fpSmspEstim = 0;

/*-------------------------------*/
TFitting::TFitting(int verbosity)
/*-------------------------------*/
{
  InitFitting(1);
  fMinuit       = new TMinuit(kFitNumber);
  fArgList      = new double[(int)kFitNumber];
  fVStart       = new double[(int)kFitNumber];
  fVStep        = new double[(int)kFitNumber];
  fMin          = new double[(int)kFitNumber];
  fMax          = new double[(int)kFitNumber];
  fParameters   = new double[(int)kFitNumber];
  fParameterSet.resize(kFitNumber);
  for( int i=0;i<kFitNumber;i++ ) {
    fParameterSet[i] = false;
    fVStart[i] = 0;
    fVStep[i] = 0;
    fMin[i] = 0;
    fMax[i] = 0;
  }
  // set defaults values
  fIerFlg = 0;
  fCount = 0;
  SetContext(verbosity);
  SetInstance();
}

/*-------------------*/
TFitting::~TFitting()
/*-------------------*/
{
  delete [] fArgList;
  delete [] fVStart;
  delete [] fVStep;
  delete [] fMin;
  delete [] fMax;
  delete [] fParameters;
  delete fMinuit;
}

/*--------------------------------------------------------------------*/
void TFitting::SetContext(int verbosity, int warnings, string method, string statistics)
/*--------------------------------------------------------------------*/
{
  SetVerbosity(verbosity);
  SetWarnings(warnings);
  SetMinimizationMethod(method);
  SetStatistics(statistics);
}

/*--------------------------------------------------------------------*/
void TFitting::SetVerbosity(int verbosity)
/*--------------------------------------------------------------------*/
{
  fVerbosity = verbosity;
  fArgList[0] = fVerbosity;
  fMinuit->mnexcm("SET PRI",fArgList,1,fIerFlg);
}

/*--------------------------------------------------------------------*/
void TFitting::SetWarnings(bool warnings)
/*--------------------------------------------------------------------*/
{
  fWarnings = warnings;
  if( fWarnings )
    fMinuit->mnexcm("SET WAR",fArgList,1,fIerFlg);
  else
    fMinuit->mnexcm("SET NOW",fArgList,1,fIerFlg);
}

/*--------------------------------------------------------------------*/
void TFitting::SetMinimizationMethod(string method)
/*--------------------------------------------------------------------*/
{
  fMethod = method;
}

/*--------------------------------------------------------------------*/
void TFitting::SetStatistics(string method)
/*--------------------------------------------------------------------*/
{
  fStatistics = method;
  if( fStatistics == "CHI2" ) {
    fArgList[0] = 1;
    fMinuit->mnexcm("SET ERR",fArgList,1,fIerFlg);
  } else if( fStatistics == "LOG" ) {
    fArgList[0] = 0.5;
    fMinuit->mnexcm("SET ERR",fArgList,1,fIerFlg);
  } else {
    cout << "TFitting::SetStatistics : method = " << method
    << " not understood. Returning." << endl;
    return;
  }
}

/*--------------------------------------------------------------------*/
void TFitting::SetLDFType(string ldf)
/*--------------------------------------------------------------------*/
{
  fLDFType = ldf;
}

/*--------------------------------------------------------------------*/
void TFitting::SetFCN(void (*fcn)(Int_t &, Double_t *, Double_t &f, Double_t *, Int_t))
/*--------------------------------------------------------------------*/
{
  fMinuit->fFCN = fcn; // set the adress of minimization function
}

/*--------------------------------------------------------------------*/
void TFitting::SetInstance()
/*--------------------------------------------------------------------*/
{
  // in case of nested fits
  // reset gFitter address to mine
  gFitter = this;
}

/*--------------------------------------------------------------------*/
void TFitting::SetFreeParameter(int parameter, double start, double step, double min, double max)
/*--------------------------------------------------------------------*/
{
  if( !fParameterSet[parameter] )
    fCount++;
  fParameterSet[parameter] = true;
  fVStart[parameter] = start;
  fVStep[parameter] = step;
  fMin[parameter] = min;
  fMax[parameter] = max;
}

/*--------------------------------------------------------------------*/
void TFitting::SetFreeParameters(int numpar, int * parameters, double * vstart,
                                 double * step, double * min, double * max)
/*--------------------------------------------------------------------*/
{
  int rank;
  for( int i=0;i<numpar;i++ ) {
    rank = parameters[i];
    if( !fParameterSet[rank] )
      fCount++;
    fParameterSet[rank] = true;
    fVStart[rank] = vstart[i];
    fVStep[rank] = step[i];
    if( min != 0x0 )
      fMin[rank] = min[i];
    if( max != 0x0 )
      fMax[rank] = max[i];
  }
}

/*--------------------------------------------------------------------*/
void TFitting::SetMin(int numpar, double min)
/*--------------------------------------------------------------------*/
{
  cout << "TFitting::changing min limit on parameter "
  << gParameterName[numpar] << ", giving value : " << min << endl;
  fMin[numpar] = min;
}

/*--------------------------------------------------------------------*/
void TFitting::SetMax(int numpar, double max)
/*--------------------------------------------------------------------*/
{
  cout << "TFitting::changing max limit on parameter "
  << gParameterName[numpar] << ", giving value : " << max << endl;
  fMax[numpar] = max;
}

/*--------------------------------------------------------------------*/
void TFitting::ReleaseParameter(int parameter, double value, double step)
/*--------------------------------------------------------------------*/
{
  fMinuit->Release(parameter);
  if( step == -1 && fParameterSet[parameter] ) {
    // then use the previous step used
    step = fVStep[parameter];
  } else {
    cout << "You SHOULD specify a step for parameter " << gParameterName[parameter] << endl;
    cout << "Taking step = 1e-4" << endl;
    step = 1e-4;
  }
  SetFreeParameter(parameter,value,step);
}

/*--------------------------------------------------------------------*/
void TFitting::RemoveParameter(int parameter)
/*--------------------------------------------------------------------*/
{
  if( fParameterSet[parameter] == false ) {
    cout << "Parameter " << gParameterName[parameter]
    << " (" << parameter << ") has not been set." << endl;
    cout << "Can't remove it, returning." << endl;
    return;
  } else
    cout << "Removing parameter " << gParameterName[parameter] << endl;
  TString tt(gParameterName[parameter].c_str());
  // remove all and set parameter to remove to false
  fMinuit->mncler();
  fParameterSet[parameter] = false;
  fCount--;
}

/*--------------------------------------------------------------------*/
void TFitting::RemoveParameters()
/*--------------------------------------------------------------------*/
{
  fMinuit->mncler();
  for( int i=0;i<kFitNumber;i++ ) {
    fParameterSet[i] = false;
    fArgList[i] = 0;
    fVStart[i] = 0;
    fVStep[i] = 0;
    fMin[i] = 0;
    fMax[i] = 0;
    fParameters[i] = 0;
  }
  fCount = 0;
}

/*--------------------------------------------------------------------*/
void TFitting::SetFixParameter(int parameter, double value)
/*--------------------------------------------------------------------*/
{
  fParameterSet[parameter] = true;
  fVStart[parameter] = value;
  fVStep[parameter] = 0;
  fMin[parameter] = 0;
  fMax[parameter] = 0;
}

/*--------------------------------------------------------------------*/
void TFitting::PrepareParameters()
/*--------------------------------------------------------------------*/
{
  for( int i=0;i<kFitNumber;i++ ) {
    if( fParameterSet[i] ) {
      fMinuit->mnparm(i,gParameterName[i].c_str(),fVStart[i],fVStep[i],fMin[i],fMax[i],fIerFlg);
      if( fMin[i] > fMax[i] ) {
        cout << "PROBLEM for parameter: " << gParameterName[i] << " max < min !!!!!" << endl;
      }
    }
  }
}

/*--------------------------------------------------------------------*/
void TFitting::Minimize()
/*--------------------------------------------------------------------*/
{
  PrepareParameters();
  InitFitResults(); // very important in the case of an iterative fit (because of the push_back of the fit results)
  fArgList[0] = 500; // iterations number
  fArgList[1] = 1.;

  /* Redirection of Minuit stdout to stderr */
  int outfd = dup(1);
  close(1);
  dup2(2,1);

  fMinuit->mnexcm("MIGRAD",fArgList,2,fIerFlg);
  if( fMethod == "MINOS" )
    fMinuit->mnexcm("MINOS",fArgList,0,fIerFlg);
  Double_t amin, edm, errdef;
  Int_t nvpar, nparx, icstat;
  fMinuit->mnstat(amin,edm,errdef,nvpar,nparx,icstat);

  /* End of stderr redirection */
  close(1);
  dup2(outfd,1);
  close(outfd);

  if( fVerbosity > -1 )
    cout << "TFitting :: (" << fNTanks << " tanks used): ";
  fConvergenceStatus = icstat;
  if( icstat == 3 ) {
    if( fVerbosity > -1 )
      cout << "Minimisation has converged." << endl;
    fHasConverged = true;
  } else if( icstat == 2 ) {
    if( fVerbosity > -1 )
      cout << "Minimisation has converged but covariance matrix forced positive definite." << endl;
    fHasConverged = true;
  } else if( icstat == 1 ) {
    if( fVerbosity > -1 )
      cout << "Minimisation has converged but get only a diagonal approximation of the covariance matrix." << endl;
    fHasConverged = true;
  } else {
    if( fVerbosity > -1 )
      cout << "Minimisation has NOT converged." << endl;
    fHasConverged = false;
  }
  if( fVerbosity > -1 )
    fMinuit->mnprin(3,amin);
  fChi2 = fMinuit->fAmin;
  int nddl = 0;
  // get fit results
  if( GetValues(kFitX0)   ) {
    nddl++;
    fMinuit->GetParameter(kFitX0,fXCore,fdXCore);
    fParameters[kFitX0] = fXCore;
  }
  if( GetValues(kFitY0)   ) {
    nddl++;
    fMinuit->GetParameter(kFitY0,fYCore,fdYCore);
    fParameters[kFitY0] = fYCore;
  }
  if( GetValues(kFitZ0)   ) {
    nddl++;
    fMinuit->GetParameter(kFitZ0,fZCore,fdZCore);
    fParameters[kFitZ0] = fZCore;
  }
  if( GetValues(kFitT0)   ) {
    nddl++;
    fMinuit->GetParameter(kFitT0,fT0s,fdT0s);
    fParameters[kFitT0] = fT0s;
  }
  if( GetValues(kFitU)    ) {
    nddl++;
    fMinuit->GetParameter(kFitU,fU,fdU);
    fParameters[kFitU] = fU;
  }
  if( GetValues(kFitV)    ) {
    nddl++;
    fMinuit->GetParameter(kFitV,fV,fdV);
    fParameters[kFitV] = fV;
  }
  if( GetValues(kFitS)    ) {
    nddl++;
    fMinuit->GetParameter(kFitS,fSRef,fdS1000);
    fParameters[kFitS] = fSRef;
  }
  if( GetValues(kFitE)    ) {
    nddl++;
    fMinuit->GetParameter(kFitE,fE,fdE);
    fParameters[kFitE] = fE;
  }
  if( GetValues(kFitBeta) ) {
    nddl++;
    fMinuit->GetParameter(kFitBeta,fBeta,fdBeta);
    fParameters[kFitBeta] = fBeta;
  }
  if( GetValues(kFitXMAX) ) {
    nddl++;
    fMinuit->GetParameter(kFitXMAX,fXmax,fdXmax);
    fParameters[kFitXMAX] = fXmax;
  }

  if( GetValues(kFitC) ) {
    nddl++;
    double Curv, dCurv;
    fMinuit->GetParameter(kFitC,Curv,dCurv);
    fParameters[kFitC] = Curv;
    if( Curv != 0 ) {
      fdR = dCurv/(2*Curv*Curv);
      fR =  1/(2.*Curv);
    }
  }
  fddl = fNTanks-nddl;
  if( fddl <= 0 )
    cout << "degree of freedom problem !!!!!!!!!!" << endl;
  frChi2 = fChi2/fddl;
}

/*--------------------------------------------------------------------*/
bool TFitting::GetValues(FitVariable fv)
/*--------------------------------------------------------------------*/
{
  // ignore when parameter were fixed
  return (fParameterSet[fv] == true && fVStep[fv] != 0 );
}

/*--------------------------------------------------------------------*/
void TFitting::ComputeResults()
/*--------------------------------------------------------------------*/
{
  fpSmspEstim_INIT;

  // compute theta, phi
  double theta, phi;
  fsqfU2fV2 = sqrt( fU * fU + fV * fV );
  uv2thetaphi(fU,fV,theta,phi);
  fTheta = theta * kCdasUtil::RAD2DEG;
  fPhi = phi * kCdasUtil::RAD2DEG;

  // Compute S1000 if we fit E
  if ( GetValues(kFitE) && fLDFType == "SMSP" )
    fSRef = compute_S1000(fLDFType,fE,theta,fpSmspEstim);
  
  if ( GetValues(kFitE) && (fLDFType == "2EXPO" || fLDFType == "Photon") )
    fSRef = compute_LDF2EXPOFunc(1000.,theta,fE,fPrimary); 

  // Compute E if we fit S1000
  if( GetValues(kFitS) && (fLDFType == "LOG" || fLDFType == "HP") )
    fE = compute_energy(fLDFType,fSRef,theta,fPrimary,fdS1000,&fdE);


  compute_distancesUV(fU,fV,fXCore,fYCore,fZCore,fListCoords,fAxisDistances);


  fpSmspEstim_INIT;

  double x, y, z, optval;
  if( !fParameterSet[kFitC] ) { // then it was a plane fit
    for( unsigned int i=0;i<fNTanks;i++ ) {
      x = fListCoords[i][0]-fXCore;
      y = fListCoords[i][1]-fYCore;
      z = fListCoords[i][2]-fZCore;
      optval = fT0s-FitPlane(x,y,z,fParameters);
      fListTimesResiduals.push_back(1e9*(optval-fListTimes[i]));
      fListFittedTimes.push_back(optval);
    }
  } else { // take into account curvature
    for( unsigned int i=0;i<fNTanks;i++ ) {
      x = fListCoords[i][0]-fXCore;
      y = fListCoords[i][1]-fYCore;
      z = fListCoords[i][2]-fZCore;
      optval = fT0s-FitCurvatureVariable(x,y,z,fAxisDistances[i],fParameters);
      fListTimesResiduals.push_back(1e9*(optval-fListTimes[i]));
      fListFittedTimes.push_back(optval);
    }
  }
  if( fParameterSet[kFitX0] && fParameterSet[kFitY0] ) {
    // if we are searching the core position then T0 is the actual impact time
    // if not, T0 is degenerated and it's a nonsense to compute timeresiduals
    TH1F * histo = new TH1F("toto","",50,*min_element(fListTimesResiduals.begin(),fListTimesResiduals.end()),
                            *min_element(fListTimesResiduals.begin(),fListTimesResiduals.end()));
    for( unsigned int i=0;i<fNTanks;i++ )
      histo->Fill(fListTimesResiduals[i]);
    if( fVerbosity > -1 )
      cout << "mean of time residuals : " << histo->GetMean() << endl;
    if( fVerbosity > -1 )
      cout << "rms of time residuals : " << histo->GetRMS() << endl;
    delete histo;
  }

  if( fParameterSet[kFitS] ) { // we know the energy
    int Npts = 100;
    double distmin = 30;
    //  double distmax = 2500;
    double distmax = *max_element(fAxisDistances.begin(),fAxisDistances.end());
    double diststep = (distmax-distmin)/(Npts-1);
    vector<double> vals(2);

    if( fLDFType == "HP" ) {
      for( int i=0;i<Npts;i++ ) {
        vals[0] = distmin + i * diststep;
        vals[1] = fSRef * compute_LDFHPFunc(vals[0],theta);
        fLDF.push_back(vals);
      }
    } else if( fLDFType == "LOG" ) {
      for( int i=0;i<Npts;i++ ) {
        vals[0] = distmin + i * diststep;
        vals[1] = fSRef * compute_LDFLOGFunc(vals[0],fBeta,fGamma);
        fLDF.push_back(vals);
      }
    } else if ( fLDFType == "SMSP" ) {
      TSMSPPoint lP;
      for( int i=0;i<Npts;i++ ) {
        vals[0] = distmin + i * diststep;
        lP[SMSP_TYPE_DIST] = vals[0];
        lP[SMSP_TYPE_THETA] = fTheta ;
        lP[SMSP_TYPE_ENERGY] = fE;
	vals[1] = fpSmspEstim->VEM(lP);
        fLDF.push_back(vals);
      }
    } else if( fLDFType == "2EXPO" ) {
      for( int i=0;i<Npts;i++ ) {
        vals[0] = distmin + i * diststep;
        //	      vals[1] = fSRef * compute_LDFPhotonFunc(vals[0],theta,fE);
        vals[1] = compute_LDF2EXPOFunc(vals[0],theta,fE,fPrimary);
        fLDF.push_back(vals);
      }
    } else {
      cout << "TFitting::ComputeResults() :" << endl;
      cout << "You have not set the LDF type." << endl;
      cout << "Can't compute LDF." << endl;
    }
  }
}

/*--------------------------------------------------------------------*/
double TFitting::ComputeChi2GlobalHP(double * par)
/*--------------------------------------------------------------------*/
{
  //calculate chisquare
  Double_t chisq = 0;
  Double_t delta;
  double signal;
  double theta = asin(sqrt(par[kFitU]*par[kFitU] + par[kFitV]*par[kFitV]));
  double cth2 = minval(1.,maxval(1-par[kFitU]*par[kFitU] - par[kFitV]*par[kFitV],0));
  double x, y, z, dist;

  for( unsigned int i=0;i<fNTanks;i++) {
    if( !fListSaturated[i] ) {
      // space-time part
      x = fListCoords[i][0]-par[kFitX0];
      y = fListCoords[i][1]-par[kFitY0];
      z = fListCoords[i][2]-par[kFitZ0];
      if(par[kFitU]*par[kFitU]+par[kFitV]*par[kFitV]<=1 ) {
        dist = compute_distUV(par[kFitU],par[kFitV],x,y,z);
        delta  = fListTimes[i] - par[kFitT0] + FitCurvatureVariable(x,y,z,dist,par);
        ComputeVarTimeError(i,dist,cth2);
        delta /= fListTimesErrors[i];
        chisq += delta*delta;

        // energy part
        signal =FitLDFHP(dist,theta,&par[kFitS]);
        //double signal_corrected = fListSignals[i] * compute_asymetry(x,y,dist,par[kFitU],par[kFitV]);
        double signal_corrected = signal / compute_asymetry(x,y,dist,par[kFitU],par[kFitV]);
        //delta = (signal-signal_corrected)/sqrt(signal_corrected);
        delta = (fListSignals[i]-signal_corrected)/sqrt(signal_corrected);
        chisq += delta*delta;
      } else {
        delta  = (fListTimes[i]-par[kFitT0]+FitPlane(x,y,z,par));
        delta /= fListTimesErrors[i];
        chisq += delta*delta;
      }
    }
  }
  return chisq;
}

#define LDF_DISTREF 1000.
#define LDF_DISTMAX_SILENT 4000.
/*--------------------------------------------------------------------*/
double TFitting::ComputeChi2GlobalLOG(double * par)
/*--------------------------------------------------------------------*/
{
  //calculate chisquare
  Double_t chisq = 0;
  Double_t delta;
  double signal;
  double theta = asin(sqrt(par[kFitU]*par[kFitU] + par[kFitV]*par[kFitV]));
  double cth2 = minval(1.,maxval(1-par[kFitU]*par[kFitU] - par[kFitV]*par[kFitV],0));
  double x, y, z, dist;
  double energy;
  energy=compute_energy("LOG",maxval(0,par[kFitS]),theta,fPrimary,fdS1000,&fdE);
  for( unsigned int i=0;i<fNTanks;i++ ) {
    if( !fListSaturated[i] ) {
      // space-time part
      x = fListCoords[i][0]-par[kFitX0];
      y = fListCoords[i][1]-par[kFitY0];
      z = fListCoords[i][2]-par[kFitZ0];
      if(par[kFitU]*par[kFitU]+par[kFitV]*par[kFitV]<=1) {
        dist = compute_distUV(par[kFitU],par[kFitV],x,y,z);
        delta  = fListTimes[i] - par[kFitT0] + FitCurvatureVariable(x,y,z,dist,par);
        ComputeVarTimeError(i,dist,cth2);
        delta /= fListTimesErrors[i];
        chisq += delta*delta;
        signal = FitLDFLOG(dist,theta,&par[kFitS]);
        double signal_corrected = signal / compute_asymetry(x,y,dist,par[kFitU],par[kFitV]);
        delta =(fListSignals[i]-signal_corrected)/sqrt(signal_corrected);
        chisq += delta*delta;
        // silent stations
#if 0
        if( GetValues(kFitS) )
        {
          //	    cout << "Try SILENT HERE" << endl;
          // mettre un truc du genre :
          static int first = 1;
          vector<bool> SilentInFit(fListCoordsSilent.size());
          for(unsigned int i=0; i<fListCoordsSilent.size(); i++) {
            double dx = fListCoordsSilent[i][0] - par[kFitX0];
            double dy = fListCoordsSilent[i][1] - par[kFitY0];
            double dz = fListCoordsSilent[i][2] - par[kFitZ0];
            double dist1 = compute_distUV(par[kFitU],par[kFitV],dx,dy,dz);
            SilentInFit[i] = (dist1 > LDF_DISTREF && dist1 < LDF_DISTMAX_SILENT);
            if( SilentInFit[i] ) {
              signal = FitLDFLOG(dist1,theta,&par[kFitS]);
              chisq += signal/3.;
            }
          }
          first = 0;
        }
#endif

      } else {
        delta = (fListTimes[i]-par[kFitT0]+FitPlane(x,y,z,par));
        delta /= fListTimesErrors[i];
        chisq += delta*delta;
      }
  }
  }
  return chisq;
}

#define LDF_DMIN_SILENT_2EXPO 500.
#define LDF_DMAX_SILENT_2EXPO 2000.

// Choose one of the asymmetry correction options: 1-CDAS default; 2-experimental one (see GAP 2006-041)
#define ASYM_CORR_1_R_THETA
//#define ASYM_CORR_2_R_THETA_E_PTYPE

/*--------------------------------------------------------------------*/
double TFitting::ComputeChi2Global2EXPO(double * par)
/*--------------------------------------------------------------------*/
{
  //calculate chisquare
  Double_t chisq = 0;
  Double_t delta;
  double signal;
  double theta = asin(sqrt(par[kFitU]*par[kFitU] + par[kFitV]*par[kFitV]));
  double cth2 = minval(1.,maxval(1-par[kFitU]*par[kFitU] - par[kFitV]*par[kFitV],0));
  double x, y, z, dist;
  double energy=maxval(0.01,par[kFitE]);
  int isilent=0;

  //cout << "new call " << endl;
  for( unsigned int i=0;i<fNTanks;i++ ) {
    //    if( !fListSaturated[i] ) {
      // space-time part
      x = fListCoords[i][0]-par[kFitX0];
      y = fListCoords[i][1]-par[kFitY0];
      z = fListCoords[i][2]-par[kFitZ0];
      if(par[kFitU]*par[kFitU]+par[kFitV]*par[kFitV]<=1) {
        dist = compute_distUV(par[kFitU],par[kFitV],x,y,z);
        delta  = fListTimes[i] - par[kFitT0] + FitCurvatureVariable(x,y,z,dist,par);
        ComputeVarTimeError(i,dist,cth2);
	//	cout << "Tank " << i << " Id " <<fListIds[i] << endl;

	//cout << i << " " << fListTimes[i] << "dist = " << dist << endl;
        //cout << "par[kFitT0] = " << par[kFitT0] << endl;
        //cout << "FitCurvatureVariable(x,y,z,dist,par) = " << FitCurvatureVariable(x,y,z,dist,par) << endl;
        //cout << "delta = " << delta << endl;
        //cout << "fListTimesErrors[i] = " << fListTimesErrors[i] << " " << endl;
        delta /= fListTimesErrors[i];
        //cout << "residu de chi2 = " << delta << " " << endl;

	chisq += delta*delta;

        // ENERGY PART
	signal = FitLDF2EXPO(dist,theta,energy,fPrimary);
        // In the new version, simulated signal must be divided by compute_asymetry and not multiplied... (Jacob)
	//                     cout << "dist = " << dist << " Qrea: " << fListSignals[i]  << " Qteo: " << signal << endl;
	//	       cout << "Qteo multiplied by " << (1./compute_asymetry(x,y,dist,par[kFitU],par[kFitV])) << endl;

#ifdef ASYM_CORR_1_R_THETA
	double signal_corrected = signal / compute_asymetry(x,y,dist,par[kFitU],par[kFitV]);
#endif

#ifdef ASYM_CORR_2_R_THETA_E_PTYPE
        double signal_corrected = signal / compute_asymetry(x,y,dist,par[kFitU],par[kFitV],(fPrimary+1),energy);
#endif

        //double ss = 0.8*pow(signal_corrected,0.95);
        //double ss = fListSignalsErrors[i];
        double ss = sqrt(signal_corrected);
	if( fListSaturated[i] ) ss*=ss;
        delta = (fListSignals[i]-signal_corrected)/ss;
        chisq += delta*delta;

        // silent stations
#if 1
	if (!isilent) {
	  isilent=1;
	  vector<bool> SilentInFit(fListCoordsSilent.size());
	  for(unsigned int i=0; i<fListCoordsSilent.size(); i++) {
	    double dx = fListCoordsSilent[i][0] - par[kFitX0];
	    double dy = fListCoordsSilent[i][1] - par[kFitY0];
	    double dz = fListCoordsSilent[i][2] - par[kFitZ0];
	    double dist1 = compute_distUV(par[kFitU],par[kFitV],dx,dy,dz);
	    SilentInFit[i] = dist1 > LDF_DMIN_SILENT_2EXPO && dist1 < LDF_DMAX_SILENT_2EXPO;

	    if( SilentInFit[i] ) {
	      // In the new version, simulated signal must be divided by compute_asymetry and not multiplied... (Jacob)
	      //        	      cout << "Silent Tank " << i << " Id " <<fListIdsSilent[i] << endl;
	      //	      cout << "dist = " << dist1 << " Qteo: " << FitLDF2EXPO(dist1,theta,energy,fPrimary) << endl;
	      //	      cout << "Qteo multiplied by " << (1./compute_asymetry(dx,dy,dist1,par[kFitU],par[kFitV])) << endl;

#ifdef ASYM_CORR_1_R_THETA
	      signal_corrected = FitLDF2EXPO(dist1,theta,energy,fPrimary)/compute_asymetry(dx,dy,dist1,par[kFitU],par[kFitV]);
#endif

#ifdef ASYM_CORR_2_R_THETA_E_PTYPE
	      signal_corrected = FitLDF2EXPO(dist1,theta,energy,fPrimary)/compute_asymetry(dx,dy,dist1,par[kFitU],par[kFitV],(fPrimary+1),energy);
#endif
	      
	      //	      if(signal_corrected > 5.) {
	      //delta = 4.86*pow(signal/compute_asymetry(dx,dy,dist1,par[kFitU],par[kFitV]),0.04);
	      delta = sqrt(signal_corrected);

	      if (fPrimary==2) delta=0.5*delta;
	      chisq += delta*delta;
	      //	      }
	    }
	  }
	}
#endif
      } else {
	delta = (fListTimes[i]-par[kFitT0]+FitPlane(x,y,z,par));
	delta /= fListTimesErrors[i];
	chisq += delta*delta;
      }
      //    } // stations saturees
  }
  return chisq;
}

/*--------------------------------------------------------------------*/
double TFitting::ComputeChi2Plane(double* par)
/*--------------------------------------------------------------------*/
{
  // calculate chisquare
  // compute U, V and stupid T0 (not the impact time since the core is not known at this step)
  Double_t chisq = 0;
  Double_t delta;
  double x, y, z;
  for( unsigned int i=0;i<fNTanks;i++ ) {
    //    if( !fListSaturated[i] ) {
      x = fListCoords[i][0]-fXCore;
      y = fListCoords[i][1]-fYCore;
      z = fListCoords[i][2]-fZCore;
      delta  = (fListTimes[i]-par[kFitT0]+FitPlane(x,y,z,par));
      delta /= fListTimesErrors[i];
      chisq += delta*delta;
      //    }
  }
  return chisq;
}

/*--------------------------------------------------------------------*/
double TFitting::ComputeChi2Curvature(double* par)
/*--------------------------------------------------------------------*/
{
  // calculate chisquare
  // compute U, V and stupid T0 (not the impact time since the core is not known at this step)
  Double_t chisq = 0;
  Double_t delta;
  double x, y, z;
  double dist;
  double cth2 = minval(1.,maxval(1-par[kFitU]*par[kFitU] - par[kFitV]*par[kFitV],0));

  for( unsigned int i=0;i<fNTanks;i++ ) {
    //    if( !fListSaturated[i] ) {
      x = fListCoords[i][0]-fXCore;
      y = fListCoords[i][1]-fYCore;
      z = fListCoords[i][2]-fZCore;
      if(par[kFitU]*par[kFitU]+par[kFitV]*par[kFitV]<=1) {
        dist = compute_distUV(par[kFitU],par[kFitV],x,y,z);
        delta  = fListTimes[i] - par[kFitT0] + FitCurvature(x,y,z,dist,par);
        ComputeVarTimeError(i,dist,cth2);
        delta /= fListTimesErrors[i];
      } else {
        delta  = (fListTimes[i]-par[kFitT0]+FitPlane(x,y,z,par));
        delta /= fListTimesErrors[i];
      }
      chisq += delta*delta;
      //   }
  }
  return chisq;
}

/*--------------------------------------------------------------------*/
double TFitting::ComputeChi2CurvatureVariable(double* par)
/*--------------------------------------------------------------------*/
{
  // calculate chisquare
  // compute U, V and stupid T0 (not the impact time since the core is not known at this step)
  Double_t chisq = 0;
  Double_t delta;
  double x, y, z;
  double dist;
  double cth2 = minval(1.,maxval(1-par[kFitU]*par[kFitU] - par[kFitV]*par[kFitV],0));

  for( unsigned int i=0;i<fNTanks;i++ ) {
    //   if( !fListSaturated[i] ) {
      x = fListCoords[i][0]-fXCore;
      y = fListCoords[i][1]-fYCore;
      z = fListCoords[i][2]-fZCore;
      if(par[kFitU]*par[kFitU]+par[kFitV]*par[kFitV]<=1) {
        dist = compute_distUV(par[kFitU],par[kFitV],x,y,z);
        delta  = fListTimes[i] - par[kFitT0] + FitCurvatureVariable(x,y,z,dist,par);
        ComputeVarTimeError(i,dist,cth2);
        delta /= fListTimesErrors[i];
      } else {
        delta  = (fListTimes[i]-par[kFitT0]+FitPlane(x,y,z,par));
        delta /= fListTimesErrors[i];
      }
      chisq += delta*delta;
      //   }
  }
  return chisq;
}

/*--------------------------------------------------------------------*/
double TFitting::FitPlane(double x, double y, double z, double * par) const
/*--------------------------------------------------------------------*/
{
  double zcoeff = sqrt(maxval(0,1-par[kFitU]*par[kFitU]-par[kFitV]*par[kFitV]));
  double value = ( par[kFitU]*x + par[kFitV]*y + zcoeff*z ) / kCdasUtil::CSPEEDSI;
  return value;
}

/*--------------------------------------------------------------------*/
double TFitting::FitCurvature(double x, double y, double z, double dist, double *par) const
/*--------------------------------------------------------------------*/
{
  double value = FitPlane(x,y,z,par) - dist*dist*par[kFitC] / kCdasUtil::CSPEEDSI ;
  return value;
}

/*--------------------------------------------------------------------*/
double TFitting::FitCurvatureVariable(double x, double y, double z, double dist, double *par) const
/*--------------------------------------------------------------------*/
{
  double R=0;
  double R0=0;
  double r2 = x*x+y*y;
  double r = sqrt(r2);
  double sq = sqrt(par[kFitU]*par[kFitU]+par[kFitV]*par[kFitV]);
  double theta = asin(sq);
  double alpha = M_PI_2+theta*(x*par[kFitU]+y*par[kFitV])/(r*sq);
  if(par[kFitC]!=0)
    R0 = 1/(2*par[kFitC]);
  R = sqrt(R0*R0+r2+2*R0*r*cos(alpha));
  double value = FitPlane(x,y,z,par);
  if(R!=0 && par[kFitC]!=0)
    value -= dist*dist/(2*R) / kCdasUtil::CSPEEDSI ;
  return value;
}

/*--------------------------------------------------------------------*/
double TFitting::FitLDFHP(double dist, double theta, double * parameter) const
/*--------------------------------------------------------------------*/
{
  return parameter[0] * compute_LDFHPFunc(dist,theta);
}

/*--------------------------------------------------------------------*/
double TFitting::FitLDFLOG(double dist, double theta, double * parameter)
/*--------------------------------------------------------------------*/
{
  //  double gamma;
  //  double val = parameter[0] * compute_LDFLOGFunc(dist,theta,energy,beta,gamma);
  double tmp_beta;
  GetBetaGammaFromTheta(0,theta,tmp_beta,fGamma);
  if( !GetValues(kFitBeta) )
    fBeta = tmp_beta;
  double val = parameter[0] * compute_LDFLOGFunc(dist,fBeta,fGamma);
  //  parameter[1] = beta;
  //  cout << "GETPARFCN = " << beta << " " << gamma << " " << parameter[1] << endl;
  return val;
  //  return parameter[0] * compute_LDFLOGFunc(dist,theta,energy,parameter[1],gamma);
}

/*--------------------------------------------------------------------*/
double TFitting::FitLDFPhoton(double dist, double theta, double energy, double* parameter) const
/*--------------------------------------------------------------------*/
{
  return parameter[0] * compute_LDFPhotonFunc(dist,theta,energy);
}

/*--------------------------------------------------------------------*/
double TFitting::FitLDF2EXPO(double dist, double theta, double energy, int primary) const
/*--------------------------------------------------------------------*/
{
  return compute_LDF2EXPOFunc(dist,theta,energy,primary);
}

/*--------------------------------------------------------------------*/
void TFitting::Print(ostream& s) const
/*--------------------------------------------------------------------*/
{
  char txt[1000];
  s << "-------------------- TFitting::Print ------------------------" << endl;
  if( fCount == 0 ) {
    s << "No parameters to display." << endl;
    return;
  }
  for( int i=0;i<kFitNumber;i++ ) {
    if( fParameterSet[i] ) {
      sprintf(txt,"%4s: value: %15.8f, start: %15.8f, step: %15.8f, min: %10.5f, max: %10.5f",
              gParameterName[i].c_str(),fParameters[i],fVStart[i],fVStep[i],fMin[i],fMax[i]);
      cout << txt << endl;
    } else {
      sprintf(txt,"%4s: is not set",gParameterName[i].c_str());
      cout << txt << endl;
    }
  }
}

/*--------------------------------------------------------------------*/
void TFitting::MinuitStatus() const
/*--------------------------------------------------------------------*/
{
  char txt[1000];
  cout << "-------------------- TFitting::MinuitStatus ------------------------" << endl;
  if( fCount == 0 ) {
    cout << "No parameters to display." << endl;
    return;
  }
  double val,err,limm,limp;
  int ii;
  for( int i=0;i<kFitNumber;i++ ) {
    TString tt(gParameterName[i].c_str());
    fMinuit->mnpout(i,tt,val,err,limm,limp,ii);
    if( fParameterSet[i] ) {
      //	  cout << "ii = " << ii << endl;
      if( ii < 0 ) {
        cout << "incompatibility: TFitting::Status found an undefined parameter: " << gParameterName[i] << endl;
        cout << "-------------------------------------" << endl;
      } else {
        sprintf(txt,"%4s ==> value: %15.8f,  error: %15.8f, min: %10.5f, max: %10.5f",
                // (ids: user %2d,  external %2d,  internal %2d)",
                gParameterName[i].c_str(),val,err,limm,limp);//,i,i+1,ii);
        cout << txt;
        if( ii == 0 )
          cout << " (not variable)";
        cout << endl;
      }
    } else {
      cout << gParameterName[i].c_str() << " not defined." << endl;
    }
  }
  if( fHasConverged )
    cout << "Chi2 = " << fChi2 << endl;
  cout << "------------------------------------------------------" << endl;
}

/*--------------------------------------------------------------------*/
void TFitting::InitFitting(bool noparams)
/*--------------------------------------------------------------------*/
{
  InitShowerParams();
  InitDataToFit();
  if( noparams == false ) // not the first time
  {
    RemoveParameters();
    SetMinimizationMethod();
  }
  fIerFlg = 0;
  fCount = 0;
}

/*--------------------------------------------------------------------*/
void FCNPlane(int&, double *, double& f, double *par, int)
/*--------------------------------------------------------------------*/
{
  f = gFitter->ComputeChi2Plane(par);
}

/*--------------------------------------------------------------------*/
void FCNCurvature(int&, double *, double& f, double *par, int)
/*--------------------------------------------------------------------*/
{
  f = gFitter->ComputeChi2Curvature(par);
}

/*--------------------------------------------------------------------*/
void FCNCurvatureVariable(int&, double *, double& f, double *par, int)
/*--------------------------------------------------------------------*/
{
  f = gFitter->ComputeChi2CurvatureVariable(par);
}

/*--------------------------------------------------------------------*/
void FCNGlobalHP(int&, double *, double& f, double *par, int)
/*--------------------------------------------------------------------*/
{
  f = gFitter->ComputeChi2GlobalHP(par);
}

/*--------------------------------------------------------------------*/
void FCNGlobalLOG(int&, double *, double& f, double *par, int)
/*--------------------------------------------------------------------*/
{
  f = gFitter->ComputeChi2GlobalLOG(par);
}

/*--------------------------------------------------------------------*/
void FCNGlobal2EXPO(int&, double *, double& f, double *par, int)
/*--------------------------------------------------------------------*/
{
  f = gFitter->ComputeChi2Global2EXPO(par);
}

/*--------------------------------------------------------------------*/
void FCNSMSP(int&, double*, double& f, double* par, int)
/*--------------------------------------------------------------------*/
{
  f = gFitter->ComputeChi2SMSP(par);
}


/* ----------------------------- */
/* -- SMSP method --------------- */
/* ----------------------------- */

double TFitting::ComputeChi2SMSP(double* par) {

  fpSmspEstim_INIT;

  Double_t chisq = 0;

  Double_t delta;
  double signal;
  double theta = asin(sqrt(par[kFitU]*par[kFitU] + par[kFitV]*par[kFitV]));
  double cth2 = minval(1.,maxval(1-par[kFitU]*par[kFitU] - par[kFitV]*par[kFitV],0));
  double x, y, z, dist;

  TSMSPPoint lP;
  lP[SMSP_TYPE_ENERGY] = par[kFitE];
  lP[SMSP_TYPE_THETA]  = theta*kCdasUtil::RAD2DEG;

  for( unsigned int i=0;i<fNTanks;i++) {
    // space-time part
    x = fListCoords[i][0]-par[kFitX0];
    y = fListCoords[i][1]-par[kFitY0];
    z = fListCoords[i][2]-par[kFitZ0];
    if(par[kFitU]*par[kFitU]+par[kFitV]*par[kFitV]<=1) {
      dist = compute_distUV(par[kFitU],par[kFitV],x,y,z);
      delta  = fListTimes[i] - par[kFitT0] + FitCurvatureVariable(x,y,z,dist,par);
      ComputeVarTimeError(i,dist,cth2);
      delta /= fListTimesErrors[i];
      chisq += delta*delta;

      lP[SMSP_TYPE_DIST] = dist;
      signal = fpSmspEstim->VEM(lP);
      /*   cout << "---- Cuve " << i << " ----" << endl;
         cout << "E = " << lP[SMSP_TYPE_ENERGY] << endl;
         cout << "Theta = " << lP[SMSP_TYPE_THETA] << endl;
         cout << "Dist = " << lP[SMSP_TYPE_DIST] << endl;
         cout << "VEM estim = " << signal << endl;
      cout << "vem ok = " << (int)fpSmspEstim->VEMok() << endl;*/
      if ( fpSmspEstim->VEMok() ) {
        double signal_corrected = signal / compute_asymetry(x,y,dist,par[kFitU],par[kFitV]);
        delta = (fListSignals[i]-signal_corrected)/sqrt(signal_corrected);
        chisq += delta*delta;
	
	//cout << "chisq = " << chisq << endl;
      } else {
        // //do not use the tank because out of range of smsp data.
        chisq += 1000;
        //cout << "chisq = " << chisq << endl;
      }
    } else {
      delta  = (fListTimes[i]-par[kFitT0]+FitPlane(x,y,z,par));
      delta /= fListTimesErrors[i];
      chisq += delta*delta;
    }
  }

//  cout << "SMSP Chi2 returning " << chisq << endl;
    
  return chisq;
}
