/**
 *    @file TSmSp.cc
 *    @author Tristan Beau, beau@in2p3.fr
 *    @brief SMSP Data and Estimation class 
*/
#ifndef __DARWIN__
#include <netinet/in.h>
#endif
#include <assert.h>
#include "TSmSp.h"

#include <algorithm>

Short_t TSwapper::Short(Short_t a) {
  return (Short_t) ( ((a >> 8)&0xff) | ((a << 8)&0xff00) ) ;
}

UShort_t TSwapper::UShort(UShort_t a) {
  return (UShort_t) ( ((a >> 8)&0xff) | ((a << 8)&0xff00) ) ;
}

Int_t TSwapper::Long(Int_t a) {
  return	(Int_t) ( ((a >> 24)&0xff)      | ((a >> 8 )&0xff00 ) |
                   ((a << 8 )&0xff0000 ) | ((a << 24)&0xff000000) ) ;
}

UInt_t TSwapper::ULong(UInt_t a) {
  return	(UInt_t) ( ((a >> 24)&0xff)      | ((a >> 8 )&0xff00 ) |
                    ((a << 8 )&0xff0000 ) | ((a << 24)&0xff000000) ) ;
}

float TSwapper::Float(float a) {
  Int_t *b = (Int_t *)(&a) ;
  *b = ( ((*b >> 24)&0xff)      | ((*b >> 8 )&0xff00 ) |
         ((*b << 8 )&0xff0000 ) | ((*b << 24)&0xff000000) ) ;
  float *c = (float*)b;
  return *c;
}

bool TSwapper::ToSwap() { // Test if byte inversion is needed
  bool lB = ( ntohs(1234) == 1234 ) ;

  assert ( sizeof(Int_t) == 4 );
  assert ( sizeof(Short_t) == 2);

#ifdef SMSP_DEBUG
  cout << "Swap ? " << flush;
  cout << ((lB)?" yes ":" no ") << endl;
#endif

  return lB;
}

ostream &operator<<(ostream &stream, const TSMSPParam &aP) {
  stream << "Param ";
  switch( aP.fType ) {
  case SMSP_TYPE_ENERGY :
    stream << "Energy" ;
    break;
  case SMSP_TYPE_THETA  :
    stream << "Theta" ;
    break;
  case SMSP_TYPE_DIST   :
    stream << "Distance" ;
    break;
  case SMSP_TYPE_NATURE :
    stream << "Nature" ;
    break;
  default:
    stream << "Unknown type = " << (int)aP.fType << " !" ;
  }
  stream << " with repart " ;
  switch ( aP.fRepart ) {
  case SMSP_REPART_LOG :
    stream << "in log." ;
    break;
  case SMSP_REPART_LIN :
    stream << "linear." ;
    break;
  default:
    stream << "Unknown repart = " << (int)aP.fRepart << " !";
  }
  stream << endl << "Min = " << aP.fMin << " ,Max = " << aP.fMax << " and N = " << aP.fN << endl;

  return stream;
}

istream &operator>>(istream &stream, TSMSPParam &aP) {
  stream.read((char*)&aP,sizeof(TSMSPParam));
  if ( stream.gcount() != sizeof(TSMSPParam) ) {
    perror("Error while reading a TSMSPParam");
    exit(errno);
  }
  if ( TSwapper::ToSwap()) {
    aP.fN        = TSwapper::ULong(  aP.fN );
    aP.fRepart   = aP.fRepart ;
    aP.fReserved = TSwapper::UShort( aP.fReserved );
    aP.fMin      = TSwapper::Float(  aP.fMin );
    aP.fMax      = TSwapper::Float(  aP.fMax );
  }

  if ( aP.fRepart != SMSP_REPART_LOG && aP.fRepart != SMSP_REPART_LIN ) {
    cerr << "Parameter -> repart = " << (int)(aP.fRepart) << endl;
    cerr << "Bad repartition for parameter. Exiting" << endl;
    exit(1);
  }

  /*  if ( aP.fType == SMSP_TYPE_ENERGY ) {
      aP.fRepart = SMSP_REPART_LOG;
      aP.fMin = alog10((float)(aP.fMin-18));
      aP.fMax = alog10((float)(aP.fMax-18));
    }*/

#ifdef SMSP_DEBUG
  cout << aP ;
#endif

  return stream;
}

ostream &operator<<(ostream &stream, const TSMSPHeader &aH) {
  stream << "Number of param = " << aH.fNParam << endl;
  stream << "VEM in " << ((aH.fRepart==SMSP_REPART_LIN)?"linear":"log10") << endl;
  stream << "Comment = " << aH.fComment << endl;
  for ( ULong_t i=0; i<aH.fNParam; i++) {
    stream << "--------------------" << endl;
    stream << "Param " << i << endl;
    stream << aH.fpParam[i] << endl;
  }
  stream << "=================================" << endl;
  return stream;
}

istream &operator>>(istream &stream, TSMSPHeader &aH) {

  stream.read((char*)&aH.fNParam,sizeof(UShort_t));
  if ( stream.gcount() != sizeof(UShort_t)) {
    perror("Error while reading in TSMSPHeader the number of param");
    exit(errno);
  }

  stream.read((char*)&aH.fRepart,sizeof(unsigned char));
  if ( stream.gcount() != sizeof(unsigned char)) {
    perror("Error while reading in TSMSPHeader the value repartition type");
    exit(errno);
  }

  if ( aH.fRepart != SMSP_REPART_LOG && aH.fRepart != SMSP_REPART_LIN ) {
    cerr << "aH.fRepart = " << (int)(aH.fRepart) << endl;
    cerr << "Bad repartition (either log or lin) of data. Exiting." << endl;
    exit(1);
  }

  stream.read((char*)&aH.fReserved,sizeof(unsigned char));
  if ( stream.gcount() != sizeof(unsigned char)) {
    perror("Error while reading in TSMSPHeader the reserved byte");
    exit(errno);
  }

  stream.read((char*)aH.fComment,SMSP_COMMENT_SIZE);
  if ( stream.gcount() != SMSP_COMMENT_SIZE ) {
    perror("Error while reading a TSMSPHeader comment");
    exit(errno);
  }

  if ( TSwapper::ToSwap()) {
    aH.fNParam   = TSwapper::UShort ( aH.fNParam );
  }

  if ( aH.fpParam ) {
    delete[] aH.fpParam;
  }

  aH.fpParam = new TSMSPParam[aH.fNParam];
  for ( UInt_t i=0; i<aH.fNParam; i++)
    stream >> aH.fpParam[i];

#ifdef SMSP_DEBUG

  cout << aH ;
#endif

  return stream;
}

ostream &operator<<(ostream &stream, const TSMSPData &aD) {
  stream << aD.fH << endl;

  return stream;
}

istream &operator>>(istream &stream, TSMSPData &aD) {
  ULong_t k,n=1;

#ifdef SMSP_DEBUG
  cout << "---Reading TSMSPData " << endl ;
#endif

  assert ( sizeof(float) == 4 ); //if not, it will crash...

  stream >> aD.fH;

  if ( aD.fpData )
    delete[] aD.fpData;

  for ( ULong_t i=0; i<aD.fH.fNParam; i++ )
    n *= aD.fH.fpParam[i].fN;
  aD.fpData = new float[n];

  stream.read((char*)aD.fpData,sizeof(float)*n);
  if ( stream.gcount() != (streamsize)(sizeof(float)*n) ) {
    perror("Error while reading SMSPData");
    exit(errno);
  }

  if ( TSwapper::ToSwap() )
    for ( k=0; k<n; k++ )
      aD.fpData[k] = TSwapper::Float( aD.fpData[k] );

#ifdef SMSP_DEBUG
  cout << aD << endl;
  cout << "---End of reading TSMSPData" << endl;
#endif

  vector<float> f;
  for ( k=0; k<n; k++)
    f.push_back(aD.fpData[k]);

  //	cout << "min = " << *min_element(f.begin(),f.end()) << endl;
  //	cout << "max = " << *max_element(f.begin(),f.end()) << endl;

  return stream;
}

TSMSPEstim::TSMSPEstim(string aSMSPFileName) {
  ifstream lF( (fFileName=aSMSPFileName).c_str() ) ;
  if ( ! lF ) {
    cout << "Error while opening " << fFileName.c_str() << ". Exiting." << endl;
    perror("ifstream open error");
    exit(errno);
  }
  lF >> fData;
  cerr << "Read SMSP info from " << fFileName.c_str() << endl;
#ifdef SMSP_DEBUG
  cout << fData;
#endif

  lF.close();
  fExceed = true;

  FillParamVector();
  CreateInterpCoef(0,TSMSPInterpCoef(),fInterpCoef);
}

TSMSPEstim::~TSMSPEstim() {
  for ( UInt_t i=0; i<fParamVector.size(); i++) {
    fParamVector[i]->clear();
    delete fParamVector[i];
  }
}

void TSMSPEstim::FillParamVector() {
  vector<float> *f;

  for ( UInt_t i=0; i<fData.fH.fNParam; i++) {
    f = new vector<float>;
    fParamVector.push_back(f);
    f->push_back ( fData.fH.fpParam[i].fMin );

#ifdef SMSP_DEBUG
    cout << "type = " << (int)fData.fH.fpParam[i].fType << endl;
    cout << "min =  " << fData.fH.fpParam[i].fMin << endl;
    cout << "max =  " << fData.fH.fpParam[i].fMax << endl;
    cout << i << " 0 " << f->back() << endl;
#endif

    MinMap[fData.fH.fpParam[i].fType] = fData.fH.fpParam[i].fMin ;
    MaxMap[fData.fH.fpParam[i].fType] = fData.fH.fpParam[i].fMax ;
    if ( fData.fH.fpParam[i].fRepart == SMSP_REPART_LOG ) {
      MinMap[fData.fH.fpParam[i].fType] = alog10(MinMap[fData.fH.fpParam[i].fType]);
      MaxMap[fData.fH.fpParam[i].fType] = alog10(MaxMap[fData.fH.fpParam[i].fType]);
    }

    for ( ULong_t k=1; k < fData.fH.fpParam[i].fN; k++ ) {
      f->push_back
      ( fData.fH.fpParam[i].fMin + ((float)k) / (fData.fH.fpParam[i].fN-1)  *
        (fData.fH.fpParam[i].fMax - fData.fH.fpParam[i].fMin) );

#ifdef SMSP_DEBUG
      cout << i << " " << k << " " << f->back() << endl;
#endif

      fTypeIndex[fData.fH.fpParam[i].fType] = i;
    }
  }
}

ULong_t TSMSPEstim::GetDataIndex(TSMSPIndex &aV) {
  ULong_t j=0;
  for ( UInt_t p=0; p<fData.fH.fNParam; p++) {
    j *= fData.fH.fpParam[p].fN ;
    j += aV[p];
  }
  return j;
}

float TSMSPEstim::PhiFunc(const TSMSPInterpCoef &k, const TSMSPInterpPoint &P) const {
  // k -> coef -1 or 1, P the point in [-1 ; 1]
  assert ( k.size() == P.size() );

  float ret=1.;
  for( ULong_t i=0; i<k.size(); i++) {
    ret *= ( 1 + k[i]* ( (fData.fH.fpParam[i].fRepart == SMSP_REPART_LOG)?log10(P[i]):P[i]) ) ;
  }

  return ret / ( 1<<k.size() ) ;
}

float TSMSPEstim::InterpWithPhi(const TSMSPInterpPoint &aPoint, const TSMSPInterpValue &Values) const {

  assert ( (TSMSPInterpPoint::size_type)(1 << aPoint.size()) == Values.size() );
  float ret = 0.;
  for ( TSMSPInterpValueIter v=Values.begin(); v!=Values.end(); v++ )
    ret += v->second * PhiFunc( v->first , aPoint ) ;

  return ret;
}

void TSMSPEstim::CreateInterpCoef( UInt_t l , TSMSPInterpCoef k , vector<TSMSPInterpCoef> &V) {
  if ( l == fData.fH.fNParam ) {
    V.push_back(k);
  } else {
    for ( int j=-1; j<2; j+=2 ) {
      TSMSPInterpCoef ktemp = k;
      ktemp.push_back (j);
      CreateInterpCoef(l+1,ktemp,V);
    }
  }
}

float TSMSPEstim::VEM(TSMSPPoint &aP) {

  vector<vector<float>::iterator> i;
  TSMSPIndex pos;
  vector<float> coef;

  for ( UInt_t p=0; p<fData.fH.fNParam; p++) {
    unsigned char lType = fData.fH.fpParam[p].fType;

    vector<float>::iterator lIter =
      find_if ( fParamVector[ fTypeIndex[lType] ]->begin(),
                fParamVector[ fTypeIndex[lType] ]->end(),
                bind2nd(greater<float>(),
                        (fData.fH.fpParam[p].fRepart==SMSP_REPART_LOG) ? log10(aP[lType]):aP[lType]
                       ) ) ;

    if ( lIter == fParamVector[ fTypeIndex[lType] ]->end() ) {
      lIter --;
      fExceed = true;
    } else if ( lIter == fParamVector[ fTypeIndex[lType] ]->begin() ) fExceed = true;
    else fExceed = false;

    pos.push_back( (ULong_t)(lIter - fParamVector[ fTypeIndex[lType] ]->begin()) );

  }

  TSMSPInterpValue        lVal;
  TSMSPInterpPoint        lPoint;
  TSMSPIndex              tempPos;
  //Create InterpValues
  for ( vector<TSMSPInterpCoef>::const_iterator k=fInterpCoef.begin();
        k < fInterpCoef.end() ;
        k++ ) {
    tempPos = pos;
    for ( ULong_t i=0; i<tempPos.size(); i++ ) {
      if ( fParamVector[i]->
           size() != 1 ) { // standard case. If not, nothing to do.
        if ( (*k)[i] ==  1 && ! tempPos[i] )
          tempPos[i]++;
        if ( (*k)[i] == -1 &&  tempPos[i]  )
          tempPos[i]-- ;
      }
    }
    lVal [ *k ] = fData.fpData [ GetDataIndex( tempPos ) ] ;

#ifdef SMSP_DEBUG
    for ( vector<int>::const_iterator z=(*k).begin(); z<(*k).end(); z++) {
      cout << " " << *z ;
    }
    cout << " val = " << lVal[*k] << endl;
#endif

  }

  //Create InterpPoint position
  for ( UInt_t p=0; p<fData.fH.fNParam; p++) {

    float pos0 = fParamVector[p]
                 ->operator[]( pos[p] ? pos[p]-1 : pos[p]   ) ;
    float pos1 = fParamVector[p]->operator[]( pos[p] ? pos[p] : pos[p]+1 ) ;
    float val;

    val = ( fParamVector[p]->size() == 1) ? 0. :
          -1. + 2. * ( aP[ fData.fH.fpParam[p].fType ] - pos0 ) / ( pos1 - pos0 ) ;

    if ( val > 1. )
      val = 1.;
    if ( val < -1.)
      val = -1.;

    lPoint.push_back( val ) ;

#ifdef SMSP_DEBUG

    cout << "p=" << p << " " << lPoint.back() << " " << endl;
    if ( fabs(lPoint.back()) > 1 ) {
      cout << "pos0 = " << pos0 << " " << "pos1 = " << pos1 << " posp=" << pos[p] << " N=" <<  fData.fH.fpParam[p].fN << endl;
      cout << "aP[ fData.fH.fpParam[p].fType ] " << aP[ fData.fH.fpParam[p].fType ] << endl;
      cout << "fParamVector[p]->operator[]( pos[p] )" << fParamVector[p]->operator[]( pos[p] ) << endl;
      cout << "fParamVector[p]->operator[]( pos[p]-1 )" << fParamVector[p]->operator[]( pos[p] -1 ) << endl;
      cout << "fParamVector[p]->operator[]( pos[p]+1 )" << fParamVector[p]->operator[]( pos[p] +1 ) << endl;
    }
#endif

  }

#ifdef SMSP_DEBUG
  cout << "interp = " <<  InterpWithPhi( lPoint, lVal)
  << " nointerp = " << fData.fpData [ GetDataIndex(pos) ] << endl;
#endif

  return ( fData.fH.fRepart == SMSP_REPART_LOG ) ?
         alog10(InterpWithPhi( lPoint, lVal)) :
         InterpWithPhi(lPoint, lVal) ;

  //    return fData.fpData [ GetDataIndex (pos) ] ;

}
