/**
 *    @file TSmSp.h
 *    @author Tristan Beau, beau@in2p3.fr
 *    @brief Structure definition for data used in SMSP functions
*/


#ifndef TSMSP_H
#define TSMSP_H

#include <math.h>

#include <iostream>
#include <fstream>
#include <vector>
#include <map>
#include <algorithm>

#include <string>
#include <cstdio>
#include <errno.h>

#include <TROOT.h>

#define SMSP_COMMENT_SIZE 64

#define SMSP_REPART_LOG 0x1
#define SMSP_REPART_LIN 0x2

#define SMSP_TYPE_ENERGY 0x1
#define SMSP_TYPE_THETA  0x2
#define SMSP_TYPE_DIST   0x4
#define SMSP_TYPE_NATURE 0x8

//#define SMSP_DEBUG 1

#define alog10(A) pow( (float)10., (A) )

using namespace std ;

class TSwapper {
public:
  static Short_t Short(Short_t ) ;
  static UShort_t UShort(UShort_t) ;
  static Int_t Long(Int_t);
  static UInt_t ULong(UInt_t);
  static float Float(float);

  static bool ToSwap() ;
} ;

class TSMSPParam {
public:
  UInt_t       fN;
  unsigned char  fRepart; //log or lin
  unsigned char  fType  ; // E or theta or...
  UShort_t       fReserved;
  float          fMin ;
  float          fMax ;
};

ostream& operator<<(ostream &, const TSMSPParam &) ;
istream& operator>>(istream &, TSMSPParam &) ;

class TSMSPHeader {
public:
  TSMSPHeader()  {
    fpParam = 0 ;
  }
  ~TSMSPHeader() { }

  UShort_t       fNParam;
  unsigned char  fRepart;
  unsigned char  fReserved;
  char           fComment[SMSP_COMMENT_SIZE];
  TSMSPParam     *fpParam;
};

ostream& operator<<(ostream &, const TSMSPHeader &);
istream& operator>>(istream &, TSMSPHeader &);

class TSMSPData {
public:
  TSMSPData () {
    fpData = 0;
  }
  TSMSPHeader fH;
  float     *fpData;
};


ostream& operator<<(ostream &, const TSMSPData &);
istream& operator>>(istream &, TSMSPData &);

typedef map<unsigned char, float> TSMSPPoint;
typedef vector<float>             TSMSPInterpPoint;
typedef vector<int>               TSMSPInterpCoef;
typedef map<TSMSPInterpCoef,float> TSMSPInterpValue;
typedef TSMSPInterpValue::const_iterator TSMSPInterpValueIter;
typedef vector<ULong_t>     TSMSPIndex;

class TSMSPEstim {
public:
  TSMSPEstim(string );
  ~TSMSPEstim() ;
  float VEM( TSMSPPoint &)  ;
  map<unsigned char, float> MinMap ;
  map<unsigned char, float> MaxMap ;
  bool VEMok() const {
    return (! fExceed) ;
  }

protected:
  void FillParamVector();
  ULong_t GetDataIndex(TSMSPIndex &);
  float PhiFunc(const TSMSPInterpCoef &, const TSMSPInterpPoint &) const ;
  float InterpWithPhi(const TSMSPInterpPoint &, const TSMSPInterpValue &) const ;
  void CreateInterpCoef( UInt_t , TSMSPInterpCoef , vector<TSMSPInterpCoef> & );

private:
  string    fFileName;
  TSMSPData  fData;
  bool      fExceed;
  vector< vector<float>* > fParamVector;
  map< unsigned char, UInt_t > fTypeIndex ;
  vector<TSMSPInterpCoef>    fInterpCoef;
};

#endif
