/**
   @file ToolsLib.cc
   @author Benoit Revenu
   @brief this file implements the functions defined in ToolsLib.h
*/

#include <algorithm>
#include <fstream>
#include <utility>
#include <vector>
#include <math.h>
#include "TCdasUtil.h"
#include "STCconstants.h"
#include "ToolsLib.h"

template<class T>
ostream& operator<<(ostream& s, const vector<T>& vc) {
  long size = vc.size();
  if( size>0 )
    for( int i=0;i<size;i++ )
      s << vc[i] << " ";
  else
    cerr << endl;
  return s;
}

template ostream& operator<<(ostream&, const vector<char>&);
template ostream& operator<<(ostream&, const vector<string>&);
template ostream& operator<<(ostream&, const vector<int>&);
template ostream& operator<<(ostream&, const vector<unsigned int>&);
template ostream& operator<<(ostream&, const vector<long>&);
template ostream& operator<<(ostream&, const vector<double>&);

void uv2thetaphi(double u, double v, double& theta, double& phi) {
  double sq = sqrt( u*u + v*v );
  theta = asin(sq);
  phi = fmod(atan2(v,u)+kCdasUtil::TwoPi,kCdasUtil::TwoPi);
}

void thetaphi2uv(double theta, double phi, double& u, double& v) {
  u = sin(theta)*cos(phi);
  v = sin(theta)*sin(phi);
}

void compute_uv(const vector< vector<double> > & coords,
                const vector<double> & times, double * t0, double *u, double *v) {
  // we assume here that all the tanks are at the same altitude (1400 m)
  double a1 = 0, a2 = 0, a3 = 0, a4 = 0, a5 = 0, a6 = 0, a7 = 0, a8 = 0, a9 = 0;
  int numstation=coords.size();
  vector<double> si2(numstation);
  double rmstime = 35.e-9;

  for( int i=0;i<numstation;i++ ) {
    si2[i]=pow(rmstime,2.);
    a1 += pow(coords[i][0],2.)      /si2[i];
    a2 += coords[i][0]*coords[i][1] /si2[i];
    a3 += coords[i][0]              /si2[i];
    a4 += pow(coords[i][1],2.)      /si2[i];
    a5 += coords[i][1]              /si2[i];
    a6 += 1.                        /si2[i];
    a7 += coords[i][0]*times[i]     /si2[i];
    a8 += coords[i][1]*times[i]     /si2[i];
    a9 += times[i]                  /si2[i];
  }
  a1 /= kCdasUtil::CSPEEDSI;
  a2 /= kCdasUtil::CSPEEDSI;
  a4 /= kCdasUtil::CSPEEDSI;

  double denominateur = a1*a4*a6*kCdasUtil::CSPEEDSI
                        -a1*a5*a5-a2*a2*a6*kCdasUtil::CSPEEDSI
                        +2*a2*a3*a5-a3*a3*a4;

  *u = (-a7*(a4*a6*kCdasUtil::CSPEEDSI-a5*a5)
        +a8*(a2*a6*kCdasUtil::CSPEEDSI-a3*a5)
        -a9*(a2*a5-a3*a4)*kCdasUtil::CSPEEDSI) / denominateur;

  *v = (a7*(a2*a6*kCdasUtil::CSPEEDSI-a3*a5)
        -a8*(a1*a6*kCdasUtil::CSPEEDSI-a3*a3)
        +a9*(a1*a5-a2*a3)*kCdasUtil::CSPEEDSI) / denominateur;

  *t0 = (a7*(a2*a5-a3*a4)
         -a8*(a1*a5-a2*a3)
         +a9*(a1*a4-a2*a2)*kCdasUtil::CSPEEDSI) / denominateur;
}

void compute_barycentre(const vector< vector<double> > & coords,
                        const vector<double> & pmt, const vector<double> times,
                        double *x, double *y, double *z, double *t) {
  *x = 0;
  *y = 0;
  *z = 0;
  *t = 0;
  double totw = 0;
  double tw;
  int numstation = pmt.size();
  int nst = 0;

  vector<int> idx;
  vector<double> sorted_signals = pmt;
  sort(sorted_signals.begin(),sorted_signals.end());
  reverse(sorted_signals.begin(),sorted_signals.end());
  for( int i=0;i<numstation;i++ )
    for( int j=0;j<numstation;j++ )
      if( pmt[j] == sorted_signals[i] )
        idx.push_back(j);

  for( int i=0;i<numstation;i++ ) {
    tw = pow(pmt[idx[i]],.33333333);
    totw += tw;
    *x += tw*coords[idx[i]][0];
    *y += tw*coords[idx[i]][1];
    *z += tw*coords[idx[i]][2];
    *t += tw*times[idx[i]];
    nst++;
    if( nst >= 3 )
      break;
  }

  *x /= totw;
  *y /= totw;
  *z /= totw;
  *t /= totw;
}

void compute_distances(double theta, double phi, double xav, double yav, double zav,
                       const vector< vector<double> > & listcoords,
                       vector<double> & listdist) {
  int nstat = listcoords.size();
  listdist.clear();
  for( int i=0;i<nstat;i++ ) {
    listdist.push_back(compute_dist(theta,phi,
                                    listcoords[i][0]-xav,
                                    listcoords[i][1]-yav,
                                    listcoords[i][2]-zav));
  }
}

void compute_distancesUV(double u, double v, double xav, double yav, double zav,
                         const vector< vector<double> > & listcoords,
                         vector<double> & listdist) {
  int nstat = listcoords.size();
  listdist.clear();
  for( int i=0;i<nstat;i++ ) {
    listdist.push_back(compute_distUV(u,v,
                                      listcoords[i][0]-xav,
                                      listcoords[i][1]-yav,
                                      listcoords[i][2]-zav));
  }
}

double compute_dmax(const vector< vector<double> >& listcoords) {
  int size = listcoords.size();
  double dx, dy, dz, dist;
  double dmax = 0;
  for( int i=0;i<size;i++ ) {
    for( int j=i+1;j<size;j++ ) {
      dx = listcoords[j][0]-listcoords[i][0];
      dy = listcoords[j][1]-listcoords[i][1];
      dz = listcoords[j][2]-listcoords[i][2];
      dist = dx*dx + dy*dy + dz*dz;
      if( dist > dmax )
        dmax = dist;
    }
  }
  return sqrt(dmax);
}

double compute_surface(const vector< vector<double> >& listcoords) {
  double dx01 = listcoords[0][0]-listcoords[1][0];
  double dy01 = listcoords[0][1]-listcoords[1][1];
  double dz01 = listcoords[0][2]-listcoords[1][2];

  double dx02 = listcoords[0][0]-listcoords[2][0];
  double dy02 = listcoords[0][1]-listcoords[2][1];
  double dz02 = listcoords[0][2]-listcoords[2][2];

  double dx12 = listcoords[1][0]-listcoords[2][0];
  double dy12 = listcoords[1][1]-listcoords[2][1];
  double dz12 = listcoords[1][2]-listcoords[2][2];

  double a = sqrt(dx01*dx01 + dy01*dy01 + dz01*dz01);
  double b = sqrt(dx02*dx02 + dy02*dy02 + dz02*dz02);
  double c = sqrt(dx12*dx12 + dy12*dy12 + dz12*dz12);

  double s = (a+b+c)/2;

  return sqrt(s*(s-a)*(s-b)*(s-c));
}

double compute_angleXY(double x1, double y1, double x2, double y2) {
  double d1 = sqrt(x1*x1+y1*y1);
  double d2 = sqrt(x2*x2+y2*y2);
  return acos((x1*x2+y1*y2)/(d1*d2));
}

double max_angle_in_triangle(const vector<double>& a,
                             const vector<double>& b,
                             const vector<double>& c) {
  vector<double> angles;
  // angle ba-ca
  angles.push_back(compute_angleXY(b[0]-a[0],b[1]-a[1],c[0]-a[0],c[1]-a[1]));
  // angle cb-ab
  angles.push_back(compute_angleXY(c[0]-b[0],c[1]-b[1],a[0]-b[0],a[1]-b[1]));
  // angle ac-bc
  angles.push_back(compute_angleXY(a[0]-c[0],a[1]-c[1],b[0]-c[0],b[1]-c[1]));
  return *max_element(angles.begin(),angles.end());
}


double compute_dist(double theta, double phi, double dxstat, double dystat, double dzstat) {
  double st = sin(theta);
  double u = st*cos(phi);
  double v = st*sin(phi);
  double w = cos(theta);
  return sqrt(dxstat*dxstat+dystat*dystat+dzstat*dzstat-(dxstat*u+dystat*v+dzstat*w)*(dxstat*u+dystat*v+dzstat*w));
}

double compute_distUV(double u, double v, double dxstat, double dystat, double dzstat) {
  // compute distance to axis
  // neglect z
  //  double w = 0;
  //  dzstat = 0;
  double w = sqrt(1-u*u-v*v);
  return sqrt(dxstat*dxstat+dystat*dystat+dzstat*dzstat-
              (dxstat*u+dystat*v+dzstat*w)*(dxstat*u+dystat*v+dzstat*w));
}

void compute_error_angle(double u,double du,double v,double dv,
                         double *dtheta,double *dphi) {
  *dtheta = (fabs(u*du)+fabs(v*dv))/(sqrt(u*u+v*v)*sqrt(1.-u*u-v*v)) * kCdasUtil::RAD2DEG;
  *dphi = (fabs(u*dv)+fabs(v*du))/(u*u+v*v) * kCdasUtil::RAD2DEG;
}

double compute_LDFHPFunc(double dist, double theta) {
  // theta is in radians
  static double first_theta = -1;
  static double eta = 0;
  if( theta != first_theta ) {
    first_theta = theta;
    if( theta < 60*kCdasUtil::DEG2RAD )
      eta = 4.735-1.236/cos(theta);
    else
      eta = 2.672-0.161/cos(theta);
  }
  // dist must be in meters
  return pow(dist/1000.,-eta);
}

double compute_LDFLOGFunc(double dist, double beta, double gamma) {
	  if( dist >= kDmin ) return pow(dist/kDref,beta+gamma*log(dist/kDref));
	  else
	    {
	      static double rappD = kDmin/kDref;
	      static double lrappD = log(rappD);
	      double fact = pow(rappD,beta+gamma*lrappD);
	      return fact*pow(dist/kDmin,beta+2*gamma*lrappD);
	    }
}

void GetBetaGammaFromTheta(int primary, double theta, double& n_beta, double& gamma) {
  // primary = 0 : proton
  // primary = 1 : iron
  if( primary == 0) {
    if(theta*kCdasUtil::RAD2DEG <= 25) {
      n_beta = interpol(3, 25, -4.4, -4.08, theta*kCdasUtil::RAD2DEG);
      gamma = interpol(3, 25, -0.8, -0.681, theta*kCdasUtil::RAD2DEG);
    } else if(theta*kCdasUtil::RAD2DEG<=36) {
      n_beta = interpol(25, 36, -4.08, -3.73, theta*kCdasUtil::RAD2DEG);
      gamma = interpol(25, 36, -0.681, -0.430, theta*kCdasUtil::RAD2DEG);
    } else if(theta*kCdasUtil::RAD2DEG<=45) {
      n_beta = interpol(36, 45, -3.73, -3.53, theta*kCdasUtil::RAD2DEG);
      gamma = interpol(36, 45, -0.430, -0.362, theta*kCdasUtil::RAD2DEG);
    } else if(theta*kCdasUtil::RAD2DEG<=53) {
      n_beta = interpol(45, 53, -3.53, -3.09, theta*kCdasUtil::RAD2DEG);
      gamma = interpol(45, 53, -0.362, -0.374, theta*kCdasUtil::RAD2DEG);
    } else if(theta*kCdasUtil::RAD2DEG<=60) {
      n_beta = interpol(53, 60, -3.09, -2.6, theta*kCdasUtil::RAD2DEG);
      gamma = interpol(53, 60, -0.374, -0.391, theta*kCdasUtil::RAD2DEG);
    } else if(theta*kCdasUtil::RAD2DEG<=66) {
      n_beta = interpol(60, 66, -2.6, -2.4, theta*kCdasUtil::RAD2DEG);
      gamma = interpol(60, 66, -0.391, -0.4, theta*kCdasUtil::RAD2DEG);
    } else {
      n_beta = interpol(60, 66, -2.6, -2.4, theta*kCdasUtil::RAD2DEG);
      gamma = -0.4;
    }
  } else if( primary == 1 ) {
    if(theta*kCdasUtil::RAD2DEG <= 25) {
      n_beta = interpol(3, 25, -4.3, -3.92, theta*kCdasUtil::RAD2DEG);
      gamma = interpol(3, 25, -0.8, -0.654, theta*kCdasUtil::RAD2DEG);
    } else if(theta*kCdasUtil::RAD2DEG<=36) {
      n_beta = interpol(25, 36, -3.92, -3.54, theta*kCdasUtil::RAD2DEG);
      gamma = interpol(25, 36, -0.654, -0.422, theta*kCdasUtil::RAD2DEG);
    } else if(theta*kCdasUtil::RAD2DEG<=45) {
      n_beta = interpol(36, 45, -3.54, -3.31, theta*kCdasUtil::RAD2DEG);
      gamma = interpol(36, 45, -0.422, -0.369, theta*kCdasUtil::RAD2DEG);
    } else if(theta*kCdasUtil::RAD2DEG<=53) {
      n_beta = interpol(45, 53, -3.31, -2.91, theta*kCdasUtil::RAD2DEG);
      gamma = interpol(45, 53, -0.369, -0.4, theta*kCdasUtil::RAD2DEG);
    } else if(theta*kCdasUtil::RAD2DEG<=60) {
      n_beta = interpol(53, 60, -2.91, -2.58, theta*kCdasUtil::RAD2DEG);
      gamma = interpol(53, 60, -0.4, -0.436, theta*kCdasUtil::RAD2DEG);
    } else if(theta*kCdasUtil::RAD2DEG<=66) {
      n_beta = interpol(60, 66, -2.58, -2.4, theta*kCdasUtil::RAD2DEG);
      gamma = interpol(60, 66, -0.436, -0.4, theta*kCdasUtil::RAD2DEG);
    } else {
      n_beta = interpol(60, 66, -2.58, -2.4, theta*kCdasUtil::RAD2DEG);
      gamma = -0.4;
    }
  }
  // FIXME overwriting beta to fit data. For next CDAS, add a eData to primary
  if( primary == 0 ) {
    double x=cos(theta);
    double a=2.5, b=1.1, c=-3.9;
    n_beta=a*(1-x)*(1-x)+b*(1-x)+c;
  }
}

double compute_LDFPhotonFunc(double dist, double theta, double energy) {
  double cost = cos(theta);
  double l10 = log10(energy); // en supposant que energy est en EeV

  double p1 =-16.9794, p2 =  8.6100;
  double p3 = 55.1539, p4 =-14.5217;
  double p5 =-30.8004, p6 =  6.7653;

  double p7 =  0.4692, p8 = -0.4075;
  double p9 = -2.5513, p10=  1.1413;
  double p11=  1.3864, p12= -0.6575;

  double p13=-10.3210, p14=  7.6583;
  double p15= 22.3469, p16=-12.9432;
  double p17=-11.5959, p18=  6.6693;

  double p19=  0.5611, p20= -0.3654;
  double p21= -1.1863, p22=  0.6917;
  double p23=  0.5988, p24= -0.3538;

  double a10=p1+p2*l10;
  double a11=p3+p4*l10;
  double a12=p5+p6*l10;

  double a1=a10+a11*cost+a12*cost*cost;

  double b10=0.01*(p7+p8*l10);
  double b11=0.01*(p9+p10*l10);
  double b12=0.01*(p11+p12*l10);
  double b1=b10+b11*cost+b12*cost*cost;

  double a20=p13+p14*l10;
  double a21=p15+p16*l10;
  double a22=p17+p18*l10;

  double a2=a20+a21*cost+a22*cost*cost;

  double b20=0.01*(p19+p20*l10);
  double b21=0.01*(p21+p22*l10);
  double b22=0.01*(p23+p24*l10);

  double b2=b20+b21*cost+b22*cost*cost;

  return exp(a1+b1*dist)+exp(a2+b2*dist);
}

double compute_LDF2EXPOFunc(double dist, double theta, double energy, int primary) {
  double cost = cos(theta);
  double l10 = log10(energy); // en supposant que energy est en EeV

  double a1,b1,a2,b2;

  // PRIMARY NUCLEI
  if (primary == 0 || primary == 1) {

    double p1,p2,p3,p4,p5,p6,p7,p8,p9,p10,p11,p12,p13,p14,p15,p16,p17,p18,p19,p20,p21,p22,p23,p24;
    
    if (primary == 0) {
      // Assuming proton in 2EXPO energy reconstruction
      /*
      p1 =   -0.69621;
      p2 =   -0.04145;
      p3 =   11.68875;
      p4 =    6.88068;
      p5 =   -5.66083;
      p6 =   -4.16986;
      
      p7 =   -0.00306;
      p8 =    0.00216;
      p9 =   -0.00023;
      p10=   -0.00724;
      p11=    0.00009;
      p12=    0.00405;

      p13=   -0.70534;
      p14=   -0.78164;
      p15=   -4.81297;
      p16=   12.85703;
      p17=    1.74534;
      p18=   -6.45722;
      
      p19=    0.00117;
      p20=   -0.00006;
      p21=   -0.00031;
      p22=   -0.00313;
      p23=    0.00054;
      p24=    0.00134;
      */
      p1 =    3.21863716;
      p2 =   -2.19019405;
      p3 =    5.72321361;
      p4 =    9.25733089;
      p5 =   -3.39605489;
      p6 =   -4.46671643;
      	     
      p7 =   -0.01031064;
      p8 =    0.00619227;
      p9 =    0.01169249;
      p10=   -0.01245066;
      p11=   -0.00494882;
      p12=    0.00535546;
	      
      p13=    6.44868803;
      p14=   -5.34460513;
      p15=  -16.62003874;
      p16=   19.65463934;
      p17=    8.28156028;
      p18=   -9.63529474;
      	     
      p19=   -0.00106699;
      p20=    0.00133732;
      p21=    0.00265568;
      p22=   -0.00468953;
      p23=   -0.00110071;
      p24=    0.00196676;

    } else if (primary == 1) {
      // Assuming iron in 2EXPO energy reconstruction
      p1 =   0.25818585;
      p2 =   0.59633787;
      p3 =   8.05644965;
      p4 =   5.68453320;
      p5 =  -2.20712530;
      p6 =  -3.99121251;
      
      p7 =  -0.00130040;
      p8 =   0.00038603;
      p9 =  -0.00290589;
      p10=  -0.00371885;
      p11=   0.00026366;
      p12=   0.00275289;

      p13=  -6.29267612;
      p14=   4.07474252;
      p15=  10.15733260;
      p16=   0.52884399;
      p17=  -4.37873507;
      p18=  -0.46768036;
      
      p19=   0.00286034;
      p20=  -0.00163844;
      p21=  -0.00522679;
      p22=   0.00114697;
      p23=   0.00225134;
      p24=  -0.00059132;
    }

    double a10=p1+p2*l10;
    double a11=p3+p4*l10;
    double a12=p5+p6*l10;

    a1=a10+a11*cost+a12*cost*cost;

    double b10=(p7+p8*l10);
    double b11=(p9+p10*l10);
    double b12=(p11+p12*l10);

    b1=b10+b11*cost+b12*cost*cost;

    double a20=p13+p14*l10;
    double a21=p15+p16*l10;
    double a22=p17+p18*l10;

    a2=a20+a21*cost+a22*cost*cost;

    double b20=(p19+p20*l10);
    double b21=(p21+p22*l10);
    double b22=(p23+p24*l10);

    b2=b20+b21*cost+b22*cost*cost;

    // PRIMARY PHOTONS
  } else if (primary  == 2 ) {
    // Assuming photon in 2EXPO energy reconstruction
    double p1 =-16.9794, p2 =  8.6100;
    double p3 = 55.1539, p4 =-14.5217;
    double p5 =-30.8004, p6 =  6.7653;

    double p7 =  0.4692, p8 = -0.4075;
    double p9 = -2.5513, p10=  1.1413;
    double p11=  1.3864, p12= -0.6575;

    double p13=-10.3210, p14=  7.6583;
    double p15= 22.3469, p16=-12.9432;
    double p17=-11.5959, p18=  6.6693;

    double p19=  0.5611, p20= -0.3654;
    double p21= -1.1863, p22=  0.6917;
    double p23=  0.5988, p24= -0.3538;

    double a10=p1+p2*l10;
    double a11=p3+p4*l10;
    double a12=p5+p6*l10;

    a1=a10+a11*cost+a12*cost*cost;

    double b10=0.01*(p7+p8*l10);
    double b11=0.01*(p9+p10*l10);
    double b12=0.01*(p11+p12*l10);

    b1=b10+b11*cost+b12*cost*cost;

    double a20=p13+p14*l10;
    double a21=p15+p16*l10;
    double a22=p17+p18*l10;

    a2=a20+a21*cost+a22*cost*cost;

    double b20=0.01*(p19+p20*l10);
    double b21=0.01*(p21+p22*l10);
    double b22=0.01*(p23+p24*l10);

    b2=b20+b21*cost+b22*cost*cost;

  } else {
    cout << "Unknown primary type" << primary << "in compute_LDF2EXPOFunc" << endl;
    cout << "##################################################################################" << endl;
    cout << "##################################################################################" << endl;
    cout << "PRIMARY TYPE IS NOT CORRECT ! YOU CAN CHOOSE ONLY BETWEEN :" << endl;
    cout << "0 (proton, default), 1 (iron) or 2 (photon)." << endl;
    cout << "##################################################################################" << endl;
    cout << "##################################################################################" << endl;
    exit(-1);
  }

  return exp(a1+b1*dist)+exp(a2+b2*dist);
}

double compute_energy(string ldftype, double SRef, double theta, 
		      int primary, double dsRef, double* de) {
  static int first = 1;
  double energy;
  double cost = cos(theta);
  double power = 1./0.95;
  // compute energy
  // NKG-like ledf. The energy relation used here is: 
  // a = 0.37-0.51 sec(theta) + 0.30 sec^2(theta),
  // b = 1.27-0.27 sec(theta) + 0.08 sec^2(theta),
  // E =a x (S_1000)^b	(in [EeV]).
  if (ldftype == "NKG") {
    double a = 0.37-0.51/cost + 0.30/cost/cost;
    double b = 1.27-0.27/cost + 0.08/cost/cost;
    energy = a*pow(SRef, b);
    if (de != 0x0) *de = energy*b*dsRef/SRef;
    return energy;
  }
  else if( ldftype == "HP" ) {
    energy = pow(SRef*sqrt(1+11.8*pow((1/cost-1),2))/7.8,power);
    if( de != 0x0 )
      *de = pow(SRef*sqrt(1+11.8*pow((1/cost-1),2))/7.8,power-1.)*dsRef;
  } else if( ldftype == "LOG" ) {
    // proton : 0
    // iron : 1
    double n_sRef;
    if( primary == 0 ) {
      if(theta*kCdasUtil::RAD2DEG <= 25) {
        n_sRef = interpol(3, 25, 6.56, 5.95, theta*kCdasUtil::RAD2DEG);
      } else if(theta*kCdasUtil::RAD2DEG<=36) {
        n_sRef = interpol(25, 36, 5.95, 5.3, theta*kCdasUtil::RAD2DEG);
      } else if(theta*kCdasUtil::RAD2DEG<=45) {
        n_sRef = interpol(36, 45, 5.3, 4.0, theta*kCdasUtil::RAD2DEG);
      } else if(theta*kCdasUtil::RAD2DEG<=53) {
        n_sRef = interpol(45, 53, 4.0, 2.78, theta*kCdasUtil::RAD2DEG);
      } else if(theta*kCdasUtil::RAD2DEG<=60) {
        n_sRef = interpol(53, 60, 2.78, 1.78, theta*kCdasUtil::RAD2DEG);
      } else if(theta*kCdasUtil::RAD2DEG<=66) {
        n_sRef = interpol(60, 66, 1.78, 1.4, theta*kCdasUtil::RAD2DEG);
      } else {
        n_sRef = interpol(66, 70, 1.4, 1.14, theta*kCdasUtil::RAD2DEG);
      }
    } else if( primary == 1 ) {
      if(theta*kCdasUtil::RAD2DEG <= 25) {
        n_sRef = interpol(3, 25, 8.0, 7.2, theta*kCdasUtil::RAD2DEG);
      } else if(theta*kCdasUtil::RAD2DEG<=36) {
        n_sRef = interpol(25, 36, 7.2, 6.0, theta*kCdasUtil::RAD2DEG);
      } else if(theta*kCdasUtil::RAD2DEG<=45) {
        n_sRef = interpol(36, 45, 6.0, 4.5, theta*kCdasUtil::RAD2DEG);
      } else if(theta*kCdasUtil::RAD2DEG<=53) {
        n_sRef = interpol(45, 53, 4.5, 3.15, theta*kCdasUtil::RAD2DEG);
      } else if(theta*kCdasUtil::RAD2DEG<=60) {
        n_sRef = interpol(53, 60, 3.15, 2.18, theta*kCdasUtil::RAD2DEG);
      } else if(theta*kCdasUtil::RAD2DEG<=66) {
        n_sRef = interpol(60, 66, 2.18, 1.73, theta*kCdasUtil::RAD2DEG);
      } else {
        n_sRef = interpol(66, 70, 1.73, 1.5, theta*kCdasUtil::RAD2DEG);
      }
    } else {
      if( first ) {
        cout << "##################################################################################" << endl;
        cout << "##################################################################################" << endl;
        cout << "PRIMARY TYPE IS NOT CORRECT : YOU CAN CHOOSE ONLY 0 (proton, default) or 1 (IRON)." << endl;
        cout << "CONTINUING WITH PROTON...." << endl;
        cout << "##################################################################################" << endl;
        cout << "##################################################################################" << endl;
        first = 0;
      }
      if( de != 0x0 )
        *de = -1;
      return compute_energy(ldftype,SRef,theta,0);
    }
    energy = pow(SRef/n_sRef,power);
    if( de != 0x0 )
      *de = power*energy*dsRef/SRef;
  } else if( ldftype == "Photon" ) {
    double a= 1605.0+(-6763.4+( 8469.4-3345.2*cost)*cost)*cost;
    double b=-1650.4+( 6845.0+(-8405.5+3256.3*cost)*cost)*cost;
    energy=pow(10.,((SRef-a)/b));
    if( de != 0x0 )
      *de = energy*dsRef*log(10.)/b;
  } else if( ldftype == "2EXPO" ) {
    double a,b;
    if(primary==0)
      //protons
      {
	a= -344.6+( 2135.5+(-4580.3+2503.5*cost)*cost)*cost;
	b=  412.7+(-2461.1+( 5124.1-2758.4*cost)*cost)*cost;
      } 
    else if(primary==1)   
      //iron
      {
	double P00,P01,P02,P03,P10,P11,P12,P13;
	P00= -592.2355957;
	P01= 2971.2131348;
	P02=-5420.1337891;
	P03= 2681.7363281;
	P10=  689.7931519;
	P11=-3410.3225098;
	P12= 6115.4399414;
        P13=-2995.1457520;
	a= P00+(P01+(P02+P03*cost)*cost)*cost;
	b= P10+(P11+(P12+P13*cost)*cost)*cost;
      } 
    else if(primary==2)
      //photons
      {
	a= 1605.0+(-6763.4+( 8469.4-3345.2*cost)*cost)*cost;
	b=-1650.4+( 6845.0+(-8405.5+3256.3*cost)*cost)*cost;
      } 
    else {
      cout << "Unknown primary type" << primary << "in compute_energy" << endl;
      cout << "##################################################################################" << endl;
      cout << "##################################################################################" << endl;
      cout << "PRIMARY TYPE IS NOT CORRECT ! YOU CAN CHOOSE ONLY BETWEEN :" << endl;
      cout << "0 (proton, default), 1 (iron) or 2 (photon)." << endl;
      cout << "##################################################################################" << endl;
      cout << "##################################################################################" << endl;
      exit(-1);
    }
    energy=pow(10.,((SRef-a)/b));
  } else if (ldftype=="CIC") {
    // 18.6, 1.41, 1.0, 0.21
    //double S38=18.6;
    double Xg = 1.41;
    double Xm = 1.0;
    double eMu=0.21;
    double TL38=0.953; 
    double cic38 = exp(-Xg*(1.0-Xm))+eMu*TL38;
    double x=1/cos(theta);
    double TL=x/(1.+0.4244*sqrt(x*x-1));
    double cic=(exp(-Xg*(x/1.27-Xm))+eMu*TL)/cic38;
    double S = SRef/cic;
    energy = 0.162*exp(1.063*log(S));
  } 
  else if (ldftype=="CICINFILL") {
    // 18.6, 1.41, 1.0, 0.21
    // double S38=18.6;
    // const double cos38 = cos(38.*kCdasUtil::DEG2RAD); 
    // const double x = 1/cos(theta);
    // const double cic38 = exp(-1.83*(x-1./cos38));
    // const double S = SRef/cic38;
    // energy = 0.00919*exp(1.0849*log(S));
    
    // energy conversion compatible with ICRC 2011
    const double a = 1.56; //+- 0.04
    const double b = -1.0; //+- 0.15
    const double cossq= pow(cos(theta), 2 ) - 
      pow(cos( 35.*kCdasUtil::DEG2RAD), 2);
    const double s35= SRef/(1.+a*cossq+ b*cossq*cossq);
    const double A = 0.0107367; //EeV
    const double B = 1.02826;
    energy = A* pow(s35, B);
    if (de != 0x0) 
      *de = energy*B*dsRef/SRef;
    
  } 
  else if (ldftype=="NEWFD") {
    // 18.6, 1.41, 1.0, 0.21
    //double S38=18.6;
    double Xg = 1.41;
    double Xm = 1.0;
    double eMu=0.21;
    double TL38=0.953; 
    double cic38 = exp(-Xg*(1.0-Xm))+eMu*TL38;
    double x=1/cos(theta);
    double TL=x/(1.+0.4244*sqrt(x*x-1));
    double cic=(exp(-Xg*(x/1.27-Xm))+eMu*TL)/cic38;
    double S = SRef/cic;
    energy = 0.143*pow(S, 1.078); // in EeV
    const double b = 1.078;
    if (de != 0x0) *de = energy*b*dsRef/SRef;
    
  } else if (ldftype=="MCPLUSMU") {
    // From Maximo: It takes the EM part calculated from the models with the
    // Xmax corresponding to the elongation rate, and a Muonic part that
    // corresponds to Xmax from data and a normalization of 1.37 (i.e. 37% more
    // muons that proton QGSJetII), which is what we get at 10 EeV.
    double sth = 1./cos(theta);
    double E = 1., E_p = 0.; // in 10 EeV

    // precision of iteration 0.1%
    while (fabs(E - E_p)/E >= 0.001) {
      E_p = E;
      // get Xmax and Nmu for previous energy estimate
      double lx = log10(E_p) + 0.3;
      double xmax = (lx > 0) ? 0.849 + 0.0537*lx : 0.849 + 0.0909 * lx;
      double nmu = 1.37 * pow(E_p, -0.04);
  
      double f_em = pow((sth - xmax + 0.689)/0.802, 0.802/0.113)
        * exp((0.113 - sth + xmax)/0.113);
      double f_mu = pow((sth - xmax + 0.228)/0.573, 0.573/1.267)
        * exp((0.345 - sth + xmax)/1.267);
      E = pow(SRef / (22.3*f_em + 15.6*nmu*f_mu), 1./0.97);
    }
    energy= E * 10.; // in EeV
  } else {
    cout << "Unknown SRef -> Energy conversion for ldf = " << ldftype << endl;
    cout << "Energy set to -1" << endl;
    if( de != 0x0 )
      *de = -1;
    energy = -1.;
  }

  return energy;
}

double compute_S1000 ( string ldftype, double E, double theta, TSMSPEstim *apSmspEstim , int primary ) { //theta in rad.
  double s1000 = -1;
  if ( ldftype == "SMSP" ) {
    if ( ! apSmspEstim ) {
      cout << "Invalid TSMSPEstim pointer in s1000 estimation." << endl;
      cout << "S1000 set to default incorrect value " << s1000 << endl;
    } else {
      TSMSPPoint lP;
      lP[SMSP_TYPE_ENERGY] = E;
      lP[SMSP_TYPE_THETA]  = theta * kCdasUtil::RAD2DEG; // conv to deg
      lP[SMSP_TYPE_DIST] = 1000;
      s1000 = apSmspEstim->VEM(lP);
    }
  } else if( ldftype == "2EXPO" ) {
    double cost = cos(theta);
    double a,b;

    if (primary==0) {
      // protons      
      a= -344.6+( 2135.5+(-4580.3+2503.5*cost)*cost)*cost;
      b=  412.7+(-2461.1+( 5124.1-2758.4*cost)*cost)*cost;
    } else if (primary==1) {      
      // iron
      double P00,P01,P02,P03,P10,P11,P12,P13;
      P00= -592.2355957;
      P01= 2971.2131348;
      P02=-5420.1337891;
      P03= 2681.7363281;
      P10=  689.7931519;
      P11=-3410.3225098;
      P12= 6115.4399414;
      P13=-2995.1457520;
      a= P00+(P01+(P02+P03*cost)*cost)*cost;
      b= P10+(P11+(P12+P13*cost)*cost)*cost;
    } else if (primary==2) {
    // photons
      a= 1605.0+(-6763.4+( 8469.4-3345.2*cost)*cost)*cost;
      b=-1650.4+( 6845.0+(-8405.5+3256.3*cost)*cost)*cost;
    } else {
      cout << "Unknown primary type" << primary << "in compute_S1000" << endl;
      cout << "##################################################################################" << endl;
      cout << "##################################################################################" << endl;
      cout << "PRIMARY TYPE IS NOT CORRECT ! YOU CAN CHOOSE ONLY BETWEEN :" << endl;
      cout << "0 (proton, default), 1 (iron) or 2 (photon)." << endl;
      cout << "##################################################################################" << endl;
      cout << "##################################################################################" << endl;
      exit(-1);
    }

    s1000=a+b*log10(E);
  } else {
    cout << "Unknown Energy->S1000 conversion for ldf = " << ldftype << endl;
    cout << "S1000 set to " << s1000 << endl;
  }
  return s1000 ;
}

void compute_error_energy(string ldftype, double SRef,double dSRef,double theta,double dtheta,
                          double *de) {
  double u,v,w,x;
  double du,dv,dw,dx;
  *de = 0;
  if( ldftype == "HP" ) {
    u=1/cos(kCdasUtil::DEG2RAD*theta);
    du=fabs(sin(kCdasUtil::DEG2RAD*theta))/(cos(kCdasUtil::DEG2RAD*theta)*cos(kCdasUtil::DEG2RAD*theta))*kCdasUtil::DEG2RAD*dtheta;
    v=(u-1)*(u-1);
    dv=fabs(2*(u-1))*du;
    w=sqrt(1+11.8*v);
    dw=11.8/(2*w)*dv;
    x=SRef*w/7.8;
    dx=fabs(SRef/7.8*dw)+fabs(dSRef/7.8*w);
    *de=1.05*pow(x,0.05)*dx;
  } else if( ldftype == "LOG" ) {}
  else if( ldftype == "2EXPO" ) {}
}

void compute_dE_dS1000_2EXPO(double E, double theta, double SRef, double *de, double *ds1000) {
  double coszen=cos(kCdasUtil::DEG2RAD*theta);
  double log10e=log10(E);

  double p1= 0.7353862;
  double p2=-0.1555241;
  double p3=-1.4084034;
  double p4= 0.2701467;
  double p5= 0.9527095;
  double p6=-0.1998810;

  double c0=p1+p2*log10e;
  double c1=p3+p4*log10e;
  double c2=p5+p6*log10e;

  double sig=c0+c1*coszen+c2*coszen*coszen;
  *de=E*(pow(M_E,sig)-1.);
  *ds1000=SRef*(pow(M_E,0.95*sig)-1.); 
}

void fit_ellipse(unsigned int npts, double *x, double *y, double *w,
		 double & xm, double & ym, double &sa, double &sb, double &theta,
		 string method_orientation, string method_area)
{
  double a,b;
  GetPointsOrientation(npts,x,y,w,xm,ym,a,b,sa,sb,theta,method_orientation);
  double e = sa/sb;
  double saa(0), sbb(0);
  if( method_area == "triangle" )
    {  
      double surface_triangle = 1500.*1500.*sqrt(3.)/4;
      double surface_cuve = (npts-2)*surface_triangle;      
      sbb=sqrt(surface_cuve/(M_PI*e)); // rescale ellipse axis
      saa=e*sbb;
      //      cout << "triangle = " << surface_cuve << endl;
    }
  else if( method_area == "smax" )
    {
      vector<int> edges = get_edges(npts,x,y);
      double * xe = new double[edges.size()];
      double * ye = new double[edges.size()];
      for( unsigned int i=0;i<edges.size();i++ )
	{
	  xe[i]=x[edges[i]];
	  ye[i]=y[edges[i]];
	  //	  cout << xe[i] << " " << ye[i] << endl;
	}
      double area = ComputePolygonArea((int)edges.size(),xe,ye);
      double e = sa/sb;
      //      cout<< "smax : polygon area " << area << endl;
      sbb=sqrt(area/(M_PI*e));
      saa=e*sbb;
      //      cout << "ellipse area : " << M_PI*saa*sbb << endl;
    }
  else if( method_area == "entoure")//tous les points sont dans l'ellipse
    {
      double saa=1400;
      double sbb=saa/e;
      double * toto = new double[npts];
      for(unsigned int k=0;k<npts;k++) toto[k]=0;
      
      for(unsigned int i=0;i<npts;i++)
	{
	  for(unsigned int j=0;j<npts;j++)
	    {
	      double distanceij=sqrt((x[i]-x[j])*(x[i]-x[j])+(y[i]-y[j])*(y[i]-y[j]));
	      if(distanceij<2000) toto[i]=1000000;
	    }
	}
      for(unsigned int i=0;i<npts;i++)
	{
	  if(toto[i]==1000000)
	    {
	      double phi=atan2(y[i]-ym,x[i]-xm);
	      double xellipse=saa*cos(phi)*cos(theta)-sbb*sin(phi)*sin(theta)+xm;
	      double yellipse=sbb*sin(phi)*cos(theta)+saa*cos(phi)*sin(theta)+ym;
	      //double distance=sqrt((x[i]-xm)*(x[i]-xm)+(y[i]-ym)*(y[i]-ym));
	      //	      double distanceellipse=sqrt((xellipse-xm)*(xellipse-xm)+(yellipse-ym)*(yellipse-ym));
	      while(xellipse<x[i] || yellipse<y[i]) saa+=10;
	      sbb=saa/e;
	    }
  	}
    }
  else if( method_area == "points" ) // comptage du nombre de plus proches voisins
    {
      double * u = new double[npts];
      double * v = new double[npts];	
      for(unsigned int i=0;i<npts;i++)
	{
	  int n=0;
	  for(unsigned int j=0;j<npts;j++)
	    {
	      if( j != i )
		{
		  double d=sqrt((x[i]-x[j])*(x[i]-x[j])+(y[i]-y[j])*(y[i]-y[j]));
		  if(d<2000) n++;
		}
	    }
	  if( n<6 )
	    {
	      u[i]=x[i];
	      v[i]=y[i];
	    }
	  else
	    {
	      u[i]=1000000000;
	      v[i]=1000000000;
	    }
	}
      double umax=0;
      double vmax(0);
      for(unsigned int i=0;i<npts;i++)
	{
	  if(u[i]!=1000000000 || v[i]!=1000000000)
	    {
	      if(u[i]>umax)
		{
		  umax=u[i];
		  vmax=v[i];
		}
	    }
	}
      double surface1=0;
      double surface2=0;
      double vbis=vmax;
      double ubis=umax;
      double phi;
      double x1(0);
      double y1(0);
      double phiprec=atan2((vmax-ym),(umax-xm));
      if(phiprec<0) phiprec=fmod(phiprec+kCdasUtil::TwoPi,kCdasUtil::TwoPi);
      do
	{
	  double phimin=100000000.;
	  for (unsigned int k=0;k<npts;k++)
	    {	
	      if(u[k]!=1000000000 || v[k]!=1000000000)
		{					
		  phi=atan2((v[k]-ym),(u[k]-xm))-phiprec;
		  if(phi<0) phi=fmod(phi+kCdasUtil::TwoPi,kCdasUtil::TwoPi);		  
		  if(phimin>phi && phi!=0)
		    {
		      x1=x[k];
		      y1=y[k];
		      phimin=phi;	
		    }		  
		  double d1=sqrt((x1-xm)*(x1-xm)+(y1-ym)*(y1-ym));
		  double d2=sqrt((ubis-xm)*(ubis-xm)+(vbis-ym)*(vbis-ym));
		  double r=sqrt((ubis-x1)*(ubis-x1)+(vbis-y1)*(vbis-y1));
		  double angle=acos((-d1*d1+d2*d2+r*r)/(2*r*d2));
		  surface1+=r*d2*sin(angle)/2;
		  phiprec+=phimin;
		  phiprec=fmod(phiprec+kCdasUtil::TwoPi,kCdasUtil::TwoPi);	
		  ubis=x1;
		  vbis=y1;
		}
	    }	  
	} while(ubis!=umax || vbis!=vmax);      
      for(unsigned int i=0;i<npts-1;i++)
	{
	  if(u[i]!=1000000000 || v[i]!=1000000000)
	    {
	      surface2+=u[i]*v[i+1]-u[i+1]*v[i];
	    }
	  surface2 += u[npts-1]*v[0] - u[0]*v[npts-1];
	}
      surface2/=2;      
      double e = sa/sb;
      sbb=sqrt(surface1/(M_PI*e));
      saa=e*sbb;
      //      cout<< "points=" << surface1 << " " << surface2 << " " << endl;
      delete [] u;
      delete [] v;
    }
  sa = saa;
  sb = sbb;
  return;
}

// theta is between 0 and 360
void GetPointsOrientation(int npts, double * x, double * y, double * w, double & xm, double & ym,
			  double & a, double & b, double & sa, double & sb, double & theta,
			  string method)
{
  xm=0;
  ym=0;
  double sx2=0,sy2=0,x2m=0,y2m=0,xym=0,cxy=0;
  double theta1(0), theta2(0), delta, ar1, ar2, br1, br2;
  double totw = 0;
  double distances1=0,distances2=0;
  double dcarree1=0,dcarree2=0;
  double saa=0,sbb=0;
  for( int i=0;i<npts;i++ )
    {
      xm += w[i]*x[i];
      ym += w[i]*y[i];
      x2m += w[i]*x[i]*x[i];
      y2m += w[i]*y[i]*y[i];
      xym += w[i]*x[i]*y[i];
      totw += w[i];
    }
  xm /= totw;
  ym /= totw;
  x2m /= totw;
  y2m /= totw;
  xym /= totw;
  
  sx2=x2m-(xm*xm);
  sy2=y2m-(ym*ym);	
  cxy=xym-(xm*ym);

  if( method == "distances" )
    {
      delta=(sx2-sy2)*(sx2-sy2)+4*(cxy*cxy);
      ar1=(-sx2+sy2+sqrt(delta))/(2*cxy);
      ar2=(-sx2+sy2-sqrt(delta))/(2*cxy);
      br1=ym-ar1*xm;
      br2=ym-ar2*xm;
      theta1=atan(ar1);
      theta2=atan(-1/ar1);
      a = ar1;
      b = br1;
    }
  else if( method =="linear" )
    {
      a=(xym-xm*ym)/sx2;
      b=ym-a*xm;
      theta1=atan(a);
      theta2=atan(-1/a);
    }
  for( int j=0;j<npts;j++ )
    {
      double tmp1 = fabs((y[j]-b-a*x[j])/(sqrt(1+a*a)));
      double tmp2 = fabs((y[j]-ym-(xm/a)+(x[j]/a))/(sqrt(1+(1/(a*a)))));
      distances1 += w[j] * tmp1;
      distances2 += w[j] * tmp2;
      dcarree1 += distances1 * tmp1;
      dcarree2 += distances2 * tmp2;
    }      
  distances1 /= totw;
  distances2 /= totw;
  dcarree1   /= totw;
  dcarree2   /= totw;

  sbb=sqrt( dcarree1-( distances1*distances1 ))/2; // 1/2 grand axe
  saa=sqrt( dcarree2-( distances2*distances2 ))/2; // 1/2 petit axe      

  if( saa>sbb)
    {
      theta = theta1*kCdasUtil::RAD2DEG;
      sa=saa;
      sb=sbb;
    }
  else 
    {
      theta = theta2*kCdasUtil::RAD2DEG;
      sa=sbb;
      sb=saa;
    }
  theta = fmod(theta+360.,360);
  double nt = theta;
  if( nt >= 90 && nt < 270 ) nt -= 180;
  else if( nt >= 270 && nt < 360 ) nt -= 360;
  theta=nt;
  return;
}

double ComputePolygonArea(int npts, double * x, double * y)
{
  double surface=0; 	
  for( int i=0;i<npts-1;i++ ) surface+=x[i]*y[i+1]-x[i+1]*y[i];
  surface += x[npts-1]*y[0]-x[0]*y[npts-1];	
  surface/=2;
  return fabs(surface)/1e6;
}

vector<int> get_edges(const vector<int>& lsids, const vector< vector<double> >& coords)
{  
  unsigned int npts = lsids.size();
  double * x = new double[npts];
  double * y = new double[npts];
  for( unsigned int i=0;i<npts;i++ )
    {
      x[i] = coords[i][0];
      y[i] = coords[i][1];
    }
  vector<int> tmp = get_edges(npts,x,y);
  delete [] x;
  delete [] y;
  vector<int> edges;
  for( unsigned int i=0;i<tmp.size();i++ ) edges.push_back(lsids[tmp[i]]);
  return edges;
}
 
vector<int> get_edges(int npts, double * x, double * y)
{
  vector<int> edges;
  // get tank with minimum y
  double ymin = 1e11, ymax = -ymin;
  unsigned int lsid_idx_ymin(0), lsid_idx, next_idx, lsid_idx_prev(0), lsid_idx_ymax(0);
  for( int i=0;i<npts;i++ )
    {
      if( y[i] < ymin )
	{
	  ymin = y[i];
	  lsid_idx_ymin = i;
	}
      if( y[i] > ymax )
	{
	  ymax = y[i];
	  lsid_idx_ymax = i;
	}
    }
  double prev_phi = 0, phimin;
  bool after_max = false;
  lsid_idx = lsid_idx_ymin; // initialisation
  do
    {
      next_idx = lsid_idx;
      phimin = 1e9;
      for( unsigned int i=0;i<(unsigned int)npts;i++ )
	{	  
	  if( i == lsid_idx || i == lsid_idx_prev ) continue;
	  double phi=fmod(kCdasUtil::RAD2DEG*atan2(y[i]-y[lsid_idx],x[i]-x[lsid_idx])+360.,360.);
	  if( lsid_idx == lsid_idx_ymax )
	    {
	      after_max = true;
	      prev_phi = 0;
	    }
	  if( phi < phimin )
	    {
	      if( !after_max || (after_max && phi > fmod(prev_phi+180.,360)) )
		{
		  next_idx = i;
		  phimin = phi;
		}
	    }
	}
      edges.push_back(lsid_idx);
      lsid_idx_prev = lsid_idx;
      lsid_idx = next_idx;
      prev_phi = phimin;
    } while( lsid_idx != lsid_idx_ymin );
  return edges;
}

float a25(float r) {
  return //0.203E-01 +  0.856E-04*r - 0.387E-07*r*r +  0.245E-11*pow(r,3);//corsika 100 EeV
    0.152E-01 + 0.138E-03*r - 0.555E-07*r*r + 0.570E-11*pow(r,3);//aires all energies
}
float a36(float r) {
  return //0.480E-01 +  0.202E-03*r - 0.114E-06*r*r +  0.167E-10*pow(r,3);//corsika
    0.465E-01 + 0.229E-03*r - 0.112E-06*r*r + 0.166E-10*pow(r,3);//aires
}
float a45(float r) {
  return //0.911E-01 +  0.295E-03*r - 0.189E-06*r*r +  0.356E-10*pow(r,3);//corsika
    0.913E-01 + 0.309E-03*r - 0.185E-06*r*r + 0.320E-10*pow(r,3);//aires
}
float a53(float r) {
  return //0.179 + 0.166E-03*r - 0.145E-06*r*r +  0.369E-10*pow(r,3);//corsika
    0.155 + 0.220E-03*r - 0.171E-06*r*r + 0.343E-10*pow(r,3);//aires
}
float a60(float r) {
  return //0.136 + 0.969E-05*r + 0.336E-08*r*r;//corsika
    0.141 + 0.376E-04*r - 0.113E-07*r*r;//aires
}
float a66(float r) {
  return 0.279E-01 + 0.693E-04*r - 0.908E-08*r*r;//aires
}
float a72(float r) {
  return 0.219E-01 + 0.370E-04*r - 0.149E-08*r*r;//aires
}

double compute_asymetry(double x, double y, double r, double u, double v, int method, double e) {
#if 0
  double alpha, uksi, vksi, cosksi;
  double sq = sqrt(u*u + v*v);
  //  double costheta = sqrt(1 - u*u + v*v); // BUG ????????????????
  double costheta = sqrt(1 - u*u - v*v); // BUG ????????????????
  double cosphi = u/sq;
  double sinphi = v/sq;

  if(costheta<0.3) alpha = 0.05;
  else alpha = 1.56*(1-costheta)*(costheta-0.25);
  uksi = costheta*cosphi*x+costheta*sinphi*y;
  vksi = -sinphi*x+cosphi*y;
  cosksi = uksi/sqrt(uksi*uksi+vksi*vksi);
  //  cout << "multiply by : " << (1+alpha*atan(r/300.)*cosksi) << endl;
  return (1+alpha*atan(r/300.)*cosksi);
#else

  double sq = sqrt(u*u + v*v);
  double costheta = sqrt(1 - u*u - v*v);
  double th = acos(costheta)*kCdasUtil::RAD2DEG;
  double cosphi = u/sq;
  double sinphi = v/sq;
  double Y = cosphi*y-sinphi*x;
  double X = (sinphi*y+cosphi*x)*costheta;
  double azimut = atan2(Y,X);
  double alpha = 0;
  double a1=0,a2=0,b1=0,b2=0;

  if (method==0 || costheta<0.5) {
    // Parametrisation initiale CDAS: Amplitude alpha=alpha(theta,r) 
  if(th<=25) {
    b1 = 0;
    a1 = 0;
    b2 = 25;
    a2 = a25(r);
  } else if(th<=36) {
    b1 = 25;
    a1 = a25(r);
    b2 = 36;
    a2 = a36(r);
  } else if(th<=45) {
    b1 = 36;
    a1 = a36(r);
    b2 = 45;
    a2 = a45(r);
  } else if(th<=53) {
    b1 = 45;
    a1 = a45(r);
    b2 = 53;
    a2 = a53(r);
  } else if(th<=60) {
    b1 = 53;
    a1 = a53(r);
    b2 = 60;
    a2 = a60(r);
  } else if(th<=66) {
    b1 = 60;
    a1 = a60(r);
    b2 = 66;
    a2 = a66(r);
  } else if(th<=72) {
    b1 = 66;
    a1 = a66(r);
    b2 = 72;
    a2 = a72(r);
  } else {
    b1 = 72;
    a1 = a72(r);
    b2 = 90;
    a2 = 0;
  }
  alpha = a2 + (th-b2)*(a2-a1)/(b2-b1);
  } else {
    // Amplitude alpha=a*sin(pi*r/b), a=a(primtype,energie,theta), de meme pour b
    // si l argument de sinus >= pi, alpha=0
    // parametrisation en trois branches separees pour les protons (method==1), 
    // les noyaux de fer (method==2) et les photons (method==3) 
    double cz[7],l10e[3];
    double a[7][3],b[7][3];

    cz[0]=0.50000;
    cz[1]=0.64279;
    cz[2]=0.76604;
    cz[3]=0.86602;
    cz[4]=0.93969;
    cz[5]=0.98481;
    cz[6]=1.00000;

    l10e[0]=1.00000;
    l10e[1]=1.69897;
    l10e[2]=2.00000;

    if (method==1) {
      // protons
       a[0][0]= 0.12931;
       a[1][0]= 0.25065;
       a[2][0]= 0.21910;
       a[3][0]= 0.14449;
       a[4][0]= 0.07350;
       a[5][0]= 0.03472;
       a[6][0]= 0.00101;

       a[0][1]= 0.19832;
       a[1][1]= 0.28346;
       a[2][1]= 0.24104;
       a[3][1]= 0.15231;
       a[4][1]= 0.08472;
       a[5][1]= 0.03995;
       a[6][1]= 0.00589;

       a[0][2]= 0.21101;
       a[1][2]= 0.29106;
       a[2][2]= 0.23047;
       a[3][2]= 0.13686;
       a[4][2]= 0.07286;
       a[5][2]= 0.04207;
       a[6][2]= 0.00352;

       b[0][0]= 2066.01855;
       b[1][0]= 2103.47754;
       b[2][0]= 2164.13330;
       b[3][0]= 2180.20410;
       b[4][0]= 2688.58716;
       b[5][0]= 3200.00000;
       b[6][0]= 2275.96802;

       b[0][1]= 2797.88696;
       b[1][1]= 2813.93506;
       b[2][1]= 2854.52930;
       b[3][1]= 2937.42334;
       b[4][1]= 2924.92041;
       b[5][1]= 3199.99829;
       b[6][1]= 1535.51892;

       b[0][2]= 3200.00000;
       b[1][2]= 3059.18921;
       b[2][2]= 3070.81104;
       b[3][2]= 3171.65527;
       b[4][2]= 3200.00000;
       b[5][2]= 2889.60986;
       b[6][2]= 1561.53748;

    } else if(method==2) {
      // fer
       a[0][0]= 0.08376;
       a[1][0]= 0.18481;
       a[2][0]= 0.19162;
       a[3][0]= 0.15225;
       a[4][0]= 0.08995;
       a[5][0]= 0.04494;
       a[6][0]= 0.01219;

       a[0][1]= 0.14449;
       a[1][1]= 0.23463;
       a[2][1]= 0.20980;
       a[3][1]= 0.14563;
       a[4][1]= 0.09291;
       a[5][1]= 0.04292;
       a[6][1]= 0.00000;

       a[0][2]= 0.15945;
       a[1][2]= 0.23970;
       a[2][2]= 0.21186;
       a[3][2]= 0.14244;
       a[4][2]= 0.08641;
       a[5][2]= 0.03994;
       a[6][2]= 0.01088;

       b[0][0]=  2046.83887;
       b[1][0]=  2293.80078;
       b[2][0]=  2333.30688;
       b[3][0]=  2248.89526;
       b[4][0]=  2491.44556;
       b[5][0]=  2077.66797;
       b[6][0]=  1744.40552;

       b[0][1]=  3200.00000;
       b[1][1]=  2902.01318;
       b[2][1]=  2924.05762;
       b[3][1]=  2828.77075;
       b[4][1]=  2986.04077;
       b[5][1]=  2953.61035;
       b[6][1]=  1998.14819;

       b[0][2]=  3200.00000;
       b[1][2]=  3200.00000;
       b[2][2]=  3091.53223;
       b[3][2]=  3199.99976;
       b[4][2]=  3048.40796;
       b[5][2]=  2820.80884;
       b[6][2]=  1200.00000;

    } else if(method==3) {
      // photons

       a[0][0]= 0.48123;
       a[1][0]= 0.46799;
       a[2][0]= 0.28095;
       a[3][0]= 0.16786;
       a[4][0]= 0.09240;
       a[5][0]= 0.00000;
       a[6][0]= 0.00000;

       a[0][1]= 0.58613;
       a[1][1]= 0.46542;
       a[2][1]= 0.25768;
       a[3][1]= 0.15662;
       a[4][1]= 0.05738;
       a[5][1]= 0.02143;
       a[6][1]= 0.00000;

       a[0][2]= 0.64512;
       a[1][2]= 0.44787;
       a[2][2]= 0.27802;
       a[3][2]= 0.14083;
       a[4][2]= 0.05302;
       a[5][2]= 0.06251;
       a[6][2]= 0.00127;

       b[0][0]= 1384.85583;
       b[1][0]= 1787.19067;
       b[2][0]= 2183.20093;
       b[3][0]= 1896.06360;
       b[4][0]= 2800.00000;
       b[5][0]= 1348.26135;
       b[6][0]= 1200.79565;

       b[0][1]= 2189.32471;
       b[1][1]= 2604.62939;
       b[2][1]= 2800.00000;
       b[3][1]= 2800.00000;
       b[4][1]= 2800.00000;
       b[5][1]= 1200.00403;
       b[6][1]= 1355.99451;

       b[0][2]= 2520.32471;
       b[1][2]= 2800.00000;
       b[2][2]= 2800.00000;
       b[3][2]= 2799.99390;
       b[4][2]= 2799.99707;
       b[5][2]= 2799.97681;
       b[6][2]= 1602.07251;

    } else {
      cout << "##################################################################################" << endl;
      cout << "##################################################################################" << endl;
      cout << "compute_asymetry(...): unknown method " << method << endl;
      cout << "when calling the finction, use method=fPrimary+1, fPrimary is the primary type" << endl;
      cout << "YOU CAN CHOOSE BETWEEN : 0 (proton, default), 1 (iron) or 2 (photon)." << endl;
      cout << "##################################################################################" << endl;
      cout << "##################################################################################" << endl;
      exit(-1);
    }

    int iz1,iz2;
    if (costheta < 0.5) {
      iz1=0;
      iz2=1;
    } else {
      for (int iz=0; iz<6; iz++) {
	iz1=iz;
	iz2=iz+1;
	if (costheta >= cz[iz1] && costheta < cz[iz2])
	  break;	
      }
    }

    double loge=log10(e);
    int ie1,ie2;

    if (loge > 2.) {
      ie1=1;
      ie2=2;
    } else if (loge < 1.) {
      ie1=0;
      ie2=1;      
    } else {
      for (int ie=0; ie<2; ie++) {
	ie1=ie; 
	ie2=ie+1;
	if (loge >= l10e[ie1] && loge < l10e[ie2])
	  break;
      }
    }

    //    cout << iz1 << iz2 << ie1 << ie2 << endl;

    double ai,bi;
    intp2dim(costheta,loge,cz[iz1],cz[iz2],l10e[ie1],l10e[ie2],a[iz1][ie1],a[iz1][ie2],a[iz2][ie1],a[iz2][ie2],ai);
    intp2dim(costheta,loge,cz[iz1],cz[iz2],l10e[ie1],l10e[ie2],b[iz1][ie1],b[iz1][ie2],b[iz2][ie1],b[iz2][ie2],bi);

    if (r > bi) {
      alpha=0.;
    } else {
      alpha=ai*sin(M_PI*r/bi);
    }

    if(alpha < 0.)
      alpha=0.;

    //    cout << "Asym: " << " " << alpha << " " << ai << " " << bi << " " << loge << " " << costheta << " " << r << " " << azimut << endl;

  }
  return 1./(1+cos(azimut)*alpha);
#endif
}

void intp2dim(double x, double y, double x1, double x2, double y1, double y2, double z11, double z12, double z21, double z22, double &z) {

  double dx1=x-x1;
  double dx2=x2-x;

  double dy1=y-y1;
  double dy2=y2-y;
 
  z=(z11*dy2+z12*dy1)*dx2+(z21*dy2+z22*dy1)*dx1;
  z/=(x2-x1);
  z/=(y2-y1);
}


double lnStirling(int n) {
  if(n==0 || n==1)
    return log(1.);
  else if(n==2)
    return log(2.);
  else if(n==3)
    return log(6.);
  else if(n==4)
    return log(24.);
  else if(n==5)
    return log(120.);

  return 0.5*log(kCdasUtil::TwoPi*n)+n*log(n/M_E);
}

double ProbPoisson(int k, double mu) {
  if (k == 0) return exp(-mu);

  double lnst = lnStirling(k);
  double lnPoisson = k*log(mu)-lnst-mu;
  double poisson = exp(lnPoisson);

  return poisson;
}

double distance2Core(double dx, double dy, double u, double v) {
  return sqrt(dx * dx + dy * dy - sqr(dx * u + dy * v));
}


float interpol(float x1, float x2, float y1, float y2, float x) {
  return (((y2-y1)/(x2-x1))*(x-x2) + y2);
}

//inversion of 4x4 symetric matrix
int matinv4(double *a, double *b) {
  double r11,
  r12, r22,
  r13, r23, r33,
  r14, r24, r34, r44,
  s11, s12, s22, s13, s23, s33, s14, s24, s34, s44;

  if (a[0] <= 0)
    return 0;
  r11 = sqrt(a[0]);
  r12 = a[1] / r11;
  r13 = a[3] / r11;
  r14 = a[6] / r11;
  if ((r22 = a[2] - r12 * r12) <= 0)
    return 0;
  r22 = sqrt(r22);
  r23 = (a[4] - r12 * r13) / r22;
  r24 = (a[7] - r12 * r14) / r22;
  if ((r33 = a[5] - r13 * r13 - r23 * r23) <= 0)
    return 0;
  r33 = sqrt(r33);
  r34 = (a[8] - r13 * r14 - r23 * r24) / r33;
  if ((r44 = a[9] - r14 * r14 - r24 * r24 - r34 * r34) <= 0)
    return 0;
  r44 = sqrt(r44);

  s11 = 1 / r11;
  s12 = -s11 * r12 / r22;
  s13 = -(s11 * r13 + s12 * r23) / r33;
  s14 = -(s11 * r14 + s12 * r24 + s13 * r34) / r44;
  s22 = 1 / r22;
  s23 = -(s22 * r23) / r33;
  s24 = -(s22 * r24 + s23 * r34) / r44;
  s33 = 1 / r33;
  s34 = -s33 * r34 / r44;
  s44 = 1 / r44;

  b[0] = s11 * s11 + s12 * s12 + s13 * s13 + s14 * s14;
  b[1] = s12 * s22 + s13 * s23 + s14 * s24;
  b[2] = s22 * s22 + s23 * s23 + s24 * s24;
  b[3] = s13 * s33 + s14 * s34;
  b[4] = s23 * s33 + s24 * s34;
  b[5] = s33 * s33 + s34 * s34;
  b[6] = s14 * s44;
  b[7] = s24 * s44;
  b[8] = s34 * s44;
  b[9] = s44 * s44;

  return 1;
}
