/**
   \file
   Declaration of HybridEvent
 
   \author Stefano Argir Antoine Letessier-Selvon
*/

#ifndef __AugerEvent_h_
#define __AugerEvent_h_

#include <TObject.h>
#include <vector>
#include <EyeEvent.hh>

class IoSdEvent;

typedef unsigned long long AugerEventId;        // should be EventId, but we have a name conflict

/**
 * \class AugerEvent
 *
 * \brief The Auger Event
 * Data structure including one SD Event (IoSd) and 1..n FD Event   
 * (TEyeEvent)
 *
 * \todo Add boundary checks 
*/

class AugerEvent : public TObject {

public:
  enum TDataType { eReal, eCLF, eUnknown };

  AugerEvent();
  virtual ~ AugerEvent();

  AugerEvent(const AugerEvent & theAugerEvent);
  AugerEvent& operator=(const AugerEvent & theAugerEvent);

  /// return the id of the event
  /** The id is assigned by the merger as an incremental number
   *
   *  more documentation here */
  AugerEventId Id() const {
    return EventId;
  };

  /// return true if the event has both FD and SD
  bool IsHybrid() const {
    return HasSd() && Eyes.size();
  };

  /// return true if the SD event exists
  bool HasSd() const {
    return (SEvent != 0);
  };

  /// return the SD event
  IoSdEvent & Sd();
  const IoSdEvent& Sd() const;

  /// The container of TEyeEvents (FD events)
  typedef std::vector < TEyeEvent > EyesContainer;

  /// A read and write  iterator over the eye events
  typedef EyesContainer::iterator EyeIterator;
  typedef EyesContainer::const_iterator ConstEyeIterator;

  /// return the first available TEyeEvent
  EyeIterator EyesBegin() { return Eyes.begin();};
  ConstEyeIterator EyesBegin() const { return Eyes.begin(); }
  
  /// return one past the last available TEyeEvent
  EyeIterator EyesEnd() {return Eyes.end();};
  ConstEyeIterator EyesEnd() const { return Eyes.end(); }
  
  /// return the TEyeEvent at position @param pos in the list
  TEyeEvent & Eye(unsigned int pos) {return Eyes[pos];};
  const TEyeEvent& Eye(unsigned int pos) const { return Eyes[pos]; }
  
  /// Add the SD event to the hybrid
  void PushEvent(const IoSdEvent & sdevent);

  /// Add the FD event to the hybrid
  void PushEvent(const TEyeEvent & eyeevent) {Eyes.push_back(eyeevent);}

public:
  TDataType Type;             //! does not go to file (carried by the AugerKey)
  AugerEventId EventId;       //! should not go to file. Id is read from the key
  IoSdEvent *SEvent;
  EyesContainer Eyes;


  ClassDef(AugerEvent, 2);

};                              // AugerEvent



#endif                          // __AugerEvent_h_
