/**
   \file
   Implementation of AugerFile
 
   \author Stefano Argir Antoine Letessier-Selvon
*/

#include <AugerFile.h>
#include <AugerEvent.h>
#include <AugerKey.h>
#include <IoSdData.h>

#include <string>
#include <iostream>
#include <TFile.h>
#include <TObjArray.h>
#include <TKey.h>

#include <iostream>
#include <sstream>
#include <string>

using namespace std;

AugerFile::AugerFile():
fCurrentKey(0) {}
;

AugerFile::~AugerFile() {
  // delete all open files
  for (FileContainer::iterator it = fFiles.begin();
       it != fFiles.end(); it++)
    delete *it;
  fFiles.clear();
  delete fCurrentKey;
};

AugerFile::AugerFile(int argc, char **argv) {
  fCurrentKey = 0;
  for (int i = 0; i < argc; i++)
    this->Open(argv[i], AugerFile::eRead);
};

AugerFile::AugerFile(vector<string> v) {
  fCurrentKey = 0;
  for (unsigned int i = 0; i < v.size(); i++)
    this->Open(v[i].c_str(), AugerFile::eRead);
};

AugerFile::AugerFile(const char *filename, Mode mode) {

  fCurrentKey = 0;
  this->Open(filename, mode);
};

AugerFile::Status AugerFile::Open(const char *filename, Mode mode) {

  // translate mode for ROOT
  std::string rootmode;
  switch (mode) {

  case eRead:
    rootmode = "read";
    break;

  case eWrite:
    rootmode = "recreate";
    break;

  case eAppend:
    rootmode = "update";
    break;

  default:
    std::cerr << "AugerFile:: Unknown mode " << mode << std::endl;
    return eFailure;
  }

  TFile *file = TFile::Open(filename, rootmode.c_str());

  if (!file)
    return eFailure;

  fFiles.push_back(file);

  // set our pointers to beginning
  fCurrentKey = new TIter(file->GetListOfKeys());
  fCurrentFile = fFiles.begin();
  (*fCurrentFile)->cd();
  return eSuccess;
};

unsigned int AugerFile::NumberOfEvents() {
  unsigned int ret=0;
  for (unsigned int i=0;i<fFiles.size();i++) {
    ret+=fFiles[i]->GetNkeys();
  }
  return ret;
}

AugerFile::Status AugerFile::Open(const char *filenames) {

  stringstream files;
  files << filenames;
  string buffer;

  // separate the file names assuming they are separated by space
  // create the TFiles and push them in the list
  while (files >> buffer) {
    TFile *file = TFile::Open(buffer.c_str(), "read");
    if (file->IsZombie())
      return eFailure;
    fFiles.push_back(file);
  }

  // set pointers to beginning
  fCurrentFile = fFiles.begin();
  (*fCurrentFile)->cd();
  fCurrentKey = new TIter((*fCurrentFile)->GetListOfKeys());
  return eSuccess;
};


static AugerEventId __eventId(const char *Name) {
  return strtoull(Name, NULL, 10);
};

AugerFile::Status AugerFile::ReadObj(TKey * key, AugerEvent & event) {
  (*fCurrentFile)->cd();
  // read the object with that key
  AugerEvent *tmp = dynamic_cast < AugerEvent * >(key->ReadObj());
  if (!tmp)
    return eFailure;

  event = *tmp;
  event.EventId = __eventId(key->GetName());
  if (event.SEvent) {
    event.SEvent->AugerId = event.EventId;
    event.SEvent->_SetUp();
  }
  delete tmp;

  return eSuccess;
};

AugerFile::Status AugerFile::ReadNext(AugerEvent & event) {

  // try to grab the next key
  TKey *key = dynamic_cast < TKey * >((*fCurrentKey) ());

  // if it is empty, go to next file
  if (!key) {
    delete fCurrentKey;
    fCurrentKey = 0;
    fCurrentFile++;
    if (fCurrentFile == fFiles.end())
      return eEOF;
    fCurrentKey = new TIter((*fCurrentFile)->GetListOfKeys());
    key = dynamic_cast < TKey * >((*fCurrentKey) ());
  }
  // read the object with that key
  return ReadObj(key, event);
};

AugerFile::Status AugerFile::ReadNextHybrid(AugerEvent & event) {

  // try to grab the next key
  TKey *key = dynamic_cast < TKey * >((*fCurrentKey) ());

  // if it is empty, go to next file
  if (!key) {
    delete fCurrentKey;
    fCurrentKey = 0;
    fCurrentFile++;
    if (fCurrentFile == fFiles.end())
      return eEOF;
    fCurrentKey = new TIter((*fCurrentFile)->GetListOfKeys());
    key = dynamic_cast < TKey * >((*fCurrentKey) ());
  }

  if (!strstr(key->GetName(), "Hybrid"))
    return ReadNextHybrid(event);

  // read the object with that key
  return ReadObj(key, event);
};

AugerFile::Status AugerFile::Read(AugerEvent & event, AugerKey & key) {


  const char *name = key.AsString().c_str();

  FileContainer::iterator file = fFiles.begin();

  while (file != fFiles.end()) {

    // randomly access the key
    TKey *rootkey = (*file)->FindKey(name);
    // if not availble, try next file
    if (!rootkey) {
      file++;
      continue;
    }
    // read the object with that key
    return ReadObj(rootkey, event);
  }

  return eEOF;
};




AugerFile::Status AugerFile::Write(AugerEvent & event, bool verbose) {

  // always write in the first file of the list
  // If it does not exist, then cannot write
  if (!fFiles[0])
    return eFailure;

  TFile *file = fFiles[0];
  file->cd();

  // generate the key for that event
  AugerKey key(event);
  if (verbose)
    std::cout << "Writing event with key :" << key.
    AsString() << std::endl;
  event.Write(key.AsString().c_str());

  return eSuccess;

};

void AugerFile::Close() {

  // close all files
  for (FileContainer::iterator it = fFiles.begin();
       it != fFiles.end(); it++)
    (*it)->Close();

};
