/**
   \file
   Implementation of AugerKey
 
   \author Stefano Argir Antoine Letessier-Selvon
*/

#include <AugerKey.h>
#include <AugerEvent.h>
#include <IoSdData.h>
#include <sstream>


const
std::string AugerKey::fgStringTypes[fgNTypes] = { "Sd-only",
    "Fd-only",
    "Hybrid",
    "Laser",
    "FdLaser",
    "None"
                                                };

AugerKey::AugerKey():
  EventId(0),
  NumberOfStations(0),
  NumberOfEyes(0),
  NumberOfPixels(0), 
  EyesId(0), 
  SdTriggerAlgo(""), 
  Type(eNone) 
{};

AugerKey::AugerKey(AugerEvent & event) {

  EventId = event.Id();
  NumberOfStations =
    event.HasSd()? event.Sd().numberoferrorzerostation() : 0;
  NumberOfEyes = event.Eyes.size();

  if (event.HasSd() && NumberOfEyes) 
    Type = eHybrid;
  else if (event.HasSd() && !NumberOfEyes) 
    Type = eSdOnly;
  else if (!event.HasSd()) 
    Type = eFdOnly;
  if (event.HasSd() && strstr(event.Sd().Trigger.Algo.c_str(), "Laser")) 
    Type = eLaser;
  if (event.HasSd()) 
    SdTriggerAlgo = event.Sd().Trigger.Algo;
  
  EyesId = 0;
  NumberOfPixels = 0;
  AugerEvent::EyeIterator iter;
  for (iter = event.EyesBegin(); iter != event.EyesEnd(); ++iter) {
    EyesId |= (0x1 << (iter->GetEventHeader()->GetEyeNo() - 1));
    NumberOfPixels += iter->GetPixelList()->GetNumPixels();
  }
};


std::string AugerKey::AsString() {
  int nEyes = int ((EyesId & 0x10) >> 4) +
    int ((EyesId & 0x8) >> 3) + int ((EyesId & 0x4) >> 2) +
    int ((EyesId & 0x2) >> 1) + int (EyesId & 0x1);
  std::ostringstream tmp;
  tmp << EventId << "#"
      << NumberOfStations << "#"
      << NumberOfEyes << "#"
      << NumberOfPixels << "#" 
      << int ((EyesId & 0x10) >> 4)
      << int ((EyesId & 0x8) >> 3)
      << int ((EyesId & 0x4) >> 2)
      << int ((EyesId & 0x2) >> 1) 
      << int (EyesId & 0x1) <<
    "#" << fgStringTypes[Type];
  if (Type == eFdLaser) {
    switch (nEyes) {
    case 1:
      tmp << "-Mono";
      break;
    case 2:
      tmp << "-Stereo";
      break;
    case 3:
      tmp << "-Trio";
      break;
    case 4:
      tmp << "-Quadri";
      break;
    case 5:
      tmp << "-Penta";
      break;
    }
  }
  if (Type != eFdOnly && Type != eFdLaser)
    tmp << "#" << SdTriggerAlgo;

  return tmp.str();
};
