
//
// File   : EyeEvent.cc
//
// Purpose: Implementation of Eye event class
//
// $Id: EyeEvent.cc 11976 2011-01-28 15:10:38Z mathes $
//


#include <iostream>
#include <iomanip>
#include <cassert>

/** @file EyeEvent.cc
  * Implementation of the class TEyeEvent.
  * @author H.-J. Mathes, FzK
  */

#include <TBuffer.h>
#include <TClass.h>

/** Fixed in ROOT 5.19/04, see svn logs */
#if ROOT_VERSION_CODE < ROOT_VERSION(5,19,4)

# if ROOT_VERSION_CODE >= ROOT_VERSION(5,10,0)
/** Enable MapObject 'bug' work-around. */
#  define ENABLE_MAPOBJECT_WORKAROUND
#  warning ROOT MapObject 'bug' work-around is enabled!
# else
#  define _ENABLE_MAPOBJECT_WORKAROUND
# endif // ROOT_VERSION_CODE

// found another (Better?) work-around in class TEyeEventFile...
# undef ENABLE_MAPOBJECT_WORKAROUND

#endif // ROOT_VERSION_CODE

#include <FdNumbering.hh>
#include <FdGeometry.hh>
#include <ErrorLogging.hh>
#include <TimeConvert.hh>
#include <TTimeStamp.hh>

#include <MiEvent.hh>
#include <FadcData.hh>

#include "CheckVersion.hh"

#include "EyeGeometryData.hh"
#include "EyeTriggerData.hh"
#include "EyeT3Data.hh"

#include "EyeEvent.hh"

using namespace std;
using FdUtil::Fd;
using FdUtil::FdGeometry;
using FdUtil::TimeConvert;

// -------------------------------------------------------------------------
// -------------------------------------------------------------------------
//
// class TEyeEvent - implementation
//

unsigned int TEyeEvent::fgPrintLevel = 0;

TEyeEvent::TEyeEvent()
 : fStreamerVersion(0), 
   fEventHeader(NULL), fFADCData(NULL), fPixelList(NULL), fPixelData(NULL), 
   fTriggerData(NULL), fT3Data(NULL), fGeometryData(NULL)
 {
  if ( fgPrintLevel >= 2 )
    FD_CDEBUG(2) << "TEyeEvent::TEyeEvent()" << endl;
#ifdef TRACE_CONSTRUCTOR
  FD_CINFO << "TEyeEvent(): " << this << endl;
#endif // TRACE_CONSTRUCTOR
}

// -------------------------------------------------------------------------

TEyeEvent::TEyeEvent(const TEyeEvent& rhs)
 : TObject(rhs),
   fEventHeader(NULL), fFADCData(NULL), fPixelList(NULL), fPixelData(NULL), 
   fTriggerData(NULL), fT3Data(NULL), fGeometryData(NULL)
 {
  *this = rhs;
#ifdef TRACE_CONSTRUCTOR
  FD_CINFO << "TEyeEvent(TEyeEvent&): " << this << endl;
#endif // TRACE_CONSTRUCTOR
}

// -------------------------------------------------------------------------

TEyeEvent& TEyeEvent::operator=(const TEyeEvent& rhs)
 {
#ifdef TRACE_CONSTRUCTOR
  FD_CINFO << "TEyeEvent(=): " << this << endl;
#endif // TRACE_CONSTRUCTOR

  // --- check for self assignment
  
  if ( this == &rhs ) return *this;
  
  // --- copy contained objects (=> deep copy)
  
  if ( rhs.fEventHeader != NULL ) {
    TEyeEventHeader *header = GetEventHeader();
    *header = *rhs.fEventHeader;
  }
  else {
    delete fEventHeader;
    fEventHeader = NULL;
  }
  
  if ( rhs.fPixelData != NULL ) {
    TEyePixelData *pdata = GetPixelData();
    *pdata = *rhs.fPixelData;
  }
  else {
    delete fPixelData;
    fPixelData = NULL;
  }

  if ( rhs.fPixelList != NULL ) {
    TEyePixelList *plist = GetPixelList();
    *plist = *rhs.fPixelList;
  }
  else {
    delete fPixelList;
    fPixelList = NULL;
  }

  if ( rhs.fFADCData != NULL ) {
    TEyeFADCData *fadc = GetFADCData();
    *fadc = *rhs.fFADCData;
  }
  else {
    delete fFADCData;
    fFADCData = NULL;
  }

  if ( rhs.fTriggerData != NULL ) {
#if 0
    TEyeTriggerData * tdata = GetTriggerData();
    *tdata = *rhs.fTriggerData;
#else
    SetTriggerData( rhs.fTriggerData );
#endif
  }
  else {
    delete fTriggerData;
    fTriggerData = NULL;
  }
  
  if ( rhs.fT3Data != NULL ) {
    TEyeT3Data *t3_data = GetT3Data();
    *t3_data = *rhs.fT3Data;
  }
  else {
    delete fT3Data;
    fT3Data = NULL;
  }
  
  if ( rhs.fGeometryData != NULL ) {
    TEyeGeometryData *geo_data = GetGeometryData();
    *geo_data = *rhs.fGeometryData;
  }
  else {
    delete fGeometryData;
    fGeometryData = NULL;
  }
  
  // --- copy remaining data members ...
  
  // none so far !
  
  return *this;
}

// -------------------------------------------------------------------------

TEyeEvent::~TEyeEvent()
 {
  if ( fgPrintLevel >= 2 )
    FD_CDEBUG(2) << "TEyeEvent::~TEyeEvent()" << endl;

#ifdef TRACE_CONSTRUCTOR
  FD_CINFO << "~TEyeEvent(): " << this << endl;
#endif // TRACE_CONSTRUCTOR

  delete fPixelData;
  delete fPixelList;
  delete fFADCData;
  delete fEventHeader;
  
  delete fTriggerData;
  delete fT3Data;
  delete fGeometryData;
}

// -------------------------------------------------------------------------

bool TEyeEvent::AddEvent(TMirrorEvent *event,bool force)
 {
  // --- process TMirrorEventHeader ...
  
  TMirrorEventHeader* mi_header = event->GetEventHeader();

  // --- composite mirror number

  Fd::MirrorNumber mi_number = mi_header->GetMirrorNumber();

  if ( fgPrintLevel >= 1 )
    FD_CDEBUG(1) << "TEyeEvent::AddEvent() - "
                 << "adding data from eye= " << Fd::GetEyeNo( mi_number )
                 << ", mirror= " << Fd::GetMirrorNo( mi_number ) << endl;

  // --- Mirror and Eye number
  
  unsigned int eye_no = Fd::GetEyeNo( mi_number );
  unsigned int mirror_no = mi_header->GetMirrorNo();

  TEyeEventHeader *eye_header = this->GetEventHeader();
  
  // ... set eye number only from the first TMirrorEvent
  
  if ( eye_header->fMirrorDataBits == 0 )
    eye_header->SetEyeNo( eye_no );
  else {
    // paranoid check
    if ( eye_no != eye_header->GetEyeNo() ) {
      FD_CERR << "TEyeEvent::AddEvent() - "
              << "trying to add data from wrong eye " << eye_no << " !" << endl;
      return false;
    }
  }

  // --- check for duplicate telescope (if 'force' is not set)
  
  if ( eye_header->IsMirrorDataPresent( mirror_no ) ) {
    if ( force ) {  // issue warning and add it
      FD_CWARN << "TEyeEvent::AddEvent() - "
               << "adding data for telescope " << mirror_no << " again ..." 
	       << endl;
    }
    else {          // issue error message, don't add it and return
      FD_CERR << "TEyeEvent::AddEvent() - "
              << "cannot add data for telescope " << mirror_no 
	      << " a second time!" << endl;
      return false;
    }
  }
  
  // Note: We use SetMirrorPresent() in order to establish 
  //       some default values, but it should be set from outside !

  eye_header->SetMirrorPresent( mirror_no );
  eye_header->SetMirrorDataPresent( mirror_no );
  
  if ( fgPrintLevel ) {
    FD_CDEBUG(1) << "TEyeEvent::AddEvent() - "
         << "Setting TrgSrc & VETO/Dead time for mi_no= "
	 << mi_header->GetMirrorNo() << " ..." 
	 << endl;
    FD_CDEBUG(1) << setfill('0');
    FD_CDEBUG(1) << "                        "
         << "TrgSrc: 0x" 
	 << hex << setw(2) << mi_header->GetTriggerSource() << dec;
    FD_CDEBUG(1) << setfill(' ');
    FD_CDEBUG(1) << "  RunTime: " << setw(20) 
         << mi_header->GetRunTimeCounter() << endl;
    FD_CDEBUG(1) << " VetoTime: " << setw(20) 
         << mi_header->GetVetoTimeCounter() << endl;
    FD_CDEBUG(1) << "                                    "
	 << " DeadTime: " << setw(20) << mi_header->GetDeadTimeCounter()
	 << endl;
  }
  
  eye_header->SetTriggerSource( mi_header->GetTriggerSource(), mirror_no );
  
  eye_header->SetRunTimeCounter( mi_header->GetRunTimeCounter(), mirror_no );
  eye_header->SetDeadTimeCounter( mi_header->GetDeadTimeCounter(), mirror_no );
  eye_header->SetVetoTimeCounter( mi_header->GetVetoTimeCounter() );

  eye_header->SetTriggerTimeStamp( mi_header->GetTriggerTimeStamp(), mirror_no );

  eye_header->SetMirrorDaqDeadTime( mi_header->GetDaqDeadTime(), mirror_no );
  eye_header->SetMirrorNSkippedTriggers( mi_header->GetNumSkippedTriggers(), mirror_no );

  // --- map Mirror event type to Eye event type (reflects trigger type)
  
  if ( eye_header->GetEventType() == TEyeEventHeader::kUnknown ) {
  
    switch ( mi_header->GetEventType() ) {

      case TMirrorEventHeader::kUnknownEvent :
           eye_header->SetEventType( TEyeEventHeader::kUnknown );
	   break;

      case TMirrorEventHeader::kPhysicsEvent :
           eye_header->SetEventType( TEyeEventHeader::kPhysics );
	   break;

      case TMirrorEventHeader::kCalibrationEvent :
           eye_header->SetEventType( TEyeEventHeader::kCalibration );
	   break;

      case TMirrorEventHeader::kSimulatedEvent :
           eye_header->SetEventType( TEyeEventHeader::kSimulated );
	   break;

      default:  // all unhandled event types
           eye_header->SetEventType( TEyeEventHeader::kInvalidType );
    }

  }
    
  // --- map Mirror event label (TLT) to Eye event class (pre-T3)
  
  eye_header->SetMirrorEventLabel( mi_header->GetEventLabel(), mirror_no );  // new v1r5
  
  ConvertMirrorEventLabel( mi_header->GetEventLabel() );

  // --- set the time(s), TTimeStamp works with Unix time
  
  FdRoot::TTimeStamp time_stamp( FdRoot::TTimeStamp::FromGPS(
                                         mi_header->GetNextPageTime(), 
                                         mi_header->GetNextPageNanoTime()) );

//  FD_CDEBUG(1) << "time_stamp.GetSec= " << time_stamp.GetSec() << endl;
  
  if ( mi_header->IsOldTimingFormat() ) {
    
    if ( TEyeEvent::fgPrintLevel ) {
      FD_CWARN << "TMirrorEventHeader::Version() = "
               << mi_header->GetVersion() << " --> old timing format !!!" 
	       << endl;
    }
    
    eye_header->SetTimeCorrectionParameters( 100., 0. );
    eye_header->SetTimeCorrectionQuality( TEyeEventHeader::kFullQuality );
  }

#ifdef DEBUG_TIMING
  FD_CDEBUG(1) << "Mirror[" << mi_header->GetMirrorNo() << "] raw time= " 
               << time_stamp.GetGPSSec() 
	       << " nano= " << time_stamp.GetNanoSec()
	       << endl;
#endif
  
  eye_header->SetRawMirrorTime( mi_header->GetMirrorNo(), time_stamp );
  
  eye_header->AdjustMirrorTimes();
  
//  FD_CDEBUG(1) << "after AdjustMirrorTimes GPSsec= " 
//               << this->GetEventHeader()->GetDateTime()->GetGPSSec() << endl;
       
  // --- process TMirrorPixelData ...
  
  mirror_no = Fd::GetMirrorNo( mi_number ); // 1..6

  if ( fgPrintLevel ) {
    FD_CDEBUG(1) << "TEyeEvent::AddEvent() - "
  		 << "adding pixel data for Mirror " << mirror_no << " ..." 
        	 << endl;
  }
  
  TMirrorPixelData* mi_pixeldata = event->GetPixelData();
  
  // --- set the Mirror Id correctly in the data (old data or readout bug?) -> see note below!
  
  mi_pixeldata->SetMirrorNumber( mi_number );
  
  this->GetPixelData()->AddPixelData( mi_pixeldata, mirror_no, kFALSE );
  
  // --- correct the Mirrors pixel list, MirrorId not correctly set !?

  // Note: The MirrorId (fMirrorNumber) is not set because fMirrorNumber 
  //       is not streamed at all. Thus it must be recreated using some 
  //       other information...
  //       -> TEyeEventFile::ReadEvent() etc.
  //       -> static void TEyeEvent::DoCorrections(TEyeEvent*);

  eye_no = Fd::GetEyeNo( mi_number );
  mirror_no = Fd::GetMirrorNo( mi_number );
  
  TMirrorPixelList* pixel_list = event->GetPixelList();
  
  for ( unsigned int i=0; i<pixel_list->GetNumPixels(); i++) {
    
    Fd::PixelNumber pixel = pixel_list->GetPixel( i );
    // FD_CDEBUG(1) << "MirrorId: " << Fd::GetMirrorNo( pixel ) << endl;
    
    Fd::SetEyeNo( pixel, eye_no );
    Fd::SetMirrorNo( pixel, mirror_no );
  }
  
  // --- process TMirrorPixelList and TMirrorFADCData ...
  
  this->GetPixelList()->SetEyeNo( eye_no );
  
  this->GetPixelList()->AddPixelList( event->GetPixelList() );
  
  TEyeFADCData* eye_fadc = this->GetFADCData();
  TMirrorFADCData* fadc_data = event->GetFADCData();

  // --- perform loop over all pixels in the Mirror pixel list

  Fd::PixelNumber pixel;
  TFADCData* fadc;
  for ( unsigned int pixnum=0; pixnum<pixel_list->GetNumPixels(); pixnum++ ) {
    
    pixel = pixel_list->GetPixel( pixnum );
    
    fadc = fadc_data->GetPixelFADCData( pixel );
    
    // correct the pixel numbering information
    
    Fd::SetEyeNo( fadc->GetPixelNumber(), eye_no );
    Fd::SetMirrorNo( fadc->GetPixelNumber(), mirror_no );
    
    eye_fadc->AddFADCData( fadc );
    
    delete fadc;
  }
  
  // --- process the list of TFADCData to set their trace start times
  //     relative to the time of the event
  
#if 0
  TIterator *iter = eye_fadc->GetIterator();
  
  while ( ( fadc = (TFADCData*)iter->Next() ) != NULL ) {
    
    FD_COUT << Fd::GetMirrorNo( fadc->GetPixelNumber() ) 
            << ", " << Fd::GetEyePixelNo( fadc->GetPixelNumber() ) << endl;
  }
#endif

  // --- true to indicate 'the TMirrorEvent has been added' !
  
  return true;
}

// -------------------------------------------------------------------------

void TEyeEvent::Clear(Option_t *)
 {
  delete fEventHeader; fEventHeader = NULL;
  delete fFADCData; fFADCData = NULL;
  delete fPixelList; fPixelList = NULL;
  delete fPixelData; fPixelData = NULL;
  
  // optional fields ...
  
  delete fTriggerData; fTriggerData = NULL;
  delete fT3Data; fT3Data = NULL;
  delete fGeometryData; fGeometryData = NULL;
}

// -------------------------------------------------------------------------

typedef struct _LabelClassMapEntry {

  TEyeEventHeader::EEventClass       fInitialClass;
  TMirrorEventHeader::EMiEventLabel  fLabel;
  TEyeEventHeader::EEventClass       fFinalClass;

} LabelClassMapEntryRec, *LabelClassMapEntry;

LabelClassMapEntryRec gLabelClassMap[] = {

   { TEyeEventHeader::kUnClassified, TMirrorEventHeader::kUnlabelled , TEyeEventHeader::kUnClassified },
   { TEyeEventHeader::kUnClassified, TMirrorEventHeader::kAcceptedEvent, TEyeEventHeader::kUnhandledSignature },
   { TEyeEventHeader::kUnClassified, TMirrorEventHeader::kMuonEvent, TEyeEventHeader::kIsMuon },
   { TEyeEventHeader::kUnClassified, TMirrorEventHeader::kShowerCandidate, TEyeEventHeader::kShowerCandidate },
   { TEyeEventHeader::kUnClassified, TMirrorEventHeader::kShowerEvent, TEyeEventHeader::kIsShower },
   { TEyeEventHeader::kUnClassified, TMirrorEventHeader::kRandomEvent, TEyeEventHeader::kNoise },
   { TEyeEventHeader::kUnClassified, TMirrorEventHeader::kManyPixelEvent, TEyeEventHeader::kLargeEvent },
   { TEyeEventHeader::kUnClassified, TMirrorEventHeader::kRejected, TEyeEventHeader::kRejected },
   
   { TEyeEventHeader::kUnhandledSignature, TMirrorEventHeader::kUnlabelled, TEyeEventHeader::kUnhandledSignature },
   { TEyeEventHeader::kUnhandledSignature, TMirrorEventHeader::kAcceptedEvent, TEyeEventHeader::kUnhandledSignature },
   { TEyeEventHeader::kUnhandledSignature, TMirrorEventHeader::kMuonEvent, TEyeEventHeader::kIsMuon },
   { TEyeEventHeader::kUnhandledSignature, TMirrorEventHeader::kShowerCandidate, TEyeEventHeader::kShowerCandidate },
   { TEyeEventHeader::kUnhandledSignature, TMirrorEventHeader::kShowerEvent, TEyeEventHeader::kIsShower },
   { TEyeEventHeader::kUnhandledSignature, TMirrorEventHeader::kRandomEvent, TEyeEventHeader::kNoise },
   { TEyeEventHeader::kUnhandledSignature, TMirrorEventHeader::kManyPixelEvent, TEyeEventHeader::kLargeEvent },
   { TEyeEventHeader::kUnhandledSignature, TMirrorEventHeader::kRejected, TEyeEventHeader::kUnhandledSignature },
   
   { TEyeEventHeader::kIsMuon, TMirrorEventHeader::kUnlabelled, TEyeEventHeader::kIsMuon },
   { TEyeEventHeader::kIsMuon, TMirrorEventHeader::kAcceptedEvent, TEyeEventHeader::kIsMuon },
   { TEyeEventHeader::kIsMuon, TMirrorEventHeader::kMuonEvent, TEyeEventHeader::kIsMuon },
   { TEyeEventHeader::kIsMuon, TMirrorEventHeader::kShowerCandidate, TEyeEventHeader::kShowerCandidate },
   { TEyeEventHeader::kIsMuon, TMirrorEventHeader::kShowerEvent, TEyeEventHeader::kIsShower },
   { TEyeEventHeader::kIsMuon, TMirrorEventHeader::kRandomEvent, TEyeEventHeader::kMuonAndNoise },
   { TEyeEventHeader::kIsMuon, TMirrorEventHeader::kManyPixelEvent, TEyeEventHeader::kIsMuon },
   { TEyeEventHeader::kIsMuon, TMirrorEventHeader::kRejected, TEyeEventHeader::kIsMuon },
   
   { TEyeEventHeader::kMuonAndNoise, TMirrorEventHeader::kUnlabelled, TEyeEventHeader::kMuonAndNoise },
   { TEyeEventHeader::kMuonAndNoise, TMirrorEventHeader::kAcceptedEvent, TEyeEventHeader::kMuonAndNoise },
   { TEyeEventHeader::kMuonAndNoise, TMirrorEventHeader::kMuonEvent, TEyeEventHeader::kMuonAndNoise },
   { TEyeEventHeader::kMuonAndNoise, TMirrorEventHeader::kShowerCandidate, TEyeEventHeader::kShowerCandidate },
   { TEyeEventHeader::kMuonAndNoise, TMirrorEventHeader::kShowerEvent, TEyeEventHeader::kIsShower },
   { TEyeEventHeader::kMuonAndNoise, TMirrorEventHeader::kRandomEvent, TEyeEventHeader::kMuonAndNoise },
   { TEyeEventHeader::kMuonAndNoise, TMirrorEventHeader::kManyPixelEvent, TEyeEventHeader::kMuonAndNoise },
   { TEyeEventHeader::kMuonAndNoise, TMirrorEventHeader::kRejected, TEyeEventHeader::kMuonAndNoise },
   
   { TEyeEventHeader::kShowerCandidate, TMirrorEventHeader::kUnlabelled, TEyeEventHeader::kShowerCandidate },
   { TEyeEventHeader::kShowerCandidate, TMirrorEventHeader::kAcceptedEvent, TEyeEventHeader::kShowerCandidate },
   { TEyeEventHeader::kShowerCandidate, TMirrorEventHeader::kMuonEvent, TEyeEventHeader::kShowerCandidate },
   { TEyeEventHeader::kShowerCandidate, TMirrorEventHeader::kShowerCandidate, TEyeEventHeader::kShowerCandidate },
   { TEyeEventHeader::kShowerCandidate, TMirrorEventHeader::kShowerEvent, TEyeEventHeader::kIsShower },
   { TEyeEventHeader::kShowerCandidate, TMirrorEventHeader::kRandomEvent, TEyeEventHeader::kShowerCandidate },
   { TEyeEventHeader::kShowerCandidate, TMirrorEventHeader::kManyPixelEvent, TEyeEventHeader::kShowerCandidate },
   { TEyeEventHeader::kShowerCandidate, TMirrorEventHeader::kRejected, TEyeEventHeader::kShowerCandidate },
   
   { TEyeEventHeader::kIsShower, TMirrorEventHeader::kUnlabelled, TEyeEventHeader::kIsShower },
   { TEyeEventHeader::kIsShower, TMirrorEventHeader::kAcceptedEvent, TEyeEventHeader::kIsShower },
   { TEyeEventHeader::kIsShower, TMirrorEventHeader::kMuonEvent, TEyeEventHeader::kIsShower },
   { TEyeEventHeader::kIsShower, TMirrorEventHeader::kShowerCandidate, TEyeEventHeader::kIsShower },
   { TEyeEventHeader::kIsShower, TMirrorEventHeader::kShowerEvent, TEyeEventHeader::kIsShower },
   { TEyeEventHeader::kIsShower, TMirrorEventHeader::kRandomEvent, TEyeEventHeader::kIsShower },
   { TEyeEventHeader::kIsShower, TMirrorEventHeader::kManyPixelEvent, TEyeEventHeader::kIsShower },
   { TEyeEventHeader::kIsShower, TMirrorEventHeader::kRejected, TEyeEventHeader::kIsShower },
   
   { TEyeEventHeader::kNoise, TMirrorEventHeader::kUnlabelled, TEyeEventHeader::kNoise },
   { TEyeEventHeader::kNoise, TMirrorEventHeader::kAcceptedEvent, TEyeEventHeader::kUnhandledSignature },
   { TEyeEventHeader::kNoise, TMirrorEventHeader::kMuonEvent, TEyeEventHeader::kIsMuon },
   { TEyeEventHeader::kNoise, TMirrorEventHeader::kShowerCandidate, TEyeEventHeader::kShowerCandidate },
   { TEyeEventHeader::kNoise, TMirrorEventHeader::kShowerEvent, TEyeEventHeader::kIsShower },
   { TEyeEventHeader::kNoise, TMirrorEventHeader::kRandomEvent, TEyeEventHeader::kNoise },
   { TEyeEventHeader::kNoise, TMirrorEventHeader::kManyPixelEvent, TEyeEventHeader::kLargeEvent },
   { TEyeEventHeader::kNoise, TMirrorEventHeader::kRejected, TEyeEventHeader::kNoise },
   
   { TEyeEventHeader::kLargeEvent, TMirrorEventHeader::kUnlabelled, TEyeEventHeader::kLargeEvent },
   { TEyeEventHeader::kLargeEvent, TMirrorEventHeader::kAcceptedEvent, TEyeEventHeader::kLargeEvent },
   { TEyeEventHeader::kLargeEvent, TMirrorEventHeader::kMuonEvent, TEyeEventHeader::kLargeEvent },
   { TEyeEventHeader::kLargeEvent, TMirrorEventHeader::kShowerCandidate, TEyeEventHeader::kShowerCandidate },
   { TEyeEventHeader::kLargeEvent, TMirrorEventHeader::kShowerEvent, TEyeEventHeader::kIsShower },
   { TEyeEventHeader::kLargeEvent, TMirrorEventHeader::kRandomEvent, TEyeEventHeader::kLargeEvent },
   { TEyeEventHeader::kLargeEvent, TMirrorEventHeader::kManyPixelEvent, TEyeEventHeader::kLargeEvent },
   { TEyeEventHeader::kLargeEvent, TMirrorEventHeader::kRejected, TEyeEventHeader::kLargeEvent },
   
   { TEyeEventHeader::kRejected, TMirrorEventHeader::kUnlabelled, TEyeEventHeader::kUnClassified },
   { TEyeEventHeader::kRejected, TMirrorEventHeader::kAcceptedEvent, TEyeEventHeader::kUnhandledSignature },
   { TEyeEventHeader::kRejected, TMirrorEventHeader::kMuonEvent, TEyeEventHeader::kIsMuon },
   { TEyeEventHeader::kRejected, TMirrorEventHeader::kShowerCandidate, TEyeEventHeader::kShowerCandidate },
   { TEyeEventHeader::kRejected, TMirrorEventHeader::kShowerEvent, TEyeEventHeader::kIsShower },
   { TEyeEventHeader::kRejected, TMirrorEventHeader::kRandomEvent, TEyeEventHeader::kNoise },
   { TEyeEventHeader::kRejected, TMirrorEventHeader::kManyPixelEvent, TEyeEventHeader::kLargeEvent },
   { TEyeEventHeader::kRejected, TMirrorEventHeader::kRejected, TEyeEventHeader::kRejected },
};
 
void TEyeEvent::ConvertMirrorEventLabel(TMirrorEventHeader::EMiEventLabel label)
 {
  TEyeEventHeader *eye_header = this->GetEventHeader();

  unsigned int map_size = sizeof(gLabelClassMap)/sizeof(LabelClassMapEntryRec);
  
  bool found = false;
  
  for ( unsigned int i=0; i<map_size; i++ ) {
  
    if ( gLabelClassMap[i].fInitialClass == eye_header->GetEventClass() ) {
    
      if ( gLabelClassMap[i].fLabel == label ) {

	TEyeEventHeader::EEventClass eclass = gLabelClassMap[i].fFinalClass;
#if 0
        FD_COUT << "Initial EyeEvent class= " 
             << TEyeEventHeader::GetVerboseEventClass( eye_header->GetEventClass() )
             << ", MiEventLabel= " 
             << TMirrorEventHeader::GetVerboseEventLabel( label )
             << ", final class= " 
             << TEyeEventHeader::GetVerboseEventClass( eclass ) 
             << endl;
#endif          
        eye_header->SetEventClass( eclass );
	found = true;
        break;
      }
    }
  }
  
  if ( !found ) {
    FD_CERR << "Initial EyeEvent class= " 
	 << TEyeEventHeader::GetVerboseEventClass( eye_header->GetEventClass() )
	 << ", MiEventLabel= " 
	 << TMirrorEventHeader::GetVerboseEventLabel( label )
	 << " --> final class= ???" 
	 << endl;
  }
}

// -------------------------------------------------------------------------

void TEyeEvent::DoCorrections(TEyeEvent * event)
 {
  assert( event != NULL );
  
  TEyePixelData * pixel_data = event->GetPixelData();
  TEyeEventHeader * header = event->GetEventHeader();
 
  // --- set eye and mirror id in the TMirrorPixelData (since v2r1)

  // see also: note in TEyeEvent::AddEvent() and file CHANGES
 
  TMirrorPixelData *mi_pixel_data;
    
  for ( unsigned int m=Fd::kEYE_FIRST_MIRROR;
  		     m<=Fd::kEYE_LAST_MIRROR; m++ ) {
    
    if ( (mi_pixel_data = pixel_data->GetPixelData( m )) == NULL ) continue;
    
    if (    mi_pixel_data->GetEyeNo() != header->GetEyeNo()
         || mi_pixel_data->GetMirrorNo() != m ) {
      
      if ( TEyeEvent::fgPrintLevel )
        FD_CWARN << "TEyeEvent::DoCorrections() - "
                 << "corrected TMirrorPixelData::(eye_id,mirror_id) from "
	         << "(" << mi_pixel_data->GetEyeNo() << "," 
	         << mi_pixel_data->GetMirrorNo() << ") to (" 
	         << header->GetEyeNo() << "," << m << ")!" << endl;
      
      mi_pixel_data->SetEyeNo( header->GetEyeNo() );
      mi_pixel_data->SetMirrorNo( m );
    }
    
    if ( TEyeEvent::fgPrintLevel > 1 )
      mi_pixel_data->Show();
  }

  // --- set TEyeGeometryData for old data to reasonable default values
  
  if ( event->fGeometryData == NULL && header->GetEyeNo() != 5 ) {
    
    double az, el;
    
    TEyeGeometryData *geom = event->GetGeometryData();
    
    for ( unsigned int mirror_id = Fd::kEYE_FIRST_MIRROR; 
                       mirror_id <= Fd::kEYE_LAST_MIRROR; mirror_id++ ) {
      
      FdGeometry::GetTelescopeAxis( 
              Fd::GetFdMirrorNo( header->GetEyeNo(), mirror_id ), az, el );
    
      geom->SetAxisDirection( mirror_id, az, el );
    }
  }
}

// -------------------------------------------------------------------------

TEyeFADCData* TEyeEvent::GetFADCData()
 {
  if ( !fFADCData ) fFADCData = new TEyeFADCData();
  return fFADCData;
}

// -------------------------------------------------------------------------

TEyeEventHeader * TEyeEvent::GetEventHeader()
 {
  if ( !fEventHeader ) fEventHeader = new TEyeEventHeader();
  return fEventHeader;
}

// -------------------------------------------------------------------------

TEyeGeometryData * TEyeEvent::GetGeometryData()
 { 
  if ( !fGeometryData )
    fGeometryData = new TEyeGeometryData(); 

  return fGeometryData; 
}

// -------------------------------------------------------------------------

TMirrorFADCData* TEyeEvent::GetMirrorFADCData(UInt_t mirror_id)
 {
  if ( mirror_id < Fd::kEYE_FIRST_MIRROR || mirror_id > Fd::kEYE_LAST_MIRROR ) {
    FD_CERR << "TEyeEvent::GetMirrorFADCData(" << mirror_id 
            << "): illegal mirror id passed !" << endl;
    return NULL;
  }
  if ( !GetEventHeader()->IsMirrorDataPresent( mirror_id ) ) {
    // requested Mirrors data not present in this event
    return NULL;
  }
  
  FD_CSEV << "TEyeEvent::GetMirrorFADCData() not yet implemented !" << endl;
  
  return NULL;
}

// -------------------------------------------------------------------------

TEyePixelData* TEyeEvent::GetPixelData()
 {
  static bool warning_printed = false;
  
  if ( !fPixelData ) fPixelData = new TEyePixelData();

  // temporary workaround for TMirrorPixelData v1 .. v2
  
  if (    TMirrorPixelData::Class_Version() == MiEVENTVERSIONv1
       || TMirrorPixelData::Class_Version() == MiEVENTVERSIONv2 ) {
    
    unsigned int eye_no = this->GetEventHeader()->GetEyeNo();
    
    for ( unsigned int id=0; id<Fd::kEYE_NMIRRORS; id++ ) {
      
      unsigned int mirror_no = id + Fd::kEYE_FIRST_MIRROR;
      
      if ( fPixelData->GetPixelData( mirror_no ) != NULL ) {
        fPixelData->GetPixelData( mirror_no )->SetEyeNo( eye_no );
        fPixelData->GetPixelData( mirror_no )->SetMirrorNo( mirror_no );
      }
    }
    
    if ( fgPrintLevel && !warning_printed ) {
      FD_CWARN << "TEyeEvent::GetPixelData() - "
           << "using workaround for TMirrorPixelData::fMirrorNo !"
	   << endl;
	   
      warning_printed = true;
    }
  }

  return fPixelData;
}

// -------------------------------------------------------------------------

TEyePixelList* TEyeEvent::GetPixelList()
 {
  static int warning_printed = false;
 
  if ( !fPixelList ) fPixelList = new TEyePixelList();
  
  // workaround for TEyePixelList v1 .. v3 (they didn't have fEyeNo field
  // workaround for v4, fEyeNo inproperly initialised/filled
  
  if (    fPixelList->fStreamerVersion == EyeEVENTVERSIONv1
       || fPixelList->fStreamerVersion == EyeEVENTVERSIONv2
       || fPixelList->fStreamerVersion == EyeEVENTVERSIONv3
       || fPixelList->fStreamerVersion == EyeEVENTVERSIONv4 ) {
    
    unsigned int eye_no = this->GetEventHeader()->GetEyeNo();
    
    if ( fgPrintLevel && !warning_printed ) {
      FD_CWARN << "TEyeEvent::GetPixelList() - "
           << "using workaround for TEyePixelList::fEyeNo," << endl
	   << "                          - "
	   << "setting fEyeNo= " << eye_no << " from event header !" 
	   << endl;
	   
      warning_printed = true;
    }
     
    fPixelList->SetEyeNo( eye_no );
  }
  
  return fPixelList;
}

// -------------------------------------------------------------------------

TEyeTriggerData * TEyeEvent::GetTriggerData()
 {
  if ( !fTriggerData )
    fTriggerData = new TEyeTriggerData(); 

  return fTriggerData;
}

// -------------------------------------------------------------------------

TEyeT3Data * TEyeEvent::GetT3Data()
 { 
  if ( !fT3Data )
    fT3Data = new TEyeT3Data(); 

  return fT3Data; 
}

// -------------------------------------------------------------------------

bool TEyeEvent::RemoveEvent(unsigned int mirror_id,TMirrorEvent** evptr)
 {
  if ( mirror_id < Fd::kEYE_FIRST_MIRROR || mirror_id > Fd::kEYE_LAST_MIRROR ) {
    FD_CERR << "TEyeEvent::RemoveEvent() - mirror_id " << mirror_id
            << " is invalid!" << endl;
    return false;
  }
  
  if ( evptr == NULL ) {
    FD_CERR << "TEyeEvent::RemoveEvent() - 'evptr' is not specified!" << endl;
    return false;
  }
  
  TMirrorEvent *event = *evptr;
  
  if ( *evptr == NULL ) {
    FD_CWARN << "TEyeEvent::RemoveEvent() - 'evptr' is NULL, "
             << "TMirrorEvent will be created!" << endl;
    event = new TMirrorEvent();
  }
  
  FD_CWARN << "TEyeEvent::RemoveEvent() - is incompletely implemented!"
           << endl;

  TEyeEventHeader *eye_header = this->GetEventHeader();
  
  // --- process TEyeEventHeader & fill TMirrorEventHeader ...
  
  TMirrorEventHeader* mi_header = event->GetEventHeader();

  mi_header->SetEyeNo( eye_header->GetEyeNo() );
  mi_header->SetMirrorNo( mirror_id );

  mi_header->SetRunNo( eye_header->GetRunNo() );
  mi_header->SetEventNo( eye_header->GetEventNo() );
  mi_header->SetTriggerNo( eye_header->GetEventNo() );  // fake !
  mi_header->SetTriggerSource( eye_header->GetTriggerSource(mirror_id) );
  mi_header->SetT3Id( eye_header->GetT3Id() );
  
  const FdRoot::TTimeStamp *tstamp = eye_header->GetRawMirrorTimeStamp(mirror_id);
  
  mi_header->SetNextPageTime( TimeConvert::GetGPSTimeFromUnixTime( tstamp->GetSec() ) );
  mi_header->SetNextPageNanoTime( tstamp->GetNanoSec() );
  mi_header->SetTriggerTimeStamp( eye_header->GetTriggerTimeStamp(mirror_id) );
  mi_header->SetNextPageDelay( 700 ); // eventually wrong !
  
  float t10, offset;
  
  eye_header->GetTimeCorrectionParameters( &t10, &offset );
  mi_header->SetTimeCorrectionParameters( t10, offset );
  // this information is lost !?
  //mi_header->SetTimeStampHigh();
  //mi_header->SetTimeStampLow();
  
  mi_header->SetDaqDeadTime( eye_header->GetMirrorDaqDeadTime(mirror_id) );
  mi_header->SetRunTimeCounter( eye_header->GetRunTimeCounter(mirror_id) );
  mi_header->SetDeadTimeCounter( eye_header->GetDeadTimeCounter(mirror_id) );
  mi_header->SetVetoTimeCounter( eye_header->GetVetoTimeCounter() );
  mi_header->SetNumSkippedTriggers( eye_header->GetMirrorNSkippedTriggers(mirror_id) );
  
  // we need another conversion table
  mi_header->SetEventType( TEyeEventHeader::GetMirrorEventType(eye_header->GetEventType()) );
  mi_header->SetEventLabel( (TMirrorEventHeader::EMiEventLabel)eye_header->GetMirrorEventLabel(mirror_id) );
  
  // and erase the data finally
  
  unsigned int gps_zero = TimeConvert::GetUnixTimeFromGPSTime( 0 );
  
  FdRoot::TTimeStamp gps_zero_time_stamp((time_t)gps_zero, 0);

  eye_header->fRunTimeCounter[mirror_id-1] = 0LL;
  eye_header->fDeadTimeCounter[mirror_id-1] = 0LL;
  eye_header->fMirrorDataBits &= ~(1<<(mirror_id-1));
  eye_header->fMirrorTimeOffset[mirror_id-1] = 0;
  eye_header->fMirrorDaqDeadTime[mirror_id-1] = 0LL;
  eye_header->fMirrorNSkippedTriggers[mirror_id-1] = 0;
  eye_header->fMirrorEventLabel[mirror_id-1] = TMirrorEventHeader::kUnlabelled;
  eye_header->fRawMirrorTime[mirror_id-1] = gps_zero_time_stamp;
  eye_header->fTriggerSource[mirror_id-1] = (UChar_t)TMirrorEventHeader::kUnknownTrigger;
  eye_header->fTriggerTimeStamp[mirror_id-1] = 0;

  eye_header->AdjustMirrorTimes();
  
  // --- process TEyePixelData and fill TMirrorPixelData ...

  TEyePixelData *eye_pixel_data = this->GetPixelData();
  TMirrorPixelData *mi_pixel_data = event->GetPixelData();
  
  *mi_pixel_data = *(eye_pixel_data->GetPixelData( mirror_id ));
  
  eye_pixel_data->AddPixelData( NULL, mirror_id );
  
  // --- process TEyePixelList and fill TMirrorPixelList ...

  TEyePixelList *eye_pixel_list = this->GetPixelList();
  
  TMirrorPixelList *mi_pixel_list = event->GetPixelList();
  
  FdUtil::Fd::FdPixelNumber fdpixel;
  
  for ( unsigned int npix=0; npix<=eye_pixel_list->GetNumPixels(); npix++ ) {
  
    fdpixel = eye_pixel_list->GetPixel( npix );
    
    if ( Fd::GetMirrorNo( fdpixel ) == mirror_id ) {
      mi_pixel_list->AddPixel( Fd::GetPixelNumber( fdpixel ) );
    }
  }
  
  // --- process TEyeFADCData and fill TMirrorFADCData
  
  TEyeFADCData *eye_fadc_data = this->GetFADCData();
  
  TMirrorFADCData *mi_fadc_data = event->GetFADCData();
  
  for ( unsigned int npix=0; npix<=mi_pixel_list->GetNumPixels(); npix++ ) {
  
    Fd::PixelNumber pixel = mi_pixel_list->GetPixel( npix );
    
    //fdpixel = Fd::GetFdPixelNumber( pixel );
    //const TFADCData * adc = eye_fadc_data->GetFADCData( fdpixel );
    const TFADCData * adc = eye_fadc_data->GetFADCData( pixel );
    
    mi_fadc_data->SetPixelFADCData( pixel, adc );
    
    eye_fadc_data->RemoveFADCData( adc );
  }
  
  // --- another (final) loop to remove the pixels from TEyePixelList
  
  for ( unsigned int npix=0; npix<=mi_pixel_list->GetNumPixels(); npix++ ) {
  
    Fd::PixelNumber pixel = mi_pixel_list->GetPixel( npix );
    
    eye_pixel_list->RemovePixel( Fd::GetFdPixelNumber( pixel ) );
  }
  
  return false;
}

// -------------------------------------------------------------------------

// Reset() = Clear() ???
void TEyeEvent::Reset()
 {
  delete fEventHeader;
  delete fFADCData;
  delete fPixelList;
  delete fPixelData;
  delete fTriggerData;
  delete fT3Data;
  delete fGeometryData;
  
  fEventHeader  = NULL;
  fFADCData     = NULL;
  fPixelList    = NULL;
  fPixelData    = NULL;
  fTriggerData  = NULL;
  fT3Data       = NULL;
  fGeometryData = NULL;
}

// -------------------------------------------------------------------------

void TEyeEvent::Show(std::ostream& ostr)
 {
  if ( fgPrintLevel >= 2 )
    ostr << "TEyeEvent::Show()" << endl;

  ostr << "--- TEyeEvent for Eye " 
       << GetEventHeader()->GetEyeNo() << " ---" << endl;

  ostr << " Streamer version: " << fStreamerVersion << endl;
  ostr << " Class version   : " << Class_Version() << endl;
  ostr << endl;
  
  ostr << "The following (optional) sub-objects are present:" << endl;
  if ( fPixelData )
    ostr << "\tTEyePixelData ..." << endl;
  if ( fTriggerData )
    ostr << "\tTEyeTriggerData ..." << endl;
  if ( fT3Data )
    ostr << "\tTEyeT3Data ..." << endl;
  if ( fGeometryData )
    ostr << "\tTEyeGeometryData ..." << endl;
  ostr << endl;
  
  unsigned int eye_no = GetEventHeader()->GetEyeNo();
  
  GetEventHeader()->Show( ostr );

  GetPixelData()->Show( ostr );

  GetPixelList()->Show( ostr );
  
  GetFADCData()->Show( ostr );
  
  if ( fTriggerData != NULL ) fTriggerData->Show( ostr );
  if ( fT3Data != NULL ) fT3Data->Show( ostr );
  if ( fGeometryData != NULL ) fGeometryData->Show( ostr, eye_no  );
}

// -------------------------------------------------------------------------

void TEyeEvent::SetGeometryData(TEyeGeometryData* data)
 {
  if ( fGeometryData != NULL ) delete fGeometryData;
  if ( data != NULL )
    fGeometryData = new TEyeGeometryData( *data );
  else
    fGeometryData = NULL;
}

// -------------------------------------------------------------------------

void TEyeEvent::SetTriggerData(TEyeTriggerData *data)
 {
  if ( fTriggerData != NULL ) delete fTriggerData;
  if ( data != NULL )
    fTriggerData = new TEyeTriggerData( *data );
  else
    fTriggerData = NULL;
}

// -------------------------------------------------------------------------

void TEyeEvent::SetT3Data(TEyeT3Data *data)
 {
  if ( fT3Data != NULL ) delete fT3Data;
  if ( data != NULL )
    fT3Data = new TEyeT3Data( *data );
  else
    fT3Data = NULL;
}

// -------------------------------------------------------------------------

void TEyeEvent::Streamer(TBuffer &R__b)
 {
  // Stream an object of class TEyeEventHeader.

  unsigned int R__s, R__c;
  
  if ( R__b.IsReading() ) {
     
    Version_t R__v = fStreamerVersion = R__b.ReadVersion(&R__s, &R__c);

    if ( TEyeEvent::fgPrintLevel ) {
      FD_CDEBUG(1) << "TEyeEvent::Streamer() - R__v= " << R__v
                   << " R__s= " << R__s << " R__c= " << R__c << endl;
    }
    
    if ( R__v <= EyeEVENTVERSIONv5 ) {
    
      TEyeEvent::Class()->ReadBuffer(R__b, this, R__v, R__s, R__c);
      
      // do some corrections to the data just read:
      // - update missing TEyeGeometryData
      // - fill TMirrorPixelData::fMirrorNumber
      
      TEyeEvent::DoCorrections( this );
    }
    else {
      FD_CWARN << "TEyeEvent::Streamer() - no corrections applied" << endl;
      
      TEyeEvent::Class()->ReadBuffer(R__b, this, R__v, R__s, R__c);
    }
  }
  else { // support new Streamer scheme of ROOT v 3.x
    
    TEyeEvent::Class()->WriteBuffer(R__b, this);

    // see http://root.cern.ch/root/Version516.news.html
    //R__b.WriteClassBuffer(TEyeEvent::Class(),this);
  }
}

// -------------------------------------------------------------------------
// work-around for a presumed bug on the way to TBuffer::MapObject()
// -------------------------------------------------------------------------

#if ROOT_VERSION_CODE >= ROOT_VERSION(5,15,4)
# include <TBufferFile.h>
# include <TExMap.h>
# include <TString.h>    // TString::Hash()
#else
# include <TExMap.h>
# include <TMath.h>      // TMath::Hash()
#endif // ROOT_VERSION_CODE

// -------------------------------------------------------------------------

#if ROOT_VERSION_CODE >= ROOT_VERSION(5,15,4)

static inline ULong_t Void_Hash(const void *ptr)
{
   // Return hash value for this object.

   return TString::Hash(&ptr, sizeof(void*));
}

// -------------------------------------------------------------------------

void TBufferFile::MapObject(const TObject *obj, UInt_t offset)
{
   // Add object to the fMap container.
   // If obj is not 0 add object to the map (in read mode also add 0 objects to
   // the map). This method may only be called outside this class just before
   // calling obj->Streamer() to prevent self reference of obj, in case obj
   // contains (via via) a pointer to itself. In that case offset must be 1
   // (default value for offset).

   if (IsWriting()) {
      if (!fMap) InitMap();

      if (obj) {
         CheckCount(offset);
         ULong_t hash = Void_Hash(obj);
         fMap->Add(hash, (Long_t)obj, offset);
         // No need to keep track of the class in write mode
         // fClassMap->Add(hash, (Long_t)obj, (Long_t)((TObject*)obj)->IsA());
         fMapCount++;
      }
   } else {
// THJM >>>
//      cout << "TBufferFile::MapObject(TObject*,UInt_t) - "
//           << hex << obj << dec 
//	   << "," << offset << endl;
// <<< THJM
      if (!fMap || !fClassMap) InitMap();

      fMap->Add(offset, (Long_t)obj);
// THJM >>>
#ifndef ENABLE_MAPOBJECT_WORKAROUND
// original code
      fClassMap->Add(offset,
             (obj && obj != (TObject*)-1) ? (Long_t)((TObject*)obj)->IsA() : 0);
#else
# warning Workaround in TBufferFile::MapObject() is active !
      fClassMap->Add(offset,
             (obj && obj != (TObject*)-1 && ((((int)obj) % sizeof(void*)) == 0)) ? (Long_t)((TObject*)obj)->IsA() : 0);
#endif // ENABLE_MAPOBJECT_WORKAROUND
// <<< THJM
      fMapCount++;
   }
}

// -------------------------------------------------------------------------

#else // ROOT_VERSION_CODE

static inline ULong_t Void_Hash(const void *ptr)
{
   // Return hash value for this object.

   return TMath::Hash(&ptr, sizeof(void*));
}

// -------------------------------------------------------------------------

void TBuffer::MapObject(const TObject *obj, UInt_t offset)
{
   // Add object to the fMap container.
   // If obj is not 0 add object to the map (in read mode also add 0 objects to
   // the map). This method may only be called outside this class just before
   // calling obj->Streamer() to prevent self reference of obj, in case obj
   // contains (via via) a pointer to itself. In that case offset must be 1
   // (default value for offset).

   if (IsWriting()) {
      if (!fMap) InitMap();

      if (obj) {
         CheckCount(offset);
         ULong_t hash = Void_Hash(obj);
         fMap->Add(hash, (Long_t)obj, offset);
         // No need to keep track of the class in write mode
         // fClassMap->Add(hash, (Long_t)obj, (Long_t)((TObject*)obj)->IsA());
         fMapCount++;
      }
   } else {
// THJM >>>
//      cout << "TBuffer::MapObject(TObject*,UInt_t) - "
//           << hex << obj << dec 
//	   << "," << offset << endl;
// <<< THJM
      if (!fMap || !fClassMap) InitMap();

      fMap->Add(offset, (Long_t)obj);
// THJM >>>
#ifndef ENABLE_MAPOBJECT_WORKAROUND
// original code
      fClassMap->Add(offset,
             (obj && obj != (TObject*)-1) ? (Long_t)((TObject*)obj)->IsA() : 0);
#else
# warning Workaround in TBuffer::MapObject() is active !
      fClassMap->Add(offset,
             (obj && obj != (TObject*)-1 && ((((int)obj) % sizeof(void*)) == 0)) ? (Long_t)((TObject*)obj)->IsA() : 0);
#endif // ENABLE_MAPOBJECT_WORKAROUND
// <<< THJM
      fMapCount++;
   }
}

#endif // ROOT_VERSION_CODE

ClassImp(TEyeEvent)

// -------------------------------------------------------------------------
// -------------------------------------------------------------------------
