
//
// File   : EyeEventFile.cc
//
// Purpose: Declaration file for class TEyeEventFile
//
// $Id: EyeEventFile.cc 8757 2010-09-27 12:16:09Z mathes $
//


#include <iostream>
#include <iomanip>
#include <cassert>
#include <algorithm>
#include <list>
#include <cerrno>

#include <unistd.h>	// getpid(), kill()

#ifdef HAVE_CONFIG_H
// proposed by L.Nellen for OS X 10.5:
# include "../../config.h"
# ifdef HAVE_SIGNAL_H
#  include <signal.h>  // signal definitions
# endif // HAVE_SIGNAL_H
#else // HAVE_CONFIG_H
# define HAVE_SIGNAL_H
# include <signal.h>   // SCons: just include it
#endif // HAVE_CONFIG_H


/** @file EyeEventFile.cc
  * Declaration file for class TEyeEventFile.
  * @author H.-J. Mathes, FzK
  */

#include <TBranch.h>
#include <TTree.h>
#include <TStreamerInfo.h>
#include <TSysEvtHandler.h>
#include <TSystem.h>
#include <TUnixSystem.h>

#include <MiEvent.hh>
#include <MiEventFile.hh>
#include <ErrorLogging.hh>

#include "CheckVersion.hh"

#include "EyeEvent.hh"
#include "EyeRunHeader.hh"
#include "EyeRunTrailer.hh"

#include "EyeEventFile.hh"

using namespace std;

// -------------------------------------------------------------------------
// -------------------------------------------------------------------------
//
// class TEyeEventFileSignalHandler - declaration & implementation
//

class TEyeEventFileSignalHandler : public TSignalHandler {
public:
    /** Constructor for the class TEyeEventSignalHandler.
      *
      * This type of signal handler is always async.
      */
    TEyeEventFileSignalHandler(ESignals sig)
     : TSignalHandler(sig, kFALSE) { }

    /**  */
    Bool_t Notify();
};

Bool_t TEyeEventFileSignalHandler::Notify()
 {
  FD_CWARN << "TEyeEventFileSignalHandler::Notify() ..." << endl;
  
  TEyeEventFile::ExitHandler();
  
  // switch back to the Unix default behaviour ...
  
  switch ( this->GetSignal() ) {
    
    case kSigQuit:
         gSystem->ResetSignal( kSigQuit );
	 kill( getpid(), SIGQUIT );
	 break;

    case kSigInterrupt:
         gSystem->ResetSignal( kSigInterrupt );
	 kill( getpid(), SIGINT );
	 break;

    case kSigTermination:
         gSystem->ResetSignal( kSigTermination );
	 kill( getpid(), SIGTERM );
	 break;

    default:
         break;
  }
  
  if ( TEyeEventFile::fgPrintLevel > 1 )
    FD_CINFO << "TEyeEventFileSignalHandler::Notify() - done !" << endl;

  return kFALSE;
}

// -------------------------------------------------------------------------
// -------------------------------------------------------------------------
//
// class TEyeEventFile - implementation
//

ClassImp(TEyeEventFile)

static const char *gEyeTreeName = "EyeEventTree";
static const char *gEyeBranchName = "EyeEvent";

int TEyeEventFile::fgPrintLevel = 0;
list<TEyeEventFileSignalHandler*> TEyeEventFile::fgSignalHandlerList;
#ifndef DPA
struct sigaction TEyeEventFile::fgOldSigSegvAction;
#endif // DPA

list<TEyeEventFile*> TEyeEventFile::fgFileList;

// -------------------------------------------------------------------------

TEyeEventFile::TEyeEventFile()
 : fBranch(0), fCurrentEntry(0), fPartialEvent(0), fEventBuffer(0), fTree(0),
   fEventHeaderBranch(0) /*, fFADCDataBranch(0),
   fPixelDataBranch(0), fPixelListBranch(0) */
 {
  if ( fgPrintLevel > 1 )
    FD_COUT << "EyeEventFile::TEyeEventFile() - " 
            << hex << this << dec << endl;
  
  Setup();
}

// -------------------------------------------------------------------------

TEyeEventFile::TEyeEventFile(const char *fname, Option_t *option, 
	                     const char *ftitle, Int_t compress )
 : TFile(fname, option, ftitle, compress),
   fBranch(0) ,fCurrentEntry(0), fPartialEvent(0), fEventBuffer(0), fTree(0),
   fEventHeaderBranch(0) /*, fFADCDataBranch(0),
   fPixelDataBranch(0), fPixelListBranch(0) */
 {
  if ( fgPrintLevel > 1 )
    FD_COUT << "EyeEventFile::TEyeEventFile('" << fname << "') - " 
            << hex << this << dec << endl;
  
  assert( option != NULL );
  
  TString optionstr( option);
  optionstr.ToUpper();
   
  if (    optionstr == "CREATE" || optionstr == "RECREATE" 
       || optionstr == "UPDATE" ) {
    fgFileList.push_back( this );
    if ( fgPrintLevel )
      FD_COUT << "EyeEventFile::TEyeEventFile() - added " 
              << hex << this << dec << " to the list !" << endl;
  }

  Setup();
}

// -------------------------------------------------------------------------

TEyeEventFile::~TEyeEventFile()
 {
  if ( fgPrintLevel > 1 )
    FD_COUT << "TEyeEventFile::~TEyeEventFile() - " << hex << this << dec << endl;
    
  if ( fTree ) Close();
  // if ( fPartialEvent ) delete fPartialEvent;
  
  // flag this file as closed in the global list
  
  list<TEyeEventFile*>::iterator iter =
                        find( fgFileList.begin(), fgFileList.end(), this );
  
  if ( iter != fgFileList.end() ) {
    if ( fgPrintLevel )
      FD_COUT << "TEyeEventFile::~TEyeEventFile() - removed "
              << hex << this << dec << " from the list !" << endl;
    *iter = NULL;
  }
}

// -------------------------------------------------------------------------

void TEyeEventFile::Close(Option_t *option)
 {
  if ( fgPrintLevel > 1 )
    FD_COUT << "TEyeEventFile::Close('" << option << "') - " << endl;
    
  if ( fgPrintLevel > 1 ) {
    FD_COUT << "\tbytes written: " 
            << setw(10) << (UInt_t)TFile::GetBytesWritten() << endl;
    FD_COUT << "\tbytes read:    " 
            << setw(10) << (UInt_t)TFile::GetBytesRead() << endl;
  }
  
  // --- file was opened in write mode ...
  
  if ( TFile::GetBytesWritten() > 0. ) {
    TFile::Write( 0, TObject::kOverwrite );
    TFile::Close( option );
  }
  
  // --- file was opened in read mode ...

  if ( TFile::GetBytesRead() > 0. ) {  
    TFile::Close( option );
  }

  fTree   = 0;
  fBranch = 0;

  fEventBuffer  = 0;  // now this address is invalid
}

// -------------------------------------------------------------------------

void TEyeEventFile::Show(std::ostream& ostr)
 {
  ostr << "--- TEyeEventFile ---" << endl;
  
  bool read_mode = ( (TFile::GetBytesRead() > 0.) ? true : false );

  ostr << " File " << GetName() << " opened for ";
  if ( read_mode )
    ostr << "reading" << endl;
  else 
    ostr << "writing" << endl;

  ostr << " CompressionLevel  = " << GetCompressionLevel() << endl;
  ostr << " CompressionFactor = " << GetCompressionFactor() << endl;
  ostr << " Number of Entries = " << GetNEvents() << endl;
  
  if ( read_mode )
    ostr << " Bytes on file     = " << TFile::GetSize() << endl;
  else
    ostr << " Bytes written     = " << TFile::GetBytesWritten() << endl;

  ostr << "---------------------------------" << endl << endl;
}

// -------------------------------------------------------------------------

void TEyeEventFile::ExitHandler()
 {
  if ( fgPrintLevel > 1 )
    FD_CINFO << "TEyeEventFile::ExitHandler() - "
             << "closing files open for writing ..." << endl;
  
  // --- clear list of files ...
  
  if ( !fgFileList.empty() ) {
  
    list<TEyeEventFile*>::iterator it1 = fgFileList.begin();
  
    while ( it1 != fgFileList.end() ) {
      if ( *it1 != NULL && fgPrintLevel > 1 ) 
        FD_CDEBUG(1) << "\t\t... closing: " << hex << *it1 << dec << endl;
      delete *it1;  // works also for NULL ptrs
      it1 = fgFileList.erase( it1 );
    }
  }
  
  // --- clear list of signal handlers ...
    
  list<TEyeEventFileSignalHandler*>::iterator it2;
  it2 = fgSignalHandlerList.begin();
  
  while ( it2 != fgSignalHandlerList.end() ) {
    (*it2)->Remove();
    delete *it2;
    it2 = fgSignalHandlerList.erase( it2 );
  }
  
  if ( fgPrintLevel > 1 )
    FD_COUT << "TEyeEventFile::ExitHandler() - done !" << endl;
}

// -------------------------------------------------------------------------

UInt_t TEyeEventFile::GetCurrentEntryId() const
 {
  if ( fNEntries == 0 ) return 0;
  if ( fCurrentEntry >= fNEntries )
    return fNEntries - 1;
  if ( fCurrentEntry == 0 )
    return 0;
  else
    return fCurrentEntry - 1;
}

// -------------------------------------------------------------------------

UInt_t TEyeEventFile::GetNEvents()
 {
  if ( TFile::GetBytesRead() ) {  // file was opened for reading
    if ( fTree == 0 ) GetTree();
    if ( fTree == 0 )
      return 0;
    else
      return (UInt_t)fTree->GetEntries();
  }
  else {                          // file was opened for writing
    if ( fTree == 0 )
      return 0;
    else
      return fCurrentEntry;
  }
}

// -------------------------------------------------------------------------

Bool_t TEyeEventFile::GetBranches(TEyeEvent **event)
 {
  if ( !fBranch ) {
    fBranch = fTree->GetBranch( gEyeBranchName );
    if ( !fBranch ) {
      FD_CERR << "TEyeEventFile::GetBranches() - "
              << "no branch with name '" << gEyeBranchName << "' found !" << endl;
      return kFALSE;
    }
  }
  
  fBranch->SetAddress( event );

  if ( !fEventHeaderBranch ) {
    fEventHeaderBranch = fTree->GetBranch( "fEventHeader" );
    if ( !fEventHeaderBranch ) {
      FD_CERR << "TEyeEventFile::GetBranches() - "
              << "no branch with name 'fEventHeader' found !" << endl;
      return kFALSE;
    }
  }

  return kTRUE;
}

// -------------------------------------------------------------------------

Bool_t TEyeEventFile::GetTree()
 {
  if ( !fTree ) {
    fTree = (TTree *)fFile->Get( gEyeTreeName );
    if ( !fTree ) {

      FD_CERR << "TEyeEventFile::GetTree() - no Tree with name '" 
	      << gEyeTreeName << "' found !" << endl;
      
      return kFALSE;
    }
    
    if ( fgPrintLevel > 2 ) fTree->Print();

#if ((ROOT_VERSION_CODE >= ROOT_VERSION(5,10,0)) && \
     (ROOT_VERSION_CODE < ROOT_VERSION(5,19,4)))
    // ROOT >= 5.10.0, cannot skip non-existing branch with splitlevel>0
    // -> disable reading from obsolete branch 'fWeatherData'
#warning TEyeEventFile - TTree workaround for 'splitlevel>0' activated!
    fTree->SetBranchStatus( "fWeatherData", 0 );
#endif // ROOT_VERSION_CODE

    fNEntries = (Int_t)fTree->GetEntries();
  }
  
  return kTRUE;
}

// -------------------------------------------------------------------------

Bool_t TEyeEventFile::ReadEvent(TEyeEvent** event)
 {
  bool valid_event = false, file_end = false;
  
  cd();  // set active directory
  
  if ( !GetTree() ) return kFALSE;

  if ( !GetBranches( event ) ) return kFALSE;
  
  // --- reading TEyeEvent from TTree in TEyeEventFile
  
  int debug_save = gDebug;
  
  if ( fgPrintLevel > 3 ) gDebug = 2;

  while ( !valid_event && !file_end ) {
    
    if ( fCurrentEntry < fNEntries ) {
      valid_event = true;
      
      fTree->GetEvent( fCurrentEntry );
      fCurrentEntry++;
    }
    else {
      FD_CWARN << "TEyeEventFile::ReadEvent() - end of file reached !" << endl;
      
      file_end = true;
    }
  } // while ( ... )
  
  gDebug = debug_save;

  if ( valid_event ) {
  
    TEyeEvent::DoCorrections( *event );
  }
  
  if ( valid_event )
    return kTRUE;
  else
    return kFALSE;
}

// -------------------------------------------------------------------------

Bool_t TEyeEventFile::ReadEventHeader(TEyeEventHeader** header)
 {   
  assert( header != NULL );
  
  Bool_t valid = kFALSE, file_end = kFALSE;
  
  cd();  // set active directory
  
  if ( !GetTree() ) return kFALSE;

  if ( !GetBranches( &fPartialEvent ) ) return kFALSE;

  while ( !valid && !file_end ) {

    if ( fCurrentEntry < fNEntries ) {

      valid = kTRUE;
      
      Int_t n_bytes = fEventHeaderBranch->GetEntry( fCurrentEntry );
      
      if ( n_bytes == 0 ) return kFALSE; // nothing read
      
      *header = fPartialEvent->GetEventHeader();
      
      fCurrentEntry++;
    }
    else {
      FD_CERR << "TEyeEventFile::ReadEventHeader() - "
              << "end of file reached !" << endl;
      
      file_end = kTRUE;
    }
  }
  
  if ( valid )
    return kTRUE;
  else
    return kFALSE;
}

// -------------------------------------------------------------------------

Bool_t TEyeEventFile::ReadRunHeader(TEyeRunHeader** run_header)
 {
  cd();  // set active directory
  
  if ( *run_header ) {
    delete *run_header;
    *run_header = NULL;
  }
  
  if ( ( *run_header = (TEyeRunHeader*)Get( "RunHeader" ) ) == NULL ) {
    FD_CERR << "TEyeEventFile::ReadRunHeader() - "
            << "reading 'RunHeader' from file was unsuccessful !" << endl;
    return kFALSE;
  }
  
  // (*run_header)->Show();

  return kTRUE;
}

// -------------------------------------------------------------------------

Bool_t TEyeEventFile::ReadRunTrailer(TEyeRunTrailer** run_trailer)
 {
  cd();  // set active directory
  
  if ( *run_trailer ) {
    delete *run_trailer;
    *run_trailer = NULL;
  }
  
  if ( ( *run_trailer = (TEyeRunTrailer*)Get( "RunTrailer" ) ) == NULL ) {
    FD_CERR << "TEyeEventFile::ReadRunTrailer() - "
            << "reading 'RunTrailer' from file was unsuccessful !" << endl;
    return kFALSE;
  }

  return kTRUE;
}

// -------------------------------------------------------------------------

static bool gHandlerInstalled = false;

void TEyeEventFile::Setup()
 {
  if ( !gHandlerInstalled ) {
  
    // --- install exit handler ...

    if ( atexit( ExitHandler ) < 0 ) {
      FD_CERR << "TEyeEventFile::Setup() - "
              << "trying installing 'atexit' handler with error, "
              << errno << " - " << strerror( errno ) << endl;
    }
    else if ( fgPrintLevel ) {
      FD_CINFO << "TEyeEventFile::Setup() - "
               << "'atexit' handler successfully installed !" << endl;
    }

#ifndef DPA
    // --- install signal handlers
    
    TEyeEventFileSignalHandler *sh;
    
    sh = new TEyeEventFileSignalHandler( kSigQuit );
    sh->Add();
    
    fgSignalHandlerList.push_back( sh );
  
# warning Handler for SIGINT disabled to make Ctrl-C available
#if 0
    sh = new TEyeEventFileSignalHandler( kSigInterrupt );
    sh->Add();
    
    fgSignalHandlerList.push_back( sh );
#endif

    sh = new TEyeEventFileSignalHandler( kSigTermination );
    sh->Add();
    
    fgSignalHandlerList.push_back( sh );

#ifdef HAVE_SIGNAL_H
    // --- special case: handler for SIGSEGV (cannot be replaced in ROOT)

    struct sigaction sa;
  
    sa.sa_handler = SigSegvHandler;
    sigemptyset ( &sa.sa_mask );
    sa.sa_flags = SA_RESTART;
    sa.sa_restorer = NULL; 

    int status = sigaction ( SIGSEGV, &sa, &fgOldSigSegvAction );
  
    if ( status < 0 ) {
      FD_CERR << "TEyeEventFile::Setup() - "
              << "Handler for SIGSEGV not installed, status= " 
	      << status << " errno= " << errno << endl;
    }
    else if ( fgPrintLevel ) {
      FD_CINFO << "TEyeEventFile::Setup() - "
               << "Handler for SIGSEGV installed, status = " << status
               << endl;
    }
#endif // HAVE_SIGNAL_H
#endif // !DPA

    gHandlerInstalled = true;
  }
}

// -------------------------------------------------------------------------

#ifndef DPA
void TEyeEventFile::SigSegvHandler(int sig)
 {
  FD_CFATAL << "TEyeEventFile::SigSegvHandler(sig= " << sig << ") - " << endl;
  
#ifdef HAVE_SIGNAL_H
  // there might be more SegFaults --> deinstall this handler
#if 0
  
  struct sigaction sa;
  
  sa.sa_handler = fgOldSigSegvAction.sa_handler;
  sigemptyset ( &sa.sa_mask );
  sa.sa_flags = SA_RESTART;
  sa.sa_restorer = NULL; 

  sigaction( SIGSEGV, &sa, NULL );
#else
  struct sigaction sa;
  
  sa.sa_handler = SIG_DFL;
  sigemptyset ( &sa.sa_mask );
  sa.sa_flags = SA_RESTART;
  sa.sa_restorer = NULL; 

  sigaction( SIGSEGV, &sa, NULL );
#endif
#endif // HAVE_SIGNAL_H

  // this handler will close our open files (TEyeEventFile type)
  
  ExitHandler();
  
#if 1
  // try to call the old handler ...
  
  if (    fgOldSigSegvAction.sa_handler != SIG_DFL
       && fgOldSigSegvAction.sa_handler != SIG_IGN )
    fgOldSigSegvAction.sa_handler( sig );
#endif

  // terminate finally ... ?
}
#endif // DPA

// -------------------------------------------------------------------------

Bool_t TEyeEventFile::WriteEvent(TEyeEvent** event)
 {
  if ( fgPrintLevel > 1 )
    FD_COUT << "TEyeEventFile::WriteEvent(): " << hex << *event << dec << endl;

  // if ( fTree ) fTree->GetCurrentFile()->cd(); // set active directory
  
  if ( !fTree ) {
  
    cd();  // attach TTree to this file !
    
    if ( fgPrintLevel > 1 )
      FD_COUT << "TEyeEventFile::WriteEvent() - "
              << "set fEventBuffer to " << hex << fEventBuffer << dec << endl;

    fEventBuffer = *event;
    
    fTree = new TTree( gEyeTreeName, "TEyeEvent tree" );
    if ( !fTree ) {
      FD_CERR << "TEyeEventFile::WriteEvent() - "
              << "could not create TTree !" << endl;
      return kFALSE;
    }
    
    if ( fgPrintLevel > 1 )
      FD_COUT << "TEyeEventFile::WriteEvent() - "
              << "set fTree to " << hex << fTree << dec << endl;
        
    fTree->SetAutoSave( 20000 );  // autosave when 20K written
  
    // create branch in split mode !
    // address where the pointer to the TEyeEvent is, mustn't be modified !
    
    TBranch * eventbranch = fTree->GetBranch( gEyeBranchName );
	 
    if ( !eventbranch )
      fTree->Branch( gEyeBranchName, "TEyeEvent", event, 500, 1 );
    else
      eventbranch->SetAddress( event );
	  
    TBranch *branch = fTree->GetBranch( "fFADCData" );
    if ( branch ) branch->SetBasketSize( 500000 );
    branch = fTree->GetBranch( "fPixelList" );
    if ( branch ) branch->SetBasketSize( 5000 );
    branch = fTree->GetBranch( "fPixelData" );
    if ( branch ) branch->SetBasketSize( 50000 );
  }
  
  if ( fEventBuffer != NULL && *event != fEventBuffer ) {
    FD_CERR << "TEyeEventFile::WriteEvent() - "
            << "address of data buffer " << hex << fEventBuffer 
	    << dec << " was changed !" << endl;
    FD_CERR << " -----> No data written to file !" << endl;
    return kFALSE;
  }

  int debug_save = gDebug;
  
  if ( fgPrintLevel > 3 ) gDebug = 2;

  gSystem->IgnoreSignal( kSigInterrupt );
  
   fTree->Fill();

  gSystem->IgnoreSignal( kSigInterrupt, kFALSE );

  if ( fgPrintLevel > 2 ) fTree->Print();
  
  gDebug = debug_save;
  
  return kTRUE;
}

// -------------------------------------------------------------------------

Bool_t TEyeEventFile::WriteRunHeader(TEyeRunHeader* run_header)
 {
  if ( fgPrintLevel > 1 )
    FD_COUT << "TEyeEventFile::WriteRunHeader()" << endl;

  if ( !run_header ) return ( kFALSE );
  
  cd();  // set active directory
  
  run_header->Write( "RunHeader", TObject::kOverwrite );
  
  return kTRUE;
}

// -------------------------------------------------------------------------

Bool_t TEyeEventFile::WriteRunTrailer(TEyeRunTrailer* run_trailer)
 {
  if ( fgPrintLevel > 1 )
    FD_COUT << "TEyeEventFile::WriteRunTrailer()" << endl;

  if ( !run_trailer ) return ( kFALSE );
  
  cd();  // set active directory
  
  run_trailer->Write( "RunTrailer", TObject::kOverwrite );
  
  return kTRUE;
}

// -------------------------------------------------------------------------
// -------------------------------------------------------------------------
