
//
// File   : EyeEventHeader.cc
//
// Purpose: Implementation of Eye event header class
//
// $Id: EyeEventHeader.cc 11941 2011-01-25 13:19:58Z mathes $
//


#include <cstring>
#include <climits>
#include <string>
#include <iostream>
#include <iomanip>
#include <map>
#include <cassert>


/** @file EyeEventHeader.cc
  * Implementation of the class TEyeEventHeader.
  * @author H.-J. Mathes, FzK
  */

#include <TObject.h>
#include <TClass.h>  // TClass::ReadBuffer(), TClass::WriteBuffer()
#include <TBuffer.h>

#include <FdConfig.hh>
#include <ErrorLogging.hh>
#include <FdNumbering.hh>

#include "CheckVersion.hh"

#include "EyeEventHeader.hh"

using namespace std;
using namespace FdUtil;

// -------------------------------------------------------------------------
// -------------------------------------------------------------------------
//
// class TEyeEventHeader - implementation
//

TEyeEventHeader::TEyeEventHeader()
 : fStreamerVersion(0), 
   fEventNumber(0), fEyeNumber(FdUtil::GetEyeId()), 
   fEventClass(kUnClassified), fEventType(kUnknown), 
   fMirrorDataBits(0), fMirrorStatus(0), fRunNo(0), fStatusBits(0),
   // preset with meaningful default values
   fSecStartOffset(0), fSecEndOffset(0), fTicksPerSecond(kTicksPerSecond),
   fPPSPhase(0),
   // preset with defaults --> i.e. time corr. parms not set !!!
   fTimeCorrectionQuality(kCorrectionNotSet), fT10Mhz(0.0), fOffset(0.0),
   fT3Id(0), fNRejectedT3(0),
   fTotalT3DeadTime(0.0)
 {
#ifdef TRACE_CONSTRUCTOR
  FD_CINFO << "TEyeEventHeader(): " << this << endl;
#endif // TRACE_CONSTRUCTOR

  unsigned int gps_zero = TimeConvert::GetUnixTimeFromGPSTime( 0 );
  
  FdRoot::TTimeStamp gps_zero_time_stamp((time_t)gps_zero, 0);
  
  fTimeStamp = gps_zero_time_stamp;
  fT3TimeAtGround = gps_zero_time_stamp;
  
  for ( UInt_t i=0; i<Fd::kEYE_NMIRRORS; i++ ) {
    fRunTimeCounter[i]         = 0LL;
    fDeadTimeCounter[i]        = 0LL;
    fMirrorTimeOffset[i]       = 0;
    fRawMirrorTime[i]          = gps_zero_time_stamp;
    fTriggerSource[i]          = (UChar_t)TMirrorEventHeader::kUnknownTrigger;
    fTriggerTimeStamp[i]       = 0;
    fMirrorDaqDeadTime[i]      = 0LL;
    fMirrorNSkippedTriggers[i] = 0;
    fMirrorEventLabel[i]       = TMirrorEventHeader::kUnlabelled;
  }
  
  fVetoTimeCounter[0] = 0LL;
  fEyeDaqDeadTime     = 0LL;
  
  // FD_CDEBUG(1) << "TEyeEventHeader::TEyeEventHeader()" << endl;
}

// -------------------------------------------------------------------------

TEyeEventHeader::TEyeEventHeader(const TEyeEventHeader& rhs)
 : TObject(rhs)
 {
#ifdef TRACE_CONSTRUCTOR
  FD_CINFO << "TEyeEventHeader(TEyeEventHeader& " << this << "): rhs= " 
           << &rhs << endl;
#endif // TRACE_CONSTRUCTOR
  *this = rhs;
}

// -------------------------------------------------------------------------

TEyeEventHeader& TEyeEventHeader::operator=(const TEyeEventHeader& rhs)
 {
#ifdef TRACE_CONSTRUCTOR
  FD_CINFO << "TEyeEventHeaderoperator=(" << &rhs 
           << " -> " << this << ")" << endl;
#endif // TRACE_CONSTRUCTOR

  if ( this == &rhs ) return *this; 

  // now copy all data members, TObject stuff is not copied !
  
    // --- general eye related stuff

  fEventNumber    = rhs.fEventNumber;  
  fRunNo          = rhs.fRunNo;
  fEyeNumber      = rhs.fEyeNumber;
  fEventClass     = rhs.fEventClass;
  fEventType      = rhs.fEventType;
  fMirrorDataBits = rhs.fMirrorDataBits;
  fMirrorStatus   = rhs.fMirrorStatus;
  fStatusBits     = rhs.fStatusBits;

    // --- event time related stuff

  fTimeStamp      = rhs.fTimeStamp;
  
  fSecStartOffset = rhs.fSecStartOffset;
  fSecEndOffset   = rhs.fSecEndOffset;
  fTicksPerSecond = rhs.fTicksPerSecond;
  fPPSPhase       = rhs.fPPSPhase;
  
  fTimeCorrectionQuality = rhs.fTimeCorrectionQuality;

  fT10Mhz         = rhs.fT10Mhz;
  fOffset         = rhs.fOffset;
  
  fVetoTimeCounter[0] = rhs.fVetoTimeCounter[0];
  
  for ( unsigned int i=0; i<FdUtil::Fd::kEYE_NMIRRORS; i++ ) {
    fRunTimeCounter[i]         = rhs.fRunTimeCounter[i];
    fDeadTimeCounter[i]        = rhs.fDeadTimeCounter[i];
    fMirrorTimeOffset[i]       = rhs.fMirrorTimeOffset[i];
    fRawMirrorTime[i]          = rhs.fRawMirrorTime[i];
    fTriggerSource[i]          = rhs.fTriggerSource[i];
    fTriggerTimeStamp[i]       = rhs.fTriggerTimeStamp[i];
    fMirrorDaqDeadTime[i]      = rhs.fMirrorDaqDeadTime[i];
    fMirrorNSkippedTriggers[i] = rhs.fMirrorNSkippedTriggers[i];
  }

  fEyeDaqDeadTime = rhs.fEyeDaqDeadTime;
  
    // ---
    
  for ( UInt_t i=0; i<FdUtil::Fd::kEYE_NMIRRORS; i++ ) {
    fMirrorEventLabel[i]       = rhs.fMirrorEventLabel[i];
  }

    // --- T3 related stuff

  fT3Id                   = rhs.fT3Id;
  fT3TimeAtGround         = rhs.fT3TimeAtGround;
  fNRejectedT3            = rhs.fNRejectedT3;
  fTotalT3DeadTime        = rhs.fTotalT3DeadTime;
  
  return *this;
}

// -------------------------------------------------------------------------

TEyeEventHeader::~TEyeEventHeader()
 {
#ifdef TRACE_CONSTRUCTOR
  FD_CINFO << "~TEyeEventHeader(): " << this << endl;
#endif // TRACE_CONSTRUCTOR
}

// -------------------------------------------------------------------------

void TEyeEventHeader::AdjustMirrorTimes()
 {
  FdRoot::TTimeStamp first((time_t)INT_MAX,0);
  unsigned int first_mirror = Fd::kEYE_FIRST_MIRROR;
  
#ifdef DEBUG_TIMING
  FD_CDEBUG(1) << "TTimeStamp(INT_MAX)::GetSec()= " << first.GetSec() << endl;
#endif
  
  // --- find the earliest mirror, 
  //     but make sure to exclude unavailable Mirrors
  
  for ( unsigned int i=0; i<Fd::kEYE_NMIRRORS; i++ ) {
  
    // Mirror present ?
    if ( IsMirrorDataPresent( i + Fd::kEYE_FIRST_MIRROR ) ) { 
#ifdef DEBUG_TIMING
      FD_CDEBUG(1) << "Mirror[" << i+1 << "] sec,nsec= " 
                   << fRawMirrorTime[i].GetGPSSec() << ","
	           << fRawMirrorTime[i].GetNanoSec() << endl;
#endif
      if ( fRawMirrorTime[i] < first ) {
        first = fRawMirrorTime[i];
	first_mirror = i;
      }
    }
  }

#ifdef DEBUG_TIMING
  FD_CDEBUG(1) << "First mirror is: " << first_mirror + 1 << endl;
#endif

  // --- correct the time of the first mirror
  
  unsigned int sec_out, nsec_out;
  
  DoTimeCorrection( first.GetSec(), first.GetNanoSec(), &sec_out, &nsec_out );
  
#ifdef DEBUG_TIMING
  FD_CDEBUG(1) << "sec_out= " << sec_out << endl;
  FD_CDEBUG(1) << "nsec_out= " << nsec_out << endl;
#endif

  // TTimeStamp has internally Unix time format
  
  FdRoot::TTimeStamp corrected( FdRoot::TTimeStamp::FromGPS( sec_out, nsec_out ) );
  
  fTimeStamp = corrected;
  
  // --- adjust now the offsets for all mirrors which are present
  
  for ( unsigned int i=0; i<Fd::kEYE_NMIRRORS; i++ ) {
  
    if (    IsMirrorDataPresent( i + Fd::kEYE_FIRST_MIRROR )
         && i != first_mirror ) {
    
      DoTimeCorrection( fRawMirrorTime[i].GetSec(),
                        fRawMirrorTime[i].GetNanoSec(), &sec_out, &nsec_out );

      // write corrected mirror time back ?   --> NO !!!
      
      // now calculate offset, might be in next second
      
      if ( sec_out > (unsigned int)fTimeStamp.GetSec() )
        fMirrorTimeOffset[i] = kNanoSecPerSecond + nsec_out - fTimeStamp.GetNanoSec();
      else
        fMirrorTimeOffset[i] = nsec_out - fTimeStamp.GetNanoSec();
   
    } // if ( IsMirrorDataPresent(..) & .. ) 
  }
}

// -------------------------------------------------------------------------

void TEyeEventHeader::ConvertTimeCorrectionParameters()
 {
#ifdef DEBUG_TIMING
  FD_CDEBUG(1) << "TEyeEventHeader::ConvertTimeCorrectionParameters() - " << endl;
#endif // DEBUG_TIMING

  // --- duration of time interval between 2 successive 1 PPS signals
  
  Double_t sec_len = kNanoSecPerSecond - fSecStartOffset + fSecEndOffset;
  
#ifdef DEBUG_TIMING
  FD_CDEBUG(1) << "sec_len= " << sec_len << endl;
#endif // DEBUG_TIMING

#ifdef DEBUG_TIMING
  // --- duration of one tick
  
  Double_t tick_len = (Double_t)kNanoSecPerSecond / (Double_t)kTicksPerSecond;
  
  FD_CDEBUG(1) << "tick_len= " << tick_len << endl;
#endif // DEBUG_TIMING
  
  // --- mean oscillator cycle duration
  
  Double_t osc_mean = sec_len / fTicksPerSecond;
  
#ifdef DEBUG_TIMING
  FD_CDEBUG(1) << "osc_mean= " << osc_mean << endl;
#endif // DEBUG_TIMING

  fT10Mhz = (Double_t)kTicksPerSecond/10000000. * osc_mean;
  fOffset = (Double_t)fPPSPhase + (Double_t)fSecStartOffset;

#ifdef DEBUG_TIMING
  FD_CDEBUG(1) << "fT10Mhz= " << fT10Mhz << endl;
  FD_CDEBUG(1) << "fOffset= " << fOffset << endl;
#endif // DEBUG_TIMING
}

// -------------------------------------------------------------------------

bool TEyeEventHeader::DoTimeCorrection(UInt_t sec_in,Int_t nsec_in,
                                       UInt_t* sec_out,UInt_t* nsec_out)
 {
#ifdef DEBUG_TIMING
  FD_CDEBUG(1) << "TEyeEventHeader::DoTimeCorrection() - "
               << "sec_in= " << sec_in << " nsec_in= " << nsec_in << endl;
  FD_CDEBUG(1) << "\tfT10Mhz= " << fT10Mhz << " fOffset= " << fOffset << endl;
#endif // DEBUG_TIMING
  
  assert( sec_out != NULL && nsec_out != NULL );

  if ( IsTimeCorrectionSet() ) {  // time corr. params set
    
    double real_nano_time = ((double)nsec_in * fT10Mhz) / 100. + fOffset;
  
    if ( real_nano_time < 0 ) {
      *sec_out = sec_in - 1;
      *nsec_out = (UInt_t)(real_nano_time + kNanoSecPerSecond);
    }
    else {
      *sec_out = sec_in;
      *nsec_out = (UInt_t)real_nano_time;
    }
    
    return true;
  }
  else {
    if ( TEyeEvent::fgPrintLevel ) {
      FD_CWARN << "TEyeEventHeader::DoTimeCorrection() - "
               << "excact time calculation is not possible !" << endl;
    }
    
    // use old procedure as best guess ...
    
    if ( nsec_in < 0 ) {
      *sec_out = sec_in - 1;
      *nsec_out = nsec_in + kNanoSecPerSecond;
    }
    else {
      *sec_out = sec_in;
      *nsec_out = nsec_in;
    }
    
    return false;
  }
  
  return false;
}

// -------------------------------------------------------------------------

unsigned long long 
TEyeEventHeader::GetDeadTimeCounter(unsigned int mirror_id) const
 {
  if ( mirror_id >= Fd::kEYE_FIRST_MIRROR && mirror_id <= Fd::kEYE_LAST_MIRROR )
    return fDeadTimeCounter[mirror_id - Fd::kEYE_FIRST_MIRROR];
  else {
    FD_CERR << "TEyeEventHeader::GetDeadTimeCounter(" << mirror_id << ") - "
            << "illegal Mirror id passed !" << endl;
    return 0LL;
  }
}

// -------------------------------------------------------------------------

unsigned int TEyeEventHeader::GetFADCBinNanoTime(const TFADCData* fadc,
                                                 unsigned int bin)
 {
  if ( bin >= fadc->GetNFADCBins() ) {
    FD_CERR << "TEyeEventHeader::GetFADCBinNanoTime() - "
            << "the specified bin " << bin << " is not valid !" << endl;
    return 0;
  }
  
  int mirror_no =   FdUtil::Fd::GetMirrorNo( fadc->GetPixelNumber() )
                  - FdUtil::Fd::kEYE_FIRST_MIRROR;
  
  int nano_sec =   fRawMirrorTime[mirror_no].GetNanoSec() 
                 - 100 * (fadc->GetNFADCBins() - bin);

  unsigned int sec, nsec;
  
  DoTimeCorrection( fRawMirrorTime[mirror_no].GetSec(), nano_sec, &sec, &nsec );
  
  return nsec;
}

// -------------------------------------------------------------------------

unsigned int TEyeEventHeader::GetFADCBinTime(const TFADCData* fadc,
                                             unsigned int bin)
 {
  if ( bin >= fadc->GetNFADCBins() ) {
    FD_CERR << "TEyeEventHeader::GetFADCBinTime() - "
            << "the specified bin " << bin << " is not valid !" << endl;
    return 0;
  }

  int mirror_no =   FdUtil::Fd::GetMirrorNo( fadc->GetPixelNumber() )
                  - FdUtil::Fd::kEYE_FIRST_MIRROR;
  
  int nano_sec =   fRawMirrorTime[mirror_no].GetNanoSec() 
                 - 100 * (fadc->GetNFADCBins() - bin);

  unsigned int sec, nsec;
  
  DoTimeCorrection( fRawMirrorTime[mirror_no].GetSec(), nano_sec, &sec, &nsec );
  
  return sec;
}

// -------------------------------------------------------------------------

unsigned long long 
TEyeEventHeader::GetMirrorDaqDeadTime(unsigned int mirror_id) const
 {
  if ( mirror_id < Fd::kEYE_FIRST_MIRROR || mirror_id > Fd::kEYE_LAST_MIRROR ) {
    FD_CERR << "TEyeEventHeader::GetMirrorDaqDeadTime(" << mirror_id 
            << "): illegal mirror id passed !" << endl;
    return 0LL;
  }
  else
    return fMirrorDaqDeadTime[mirror_id - Fd::kEYE_FIRST_MIRROR];
}

// -------------------------------------------------------------------------

int
TEyeEventHeader::GetMirrorEventLabel(unsigned int mirror_id) const
 {
  if ( mirror_id < Fd::kEYE_FIRST_MIRROR || mirror_id > Fd::kEYE_LAST_MIRROR ) {
    FD_CERR << "TEyeEventHeader::GetMirrorEventLabel(" << mirror_id 
            << "): illegal mirror id passed !" << endl;
    return TMirrorEventHeader::kUnlabelled;
  }
  else
    return fMirrorEventLabel[mirror_id - Fd::kEYE_FIRST_MIRROR];
}

// -------------------------------------------------------------------------

TMirrorEventHeader::EMiEventType
TEyeEventHeader::GetMirrorEventType(TEyeEventHeader::EEventType eye_etype)
 {
  switch ( eye_etype ) {

    case kUnknown:
         return TMirrorEventHeader::kUnknownEvent;
      
    case kPhysics:
         return TMirrorEventHeader::kPhysicsEvent;
      
    case kCalibration:
         return TMirrorEventHeader::kCalibrationEvent;
      
    case kSimulated:
         return TMirrorEventHeader::kSimulatedEvent;
    
    // those event types are assigned to 'normal' events using their time
    case kLaserOffset:
    case kFlasherOffset:
    case kCalibrationOffset:
    case kLidarOffset:		
    case kAPFOffset:
         return TMirrorEventHeader::kPhysicsEvent;
      
    case kInvalidType:
    default:
         return TMirrorEventHeader::kUnknownEvent;
  }
}

// -------------------------------------------------------------------------

unsigned int 
TEyeEventHeader::GetMirrorNSkippedTriggers(unsigned int mirror_id) const
 {
  if ( mirror_id < Fd::kEYE_FIRST_MIRROR || mirror_id > Fd::kEYE_LAST_MIRROR ) {
    FD_CERR << "TEyeEventHeader::GetMirrorNSkippedTriggers(" << mirror_id 
            << "): illegal mirror id passed !" << endl;
    return 0;
  }
  else
    return fMirrorNSkippedTriggers[mirror_id - Fd::kEYE_FIRST_MIRROR];
}

// -------------------------------------------------------------------------

unsigned int TEyeEventHeader::GetMirrorTimeOffset(UInt_t mirror_id) const
 {
  if ( mirror_id < Fd::kEYE_FIRST_MIRROR || mirror_id > Fd::kEYE_LAST_MIRROR ) {
    FD_CERR << "TEyeEventHeader::GetMirrorTimeOffset(" << mirror_id 
            << "): illegal mirror id passed !" << endl;
    return 0;
  }
  else
    return fMirrorTimeOffset[mirror_id - Fd::kEYE_FIRST_MIRROR];
}

// -------------------------------------------------------------------------

const FdRoot::TTimeStamp *
TEyeEventHeader::GetRawMirrorTimeStamp(UInt_t mirror_id) const
 {
  if ( mirror_id < Fd::kEYE_FIRST_MIRROR || mirror_id > Fd::kEYE_LAST_MIRROR ) {
    FD_CERR << "TEyeEventHeader::GetRawMirrorTimeStamp(" << mirror_id 
            << "): illegal mirror id passed !" << endl;
    return NULL;
  }
  else if ( !(fMirrorDataBits & ( 1 << (mirror_id - Fd::kEYE_FIRST_MIRROR)) ) )
    return NULL;
  else
    return &fRawMirrorTime[mirror_id - Fd::kEYE_FIRST_MIRROR];
}

// -------------------------------------------------------------------------

unsigned long long 
TEyeEventHeader::GetRunTimeCounter(UInt_t mirror_id) const
 {
  if ( mirror_id >= Fd::kEYE_FIRST_MIRROR && mirror_id <= Fd::kEYE_LAST_MIRROR )
    return fRunTimeCounter[mirror_id - Fd::kEYE_FIRST_MIRROR];
  else {
    FD_CERR << "TEyeEventHeader::GetRunTimeCounter(" << mirror_id << ") - "
            << "illegal Mirror id passed !" << endl;
    return 0LL;
  }
}

// -------------------------------------------------------------------------

void TEyeEventHeader::GetTimeCorrectionParameters(Float_t *t10,Float_t *offset) const
 {
  assert( t10 != NULL && offset != NULL );
  
  *t10    = fT10Mhz;
  *offset = fOffset;
}

// -------------------------------------------------------------------------

TMirrorEventHeader::EMiTriggerSource 
TEyeEventHeader::GetTriggerSource(UInt_t mirror_id) const
 {
  if ( mirror_id >= Fd::kEYE_FIRST_MIRROR && mirror_id <= Fd::kEYE_LAST_MIRROR )
    return (TMirrorEventHeader::EMiTriggerSource)fTriggerSource[mirror_id - Fd::kEYE_FIRST_MIRROR];
  else {
    FD_CERR << "TEyeEventHeader::GetTriggerSource(" << mirror_id << ") - "
            << "illegal Mirror id specified !" << endl;
    return TMirrorEventHeader::kUnknownTrigger;
  }
}

// -------------------------------------------------------------------------

UInt_t TEyeEventHeader::GetTriggerTimeStamp(UInt_t mirror_id) const
 {
  if ( mirror_id >= Fd::kEYE_FIRST_MIRROR && mirror_id <= Fd::kEYE_LAST_MIRROR )
    return fTriggerTimeStamp[mirror_id - Fd::kEYE_FIRST_MIRROR];
  else {
    FD_CERR << "TEyeEventHeader::GetTriggerTimeStamp(" << mirror_id << ") - "
            << "illegal Mirror id specified !" << endl;
    return 0;
  }
}

// -------------------------------------------------------------------------

static map<TEyeEventHeader::EEventClass,string> gEventClassMap;

const char * 
TEyeEventHeader::GetVerboseEventClass(TEyeEventHeader::EEventClass eclass)
 {
  static bool first = true;
  
  // --- initialize the gEventClassMap
  
  if ( first ) {
  
    gEventClassMap.insert( make_pair( TEyeEventHeader::kUnClassified, "'Unclassified'" ) );
    gEventClassMap.insert( make_pair( TEyeEventHeader::kNoPhysicsEvent, "'NoPhys'" ) );
    gEventClassMap.insert( make_pair( TEyeEventHeader::kLargeEvent, "'Large Event'" ) );
    gEventClassMap.insert( make_pair( TEyeEventHeader::kUnhandledSignature, "'Unhandled Signature'" ) );
    gEventClassMap.insert( make_pair( TEyeEventHeader::kEmpty, "'Empty'" ) );
    gEventClassMap.insert( make_pair( TEyeEventHeader::kIsMuon, "'Muon'" ) );
    gEventClassMap.insert( make_pair( TEyeEventHeader::kLongMuon, "'Long Muon'" ) );
    gEventClassMap.insert( make_pair( TEyeEventHeader::kCloseShower, "'Close Shower'" ) );
    gEventClassMap.insert( make_pair( TEyeEventHeader::kNoise, "'Noise'" ) );
    gEventClassMap.insert( make_pair( TEyeEventHeader::kMuonAndNoise, "'Muon + Noise'" ) );
    gEventClassMap.insert( make_pair( TEyeEventHeader::kVetoedEvent, "'Veto'" ) );
    gEventClassMap.insert( make_pair( TEyeEventHeader::kShowerCandidate, "'Shower Candidate'" ) );
    gEventClassMap.insert( make_pair( TEyeEventHeader::kHorizontalShower, "'Horizontal Shower'" ) );
    gEventClassMap.insert( make_pair( TEyeEventHeader::kIsShower, "'Shower'" ) );

    gEventClassMap.insert( make_pair( TEyeEventHeader::kUBANoise, "'UBA T3 noise'" ) );
    gEventClassMap.insert( make_pair( TEyeEventHeader::kUserDefined2, "'User defined 2'" ) );
    gEventClassMap.insert( make_pair( TEyeEventHeader::kUserDefined3, "'User defined 3'" ) );
    gEventClassMap.insert( make_pair( TEyeEventHeader::kUserDefined4, "'User defined 4'" ) );

    gEventClassMap.insert( make_pair( TEyeEventHeader::kRejected, "'Rejected'" ) );
    gEventClassMap.insert( make_pair( TEyeEventHeader::kInvalidClass, "'Invalid event class'" ) );

    first = false;
  }
  
  map<TEyeEventHeader::EEventClass,string>::iterator found;
  
  found = gEventClassMap.find( eclass );
  
  if ( found == gEventClassMap.end() )
    return "'Invalid event class'";
  else
    return found->second.c_str();
}

// -------------------------------------------------------------------------

static map<TEyeEventHeader::EEventType,string> gEventTypeMap;

const char *
TEyeEventHeader::GetVerboseEventType(TEyeEventHeader::EEventType etype)
 {
  static bool first = true;
  
  // --- initialize the gEventTypeMap
  
  if ( first ) {

    gEventTypeMap.insert( make_pair( TEyeEventHeader::kUnknown, "'Unknown'" ) );
    gEventTypeMap.insert( make_pair( TEyeEventHeader::kPhysics, "'Physics - Int or L/R trigger'" ) );
    gEventTypeMap.insert( make_pair( TEyeEventHeader::kCalibration, "'Calibration - Ext trigger'" ) );
    gEventTypeMap.insert( make_pair( TEyeEventHeader::kSimulated, "'Simulated - Sw trigger'" ) );
    gEventTypeMap.insert( make_pair( TEyeEventHeader::kLaserOffset, "'Laser Offset'" ) );
    gEventTypeMap.insert( make_pair( TEyeEventHeader::kFlasherOffset, "'Flasher Offset'" ) );
    gEventTypeMap.insert( make_pair( TEyeEventHeader::kCalibrationOffset, "'Calibration Offset'" ) );
    gEventTypeMap.insert( make_pair( TEyeEventHeader::kLidarOffset, "'Lidar Offset'" ) );
    gEventTypeMap.insert( make_pair( TEyeEventHeader::kAPFOffset, "'APF Offset'" ) );

    gEventTypeMap.insert( make_pair( TEyeEventHeader::kInvalidType, "'Invalid'" ) );

    first = false;
  }
  
  map<TEyeEventHeader::EEventType,string>::iterator found;
  
  found = gEventTypeMap.find( etype );
  
  if ( found == gEventTypeMap.end() )
    return "Invalid Type";
  else
    return found->second.c_str();
}

// -------------------------------------------------------------------------

const char* 
TEyeEventHeader::GetVerboseTimeCorrectionQuality(TEyeEventHeader::ETimeCorrectionQuality quality)
 {
  switch ( quality ) {
    case kCorrectionNotSet:
         return "'NotSet'";
    case kAveragedPeriod:
         return "'Averaged'";
    case kFullQuality:
         return "'Full'";
    case kUnknownQuality:
    default:
         return "'Unknown'";
  }
}

// -------------------------------------------------------------------------

bool TEyeEventHeader::IsMirrorDataPresent(unsigned int mirror_id) const
 {
  if ( mirror_id < Fd::kEYE_FIRST_MIRROR && mirror_id > Fd::kEYE_LAST_MIRROR )
    return kFALSE;
  if ( fMirrorDataBits & ( 1 << (mirror_id - Fd::kEYE_FIRST_MIRROR) ) )
    return kTRUE;
  else
    return kFALSE;
}

// -------------------------------------------------------------------------

bool TEyeEventHeader::IsMirrorPresent(unsigned int mirror_id) const
 {
  if ( mirror_id < Fd::kEYE_FIRST_MIRROR && mirror_id > Fd::kEYE_LAST_MIRROR )
    return kFALSE;
  if ( fMirrorStatus & ( 1 << (mirror_id - Fd::kEYE_FIRST_MIRROR) ) )
    return kTRUE;
  else
    return kFALSE;
}

// -------------------------------------------------------------------------

bool TEyeEventHeader::IsTimeCorrectionSet()
 {
  if ( fT10Mhz <= 50.0 || fT10Mhz > 150.0 ) {
    // check if 'raw' parameters have been set (not defaults)
    if (    fSecStartOffset == 0 && fSecEndOffset == 0
         && fTicksPerSecond == kTicksPerSecond && fPPSPhase == 0 ) {
      if ( TEyeEvent::fgPrintLevel ) {
        FD_CWARN << "TEyeEventHeader::IsTimeCorrectionSet() - "
                 << "raw time correction parameters are set to defaults !" << endl;
      }
      return false;
    }

    ConvertTimeCorrectionParameters();
  }

  return (fT10Mhz > 50.0);
}

// -------------------------------------------------------------------------

void TEyeEventHeader::SetDeadTimeCounter(unsigned long long counter,
                                         unsigned  mirror_id)
 {
  if ( mirror_id >= Fd::kEYE_FIRST_MIRROR && mirror_id <= Fd::kEYE_LAST_MIRROR )
    fDeadTimeCounter[mirror_id - Fd::kEYE_FIRST_MIRROR] = counter;
  else {
    FD_CERR << "TEyeEventHeader::SetDeadTimeCounter(," << mirror_id << ") - "
            << "illegal Mirror id passed !" << endl;
  }
}

// -------------------------------------------------------------------------

void TEyeEventHeader::SetMirrorDaqDeadTime(unsigned long long dead_time,
                                           unsigned int mirror_id)
 {
  if ( mirror_id >= Fd::kEYE_FIRST_MIRROR && mirror_id <= Fd::kEYE_LAST_MIRROR )
    fMirrorDaqDeadTime[mirror_id - Fd::kEYE_FIRST_MIRROR] = dead_time;
  else {
    FD_CERR << "TEyeEventHeader::SetMirrorDaqDeadTime(" << mirror_id << ") - "
            << "illegal Mirror id passed !" << endl;
  }
}

// -------------------------------------------------------------------------

void TEyeEventHeader::SetMirrorDataPresent(unsigned int mirror_id)
 {
  if ( mirror_id >= Fd::kEYE_FIRST_MIRROR && mirror_id <= Fd::kEYE_LAST_MIRROR )
    fMirrorDataBits |= ( 1 << (mirror_id - Fd::kEYE_FIRST_MIRROR) );
  else {
    FD_CERR << "TEyeEventHeader::SetMirrorDataPresent(" << mirror_id << ") - "
            << "illegal Mirror id passed !" << endl;
  }
}

// -------------------------------------------------------------------------

void TEyeEventHeader::SetMirrorEventLabel(int evlabel,unsigned int mirror_id)
 {
  if ( mirror_id >= Fd::kEYE_FIRST_MIRROR && mirror_id <= Fd::kEYE_LAST_MIRROR )
    fMirrorEventLabel[mirror_id - Fd::kEYE_FIRST_MIRROR] = evlabel;
  else {
    FD_CERR << "TEyeEventHeader::SetMirrorEventLabel(" << mirror_id << ") - "
            << "illegal Mirror id passed !" << endl;
  }
}

// -------------------------------------------------------------------------

void TEyeEventHeader::SetMirrorNSkippedTriggers(unsigned int n_skipped,
                                                unsigned int mirror_id)
 {
  if ( mirror_id >= Fd::kEYE_FIRST_MIRROR && mirror_id <= Fd::kEYE_LAST_MIRROR )
    fMirrorNSkippedTriggers[mirror_id - Fd::kEYE_FIRST_MIRROR] = n_skipped;
  else {
    FD_CERR << "TEyeEventHeader::SetMirrorNSkippedTriggers(" << mirror_id << ") - "
            << "illegal Mirror id passed !" << endl;
  }
}

// -------------------------------------------------------------------------

void TEyeEventHeader::SetMirrorPresent(unsigned int mirror_id)
 {
  if ( mirror_id >= Fd::kEYE_FIRST_MIRROR && mirror_id <= Fd::kEYE_LAST_MIRROR )
    fMirrorStatus |= ( 1 << (mirror_id - Fd::kEYE_FIRST_MIRROR) );
  else {
    FD_CERR << "TEyeEventHeader::SetMirrorPresent(" << mirror_id << ") - "
            << "illegal Mirror id passed !" << endl;
  }
}

// -------------------------------------------------------------------------

void TEyeEventHeader::SetRawMirrorTime(unsigned int mirror_id,
                                       FdRoot::TTimeStamp& timestamp)
 {
  if ( mirror_id >= Fd::kEYE_FIRST_MIRROR && mirror_id <= Fd::kEYE_LAST_MIRROR )
    fRawMirrorTime[mirror_id - Fd::kEYE_FIRST_MIRROR] = timestamp;
  else {
    FD_CERR << "TEyeEventHeader::SetRawMirrorTime(" << mirror_id << ",) - "
            << "illegal Mirror id passed !" << endl;
  }
}

// -------------------------------------------------------------------------

void TEyeEventHeader::SetRunTimeCounter(unsigned long long counter,
                                        unsigned  mirror_id)
 {
  if ( mirror_id >= Fd::kEYE_FIRST_MIRROR && mirror_id <= Fd::kEYE_LAST_MIRROR )
    fRunTimeCounter[mirror_id - Fd::kEYE_FIRST_MIRROR] = counter;
  else {
    FD_CERR << "TEyeEventHeader::SetRunTimeCounter(," << mirror_id << ") - "
            << "illegal Mirror id passed !" << endl;
  }
}

// -------------------------------------------------------------------------

void TEyeEventHeader::SetTimeCorrectionParameters(short start_offset, 
                                                  short end_offset, 
						  unsigned int ticks_per_sec, 
						  unsigned int phase)
 {
#ifdef DEBUG_TIMING
  FD_CDEBUG(1) << "TEyeEventHeader::SetTimeCorrectionParameters() - "
               << "start_offset,end_offset,ticks_per_sec,phase "
               << start_offset << "," << end_offset << "," 
               << ticks_per_sec << "," <<phase << endl;
#endif

  // --- check first if parameters are in reasonable range
  
  if ( start_offset < -128 || start_offset > 127 ) {
    FD_CERR << "TEyeEventHeader::SetTimeCorrectionParameters() - "
            << "value '" << start_offset 
	    << "' passed for 'start_offset' is not plausible !" << endl;
    return;
  }
  
  if ( end_offset < -128 || end_offset > 127 ) {
    FD_CERR << "TEyeEventHeader::SetTimeCorrectionParameters() - "
            << "value '" << end_offset 
	    << "' passed for 'end_offset' is not plausible !" << endl;
    return;
  }
  
  unsigned int tick_error = (unsigned int)(1E-02 * (Double_t)kTicksPerSecond);
  
  if (   ticks_per_sec < (kTicksPerSecond - tick_error)
      || ticks_per_sec > (kTicksPerSecond + tick_error) ) {
    FD_CERR << "TEyeEventHeader::SetTimeCorrectionParameters() - "
            << "value '" << ticks_per_sec 
	    << "' passed for 'ticks_per_sec' is not plausible !" << endl;
    return;
  }
  
  if ( phase > 100 ) {
    FD_CERR << "TEyeEventHeader::SetTimeCorrectionParameters() - "
            << "value '" << phase 
	    << "' passed for 'phase' is not plausible !" << endl;
    return;
  }
  
  // --- and store them
  
  fSecStartOffset = start_offset;
  fSecEndOffset   = end_offset;
  fTicksPerSecond = ticks_per_sec;
  fPPSPhase       = phase;
  
  // --- now calculate the internally used (and stored) parameters
  
  ConvertTimeCorrectionParameters();
  
  // --- set time correction quality, DPA only averaged values !
  
#ifndef DPA
  SetTimeCorrectionQuality( kFullQuality );
#else
  SetTimeCorrectionQuality( kAveragedPeriod );
#endif // DPA
}

// -------------------------------------------------------------------------

void TEyeEventHeader::SetTimeCorrectionParameters(Float_t t10,Float_t offset)
 {
  // --- check first if parameters are in reasonable range
  
  if ( t10 < 50. || t10 > 150. ) {
    FD_CERR << "TEyeEventHeader::SetTimeCorrectionParameters() - "
            << "value '" << t10 
	    << "' passed for 'T10' is not plausible !" << endl;
    return;
  }
  
  // -128 < neg.sawtooth < 127 and 0 <= offset <= 100
  
  if ( offset < -128. || offset > 227. ) {
    FD_CERR << "TEyeEventHeader::SetTimeCorrectionParameters() - "
            << "value '" << offset 
	    << "' passed for 'offset' is not plausible !" << endl;
    return;
  }
  
  // --- and store them
  
  fT10Mhz = t10;
  fOffset = offset;
  
  // --- set time correction quality, DPA only averaged values !

#ifndef DPA
  SetTimeCorrectionQuality( kFullQuality );
#else
  SetTimeCorrectionQuality( kAveragedPeriod );
#endif // DPA
}

// -------------------------------------------------------------------------

void TEyeEventHeader::SetTimeStamp(unsigned int sec,unsigned int nsec)
 {
  FdRoot::TTimeStamp ts( FdRoot::TTimeStamp::FromGPS( sec, nsec ) );
  
  SetTimeStamp( ts );
}

// -------------------------------------------------------------------------

void TEyeEventHeader::SetTriggerSource(TMirrorEventHeader::EMiTriggerSource src,
                                       unsigned int mirror_id)
 {
  if ( TEyeEvent::fgPrintLevel ) {
    FD_CDEBUG(1) << "TEyeEventHeader::SetTriggerSource() - "
                 << "src= " << hex << src << dec << ", id= " << mirror_id 
		 << endl;
  }

  if ( mirror_id >= Fd::kEYE_FIRST_MIRROR && mirror_id <= Fd::kEYE_LAST_MIRROR )
    fTriggerSource[mirror_id - Fd::kEYE_FIRST_MIRROR] = src;
  else {
    FD_CERR << "TEyeEventHeader::SetTriggerSource(," << mirror_id << ") - "
            << "illegal Mirror id passed !" << endl;
  }
}

// -------------------------------------------------------------------------

void TEyeEventHeader::SetTriggerTimeStamp(unsigned int ts,unsigned int mirror_id)
 {
  if ( mirror_id >= Fd::kEYE_FIRST_MIRROR && mirror_id <= Fd::kEYE_LAST_MIRROR )
    fTriggerTimeStamp[mirror_id - Fd::kEYE_FIRST_MIRROR] = ts;
  else {
    FD_CERR << "TEyeEventHeader::SetTriggerTimeStamp(" << mirror_id << ") - "
            << "illegal Mirror id passed !" << endl;
  }
}

// -------------------------------------------------------------------------

void TEyeEventHeader::SetT3TimeStamp(unsigned int sec,unsigned int nsec)
 {
  FdRoot::TTimeStamp t3( FdRoot::TTimeStamp::FromGPS( sec, nsec ) );
  
  SetT3TimeStamp( t3 );
}

// -------------------------------------------------------------------------

void TEyeEventHeader::Show(std::ostream& ostr) const
 {
  ostr << "--- TEyeEventHeader for Eye " << (int)fEyeNumber << " ---" << endl;

  ostr << " Streamer version: " << fStreamerVersion << endl;
  ostr << " Class version   : " << Class_Version() << endl;
  ostr << " Recorded at     : " << fTimeStamp.AsString() << endl;
  ostr << " RUN number      : " << fRunNo << endl;

  ostr << " Timestamp       : " << setw(10) << fTimeStamp.GetSec()
       << " [s] " << setw(10) << fTimeStamp.GetNanoSec() 
       << " [ns] (corrected)" << endl;
  ostr << " Event number    : " << setw(10) << fEventNumber << endl;
  
  unsigned int mask = 0x80000000;
  ostr << " Status bits     : 0x" << setw(8) << hex << setfill('0') << fStatusBits
                                  << setfill(' ') << dec << " = ";
  for ( int i=31; i>=0; --i ) {
    ostr << ( ( (fStatusBits & mask) != 0 ) ? 1 : 0 );
    if ( !(i % 8 ) ) ostr << " ";
    mask >>= 1;
  }
  ostr << endl;
  
  ostr << " Assigned T3 Id  : " << setw(10) << fT3Id;

  if ( !fT3Id )
    ostr << "   (no T3 !)" << endl;
  else {
    ostr << "   (T3 !!!)" << endl;
    ostr << " TimeAtGround    : " << setw(10) << fT3TimeAtGround.GetSec()
         << " [s] " << setw(10) << fT3TimeAtGround.GetNanoSec() 
         << " [ns] ((un)corrected ?)" << endl;
    ostr << " rejected by T3-CDAS VETO: " 
         << (IsRejectedT3() ? "YES" : "NO") << endl;
  }

  ostr << " Event type      : " << GetVerboseEventType( GetEventType() ) << endl;
  ostr << " Event class     : " << GetVerboseEventClass( GetEventClass() ) << endl;
  
  // max number of mirrors differ between HEAT and standard sites
  
  unsigned int n_mirrors;
  
  if ( (int)fEyeNumber < 5 )
    n_mirrors = Fd::kEYE_NMIRRORS;
  else
    n_mirrors = Fd::kHEAT_NMIRRORS;
  
  string trigger_str( " TriggerSources  : " );
  for ( unsigned int i=0; i<n_mirrors; i++ ) {
    char source[8];
    if ( fMirrorDataBits & (1<<i) ) {
      sprintf( source, " 0x%02x", fTriggerSource[i] );
      trigger_str += source;
    }
    else
      trigger_str += " --- ";
  }
  ostr << trigger_str << endl;
  
  ostr << " accum VETO time: " << setw(20) << fVetoTimeCounter[0] 
       << " [* 100 ns]" << endl;
  if ( fEyeDaqDeadTime > 0LL )
    ostr << " Eye DAQ Dead time : " 
         << setw(20) << fEyeDaqDeadTime << " [* 100 ns]" << endl;
  
  string s1( " The following Mirrors participated in the DAQ  :" );
  string s2( " The following Mirrors are present in this event:" );
  
  for ( unsigned int i=0; i<n_mirrors; i++ ) {
    char mirror[5];
    if ( fMirrorStatus & (1<<i) ) {
      sprintf( mirror, " %d", i + Fd::kEYE_FIRST_MIRROR );
      s1 += mirror;
    }
    if ( fMirrorDataBits & (1<<i) ) {
      sprintf( mirror, " %d", i + Fd::kEYE_FIRST_MIRROR );
      s2 += mirror;
    }
  }
  
  ostr << s1 << endl;
  ostr << s2 << endl;
  
  for ( unsigned int i=0; i<n_mirrors; i++ ) {
  
    if ( fMirrorDataBits & (1<<i) ) {
      ostr << " Time Offset[" << i + Fd::kEYE_FIRST_MIRROR << "]  : " 
           << setw(10) << fMirrorTimeOffset[i] << " [ns]" 
	   << endl;
      ostr << " Timestamp[" << i + Fd::kEYE_FIRST_MIRROR << "]    : " 
           << setw(10) << fRawMirrorTime[i].GetSec()
	   << " [s] " << setw(10) << fRawMirrorTime[i].GetNanoSec() 
	   << " [ns] (uncorrected)" 
	   << endl;
      ostr << " TrgTimeStamp[" << i + Fd::kEYE_FIRST_MIRROR << "] : " 
           << setw(10) << hex << fTriggerTimeStamp[i] << dec << " hex" 
	   << endl;
      ostr << " Dead time[" << i + Fd::kEYE_FIRST_MIRROR << "]    : " 
           << setw(20) << fDeadTimeCounter[i] << " [* 100 ns]" 
	   << endl;
      ostr << "  Run time[" << i + Fd::kEYE_FIRST_MIRROR << "]    : " 
           << setw(20) << fRunTimeCounter[i] << " [* 100 ns]" 
	   << endl;
      if ( fMirrorDaqDeadTime[i] > 0.0 )
        ostr << " DaqDead time[" << i + Fd::kEYE_FIRST_MIRROR << "] : " 
             << setw(20) << fMirrorDaqDeadTime[i] << " [* 100 ns]" 
	     << endl;
      if ( fMirrorNSkippedTriggers[i] > 0 )
        ostr << " Skpd triggers[" << i + Fd::kEYE_FIRST_MIRROR << "]: " 
             << fMirrorNSkippedTriggers[i]
	     << endl;
   }
  }
  
  if ( fT10Mhz > 50.0 || GetTimeCorrectionQuality() == kCorrectionNotSet ) {
    ostr << " Time correction quality: " 
         << GetVerboseTimeCorrectionQuality( GetTimeCorrectionQuality() ) 
	 << endl;
    if ( GetTimeCorrectionQuality() != kCorrectionNotSet ) {
      ostr << " T_10MHz        : " << setw(10) 
           << fT10Mhz << " [ns]" << endl;
      ostr << " Sec_Offset     : " << setw(10) 
           << fOffset << " [ns]" << endl;
    }
  }

  if ( GetNRejectedT3() > 0 ) {
    ostr << " T3 rejected totally: " << GetNRejectedT3()
         << " accumulated T3-CDAS dead time: " << GetTotalT3DeadTime() 
	 << " [s]" << endl;
  }
       
  ostr << "---------------------------------" << endl;
}

/** This method creates a hex dump of the 'payload' of an area in the Tbuffer.
  */
static void BufferShow(TBuffer& R__b,UInt_t R__c,ostream& ostr=std::cout)
 {
  unsigned int i, j;
  unsigned char *data = new unsigned char[R__c+1];
  
  for ( i=0; i<R__c; i++ )
    R__b >> data[i];
  
  for ( i=0; i<R__c; i+=16 ) {
  
    // output hex numbers
    for ( j=i; j<i+16; j++ ) {
      if ( j<R__c )
        ostr << " " << setfill('0') << setw(2)
             << hex << (unsigned int)(unsigned char)data[j] << dec 
	     << setfill(' ');
      else
        ostr << "   ";
    }

    ostr << "   ";
    
    // output ASCII chars if feasible
    for ( j=i; j<i+16; j++ ) {
      if ( j<R__c ) {
        if ( isprint(data[j]) )
          ostr << data[j];
        else
          ostr << ".";
      }
    }
    ostr << endl;
  }

  ostr << endl;
}

// -------------------------------------------------------------------------

// Note: Streaming fMirrorTimeStamp[6] to file is broken in root 3.02/07
//       when using the new Streamer scheme.
//       Therefor the next lines prevent us from using that ROOT version.

#if (ROOT_VERSION_CODE < ROOT_VERSION(3,3,0))
# error Using ROOT version 3.02/07 will create corrupted data !!!
#endif // ROOT_VERSION_CODE

// Table of (old) streamer versions and sizes
//
// R__v   R__s   R__c  CVS  EyeEventHeader.hh
//
//   4     103    255  v0r8    1.24
//   5     103    239  v0r9    1.27          test data faulty ?
//   6     103    267  v1r1    1.32
//   7     103    359  v1r4p2  1.44
//   8     103    383  v1r5    1.46
//   9     103    383  v1r7    1.52
//

void TEyeEventHeader::Streamer(TBuffer &R__b)
 {
  // Stream an object of class TEyeEventHeader.

  unsigned int R__s, R__c;
  
  if ( R__b.IsReading() ) {
     
    Version_t R__v = fStreamerVersion = R__b.ReadVersion(&R__s, &R__c);

    if ( TEyeEvent::fgPrintLevel ) {
      FD_CDEBUG(1) << "TEyeEventHeader::Streamer() - R__v= " << R__v
                   << " R__s= " << R__s << " R__c= " << R__c << endl;
    }
    
    if ( R__v == EyeEVENTVERSIONv4 ) {
      StreamerV4(R__b);
      R__b.CheckByteCount(R__s, R__c, TEyeEventHeader::IsA());
    }
    // new Streamer scheme of ROOT v 3.x
    // Note: ROOT 3.05/07 (and also some newer version) have two problems
    //       - the typecode for some long long types is set incorrectly
    //       - restoring a streamer with the automatic streamer to a new class
    //         layout might cause problem
    else if ( R__v == EyeEVENTVERSIONv5 ) {
      //BufferShow(R__b, R__c - sizeof(R__v));
      if ( R__c != 239 ) {
        FD_CFATAL << "TEyeEventHeader::Streamer() - Error: R__c= " << R__c 
	          << " doesn't match for V5, reading aborted!" << endl;
        BufferShow(R__b, R__c - sizeof(R__v));
	return;
      }
      StreamerV5(R__b);
      R__b.CheckByteCount(R__s, R__c, TEyeEventHeader::IsA());
    }
    else if ( R__v == EyeEVENTVERSIONv6 ) {
      StreamerV6(R__b);
      R__b.CheckByteCount(R__s, R__c, TEyeEventHeader::IsA());
    }
    else if ( R__v == EyeEVENTVERSIONv7 ) {
      StreamerV7(R__b);
      R__b.CheckByteCount(R__s, R__c, TEyeEventHeader::IsA());
    }
    else if ( R__v == EyeEVENTVERSIONv8 ) {
      StreamerV8(R__b);
      R__b.CheckByteCount(R__s, R__c, TEyeEventHeader::IsA());
    }
    else if ( R__v >= EyeEVENTVERSIONv9 ) {
      // Note: more parameters as in standard case as we read already 2 bytes
      TEyeEventHeader::Class()->ReadBuffer(R__b, this, R__v, R__s, R__c);
      
      // see http://root.cern.ch/root/Version516.news.html
      //R__b.ReadClassBuffer(TEyeEventHeader::Class(),this, R__v, R__s, R__c);

      // this->Show();
    }
    else { // no support yet for *** VERY OLD and NEW *** versions (<= 3) !
      FD_CERR << "TEyeEventHeader::Streamer() - version " 
              << R__v << " not streamable !" << endl;
    }

    // --- set some fields not available in old data

    if ( R__v <= EyeEVENTVERSIONv5 ) {

      // FD_CDEBUG(1) << "TEyeEventHeader::Streamer() - "
      //              << "setting fTimeCorrectionQuality, IsTimeCorrectionSet()= "
      //              << (int)IsTimeCorrectionSet() << endl;
	   
      if ( IsTimeCorrectionSet() )
        fTimeCorrectionQuality = kFullQuality;
      else
        fTimeCorrectionQuality = kCorrectionNotSet;

      for ( int i=0; i<Fd::kEYE_NMIRRORS; i++ )
        fTriggerTimeStamp[i] = 0;
    }
    
    if ( R__v <= EyeEVENTVERSIONv6 ) {  // fix data missing <= V6
      unsigned int gps_zero = TimeConvert::GetUnixTimeFromGPSTime( 0 );
  
      FdRoot::TTimeStamp gps_zero_time_stamp((time_t)gps_zero, 0);
      
      fStatusBits = 0;
      fEyeDaqDeadTime = 0LL;
      
      for ( unsigned int i=0; i<Fd::kEYE_NMIRRORS; i++ ) {
        fMirrorDaqDeadTime[i] = 0LL;
        fMirrorNSkippedTriggers[i] = 0;
      }
      
      fT3Id = 0;
      fT3TimeAtGround = gps_zero_time_stamp;
      fNRejectedT3 = 0;
      fTotalT3DeadTime = 0LL;
    }
    
    if ( R__v <= EyeEVENTVERSIONv7 ) {  // fix data missing <= V7
      for ( unsigned int i=0; i<Fd::kEYE_NMIRRORS; i++ ) {
        fMirrorEventLabel[i] = TMirrorEventHeader::kUnlabelled;
      }
    }
  //  if ( R__v <= EyeEVENTVERSIONv8 ) {  // fix data missing <= V8
  //  }
  }
  else { // support new Streamer scheme of ROOT v 3.x
    
    TEyeEventHeader::Class()->WriteBuffer(R__b, this);

    // see http://root.cern.ch/root/Version516.news.html
    //R__b.WriteClassBuffer(TEyeEventHeader::Class(),this);
  }

  //TEyeEventHeader::Class()->Dump();
}

ClassImp(TEyeEventHeader)

// -------------------------------------------------------------------------

void TEyeEventHeader::StreamerV4(TBuffer& R__b)
 {
  TObject::Streamer(R__b);

  if ( TEyeEvent::fgPrintLevel ) {
    FD_CDEBUG(1) << "TEyeEventHeader::StreamerV4() - " << endl;
  }
  
  R__b >> fEventNumber;
  R__b >> fEyeNumber;
  R__b >> fEventClass;
  R__b >> fEventType;
  R__b >> fMirrorDataBits;
  R__b >> fMirrorStatus;
  R__b >> fRunNo;
  R__b.ReadStaticArray((unsigned char*)fTriggerSource);
  //R__b.ReadStaticArray(fTriggerSource);
  fTimeStamp.Streamer(R__b);
  R__b >> fT10Mhz;
  R__b >> fOffset;
  R__b.ReadStaticArray((unsigned int*)fMirrorTimeOffset);
  //R__b.ReadStaticArray(fMirrorTimeOffset);
  int R__i;
  for (R__i = 0; R__i < 6; R__i++)
    fRawMirrorTime[R__i].Streamer(R__b);
  // found in TEyeEventHeader::Streamer() (for writing part):
  //R__b.WriteArray((unsigned int*)fDeadTimeCounter, 12);
  //R__b.WriteArray((unsigned int*)fVetoTimeCounter, 2);
  R__b.ReadStaticArray((unsigned int*)fDeadTimeCounter);
  R__b.ReadStaticArray((unsigned int*)fVetoTimeCounter);
  R__b >> fT3Id;
  fT3TimeAtGround.Streamer(R__b);
}

// -------------------------------------------------------------------------

void TEyeEventHeader::StreamerV5(TBuffer &R__b)
 {
  TObject::Streamer(R__b);

  if ( TEyeEvent::fgPrintLevel ) {
    FD_CDEBUG(1) << "TEyeEventHeader::StreamerV5() - " << endl;
  }

  R__b >> fEventNumber;
  R__b >> fEyeNumber;
  R__b >> fEventClass;
  R__b >> fEventType;
  R__b >> fMirrorDataBits;
  R__b >> fMirrorStatus;
  R__b >> fRunNo;
  //R__b.ReadStaticArray((unsigned char*)fTriggerSource); // --> 12 bytes
  R__b.ReadFastArray((unsigned char*)fTriggerSource,6);   // 6 bytes
  fTimeStamp.Streamer(R__b);
  R__b >> fT10Mhz;
  R__b >> fOffset;
  //R__b.ReadStaticArray((unsigned int*)fMirrorTimeOffset); // --> 4 bytes
  R__b.ReadFastArray((unsigned int*)fMirrorTimeOffset,6);   // 24 bytes
  int R__i;
  for (R__i = 0; R__i < 6; R__i++)
     fRawMirrorTime[R__i].Streamer(R__b);
  //R__b.ReadStaticArray((unsigned int*)fDeadTimeCounter);
  R__b.ReadFastArray((unsigned int*)fDeadTimeCounter,2*6);
  //R__b.ReadStaticArray((unsigned int*)fVetoTimeCounter);
  R__b.ReadFastArray((unsigned int*)fVetoTimeCounter,2);
  R__b >> fT3Id;
  fT3TimeAtGround.Streamer(R__b);
}

// -------------------------------------------------------------------------

void TEyeEventHeader::StreamerV6(TBuffer &R__b)
 {
  TObject::Streamer(R__b);

  if ( TEyeEvent::fgPrintLevel ) {
    FD_CDEBUG(1) << "TEyeEventHeader::StreamerV6() - " << endl;
  }

  R__b >> fEventNumber;
  R__b >> fEyeNumber;
  R__b >> fEventClass;
  R__b >> fEventType;
  R__b >> fMirrorDataBits;
  R__b >> fMirrorStatus;
  R__b >> fRunNo;
  //R__b.ReadStaticArray((unsigned char*)fTriggerSource);
  R__b.ReadFastArray((unsigned char*)fTriggerSource,6);
  fTimeStamp.Streamer(R__b);
  R__b >> fTimeCorrectionQuality;
  R__b >> fT10Mhz;
  R__b >> fOffset;
  //R__b.ReadStaticArray((unsigned int*)fMirrorTimeOffset);
  R__b.ReadFastArray((unsigned int*)fMirrorTimeOffset,6);
  int R__i;
  for (R__i = 0; R__i < 6; R__i++)
     fRawMirrorTime[R__i].Streamer(R__b);
  //R__b.ReadStaticArray((unsigned int*)fTriggerTimeStamp);
  R__b.ReadFastArray((unsigned int*)fTriggerTimeStamp,6);
  //R__b.ReadStaticArray((unsigned int*)fDeadTimeCounter);
  R__b.ReadFastArray((unsigned int*)fDeadTimeCounter,2*6);
  //R__b.ReadStaticArray((unsigned int*)fVetoTimeCounter);
  R__b.ReadFastArray((unsigned int*)fVetoTimeCounter,2);
  R__b >> fT3Id;
  fT3TimeAtGround.Streamer(R__b);
}

// -------------------------------------------------------------------------

// Streamer for Rev. 1.44
void TEyeEventHeader::StreamerV7(TBuffer &R__b)
 {
  TObject::Streamer(R__b);

  if ( TEyeEvent::fgPrintLevel ) {
    FD_CDEBUG(1) << "TEyeEventHeader::StreamerV7() - " << endl;
  }
  
  R__b >> fEventNumber;
  R__b >> fEyeNumber;
  R__b >> fEventClass;
  R__b >> fEventType;
  R__b >> fMirrorDataBits;
  R__b >> fMirrorStatus;
  R__b >> fRunNo;
  //R__b.ReadStaticArray((unsigned char*)fTriggerSource);
  R__b.ReadFastArray((unsigned char*)fTriggerSource,6);
  R__b >> fStatusBits;
  fTimeStamp.Streamer(R__b);
  R__b >> fTimeCorrectionQuality;
  R__b >> fT10Mhz;
  R__b >> fOffset;
  //R__b.ReadStaticArray((unsigned int*)fMirrorTimeOffset);
  R__b.ReadFastArray((unsigned int*)fMirrorTimeOffset,6);
  int R__i;
  for (R__i = 0; R__i < 6; R__i++)
     fRawMirrorTime[R__i].Streamer(R__b);
  //R__b.ReadStaticArray((unsigned int*)fTriggerTimeStamp);
  R__b.ReadFastArray((unsigned int*)fTriggerTimeStamp,6);
  //R__b.ReadStaticArray((unsigned int*)fDeadTimeCounter);
  R__b.ReadFastArray((unsigned int*)fDeadTimeCounter,2*6);
  //R__b.ReadStaticArray((unsigned int*)fVetoTimeCounter);
  R__b.ReadFastArray((unsigned int*)fVetoTimeCounter,2);
  R__b >> fEyeDaqDeadTime;
  //R__b.ReadStaticArray((unsigned long long*)fMirrorDaqDeadTime);
  R__b.ReadFastArray((unsigned long long*)fMirrorDaqDeadTime,6);
  //R__b.ReadStaticArray((unsigned int*)fMirrorNSkippedTriggers);
  R__b.ReadFastArray((unsigned int*)fMirrorNSkippedTriggers,6);
  R__b >> fT3Id;
  fT3TimeAtGround.Streamer(R__b);
  R__b >> fNRejectedT3;
  R__b >> fTotalT3DeadTime;
}

// -------------------------------------------------------------------------

#if 0
// Streamer for Rev. 1.45
#endif

#if 0
// Streamer for Rev. 1.47
#endif

#if 0
// Streamer for Rev. 1.49
#endif

// -------------------------------------------------------------------------

// Streamer for Rev. 1.50
void TEyeEventHeader::StreamerV8(TBuffer &R__b)
 {
  TObject::Streamer(R__b);

  if ( TEyeEvent::fgPrintLevel ) {
    FD_CDEBUG(1) << "TEyeEventHeader::StreamerV8() - " << endl;
  }

  R__b >> fEventNumber;
  R__b >> fEyeNumber;
  R__b >> fEventClass;
  R__b >> fEventType;
  R__b >> fMirrorDataBits;
  R__b >> fMirrorStatus;
  R__b >> fRunNo;
  //R__b.ReadStaticArray((unsigned char*)fTriggerSource);
  R__b.ReadFastArray((unsigned char*)fTriggerSource,6);
  R__b >> fStatusBits;
  fTimeStamp.Streamer(R__b);
  R__b >> fTimeCorrectionQuality;
  R__b >> fT10Mhz;
  R__b >> fOffset;
  //R__b.ReadStaticArray((unsigned int*)fMirrorTimeOffset);
  R__b.ReadFastArray((unsigned int*)fMirrorTimeOffset,6);
  int R__i;
  for (R__i = 0; R__i < 6; R__i++)
     fRawMirrorTime[R__i].Streamer(R__b);
  //R__b.ReadStaticArray((unsigned int*)fTriggerTimeStamp);
  R__b.ReadFastArray((unsigned int*)fTriggerTimeStamp,6);
  //R__b.ReadStaticArray((unsigned int*)fDeadTimeCounter);
  R__b.ReadFastArray((unsigned int*)fDeadTimeCounter,2*6);
  //R__b.ReadStaticArray((unsigned int*)fVetoTimeCounter);
  R__b.ReadFastArray((unsigned int*)fVetoTimeCounter,2);
  R__b >> fEyeDaqDeadTime;
  //R__b.ReadStaticArray((unsigned long long*)fMirrorDaqDeadTime);
  R__b.ReadFastArray((unsigned long long*)fMirrorDaqDeadTime,6);
  //R__b.ReadStaticArray((unsigned int*)fMirrorNSkippedTriggers);
  R__b.ReadFastArray((unsigned int*)fMirrorNSkippedTriggers,6);
  //R__b.ReadStaticArray((int*)fMirrorEventLabel);
  R__b.ReadFastArray((int*)fMirrorEventLabel,6);
  R__b >> fT3Id;
  fT3TimeAtGround.Streamer(R__b);
  R__b >> fNRejectedT3;
  R__b >> fTotalT3DeadTime;
}

// -------------------------------------------------------------------------

#if 0
// Streamer for Rev. 1.52
void TEyeEventHeader::StreamerV9(TBuffer &R__b)
 {
  TObject::Streamer(R__b);

  if ( TEyeEvent::fgPrintLevel ) {
    FD_CDEBUG(1) << "TEyeEventHeader::StreamerV9() - " 
                 << endl;
  }

  R__b >> fEventNumber;
  R__b >> fEyeNumber;
  R__b >> fEventClass;
  R__b >> fEventType;
  R__b >> fMirrorDataBits;
  R__b >> fMirrorStatus;
  R__b >> fRunNo;
  R__b.ReadStaticArray(fTriggerSource);
  R__b >> fStatusBits;
  fTimeStamp.Streamer(R__b);
  R__b >> fTimeCorrectionQuality;
  R__b >> fT10Mhz;
  R__b >> fOffset;
  R__b.ReadStaticArray(fMirrorTimeOffset);
  int R__i;
  for (R__i = 0; R__i < 6; R__i++)
     fRawMirrorTime[R__i].Streamer(R__b);
  R__b.ReadStaticArray(fTriggerTimeStamp);
  R__b.ReadStaticArray((unsigned int*)fDeadTimeCounter);
  R__b.ReadStaticArray((unsigned int*)fVetoTimeCounter);
  R__b >> fEyeDaqDeadTime;
  R__b.ReadStaticArray(fMirrorDaqDeadTime);
  R__b.ReadStaticArray(fMirrorNSkippedTriggers);
  R__b.ReadStaticArray(fMirrorEventLabel);
  R__b >> fT3Id;
  fT3TimeAtGround.Streamer(R__b);
  R__b >> fNRejectedT3;
  R__b >> fTotalT3DeadTime;
}
#endif

// -------------------------------------------------------------------------
// -------------------------------------------------------------------------

