
//
// File   : EyeFadcData.cc
//
// Purpose: Implementation of Eyes class TEyeFADCData
//
// $Id: EyeFadcData.cc 11941 2011-01-25 13:19:58Z mathes $
//


#include <iostream>
#include <cassert>


#include <TObject.h>
#include <TObjArray.h>

/** @file EyeFadcData.cc
  * Implementation of the class TEyeFADCData.
  * @author H.-J. Mathes, FzK
  */

#include <FdNumbering.hh>
#include <ErrorLogging.hh>

#include "CheckVersion.hh"

#include "EyeFadcData.hh"

using namespace std;
using namespace FdUtil;

// -------------------------------------------------------------------------
// -------------------------------------------------------------------------
//
// class TEyeFADCData - implementation
//

TEyeFADCData::TEyeFADCData(Int_t num_fadc)
 : fStreamerVersion(0), 
   fNTraces(0), fFADCData(new TObjArray(num_fadc,Fd::kEYE_FIRST_PIXEL))
 {
#ifdef TRACE_CONSTRUCTOR
  FD_CINFO << "TEyeFADCData(): " << this << endl;
#endif // TRACE_CONSTRUCTOR
}

// -------------------------------------------------------------------------

TEyeFADCData::TEyeFADCData(const TEyeFADCData& rhs)
 : TObject(rhs), fNTraces(0), fFADCData(NULL)
 {
#ifdef TRACE_CONSTRUCTOR
  FD_CINFO << "TEyeFADCData(TEyeFADCData& " << this << "): rhs= " 
           << &rhs << endl;
#endif // TRACE_CONSTRUCTOR

  fFADCData = new TObjArray( rhs.fFADCData->GetSize(), Fd::kEYE_FIRST_PIXEL );
  *this = rhs;
}

// -------------------------------------------------------------------------

TEyeFADCData& TEyeFADCData::operator=(const TEyeFADCData& rhs)
 {
#ifdef TRACE_CONSTRUCTOR
  FD_CINFO << "TEyeFADCData::operator=(" << &rhs << " -> " << this << ")" << endl;
#endif // TRACE_CONSTRUCTOR

  if ( this == &rhs ) return *this;
  
  fNTraces = rhs.fNTraces;

  // delete all old pointers ...
  // clear destination object
#if 1
  Clear();
#else  
  for ( UInt_t i=0; i<Fd::kEYE_NEXTPIXELS; i++ ) {
    TFADCData *adc;
    if ( ( adc = (TFADCData *)fFADCData->At(i + Fd::kEYE_FIRST_PIXEL) ) != NULL ) {
      delete adc;
      fFADCData->RemoveAt( i + Fd::kEYE_FIRST_PIXEL );
    }
  }
#endif
  
  for ( Int_t i=0; i<rhs.fFADCData->GetSize(); i++ ) {
    TFADCData *adc;
    if ( ( adc = (TFADCData *)rhs.fFADCData->At(i + Fd::kEYE_FIRST_PIXEL) ) != NULL )
      fFADCData->AddAt( new TFADCData( *adc ), i + Fd::kEYE_FIRST_PIXEL );
  }
  
  return *this;
}

// -------------------------------------------------------------------------

TEyeFADCData::~TEyeFADCData()
 {
#ifdef TRACE_CONSTRUCTOR
  FD_CINFO << "~TEyeFADCData(): " << this << endl;
#endif // TRACE_CONSTRUCTOR

#if 1
  fFADCData->Delete();       // delete all objects in container

#else
  for ( Int_t i=1; i<=fFADCData->GetSize(); i++) {
    TFADCData *adc = (TFADCData *)fFADCData->At( i );
//    FD_CERR << "TEyeFADCData::~TEyeFADCData() - trying to delete TFADCData("
//            << i << ") " << hex << adc << dec << endl;
    if ( adc ) delete adc;
  }
#endif
  
  delete fFADCData;          // delete the container
}

// -------------------------------------------------------------------------

void TEyeFADCData::AddFADCData(TFADCData* fadc,bool is_owner /* false */)
 {
  // some security check: PixelNumber is in allowed range ?
  
  Fd::PixelNumber pixel = fadc->GetPixelNumber();

  if (    Fd::GetPixelNo( pixel ) < Fd::kMI_FIRST_PIXEL
       || Fd::GetPixelNo( pixel ) > Fd::kMI_LAST_EXTPIXEL ) {
    FD_CERR << "TEyeFADCData::AddFADCData() - " 
            << "tried to add illegal FADC trace of pixel "
	    << Fd::GetPixelNo( fadc->GetPixelNumber() ) << endl;
    return;
  }
  
//  FD_CERR << "TEyeFADCData::AddFADCData - "
//          << "trying to add pixel " << Fd::GetPixelNo( pixel ) 
//          << " at " << Fd::GetEyePixelNo( pixel ) << " ..." << endl;

  if ( fFADCData->At( Fd::GetEyePixelNo( pixel ) ) != NULL ) {
    FD_CWARN << "TEyeFADCData::AddFADCData() - tried to add trace for pixel "
             << Fd::GetEyePixelNo( pixel ) << " again !" << endl;
    return;
  }
  
  // check if the container is large enough (too small when streamed)
  
  if ( fFADCData->GetSize() < FdUtil::Fd::kEYE_NEXTPIXELS )
    fFADCData->Expand( FdUtil::Fd::kEYE_NEXTPIXELS );
  
  if ( is_owner )
    fFADCData->AddAt( fadc, Fd::GetEyePixelNo( pixel ) );
  else {
    fFADCData->AddAt( new TFADCData( *fadc ), Fd::GetEyePixelNo( pixel ) );
  }
  
  fNTraces++;
}

// -------------------------------------------------------------------------

void TEyeFADCData::Clear(Option_t *)
 {
  fFADCData->SetOwner( kTRUE );
  
  fFADCData->Clear();   // deletes the contained objects and 
                        // null-ifies the pointers to them
  
  fFADCData->SetOwner( kFALSE );
}

// -------------------------------------------------------------------------

const TFADCData * TEyeFADCData::GetFADCData(Fd::FdPixelNumber pixel) const
 {
  assert( pixel != NULL );
  
  // this will avoid the error message from TObjArray::At()
  
  if ( Fd::GetEyePixelNo( pixel ) > (UInt_t)fFADCData->GetSize() ) return NULL;
  
  TObject* obj = fFADCData->At( Fd::GetEyePixelNo( pixel ) );
  if ( !obj )
    return NULL;
  else {
    return (TFADCData*)obj;
  }
}

// -------------------------------------------------------------------------

const TFADCData * TEyeFADCData::GetFADCData(Fd::PixelNumber pixel) const
 {
  assert( pixel != NULL );

  // this will avoid the error message from TObjArray::At()
  
  if ( Fd::GetEyePixelNo( pixel ) > (UInt_t)fFADCData->GetSize() ) return NULL;
  
  TObject* obj = fFADCData->At( Fd::GetEyePixelNo( pixel ) );
  if ( !obj )
    return NULL;
  else {
    return (TFADCData*)obj;
  }
}

// -------------------------------------------------------------------------

void TEyeFADCData::RemoveFADCData(const TFADCData* fadc)
 {
  if ( fadc == NULL ) return;

  for ( int i=Fd::kEYE_FIRST_PIXEL; i<=fFADCData->GetSize(); i++ ) {
  
    TFADCData* fadc1 = (TFADCData*)(*fFADCData)[i];
    
    if ( fadc1 && fadc1 == fadc ) {   // pointers match
      fFADCData->RemoveAt( i );
      fNTraces--;
    }
    
    if ( fadc1 && fadc1->GetPixelNumber() == fadc->GetPixelNumber() ) {
      fFADCData->Remove( fadc1 );
      fNTraces--;
    }
  }
}

// -------------------------------------------------------------------------

void TEyeFADCData::Show(std::ostream& ostr) const
 {
  ostr << "--- TEyeFADCData ---" << endl;
  
  ostr << " Number of traces : " << fNTraces << endl;
  ostr << " Streamer version : " << fStreamerVersion << endl;
  ostr << " Class version    : " << Class_Version() << endl;
  
#if 0
  ostr << " GetEntries()     : " << fFADCData->GetEntries() << endl;
  ostr << " GetSize()        : " << fFADCData->GetSize() << endl;
  ostr << " IsEmpty()        : " << (Int_t)fFADCData->IsEmpty() << endl;
#endif
  
  TIterator* iter = GetIterator();
  TFADCData* fadc;
  
  iter->Reset();
  
  if ( (fadc = (TFADCData*)((*iter)())) != NULL && fadc != NULL ) {
    cout << " GetTraceLength() : " << fadc->GetTraceLength() << endl;
    cout << " GetNFADCBins()   : " << fadc->GetNFADCBins() << endl;
  }
  
  ostr << "--------------------" << endl;
}

ClassImp(TEyeFADCData)

// -------------------------------------------------------------------------
// -------------------------------------------------------------------------
