
//
// File   : EyeGeometryData.cc
//
// Purpose: Implementation of the class TEyeGeometryData
//
// $Id: EyeGeometryData.cc 8757 2010-09-27 12:16:09Z mathes $
//


#include <iostream>
#include <stdexcept>

/** @file EyeGeometryData.cc
  * Implementation of the class TEyeGeometryData.
  * @author H.-J. Mathes, FzK
  */


#include <FdNumbering.hh>
#include <ErrorLogging.hh>
#include <FdConfig.hh>
#include <FdGeometry.hh>

#include "CheckVersion.hh"

#include "EyeGeometryData.hh"

using namespace std;
using FdUtil::Fd;
using FdUtil::FdConfig;
using FdUtil::FdGeometry;

// -------------------------------------------------------------------------
// -------------------------------------------------------------------------
// 
// class TEyeGeometryData - implementation
// 

ClassImp(TEyeGeometryData)

const float TEyeGeometryData::kInvalidValue = 0.0;

TEyeGeometryData::TEyeGeometryData()
 {
  Reset();
}

// -------------------------------------------------------------------------

TEyeGeometryData::TEyeGeometryData(unsigned int eye_no)
 {
  if ( eye_no < Fd::kFD_FIRST_EYE || eye_no > Fd::kFD_LAST_EYE ) {
    FD_CERR << "TEyeGeometryData(int,int) - eye number " 
            << eye_no << " is outside valid range!" << endl;
    throw std::logic_error("eye number is outside valid range");
  }
  
  Reset();
  
  Fd::MirrorNumberRec mirror_num;
  
  Fd::SetEyeNo( &mirror_num, eye_no );
  
  for ( unsigned int mi_no=Fd::kEYE_FIRST_MIRROR; 
                     mi_no<=Fd::kEYE_LAST_MIRROR; mi_no++ ) {
  
    Fd::SetMirrorNo( &mirror_num, mi_no );
  
    unsigned int abs_tel_id = Fd::GetFdMirrorNo( &mirror_num );
    
    double az, el;
    FdGeometry::GetTelescopeAxis( abs_tel_id, az, el );
  
    fAxisAzimuth[mi_no - Fd::kEYE_FIRST_MIRROR] = az;
    fAxisElevation[mi_no - Fd::kEYE_FIRST_MIRROR] = el;
  }
}

// -------------------------------------------------------------------------

void TEyeGeometryData::GetAxisDirection(unsigned int tel_id,
                                        float& az,float& el)
 {
  if ( tel_id < Fd::kEYE_FIRST_MIRROR || tel_id > Fd::kEYE_LAST_MIRROR ) {
    FD_CERR << "TEyeGeometryData::GetAxisDirection() - "
            << "telescope id " << tel_id << " is outside valid range!"
	    << endl;
    az = el = kInvalidValue;
  }
  
  az = fAxisAzimuth[tel_id - Fd::kEYE_FIRST_MIRROR];
  el = fAxisElevation[tel_id - Fd::kEYE_FIRST_MIRROR];
}

// -------------------------------------------------------------------------

#if (FDUTIL_VERSION_CODE >= ModuleVersionCode(4,1,0))
void TEyeGeometryData::GetAxisDirection(unsigned int telescope_id,
                                        FdGeometry::CameraPointing_t& pointing)
 {
  float az, el;
  
  GetAxisDirection( telescope_id, az, el );
  
  pointing.fAzimuth = az;
  pointing.fElevation = el;
}
#endif // FDUTIL_VERSION_CODE

// -------------------------------------------------------------------------

void TEyeGeometryData::Reset()
 {
  for ( unsigned int i=0; i<Fd::kEYE_NMIRRORS; i++ ) {
    fAxisAzimuth[i]   = kInvalidValue;
    fAxisElevation[i] = kInvalidValue;
  }
}

// -------------------------------------------------------------------------

void TEyeGeometryData::SetAxisDirection()
 {
  // current Eye Id (use only on PAO hardware!)
  unsigned int eye_id = FdConfig::GetInstance().GetEyeId();
  
  SetAxisDirection( eye_id );
}

// -------------------------------------------------------------------------

void TEyeGeometryData::SetAxisDirection(unsigned int eye_id)
 {
  if ( eye_id < Fd::kFD_FIRST_EYE || eye_id > Fd::kFD_LAST_EYE ) {
    FD_CERR << "TEyeGeometryData::SetAxisDirection() - "
            << "eye id " << eye_id << " is outside valid range!"
	    << endl;
    throw std::logic_error("Eye id is outside valid range");
  }

  Fd::MirrorNumberRec mirrorNum;
  unsigned int abs_tel_id;
  double az, el;

  for ( unsigned int tel_id=Fd::kEYE_FIRST_MIRROR;
                     tel_id<=Fd::GetEyeLastMirrorNo( eye_id ); tel_id++ ) {
    
    Fd::SetEyeNo( &mirrorNum, eye_id );
    Fd::SetMirrorNo( &mirrorNum, tel_id );
    
    abs_tel_id = Fd::GetFdMirrorNo( &mirrorNum );
    
    FdGeometry::GetTelescopeAxis( abs_tel_id, az, el );
    
    SetAxisDirection( tel_id, az, el );
  }
}

// -------------------------------------------------------------------------

void TEyeGeometryData::SetAxisDirection(unsigned int tel_id,
                                        float az,float el)
 {
  if ( tel_id < Fd::kEYE_FIRST_MIRROR || tel_id > Fd::kEYE_LAST_MIRROR ) {
    FD_CERR << "TEyeGeometryData::SetAxisDirection() - "
            << "telescope id " << tel_id << " is outside valid range!"
	    << endl;
    az = el = kInvalidValue;
  }
  
  fAxisAzimuth[tel_id - Fd::kEYE_FIRST_MIRROR] = az;
  fAxisElevation[tel_id - Fd::kEYE_FIRST_MIRROR] = el;
}

// -------------------------------------------------------------------------

void TEyeGeometryData::Show(std::ostream& ostr,unsigned int eye_no)
 {
  ostr << "--- TEyeGeometryData ---" << endl;

  unsigned int n_mirrors;
  
  if ( eye_no < 5 )
    n_mirrors = Fd::kEYE_NMIRRORS;
  else
    n_mirrors = Fd::kHEAT_NMIRRORS;
  
  for ( unsigned int i=0; i<n_mirrors; i++ ) {
    ostr << " (az,el)[" << i + Fd::kEYE_FIRST_MIRROR << "]=      (" 
         << fAxisAzimuth[i] << "," << fAxisElevation[i] << ")" << endl;
  }

  ostr << "------------------------" << endl;
}

// -------------------------------------------------------------------------
// -------------------------------------------------------------------------
