

//
// File   : EyePixelData.cc
//
// Purpose: Implementation of Eyes class TEyePixelData
//
// $Id: EyePixelData.cc 8757 2010-09-27 12:16:09Z mathes $
//


#include <iostream>


/** @file EyePixelData.cc
  * Implementation of the class TEyePixelData.
  * @author H.-J. Mathes, FzK
  */

#include <FdNumbering.hh>
#include <ErrorLogging.hh>

#include "CheckVersion.hh"

#include "EyeEvent.hh"
#include "EyePixelData.hh"

using namespace std;
using namespace FdUtil;

// -------------------------------------------------------------------------
// -------------------------------------------------------------------------
//
// class TEyePixelData - implementation
//

const int TEyePixelData::kNoTrigger = -1;

TEyePixelData::TEyePixelData()
 : fStreamerVersion(0)
 {
#ifdef TRACE_CONSTRUCTOR
  FD_CINFO << "TEyePixelData(): " << this << endl;
#endif // TRACE_CONSTRUCTOR

  for ( unsigned int i=0; i<Fd::kEYE_NMIRRORS; i++ )
    fPixelData[i] = NULL;
}

// -------------------------------------------------------------------------

TEyePixelData::TEyePixelData(const TEyePixelData& rhs)
 : TObject(rhs)
 {
#ifdef TRACE_CONSTRUCTOR
  FD_CINFO << "TEyePixelData(TEyePixelData&): " << this << endl;
#endif // TRACE_CONSTRUCTOR

  *this = rhs;
}

// -------------------------------------------------------------------------

TEyePixelData& TEyePixelData::operator=(const TEyePixelData& rhs)
 {
#ifdef TRACE_CONSTRUCTOR
  FD_CINFO << "TEyePixelData::operator=(" << &rhs 
           << " -> " << this << ")" << endl;
#endif // TRACE_CONSTRUCTOR

  if ( this == &rhs ) return *this;

  for ( unsigned int i=0; i<Fd::kEYE_NMIRRORS; i++ ) {
    if ( rhs.fPixelData[i] == NULL ) {
      delete fPixelData[i];
      fPixelData[i] = NULL;
    }
    else {  // rhs.fPixelData[i] != NULL
      if ( fPixelData[i] )
        *fPixelData[i] = *rhs.fPixelData[i];
      else
        fPixelData[i] = new TMirrorPixelData( *rhs.fPixelData[i] );
    }
  }
  
  return *this;
}

// -------------------------------------------------------------------------

TEyePixelData::~TEyePixelData()
 {
#ifdef TRACE_CONSTRUCTOR
  FD_CINFO << "~TEyePixelData(): " << this << endl;
#endif // TRACE_CONSTRUCTOR

  for ( unsigned int i=0; i<Fd::kEYE_NMIRRORS; i++ )
    delete fPixelData[i];
}

// -------------------------------------------------------------------------

bool TEyePixelData::AddPixelData(TMirrorPixelData* pixel_data, 
                                 unsigned int mirror_id,bool is_owner)
 {
  // --- check the passed Mirror Id (1..6)
  
  if ( mirror_id < Fd::kEYE_FIRST_MIRROR || mirror_id > Fd::kEYE_NMIRRORS ) {
    FD_CERR << "TEyePixelData::AddPixelData() - "
  	    << "trying to add data for illegal mirror id !" << endl;
    return false;
  }
  
  // --- cross-check the passed Mirror Id with stored one (in TMirrorPixelData)
  
  // ...
  
  // --- check if already added and ignore/replace it ?
    
  if ( fPixelData[mirror_id-Fd::kEYE_FIRST_MIRROR] ) {
    
    FD_CWARN << "TEyePixelData::AddPixelData() - "
  	     << "pixel data for Mirror Id " << mirror_id
             << " has been already added, ignored!"
             << endl;
    
    // THJM: am I allowed to delete it ? (not used)
    if ( is_owner ) delete pixel_data;

#if 1
    return false;
#else
    delete fPixelData[mirror_id-Fd::kEYE_FIRST_MIRROR];
    fPixelData[mirror_id-Fd::kEYE_FIRST_MIRROR] = NULL;
#endif
  }
    
  if ( is_owner ) 
    fPixelData[mirror_id-Fd::kEYE_FIRST_MIRROR] = pixel_data;
  else {
    TMirrorPixelData* new_pixel_data = new TMirrorPixelData();
    *new_pixel_data = *pixel_data;
    fPixelData[mirror_id-Fd::kEYE_FIRST_MIRROR] = new_pixel_data;
  }
  
  return true; // = successfully added
}

// -------------------------------------------------------------------------

void TEyePixelData::Clear(Option_t *)
 {
  for ( unsigned int i=0; i<Fd::kEYE_NMIRRORS; i++ ) {
    delete fPixelData[i];
    fPixelData[i] = 0;
  }
}

// -------------------------------------------------------------------------

int TEyePixelData::GetFirstTriggeredTimeBin(Fd::FdPixelNumber fdpixel,
                                            unsigned int first_bin,
                                            unsigned int last_bin) const
 {
  unsigned int mirror_no = Fd::GetEyeMirrorNo( fdpixel );
  
  if ( !fPixelData[mirror_no - Fd::kEYE_FIRST_MIRROR] ) return kNoTrigger;
  
  return fPixelData[mirror_no - Fd::kEYE_FIRST_MIRROR]->GetFirstTriggeredTimeBin(
                                                 Fd::GetPixelNumber(fdpixel),
                                                 first_bin,last_bin );
}

// -------------------------------------------------------------------------

unsigned int
TEyePixelData::GetNumberOfTriggeredPixels(unsigned int start_slice,
                                          unsigned int end_slice) const
 {
  unsigned int n_triggered_pixels = 0;
  
  for ( unsigned int id=Fd::kEYE_FIRST_MIRROR;
                     id <= Fd::kEYE_LAST_MIRROR; id ++ )
    n_triggered_pixels += GetNumberOfTriggeredPixels( id, start_slice, end_slice );
    
  return n_triggered_pixels;
}

// -------------------------------------------------------------------------

unsigned int 
TEyePixelData::GetNumberOfTriggeredPixels(unsigned int mirror_id,
                                          unsigned int start_slice,
                                          unsigned int end_slice) const
 {
  TMirrorPixelData *pd;

  if ( ( pd = GetPixelData( mirror_id ) ) == NULL )
    return 0;
  else
    return pd->GetNumberOfTriggeredPixels( start_slice, end_slice );
}

// -------------------------------------------------------------------------

TMirrorPixelData* TEyePixelData::GetPixelData(unsigned int mirror_id) const
 {
  if ( mirror_id < Fd::kEYE_FIRST_MIRROR || mirror_id > Fd::kEYE_NMIRRORS ) {
    FD_CERR << "TEyePixelData::GetPixelData(" << mirror_id << ") - "
            << "illegal Mirror id passed !" << endl;
    return NULL;
  }

  return fPixelData[mirror_id - Fd::kEYE_FIRST_MIRROR];
}

// -------------------------------------------------------------------------

void TEyePixelData::Show(ostream& ostr)
 {
  ostr << "--- TEyePixelData ---" << endl;
  
  ostr << " Streamer version: " << fStreamerVersion << endl;
  ostr << " Class version   : " << Class_Version() << endl;
  
  for ( unsigned int id=0; id<Fd::kEYE_NMIRRORS; id++ ) {
    if ( fPixelData[id] ) {
      ostr << "\tFound pixel data for Mirror "
           << id + Fd::kEYE_FIRST_MIRROR 
	   << ", telescope id= " 
	   << Fd::GetFdMirrorNo( fPixelData[id]->GetMirrorNumber() )
	   << " (probably wrong!)" << endl;
    
      fPixelData[id]->Show();
    }
  }
}

ClassImp(TEyePixelData)

// -------------------------------------------------------------------------
// -------------------------------------------------------------------------
