
//
// File  : EyeRunHeader.cc
//
// Purpose: Declaration file for class %Class_name%
//
// $Id: EyeRunHeader.cc 8757 2010-09-27 12:16:09Z mathes $
//

#include <iostream>
#include <sstream>

/** @file EyeRunHeader.cc
  * Implementation for class TEyeRunHeader.
  * @author H.-J. Mathes, FzK
  */

#include <TTimeStamp.hh>
#include <FdConfig.hh>
#include <FdNumbering.hh>

#include "CheckVersion.hh"

#include "EyeRunHeader.hh"

using namespace FdUtil;
using namespace std;

// -------------------------------------------------------------------------
// -------------------------------------------------------------------------
//
// class TEyeRunHeader - implementation
//

TEyeRunHeader::TEyeRunHeader()
 {
  Clear();
}

// -------------------------------------------------------------------------

TEyeRunHeader& TEyeRunHeader::operator=(const TEyeRunHeader& rhs)
 {
  if ( this == &rhs ) return *this; 

  // now copy all data members, TObject stuff is not copied !
  
  fEyeNumber        = rhs.fEyeNumber;
  fMirrorStatus     = rhs.fMirrorStatus;
  fRunNo            = rhs.fRunNo;
  
  fStartTime        = rhs.fStartTime;

  fUserName         = TString( rhs.fUserName );
  fComment          = TString( rhs.fComment );
  
  fEvbVersion       = rhs.fEvbVersion;
  fEyeEventVersion  = rhs.fEyeEventVersion;
  fMiEventVersion   = rhs.fMiEventVersion;
  fReadoutVersion   = rhs.fReadoutVersion;
  
  return *this;
}

// -------------------------------------------------------------------------

TEyeRunHeader::~TEyeRunHeader()
 { }

// -------------------------------------------------------------------------

void TEyeRunHeader::Clear(Option_t *)
 {
  fEyeNumber = FdUtil::GetEyeId();
  fMirrorStatus = 0;
  fRunNo = 0;
  
  FdRoot::TTimeStamp t((time_t)0,0);
  fStartTime = t;
  
  fEvbVersion      = 0;
#ifdef NO_VERSION
  fMiEventVersion  = 0; 
  fEyeEventVersion = 0;
#else
  fEyeEventVersion = EYEEVENT_VERSION_CODE;
  fMiEventVersion = MIEVENT_VERSION_CODE;
#endif // NO_VERSION
  fReadoutVersion  = 0; 

  fUserName = TString( "default" );
  fComment  = TString( "not set" );
}

// -------------------------------------------------------------------------

const string TEyeRunHeader::GetVersionCodeVerbosely(UInt_t vcode)
 {
  if ( vcode == 0 ) return "'unknown'";
  
  ostringstream verbose;
  
  // ignore leading byte
  
  verbose << ((vcode & 0x00ff0000) >> 16) << ".";
  verbose << ((vcode & 0x0000ff00) >> 8) << ".";
  verbose << (vcode & 0x000000ff);
  
  return verbose.str();
}

// -------------------------------------------------------------------------

Bool_t TEyeRunHeader::IsMirrorPresent(UInt_t mirror_id)
 {
  if ( mirror_id < Fd::kEYE_FIRST_MIRROR && mirror_id > Fd::kEYE_LAST_MIRROR )
    return kFALSE;
  if ( fMirrorStatus & ( 1 << (mirror_id - Fd::kEYE_FIRST_MIRROR) ) )
    return kTRUE;
  else
    return kTRUE;
}

// -------------------------------------------------------------------------

void TEyeRunHeader::SetMirrorPresent(UInt_t mirror_id)
 {
  if ( mirror_id >= Fd::kEYE_FIRST_MIRROR && mirror_id <= Fd::kEYE_LAST_MIRROR )
    fMirrorStatus |= ( 1 << (mirror_id - Fd::kEYE_FIRST_MIRROR) );
}

// -------------------------------------------------------------------------

void TEyeRunHeader::SetVersionInformation()
 {
  fReadoutVersion  = 0;  // unspecified
  fEvbVersion      = 0;  // unspecified
#ifndef NO_VERSION
  fMiEventVersion  = MIEVENT_VERSION_CODE;
  fEyeEventVersion = EYEEVENT_VERSION_CODE;
#else
  // in case software was built without version information ...
  fMiEventVersion  = 0;  // unspecified
  fEyeEventVersion = 0;  // unspecified
#endif // NO_VERSION
}

// -------------------------------------------------------------------------

#define VERSION(_v)          (UInt_t)((_v & 0xff0000) >> 16)
#define SUB_VERSION(_v)      (UInt_t)((_v & 0xff00) >> 8)
#define SUB_SUB_VERSION(_v)  (UInt_t)(_v & 0xff)

void TEyeRunHeader::Show(std::ostream& ostr)
 {
  ostr << "--- TEyeRunHeader for Eye " << (Int_t)fEyeNumber << " ---" << endl;

  ostr << " RUN started at  : " << fStartTime.AsString() << endl;
  ostr << " RUN id          : " << fRunNo << endl;
  ostr << " Mirrors present : ";
  for ( UInt_t i=0; i<Fd::kEYE_NMIRRORS; i++ )
    if ( fMirrorStatus & (1<<i) ) ostr << i + 1 << " ";
  ostr << endl;
  
  ostr << " Users name      : " << fUserName << endl;
  ostr << " RUN comment     : " << fComment << endl;
  
  ostr << " Software versions used: " << endl;
  ostr << "  - MiEvent " << GetVersionCodeVerbosely( fMiEventVersion ) << endl;
  ostr << "  - EyeEvent " << GetVersionCodeVerbosely( fEyeEventVersion ) << endl;
  ostr << "  - MiReadout " << GetVersionCodeVerbosely( fReadoutVersion ) << endl;
  ostr << "  - Evb/MiApps " << GetVersionCodeVerbosely( fEvbVersion ) << endl;
}

ClassImp(TEyeRunHeader)

// -------------------------------------------------------------------------
// -------------------------------------------------------------------------
