
//
// File   : FadcData.cc
//
// Purpose: Implementation of Pixel FADC data class TFADCData
//
// $Id: FadcData.cc 8759 2010-09-27 12:16:36Z mathes $
//


#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <iostream>
#include <stdexcept>
#include <cassert>

#include <TClass.h>
#include <TBuffer.h>


/** @file FadcData.cc
  * Implementation of class TFADCData.
  * @author H.-J. Mathes, FzK
  */

#include <ErrorLogging.hh>
#include <FdNumbering.hh>
#include <FdConfig.hh>

#include "CheckVersion.hh"

#ifdef __ppc__
# include "PPCData.hh"
#endif // __ppc__
#include "MiEvent.hh"

using namespace std;
using FdUtil::Fd;

// -------------------------------------------------------------------------
// -------------------------------------------------------------------------
//
// class TFADCData - implementation
//

TFADCData::EFADCTraceResolution TFADCData::fgDefaultResolution = 
                                      TFADCData::kFADC_DEFAULT_RESOLUTION;

TFADCData::TFADCData(Fd::PixelNumber pixel)
 : fConstructorType(kCREATE), fFADCData(new FADCDataRec)
 {
  assert( fFADCData != NULL );
  
  Init();  // sets default pixel number
#ifdef TRACE_CONSTRUCTOR
  FD_CINFO << "TFADCData(PixelNumber): " << this << endl;
#endif
  if ( pixel ) SetPixelNumber( pixel );
}

// -------------------------------------------------------------------------

TFADCData::TFADCData(FADCData fadc_data)
 : fConstructorType( kFADCDATA )
 {
#ifdef TRACE_CONSTRUCTOR
  FD_CINFO << "TFADCData(FADCData): " << this << endl;
#endif
  fFADCData = fadc_data;
}

// -------------------------------------------------------------------------

TFADCData::TFADCData(const TFADCData& rhs)
 : TObject(rhs),
   fConstructorType(kCREATE), fFADCData(new FADCDataRec)
 {
#ifdef TRACE_CONSTRUCTOR
  FD_CINFO << "TFADCData(TFADCData&): " << this << endl;
#endif
  //Init();
  
  *this = rhs;
}

// -------------------------------------------------------------------------

TFADCData& TFADCData::operator=(const TFADCData& rhs)
 {
#ifdef TRACE_CONSTRUCTOR
  FD_CINFO << "TFADCData::operator=(" << &rhs << " -> " << this << ")" << endl;
  FD_CINFO << GetVerboseConstructorType( rhs.fConstructorType ) << " -> "
           << GetVerboseConstructorType( this->fConstructorType ) << endl;
#endif // TRACE_CONSTRUCTOR

  if ( this == &rhs ) return *this;
  
  // dangerous if different sizes for fFADCData are used !
  memcpy( this->fFADCData, rhs.fFADCData, sizeof( *rhs.fFADCData ) );

  return *this;
}

// -------------------------------------------------------------------------

TFADCData::~TFADCData()
 {
#ifdef TRACE_CONSTRUCTOR
  FD_CINFO << "~TFADCData('" << GetVerboseConstructorType( fConstructorType ) 
           << "'): " << this << endl;
#endif

  switch ( fConstructorType ) {
  
    case kCREATE:
         delete fFADCData;
         break;

    case kFADCDATA:
         fFADCData = NULL;
         break;

    default:
         FD_CERR << "TFADCData::~TFADCData() - unknown fConstructorType= " 
	      << (int)fConstructorType << endl;
  }
}

// -------------------------------------------------------------------------

void TFADCData::Clear(Option_t* /* option */)
 {
  Init();
  
  memset( fFADCData->fFadcData, 0x00, 
          GetNFADCBins( (EFADCTraceResolution)fFADCData->fTraceInfo.fBits.fResolution ) * sizeof( FADCDataWordRec ) );
  //memset( fFADCData->fFadcData, 0x00, 
  //        kMI_NFADC_BINS_HIGH * sizeof( FADCDataWordRec ) );
}

// -------------------------------------------------------------------------

bool TFADCData::ConvertFromV3(const TFADCData* src)
 {
  if ( !src ) {
    FD_CERR << "TFADCData::ConvertFromV3() - NULL pointer passed!" << endl;
    return false;
  }
  
  return ConvertToV4( src, this );
}
    
// -------------------------------------------------------------------------

bool TFADCData::ConvertToV3(const TFADCData* src,TFADCData* dest)
 {
  if ( src == NULL || dest == NULL ) {
    FD_CERR << "TFADCData::ConvertToV3() - NULL pointer passed!" << endl;
    return false;
  }

  if ( src == dest ) {
    FD_CERR << "TFADCData::ConvertToV3() - in place conversion not possible!" << endl;
    return false;
  }

  unsigned int n_src_bins = src->GetNFADCBins();

  if ( n_src_bins == TFADCData::GetNFADCBins( kFADC_RESOLUTION_1000_BINS ) ) {
    FD_CWARN << "TFADCData::ConvertToV3() - "
             << "data of pixel " << Fd::GetPixelNo( src->GetPixelNumber() ) 
             << " already with " << n_src_bins << " FADC bins," 
	     << " just copied!" << endl;
    *dest = *src;
    return true;
  }
  
  unsigned int factor = 
        GetNFADCBins( TFADCData::kFADC_RESOLUTION_1000_BINS ) / n_src_bins;

  dest->SetTraceResolution( TFADCData::kFADC_RESOLUTION_1000_BINS );
  
  dest->SetPixelNumber( src->GetPixelNumber() );
  dest->SetPixelMonitorData( src->GetPixelMonitorData() );
  
  //dest->SetTraceStartTime( src->GetTraceStartTime() );
  dest->SetTraceStartBin( src->GetTraceStartBin() / factor );
  dest->SetTraceEndBin( src->GetTraceEndBin() / factor );

  // loop over FADC bins
  
  TFADCData::FADCDataWord src_data = src->GetFADCTrace();
  TFADCData::FADCDataWord dest_data = dest->GetFADCTrace();
  
  unsigned short data;
  bool threshold_has_changed, has_pixeltrigger;
  
  for ( unsigned int n=0; n<Fd::kMI_NFADC_BINS_LOW; n++ ) {
    
    data = 0;
    threshold_has_changed = has_pixeltrigger = false;
    
    // get only the 12 bits but can use all 16 bits for calculation
    for ( unsigned int i=0; i<factor; i++ ) {
      data += FADCDataWordGetData( &src_data[n + i] );
      if ( FADCDataWordHasThresholdChanged( &src_data[n + i] ) )
        threshold_has_changed = true;
      if ( FADCDataWordHasPixelTrigger( &src_data[n + i] ) )
        has_pixeltrigger = true;
    }
    data /= factor;
    
    // out-of-range is a V3 feature only
    if ( data >= 4095 ) {
      //FADCDataWordSetOutOfRange( &dest_data[n] );
      data = 4095;
    }
    
    FADCDataWordSetData( &dest_data[n], data );
    
    if ( threshold_has_changed ) 
      FADCDataWordSetThresholdChanged( &dest_data[n] );
    if ( has_pixeltrigger )
      FADCDataWordSetPixelTrigger( &dest_data[n] );
  }
  
  return true;
}

// -------------------------------------------------------------------------

bool TFADCData::ConvertToV4(const TFADCData* src,TFADCData* dest)
 {
  if ( src == NULL || dest == NULL ) {
    FD_CERR << "TFADCData::ConvertToV4() - NULL pointer passed!" << endl;
    return false;
  }
  
  if ( src == dest ) {
    FD_CERR << "TFADCData::ConvertToV4() - in place conversion not possible!" << endl;
    return false;
  }
  
  unsigned int n_src_bins = src->GetNFADCBins();
  
  if ( n_src_bins == TFADCData::GetNFADCBins( TFADCData::kFADC_RESOLUTION_2000_BINS ) ) {
    FD_CWARN << "TFADCData::ConvertToV4() - "
             << "data of pixel " << Fd::GetPixelNo( src->GetPixelNumber() ) 
             << " already with " << n_src_bins << " FADC bins," 
	     << " just copied!" << endl;
    *dest = *src;
    return true;
  }
  else if ( n_src_bins != TFADCData::GetNFADCBins( TFADCData::kFADC_RESOLUTION_1000_BINS ) ) {
    FD_CERR << "TFADCData::ConvertToV4() - "
            << "data of pixel " << Fd::GetPixelNo( src->GetPixelNumber() ) 
            << "  with " << n_src_bins << " FADC bins," 
	     << " cannot be converted!" << endl;
    return false;
  }
  
  dest->SetTraceResolution( TFADCData::kFADC_RESOLUTION_2000_BINS );
  
  dest->SetPixelNumber( src->GetPixelNumber() );
  dest->SetPixelMonitorData( src->GetPixelMonitorData() );
  
  //dest->SetTraceStartTime( src->GetTraceStartTime() );
  dest->SetTraceStartBin( src->GetTraceStartBin() * 2 );
  if ( dest->GetTraceEndBin() == (Fd::kMI_NFADC_BINS_LOW - 1) )
    dest->SetTraceEndBin( Fd::kMI_NFADC_BINS_MED - 1 );
  else
    dest->SetTraceEndBin( src->GetTraceEndBin() * 2 + 1 );

  // loop over FADC bins
  
  TFADCData::FADCDataWord src_data = src->GetFADCTrace();
  TFADCData::FADCDataWord dest_data = dest->GetFADCTrace();
  
  for ( unsigned int j=0; j<n_src_bins; j++ ) { // 1000
    
    dest_data[2*j] = dest_data[2*j+1] = src_data[j];
    
    // clear the unused bits (not portable)
    dest_data[2*j].fBits.fUnused1 = dest_data[2*j+1].fBits.fUnused1 = 0;
    dest_data[2*j].fBits.fUnused2 = dest_data[2*j+1].fBits.fUnused2 = 0;
  }

  return true;
}

// -------------------------------------------------------------------------

void TFADCData::ConvertToV4(const FADCTraceInfoV3 src,FADCTraceInfo dest)
 {
  dest->fBits.fStartBin   = src->fBits.fStartBin;
  dest->fBits.fEndBin     = src->fBits.fEndBin;
  dest->fBits.fResolution = kFADC_RESOLUTION_1000_BINS;
}

// -------------------------------------------------------------------------

void TFADCData::ConvertToV4(const PixelMonitorDataV3 src,PixelMonitorData dest)
 {
   dest->fPrevThreshold   = src->fPrevThreshold;
   dest->fActualThreshold = src->fActualThreshold;
   dest->fActualRate      = src->fActualRate;
   dest->fMean            = src->fMean;
   dest->fVariance        = src->fVariance;
}

// -------------------------------------------------------------------------

TFADCData::EFADCBinWidth 
TFADCData::GetBinWidth(TFADCData::EFADCTraceResolution resolution)
 {
  switch ( resolution ) {
    case kFADC_RESOLUTION_1000_BINS:
         return kFADC_BINWIDTH_100NS;
    case kFADC_RESOLUTION_2000_BINS:
         return kFADC_BINWIDTH_50NS;
    case kFADC_RESOLUTION_4000_BINS:
         return kFADC_BINWIDTH_25NS;
    default:
         FD_CERR << "TFADCData::GetBinWidth() - "
	         << "called with illegal resolution " << (int)resolution
		 << " !" << endl;
         throw std::logic_error("TFADCData::GetBinWidth() - called with illegal resolution!");
  }
  
  return GetBinWidth( kFADC_DEFAULT_RESOLUTION );
}

// -------------------------------------------------------------------------

unsigned int TFADCData::GetNFADCBins(EFADCTraceResolution res)
 {
  switch ( res ) {
    
    case kFADC_RESOLUTION_1000_BINS:
         return Fd::kMI_NFADC_BINS_LOW;
    
    case kFADC_RESOLUTION_2000_BINS:
         return Fd::kMI_NFADC_BINS_MED;
    
    case kFADC_RESOLUTION_4000_BINS:
         return Fd::kMI_NFADC_BINS_HIGH;
    
    default:
         FD_CERR << "TFADCData::GetNFADCBins() - "
	        << "illegal value " << (int)res 
		<< " for 'EFADCTraceResolution' is used!" << endl;
         return Fd::kMI_NFADC_BINS_HIGH;
  }
}

// -------------------------------------------------------------------------

TFADCData::EFADCTraceResolution 
TFADCData::GetTraceResolution(unsigned int nbins)
 {
  TFADCData::EFADCTraceResolution res;
  
  if ( nbins <= Fd::kMI_NFADC_BINS_LOW )
    res = kFADC_RESOLUTION_1000_BINS;
  else if ( nbins <= Fd::kMI_NFADC_BINS_MED )
    res = kFADC_RESOLUTION_2000_BINS;
  else if ( nbins <= Fd::kMI_NFADC_BINS_HIGH )
    res = kFADC_RESOLUTION_4000_BINS;
  else {
    FD_CERR << "TFADCData::GetTraceResolution() - "
    	    << "illegal value " << nbins 
	    << " is used for the number of FADC bins!" << endl;
    return TFADCData::kFADC_DEFAULT_RESOLUTION;
  }
  
  unsigned int nbins_cooked = GetNFADCBins( res );
  
  if ( nbins != nbins_cooked ) {
    FD_CWARN << "TFADCData::GetTraceResolution() - "
             << "passed value for NFADC bins of " << nbins
	     << " must be corrected to " << nbins_cooked
	     << " !" << endl;
  }
  
  return res;
}

// -------------------------------------------------------------------------

const char *TFADCData::GetVerboseConstructorType(enum Constructor_t ct)
 {
  static const char *ct_str;

  switch ( ct ) {
    case kCREATE:
         ct_str = "kCREATE";
	 break;
    case kFADCDATA:
         ct_str = "kFADCDATA";
	 break;
    default:
         ct_str = "unknown";
  }

  return ct_str;
}

// -------------------------------------------------------------------------

void TFADCData::Init()
 {
#if 1
  int column = Fd::kMI_FIRST_COL, row = Fd::kMI_FIRST_ROW;
  Fd::PixelNumberRec pixel;
  
  Fd::SetEyeNo( &pixel, FdUtil::GetEyeId() );
  Fd::SetMirrorNo( &pixel, FdUtil::GetMirrorId() );

  Fd::SetPixelNoFromColumnRow( &pixel, column, row );
  
  Init( fFADCData, &pixel );
#else
  Init( fFADCData );
#endif
}

// -------------------------------------------------------------------------

void TFADCData::Init(TFADCData::FADCData fadc,FdUtil::Fd::PixelNumber pixel_in)
 {
  assert( fadc != NULL );
  
  Fd::PixelNumberRec pixel;
  
  if ( pixel_in == NULL ) {
    Fd::SetEyeNo( &pixel, FdUtil::GetEyeId() );
    Fd::SetMirrorNo( &pixel, FdUtil::GetMirrorId() );

    Fd::SetPixelNoFromColumnRow( &pixel, 0 /* column */, 0 /* row */ );
  }
  else
    pixel = *pixel_in;
  
  FADCDataSetPixelNumber( fadc, &pixel );
	
  fadc->fTraceInfo.fBits.fStartBin   = 0;
  fadc->fTraceInfo.fBits.fEndBin     = GetNFADCBins( fgDefaultResolution ) - 1;
  fadc->fTraceInfo.fBits.fResolution = fgDefaultResolution;

  PixelMonitorData monitor = FADCDataGetPixelMonitorData( fadc );
	
  PixelMonitorDataSetPrevThr( monitor, 0 );
  PixelMonitorDataSetActThr( monitor, 0 );
  PixelMonitorDataSetActRate( monitor, 0 );
	
  PixelMonitorDataSetMean( monitor, 0.0 );
  PixelMonitorDataSetVariance( monitor, 0.0 );
  
  fadc->fTraceStartTime = 0.0;
}

// -------------------------------------------------------------------------

void TFADCData::SetDefaultResolution(EFADCTraceResolution res)
 {
  unsigned int n_fadc_bins = GetNFADCBins( res );
  
  if ( n_fadc_bins > Fd::kMI_NFADC_BINS_DEFAULT )
    throw std::logic_error("Number of FADC bins too large");

  fgDefaultResolution = res;
}

// -------------------------------------------------------------------------

void TFADCData::SetTraceResolution(EFADCTraceResolution res)
 {
  unsigned int n_fadc_bins = GetNFADCBins( res );
  
  if ( n_fadc_bins > Fd::kMI_NFADC_BINS_DEFAULT )
    throw std::logic_error("Number of FADC bins too large");

  fFADCData->fTraceInfo.fBits.fEndBin = 
             min( fFADCData->fTraceInfo.fBits.fEndBin, n_fadc_bins - 1 );
  fFADCData->fTraceInfo.fBits.fResolution = res;
}

// -------------------------------------------------------------------------

void TFADCData::Show(TFADCData::FADCDataWord data,std::ostream& ostr)
 {
  ostr << hex << "0x" << setw(4) << setfill('0') << FADCDataWordGetWord( data ) 
       << setfill(' ') << dec 
       << " fadc: " << setw(4) << FADCDataWordGetData( data )
       << " PT: " << FADCDataWordGetPixelTrigger( data )
       << " TC: " << FADCDataWordGetThresholdChanged( data );

#ifdef HARDWARE_V3
  // special treatment for V3 data type --> not possible
  if ( dynamic_cast<TFADCData::FADCDataWordV3>(data) != NULL ) {
    
    TFADCData::FADCDataWordV3 data3 = dynamic_cast<TFADCData::FADCDataWordV3>(data);
    
    ostr << " OV: " << FADCDataWordGetOutOfRange( data3 )
         << " OV: " << FADCDataWordGetPixelTriggerStatus( data3 );
  }
#endif // HARDWARE_V3

  ostr << endl;
}

// -------------------------------------------------------------------------

void TFADCData::Show(std::ostream& ostr) const
 {
#ifdef TRACE_CONSTRUCTOR
  FD_CINFO << "TFADCData::Show(" << this << "):" << endl;
#endif

  Fd::PixelNumber pixel = GetPixelNumber();
  
  ostr << "--- TFADCData(" << 
       Fd::GetFdPixelNo( pixel ) << ") ---" << endl;
  
  ostr << " Pixel info: " << endl;
  ostr << "  (Eye,Mi) = (" << Fd::GetEyeNo( pixel ) << "," 
                           << Fd::GetMirrorNo( pixel ) << ")" << endl;
  ostr << "  PixelNo  = " << setw(3) << Fd::GetPixelNo( pixel ) << endl;
  
  ostr << " Trace info: " << endl;
  ostr << "  start bin = " << GetTraceStartBin() << endl;
  ostr << "  end bin   = " << GetTraceEndBin() << endl;
  ostr << "  length    = " << GetTraceLength() << endl;
  //ostr << "  resolution= " << GetTraceResolution() << endl;
  ostr << "  nbins     = " << GetNFADCBins() << endl;
  
  ostr << " Monitor data: " << endl;
  ostr << "  mean            = " << GetMean() << endl;
  ostr << "  variance        = " << GetVariance() << endl;
  
  if ( Fd::GetPixelNo( pixel ) > Fd::kMI_LAST_PIXEL )
    ostr << "  = virtual channel (VC)" << endl;
  else {
    ostr << "  rate            = " << GetActualRate() << endl;
    ostr << "  curr. threshold = " << GetActualThreshold() << endl;
    ostr << "  prev. threshold = " << GetPreviousThreshold() << endl;
  }
  
  TFADCData::FADCDataWordRec data_s, data_e;
  
  data_s = GetFADCTrace()[GetTraceStartBin()];
  data_e = GetFADCTrace()[GetTraceEndBin()];

  ostr << " FADC data: " << endl;
  ostr << "  data[" << setw(4) << GetTraceStartBin() << "]      = ";
  Show( &data_s );
  ostr << "  data[" << setw(4) << GetTraceEndBin() << "]      = ";
  Show( &data_e );
}

// -------------------------------------------------------------------------

/** Stream a TFADCData object. */
void TFADCData::Streamer(TBuffer &R__b)
 {
  unsigned int R__s, R__c;
  
  if ( R__b.IsReading() ) { 
    
    Version_t R__v = R__b.ReadVersion(&R__s, &R__c);

    if ( TMirrorEvent::GetPrintLevel() ) {
      FD_CDEBUG(1) << "TFADCData::Streamer() - R__v= " << R__v
                   << " R__s= " << R__s << " R__c= " << R__c << endl;
    }
    
    TObject::Streamer( R__b );
    
    if ( R__v > MiEVENTVERSIONv5 ) {
      FD_CERR << "Version mismatch in TFADCData::Streamer()!" << endl;
      FD_CERR << "Current version: " << R__v << ", max. accepted: " 
              << MiEVENTVERSIONv5 << endl;
      return;
    }
    
    R__b >> fFADCData->fPixelNumber;
    
    if ( TMirrorEvent::GetPrintLevel() > 1 ) {
      FD_CDEBUG(1) << "PixelNumber.fPixelNo = "
                   << Fd::GetPixelNo( &fFADCData->fPixelNumber ) << endl;
      FD_CDEBUG(1) << "PixelNumber.fEyeNo = "
                   << Fd::GetEyeNo( &fFADCData->fPixelNumber ) << " "
		   << "*.fMirrorNo = "
		   << Fd::GetMirrorNo( &fFADCData->fPixelNumber ) << endl;
    }
    
    // --- deal with the reading of FADCTRaceInfoRec
    
    if ( R__v >= MiEVENTVERSIONv4 ) {
      R__b >> fFADCData->fTraceInfo.fWord;
#ifdef __ppc__
      File2PPCSwapFADCTraceInfoRecV4( fFADCData->fTraceInfo );
#endif // __ppc__
    }
    else if ( R__v >= MiEVENTVERSIONv2 && R__v <= MiEVENTVERSIONv3 ) {
      FADCTraceInfoV3Rec traceinfo_V3;
      R__b >> traceinfo_V3.fWord;
#ifdef __ppc__
      File2PPCFADCTraceInfoRecV3( traceinfo_V3 );
#endif // __ppc__
      // conversion V3 -> V4 data
      fFADCData->fTraceInfo.fBits.fStartBin   = traceinfo_V3.fBits.fStartBin;
      fFADCData->fTraceInfo.fBits.fEndBin     = traceinfo_V3.fBits.fEndBin;
      fFADCData->fTraceInfo.fBits.fResolution = kFADC_RESOLUTION_1000_BINS;
    }
    else { // v1 data: no FADCTraceInfo yet
      fFADCData->fTraceStartTime              = 0.0;
      fFADCData->fTraceInfo.fBits.fStartBin   = 0;
      // set resolution etc.
      fFADCData->fTraceInfo.fBits.fEndBin     = Fd::kMI_NFADC_BINS_v3 - 1;
      fFADCData->fTraceInfo.fBits.fResolution = kFADC_RESOLUTION_1000_BINS;
    }
    
    if ( TMirrorEvent::GetPrintLevel() > 1 ) {
      FD_CDEBUG(1) << "FADCTraceInfoRec.fWord = 0x" << setfill('0') 
                   << hex << fFADCData->fTraceInfo.fWord << dec 
		   << setfill(' ') << endl;
      FD_CDEBUG(1) << "FADCTraceInfoRec.fStartBin = " 
                   << fFADCData->fTraceInfo.fBits.fStartBin << " "
		   << "*.fEndBin = " 
		   << fFADCData->fTraceInfo.fBits.fEndBin << endl; 
      FD_CDEBUG(1) << "FADCTraceInfoRec.fResolution = " 
                   << fFADCData->fTraceInfo.fBits.fResolution << " # "
		   << GetNFADCBins( (EFADCTraceResolution)fFADCData->fTraceInfo.fBits.fResolution )
		   << " bins" << endl;
    }
    
    // --- deal with the reading of the PixelMonitorData
    
    if ( R__v <= MiEVENTVERSIONv4 ) {  // V3 format of PixelMonitorData
    
      PixelMonitorDataV3Rec v3_data;
      
      R__b >> v3_data.fPrevThreshold;
      R__b >> v3_data.fActualThreshold;
      R__b >> v3_data.fActualRate;
    
      if ( R__v >= MiEVENTVERSIONv2 ) {
        R__b >> v3_data.fMean;
        R__b >> v3_data.fVariance;
      }
      else { // v1 data
        v3_data.fMean     = 0.0;
        v3_data.fVariance = 0.0;
      }
      
      // convert V3 to V4 format
      ConvertToV4( &v3_data, &fFADCData->fMonitor );
    }
    else { // V4 format
      R__b >> fFADCData->fMonitor.fPrevThreshold;
      R__b >> fFADCData->fMonitor.fActualThreshold;
      R__b >> fFADCData->fMonitor.fActualRate;
      R__b >> fFADCData->fMonitor.fMean;
      R__b >> fFADCData->fMonitor.fVariance;
    }
    
    // --- need to evaluate 'fTraceInfo' first !
    
    unsigned int start_idx = GetTraceStartBin();
    unsigned int end_idx   = GetTraceEndBin();

    // check that the indices are set reasonably
    
    if ( start_idx > end_idx ) {
      FD_CWARN << "TFADCData::Streamer(" 
               << Fd::GetPixelNo( &fFADCData->fPixelNumber )
               << ") - swapped start and end bin!" << endl;
      swap( start_idx, end_idx );
    }
    
    bool index_overflow = false;
    
    // memory size should fit for chosen resolution !
    
    if ( end_idx >= GetNFADCBins( (EFADCTraceResolution)fFADCData->fTraceInfo.fBits.fResolution ) ) {
      FD_CERR << "TFADCData::Streamer(" 
              << Fd::GetPixelNo( &fFADCData->fPixelNumber ) << ") - "
              << "end bin(" << end_idx << ") is outside valid range ("
	      << GetNFADCBins( (EFADCTraceResolution)fFADCData->fTraceInfo.fBits.fResolution )
	      << ") !" << endl;
      index_overflow = true;
    }

    // --- deal with the reading of the FADCData
    
    if ( sizeof(FADCDataWordRec) != 2 ) {  // check for the data size (16 Bits)
      FD_CSEV << "TFADCData::Streamer() - "
              << "the data type FADCDataWordRec has wrong size= "
              << sizeof(FADCDataWordRec) << endl;
      return;
    }
  
    if ( index_overflow ) {
    
      SetTraceEndBin( GetNFADCBins( GetTraceResolution() ) - 1 );
      
      FD_CWARN << "TFADCData::Streamer() - corrected FADC trace end bin, "
               << "new value= " << GetTraceEndBin() << endl;
      
      FADCDataWord fadc = new FADCDataWordRec[end_idx - start_idx + 1];
      
      R__b.ReadFastArray( &fadc[0].fWord, end_idx - start_idx + 1 );
      
      end_idx = GetTraceEndBin();
      for ( unsigned int i=0; i<end_idx - start_idx + 1; i++)
        fFADCData->fFadcData[start_idx + i] = fadc[i];
      
      delete [] fadc;
    }
    else
      R__b.ReadFastArray( &fFADCData->fFadcData[start_idx].fWord, 
                          end_idx - start_idx + 1 );

#ifdef __ppc__
    for ( unsigned int i=start_idx; i<=end_idx; i++ ) {
      File2PPCFADCDataWordRec( fFADCData->fFadcData[i] );
    }
#endif // __ppc__

    if ( R__v >= MiEVENTVERSIONv3 ) {
      R__b >> fFADCData->fTraceStartTime;
    }
    
    if ( R__v >= MiEVENTVERSIONv4 ) {
      // ...
    }
    
  } 
  else {
    R__b.WriteVersion( TFADCData::IsA() );
    TObject::Streamer( R__b );

    R__b << fFADCData->fPixelNumber;
    
#ifdef __ppc__
    TFADCData::FADCTraceInfoRec traceinfo = 
               PPC2FileFADCTraceInfoRec( fFADCData->fTraceInfo );
    R__b << traceinfo.fWord;
#else
    R__b << fFADCData->fTraceInfo.fWord;
#endif // __ppc__
    
    R__b << fFADCData->fMonitor.fPrevThreshold;
    R__b << fFADCData->fMonitor.fActualThreshold;
    R__b << fFADCData->fMonitor.fActualRate;
    R__b << fFADCData->fMonitor.fMean;
    R__b << fFADCData->fMonitor.fVariance;
    
    unsigned int start_idx = fFADCData->fTraceInfo.fBits.fStartBin;
    unsigned int end_idx   = fFADCData->fTraceInfo.fBits.fEndBin;
    
    // check that the indices are set reasonably
    
    if ( start_idx > end_idx ) {
      FD_CWARN << "TFADCData::Streamer() - swapped start and end bin!" << endl;
      swap( start_idx, end_idx );
    }

    if ( end_idx >= GetNFADCBins( (EFADCTraceResolution)fFADCData->fTraceInfo.fBits.fResolution ) ) {
      FD_CWARN << "TFADCData::Streamer() - "
               << "end bin (" << end_idx << ") is outside valid range ("
	       << GetNFADCBins( (EFADCTraceResolution)fFADCData->fTraceInfo.fBits.fResolution )
	       << ") !" << endl;
      end_idx = min( end_idx, (unsigned int)(Fd::kMI_NFADC_BINS_DEFAULT - 1) );
    }

#ifdef __ppc__
    TFADCData::FADCDataWordRec fadc[FdUtil::Fd::kMI_NFADC_BINS_DEFAULT];
    
    for ( unsigned int i=start_idx; i<=end_idx; i++ ) {
      fadc[i] = PPC2FileFADCDataWordRec( fFADCData->fFadcData[i] );
    }
    
    R__b.WriteFastArray( &copy[start_idx].fWord, end_idx - start_idx + 1 );
#else
    R__b.WriteFastArray( &fFADCData->fFadcData[start_idx].fWord, 
                         end_idx - start_idx + 1 );
#endif // __ppc__

    // new for MiEVENTVERSIONv3
    
    R__b << fFADCData->fTraceStartTime;

    // new for MiEVENTVERSIONv4
    
    // ... not yet existing
  }
}

// -------------------------------------------------------------------------

bool TFADCData::ThresholdHasChanged() const
 {
  FADCDataWord first = FADCDataGetFADCDataWord( fFADCData, 
                               fFADCData->fTraceInfo.fBits.fStartBin );
  FADCDataWord last = FADCDataGetFADCDataWord( fFADCData, 
                               fFADCData->fTraceInfo.fBits.fEndBin );

  if ( first->fBits.fThresholdChanged == last->fBits.fThresholdChanged )
    return false;
  else
    return true;
}

ClassImp(TFADCData)

// -------------------------------------------------------------------------
// -------------------------------------------------------------------------
