
//
// File   : FadcData.hh
//
// Purpose: Declaration for the Pixel FADC data class TFADCData
//
// $Id: FadcData.hh 12082 2011-02-21 17:11:29Z mathes $
//


/** @file FadcData.hh
  * Declaration of class TFADCData.
  * @author H.-J. Mathes, FzK
  */


#ifndef _FadcData_hh_
#define _FadcData_hh_

#include <iostream>

// --> prevent users from including that file directly
#ifndef __CINT__
# if (!defined _MiEvent_hh_) && (!defined _VMiEvent_hh_) && (!defined _FEMiEvent_hh_)
#  error You are not supposed to include that header file !
#  error Please use MiEvent.hh instead !
# endif // !_MiEvent_hh_ && !_VMiEvent_hh_
#endif // __CINT__

#include <TObject.h>

#include <FdNumbering.hh>
#include <MiEventVersionDefines.hh>

/** Class for pixel ADC raw data.
  *
  * This class describes the ADC raw data (ADC data trace) of a single 
  * pixel consisting of:
  *  @li the number of the pixel (PixelNumber),
  *  @li the monitoring data of that pixel (PixelMonitorData),
  *  @li the FADC trace (FADCDataWord),
  *  @li and additional information about the way the trace is stored
  *      (number of time bins, first time bin, ...)
  *
  * Notes:
  * @li When a TFADCData object is created on the heap, the trace
  *     characteristics are automatically set to the default values. This means 
  *     that in this case the full number of bins is used.
  */
class TFADCData : public TObject {
    friend class TMirrorFADCData;
    
public:

    /** Single FADC data word format V3. */
    typedef union _FADCDataWordV3Rec {
    
      /** Describes the whole data (16 Bits). */
      unsigned short    fWord;

# ifndef __CINT__
      /** Format of FADC data woord, firmware revision V3. */
      struct {

        /** This are the 12-Bit data resulting from the conversion. */
        unsigned short  fData		  : 12;
 
        /** Shows that the conversion led to an result outside the
          * FADC conversion range.
          */
        unsigned short   fOutOfRange	   : 1;
      
        /** Pixel trigger was switched off for the current pixel ? */
        unsigned short   fPixelTriggerOff  : 1;

        /** Threshold changed within the time of that trace. */
        unsigned short   fThresholdChanged : 1;

        /** Pixel trigger (FLT) was active at that time from current pixel. */
        unsigned short   fPixelTrigger	   : 1;

      } fBits;
# endif // __CINT__

    } FADCDataWordV3Rec, *FADCDataWordV3;


    /** Single FADC data word format V4.
       
        This structure defines the format of a single FADC data word as it is 
        read out from the FLT memories. Be aware that this format is highly 
        dependent on the hardware (FPGA) version used !
        
        Matches the documentation: FLT-UserManualV3-230402.pdf, pg. 12
       
        It is coded as an union in order to make a fast access possible.
        
        This data structure has to be accessed using:
        @li FADCDataWordGetData(), FADCDataWordSetData()
        @li FADCDataWordGetOutOfRange(), FADCDataWordIsOutOfRange(), 
	    FADCDataWordSetOutOfRange()
        @li FADCDataWordGetPixelTrigger(), FADCDataWordHasPixelTrigger(),
	    FADCDataWordSetPixelTrigger()
        @li FADCDataWordGetThresholdChanged(), FADCDataWordHasThresholdChanged(),
	    FADCDataWordSetThresholdChanged()
        @li FADCDataWordGetPixelTriggerStatus(), FADCDataWordIsPixelTriggerOn(), 
	    FADCDataWordIsPixelTriggerOff(), FADCDataWordSetPixelTriggerStatus()
        @li FADCDataWordGetWord(), FADCDataWordSetWord()
      */
    typedef union _FADCDataWordRec {
    
      /** Describes the whole data (16 Bits). */
      unsigned short    fWord;

# ifndef __CINT__
      /** Format of FADC data word, firmware revision V4. */
      struct {

        /** This are the 12-Bit data resulting from the conversion. */
        unsigned short   fData		   : 12;

        /** Not used (was fOutOfRange before). */
        unsigned short   fUnused1          : 1;
      
        /** Not used (was fPixelTriggerOff before). */
        unsigned short   fUnused2          : 1;
      
        /** Threshold changed within the time of that trace. */
        unsigned short   fThresholdChanged : 1;

        /** Pixel trigger (FLT) was active at that time from current pixel. */
        unsigned short   fPixelTrigger	   : 1;

      } fBits;
# endif // __CINT__

    } FADCDataWordRec, *FADCDataWord;


public:

    /** Information (Hardware V3) about the way the FADC trace is stored.
      *
      * Usually a full length trace is stored (MI_NFADC_BINS) but it is also
      * possible (in order to save storage space and processing time) to store 
      * only the interesting part of a FADC trace.
      *
      * The fields of this data structure should be accessed using:
      * @li FADCTraceSetStartBin(), FADCTraceStartBin() or FADCTraceGetStartBin()
      * @li FADCTraceSetEndBin(), FADCTraceEndBin() or FADCTraceGetEndBin()
      * @li FADCTraceLength() or FADCTraceGetLength()
      *
      * The length of the trace is a read-only parameter as internally only 
      * start and end bin are stored.
      */
    typedef union _FADCTraceInfoV3Rec {
    
      /** Describes the whole _FADCTraceInfoRec data (32 Bits). */
      unsigned int fWord;

# ifndef __CINT__
      /** Explicit default constructor for the type _FADCTraceInfoRec.
        * It guarantees a proper default initialisation ...
        */
      _FADCTraceInfoV3Rec(unsigned int start=0, 
                          unsigned int stop=FdUtil::Fd::kMI_NFADC_BINS_v3-1)
       { fBits.fStartBin = start; fBits.fEndBin = stop; }

      /** An alternative bit-field overlayed our data word. */
      struct {
        /** Original time bin where the trace starts. */
        unsigned  fStartBin : 10;
        /** Original time bin where the trace stops. */
        unsigned  fEndBin : 10;
      } fBits;
# endif  //  __CINT__
    } FADCTraceInfoV3Rec, *FADCTraceInfoV3;

    /** Possible values for the FADC resolution. */
    typedef enum FADCTraceResolution {
      
      kFADC_RESOLUTION_1000_BINS = 0,
      kFADC_RESOLUTION_2000_BINS,
      kFADC_RESOLUTION_4000_BINS,
      
      kFADC_DEFAULT_RESOLUTION = kFADC_RESOLUTION_2000_BINS
    
    } EFADCTraceResolution;
    
    /** Possible values for the FADC bin width. */
    typedef enum FADCBinWidth {
    
      kFADC_BINWIDTH_25NS = 25,
      kFADC_BINWIDTH_50NS = 50,
      kFADC_BINWIDTH_100NS = 100
      
    } EFADCBinWidth;
    
    /** Information (Hardware V4) about the way the FADC trace is stored. */
    typedef union _FADCTraceInfoRec {
    
      /** Describes the whole _FADCTraceInfoRec data (32 Bits). */
      unsigned int fWord;

# ifndef __CINT__
      /** Explicit default constructor for the type _FADCTraceInfoRec.
        * It guarantees a proper default initialisation ...
        */
      _FADCTraceInfoRec(unsigned int start=0, 
                        unsigned int stop=FdUtil::Fd::kMI_NFADC_BINS_DEFAULT-1,
			EFADCTraceResolution res=kFADC_DEFAULT_RESOLUTION)
       { fBits.fStartBin = start;
         fBits.fEndBin = stop;
         fBits.fResolution = (int)res;
      }
      
      /** An alternative bit-field overlayed our data word. */
      struct {
        /** Original time bin where the trace starts. */
        unsigned   fStartBin   : 12;
        /** Original time bin where the trace stops. */
        unsigned   fEndBin     : 12;
	/** Resolution specifier. */
	unsigned   fResolution : 2;
      } fBits;
# endif  //  __CINT__

    } FADCTraceInfoRec, *FADCTraceInfo;

public:

    /** Single pixel monitoring data.
      * 
      * This structure contains the monitoring data of a single pixel consisting
      * of the threshold settings and the pixel trigger rate (FLT rate).
      *
      * This data structure has to be accessed using:
      * @li PixelMonitorDataGetPrevThr(), PixelMonitorDataSetPrevThr()
      * @li PixelMonitorDataGetActThr(), PixelMonitorDataSetActThr()
      * @li PixelMonitorDataGetActRate(), PixelMonitorDataSetActRate()
      * @li PixelMonitorDataGetMean(), PixelMonitorDataSetMean()
      * @li PixelMonitorDataGetVariance(), PixelMonitorDataSetVariance()
      */
    typedef struct _PixelMonitorDataRec {
    
      /** Previous threshold value. */
      unsigned int    fPrevThreshold;
      /** Current threshold value. */
      unsigned int    fActualThreshold;
      /** Current pixel trigger rate. */
      unsigned int    fActualRate;
      /** Mean calculated by the FLT statistics algorithm. */
      double          fMean;
      /** Variance calculated by the FLT statistics algorithm. */
      double          fVariance;
      
    } PixelMonitorDataRec, *PixelMonitorData;


    /** Single pixel monitoring data for hardware V3.
      * 
      * This structure contains the monitoring data of a single pixel consisting
      * of the threshold settings and the pixel trigger rate (FLT rate).
      */
    typedef struct _PixelMonitorDataV3Rec {
    
      /** Previous threshold value. */
      unsigned short    fPrevThreshold;
      /** Current threshold value. */
      unsigned short    fActualThreshold;
      /** Current pixel trigger rate. */
      unsigned short    fActualRate;
      /** Mean calculated by the FLT statistics algorithm. */
      double            fMean;
      /** Variance calculated by the FLT statistics algorithm. */
      double            fVariance;
      
    } PixelMonitorDataV3Rec, *PixelMonitorDataV3;


   
    /** Single PMTs data structure.
      * 
      * This structure contains the FADC data of a single PMT plus
      * additional information necessary for monitoring and calibration.
      *
      * This data structure has to be accessed using:
      * @li FADCDataGetPixelNo(), FADCDataSetPixelNo()
      * @li FADCDataGetPixelMonitorData()
      * @li FADCDataGetTraceInfo()
      * @li FADCDataGetFADCDataWord()
      */
    typedef struct _FADCDataRec {
    
      //static void * operator new(size_t);
      //static void operator delete(void *,size_t);
      
      /** Constructor of the struct _FADCDataRec. */
      //_FADCDataRec(unsigned int initial_size=FdUtil::Fd::kMI_NFADC_BINS_LOW);
      
      /** Copy-constructor of the struct _FADCData. */
      //_FADCDataRec(const _FADCData&);
      
      /** The number of the current pixel. */
      FdUtil::Fd::PixelNumberRec   fPixelNumber;
      
      /** Structure containing the pixels monitoring data. */
      PixelMonitorDataRec          fMonitor;

      /** Information about how the trace is stored. */
      FADCTraceInfoRec             fTraceInfo;  // Hardware V4
      
      /** The absolute time of the first bin of a trace. */
      Float_t                      fTraceStartTime;

      /** The FADC raw data - Hardware V4. */
      FADCDataWordRec      fFadcData[FdUtil::Fd::kMI_NFADC_BINS_DEFAULT];

    } FADCDataRec, *FADCData;


public:
    
    /** Constructor for an object of class TFADCData. */
    TFADCData(FdUtil::Fd::PixelNumber pixel=NULL);

    /** Constructor for class TFADCData from existing C-struct FADCData. */
    TFADCData(FADCData);

    /** Copy constructor of the class TFADCData. */
    TFADCData(const TFADCData&);

    /** operator=() of the class TFADCData. */
    TFADCData& operator=(const TFADCData&);

    /** Destructor for an object of type TFADCData. 
      *
      * This destructor destroys the object in a way which is compatible to 
      * the specific method of its construction.
      */
    virtual ~TFADCData();

    /** Clear the contents of a TFADCData object. */
    virtual void Clear(Option_t *o="");
    
    /** Convert V4 data to V3 data.
      *
      * This method is only necessary for some software tests.
      */
    static bool ConvertToV3(const TFADCData*,TFADCData*);

    /** Initialize the FADCs data structures by converting from V3 data.
      *
      * @see also the static member function !
      */
    virtual bool ConvertFromV3(const TFADCData*);
    
    /** Convert V3 data to V4 data.
      *
      * If the conversion is not possible, return false, else true. This
      * might be the case if:
      * @li the resolution is matching
      * @li an unsupported resolution is requested/used.
      */
    static bool ConvertToV4(const TFADCData*,TFADCData*);
    
    /** Convert the FADCTraceInfo from V3 to V4. */
    static void ConvertToV4(const FADCTraceInfoV3,FADCTraceInfo);
    
    /** oOnvert the PixelMonitorData from V3 to V4. */
    static void ConvertToV4(const PixelMonitorDataV3,PixelMonitorData);
    
    
public:    // --- getters ...
    
    /** Return the actual trigger rate of the selected pixel. */
    unsigned int GetActualRate() const
     { return ((TFADCData *)this)->GetPixelMonitorData()->fActualRate; }

    /** Return the actual threshold settings stored in PixelMonitorData. */
    unsigned int GetActualThreshold() const
     { return ((TFADCData *)this)->GetPixelMonitorData()->fActualThreshold; }
   
    /** Get the bin width associated with a certain resolution. */
    static EFADCBinWidth GetBinWidth(EFADCTraceResolution);
    
    /** Get the bin width of the current FADC trace. */
    EFADCBinWidth GetBinWidth() const
     { return GetBinWidth( GetTraceResolution() ); }
    
    /** Get the default resolution. */
    static EFADCTraceResolution GetDefaultResolution()
     { return fgDefaultResolution; }

    /** Return a pointer to the FADC data trace. */
    FADCDataWord GetFADCTrace() const
     { return &fFADCData->fFadcData[0]; }

    /** Return the mean value of the FLT statistics calculation. */
    float GetMean() const
     { return ((TFADCData *)this)->GetPixelMonitorData()->fMean; }

    /** Get the current number of FADC bins. */
    unsigned int GetNFADCBins() const
     { 
      return GetNFADCBins( (EFADCTraceResolution)fFADCData->fTraceInfo.fBits.fResolution );
    }
    
    /** Get the number of FADC bins corresponding to the resolution. */
    static unsigned int GetNFADCBins(EFADCTraceResolution);
    
    /** Return the PixelNumber of the current pixel. */
    FdUtil::Fd::PixelNumber GetPixelNumber() const
     { return &fFADCData->fPixelNumber; }

    /** Return a reference to the pixels monitoring data.
      * This might be useful for first level calibration or zero suppression.
      */
    PixelMonitorData GetPixelMonitorData() const
     { return &fFADCData->fMonitor; }

    /** Get the bin number where the FADC trace ends. */
    unsigned int GetTraceEndBin() const
     { return fFADCData->fTraceInfo.fBits.fEndBin; }

    /** Get the additional information which is stored about the FADC trace. */
    TFADCData::FADCTraceInfo GetTraceInfo()
     { return &fFADCData->fTraceInfo; }

    /** Get the length of the ADC data traces.
      *
      * For a full mirror readout the returned value equals to MI_NFADC_BINS. 
      * For a specially selected readout of a limited number of ADC samples, a 
      * value smaller than MI_NFADC_NBINS might be returned.
      */
    unsigned int GetTraceLength() const
     { return fFADCData->fTraceInfo.fBits.fEndBin - 
              fFADCData->fTraceInfo.fBits.fStartBin + 1; }

    /** Get the resolution settings for this TFADCData trace. */
    TFADCData::EFADCTraceResolution GetTraceResolution() const
     { return (EFADCTraceResolution)fFADCData->fTraceInfo.fBits.fResolution; }

    /** Get the resolution which corresponds to a certain number of bins. */
    static TFADCData::EFADCTraceResolution GetTraceResolution(unsigned int);

    /** Get the bin number where the FADC trace begins. */
    unsigned int GetTraceStartBin() const
     { return fFADCData->fTraceInfo.fBits.fStartBin; }

    /** Return the variance of the FLT statistics calculation. */
    float GetVariance() const
     { return ((TFADCData *)this)->GetPixelMonitorData()->fVariance; }

    /** Return the previous threshold settings stored in PixelMonitorData. */
    unsigned int GetPreviousThreshold() const
     { return ((TFADCData *)this)->GetPixelMonitorData()->fPrevThreshold; }
   
    /** Return kTRUE, if the trigger threshold (FLT) has been changed during
      * that particular trace.
      */
    bool ThresholdHasChanged() const;
    
    /** Initialize the structure with some reasonable data.
      *
      * If pixelnumber==NULL, the local ones are taken - only Eye and Mirror Id.
      */
    static void Init(TFADCData::FADCData,FdUtil::Fd::PixelNumber =NULL);
    
public:    // --- setters ...
    
    /** Set the default number of FADC bins (resolution). */
    static void SetDefaultResolution(EFADCTraceResolution res);

    /** Set the eye number stored in the PixelNumber field. */
    void SetEyeNo(unsigned int eye_no)
     { FdUtil::Fd::SetEyeNo( &fFADCData->fPixelNumber, eye_no ); }

    /** Set the mean value of the statistics analysis stored in the
      * PixelMonitorData 'object'.
      */
    void SetMean(float mean)
     { GetPixelMonitorData()->fMean = mean; }

    /** Set the mirror number stored in the PixelNumber field. */
    void SetMirrorNo(unsigned int mirror_no)
     { FdUtil::Fd::SetMirrorNo( &fFADCData->fPixelNumber, mirror_no ); }

    /** Set the pixels monitor data. */
    void SetPixelMonitorData(PixelMonitorData pmd)
     { *(GetPixelMonitorData()) = *pmd; }
    
    /** Set the actual (compound) pixel number stored in the TFADCData object. 
      */
    void SetPixelNumber(FdUtil::Fd::PixelNumber pixel)
     { fFADCData->fPixelNumber = *pixel; }

    /** Set the actual (compound) pixel number stored in the TFADCData object. */
    void SetPixelNumber(unsigned int col,unsigned int row)
     { FdUtil::Fd::SetPixelNoFromColumnRow( &fFADCData->fPixelNumber, col, row ); }

    /** Set the number of FADC bins (resolution). */
    void SetTraceResolution(EFADCTraceResolution);

    /** Set the end bin for the FADC trace. */
    void SetTraceEndBin(unsigned int end_bin)
     { fFADCData->fTraceInfo.fBits.fEndBin = end_bin; }

#ifndef __CINT__
    /** Set the additional information about the FADC trace. */
    void SetTraceInfo(TFADCData::FADCTraceInfo info)
     { fFADCData->fTraceInfo = *info; }
#endif // __CINT__

    /** Set the start bin for the FADC trace. */
    void SetTraceStartBin(UInt_t start_bin)
     { fFADCData->fTraceInfo.fBits.fStartBin = start_bin; }

    /** Set the variance value of the statistics analysis stored in 
      * PixelMonitorData.
      */
    void SetVariance(float var)
     { GetPixelMonitorData()->fVariance = var; }

    // --- iterators, operators to access directly the data ...

    // ...

    /** Print the contents of the TFADCData to the specified ostream. */
    virtual void Show(std::ostream& ostr=std::cout) const;
    
    /** Print the contents of a single FADCDataWord data item. */
    static void Show(TFADCData::FADCDataWord,std::ostream& ostr=std::cout);
         
protected:

    /** Initialize a FADCData object. 
      *
      * Column and row number are set to invalid values (0,0), the start/end bins
      * of the FADC trace is set to 0/MI_NFADC_BINS. All fields of the 
      * PixelMonitorData are initialized with 0.
      */
    void Init();

private:
    enum Constructor_t { kCREATE, kFADCDATA };
    
    static const char *GetVerboseConstructorType(enum Constructor_t);
    
    enum Constructor_t fConstructorType;          //! how object was created
    FADCData           fFADCData;                 //! pointer to C-struct

    static EFADCTraceResolution fgDefaultResolution;   //!
    
    ClassDef(TFADCData,MiEVENTVERSIONv5)
};


// -------------------------------------------------------------------------
// -------------------------------------------------------------------------
//
// defines and inlines for the public data structures
//

/* --- accessing single FADC data word --- */

/** Get the complete data word (16 bits). */
#ifndef _MIEVENT_USE_MACROS_
inline unsigned short FADCDataWordGetWord(TFADCData::FADCDataWord fdw)
 { return fdw->fWord; }
#else
# define FADCDataWordGetWord(_fdw)        (_fdw)->fWord
#endif // _MIEVENT_USE_MACROS_

/** Set the complete data word (16 bits). */
#ifndef _MIEVENT_USE_MACROS_
inline void FADCDataWordSetWord(TFADCData::FADCDataWord fdw,unsigned short word)
 { fdw->fWord = word; }
#else
# define FADCDataWordSetWord(_fdw,_x)     (_fdw)->fWord = _x
#endif // _MIEVENT_USE_MACROS_

/** Get the FADC data word (12 bits). */
#ifndef _MIEVENT_USE_MACROS_
inline unsigned short FADCDataWordGetData(TFADCData::FADCDataWord fdw)
 { return fdw->fBits.fData; }
#else
# define FADCDataWordGetData(_fdw)        (_fdw)->fBits.fData
#endif // _MIEVENT_USE_MACROS_

/** Set the FADC data word (12 bits). */
#ifndef _MIEVENT_USE_MACROS_
inline void FADCDataWordSetData(TFADCData::FADCDataWord fdw,unsigned short data)
 { fdw->fBits.fData = data; }
#else
# define FADCDataWordSetData(_fdw,_x)     (_fdw)->fBits.fData = _x
#endif // _MIEVENT_USE_MACROS_

/** Get the out-of-range bit (V3). */
# ifdef _MIEVENT_USE_MACROS_
inline unsigned int FADCDataWordGetOutOfRange(TFADCData::FADCDataWordV3 fdw)
 { return fdw->fBits.fOutOfRange; }
# else
#  define FADCDataWordGetOutOfRange(_fdw)     (_fdw)->fBits.fOutOfRange
# endif // _MIEVENT_USE_MACROS_

/** Return true, if the FADC has to convert a value outside
  * its conversion range (V3).
  */
# ifdef _MIEVENT_USE_MACROS_
inline bool FADCDataWordIsOutOfRange(TFADCData::FADCDataWordV3 fdw)
 { return (fdw->fBits.fOutOfRange == 1); }
# else
#  define FADCDataWordIsOutOfRange(_fdw)        ((_fdw)->fBits.fOutOfRange == 1)
# endif // _MIEVENT_USE_MACROS_

/** Set the out-of-range bit. (V3). */
# ifdef _MIEVENT_USE_MACROS_
inline void FADCDataWordSetOutOfRange(TFADCData::FADCDataWordV3 fdw)
 { fdw->fBits.fOutOfRange = 1; }
# else
#  define FADCDataWordSetOutOfRange(_fdw)        (_fdw)->fBits.fOutOfRange = 1
# endif // _MIEVENT_USE_MACROS_

# ifdef _MIEVENT_USE_MACROS_
/** Get the trigger status bit, i.e. if the pixel
  * trigger was on or off for the specified time bin (V3).
  */
inline unsigned int FADCDataWordGetPixelTriggerStatus(TFADCData::FADCDataWordV3 fdw)
 { return fdw->fBits.fPixelTriggerOff; }
# else
#  define FADCDataWordGetPixelTriggerStatus(_fdw) (_fdw)->fBits.fPixelTriggerOff
# endif // _MIEVENT_USE_MACROS_

/** Return true, if the pixel trigger is on in the data (V3). */
# ifdef _MIEVENT_USE_MACROS_
inline bool FADCDataWordIsPixelTriggerOn(TFADCData::FADCDataWordV3 fdw)
 { return (fdw->fBits.fPixelTriggerOff != 1); }
# else
#  define FADCDataWordIsPixelTriggerOn(_fdw)    ((_fdw)->fBits.fPixelTriggerOff != 1)
# endif // _MIEVENT_USE_MACROS_

/** Return true, if the pixel trigger is off in the data (V3). */
# ifdef _MIEVENT_USE_MACROS_
inline bool FADCDataWordIsPixelTriggerOff(TFADCData::FADCDataWordV3 fdw)
 { return (fdw->fBits.fPixelTriggerOff == 1); }
# else
#  define FADCDataWordIsPixelTriggerOff(_fdw)   ((_fdw)->fBits.fPixelTriggerOff == 1)
# endif // _MIEVENT_USE_MACROS_

# ifdef _MIEVENT_USE_MACROS_
/** Set the trigger status bit, i.e. if the pixel
  * trigger was on or off for the specified time bin (V3).
  */
inline void FADCDataWordSetPixelTriggerStatus(TFADCData::FADCDataWordV3 fdw,int status)
 { fdw->fBits.fPixelTriggerOff = (0x01 & status); }
# else
#  define FADCDataWordPixelTriggerStatus(_fdw,_st) \
	(_fdw)->fBits.fPixelTriggerOff = ((_st) & 0x01)
# endif // _MIEVENT_USE_MACROS_

# ifdef _MIEVENT_USE_MACROS_
/** Set the trigger status bit, i.e. if the pixel
  * trigger was on or off for the specified time bin (V3).
  */
inline void FADCDataWordSetPixelTriggerOn(TFADCData::FADCDataWordV3 fdw)
 { fdw->fBits.fPixelTriggerOff = 0; }

inline void FADCDataWordSetPixelTriggerOff(TFADCData::FADCDataWordV3 fdw)
 { fdw->fBits.fPixelTriggerOff = 1; }
# else
#  define FADCDataWordSetPixelTriggerOn(_fdw) \
	(_fdw)->fBits.fPixelTriggerOff = 0
#  define FADCDataWordSetPixelTriggerOff(_fdw) \
	(_fdw)->fBits.fPixelTriggerOff = 1
# endif // _MIEVENT_USE_MACROS_

/** Get the threshold changed bit. */
#ifndef _MIEVENT_USE_MACROS_
inline unsigned int FADCDataWordGetThresholdChanged(TFADCData::FADCDataWord fdw)
 { return fdw->fBits.fThresholdChanged; }
#else
# define FADCDataWordGetThresholdChanged(_fdw) (_fdw)->fBits.fThresholdChanged
#endif // _MIEVENT_USE_MACROS_

/** Return true, if the threshold has changed. 
  * The new threshold has the name ActualThreshold and the old one
  * PreviousThreshold
  * @see PixelMonitorData
  */
#ifndef _MIEVENT_USE_MACROS_
inline bool FADCDataWordHasThresholdChanged(TFADCData::FADCDataWord fdw)
 { return (fdw->fBits.fThresholdChanged == 1); }
#else
# define FADCDataWordHasThresholdChanged(_fdw)  ((_fdw)->fBits.fThresholdChanged == 1)
#endif // _MIEVENT_USE_MACROS_

/** Set the threshold changed bit. */
#ifndef _MIEVENT_USE_MACROS_
inline void FADCDataWordSetThresholdChanged(TFADCData::FADCDataWord fdw)
 { fdw->fBits.fThresholdChanged = 1; }
#else
# define FADCDataWordSetThresholdChanged(_fdw) (_fdw)->fBits.fThresholdChanged = 1
#endif // _MIEVENT_USE_MACROS_

/** Get the bit which indicates that the boxcar summed signal was
  * above the threshold.
  */
#ifndef _MIEVENT_USE_MACROS_
inline unsigned int FADCDataWordGetPixelTrigger(TFADCData::FADCDataWord fdw)
 { return fdw->fBits.fPixelTrigger; }
#else
# define FADCDataWordGetPixelTrigger(_fdw)   (_fdw)->fBits.fPixelTrigger
#endif // _MIEVENT_USE_MACROS_

/** Return true, if a pixel trigger has occurred in the data.  */
#ifndef _MIEVENT_USE_MACROS_
inline bool FADCDataWordHasPixelTrigger(TFADCData::FADCDataWord fdw)
 { return (fdw->fBits.fPixelTrigger == 1); }
#else
# define FADCDataWordHasPixelTrigger(_fdw)      ((_fdw)->fBits.fPixelTrigger == 1)
#endif // _MIEVENT_USE_MACROS_

/** Set the bit which indicates that the boxcar summed signal was
  * above the threshold.
  */
#ifndef _MIEVENT_USE_MACROS_
inline void FADCDataWordSetPixelTrigger(TFADCData::FADCDataWord fdw)
 { fdw->fBits.fPixelTrigger = 1; }
#else
# define FADCDataWordSetPixelTrigger(_fdw)   (_fdw)->fBits.fPixelTrigger = 1
#endif // _MIEVENT_USE_MACROS_

/* --- accessing data structures of FADC structure --- */

/** Get the PixelNumber of the single FADC trace. */
inline FdUtil::Fd::PixelNumber FADCDataGetPixelNumber(TFADCData::FADCData fadc)
 { return &fadc->fPixelNumber; }
 
/** Get the pixel number of the single FADC trace. */
inline unsigned int FADCDataGetPixelNo(TFADCData::FADCData fadc)
 { return FdUtil::Fd::GetPixelNo( &fadc->fPixelNumber ); }

/** Get pointer to PixelMonitorData of the single FADC trace. */
inline TFADCData::PixelMonitorData
       FADCDataGetPixelMonitorData(TFADCData::FADCData fadc)
 { return &fadc->fMonitor; }
 
/** Get pointer to single data word at [] of single FADC. 
  * So FADCDataGetFADCDataWord(fadc,0) will get the pointer
  * to the complete data array (obsolete).
  */
inline TFADCData::FADCDataWord 
       FADCDataGetFADCDataWord(TFADCData::FADCData fadc, int ind)
 { return &fadc->fFadcData[ind]; }

/** Get pointer to FADCTraceInfo of the single FADC trace. */
inline TFADCData::FADCTraceInfo
       FADCDataGetTraceInfo(TFADCData::FADCData fadc)
 { return &fadc->fTraceInfo; }

/** Set the pixel number for the single FADC. */
inline void FADCDataSetPixelNo(TFADCData::FADCData fadc, UInt_t p)
 { FdUtil::Fd::SetPixelNo( &fadc->fPixelNumber, p ); }

/** Set the PixelNumber for the single FADC. */
inline void FADCDataSetPixelNumber(TFADCData::FADCData fadc,
                                   FdUtil::Fd::PixelNumber p)
 { fadc->fPixelNumber = *p; }

/* --- accessing PixelMonitorData --- */

/** Get the previous threshold value of the pixels monitoring data. */
inline unsigned int PixelMonitorDataGetPrevThr(TFADCData::PixelMonitorData pmd)
 { return pmd->fPrevThreshold; }

/** Get the previous threshold value of the pixels monitoring data (obsolete). */
inline unsigned int PixelMonitorDataPrevThr(TFADCData::PixelMonitorData pmd)
 { return PixelMonitorDataGetPrevThr( pmd ); }

/** Get the actual threshold value of the pixels monitoring data. */
inline unsigned int PixelMonitorDataGetActThr(TFADCData::PixelMonitorData pmd)
 { return pmd->fActualThreshold; }

/** Get the actual threshold value of the pixels monitoring data (obsolete). */
inline unsigned int PixelMonitorDataActThr(TFADCData::PixelMonitorData pmd)
 { return PixelMonitorDataGetActThr( pmd ); }

/** Get the actual pixel rate value of the pixels monitoring data. */
inline unsigned int PixelMonitorDataGetActRate(TFADCData::PixelMonitorData pmd)
 { return pmd->fActualRate; }

/** Get the actual pixel rate value of the pixels monitoring data (obsolete). */
inline unsigned int PixelMonitorDataActRate(TFADCData::PixelMonitorData pmd)
 { return PixelMonitorDataGetActRate( pmd ); }

/** Get the mean value calculated by the FLT statistics algorithm.
  */
inline double PixelMonitorDataGetMean(TFADCData::PixelMonitorData pmd)
 { return pmd->fMean; }

/** Get the mean value calculated by the FLT statistics algorithm (obsolete). */
inline double PixelMonitorDataMean(TFADCData::PixelMonitorData pmd)
 { return PixelMonitorDataGetMean( pmd ); }

/** Get the variance calculated by the FLT statistics algorithm. */
inline double PixelMonitorDataGetVariance(TFADCData::PixelMonitorData pmd)
 { return pmd->fVariance; }

/** Get the variance calculated by the FLT statistics algorithm (obsolete). */
inline double PixelMonitorDataVariance(TFADCData::PixelMonitorData pmd)
 { return PixelMonitorDataGetVariance( pmd ); }

/** Set the value for the previous threshold settings. */
inline void PixelMonitorDataSetPrevThr(TFADCData::PixelMonitorData pmd,unsigned int thr)
 { pmd->fPrevThreshold = thr; }
 
/** Set the value for the actual threshold settings. */
inline void PixelMonitorDataSetActThr(TFADCData::PixelMonitorData pmd,unsigned int thr)
 { pmd->fActualThreshold = thr; }
 
/** Set the value of the actual pixel rate. */
inline void PixelMonitorDataSetActRate(TFADCData::PixelMonitorData pmd,unsigned int rate)
 { pmd->fActualRate = rate; }

/** Set the mean value calculated by the FLT statistics algorithm. */
inline void PixelMonitorDataSetMean(TFADCData::PixelMonitorData pmd, double mean)
 { pmd->fMean = mean; }

/** Set the variance calculated by the FLT statistics algorithm.
  */
inline void PixelMonitorDataSetVariance(TFADCData::PixelMonitorData pmd, double var)
 { pmd->fVariance = var; }


/* --- accessing FADCTraceInfo --- */

/** Set the start bin of the FADC data trace. */
inline void FADCTraceSetStartBin(TFADCData::FADCTraceInfo info,unsigned int start_bin)
 { info->fBits.fStartBin = start_bin; }

/** Return the start bin of the FADC data trace. */
inline unsigned int FADCTraceGetStartBin(TFADCData::FADCTraceInfo info)
 { return info->fBits.fStartBin; }

/** Set the end bin of the FADC data trace. */
inline void FADCTraceSetEndBin(TFADCData::FADCTraceInfo info,unsigned int end_bin)
 { info->fBits.fEndBin = end_bin; }

/** Return the end bin of the FADC data trace. */
inline unsigned int FADCTraceGetEndBin(TFADCData::FADCTraceInfo info)
 { return info->fBits.fEndBin; }

/** Return the length of the FADC data trace. */
inline unsigned int FADCTraceGetLength(TFADCData::FADCTraceInfo info)
 { return info->fBits.fEndBin - info->fBits.fStartBin + 1; }

/** Get the resolution of the FADCDataTraceInfo. */
inline TFADCData::FADCTraceResolution
       FADCTraceGetResolution(TFADCData::FADCTraceInfo info)
 { return (TFADCData::FADCTraceResolution)info->fBits.fResolution; }

/** Set the resolution of the FADCDataTraceInfo. */
inline void FADCTraceSetResolution(TFADCData::FADCTraceInfo info,
                                   TFADCData::EFADCTraceResolution res)
 { info->fBits.fResolution = res; }

#endif  // _FadcData_hh_
