
//
// File   : MiBitArray.cc
//
// Purpose: Implementation of the bit array class
//
// $Id: MiBitArray.cc 8759 2010-09-27 12:16:36Z mathes $
//


#include <iostream>
#include <algorithm>
#include <stdexcept>
#include <cassert>


/** @file MiBitArray.cc
  * Implementation of class TMirrorBitArray
  * @author H.-J. Mathes, FzK
  */

#include <ErrorLogging.hh>
#include <FdNumbering.hh>

#include "CheckVersion.hh"

#include "MiEvent.hh"
#include "MiBitArray.hh"

using namespace std;
using namespace FdUtil;

// -------------------------------------------------------------------------
// -------------------------------------------------------------------------
//
// class TMirrorBitArray - implementation
//

TMirrorBitArray::TMirrorBitArray()
 {
#ifdef TRACE_CONSTRUCTOR
  FD_CINFO << "TMirrorBitArray(): " << this << endl;
#endif
  Clear();
}

// -------------------------------------------------------------------------

TMirrorBitArray::TMirrorBitArray(const TMirrorPixelList *pixel_list)
 {
#ifdef TRACE_CONSTRUCTOR
  FD_CINFO << "TMirrorBitArray(TMirrorPixelList*): " << this << endl;
#endif
  assert( pixel_list != NULL );
  Clear();
  Set( pixel_list );
}
 
// -------------------------------------------------------------------------

TMirrorBitArray::TMirrorBitArray(const TMirrorPixelData *pixel_data, 
                                 unsigned int start_slice,unsigned int end_slice)
 {
#ifdef TRACE_CONSTRUCTOR
  FD_CINFO << "TMirrorBitArray(TMirrorPixelData*): " << this << endl;
#endif
  assert( pixel_data != NULL );
  Clear();
  Set( pixel_data, start_slice, end_slice );
}

// -------------------------------------------------------------------------

void TMirrorBitArray::Clear()
 {
  for ( int col=0; col<Fd::kMI_NCOLS; col++ ) fBits[col] = 0;
}
 
// -------------------------------------------------------------------------

bool TMirrorBitArray::_Check(unsigned int col,unsigned int row) const
 {
  if ( col < Fd::kMI_FIRST_COL || col > Fd::kMI_LAST_COL ) return false;
  if ( row < Fd::kMI_FIRST_ROW || row > Fd::kMI_LAST_ROW ) return false;
  
  return true;
}

// -------------------------------------------------------------------------

void TMirrorBitArray::Extend(unsigned int extcol,unsigned int extrow)
 {
  if ( !extcol && !extrow ) return; // nothing to do ...
  
  TMirrorBitArray work(*this);
  
  unsigned int rmin, rmax, cmin, cmax;
  
  for ( unsigned int col=Fd::kMI_FIRST_COL; col<=Fd::kMI_LAST_COL; col++ ) {
    for ( unsigned int row=Fd::kMI_FIRST_ROW; row<=Fd::kMI_LAST_ROW; row++ ) {
      
      if ( Test( col, row ) ) {
	
	rmin = max( row - extrow, (unsigned int)Fd::kMI_FIRST_ROW );
	rmax = min( row + extrow, (unsigned int)Fd::kMI_LAST_ROW );
	cmin = max( col - extcol, (unsigned int)Fd::kMI_FIRST_COL );
	cmax = min( col + extcol, (unsigned int)Fd::kMI_LAST_COL );
	
#if 0
	if ( extrow > 0 ) {
	  for ( unsigned int r=rmin; r<=rmax; r++ ) work.Set( col, r ); 
        }
	if ( extcol > 0 ) {
          for ( unsigned int c=cmin; c<=cmax; c++ ) work.Set( c, row ); 
	}
#else
        for ( unsigned int c=cmin; c<=cmax; c++ ) 
	  for ( unsigned int r=rmin; r<=rmax; r++ ) work.Set( c, r ); 
#endif
      } // if ( Test(col,row) ) ...
    } // for ( col=..,) ...
  } // for ( row=..,) ...
  
  Merge( &work );
}

// -------------------------------------------------------------------------

// bits to be added for even column index
static short gExtEven[6][2] = 
  { { 0, 1 }, { 1, 1 }, { -1, 0 }, { 1, 0 }, { 0, -1 }, { 1, -1 } };
// bits to be added for odd column index
static short gExtOdd[6][2] = 
  { { -1, 1 }, { 0, 1 }, { -1, 0 }, { 1, 0 }, { -1, -1 }, { 0, -1 } };

void TMirrorBitArray::ExtendHexagonal(unsigned int extcrown)
 {
  if ( !extcrown ) return; // nothing to do ...
  
  if ( extcrown > 3 ) {  // switch to 'orthogonal' algorithm
    Extend( extcrown, extcrown );
    return;
  }
  
  for ( unsigned int n=0; n<extcrown; n++ ) {
  
    TMirrorBitArray work(*this);

    for ( unsigned int col=Fd::kMI_FIRST_COL; col<=Fd::kMI_LAST_COL; col++ ) {
      for ( unsigned int row=Fd::kMI_FIRST_ROW; row<=Fd::kMI_LAST_ROW; row++ ) {
      
        if ( Test( col, row ) ) {

          short *ext = (( col % 2 ) ? &gExtOdd[0][0] : &gExtEven[0][0] );
	
	  for ( int k=0; k<12; k+=2 ) {
	
	    int c = col + ext[k + 0];
	    int r = row + ext[k + 1];
	  
	    if ( !c && c > Fd::kMI_LAST_COL ) continue;
	    if ( !r && r > Fd::kMI_LAST_ROW ) continue;
	  
	    work.Set( c, r );
	  }

        } // if ( Test(col,row) ) ...
      } // for ( col=..,) ...
    } // for ( row=..,) ...
  
    Merge( &work );
    
  } // for ( n=..,) ...
}

// -------------------------------------------------------------------------

unsigned int TMirrorBitArray::GetNumberOfBits() const
 {
  unsigned int bits_set = 0;
  
  for ( unsigned int col=Fd::kMI_FIRST_COL; col<=Fd::kMI_LAST_COL; col++ )
    bits_set += GetNumberOfBitsInColumn( col );

  return bits_set;
}

// -------------------------------------------------------------------------

unsigned int TMirrorBitArray::GetNumberOfBitsInColumn(unsigned int col) const
 {
  unsigned int bits_set = 0;
  unsigned int mask = 1;
  
  for ( unsigned int row=0; row<Fd::kMI_NROWS; row++, mask <<= 1 )
    if ( fBits[col - Fd::kMI_FIRST_COL] & mask ) bits_set++;

  return bits_set;
}

// -------------------------------------------------------------------------

unsigned int TMirrorBitArray::GetNumberOfBitsInRow(unsigned int row) const
 {
  unsigned int bits_set = 0;
  unsigned int mask = 1 << (row - Fd::kMI_FIRST_ROW);
  
  for ( unsigned int col=0; col<Fd::kMI_NCOLS; col++ )
    if ( fBits[col] & mask ) bits_set++;

  return bits_set;
}

// -------------------------------------------------------------------------

int TMirrorBitArray::GetRowData(unsigned int row) const
 {
  if ( row < Fd::kMI_FIRST_ROW || row > Fd::kMI_LAST_ROW ) return 0;
  
  int data = 0, mask = 1;
  
  for ( unsigned int col=0; col<Fd::kMI_NROWS; col++) {
    if ( Test( col + Fd::kMI_FIRST_COL, row ) )
      data |= mask;
    mask <<= 1;
  }
  
  return data;
}

// -------------------------------------------------------------------------

void TMirrorBitArray::Merge(TMirrorBitArray* ba)
 {
  for ( int col=0; col<Fd::kMI_NCOLS; col++ )
    fBits[col] |= ba->fBits[col];
}

// -------------------------------------------------------------------------

void TMirrorBitArray::Set(const TMirrorPixelData *pixel_data,
                          unsigned int start_slice,unsigned int end_slice)
 {
  assert( pixel_data != NULL );

  if ( start_slice > end_slice ) {
  
    swap( start_slice, end_slice );

    FD_CWARN << "TMirrorBitArray::Set() - "
             << "start and end slice exchanged !" << endl;
  }
  
  if ( start_slice > Fd::kMI_NSLT_BINS || end_slice > Fd::kMI_NSLT_BINS ) {
    FD_CWARN << "TMirrorBitArray::Set() - "
             << "start and/or end slice exceed max. value !" << endl;
    throw std::range_error("slice parameter exceeds max. value");
  }
  
  // need to have access to private variables ...
  
  TMirrorPixelData::PixelDataWord pd = 
    &pixel_data->fPixelData->fPixelData[0][0];
  
  for ( unsigned int col=Fd::kMI_FIRST_COL; col<=Fd::kMI_LAST_COL; col++ ) {
	
    int data = 0;
	  
    for ( unsigned int i=start_slice; i<end_slice; i++ ) {
      data |= pd[i*Fd::kMI_NCOLS + col-Fd::kMI_FIRST_COL].fBits.fPixel;
    }
  
    fBits[col-Fd::kMI_FIRST_COL] = data;
  }
}

// -------------------------------------------------------------------------

void TMirrorBitArray::Set(const TMirrorPixelList* pixel_list)
 {
  for ( unsigned int i=0; i<Fd::kMI_NCOLS; i++ ) fBits[i] = 0;
  
  Fd::PixelNumber pix;
  for ( unsigned int i=0; i<pixel_list->GetNumPixels(); i++ ) {
    pix = pixel_list->GetPixel( i );
    Set( Fd::GetColumnNo( pix ), Fd::GetRowNo( pix ) );
  }
}

// -------------------------------------------------------------------------

void TMirrorBitArray::SetRowData(unsigned int row,int data)
 {
  if ( row < Fd::kMI_FIRST_ROW || row > Fd::kMI_LAST_ROW ) return;
  
  unsigned int mask = 1;
  for ( unsigned int col=0; col<Fd::kMI_NCOLS; col++ ) {
    if ( data && mask )
      Set( col + Fd::kMI_FIRST_COL, row );
    else
      Clear( col + Fd::kMI_FIRST_COL, row );
    mask <<= 1;
  }
}

// -------------------------------------------------------------------------

void TMirrorBitArray::Show(const char *title,char one,char zero) const
 {
  std::ostream& ostr = FD_LOGSTREAM(ErrorLogger::eFatal);
  
  ostr << "-- " << title << " --" << endl;
  ostr << " Number of set bits: " << GetNumberOfBits() << endl;

  for ( int row=0; row<Fd::kMI_NROWS; row++ ) {
    if ( row < 10 )
      ostr << " " << row << ": ";
    else
      ostr << row << ": ";
    for ( int col=0; col<Fd::kMI_NCOLS; col++ ) {
      if ( fBits[col] & (1 << row) ) 
        ostr << one;
      else
        ostr << zero;
    }
    ostr << endl;
  }
}

// -------------------------------------------------------------------------

bool TMirrorBitArray::TestColumn(unsigned int col) const
 {
  if ( col < Fd::kMI_FIRST_COL || col > Fd::kMI_LAST_COL ) return false;
  if ( fBits[col - Fd::kMI_FIRST_COL] )
    return true;
  else
    return false;
}

// -------------------------------------------------------------------------

bool TMirrorBitArray::Test(unsigned int col,unsigned int row) const
 {
  if ( _Check( col, row ) ) {
    if ( BitArrayTest ( fBits, col, row ) )
      return true;
    else
      return false;
  }
  else
    return false;
}

// -------------------------------------------------------------------------

bool TMirrorBitArray::TestRow(unsigned int row) const
 {
  if ( row < Fd::kMI_FIRST_ROW || row > Fd::kMI_LAST_ROW ) return false;
  if ( GetRowData( row ) > 0 )
    return true;
  else
    return false;
}

// -------------------------------------------------------------------------

void TMirrorBitArray::operator|=(const TMirrorBitArray& rhs)
 {
  for ( unsigned int col=0; col<Fd::kMI_NCOLS; col++ )
    fBits[col] |= rhs.fBits[col];
}

// -------------------------------------------------------------------------

void TMirrorBitArray::operator&=(const TMirrorBitArray& rhs)
 {
  for ( unsigned int col=0; col<Fd::kMI_NCOLS; col++ )
    fBits[col] &= rhs.fBits[col];
}

// -------------------------------------------------------------------------

void TMirrorBitArray::operator^=(const TMirrorBitArray& rhs)
 {
  for ( unsigned int col=0; col<Fd::kMI_NCOLS; col++ )
    fBits[col] ^= rhs.fBits[col];
}

// -------------------------------------------------------------------------

TMirrorBitArray& TMirrorBitArray::operator~()
 {
  for ( unsigned int col=0; col<Fd::kMI_NCOLS; col++ )
    fBits[col] = ~fBits[col];
  return *this;
}

ClassImp(TMirrorBitArray)

// -------------------------------------------------------------------------
// -------------------------------------------------------------------------
