
//
// File   : MiBitArray.hh
//
// Purpose: Declaration of a bit array class, this class is indented for
//          projections of the SLT memory
//
// $Id: MiBitArray.hh 8759 2010-09-27 12:16:36Z mathes $
//

#ifndef _MiBitArray_hh_
#define _MiBitArray_hh_


#include <TObject.h>
#include <FdNumbering.hh>

/** @file MiBitArray.hh
  * Declaration of class TMirrorBitArray
  * uthor H.-J. Mathes, FzK
  */
  
// ----- forward declaration

class TMirrorPixelData;
class TMirrorPixelList;



/** Set bit at position col=_c, row=_r in TMirrorBitArray _b.
  */
#define BitArraySet(_b,_c,_r)   \
        _b[_c-FdUtil::Fd::kMI_FIRST_COL] |= (1 << (_r-FdUtil::Fd::kMI_FIRST_ROW))

/** Clear bit at position col=_c, row=_r in TMirrorBitArray _b.
  */
#define BitArrayClear(_b,_c,_r) \
        _b[_c-FdUtil::Fd::kMI_FIRST_COL] &= ~(1 << (_r-FdUtil::Fd::kMI_FIRST_ROW))

/** Test bit at position col=_c, row=_r in TMirrorBitArray _b.
  */
#define BitArrayTest(_b,_c,_r)  \
        ((_b[_c-FdUtil::Fd::kMI_FIRST_COL] & (1 << (_r-FdUtil::Fd::kMI_FIRST_ROW))) ? 1 : 0)

/** Class for bit array similar to the pixel matrix.
  *
  * This class is simply a bit field which represents the pixel matrix. The
  * virtual pixels are not included in this bit field.
  * 
  * It could be useful in order to describe the pixels to be read out or 
  * the pixels which are present in a certain event. Especially this class is
  * intended as a projection of the SLT data.
  * 
  * It is also an alternative representation of the TMirrorPixelList !
  *
  * The counting for the parameters 'column' and 'row' starts from 1 except
  * for the methods where data arrays are passed !
  *
  * Changes:
  *  @li 2002/05/05 (thjm) adopted to the use of the class FdUtil::Fd
  *  @li 2001/10/08 (thjm) added operators ^, ^= and ~, speed increase in 
  *                 Set() method
  *  @li 2001/09/27 (thjm) extended interface of ctor and Set(TMirrorPixelData*)
  *  @li 2001/07/19 (ak) some modifications for MS VC
  *  @li 2001/07/23 (thjm) added Set(TMirrorPixelList*) method
  */
class TMirrorBitArray {
public:
    /** Constructor for a TMirrorBitArray. 
      * All bits of the object are initialized with 0.
      */
    TMirrorBitArray();

    /** Virtual destructor of the class TMirrorBitArray. */
    virtual ~TMirrorBitArray() { }
    
    /** Construct a TMirrorBitArray object from a  TMirrorPixelList object.
      *
      * The TMirrorBitArray object is constructed and initialized with the data 
      * extracted from the TMirrorPixelList. This means that the corresponding 
      * bit for each pixel found in the TMirrorPixelList is set.
      */
    TMirrorBitArray(const TMirrorPixelList*);

    /** Construct a TMirrorBitArray object and initialize with the 
      * TMirrorPixelData objects values.
      *
      * @param start_slice The time slice from which we start to take the data. 
      * @param end_slice The time slice where we stop to take the data. 
      *              The default is Fd::kMI_NSLT_BINS which means that all 
      *              time slices should be or'ed together.
      */
    TMirrorBitArray(const TMirrorPixelData*,unsigned int start_slice = 0,
                    unsigned int end_slice = FdUtil::Fd::kMI_NSLT_BINS);
   
    /** Clear all bits of the TMirrorBitArray object. */
    void Clear();

    /** Clear the bits in column 'col'. */
    void Clear(unsigned int col)
     { fBits[col-FdUtil::Fd::kMI_FIRST_COL] = 0; }

    /** Clear the bit selected by the indices col, row. */
    void Clear(unsigned int col,unsigned int row)
     {
      if ( _Check( col, row ) ) {
        BitArrayClear ( fBits, col, row );
      }
    }

    /** Print the contents of MirrorBitArray verbosely to the specified stream.
      *
      * @param title text to display in the title line
      * @param one character to print for a set bit
      * @param zero character to print for a cleared bit
      */
    void Show(const char* title="TMirrorBitArray",
              char one='1', char zero='0') const;

    /** Extend the bit field by neighbor bits.
      *
      * The parameters 'extcol' and 'extrow' specify by how many columns and 
      * rows bit pattern will be extended. This extension is used to steer a 
      * selective readout of untriggered pixels which may have a sub-threshold 
      * signal.
      * @param extcol number of columns to extend the bit pattern
      * @param extrow number of rows to extend the bit pattern
      */
    void Extend(unsigned int extcol=0,unsigned int extrow=0);
   
    /** Extend the bit field by neighbor bits. 
      *
      * Unlike the Extend() method, this method assumes a hexagonal
      * layout of the bits and the extension is done in terms of crowns.
      *
      * Extending a single bit by one crown yields 6 additional bits whereas
      * extending it by one column and one row yields 8 additional bits.
      *
      * If more than 3 crowns are selected, the 'orthogonal' method is used.
      */
    void ExtendHexagonal(unsigned int extcrown = 1);
    
    /** Get the data of the specified column.
      *
      * @param col column index, range Fd::kMI_FIRST_COL .. Fd::kMI_LAST_COL
      * @return data to for the specified column, bit 0 corresponds to
      *             row 1. Will return 0, if wrong column selected.
      */
    int GetColumnData(unsigned int col) const
     {
      if ( col < FdUtil::Fd::kMI_FIRST_COL || col > FdUtil::Fd::kMI_LAST_COL ) return 0;
      return fBits[col - FdUtil::Fd::kMI_FIRST_COL];
    }

    /** Get the number of bits set in the TMirrorBitArray object. */
    unsigned int GetNumberOfBits() const;

    /** Get the number of bits set in the specified column of the 
      * TMirrorBitArray object.
      */
    unsigned int GetNumberOfBitsInColumn(unsigned int col) const;

    /** Get the number of bits set in the specified row of the 
      * TMirrorBitArray object.
      */
    unsigned int GetNumberOfBitsInRow(unsigned int row) const;

    /** Get the data of the specified row.
      *
      * @param row row index, range Fd::kMI_FIRST_ROW .. Fd::kMI_LAST_ROW
      * @return data to for the specified row, bit 0 corresponds to
      *             row 1. Will return 0, if wrong row selected.
      */
    int GetRowData(unsigned int row) const;

    /** Merge the contents of two TMirrorBitArray objects together. */
    void Merge(TMirrorBitArray*);

    /** Set the bit selected by the indices col, row. */
    void Set(unsigned int col,unsigned int row)
     {
      if ( _Check( col, row ) ) {
        BitArraySet ( fBits, col, row );
      }
    }

    /** Initialize a TMirrorBitArray object with the TMirrorPixelData 
      * objects values.
      *
      * @param start_slice The time slice from which we start to take the data. 
      * @param end_slice The time slice where we stop to take the data. 
      *              The default is Fd::kMI_NSLT_BINS which means that all time 
      *              slices should be or'ed together.
      */
    void Set(const TMirrorPixelData*,
             unsigned int start_slice = 0,
             unsigned int end_slice = FdUtil::Fd::kMI_NSLT_BINS);

    /** Initialize a TMirrorBitArray object with the TMirrorPixelList content.
      */
    void Set(const TMirrorPixelList*);

    /** Set the data of the specified column.
      *
      * @param col column index, range Fd::kMI_FIRST_COL .. Fd::kMI_LAST_COL
      * @param data data to stored for the specified column, bit 0 corresponds 
      *             to row 1
      */
    void SetColumnData(unsigned int col,int data)
     {
      if (   col < FdUtil::Fd::kMI_FIRST_COL
          || col > FdUtil::Fd::kMI_LAST_COL ) return;
      fBits[col - FdUtil::Fd::kMI_FIRST_COL] = data;
    }

    /** Set the data of the specified row.
      *
      * @param row row index, range Fd::kMI_FIRST_ROW .. Fd::kMI_LAST_ROW
      * @param data data to stored for the specified row, bit 0 corresponds to
      *             column 1
      */
    void SetRowData(unsigned int row,int data);

    /** Test if the bit, selected by (col, row), is set. */
    bool Test(unsigned int col,unsigned int row) const;

    /** Test if any bit in column, selected by (col), is set.
      *
      * @param col column index, valid range Fd::kMI_FIRST_COL ... Fd::kMI_LAST_COL
      * @return true if set, false otherwise.
      */ 
    bool TestColumn(unsigned int col) const;

    /** Test if any bit in row, selected by (row), is set.
      *
      * @param row row index, valid range Fd::kMI_FIRST_ROW ... Fd::kMI_LAST_ROW
      * @return kTRUE if set, kFALSE otherwise.
      */ 
    bool TestRow(unsigned int) const;

    /** Perform a bitwise OR with the content of the passed TMirrorBitArray 
      * object and return a reference to the resulting value.
      * @see TMirrorBitArray::Merge()
      */
    void operator|=(const TMirrorBitArray&);

    /** Perform a bitwise OR with the content of the passed TMirrorBitArray 
      * object.
      */
    TMirrorBitArray& operator|(const TMirrorBitArray& rhs)
     { operator|=(rhs); return *this; }

    /** Perform a bitwise AND with the content of the passed TMirrorBitArray 
      * object and return a reference to the resulting value.
      */
    TMirrorBitArray& operator&(const TMirrorBitArray& rhs)
     { operator&=(rhs); return *this; }
    
    /** Perform a bitwise AND with the content of the passed TMirrorBitArray 
      * object.
      */
    void operator&=(const TMirrorBitArray&);
   
    /** Perform a bitwise XOR with the content of the passed TMirrorBitArray 
      * object and return a reference to the resulting value.
      */
    TMirrorBitArray& operator^(const TMirrorBitArray& rhs)
     { operator^=(rhs); return *this; }

    /** Perform a bitwise XOR with the content of the passed TMirrorBitArray 
      * object.
      */
    void operator^=(const TMirrorBitArray&);

    /** Perform a bitwise NOT and return a reference to the resulting value. */
    TMirrorBitArray& operator~(void); 

private:

    /** This is an internal check function which does also the parameter 
      * checking for the col, row indices.
      */
    bool  _Check(unsigned int,unsigned int) const;

    int     fBits[FdUtil::Fd::kMI_NCOLS]; // the matrix of bits
    
    ClassDef(TMirrorBitArray,0)
};

#endif  // _MiBitArray_hh_
