
//
// File   : MiEventFile.cc
//
// Purpose: Implementation of the class TMirrorEventFile
//
// $Id: MiEventFile.cc 8759 2010-09-27 12:16:36Z mathes $
//
 

#include <cstdlib>
#include <cstring>
#include <iostream>

#include <TFile.h>
#include <TString.h>
#include <TTree.h>


/** @file MiEventFile.cc
  * Implementation of class TMirrorEventFile
  * @author H.-J. Mathes, FzK
  */

#include <ErrorLogging.hh>

#include "CheckVersion.hh"

#include "MiEventFile.hh"

using namespace std;

// -------------------------------------------------------------------------
// -------------------------------------------------------------------------
//
// class TMirrorEventFile - implementation
//

static const char *gMirrorTreeName = "EventTree";
static const char *gMirrorBranchName = "EventBranch";

TMirrorEventFile::TMirrorEventFile()
 : TFile("data.root"),
   fBranch(0), fCurrentEvent(0), fEvent(0), 
   fFileName("data.root"), fTree(0)
 { }

// -------------------------------------------------------------------------

TMirrorEventFile::TMirrorEventFile(const char *fname, Option_t *option,
	                           const char *ftitle, Int_t compress)
 : TFile(fname, option, ftitle, compress),
   fBranch(0), fCurrentEvent(0), fEvent(0), 
   fFileName(fname), fTree(0),
   fIsReading(false)
 {
  TString foption = option;
  
  Bool_t create = kFALSE;
  
  if (    !fOption.CompareTo( "NEW", TString::kIgnoreCase )
       || !fOption.CompareTo( "CREATE", TString::kIgnoreCase ) )
       create = kTRUE;
  
  Bool_t recreate = foption.CompareTo( "RECREATE", TString::kIgnoreCase )
                                          ? kFALSE : kTRUE;
  
  // set some variables in case of reading from the file
  
  if ( !create && !recreate ) {
  
    fTree = (TTree *)fFile->Get( gMirrorTreeName );
    if ( !fTree ) {
      FD_CSEV << "TMirrorEventFile::TMirrorEventFile() - "
              << "no TTree with name '" << gMirrorTreeName << "' found !" << endl;
      fNEntries = 0;
    }
    else 
      fNEntries = (Int_t)fTree->GetEntries();
  
    fIsReading = true;
  }
}

// -------------------------------------------------------------------------

TMirrorEventFile::~TMirrorEventFile()
 {
  // fTree is owned by TFile and is deleted by TFile::Close
  // => so don't delete fTree !
  
  // make sure, that file is properly closed (i.e. all written to disk)
  
  if ( fTree ) Close();
}

// -------------------------------------------------------------------------

void TMirrorEventFile::Close(Option_t *option)
 {
  // file was opened in write mode ...
  
  if ( TFile::GetBytesWritten() ) {
    TFile::Write( 0, TObject::kOverwrite );

    fTree   = 0;
    fBranch = 0;
  }
  
  // file was opened in read mode ...

  if ( TFile::GetBytesRead() ) {
    fTree   = 0;
    fBranch = 0;
  }
  
  TFile::Close( option );

  fEvent  = 0;  // now this address is invalid
}

// -------------------------------------------------------------------------

UInt_t TMirrorEventFile::GetNEvents()
 {
  if ( fIsReading ) {  // file was opened for reading
    if ( fTree == 0 ) fTree = (TTree *)fFile->Get( gMirrorTreeName );
    if ( fTree == 0 )
      return 0;
    else
      return (UInt_t)fNEntries;
  }
  else {                          // file was opened for writing
    if ( fTree == 0 )
      return 0;
    else
      return fCurrentEvent;
  }
}

// -------------------------------------------------------------------------

Bool_t TMirrorEventFile::ReadMirrorEvent(TMirrorEvent** event)
 {
  Bool_t valid_event = kFALSE, file_end = kFALSE;
  
  cd();  // set active directory
  
  if ( !fTree ) {
    fTree = (TTree *)fFile->Get( gMirrorTreeName );
    if ( !fTree ) {
      FD_CSEV << "TMirrorEventFile::ReadMirrorEvent() - "
              << "no TTree with name '" << gMirrorTreeName << "' found !" 
	      << endl;
      return kFALSE;
    }
    
    fNEntries = (Int_t)fTree->GetEntries();
  }
  
  fBranch = fTree->GetBranch( gMirrorBranchName );
  if ( !fBranch ) {
    FD_CSEV << "TMirrorEventFile::ReadMirrorEvent() - " 
            << "no TBranch with name '" << gMirrorBranchName << "' found !" 
	    << endl;
    return kFALSE;
  }
  
  while ( fBranch && !valid_event && !file_end ) {
    fBranch->SetAddress( event );
    
    if ( fCurrentEvent < fNEntries ) {
      valid_event = kTRUE;
      
      fTree->GetEvent( fCurrentEvent );
      fCurrentEvent++;
    }
    else {
      FD_CINFO << "TMirrorEventFile::ReadMirrorEvent() - "
               << "End of file reached !" << endl;
      
      file_end = kTRUE;
    }
  }

  if ( valid_event )
    return kTRUE;
  else
    return kFALSE;
}

// -------------------------------------------------------------------------

Bool_t TMirrorEventFile::ReadRunHeader(TMirrorRunHeader** header)
 {
  cd();  // set active directory
  
  if ( *header ) {
    delete *header;
    *header = NULL;
  }
  
#if 0
  TFile::Get( "RunHeader" )->Dump();
#endif
  
  if ( ( *header = (TMirrorRunHeader*)Get( "RunHeader" ) ) == NULL ) {
    FD_CERR << "TMirrorEventFile::ReadRunHeader() - "
            << "reading 'RunHeader' from file was unsuccessful !" << endl;
    return kFALSE;  // no success
  }
  
  // the above cast might fail...
  if ( (*header)->IsA() != TMirrorRunHeader::Class() ) return kFALSE;
  
  return kTRUE;  // object found
}

// -------------------------------------------------------------------------

Bool_t TMirrorEventFile::ReadRunTrailer(TMirrorRunTrailer** trailer)
 {
  cd();  // set active directory
  
  if ( *trailer ) {
    delete *trailer;
    *trailer = NULL;
  }
  
#if 0
  TFile::Get( "RunTrailer" )->Dump();
#endif

  if ( ( *trailer = (TMirrorRunTrailer*)Get( "RunTrailer" ) ) == NULL ) {
    FD_CERR << "TMirrorEventFile::ReadRunTrailer() - "
            << "reading 'RunTrailer' from file was unsuccessful !" << endl;
    return kFALSE;  // no success
  }
  
  // the above cast might fail...
  if ( (*trailer)->IsA() != TMirrorRunTrailer::Class() ) return kFALSE;

  return kTRUE;  // object found
}

// -------------------------------------------------------------------------

#if 0
// produced by rootcint if desired (old scheme)
void TMirrorEventFile::Streamer(TBuffer &R__b)
 {
  // Stream an object of class TMirrorEventFile.

  TFile::Streamer(R__b);
}
#endif

// -------------------------------------------------------------------------

#if 0
// produced by rootcint if desired (new scheme)
void TMirrorEventFile::Streamer(TBuffer &R__b)
 {
  // Stream an object of class TMirrorEventFile.

  if (R__b.IsReading()) {
     TMirrorEventFile::Class()->ReadBuffer(R__b, this);
  } else {
     TMirrorEventFile::Class()->WriteBuffer(R__b, this);
  }
}
#endif

// -------------------------------------------------------------------------

Bool_t TMirrorEventFile::WriteMirrorEvent( TMirrorEvent** event )
 {
  cd();  // set active directory
  
  if ( !fTree ) {
    fEvent = *event;
    
    fTree = new TTree( gMirrorTreeName, "Tree of TMirrorEvents" );
    if ( !fTree ) {
      FD_CSEV << "TMirrorEventFile::WriteMirrorEvent() - "
              << "could not create Tree !" << endl;
      return kFALSE;
    }
    
    fTree->SetAutoSave( 1000000 );  // autosave when 1M written
  
    // create branch in non-split mode !
    // Note: This calls internally TMirrorEvent() and ~TMirrorEvent()
    //       to create the dictionary of TMirrorEvent
  
    fTree->Branch( gMirrorBranchName, "TMirrorEvent", event, 64000, 0 );
  }
  
  if ( *event != fEvent ) {
    FD_CSEV << "TMirrorEventFile::WriteMirrorEvent() - "
            << "address of data buffer between subsequent calls was changed !" 
	    << endl;
    return kFALSE;
  }
  
  fTree->Fill();

  return kTRUE;
}

// -------------------------------------------------------------------------

Bool_t TMirrorEventFile::WriteRunHeader( TMirrorRunHeader* header )
 {
  cd();  // set active directory

  if ( !header ) return kFALSE;
  
  header->Write( "RunHeader", TObject::kOverwrite );
  
  return kTRUE;
}

// -------------------------------------------------------------------------

Bool_t TMirrorEventFile::WriteRunTrailer( TMirrorRunTrailer* trailer )
 {
  cd();  // set active directory

  if ( !trailer ) return kFALSE;
  
  trailer->Write( "RunTrailer", TObject::kOverwrite );
  
  return kTRUE;
}

ClassImp(TMirrorEventFile)

// -------------------------------------------------------------------------
// -------------------------------------------------------------------------
