
//
// File   : MiEventHeader.cc
//
// Purpose: Implementation of Mirror event header class
//
// $Id: MiEventHeader.cc 12216 2011-03-16 21:23:05Z mathes $
//


#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <iostream>
#include <sstream>
#include <iomanip>
#include <stdexcept>
#include <cassert>

/** @file MiEventHeader.cc
  * Implementation of class TMirrorEventHeader.
  * @author H.-J. Mathes, FzK
  */

#include <TClass.h>
#include <TBuffer.h>

#include <ErrorLogging.hh>
#include <FdNumbering.hh>
#include <FdConfig.hh>

#include "CheckVersion.hh"

#include "MiEvent.hh"

using namespace std;
using FdUtil::Fd;
using namespace MiEvent;

// -------------------------------------------------------------------------

// these operators are still missing in ROOT (before 3.05.05)
// 3.05.03 at least does not have these operators ...

#if (ROOT_VERSION_CODE < ROOT_VERSION(3,5,5))
void operator<<(TBuffer& b,unsigned long long ul)
 { Int_t *i = (Int_t *)&ul; b << i[1]; b << i[0]; }
void operator<<(TBuffer& b,long long l)
 { Int_t *i = (Int_t *)&l; b << i[1]; b << i[0]; }
void operator>>(TBuffer& b,unsigned long long ul)
 { Int_t *i = (Int_t *)&ul; b >> i[1]; b >> i[0]; }
void operator>>(TBuffer& b,long long l)
 { Int_t *i = (Int_t *)&l; b >> i[1]; b >> i[0]; }
#endif // ROOT_VERSION_CODE

// -------------------------------------------------------------------------
// -------------------------------------------------------------------------
//
// class TMirrorEventHeader - implementation
//

bool TMirrorEventHeader::fgTimeCorrectionWarningDisabled = false;

TMirrorEventHeader::TMirrorEventHeader()
 : fConstructorType(kCREATE)
 { 
#ifdef TRACE_CONSTRUCTOR
  FD_CINFO << "TMirrorEventHeader(): " << this << endl;
#endif
  Construct();
  Init();
}

// -------------------------------------------------------------------------

TMirrorEventHeader::TMirrorEventHeader(void *ptr)
 : fConstructorType(kMIEVENTHEADER)
 {
#ifdef TRACE_CONSTRUCTOR
  FD_CINFO << "TMirrorEventHeader(void*): " << this << endl;
#endif
  fHeader = (MiEventHeader)ptr;
}

// -------------------------------------------------------------------------

TMirrorEventHeader::TMirrorEventHeader(const TMirrorEventHeader& rhs)
 : TObject(rhs), fConstructorType(kCREATE)
 {
#ifdef TRACE_CONSTRUCTOR
  FD_CINFO << "TMirrorEventHeader(TMirrorEventHeader&): " << this << endl;
#endif
  *this = rhs;
}

// -------------------------------------------------------------------------

TMirrorEventHeader::~TMirrorEventHeader()
 {
#ifdef TRACE_CONSTRUCTOR
  FD_CINFO << "~TMirrorEventHeader(): " << this << endl;
#endif

  switch ( fConstructorType ) {
  
    case kCREATE:
         Destruct();
	 break;

    case kSHMEVENT:
    case kMIEVENTHEADER:
         fHeader = NULL;
	 break;

    default:
         FD_CERR << "TMirrorEventHeader::~TMirrorEventHeader() - "
	         << "unknown fConstructorType= " 
	         << (int)fConstructorType << endl;
  }
}

// -------------------------------------------------------------------------

TMirrorEventHeader& 
TMirrorEventHeader::operator=(const TMirrorEventHeader& rhs)
 { 
  if ( this == &rhs ) return *this;
  
  *fHeader = *rhs.fHeader;
  
  return *this;
}

// -------------------------------------------------------------------------

bool 
TMirrorEventHeader::DoTimeCorrection(unsigned int sec_in,
                                     int nsec_in,
                                     unsigned int * sec_out,
				     unsigned int * nsec_out) const
 {
  assert( sec_out != NULL && nsec_out != NULL );

  if ( IsOldTimingFormat() ) {         // old format
    if ( nsec_in < 0 ) {
      *sec_out = sec_in - 1;
      *nsec_out = nsec_in + kNanoSecPerSecond;
    }
    else {
      *sec_out = sec_in;
      *nsec_out = nsec_in;
    }
    
    return true;
  }
  else if ( IsTimeCorrectionSet() ) {  // new format and corr. set
    
    double real_nano_time =   
           ((double)nsec_in * fHeader->fTimeCorrection.fT10Mhz) / 100.
	 + fHeader->fTimeCorrection.fOffset;
  
    if ( real_nano_time < 0 ) {
      *sec_out = sec_in - 1;
      *nsec_out = (UInt_t)(real_nano_time + (double)kNanoSecPerSecond);
    }
    else {
      *sec_out = sec_in;
      *nsec_out = (UInt_t)real_nano_time;
    }
    
    return true;
  }
  else {
    if ( !fgTimeCorrectionWarningDisabled ) {
      FD_CWARN << "TMirrorEventHeader::DoTimeCorrection() - "
               << "time calculation is not possible !" << endl;
      fgTimeCorrectionWarningDisabled = true;
    }
    
    // use old procedure as best guess ...
    
    if ( nsec_in < 0 ) {
      *sec_out = sec_in - 1;
      *nsec_out = nsec_in + kNanoSecPerSecond;
    }
    else {
      *sec_out = sec_in;
      *nsec_out = nsec_in;
    }
    
    return false;
  }
  
  return false;
}

// -------------------------------------------------------------------------

unsigned int 
TMirrorEventHeader::GetFADCBinNanoTime(TFADCData* fadc, 
                                       unsigned int bin) const
 {
  if ( bin >= fadc->GetNFADCBins() ) {
    FD_CERR << "TMirrorEventHeader::GetFADCBinNanoTime() - "
            << "the specified bin " << bin << " is not valid !" << endl;
    return 0;
  }
  
  int nano_sec =   fHeader->fNextPageNanoTime 
                 - 100 * (fadc->GetNFADCBins() - bin);

  unsigned int sec, nsec;
  
  DoTimeCorrection( fHeader->fNextPageTime, nano_sec, &sec, &nsec );
  
  return nsec;
}

// -------------------------------------------------------------------------

unsigned int 
TMirrorEventHeader::GetFADCBinTime(TFADCData* fadc, unsigned int bin) const
 {
  if ( bin >= fadc->GetNFADCBins() ) {
    FD_CERR << "TMirrorEventHeader::GetFADCBinTime() - "
            << "the specified bin " << bin << " is not valid !" << endl;
    return 0;
  }
  
  int nano_time =   fHeader->fNextPageNanoTime
                  - 100 * (fadc->GetNFADCBins() - bin);
    
  unsigned int sec, nsec;
  
  DoTimeCorrection( fHeader->fNextPageTime, nano_time, &sec, &nsec );
  
  return sec;
}


//void TMirrorEventHeader::SetFADCTraceStartTime(TFADCData* fadc)
// {
//}

// -------------------------------------------------------------------------

// Note: Calculation only valid for data version <= 3 and eventually 4 !

unsigned int TMirrorEventHeader::GetGPSNanoTime() const
 {
  if ( !IsOldTimingFormat() && IsTimeCorrectionSet() ) {
  
    // new calculation method ...
    
    int nano_sec = (int)GetNextPageNanoTime() - GetNextPageDelay() * 100;
  
    if ( nano_sec < 0 )
      // extrapolate the number of ticks in the previous second ...
      return (unsigned int)( nano_sec + 
             (unsigned int)(100. * 1.0E09 / fHeader->fTimeCorrection.fT10Mhz) );
    else
      return (unsigned int)nano_sec;
  }
  else if (   (!IsOldTimingFormat() && !IsTimeCorrectionSet())
            || IsOldTimingFormat() ) {

    if (    !IsOldTimingFormat() && !IsTimeCorrectionSet() 
         && !fgTimeCorrectionWarningDisabled) {
      FD_CWARN << "TMirrorEventHeader::GetGPSNanoTime() - "
               << "time calculation is not exact !" << endl;
      fgTimeCorrectionWarningDisabled = true;
    }
  
    // Note: Calculation only valid for data version <= 3 and eventually 4 !

    int nano_sec = (int)GetNextPageNanoTime() - GetNextPageDelay() * 100;
  
    if ( nano_sec < 0 )
      return (UInt_t)(nano_sec + kNanoSecPerSecond);
    else
      return (UInt_t)nano_sec;
  }
  
  // should never come here !
  
  FD_CSEV << "TMirrorEventHeader::GetGPSNanoTime() - "
          << "entered unaccessible code, line " << __LINE__ << " !" << endl;

  return 0;
}

// -------------------------------------------------------------------------

UInt_t TMirrorEventHeader::GetGPSTime() const
 {
  int nano_sec = (int)GetNextPageNanoTime() - GetNextPageDelay() * 100;

  if ( nano_sec < 0 )
    return fHeader->fNextPageTime - 1;
  else
    return fHeader->fNextPageTime;
}

// -------------------------------------------------------------------------

void TMirrorEventHeader::GetTimeCorrectionParameters(Float_t *t10,
                                                     Float_t *offset) const
 {
  assert( t10 != NULL && offset != NULL );
  
  *t10    = fHeader->fTimeCorrection.fT10Mhz;
  *offset = fHeader->fTimeCorrection.fOffset;
}

// -------------------------------------------------------------------------

const string TMirrorEventHeader::GetVerboseEventLabel(EMiEventLabel label)
 {
  char event_label_str[20];

  switch ( label )
   {
    case kAcceptedEvent:
	 sprintf( event_label_str, "'Accepted'" );
         break;

    case kMuonEvent:
	 sprintf( event_label_str, "'Direct hit'" );
         break;

    case kShowerCandidate:
	 sprintf( event_label_str, "'Shower Candidate'" );
         break;

    case kShowerEvent:
	 sprintf( event_label_str, "'Shower'" );
         break;

    default:
	 sprintf( event_label_str, "'Unlabelled'" );
         break;
  }
  
  return string( event_label_str );
}

// -------------------------------------------------------------------------

const string TMirrorEventHeader::GetVerboseEventType(EMiEventType type)
 {
  char event_type_str[20];
  
  switch ( type )
   {
    case kPhysicsEvent:
	 sprintf( event_type_str, "'Physics'" );
         break;
    
    case kCalibrationEvent:
	 sprintf( event_type_str, "'Calibration'" );
         break;
    
    case kSimulatedEvent:
	 sprintf( event_type_str, "'Simulated'" );
         break;
    
    default:
    case kUnknownEvent:
	 sprintf( event_type_str, "'Unknown'" );
         break;
  }
  
  return string(event_type_str);
}

// -------------------------------------------------------------------------

const string 
TMirrorEventHeader::GetVerboseTriggerSource(EMiTriggerSource source)
 {
  ostringstream trigger_str;
  
  if ( (source & kTriggerMask) == kUnknownTrigger)
    return "Unknown";
    
  if ( source & kSoftwareTrigger )
    trigger_str << "Sw";
  if ( source & kRightTrigger )
    trigger_str << (!trigger_str.str().empty() ? " | " : "") << "R";
  if ( source & kLeftTrigger )
    trigger_str << (!trigger_str.str().empty() ? " | " : "")  << "L";
  if ( source & kExternalTrigger )
    trigger_str << (!trigger_str.str().empty() ? " | " : "")  << "Ext";
#ifdef USE_HW_V4
  if ( source & kFrontTrigger )
    trigger_str << (!trigger_str.str().empty() ? " | " : "")  << "Frn";
#endif
  if ( source & kInternalTrigger )
    trigger_str << (!trigger_str.str().empty() ? " | " : "")  << "Int";

  return trigger_str.str();
}

// -------------------------------------------------------------------------

bool TMirrorEventHeader::IsOldTimingFormat() const
 {
  if ( fHeader->fVersion < 4 ) return true;
  
  // this is around end of April 2002
  
  if ( fHeader->fVersion == 4 && fHeader->fGpsTime < 703800000 ) return true;
  
  // version 5 data: fHeader->fGpsTime might be 0 (when read from file)
  
  return false;
}

// -------------------------------------------------------------------------

void TMirrorEventHeader::Init()
 {
  if ( !fHeader ) {
    FD_CCRIT << "TMirrorEventHeader::Init() - no structure defined !" << endl;
    throw logic_error("TMirrorEventHeader::Init(): no structure defined !");
  }
  
  fHeader->fVersion              = Class_Version(); 

  fHeader->fFLTVersion           = 0;
  fHeader->fSLTVersion           = 0;

  SetEyeNo( FdUtil::GetEyeId() );
  SetMirrorNo( FdUtil::GetMirrorId() );
  
  fHeader->fGpsTime                = 0;
  fHeader->fGpsNanoTime            = 0;
  fHeader->fTriggerNo              = 0;
  fHeader->fEventNo                = 0;

  fHeader->fRunNo                  = 0;
  fHeader->fEventType              = kUnknownEvent;
  fHeader->fTriggerSource          = 0;
  
  fHeader->fEventType              = (int)kUnlabelled;
  fHeader->fT3Id                   = 0;
  
  fHeader->fNextPageDelay          = 0;
  fHeader->fNextPageTime           = 0;
  fHeader->fNextPageNanoTime       = 0;
  
  fHeader->fTriggerTimeStamp       = 0;
  fHeader->fNextPageTimeStampL     = 0;
  fHeader->fNextPageTimeStampH     = 0;

  fHeader->fTimeCorrection.fT10Mhz = 0.0;
  fHeader->fTimeCorrection.fOffset = 0.0;
  
  fHeader->fDeadTimeCounter        = 0LL;
  fHeader->fRunTimeCounter         = 0LL;
  fHeader->fVetoTimeCounter        = 0LL;
  
  fHeader->fDaqDeadTime            = 0LL;
  fHeader->fNSkippedTriggers       = 0;
}

// -------------------------------------------------------------------------

void TMirrorEventHeader::SetNextPageTimes()
 {
//  FD_CDEBUG(1) << "TMirrorEventHeader::SetNextPageTimes() ..." << endl;
  
//  FD_CDEBUG(1) << "\tfGpsTime                 = " << fHeader->fGpsTime << endl;
//  FD_CDEBUG(1) << "\tfGpsNanoTime             = " << fHeader->fGpsNanoTime << endl;
//  FD_CDEBUG(1) << "\tGetNextPageDelay()       = " << GetNextPageDelay() * 100 << endl;

  // calculate time of NxPage from GPS (event time) time ...
#warning SetNextPageTimes() not adapted to new hardware format V4 !
  int nano_sec = (int)fHeader->fGpsNanoTime + GetNextPageDelay() * 100;
  
  if ( nano_sec >= kNanoSecPerSecond ) {
    SetNextPageNanoTime( nano_sec - kNanoSecPerSecond );
    SetNextPageTime( fHeader->fGpsTime + 1 );
  }
  else {
    SetNextPageNanoTime( nano_sec );
    SetNextPageTime( fHeader->fGpsTime );
  }

//  FD_CDEBUG(1) << "\t-> GetNextPageNanoTime() = " << GetNextPageNanoTime() << endl;
//  FD_CDEBUG(1) << "\t-> GetNextPageTime()     = " << GetNextPageTime() << endl;
}

// -------------------------------------------------------------------------

void TMirrorEventHeader::SetTimeCorrectionParameters(Float_t t10,Float_t offs)
 {
  // --- check first if parameters are in reasonable range
  
  if ( t10 < 50. || t10 > 150. ) {
    FD_CERR << "TMirrorEventHeader::SetTimeCorrectionParameters() - "
            << "value passed for 'T10' is not plausible !" << endl;
    return;
  }
  
  // -128 < neg.sawtooth < 127 and 0 <= offset <= 100
  
  if ( offs < -128. || offs > 227. ) {
    FD_CERR << "TMirrorEventHeader::SetTimeCorrectionParameters() - "
            << "value passed for 'offset' is not plausible !" << endl;
    return;
  }
  
  // --- and store them
  
  fHeader->fTimeCorrection.fT10Mhz = t10;
  fHeader->fTimeCorrection.fOffset = offs;
}

// -------------------------------------------------------------------------

void TMirrorEventHeader::Show(std::ostream& ostr) const
 {
  if ( fHeader ) {
    
    string event_type_str = GetVerboseEventType( GetEventType() );
    
    string event_label_str = GetVerboseEventLabel( GetEventLabel() );
        
    Fd::MirrorNumber mi_num = GetMirrorNumber();
    
    ostr << "-- Eventheader of event: " << setw(8)
         << GetEventNo() << " from mirror: "
	 << Fd::GetMirrorNo( mi_num ) << ", eye: "
	 << Fd::GetEyeNo( mi_num ) << " --" << endl; 
    ostr << " Version:      " << setw(10) << GetVersion() << endl;
  
    ostr << " FLT version:  0x" << setfill('0') << hex << setw(8) 
                                << GetFltVersion() 
				<< dec << setfill(' ') << endl;
    ostr << " SLT version:  0x" << setfill('0') << hex << setw(8) 
                                << GetSltVersion() 
				<< dec << setfill(' ') << endl;
    
    ostr << " EventType:    " << event_type_str << endl;
    ostr << " EventLabel:   " << event_label_str << endl;
    ostr << " TrgNo:        " << setw(10) 
         << GetTriggerNo() << endl;
    if ( GetT3Id() )
      ostr << " T3Id:         " << setw(10) 
           << GetT3Id() << endl;
#if 0
    ostr << " UTCTime:      " << setw(10) 
         << GetUTCTime() << " [sec]" << endl;
#endif
    ostr << " GPSTime:      " << setw(10) 
         << GetGPSTime() << " [sec]" << endl;
    ostr << " GPSNanoTime:  " << setw(10) 
         << GetGPSNanoTime() << " [nsec]" << endl;
    ostr << " NxPageDelay:  " << setw(10) 
         << GetNextPageDelay()*100 << " [nsec]"<< endl;
    ostr << " NxPgTime:     " << setw(10) 
         << GetNextPageTime() << " [sec]" << endl;
    ostr << " NxPgNanoTime: " << setw(10) 
         << GetNextPageNanoTime() << " [nsec]" << endl;
    // not supported elsewhere ...
    if ( GetEyeNo() == 5 ) {
      ostr << " AccRUN time:  " << setw(20) 
           << GetRunTimeCounter() << " [* 100ns]" << endl;
    }
    ostr << " AccVETO time: " << setw(20) 
         << GetVetoTimeCounter() << " [* 100ns]" << endl;
    ostr << " AccDead time: " << setw(20) 
         << GetDeadTimeCounter() << " [* 100ns]" << endl;
    
    if ( !IsOldTimingFormat() && IsTimeCorrectionSet() ) {
      ostr << " T_10MHz:      " << setw(10) 
           << fHeader->fTimeCorrection.fT10Mhz << " [nsec]" << endl;
      ostr << " Sec_Offset:   " << setw(10) 
           << fHeader->fTimeCorrection.fOffset << " [nsec]" << endl;
    }
    else if ( !IsOldTimingFormat() )
      ostr << " TimeCorrectionParameters not set !" << endl;
  }
}
 
// -------------------------------------------------------------------------

/** Stream a TMirrorEventHeader object (SA). */
void TMirrorEventHeader::Streamer(TBuffer &R__b)
 {
  unsigned int R__s, R__c;
  
  if ( R__b.IsReading() ) {
  
    Version_t R__v = R__b.ReadVersion(&R__s, &R__c);

    if ( TMirrorEvent::GetPrintLevel() ) {
      FD_CDEBUG(1) << "TMirrorEventHeader::Streamer() - R__v= " << R__v
                   << " R__s= " << R__s << " R__c= " << R__c << endl;
    }
    
    TObject::Streamer( R__b );
    
    if ( R__v > MiEVENTVERSIONv8 ) {
      FD_CERR << "Version mismatch in TMirrorEventHeader::Streamer() !" << endl;
      FD_CERR << "Expexted <= " << MiEVENTVERSIONv8
              << ", found " << R__v << endl;      

#if 0
      // in case of a version mismatch, we should 'eat' the bytes...
      unsigned char dummy_uchar;
      
      for ( unsigned int i=0; i<(R__s - sizeof(R__v)); i++)
        R__b >> dummy_uchar;
    
      R__b.CheckByteCount(R__s, R__c, TMirrorEventHeader::IsA());
#endif

      return;
    }
    
    if ( R__v <= MiEVENTVERSIONv4 ) {
      R__b >> fHeader->fVersion;
      R__b >> fHeader->fMirrorNumber;
      R__b >> fHeader->fGpsTime;
      R__b >> fHeader->fGpsNanoTime;
      R__b >> fHeader->fTriggerNo;
      R__b >> fHeader->fEventNo;
    }
    else { // vers >= MiEVENTVERSIONv5
      R__b >> fHeader->fVersion;
      R__b >> fHeader->fMirrorNumber;
      
      fHeader->fGpsTime = 0;
      fHeader->fGpsNanoTime = 0;
      
      R__b >> fHeader->fTriggerNo;
      R__b >> fHeader->fEventNo;
    }
    
    if ( R__v >= MiEVENTVERSIONv2 && R__v <= MiEVENTVERSIONv4 ) {
      unsigned int dummy_uint;
      R__b >> fHeader->fTriggerTimeStamp;
      R__b >> dummy_uint;  // fLastTriggerTSHigh is obsolete
      R__b >> fHeader->fNextPageTimeStampL;
      R__b >> fHeader->fNextPageTimeStampH;
      R__b >> fHeader->fNextPageDelay;
    }
    else if ( R__v >= MiEVENTVERSIONv5 ) {
      R__b >> fHeader->fTriggerTimeStamp;
      R__b >> fHeader->fNextPageTimeStampL;
      R__b >> fHeader->fNextPageTimeStampH;
      R__b >> fHeader->fNextPageDelay;
    }
    
    if ( R__v >= MiEVENTVERSIONv3 ) {
      R__b >> fHeader->fTriggerSource;
      R__b >> fHeader->fEventType;
      R__b >> fHeader->fRunNo;
    }
    
    if ( R__v >= MiEVENTVERSIONv4 ) {
      R__b >> fHeader->fEventLabel;
      R__b >> fHeader->fT3Id;
      if ( R__v == MiEVENTVERSIONv4 ) {
        unsigned char dummy_uchar[40];
        R__b.ReadFastArray( dummy_uchar, sizeof(dummy_uchar) );
      }
    }

    if ( R__v >= MiEVENTVERSIONv5 ) {
      R__b >> fHeader->fNextPageTime;
      R__b >> fHeader->fNextPageNanoTime;
      R__b >> fHeader->fDeadTimeCounter;
      R__b >> fHeader->fVetoTimeCounter;
      R__b >> fHeader->fTimeCorrection.fT10Mhz;
      R__b >> fHeader->fTimeCorrection.fOffset;
    }
    else {     // calculate some new quantities from old GPS times
      SetNextPageTimes();
      
      fHeader->fDeadTimeCounter = 0LL;
      fHeader->fVetoTimeCounter = 0LL;
    }
    
    if ( R__v >= MiEVENTVERSIONv6 ) {
      R__b >> fHeader->fDaqDeadTime;
      R__b >> fHeader->fNSkippedTriggers;
    }
    
    if ( R__v >= MiEVENTVERSIONv7 ) {
      R__b >> fHeader->fFLTVersion;
      R__b >> fHeader->fSLTVersion;
    }
    else {
      fHeader->fFLTVersion = 0;
      fHeader->fSLTVersion = 0;
    }
    
    if ( R__v >= MiEVENTVERSIONv8 ) {
      R__b >> fHeader->fRunTimeCounter;
    }
    else {
      fHeader->fRunTimeCounter = 0LL;
    }
    
    R__b.CheckByteCount(R__s, R__c, TMirrorEventHeader::IsA());
  }
  else {
    R__b.WriteVersion(TMirrorEventHeader::IsA());
    
    TObject::Streamer( R__b );

    // modified for MiEVENTVERSIONv5

    R__b << fHeader->fVersion;
    R__b << fHeader->fMirrorNumber;
//    R__b << fHeader->fGpsTime;
//    R__b << fHeader->fGpsNanoTime;
    R__b << fHeader->fTriggerNo;
    R__b << fHeader->fEventNo;
    
    // new for MiEVENTVERSIONv2
    // modified for MiEVENTVERSIONv5
    
    R__b << fHeader->fTriggerTimeStamp;
//    R__b << fHeader->fLastTriggerTSHigh;
    R__b << fHeader->fNextPageTimeStampL;
    R__b << fHeader->fNextPageTimeStampH;
    R__b << fHeader->fNextPageDelay;
    
    // new for MiEVENTVERSIONv3
    
    R__b << fHeader->fTriggerSource;
    R__b << fHeader->fEventType;
    R__b << fHeader->fRunNo;
    
    // new for MiEVENTVERSIONv4
    // modified for MiEVENTVERSIONv5

    R__b << fHeader->fEventLabel;
    R__b << fHeader->fT3Id;
//    // << insert here future data and reduce the number 40 = sizeof(fHeader->fDummy)
//    R__b.WriteFastArray( fHeader->fDummy, sizeof(fHeader->fDummy) );

    // new for MiEVENTVERSIONv5
    
    R__b << fHeader->fNextPageTime;
    R__b << fHeader->fNextPageNanoTime;
    R__b << fHeader->fDeadTimeCounter;
    R__b << fHeader->fVetoTimeCounter;
    R__b << fHeader->fTimeCorrection.fT10Mhz;
    R__b << fHeader->fTimeCorrection.fOffset;

    // new for MiEVENTVERSIONv6
    
    R__b << fHeader->fDaqDeadTime;
    R__b << fHeader->fNSkippedTriggers;
    
    // new for MiEVENTVERSIONv7
    
    R__b << fHeader->fFLTVersion;
    R__b << fHeader->fSLTVersion;
    
    // new for MiEVENTVERSIONv8
    
    R__b << fHeader->fRunTimeCounter;
  }
}

ClassImp(TMirrorEventHeader)

// -------------------------------------------------------------------------
// -------------------------------------------------------------------------
