
//
// File   : MiFadcData.hh
//
// Purpose: Declaration of the Mirror FADC data class TMirrorFADCData
//
// $Id: MiFadcData.hh 12082 2011-02-21 17:11:29Z mathes $
//


#ifndef _MiFadcData_hh_
#define _MiFadcData_hh_

// --> prevent users from including that file directly
#ifndef __CINT__
# if (!defined _MiEvent_hh_) && (!defined _VMiEvent_hh_)
#  error You are not supposed to include that header file !
#  error Please use MiEvent.hh instead !
# endif // !_MiEvent_hh_ && !_VMiEvent_hh_
#endif // __CINT__

#include <iostream>
#include <cstring>

#include <TObject.h>

#include <FdNumbering.hh>
#include <MiEventTypes.hh>
#include <FadcData.hh>

/** @file MiFadcData.hh
  * Declaration of class TMirrorFADCData.
  * @author H.-J. Mathes, FzK
  */

// --- forward declarations

class TFADCData;

/** FADC data of the complete camera.
  *
  * This class describes the FADC raw data of the complete camera, 
  * consisting of:
  * @li a data format version information
  * @li the actual mirror/camera number
  * @li the FADCData (MI_NFADC_COLS=1000 time bins) for each individual 
  *     pixel
  *
  * @todo Add eventually more access methods like:
  *  @li GetFADCDataWordPtr()
  */
class TMirrorFADCData : public TObject {

    friend class TMirrorEvent;
    friend class TShmEvent;
    friend class ShmEventDescriptor;
    friend bool TestMirrorStructSizes();  // test code

protected:

    /** FADC data of PMT matrix.
      *
      * This structure contains the FADC data of the complete PMT matrix
      * organized as an array.
      */
    typedef struct _MiFADCData {
    
      /** Internally used structure version information. */
      Version_t         fVersion;

      /** Array of [Fd::kMI_NEXTROWS][Fd::kMI_NCOLS] structures describing the 
        * individual pixels FADC data. Here the virtual channels are also 
	* taken into account.
	*/
      TFADCData::FADCDataRec       fFadc[FdUtil::Fd::kMI_NEXTROWS][FdUtil::Fd::kMI_NCOLS];
  
    } MiFADCDataRec, *MiFADCData;

public:

    /** Default constructor for class TMirrorFADCData.
      *  
      * This constructor allocates space for the data using the C-function
      * MiFADCDataCreate().
      */
    TMirrorFADCData();

    /** Constructor from void pointer (useful when in SHM). */
    TMirrorFADCData(void *);

    /** Copy constructor of the class TMirrorFADCData. */
    TMirrorFADCData(const TMirrorFADCData&);

    /** Destructor of the class TMirrorFADCData. 
      *
      * The destructor destroys the object in a way which is compatible to 
      * the specific method of its construction.
      */
    virtual ~TMirrorFADCData();

    /** operator=() for class TMirrorFADCData.
      *
      * This operator has to be defined in order to perform a deep copy of
      * a TMirrorFADCData object.
      */
    TMirrorFADCData& operator=(const TMirrorFADCData& rhs)
     {
      if ( this == &rhs ) return *this;
      memcpy( fMiFADCData, rhs.fMiFADCData, sizeof( *fMiFADCData ) );
      return *this;
    }

    /** Clear the contents of a TMirrorFADCData object. */
    virtual void Clear(Option_t *o="");

    /** Print the contents of the TMirrorFADCData object to the passed stream. */
    void Show(std::ostream& ostr=std::cout) const;
    
    /** Print the contents of the TMirrorFADCData object to the passed stream. 
      *
      * Use the content of the passed TMirrorPixelList to select the pixels 
      * whose TFADCData is to be dumped.
      */
    void Show(TMirrorPixelList*,std::ostream& ostr=std::cout) const;
    
    /** Print the contents of the TMirrorFADCData object to the passed stream. 
      *
      * Use the content of the passed TMirrorPixelData (=triggered pixels only)
      * to select the pixels whose TFADCData is to be dumped.
      */
    void Show(TMirrorPixelData*,std::ostream& ostr=std::cout) const;
    
    /** Get a pointer to the single FADC data of an individual pixel
      * which is implemented at present as a C-struct.
      *
      * Regard numbering scheme (FdNumbering.hh) to apply the correct numbers
      * in order to access the data at the proper coordinates.
      */
    TFADCData::FADCData GetFADCData(UInt_t n_col, UInt_t n_row) const
     { return _GetFADCData( n_col, n_row ); }

    /** Get pointer to the single FADC data of an individual pixel
      * which is implemented at present as C-struct.
      *
      * PixelNumber pixel has to be created according to the specification in
      * FdNumbering.h.
      */
    TFADCData::FADCData GetFADCData(FdUtil::Fd::PixelNumber pixel) const
     { return _GetFADCData( FdUtil::Fd::GetColumnNo( pixel ), 
			    FdUtil::Fd::GetRowNo( pixel ) ); }

    /** Return a pointer to the internal FADC data array of type
      * @ref FADCDataWordRec in order to make a fast access to this data
      * array possible.
      */
    TFADCData::FADCDataWord 
    GetFADCDataArray(unsigned int n_col,unsigned int n_row)
     { return FADCDataGetFADCDataWord( GetFADCData(n_col, n_row), 0 ); }

    /** Return a pointer to the internal FADC data array of type
      * @ref FADCDataWordRec in order to make a fast access to this data
      * array possible.
      *
      * PixelNumber pixel has to be created according to the specification in
      * FdNumbering.h.
      */
    TFADCData::FADCDataWord GetFADCDataArray(FdUtil::Fd::PixelNumber pixel)
     {
      return FADCDataGetFADCDataWord( 
             GetFADCData( FdUtil::Fd::GetColumnNo( pixel ), 
	                  FdUtil::Fd::GetRowNo( pixel ) ), 0 );
    }

    /** Get a pointer to C-struct single pixel monitoring data.
      * 
      * The PixelMonitorData is referenced by its column and row indices.
      * The valid ranges for these indices are:
      * @li column: MI_FIRST_COL ... MI_LAST_COL
      * @li row: MI_FIRST_ROW ... MI_LAST_ROW, MI_LAST_VCROW including the
      *     virtual channels.
      */
    TFADCData::PixelMonitorData GetMonitorData(UInt_t n_col, UInt_t n_row)
     {
      if ( !IsValidPixel( n_col, n_row ) )
        return NULL;
      else
        return FADCDataGetPixelMonitorData ( GetFADCData ( n_col, n_row ) );
    }

    /** Get a pointer to C-struct single pixel monitoring data.
      * 
      * The PixelMonitorData is referenced by its PixelNumber which must be
      * constructed properly from column and row indices or which was extracted
      * from the pixel list.
      */
    TFADCData::PixelMonitorData GetMonitorData(FdUtil::Fd::PixelNumber pixel)
     { return GetMonitorData( FdUtil::Fd::GetColumnNo( pixel ), 
                              FdUtil::Fd::GetRowNo( pixel ) ); }

    /** Get a pointer to class TFADCData describing single FADC data.
      *
      * The class is created inside the function (via new()). Thus it has to 
      * be released afterwards (via delete()).
      * 
      * The TFADCData is referenced either by its column and row indices
      * or by the PixelNumber argument (see next method).<br>
      * The valid ranges for these indices are:
      * @li column: MI_FIRST_COL ... MI_LAST_COL
      * @li row: MI_FIRST_ROW ... MI_LAST_ROW, MI_LAST_VCROW including the
      *     virtual channels.
      * 
      * @todo Discuss: is new()/delete() of TFADCData objects a good idea 
      *       (concerning efficiency) ?
      */
    TFADCData* GetPixelFADCData(UInt_t n_col, UInt_t n_row) const
     {
      if ( !IsValidPixel( n_col, n_row ) )
        return NULL;
      else
        return new TFADCData( GetFADCData( n_col, n_row ) );
    }

    /** Get a pointer to class TFADCData describing single FADC data.
      * 
      * The TFADCData is referenced by its PixelNumber.
      */
    TFADCData* GetPixelFADCData(FdUtil::Fd::PixelNumber pixel) const
     { return GetPixelFADCData( FdUtil::Fd::GetColumnNo( pixel ), 
                                FdUtil::Fd::GetRowNo( pixel ) ); }
   
    /** Get the information about FADC trace specific parameters.
      *
      * This information comprises of the start and the end bin of the trace 
      * and the length of the trace. Access them with the macros/functions:
      * @li FADCTraceInfoSetStartBin(), FADCTraceInfoStartBin()
      * @li FADCTraceInfoSetEndBin(), FADCTraceInfoEndBin()
      * @li FADCTraceInfoTraceLength()
      *
      * The valid ranges for the indices column and row are:
      * @li column: MI_FIRST_COL ... MI_LAST_COL
      * @li row: MI_FIRST_ROW ... MI_LAST_ROW, MI_LAST_VCROW including the
      *     virtual channels.
      */
    TFADCData::FADCTraceInfo 
    GetFADCTraceInfo(unsigned int n_col,unsigned int n_row) const
     {
      if ( !IsValidPixel( n_col, n_row ) )
        return NULL;
      else
        return FADCDataGetTraceInfo( GetFADCData( n_col, n_row ) );
    }

    /** Get the information about FADC trace specific parameters. */
    TFADCData::FADCTraceInfo GetFADCTraceInfo(FdUtil::Fd::PixelNumber pixel)
     { return GetFADCTraceInfo( FdUtil::Fd::GetColumnNo( pixel ), 
                                FdUtil::Fd::GetRowNo( pixel ) ); }

    /** Return the version information stored internally in the C-struct. */
    Version_t GetVersion() const
     { return fMiFADCData->fVersion; }
    
    /** Set the data for the specified pixel using a TFADCData object.
      * 
      * The pixel itself is referenced by its PixelNumber.
      */
    void SetPixelFADCData(FdUtil::Fd::PixelNumber,const TFADCData*);
    
protected:
    
    /** Access the complete FADC data of the specified pixel (column,row).
      * It is checked if the indices are within the allowed parameter range,
      * otherwise NULL is returned.
      */
    TFADCData::FADCData _GetFADCData(unsigned int column, unsigned int row) const;
   
    /** Init the MiFADCData struct partially.
      *
      * This method sets:
      * @li the version number in the struct.
      * @li the length of all ADC traces to MI_NBINS.
      * @li the pixel numbering fields of the FADCData sub-structs.
      * @li all monitoring data fields to 0.
      *
      * @todo
      * @li Initialize mirror and eye numbers correctly
      */
    void Init();
    
private:

    /**  */
    void Construct();
    
    /**  */
    void Destruct();

    /** Check if the pixel addressed by col and row is valid, i.e. correctly
      * addressed.
      */
    bool IsValidPixel(unsigned int,unsigned int) const;

    MiEvent::Constructor_t    fConstructorType;     //! how object was created
    MiFADCData                fMiFADCData;          //! pointer to C-struct
   
    ClassDef(TMirrorFADCData,MiEVENTVERSIONv3)
};

#endif  // _MiFadcData_hh_
