
//
// File   : MiPixelData.cc
//
// Purpose: Implementation of Mirror pixel data class TMirrorPixelData
//
// $Id: MiPixelData.cc 11974 2011-01-28 14:33:56Z mathes $
//


#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <iostream>
#include <algorithm>
#include <cassert>


/** @file MiPixelData.cc
  * Implementation of class TMirrorPixelData
  * @author H.-J. Mathes, FzK
  */

#include <TClass.h>
#include <TBuffer.h>

#include <ErrorLogging.hh>
#include <FdNumbering.hh>
#include <FdConfig.hh>

#include "CheckVersion.hh"

#include "MiEvent.hh"

using namespace std;
using FdUtil::Fd;
using namespace MiEvent;

// -------------------------------------------------------------------------
// -------------------------------------------------------------------------
//
// class TMirrorPixelData - implementation
//

const unsigned int TMirrorPixelData::kSLTBinWidth;

const int TMirrorPixelData::kNoTrigger;

TMirrorPixelData::TMirrorPixelData()
 : fConstructorType(kCREATE)
 {
#ifdef TRACE_CONSTRUCTOR
  FD_CINFO << "TMirrorPixelData(): " << this << endl;
#endif // TRACE_CONSTRUCTOR
  Construct();
  Init();
}

// -------------------------------------------------------------------------

TMirrorPixelData::TMirrorPixelData(void *ptr)
 : fConstructorType(kMIPIXELDATA)
 { 
#ifdef TRACE_CONSTRUCTOR
  FD_CINFO << "TMirrorPixelData(void*): " << this << endl;
#endif // TRACE_CONSTRUCTOR
  fPixelData = (MiPixelData)ptr;
}

// -------------------------------------------------------------------------

TMirrorPixelData::TMirrorPixelData(const TMirrorPixelData& rhs)
 : TObject(rhs), fConstructorType(kCREATE)
 {
#ifdef TRACE_CONSTRUCTOR
  FD_CINFO << "TMirrorPixelData(TMirrorPixelData&): " << this << endl;
#endif // TRACE_CONSTRUCTOR
  Construct();
  
  *this = rhs;
}

// -------------------------------------------------------------------------

TMirrorPixelData& TMirrorPixelData::operator=(const TMirrorPixelData& rhs)
 {
  if ( this == &rhs ) return *this;
 
  memcpy( fPixelData, rhs.fPixelData, sizeof( *fPixelData ) );
  
  return *this;
}

// -------------------------------------------------------------------------

TMirrorPixelData::~TMirrorPixelData()
 {
#ifdef TRACE_CONSTRUCTOR
  FD_CINFO << "~TMirrorPixelData(): " << this << endl;
#endif // TRACE_CONSTRUCTOR

  switch ( fConstructorType ) {
    case kCREATE:
         Destruct();

    case kSHMEVENT:
    case kMIPIXELDATA:
         fPixelData = NULL;
	 break;
	 
    default:
         FD_CERR << "TMirrorPixelData::~TMirrorPixelData() - "
	         << "unknown fConstructorType= " 
	         << (int)fConstructorType << endl;
  }
}

// -------------------------------------------------------------------------

void TMirrorPixelData::Convert(const MultiplicityPageDataWord src,
                                     MultiplicityDataWord dest)
 {
  memset(dest, 0, FdUtil::Fd::kMI_NMULT_BINS*sizeof(MultiplicityDataWordRec));
  
  for ( unsigned int i=0; i<Fd::kMI_NMULT_BINS; i++ ) {
    dest[i] = (src[i].fBits.fMultiplicity > kMaxMultiplicityV4) ? 
                   (MultiplicityDataWordRec)kMaxMultiplicityV4 :
		   (MultiplicityDataWordRec)src[i].fBits.fMultiplicity;
  }
}
 
// -------------------------------------------------------------------------

void TMirrorPixelData::ConvertToV4(const TMirrorPixelData * src,
                                         TMirrorPixelData * dest)
 {
  MiPixelData src_p = src->fPixelData;
  MiPixelData dest_p = dest->fPixelData;
  
  // --- clear initially all arrays
  
  memset( dest_p->fPixelData, 0, Fd::kMI_NSLT_BINS*Fd::kMI_NCOLS );
  memset( dest_p->fMultiplicityData, 0, Fd::kMI_NMULT_BINS );
  memset( dest_p->fSLTPageData, 0, Fd::kMI_NSLT_BINS*Fd::kMI_NCOLS );
  memset( dest_p->fMultiplicityPageData, 0, Fd::kMI_NMULT_BINS );
  
  // --- convert multiplicity data
  
  memcpy( dest_p->fMultiplicityData, src_p->fMultiplicityData, 
                                     Fd::kMI_NMULT_BINS );

  for ( unsigned int bin=0; bin<Fd::kMI_NMULT_BINS; bin++ )
    dest_p->fMultiplicityPageData[bin].fBits.fMultiplicity = 
                                           src_p->fMultiplicityData[bin];

  // open: fMultiplicityPageData[i].fBits.fProjection
  
  // --- convert pixel data (SLT) - size is identical
  
  memcpy( dest_p->fPixelData, 
          src_p->fPixelData, Fd::kMI_NSLT_BINS*Fd::kMI_NCOLS );
  
  // --- convert SLT page data (identical to old data)
  
  for ( unsigned int col=0; col<Fd::kMI_NCOLS; col++ ) {
    for ( unsigned int bin=0; bin<Fd::kMI_NSLT_BINS; bin++ ) {
    
      dest_p->fPixelData[bin][col] = src_p->fPixelData[bin][col];
      
      dest_p->fSLTPageData[bin][col].fBits.fExtPixelTrigger =
                          src_p->fPixelData[bin][col].fBits.fPixel;
      dest_p->fSLTPageData[bin][col].fBits.fPattern = 
                          src_p->fPixelData[bin][col].fBits.fPatternClass;
      dest_p->fSLTPageData[bin][col].fBits.fGlobTrigger = 
                          src_p->fPixelData[bin][col].fBits.fExtTrigger;
    } // bin
  } // col

#if 0
  // --- convert PixelPageData

  dest_p->fPixelPageData[][] = ...;
#endif
}

// -------------------------------------------------------------------------

/** @return 1 (ODD) or 0 (EVEN) parity. 
  *
  * @see http://groups.google.com/group/comp.lang.c/msg/e65b8577d041435b?hl=en&
  */
static int CalculateParity(unsigned int x)
 {
  x = x ^ x >> 16;
  x = x ^ x >> 8;
  x = x ^ x >> 4;
  x = x ^ x >> 2;
  x = x ^ x >> 1;
 
  return x & 1;
}

void TMirrorPixelData::ConvertToV3(const SLTPageDataWord src,PixelDataWordV3 dest)
 {
  memset( dest, 0, Fd::kMI_NSLT_BINS*Fd::kMI_NCOLS );

  for ( unsigned int i=0; i<Fd::kMI_NSLT_BINS*Fd::kMI_NCOLS; i++ ) {
    dest[i].fPixel = src[i].fBits.fExtPixelTrigger;
    dest[i].fParity = CalculateParity( dest[i].fPixel );
    dest[i].fPatternClass = src[i].fBits.fPattern;
    //dest[i].fParityError = 0;
    dest[i].fExtTrigger = src[i].fBits.fGlobTrigger;
  } // col
}

// -------------------------------------------------------------------------

void TMirrorPixelData::ConvertToV4(const MultiplicityDataWordV3 src,
                                         MultiplicityDataWord dest)
 {
  memset(dest, 0, FdUtil::Fd::kMI_NMULT_BINS*sizeof(MultiplicityDataWordRec));
  
  for ( unsigned int i=0; i<Fd::kMI_NMULT_BINS; i++ ) {
    dest[i] = (src[i] > kMaxMultiplicityV3) ? 
                   (MultiplicityDataWordRec)kMaxMultiplicityV3 :
		   (MultiplicityDataWordRec)src[i];
  }
}

// -------------------------------------------------------------------------

void TMirrorPixelData::CreateMultiplicityData(TMirrorFADCData* fadc,
                                              TMirrorPixelList* plist,
				              MultiplicityDataWord multid)
 {
  assert( multid != NULL );

  unsigned int fadc_bin_increment = 1;
  unsigned int n_fadc_bins;
  
  memset(multid, 0, FdUtil::Fd::kMI_NMULT_BINS*sizeof(MultiplicityDataWordRec)); 
  
  Fd::PixelNumber pixel;
  TFADCData::FADCDataWord fadc_data;
  TFADCData::FADCTraceInfo trace_info;
  
  // --- loop over all pixels
  
  for ( unsigned int ipix=0; ipix<plist->GetNumPixels(); ipix++ ) {
    
    pixel = plist->GetPixel( ipix );
    
    // do not proceed with empty (= not triggered) traces ...
    //if ( GetFirstTriggeredTimeBin( pixel ) == kNoTrigger ) continue;
    
    fadc_data = fadc->GetFADCDataArray( pixel );
    trace_info = fadc->GetFADCTraceInfo( pixel );

    n_fadc_bins = TFADCData::GetNFADCBins( FADCTraceGetResolution(trace_info) );
    
#if 0
    cout << n_fadc_bins << " " 
         << FADCTraceGetStartBin(trace_info) << " "
	 << FADCTraceGetEndBin(trace_info) << endl;
#endif

    fadc_bin_increment = n_fadc_bins / Fd::kMI_NMULT_BINS;
    
    for ( unsigned int i=FADCTraceGetStartBin(trace_info); 
                       i<=FADCTraceGetEndBin(trace_info); 
		       i+=fadc_bin_increment ) {
    
      // increase multiplicity for each triggered pixel
      // V3: maximum multiplicity of hardware is limited to 63
      // V4: maximum multiplicity of hardware is limited to 440
      if (    FADCDataWordGetPixelTrigger(&fadc_data[i])
           && (multid[i] < kMaxMultiplicity) )
	multid[i]++;
    }
  }

  if ( fadc_bin_increment != 1 ) {
    
    FD_CWARN << "TMirrorPixelData::CreateMultiplicityData() - "
             << "might fail, kMI_NMULT_BINS != GetNFADCBins() !" 
	     << endl;

    FD_CWARN << "TMirrorPixelData::CreateMultiplicity() - "
             << "calculating multiplicity from every " << fadc_bin_increment
	     << "nd bin!" << endl;
  }
  
}

// -------------------------------------------------------------------------

int 
TMirrorPixelData::GetFirstTriggeredTimeBin(unsigned int col,
                                           unsigned int row,
                                           unsigned int first_bin,
					   unsigned int last_bin) const
 {
  assert( fPixelData != NULL );
  
  // --- check column and row indices
  
  if ( col < Fd::kMI_FIRST_COL || col > Fd::kMI_LAST_COL ) {
    FD_CERR << "TMirrorPixelData::GetFirstTriggeredTimeBin() - "
            << "column index " << col << " is invalid !" << endl;
    return kNoTrigger;
  }
  
  if ( row < Fd::kMI_FIRST_ROW || row > Fd::kMI_LAST_ROW ) {
    FD_CERR << "TMirrorPixelData::GetFirstTriggeredTimeBin() - "
            << "row index " << row << " is invalid !" << endl;
    return kNoTrigger;
  }

  // --- check bin indices

  if ( first_bin >= Fd::kMI_NSLT_BINS ) {
    FD_CERR << "TMirrorPixelData::GetFirstTriggeredTimeBin() - "
            << "first bin index " << first_bin << " is invalid !" << endl;
    return kNoTrigger;
  }
  
  if ( last_bin > Fd::kMI_NSLT_BINS ) {
    FD_CWARN << "TMirrorPixelData::GetFirstTriggeredTimeBin() - "
             << "last bin index " << last_bin << " too large, corrected !" 
	     << endl;
  }
  
  unsigned int start_bin = first_bin;
  unsigned int end_bin = min( last_bin, (unsigned int)(Fd::kMI_NSLT_BINS - 1) );
  
  if ( end_bin < start_bin ) {
    FD_CWARN << "TMirrorPixelData::GetFirstTriggeredTimeBin() - "
             << "bin indices are in wrong order, I swap them !" << endl;
    swap( start_bin, end_bin );
  }
  
  unsigned int row_mask = 1 << (row - Fd::kMI_FIRST_ROW);
  
  for ( unsigned int bin=start_bin; bin<=end_bin; bin++ ) {
    TMirrorPixelData::PixelDataWord word = 
                      &fPixelData->fPixelData[bin][col-Fd::kMI_FIRST_COL];
    if ( PixelDataGetPixel( word ) & row_mask ) {
//      FD_CDEBUG(1) << "First triggered time bin for col= " << col 
//           << " , row= " << row << " is: " << bin << endl;

      // last one is invalid, too
      if ( bin == Fd::kMI_NSLT_BINS ) return kNoTrigger;

      return bin * 10;
    }
  }
  
  return kNoTrigger;
}

// -------------------------------------------------------------------------

int 
TMirrorPixelData::GetFirstTriggeredTimeBin(Fd::FdPixelNumber fd_pixel,
                                           unsigned int /* first_bin */,
					   unsigned int /* last_bin */) const
 {
  Fd::PixelNumber pixel = Fd::GetPixelNumber( fd_pixel );

  assert( fPixelData != NULL );
  
  // --- check eye and mirror numbers

#if 1
  if ( Fd::GetEyeNo( &fPixelData->fMirrorNumber ) != Fd::GetEyeNo( pixel ) ) {
    return kNoTrigger;
  }
  
  if ( Fd::GetMirrorNo( &fPixelData->fMirrorNumber ) != Fd::GetMirrorNo( pixel ) ) {
    return kNoTrigger;
  }
#endif
  
  return GetFirstTriggeredTimeBin( Fd::GetColumnNo( pixel ),
                                   Fd::GetRowNo( pixel ) );
}

// -------------------------------------------------------------------------

int 
TMirrorPixelData::GetLastTriggeredTimeBin(unsigned int col, 
                                          unsigned int row,
                                          unsigned int first_bin,
					  unsigned int last_bin) const
 {
  // --- check column and row indices
  
  if ( col < Fd::kMI_FIRST_COL || col > Fd::kMI_LAST_COL ) {
    FD_CERR << "TMirrorPixelData::GetFirstTriggeredTimeBin() - "
            << "column index " << col << " is invalid !" << endl;
    return -1;
  }
  
  if ( row < Fd::kMI_FIRST_ROW || row > Fd::kMI_LAST_ROW ) {
    FD_CERR << "TMirrorPixelData::GetFirstTriggeredTimeBin() - "
            << "row index " << row << " is invalid !" << endl;
    return -1;
  }

  // --- check bin indices

  if ( first_bin >= Fd::kMI_NSLT_BINS ) {
    FD_CERR << "TMirrorPixelData::GetFirstTriggeredTimeBin() - "
            << "first bin index " << first_bin << " is invalid !" << endl;
    return -1;
  }
  
  if ( last_bin > Fd::kMI_NSLT_BINS ) {
    FD_CWARN << "TMirrorPixelData::GetFirstTriggeredTimeBin() - "
             << "last bin index " << last_bin << " too large, corrected !" 
	     << endl;
  }
  
  unsigned int start_bin = first_bin;
  unsigned int end_bin = min( last_bin, (unsigned int)(Fd::kMI_NSLT_BINS - 1));
  
  if ( end_bin < start_bin ) {
    FD_CWARN << "TMirrorPixelData::GetFirstTriggeredTimeBin() - "
             << "bin indices are in wrong order, I swap them !" << endl;
    swap( start_bin, end_bin );
  }
  
  assert( fPixelData != NULL );
  
  unsigned int row_mask = 1 << (row - Fd::kMI_FIRST_ROW);
  
  for ( unsigned int bin=end_bin; bin>=start_bin; bin-- ) {
    TMirrorPixelData::PixelDataWord word = 
                      &fPixelData->fPixelData[bin][col-Fd::kMI_FIRST_COL];
    if ( PixelDataGetPixel( word ) & row_mask ) {
//      FD_CDEBUG(1) << "Last triggered time bin for col= " << col 
//           << " , row= " << row << " is: " << bin << endl;

      if ( bin == 0 ) return -1; // the first one is invalid

      return bin;
    }
  }
  
  return -1;
}

// -------------------------------------------------------------------------

TMirrorPixelData::MultiplicityDataWord 
TMirrorPixelData::GetMultiplicityData(unsigned int bin) const
 {
  if ( bin >= Fd::kMI_NMULT_BINS ) {
    FD_CERR << "TMirrorPixelData::GetMultiplicityData() - "
            << "bin index out of range !" << endl;
    return NULL;
  }
  
  return GetMultiplicityDataFast( bin );
}

// -------------------------------------------------------------------------

TMirrorPixelData::MultiplicityPageDataWord 
TMirrorPixelData::GetMultiplicityPageData(unsigned int bin) const
 {
  if ( bin >= Fd::kMI_NMULT_BINS ) {
    FD_CERR << "TMirrorPixelData::GetMultiplicityPageData() - "
            << "bin index out of range !" << endl;
    return NULL;
  }
  
  return GetMultiplicityPageDataFast( bin );
}

// -------------------------------------------------------------------------

TMirrorPixelData::PixelDataWord 
TMirrorPixelData::GetPixelData(unsigned int column,unsigned int bin) const
 {
  if ( column < Fd::kMI_FIRST_COL || column > Fd::kMI_LAST_COL ) {
    FD_CERR << "TMirrorPixelData::GetPixelData() - "
            << "column index " << column << " out of range !" << endl;
    return NULL;
  }
  
  if ( bin >= Fd::kMI_NSLT_BINS ) {
    FD_CERR << "TMirrorPixelData::GetPixelData() - "
            << "bin index " << bin << " out of range !" << endl;
    return NULL;
  }
  
  return GetPixelDataFast( column, bin );
}

// -------------------------------------------------------------------------

TMirrorPixelData::SLTPageDataWord 
TMirrorPixelData::GetSLTPageData(unsigned int column,unsigned int bin) const
 {
  if ( column < Fd::kMI_FIRST_COL || column > Fd::kMI_LAST_COL ) {
    FD_CERR << "TMirrorPixelData::GetSLTPageData() - "
            << "column index " << column << " out of range !" << endl;
    return NULL;
  }
  
  if ( bin >= Fd::kMI_NSLT_BINS ) {
    FD_CERR << "TMirrorPixelData::GetSLTPageData() - "
            << "bin index " << bin << " out of range !" << endl;
    return NULL;
  }
  
  return GetSLTPageDataFast( column, bin );
}

// -------------------------------------------------------------------------

unsigned int 
TMirrorPixelData::GetNumberOfTriggeredPixels(unsigned int start_slice,
                                             unsigned int end_slice) const
 {
  TMirrorBitArray ba( this, start_slice, end_slice );
  
  return ba.GetNumberOfBits();
}

// -------------------------------------------------------------------------

bool 
TMirrorPixelData::HasTriggered(unsigned int col, 
                               unsigned int row,
                               unsigned int first_bin, 
			       unsigned int last_bin) const
 {
   // check parameter range ...
  
  if ( col < Fd::kMI_FIRST_COL || col > Fd::kMI_LAST_COL ) return false;
  if ( row < Fd::kMI_FIRST_ROW || row > Fd::kMI_LAST_ROW ) return false;
  if ( first_bin >= Fd::kMI_NSLT_BINS ) return false;
  if ( last_bin > Fd::kMI_NSLT_BINS ) return false;
  
  unsigned int end_bin = last_bin;
  
  if ( end_bin == Fd::kMI_NSLT_BINS ) end_bin--;
  if ( end_bin < first_bin ) {
    FD_CWARN << "TMirrorPixelData::HasTriggered() - "
             << "time bin indices are in wrong order, I swap them !" << endl;
    swap( first_bin, end_bin );
  }
  
  unsigned int p_data;
  
  for ( unsigned int t = first_bin; t <= end_bin; t++ ){
     PixelDataWord pdw = &fPixelData->fPixelData[t][col - FdUtil::Fd::kMI_FIRST_COL];
     p_data = PixelDataGetData(pdw);
     
     if ((p_data >> (row-1)) & 0x0001) return true;
  }
  
  return false;
}

// -------------------------------------------------------------------------

bool 
TMirrorPixelData::HasTriggeredColumn(unsigned int col,
                                     unsigned int first_bin, 
				     unsigned int last_bin) const
 {
  // check parameter range ...
  
  if ( col < Fd::kMI_FIRST_COL || col > Fd::kMI_LAST_COL ) return false;
  if ( first_bin >= Fd::kMI_NSLT_BINS ) return false;
  if ( last_bin > Fd::kMI_NSLT_BINS ) return false;
  
  unsigned int end_bin = last_bin;
  if ( end_bin == Fd::kMI_NSLT_BINS ) end_bin--;

  FD_CERR << "TMirrorPixelData::HasTriggeredColumn() - "
          << "is not yet implemented !" << endl;
  
  return false;
}

// -------------------------------------------------------------------------

void TMirrorPixelData::Init()
 {
  fPixelData->fVersion = Class_Version();
  
  Fd::SetEyeNo( &fPixelData->fMirrorNumber, FdUtil::GetEyeId() );
  Fd::SetMirrorNo( &fPixelData->fMirrorNumber, FdUtil::GetMirrorId() );

//  FD_COUT << "TMirrorPixelData::Init() - fPixelData: "
//          << Fd::kMI_NCOLS * Fd::kMI_NSLT_BINS * sizeof( PixelDataWordRec )
//	  << " bytes" << endl;

  memset( fPixelData->fPixelData, 0x00,
          Fd::kMI_NCOLS * Fd::kMI_NSLT_BINS * sizeof( PixelDataWordRec ) );

#if (FDUTIL_VERSION_CODE >= ModuleVersionCode(4,0,0))
//  FD_COUT << "TMirrorPixelData::Init() - fMultiplicityData: "
//          << Fd::kMI_NSLT_BINS * sizeof( MultiplicityDataWordRec )
//	  << " bytes" << endl;

  memset( fPixelData->fMultiplicityData, 0x00,
          Fd::kMI_NMULT_BINS * sizeof( MultiplicityDataWordRec ) );
#else
  memset( fPixelData->fMultiplicityData, 0x00,
          Fd::kMI_NFADC_BINS * sizeof( MultiplicityDataWordRec ) );
#endif // FDUTIL_VERSION_CODE

#if 0
  memset( fPixelData->fPixelPageData, 0x00,
          Fd::kMI_NCOLS * Fd::kMI_NSLT_BINS * sizeof( PixelDataWordRec ) );
#endif

//  FD_COUT << "TMirrorPixelData::Init() - fSLTPageData: "
//          << Fd::kMI_NCOLS * Fd::kMI_NSLT_BINS * sizeof( SLTPageDataWordRec )
//	  << " bytes" << endl;

  memset( fPixelData->fSLTPageData, 0x00,
          Fd::kMI_NCOLS * Fd::kMI_NSLT_BINS * sizeof( SLTPageDataWordRec ) );

//  FD_COUT << "TMirrorPixelData::Init() - fMultiplicityPageData: "
//          << Fd::kMI_NSLT_BINS * sizeof( MultiplicityPageDataWordRec )
//	  << " bytes" << endl;

  memset( fPixelData->fMultiplicityPageData, 0x00,
          Fd::kMI_NMULT_BINS * sizeof( MultiplicityPageDataWordRec ) );
}

// -------------------------------------------------------------------------

void TMirrorPixelData::SetMirrorNumber()
 {
  if ( FdUtil::IsMirrorPC() ) {
    Fd::SetEyeNo( &fPixelData->fMirrorNumber, FdUtil::GetEyeId() );
    Fd::SetMirrorNo( &fPixelData->fMirrorNumber, FdUtil::GetMirrorId() );
  }
  else
    FD_CWARN << "TMirrorPixelData::SetMirrorNumber() - "
             << "not running on a MirrorPC, defaults are set!" << endl;
}

// -------------------------------------------------------------------------

void TMirrorPixelData::Show(ostream& ostr,unsigned int slice) const
 {
  ostr << "-- PixelData --" << endl;
  
  int mask = 1;

  ostr << "Mirror Id= " << Fd::GetMirrorNo( &fPixelData->fMirrorNumber )
       << " Eye Id= " << Fd::GetEyeNo( &fPixelData->fMirrorNumber ) << endl;

  ostr << "    " << "           1         1         1         2" << endl;
  ostr << "    " << "   1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0" << endl;

  for ( unsigned int row=Fd::kMI_FIRST_ROW; row<=Fd::kMI_LAST_ROW; row ++ ) {
    
    if ( row > 5 && (row % 5) == 0 )
      ostr << "    " << row << " ";
    else
      ostr << "    " << " " << row % 10 << " ";
    
    for ( unsigned int col=Fd::kMI_FIRST_COL; col<=Fd::kMI_LAST_COL; col++ ) {
      int data = 0;
      if ( slice < Fd::kMI_NSLT_BINS ) 
        data = fPixelData->fPixelData[slice][col-1].fBits.fPixel;
      else {
        data = 0;
	for ( int i=0; i<Fd::kMI_NSLT_BINS; i++)
	  data |= fPixelData->fPixelData[i][col-1].fBits.fPixel;
      }
      
      if ( data & mask )
        ostr << "* ";
      else
        ostr << "  ";
    }
    
    ostr << "|" << endl;
    mask <<= 1;
  }

  ostr << "    " << "   ---------------------------------------" << endl;
  
  // --- check if pixel data is there
  
  PixelDataWord pdata = GetPixelData();
  
  int sum = 0;
  int patt_sum = 0;
  int pix_sum = 0;
  int ext_trigger = 0;
  
  for ( unsigned int m=0; m<Fd::kMI_NSLT_BINS; m++ ) {
    for ( unsigned int col=0; col<Fd::kMI_NCOLS; col++ ) {
      sum |= (int)pdata[m*Fd::kMI_NCOLS+col].fWord;
      pix_sum |= pdata[m*Fd::kMI_NCOLS+col].fBits.fPixel;
      patt_sum |= pdata[m*Fd::kMI_NCOLS+col].fBits.fPatternClass;
      ext_trigger |= pdata[m*Fd::kMI_NCOLS+col].fBits.fExtTrigger;
    }
  }
  
  ostr << " PixelData            : is" << ((sum == 0) ? " not" : "") 
       << " set:" << endl;
  ostr << "  - fPixel            :  contains" << ((pix_sum == 0) ? " no" : "") 
       << " data" << endl;
  ostr << "  - fPatternClass     :  contains" << ((patt_sum == 0) ? " no" : "") 
       << " data" << endl;
  ostr << "  - fExtTrigger       :  is" << ((patt_sum == 0) ? " not" : "") 
       << " set" << endl;

  // --- check if multiplicity data is there
  
  MultiplicityDataWord mult = GetMultiplicityData();
    
  sum = 0;
  for ( unsigned int m=0; m<Fd::kMI_NMULT_BINS; m++ )
    sum |= ((int)mult[m]) & 0x1ff;
  
  ostr << " MultiplicityData     : is" << ((sum == 0) ? " not" : "") 
       << " set." << endl;
  
  // --- check if multiplicity page data is there
  
  MultiplicityPageDataWord mp_data = GetMultiplicityPageData();
  
  sum = 0;
  int mult_sum = 0;
  int proj_sum = 0;
  
  for ( unsigned int m=0; m<Fd::kMI_NMULT_BINS; m++ ) {
    sum |= (int)mp_data[m].fWord;
    mult_sum |= mp_data[m].fBits.fMultiplicity;
    proj_sum |= mp_data[m].fBits.fProjection;
  }
  
  ostr << " MultiplicityPageData : is" << ((sum == 0) ? " not" : "") 
       << " set:" << endl;
  ostr << "  - fMultiplicity     :  contains" << ((mult_sum == 0) ? " no" : "") 
       << " data" << endl;
  ostr << "  - fProjection       :  contains" << ((proj_sum == 0) ? " no" : "") 
       << " data" << endl;

  // --- check if SLT page data is there
  
  SLTPageDataWord slt = GetSLTPageData();
  
  sum = 0;
  patt_sum = 0;
  pix_sum = 0;
  int glob_trigger = 0;
  
  for ( unsigned int m=0; m<Fd::kMI_NSLT_BINS; m++ ) {
    for ( unsigned int col=0; col<Fd::kMI_NCOLS; col++ ) {
      sum |= (int)slt[m*Fd::kMI_NCOLS+col].fWord;
      pix_sum |= slt[m*Fd::kMI_NCOLS+col].fBits.fExtPixelTrigger;
      patt_sum |= slt[m*Fd::kMI_NCOLS+col].fBits.fPattern;
      glob_trigger |= slt[m*Fd::kMI_NCOLS+col].fBits.fGlobTrigger;
    }
  }
  
  ostr << " SLTPageData          : is" << ((sum == 0) ? " not" : "") 
       << " set:" << endl;
  ostr << "  - fExtPixelTrigger  :  contains" << ((pix_sum == 0) ? " no" : "") 
       << " data" << endl;
  ostr << "  - fPattern          :  contains" << ((patt_sum == 0) ? " no" : "") 
       << " data" << endl;
  ostr << "  - fGlobTrigger      :  is" << ((patt_sum == 0) ? " not" : "") 
       << " set" << endl;

  ostr << "----------------------------------------------" << endl;
}

// -------------------------------------------------------------------------

/** Stream a TMirrorPixelData object (SA). */
void TMirrorPixelData::Streamer(TBuffer &R__b)
 {
  if ( R__b.IsReading() ) {
  
    Version_t R__v = R__b.ReadVersion();
  
    TObject::Streamer( R__b );
  
    if ( R__v > MiEVENTVERSIONv4 ) {
      FD_CSEV << "TMirrorPixelData::Streamer() - version mismatch: " << endl;
      FD_CSEV << "found version: " << R__v << " expected <= " 
              << MiEVENTVERSIONv4 << endl;
      return;
    }
    
    R__b >> fPixelData->fVersion;
    
    // check for the data size (32 Bits)
    
    if ( sizeof(PixelDataWordRec) != 4 ) {
      FD_CSEV << "Warning: The data type TMirrorPixelData::PixelDataWordRec"
              << " has wrong size (%d)!"
              << sizeof(PixelDataWordRec) << endl;
      FD_CSEV << "Continue at your own risk !" << endl;
    }
  
    // --- read 2D array of PixelDataWordRec

    R__b.ReadFastArray( &fPixelData->fPixelData[0][0].fWord, 
                        Fd::kMI_NSLT_BINS*Fd::kMI_NCOLS );

#ifdef __ppc__
    for ( int bin=0; bin<Fd::kMI_NSLT_BINS; bin++ ) {
      for ( int col=0; col<Fd::kMI_NCOLS; col++ ) {
        File2PPCPixelDataWordRec( fPixelData->fPixelData[bin][col] );
      }
    }
#endif // __ppc__
    
    // --- read 1D array of MultiplicityDataWordRec

#if 0
    // MiEVENTVERSIONv2: data needs 4 times too much space
    if ( R__v == MiEVENTVERSIONv2 ) {
    
      unsigned int data[Fd::kMI_NMULT_BINS];
      
      R__b.ReadFastArray( data, Fd::kMI_NMULT_BINS );
      
      unsigned char * data_ptr = (unsigned char *)data;
      
      for ( unsigned int i=0; i<Fd::kMI_NMULT_BINS; i++ )
        fPixelData->fMultiplicityData[i] = data_ptr[i];
    }
    // MiEVENTVERSIONv3: storage more compact
    else if ( R__v >= MiEVENTVERSIONv3 ) {
      R__b.ReadFastArray( fPixelData->fMultiplicityData, Fd::kMI_NMULT_BINS );
    }
    // MiEVENTVERSIONv4: not implemented
#else
    // patch code proposed by Antoine, v2 saved also 1000 bytes!
    if ( R__v >= MiEVENTVERSIONv2 && R__v <= MiEVENTVERSIONv3 ) {
      MultiplicityDataWordV3Rec mult[Fd::kMI_NMULT_BINS];
      R__b.ReadFastArray( mult, Fd::kMI_NMULT_BINS );
      
#if 1
      // convert multiplicity data into V4 format (trivial)
      ConvertToV4( mult, fPixelData->fMultiplicityData );
#else
      for ( unsigned int i=0; i<Fd::kMI_NMULT_BINS; i++ )
        fPixelData->fMultiplicityData[i] = mult[i];
#endif
    }
    else if ( R__v == MiEVENTVERSIONv4 ) {
      R__b.ReadFastArray( fPixelData->fMultiplicityData, Fd::kMI_NMULT_BINS );
    }
#endif

#ifdef __ppc__
      // no code for Darwin (PPC)
#endif // __ppc__
    else {  // clear all values
      memset( fPixelData->fMultiplicityData, 0, Fd::kMI_NMULT_BINS );
    }
    
    // --- read 2D array of SLTPageDataRec
    
    if ( R__v == MiEVENTVERSIONv4 ) {
      R__b.ReadFastArray( &fPixelData->fSLTPageData[0][0].fWord, 
                          Fd::kMI_NSLT_BINS*Fd::kMI_NCOLS );
    }
    else {  // clear all values
      memset( fPixelData->fSLTPageData, 0, Fd::kMI_NSLT_BINS*Fd::kMI_NCOLS );
    }
    
    // --- read 1D array of MultiplicityPageDataWordRec
    
    if ( R__v == MiEVENTVERSIONv4 ) {
      R__b.ReadFastArray( &fPixelData->fMultiplicityPageData[0].fWord, 
                          Fd::kMI_NMULT_BINS );
    }
    else {  // clear all values
      memset( fPixelData->fMultiplicityPageData, 0, Fd::kMI_NMULT_BINS );
    }
    
  }
  else {
    R__b.WriteVersion( TMirrorPixelData::IsA() );
    
    TObject::Streamer( R__b );

    R__b << fPixelData->fVersion;
    
#ifdef __ppc__
#waring This will spoil the data in memory !
    for ( int bin=0; bin<Fd::kMI_NSLT_BINS; bin++ ) {
      for ( int col=0; col<Fd::kMI_NCOLS; col++ ) {
        fPixelData->fPixelData[bin][col] = 
	            PPC2FilePixelDataWordRec( fPixelData->fPixelData[bin][col] );
      }
    }
#endif // __ppc__
    
    R__b.WriteFastArray( &fPixelData->fPixelData[0][0].fWord, 
                         Fd::kMI_NSLT_BINS*Fd::kMI_NCOLS );
  
    // new for MiEVENTVERSIONv2
    // modified for MiEVENTVERSIONv3
    // identical for MiEVENTVERSIONv4
    R__b.WriteFastArray( fPixelData->fMultiplicityData, Fd::kMI_NMULT_BINS );

    // new for MiEVENTVERSIONv4
    R__b.WriteFastArray( &fPixelData->fSLTPageData[0][0].fWord, 
                         Fd::kMI_NSLT_BINS*Fd::kMI_NCOLS );

    // new for MiEVENTVERSIONv4
    R__b.WriteFastArray( &fPixelData->fMultiplicityPageData[0].fWord, 
                         Fd::kMI_NMULT_BINS );
#ifdef __ppc__
    // no special code for Darwin (PPC)
#endif // __ppc__
  }   
}

ClassImp(TMirrorPixelData)

// -------------------------------------------------------------------------
// -------------------------------------------------------------------------
