
//
// File   : MiPixelList.cc
//
// Purpose: Implementation of Mirror pixel list class TMirrorPixelList
//
// $Id: MiPixelList.cc 8759 2010-09-27 12:16:36Z mathes $
//


#include <cstdio>
#include <cstdlib>
#include <cstring>

#include <iostream>
#include <iomanip>


/** @file MiPixelList.cc
  * Implementation of class TMirrorPixelList
  * @author H.-J. Mathes, FzK
  */

#include <TClass.h>
#include <TBuffer.h>

#include <ErrorLogging.hh>
#include <FdNumbering.hh>
#include <FdConfig.hh>

#include "CheckVersion.hh"
#include "MiBitArray.hh"

#include "MiEvent.hh"

using namespace std;
using FdUtil::Fd;
using namespace MiEvent;

// -------------------------------------------------------------------------
// -------------------------------------------------------------------------
//
// class TMirrorPixelList - implementation
//

TMirrorPixelList::TMirrorPixelList()
 : fConstructorType(kCREATE), fIsSorted(false)
 {
#ifdef TRACE_CONSTRUCTOR
  FD_CINFO << "TMirrorPixelList(): " << this << endl;
#endif
  Construct();
  Init();
}
 
// -------------------------------------------------------------------------

TMirrorPixelList::TMirrorPixelList(void *ptr)
 : fConstructorType(kMIPIXELLIST), fIsSorted(false)
 {
#ifdef TRACE_CONSTRUCTOR
  FD_CINFO << "TMirrorPixelList(void*): " << this << endl;
#endif
  fPixelList = (MiPixelList)ptr;
}

// -------------------------------------------------------------------------

TMirrorPixelList::TMirrorPixelList(const TMirrorPixelList& rhs)
 : TObject(rhs), fConstructorType(kCREATE), fIsSorted(false)
 {
#ifdef TRACE_CONSTRUCTOR
  FD_CINFO << "TMirrorPixelList(TMirrorPixelList&): " << this << endl;
#endif

  Construct();
  *this = rhs;
}

// -------------------------------------------------------------------------

TMirrorPixelList::TMirrorPixelList(const TMirrorBitArray *bit_arr)
 : fConstructorType(kCREATE), fIsSorted(false)
 {
#ifdef TRACE_CONSTRUCTOR
  FD_CINFO << "TMirrorPixelList(TMirrorBitArray*): " << this << endl;
#endif
  Construct();
  Init();

  Fd::PixelNumberRec pixel;
	
  Fd::SetEyeNo( &pixel, FdUtil::GetEyeId() );
  Fd::SetMirrorNo( &pixel, FdUtil::GetMirrorId() );

  for ( unsigned int col=Fd::kMI_FIRST_COL; col<=Fd::kMI_LAST_COL; col++ ) {

    for ( unsigned int row=Fd::kMI_FIRST_ROW; row<=Fd::kMI_LAST_ROW; row++ ) {

      if ( bit_arr->Test( col, row ) ) {
        Fd::SetPixelNoFromColumnRow( &pixel, col, row );
	AddPixel( &pixel );
      }
    }

    if ( bit_arr->TestColumn( col ) ) {
      Fd::SetPixelNoFromColumnRow( &pixel, col, Fd::kMI_VC1_ROW );
      AddPixel( &pixel );
      Fd::SetPixelNoFromColumnRow( &pixel, col, Fd::kMI_VC2_ROW );
      AddPixel( &pixel );
    }
  }
}

// -------------------------------------------------------------------------

TMirrorPixelList::TMirrorPixelList(const TMirrorPixelData* pixel_data,
                                   unsigned int start_slice, 
				   unsigned int end_slice)
 : fConstructorType(kCREATE), fIsSorted(false)
 {
#ifdef TRACE_CONSTRUCTOR
  FD_CINFO << "TMirrorPixelList(TMirrorPixelData*): " << this << endl;
#endif
  Construct();
  Init();
  
  TMirrorBitArray ba;
  
  Fd::PixelNumberRec pixel;
	
  Fd::SetEyeNo( &pixel, FdUtil::GetEyeId() );
  Fd::SetMirrorNo( &pixel, FdUtil::GetMirrorId() );

  ba.Set( pixel_data, start_slice, end_slice );

  for ( unsigned int col=Fd::kMI_FIRST_COL; col<=Fd::kMI_LAST_COL; col++ ) {

    for ( unsigned int row=Fd::kMI_FIRST_ROW; row<=Fd::kMI_LAST_ROW; row++ ) {
      if ( ba.Test( col, row ) ) {
        Fd::SetPixelNoFromColumnRow( &pixel, col, row );
	AddPixel( &pixel );
      }
    }

    if ( ba.TestColumn( col ) ) {
      Fd::SetPixelNoFromColumnRow( &pixel, col, Fd::kMI_VC1_ROW );
      AddPixel( &pixel );
      Fd::SetPixelNoFromColumnRow( &pixel, col, Fd::kMI_VC2_ROW );
      AddPixel( &pixel );
    }
  }
}

// -------------------------------------------------------------------------

TMirrorPixelList::~TMirrorPixelList()
 {
#ifdef TRACE_CONSTRUCTOR
  FD_CINFO << "~TMirrorPixelList(): " << this << endl;
#endif

  switch ( fConstructorType ) {
  
    case kCREATE:
         Destruct();

    case kSHMEVENT:
    case kMIPIXELLIST:
         fPixelList = NULL;
	 break;
	 
    default:
         FD_CERR << "TMirrorPixelList::~TMirrorPixelList() -"
	         << " unknown fConstructorType= " 
	         << (int)fConstructorType << endl;
  }
}

// -------------------------------------------------------------------------

TMirrorPixelList& TMirrorPixelList::operator=(const TMirrorPixelList& rhs )
 {
  // check for self-assignment
  if ( this == &rhs ) return *this;
  
  fPixelList->fVersion = rhs.fPixelList->fVersion;
  fPixelList->fNPixels = rhs.fPixelList->fNPixels;

  for ( UInt_t i=0; i<fPixelList->fNPixels; i++ )
    fPixelList->fPixelNumber[i] = rhs.fPixelList->fPixelNumber[i];
  
  fIsSorted = rhs.fIsSorted;
  
  return *this;
}

// -------------------------------------------------------------------------

bool TMirrorPixelList::operator==(const TMirrorPixelList& rhs) const
 {
  if ( this == &rhs ) return true;
  
  // --- initzialize empty pixel vectors
  
  unsigned int pixel_vect[kPixelVectorSize];
  unsigned int rhs_pixel_vect[kPixelVectorSize];
  
  memset( pixel_vect, 0, kPixelVectorSize * sizeof( unsigned int ) );
  memset( rhs_pixel_vect, 0, kPixelVectorSize * sizeof( unsigned int ) );
  
  // --- populate with pixels ...
  
  Fd::PixelNumber pixel;
  
  for ( unsigned int i=0; i<GetNumPixels(); i++ ) {
  
    pixel = GetPixel( i );
    
    unsigned int mi_pixel_no = Fd::GetPixelNo( pixel ) % Fd::kMI_NEXTPIXELS;
    
    pixel_vect[((mi_pixel_no & 0xfffe0) >> 5)] |= (1 << (mi_pixel_no & 0x1f));
  }
  
  for ( unsigned int i=0; i<rhs.GetNumPixels(); i++ ) {
  
    pixel = rhs.GetPixel( i );
    
    unsigned int mi_pixel_no = Fd::GetPixelNo( pixel ) % Fd::kMI_NEXTPIXELS;
    
    rhs_pixel_vect[((mi_pixel_no & 0xfffe0) >> 5)] |= (1 << (mi_pixel_no & 0x1f));
  }
  
  // --- finally compare them
  
  for ( int i=0; i<kPixelVectorSize; i++ )
    if ( pixel_vect[i] != rhs_pixel_vect[i] ) return false; // not equal

  return true;
}

// -------------------------------------------------------------------------

int TMirrorPixelList::AddPixel(Fd::PixelNumber pixel)
 { 
  if ( Fd::GetPixelNo( pixel ) < Fd::kMI_FIRST_PIXEL ) return 1;
  if ( Fd::GetPixelNo( pixel ) > Fd::kMI_NEXTPIXELS ) return 1;
  if ( fPixelList->fNPixels > Fd::kMI_NEXTPIXELS ) return 1;
  
  // now everything is OK -> add pixel
  
  fPixelList->fPixelNumber[fPixelList->fNPixels] = *pixel;
  fPixelList->fNPixels++;

  fIsSorted = false;
  
  return 0;
}

// -------------------------------------------------------------------------

int TMirrorPixelList::AddPixel(UInt_t col, UInt_t row)
 {
  if ( col < Fd::kMI_FIRST_COL || col > Fd::kMI_LAST_COL ) return 1;
  if ( row > Fd::kMI_LAST_ROW ) 
    return AddVirtualPixel( col, row - Fd::kMI_VC1_ROW + Fd::kMI_FIRST_VCHANNEL );
  if ( row < Fd::kMI_FIRST_ROW || row > Fd::kMI_LAST_ROW ) return 1;
  
  Fd::PixelNumberRec pixel;
  
  // assuming the defaults for eye and mirror number ...
  
  Fd::SetEyeNo( &pixel, FdUtil::GetEyeId() );
  Fd::SetMirrorNo( &pixel, FdUtil::GetMirrorId() );
  Fd::SetPixelNoFromColumnRow( &pixel, col, row );
  
  return AddPixel( &pixel );
}

// -------------------------------------------------------------------------

int TMirrorPixelList::AddVirtualPixel(UInt_t col, UInt_t virt)
 {
  if ( virt < Fd::kMI_FIRST_VCHANNEL || virt > Fd::kMI_LAST_VCHANNEL ) return 1;
  if ( col < Fd::kMI_FIRST_COL || col > Fd::kMI_LAST_COL ) return 1;

  Fd::PixelNumberRec pixel;

  Fd::SetEyeNo( &pixel, FdUtil::GetEyeId() );
  Fd::SetMirrorNo( &pixel, FdUtil::GetMirrorId() );
  Fd::SetPixelNoFromColumnRow( &pixel, col, Fd::kMI_NROWS + virt );
  
  return AddPixel( &pixel );
}

// -------------------------------------------------------------------------

void TMirrorPixelList::CopyFrom(TMirrorPixelList* pixel_list)
 {
  if ( !pixel_list && !pixel_list->fPixelList ) {
    FD_CERR << "TMirrorPixelList::CopyFrom() - cannot copy from NULL pointer !" 
            << endl;
    return;
  }

  fPixelList->fVersion = pixel_list->fPixelList->fVersion;
  fPixelList->fNPixels = pixel_list->fPixelList->fNPixels;
  
  memcpy ( &fPixelList->fPixelNumber[0], 
           &pixel_list->fPixelList->fPixelNumber[0], 
           pixel_list->fPixelList->fNPixels * sizeof( Fd::PixelNumberRec ) );
}

// -------------------------------------------------------------------------

void TMirrorPixelList::Construct()
 {
  fPixelList = (MiPixelList)malloc( GetMaxSize() );
  
  memset( fPixelList, 0, GetMaxSize() );
}

// -------------------------------------------------------------------------

Fd::PixelNumber TMirrorPixelList::GetPixel(UInt_t seq) const
 {
  if ( seq >= fPixelList->fNPixels ) return NULL;
  
  // Note: If the pixel list is sorted during streaming, then the
  //       pixel list is no longer in sync with the TFADCData objects.
  
  // in this case the above 'const' is incorrect ...
  // if ( !fIsSorted ) Sort();
  
  return &fPixelList->fPixelNumber[seq];
}

// -------------------------------------------------------------------------

void TMirrorPixelList::Init()
 {
  fPixelList->fVersion = Class_Version();
  fPixelList->fNPixels = 0;
}

// -------------------------------------------------------------------------

void TMirrorPixelList::Merge(TMirrorPixelList* pixel_list)
 {
  TMirrorBitArray* arr1 = new TMirrorBitArray( this );
  TMirrorBitArray* arr2 = new TMirrorBitArray( pixel_list );
  
  *arr2 |= *arr1;  // merge both bit arrays
  
  TMirrorPixelList* temp = new TMirrorPixelList( arr2 );
  *this = *temp;
  delete temp;
  
  delete arr1;
  delete arr2;
}

// -------------------------------------------------------------------------

void TMirrorPixelList::Merge(TMirrorBitArray* bit_array)
 {
  // create a TMirrorBitArray from the list and merge both bit arrays
  
  TMirrorBitArray* temp_array = new TMirrorBitArray( this );
  *temp_array |= *bit_array;
  
  // create a temporary TMirrorPixelList from the merged bit array
  
  TMirrorPixelList *temp_list = new TMirrorPixelList( temp_array );
  *this = *temp_list;
  delete temp_list;
  
  delete temp_array;
}

// -------------------------------------------------------------------------

void TMirrorPixelList::Set(TMirrorBitArray* bit_array)
 {
  // create a temporary TMirrorPixelList from the bit array
  
  TMirrorPixelList *temp_list = new TMirrorPixelList( bit_array );

  *this = *temp_list;

  delete temp_list;
}

// -------------------------------------------------------------------------

void TMirrorPixelList::Sort()
 {
  if ( fPixelList->fNPixels == 0 ) {
    fIsSorted = true;
    return;
  }
  
  size_t vec_size = Fd::kMI_NEXTROWS*Fd::kMI_NCOLS/(sizeof(int) * 8);
  
  int *vec = new int[vec_size];
  
  memset( vec, 0x00, vec_size * sizeof(int)/sizeof(char) );
//  for ( int i=0; i<vec_size; i++ ) vec[i] = 0;
  
  // --- unfold list into bit vector
  
  register UInt_t pixel_no;
  register UInt_t wordpos, bitpos, mask;
  
  for ( UInt_t seq=0; seq<fPixelList->fNPixels; seq++ ) {
    
    pixel_no =   Fd::GetPixelNo( &fPixelList->fPixelNumber[seq] )
               - Fd::kMI_FIRST_PIXEL;
    
    wordpos = pixel_no / (sizeof( UInt_t ) * 8);
    bitpos = pixel_no % (sizeof( UInt_t ) * 8);
  
    mask = 1 << bitpos;
    
    vec[wordpos] |= mask;
  }
  
  // --- fill bit vector back into list

  Fd::PixelNumberRec pixel = fPixelList->fPixelNumber[0];  // Eye & Mirror Id
  
  fPixelList->fNPixels = 0;
  
  for ( pixel_no=0; pixel_no<Fd::kMI_NEXTROWS*Fd::kMI_NCOLS; pixel_no++ ) {
  
    wordpos = pixel_no / (sizeof( UInt_t ) * 8);
    bitpos = pixel_no % (sizeof( UInt_t ) * 8);
  
    mask = 1 << bitpos;
    
    if ( (vec[wordpos] & mask) ) {
      Fd::SetPixelNo( &pixel, pixel_no + Fd::kMI_FIRST_PIXEL );
      fPixelList->fPixelNumber[fPixelList->fNPixels] = pixel;
      fPixelList->fNPixels++;
    }
  }

  delete [] vec;
 
  fIsSorted = true;
}

// -------------------------------------------------------------------------

void TMirrorPixelList::Show(ostream& ostr) const
 {
  Fd::PixelNumber pix;
  
  if ( fPixelList != NULL ) {
    ostr <<  "-- PixelList --" << endl;
    ostr <<  "Version:      " << GetVersion() << endl;
    ostr <<  "ActualSize:   " << GetSize() << endl;
    ostr <<  "MaxSize:      " << GetMaxSize() << endl;
    ostr <<  "NPixels:      " << GetNumPixels() << endl;
    
    for ( unsigned int i=0; i<GetNumPixels(); i++ ) {
      pix = GetPixel( i );
      ostr << " - Pixel[" << setw(3) << i << "] = " 
           << Fd::GetPixelNo( pix ) << endl;
    }
  }
}

// -------------------------------------------------------------------------

/** Stream a TMirrorPixelList object (SA).
  */
void TMirrorPixelList::Streamer(TBuffer &b)
 {
  if ( b.IsReading() ) {

    Version_t vers = b.ReadVersion();

    TObject::Streamer( b );

    if ( vers != MiEVENTVERSIONv1 ) {
      FD_CERR << "Version mismatch in TMirrorPixelList::Streamer()!" << endl;
      FD_CERR << "Found version: " << vers << " expected == " 
              << MiEVENTVERSIONv1 << endl;
      return;
    }

    b >> fPixelList->fVersion;
    b >> fPixelList->fNPixels;

    for ( UInt_t i=0; i<fPixelList->fNPixels; i++ )
      b >> fPixelList->fPixelNumber[i];
  }
  else {
    b.WriteVersion( TMirrorPixelList::IsA() );

    TObject::Streamer( b );

    b << fPixelList->fVersion;
    b << fPixelList->fNPixels;
    
    for ( UInt_t i=0; i<fPixelList->fNPixels; i++ )
      b << fPixelList->fPixelNumber[i];
  }
}

ClassImp(TMirrorPixelList)

// -------------------------------------------------------------------------
// -------------------------------------------------------------------------
