
//
// File   : MiRunHeader.cc
//
// Purpose: Implementation of Mirror run header class.
//
// $Id: MiRunHeader.cc 8759 2010-09-27 12:16:36Z mathes $
//


#include <ctime>
#include <stdexcept>
#include <cassert>

#include <TObject.h>
#include <TBuffer.h>

#include <iostream>


/** @file MiRunHeader.cc
  * Implementation of the class TMirrorRunHeader.
  * @author H.-J. Mathes, FzK
  */

#include <ErrorLogging.hh>

#include "CheckVersion.hh"

#include "MiRunHeader.hh"

using namespace std;

// -------------------------------------------------------------------------
// -------------------------------------------------------------------------
//
// class TMirrorRunHeader - implementation
//

TMirrorRunHeader::TMirrorRunHeader()
 : fConstructorType(kCREATE), fRunHeader(new MiRunHeaderRec)
 {
#ifdef TRACE_CONSTRUCTOR
  FD_CINFO << "TMirrorRunHeader(): " << this << endl;
#endif
  Init();
}

// -------------------------------------------------------------------------

TMirrorRunHeader::TMirrorRunHeader(MiRunHeader header)
 : fConstructorType(kRUNHEADER), fRunHeader(header)
 { 
#ifdef TRACE_CONSTRUCTOR
  FD_CINFO << "TMirrorRunHeader(MiRunHeader): " << this << endl;
#endif
  if ( fRunHeader == NULL ) 
    FD_CERR << "TMirrorRunHeader::TMirrorRunHeader(MiRunHeader) - "
            << "bad fRunHeader !" << endl;
}

// -------------------------------------------------------------------------

TMirrorRunHeader::TMirrorRunHeader(const TMirrorRunHeader& rhs)
 : TObject(rhs),
   fConstructorType(kCREATE), fRunHeader(new MiRunHeaderRec)
 {
  *this = rhs;
}

// -------------------------------------------------------------------------

TMirrorRunHeader::~TMirrorRunHeader()
 {
#ifdef TRACE_CONSTRUCTOR
  FD_CINFO << "~TMirrorRunHeader(): " << this << endl;
#endif
  switch ( fConstructorType ) {
  
    case kCREATE:
         delete fRunHeader;
         fRunHeader = 0;
	 break;

    case kRUNHEADER:
         break;

    case kSHM_RUNHEADER:
         fRunHeader = 0;
         break;

    default:
         FD_CERR << "TMirrorRunHeader::~TMirrorRunHeader() - "
	         << "illegal fConstructorType= " << fConstructorType << endl;
  }
}

// -------------------------------------------------------------------------

TMirrorRunHeader& TMirrorRunHeader::operator=(const TMirrorRunHeader& rhs)
 {
#ifdef TRACE_CONSTRUCTOR
  FD_CINFO << "TMirrorRunHeader(TMirrorRunHeader&): " << this << endl;
#endif
  if ( this->fRunHeader == rhs.fRunHeader ) return *this;
  // don't copy fConstructorType but only the rest
  memcpy( this->fRunHeader, rhs.fRunHeader, sizeof(*rhs.fRunHeader) );

  return *this;
}

// -------------------------------------------------------------------------

void TMirrorRunHeader::GetOption(unsigned int which,int& option) const
 {
  if ( which < GetMaxNumOptions() )
    option = fRunHeader->fOption[which].fIntOption;
  else {
    FD_CWARN << "TMirrorRunHeader::GetOption() - "
             << "bad option " << option << " passed!" << endl;
    option = 0;
    // throw BadHeaderOption();
  }
}

// -------------------------------------------------------------------------

void TMirrorRunHeader::GetOption(unsigned int which,unsigned int& option) const
 {
  if ( which < (unsigned int)GetMaxNumOptions() )
    option = fRunHeader->fOption[which].fUIntOption;
  else {
    FD_CWARN << "TMirrorRunHeader::GetOption() - "
             << "bad option " << option << " passed!" << endl;
    option = 0;
    // throw BadHeaderOption();
  }
}

// -------------------------------------------------------------------------

void TMirrorRunHeader::GetOption(unsigned int which,float& option) const
 {
  if ( which < GetMaxNumOptions() )
    option = fRunHeader->fOption[which].fFloatOption;
  else {
    FD_CWARN << "TMirrorRunHeader::GetOption() - "
             << "bad option " << option << " passed!" << endl;
    option = 0.;
    // throw BadHeaderOption();
  }
}

// -------------------------------------------------------------------------

void TMirrorRunHeader::Init()
 {
  if ( !fRunHeader ) {
    FD_CSEV << "TMirrorRunHeader::Init(): no structure defined !" << endl;
    throw logic_error("TMirrorRunHeader::Init(): no structure defined !");
  }

  fRunHeader->fRunNumber = 0;
  fRunHeader->fStartTime = 0;
  
  strcpy( fRunHeader->fComment, "" );

  fRunHeader->fNOptions  = 0;
}

// -------------------------------------------------------------------------

void TMirrorRunHeader::Show(ostream& ostr) const
 { 
  time_t start_time =
     FdUtil::TimeConvert::GetUnixTimeFromGPSTime( GetStartTime() );

  ostr << "-- RunHeader for run: " << GetRunNo() << endl;
  ostr << " started at (UTC): " << asctime( gmtime ( &start_time ) );
  ostr << "          GPS sec: " << GetStartTime() << endl;
  ostr << " comment:          " << GetComment() << endl;
}

// -------------------------------------------------------------------------

void TMirrorRunHeader::SetOption(unsigned int which, int option)
 {
  if ( which < (UInt_t)GetMaxNumOptions() )
    fRunHeader->fOption[which].fIntOption = option;
  else {
    FD_CWARN << "TMirrorRunHeader::SetOption() - "
             << "bad option " << option << " passed!" << endl;
    // throw BadHeaderOption();
  }
}

// -------------------------------------------------------------------------

void TMirrorRunHeader::SetOption(UInt_t which, UInt_t option)
 {
  if ( which < (UInt_t)GetMaxNumOptions() )
    fRunHeader->fOption[which].fUIntOption = option;
  else {
    FD_CWARN << "TMirrorRunHeader::SetOption() - "
             << "bad option " << option << " passed!" << endl;
    // throw BadHeaderOption();
  }
}

// -------------------------------------------------------------------------

void TMirrorRunHeader::SetOption(UInt_t which, float option)
 {
  if ( which < (UInt_t)GetMaxNumOptions() )
    fRunHeader->fOption[which].fFloatOption = option;
  else {
    FD_CWARN << "TMirrorRunHeader::SetOption() - "
             << "bad option " << option << " passed!" << endl;
    // throw BadHeaderOption();
  }
}

// -------------------------------------------------------------------------

void TMirrorRunHeader::Streamer(TBuffer &b)
 {
  if ( b.IsReading() ) {

    Version_t v = b.ReadVersion();

    if ( v > MiEVENTVERSIONv2) {
      FD_CCRIT << "TMirrorRunHeader::Streamer() - "
               << "version mismatch" << endl;
    }
    
    b >> fRunHeader->fRunNumber;
    b >> fRunHeader->fStartTime;
    b >> fRunHeader->fNOptions;

    for ( unsigned int i=0; i<(unsigned int)GetMaxNumOptions(); i++ )
      b >> fRunHeader->fOption[i].fIntOption;

    b >> fRunHeader->fComment;
    
    if ( v < MiEVENTVERSIONv2 ) {
      SetOption( RUNH_OPT_FLT_VERSION, 0 );
      SetOption( RUNH_OPT_SLT_VERSION, 0 );
    }
  }
  else {
    b.WriteVersion(TMirrorRunHeader::IsA());

    b << fRunHeader->fRunNumber;
    b << fRunHeader->fStartTime;
    b << fRunHeader->fNOptions;

    for ( unsigned int i=0; i<(unsigned int)GetMaxNumOptions(); i++ )
      b << fRunHeader->fOption[i].fIntOption;

    b << fRunHeader->fComment;
  }
}

ClassImp(TMirrorRunHeader)

// -------------------------------------------------------------------------
// -------------------------------------------------------------------------
