
//
// File   : MiRunTrailer.cc
//
// Purpose: Implementation of Mirror run trailer class.
//
// $Id: MiRunTrailer.cc 8759 2010-09-27 12:16:36Z mathes $
//


#include <ctime>
#include <iostream>
#include <stdexcept>

#include <TObject.h>
#include <TBuffer.h>


/** @file MiRunTrailer.cc
  * Implementation of the class TMirrorRunTrailer.
  * @author H.-J. Mathes, FzK
  */

#include <TimeConvert.hh>
#include <ErrorLogging.hh>

#include "CheckVersion.hh"

#include "MiRunTrailer.hh"

using namespace std;

// -------------------------------------------------------------------------

// these operators are still missing in ROOT (before 3.05/00)

#if (ROOT_VERSION_CODE < ROOT_VERSION(3,5,5))
extern void operator<<(TBuffer&,unsigned long long);
extern void operator<<(TBuffer&,long long);
extern void operator>>(TBuffer&,unsigned long long);
extern void operator>>(TBuffer&,long long);
#endif // ROOT_VERSION_CODE

// -------------------------------------------------------------------------
// -------------------------------------------------------------------------
//
// class TMirrorRunTrailer - implementation
//

TMirrorRunTrailer::TMirrorRunTrailer()
 : fConstructorType(kCREATE), fRunTrailer(new MiRunTrailerRec),
   fStreamerVersion(0)
 {
#ifdef TRACE_CONSTRUCTOR
  FD_CINFO << "TMirrorRunTrailer(): " << this << endl;
#endif // TRACE_CONSTRUCTOR
  Init();
}

// -------------------------------------------------------------------------

TMirrorRunTrailer::TMirrorRunTrailer(MiRunTrailer trailer)
 : fConstructorType(kRUNTRAILER), fRunTrailer(trailer),
   fStreamerVersion(0)
 { 
#ifdef TRACE_CONSTRUCTOR
  FD_CINFO << "TMirrorRunTrailer(MiRunTrailer) - " << this << endl;
#endif // TRACE_CONSTRUCTOR
  if ( fRunTrailer == NULL )
    FD_CERR << "TMirrorRunTrailer::TMirrorRunTrailer(MiRunTrailer) - "
            << "bad fRunTrailer !" << endl;
}

// -------------------------------------------------------------------------

TMirrorRunTrailer::TMirrorRunTrailer(const TMirrorRunTrailer& rhs)
 : TObject(rhs),
   fConstructorType(kCREATE), fRunTrailer(new MiRunTrailerRec),
   fStreamerVersion(0)
 {
  *this = rhs;
}

// -------------------------------------------------------------------------

TMirrorRunTrailer::~TMirrorRunTrailer()
 {
#ifdef TRACE_CONSTRUCTOR
  FD_CINFO << "~TMirrorRunTrailer(): " << this << endl;
#endif // TRACE_CONSTRUCTOR

  switch ( fConstructorType ) {
  
    case kCREATE:
         delete fRunTrailer;
         fRunTrailer = 0;
	 break;

    case kRUNTRAILER:
         break;

    case kSHM_RUNTRAILER:
         fRunTrailer = 0;
         break;

    default:
         FD_CERR << "TMirrorRunTrailer::~TMirrorRunTrailer() - "
	         << "illegal fConstructorType= " << fConstructorType 
		 << endl;
  }
}

// -------------------------------------------------------------------------

TMirrorRunTrailer& TMirrorRunTrailer::operator=(const TMirrorRunTrailer& rhs)
 {
#ifdef TRACE_CONSTRUCTOR
  FD_CINFO << "TMirrorRunTrailer(TMirrorRunTrailer&): " << this << endl;
#endif // TRACE_CONSTRUCTOR

  if ( this->fRunTrailer == rhs.fRunTrailer ) return *this;
  
  // don't copy fConstructorType but only the rest
  memcpy( this->fRunTrailer, rhs.fRunTrailer, sizeof(*rhs.fRunTrailer) );

  fStreamerVersion = rhs.fStreamerVersion;
  
  return *this;
}

// -------------------------------------------------------------------------

void TMirrorRunTrailer::Init()
 {
  if ( !fRunTrailer ) {
    FD_CSEV << "TMirrorRunTrailer::Init() - no structure defined !" << endl;
    throw logic_error("TMirrorRunTrailer::Init() - no structure defined !");
  }

  fRunTrailer->fStartTime         = FdUtil::TimeConvert::GetGPSTime();
  fRunTrailer->fStopTime          = FdUtil::TimeConvert::GetGPSTime();
  fRunTrailer->fTotalTime         = 0;
  
  fRunTrailer->fNEvents           = 0;
  fRunTrailer->fNTriggers         = 0;
  
  fRunTrailer->fDeadTimeCounter   = 0;
  fRunTrailer->fRunTimeCounter    = 0;
  fRunTrailer->fVetoTimeCounter   = 0;
  
  fRunTrailer->fDaqDeadTime       = 0;
  fRunTrailer->fNSkippedTriggers  = 0;
}

// -------------------------------------------------------------------------

void TMirrorRunTrailer::SetStopTime()
 {
  SetStopTime( FdUtil::TimeConvert::GetGPSTime() );
}

// -------------------------------------------------------------------------

void TMirrorRunTrailer::Show(ostream& ostr) const
 { 
  time_t start_time, stop_time;
  
  if ( fStreamerVersion == 1 ) {
    start_time = GetStartTime();
    stop_time = GetStopTime();
  }
  else {
    start_time = FdUtil::TimeConvert::GetUnixTimeFromGPSTime( GetStartTime() );
    stop_time = FdUtil::TimeConvert::GetUnixTimeFromGPSTime( GetStopTime() );
  }

  ostr << "-- RunTrailer for run: " << endl;
  ostr << " Streamer version: " << fStreamerVersion << endl;
  ostr << " started at (UTC): " << asctime( gmtime ( &start_time ) );
  ostr << " stopped at (UTC): " << asctime( gmtime ( &stop_time ) );
  ostr << " total time:       " << GetTotalTime() << " [sec]" << endl;
  ostr << " events    :       " << GetNumEvents() << endl;
  ostr << " triggers  :       " << GetNumTriggers() << endl;
  ostr << " AccVETO time:     " << setw(20) 
       << GetVetoTimeCounter() << " [* 100ns]" << endl;
  ostr << " AccDead time:     " << setw(20) 
       << GetDeadTimeCounter() << " [* 100ns]" << endl;
  ostr << " DAQ dead time:    " << GetDaqDeadTime() << " [sec]" << endl;
  ostr << " Triggers skipped: " << GetNumSkippedTriggers() << endl;
}

// -------------------------------------------------------------------------

void TMirrorRunTrailer::Streamer(TBuffer &R__b)
 {
  unsigned int R__s, R__c;
  
  if ( R__b.IsReading() ) {

    Version_t R__v = fStreamerVersion = R__b.ReadVersion(&R__s, &R__c);

    if ( TMirrorEvent::GetPrintLevel() ) {
      FD_CDEBUG(1) << "TMirrorRunTrailer::Streamer() - R__v= " << R__v
                   << " R__s= " << R__s << " R__c= " << R__c << endl;
    }

    if ( R__v > MiEVENTVERSIONv4 ) {
      FD_CCRIT << "Version mismatch in TMirrorRunTrailer::Streamer()!" << endl;
      FD_CCRIT << "Found version: " << R__v << " expected <= " 
               << MiEVENTVERSIONv4 << endl;
      return;
    }
    
    R__b >> fRunTrailer->fStartTime;
    R__b >> fRunTrailer->fStopTime;
    R__b >> fRunTrailer->fTotalTime;
    R__b >> fRunTrailer->fNEvents;
    R__b >> fRunTrailer->fNTriggers;
    
    if ( R__v >= MiEVENTVERSIONv2 ) {
      R__b >> fRunTrailer->fDeadTimeCounter;
      R__b >> fRunTrailer->fVetoTimeCounter;
    }

    if ( R__v >= MiEVENTVERSIONv3 ) {
      R__b >> fRunTrailer->fDaqDeadTime;
      R__b >> fRunTrailer->fNSkippedTriggers;
    }

    if ( R__v >= MiEVENTVERSIONv4 ) {
      R__b >> fRunTrailer->fRunTimeCounter;
    }
  }
  else {
    R__b.WriteVersion(TMirrorRunTrailer::IsA());

    R__b << fRunTrailer->fStartTime;
    R__b << fRunTrailer->fStopTime;
    R__b << fRunTrailer->fTotalTime;
    R__b << fRunTrailer->fNEvents;
    R__b << fRunTrailer->fNTriggers;
    
    // new for version MiEVENTVERSIONv2

    R__b << fRunTrailer->fDeadTimeCounter;
    R__b << fRunTrailer->fVetoTimeCounter;

    // new for version MiEVENTVERSIONv3

    R__b << fRunTrailer->fDaqDeadTime;
    R__b << fRunTrailer->fNSkippedTriggers;

    // new for version MiEVENTVERSIONv4

    R__b << fRunTrailer->fRunTimeCounter;
  }
}

ClassImp(TMirrorRunTrailer)

// -------------------------------------------------------------------------
// -------------------------------------------------------------------------
