
//
// File   : EyeEventClassifier.cc
//
// Purpose: Implementation of the class EyeEventClassifier.
//          It contains the third level trigger algorithm used to 
//          classify the events.
//
// Authors: Pedro Facal, Paolo Privitera
// 
// $Id: EyeEventClassifier.cc 11969 2011-01-27 16:53:59Z mathes $
//


#include <iostream>
#include <iomanip>
#include <vector>
#include <algorithm>

#include <cstdio>
#include <cstdlib>
#include <cmath>
#include <cassert>

#include <TVector3.h>
#include <TFile.h>
#include <TMath.h>


/** @file EyeEventClassifier.cc
  * Implementation of the class EyeEventClassifier.
  * @author P. Privitera & P. Facal, INFN Rome
  */


#include <FdConfig.hh>
#include <FdNumbering.hh>
#include <ErrorLogging.hh>
#include <FdStdioStream.hh>
#include <TTimeStamp.hh>
#if (FDUTIL_VERSION_CODE >= ModuleVersionCode(3,6,0))
# include <FdGeometry.hh>
#endif // FDUTIL_VERSION_CODE


#include <MiEvent.hh>
#include <EyeEvent.hh>
#include <EyeEventFile.hh>

#include "CheckVersion.hh"

#include "SltPattern.hh"

#if (FDUTIL_VERSION_CODE < ModuleVersionCode(3,6,0))
/* Contains the pixels azimuthal and elevation angle */
# include "CameraGeometry.hh"
#endif // FDUTIL_VERSION_CODE

#include "T3PixelSel.hh"
#include "T3PixelList.hh"

#include "EyeEventClassifier.hh"

using namespace std;
using FdUtil::Fd;
using FdUtil::FdConfig;
using FdUtil::ErrorLogger;
#if (FDUTIL_VERSION_CODE >= ModuleVersionCode(3,6,0))
using FdUtil::FdGeometry;
#endif // FDUTIL_VERSION_CODE

static int fNMuon = 0, fNLongMuon = 0, fNCloseShower = 0, fNUnclassified = 0, 
           fNNoise = 0, fNMuonAndNoise = 0, fNShowerCandidate = 0, 
	   fNHorizontalShower = 0, fNEmpty = 0, fNLargeEvent=0;

#if (FDUTIL_VERSION_CODE < ModuleVersionCode(3,6,0))
// Angle between the East and bay 1 line of each Eye 
Float_t AngletoEast[FdUtil::Fd::kFD_NEYES]; 
#endif // FDUTIL_VERSION_CODE

// -------------------------------------------------------------------------
// -------------------------------------------------------------------------
//
// class EyeEventClassifier - implementation
//

EyeEventClassifier::EyeEventClassifier()
  : fT3List(NULL), fHeader(NULL), fEyePixelList(NULL), fTimeStamp(NULL),
    fPixelData(NULL), fEyeFadcData(NULL),
    fPrintLevel(2), fLogLevel(0), fLogfile(NULL), fLogfileName(0)
 {
  FD_CINFO << "EyeEventClassifier::EyeEventClassifier() - initializing ..." << endl;

#ifdef USE_UBA_EXTENSION
# warning UBA T3 Extension is activated!
  FD_CWARN << "EyeEventClassifier::EyeEventClassifier() - "
           << "using UBA T3 extension!" << endl;
#endif // USE_UBA_EXTENSION
  
   //  NT2mT1    = new TH1F("NT2mT1","NT2mT1 ",100,0,100);
   //  NT2vT1    = new TH2F("NT2vT1","NT2vT1 ",100,0,100,100,0,100);
   //  Ssp    = new TH1F("Ssp","Ssp ",100,0,20);
  fClassificationTime.Reset();
  Clear();
  
  fSDPTheta = -9999;
  fSDPPhi   = -9999;
  fSDPPhiGround = -9999;

#if (FDUTIL_VERSION_CODE < ModuleVersionCode(3,6,0))
//  AngletoEast[0] =  30.0;   // Los Leones, old value
//  AngletoEast[1] = -60.0;   // Morados, old value
//  AngletoEast[2] = 171.875; // Loma Amarilla, old value
//  AngletoEast[3] = 114.64;  // Coihueco, new value (10/2005)
//  AngletoEast[3] = 116.64;  // Coihueco, old value
  AngletoEast[0] =  29.995;   // Los Leones, new value (ARGENSIG 25 Oct. 2007)
  AngletoEast[1] = -59.992;   // Morados, new value (ARGENSIG 25 Oct. 2007)
  AngletoEast[2] = 171.865;   // Loma Amarilla , new value (ARGENSIG 25 Oct. 2007)
  AngletoEast[3] = 116.697;   // Coihueco, new value (ARGENSIG 25 Oct. 2007)
  if ( Fd::kFD_NEYES > 4 )
    AngletoEast[5] = 42.0;  // HEAT design value for bay 2 only
  if ( Fd::kFD_NEYES > 5 ) {
    FD_CFATAL << "EyeEventClassifier::EyeEventClassifier() - "
              << "table 'AngletoEast' is not foreseen for " << Fd::kFD_NEYES
	      << " FD eyes!" << endl;
  }
#endif // FDUTIL_VERSION_CODE

  //ShowSettings( FD_LOGSTREAM( ErrorLogger::eWarning ) );
}

// -------------------------------------------------------------------------

EyeEventClassifier::~EyeEventClassifier()
 {
  Clear();
  
  PrintStatistics();
  
  if ( fLogfile ) PrintStatistics( fLogfile );
  
  if ( fLogfile ) {
    fclose( fLogfile );
    fLogfile = 0;
  }
  // Histo for debug
  //  TFile hfile("T3histos.root","RECREATE","T3 ROOT file with histograms");
  //  NT2mT1->Write();  
  //  NT2vT1->Write();  
  //  Ssp->Write();  
  //  hfile.Close();

  // reset all counting variables
  
  fNMuon = 0;
  fNLongMuon = 0;
  fNCloseShower = 0;
  fNUnclassified = 0;
  fNNoise = 0;
  fNMuonAndNoise = 0;
  fNShowerCandidate = 0;
  fNHorizontalShower = 0;
  fNEmpty = 0;
  fNLargeEvent = 0;
}

// -------------------------------------------------------------------------

void EyeEventClassifier::ClassifyEvent() 
 {
  int tpmax = 1000000000;
  float signmax = 0.;
  int imax = 0;
  int tmax = 0;
  int tmin = 1000000000;

#ifdef DEBUG_T3_CLASSIFYEVENT  
    cout << "ClassifyEvent t3 pixels  " << fT3List->GetT3Pixels() << endl;
#endif    

  T3PixelSel *s;
  
  // Find the maximum pulse time and minimum pulse time of the selected pixels.    // Find the pulse time of the pixel with max signal.  
  float ssp = 0.;
  for ( int i=0; i<fT3List->GetT3Pixels(); i++ ) {
    if ( (s = fT3List->GetPixel(i)) == NULL ) break;
    int t = s->GetTPulse();
    float sign = s->GetPeakValue();
    if ( sign > signmax) {
      signmax = sign;
      tpmax=t;
      imax = i;
    }
    if ( t > 0 ) {
      float ss = s->GetSignOvSigm();
      ssp = ss+ ssp;
    }
    if ( t<0 ) t=s->GetT1();
    if ( t>0 && t>tmax ) tmax=t;
    if ( t>0 && t<tmin ) tmin=t;
  }
  ssp = ssp/float(fT3List->GetT3Pixels());

  // Fill histos
  //	Ssp->Fill(ssp);
 
  // Cut lightning. Require that after cleaning the number of T3 pixel
  // is more than 0.15 the number of T1 pixels 
  if ( (float(fT3List->GetT3Pixels())/float(fNT1)) < 0.15 ) {
#ifdef DEBUG_T3_CLASSIFYEVENT 
      cout << "Event Rejected - Large Event - N. T3 < 0.15 N. T1 " << endl;
#endif 
    fEventClass = TEyeEventHeader::kLargeEvent;
    return;
  }

  if ( tmax==0 && tmin==1000000000 ) {
    // All pixels are below threshold => kNoise
#ifdef DEBUG_T3_CLASSIFYEVENT 
      cout << "Event Rejected - Noise - All pixels below threshold " << endl;
#endif 
    fEventClass = TEyeEventHeader::kNoise;
    return;
  }
  else if ( tmax-tmin < 250 ) {
    // The difference between maximum and minimum time is at most 200 ns 
    // => kLongMuon
#ifdef DEBUG_T3_CLASSIFYEVENT 
      cout << "Event Rejected - Long Muon - Time Difference < 250 ns " << endl;
#endif 
    fEventClass = TEyeEventHeader::kLongMuon;
    return;
  }
  
  // count how many pixels have the same time of pixel with max pulse
  int ntpmax = 1;
  float ratio = 0.;
  for ( int i=0; i<fT3List->GetT3Pixels(); i++ ) {
    if ( (s = fT3List->GetPixel(i)) == NULL ) break;
    int t = s->GetTPulse();
    if ( t>0 && i != imax ) {
      if ( abs(t-tpmax) < 200 ) ntpmax++;
    }
  }
 
  ratio = (float)ntpmax/(float)(fT3List->GetT3Pixels());

  // Simple algorithm to find muon plus a noisy pixel. 
  // Find how many pixels have the same time of 
  //  the maximum and minimum time.
  int ntmax=0, ntmin=0;
  int n=0;
  for ( int i=0; i<fT3List->GetT3Pixels(); i++ ) {
    if ( (s = fT3List->GetPixel(i)) == NULL ) break;
    int t = s->GetTPulse();
    if ( t>0 ) {
      if ( (tmax-t) < 200 ) ntmax++;
      if ( (t-tmin) < 200 ) ntmin++;
      n++;
    }
  }
  
  if ( n>2 && ((ntmax==n-1) || (ntmin==n-1)) )
    // If all pixels have the same time apart from one => kMuonAndNoise
    //#ifdef DEBUG_T3_CLASSIFYEVENT 
    //      cout << "Event Rejected - Muon + Noise - All pixels have the same time apart one " << endl;
    //#endif 
    fEventClass = TEyeEventHeader::kMuonAndNoise;
  else if ( tmax - tmin < 1000 ) {
    // If the difference between maximum and minimum time 
    // is smaller than 1 microsecond, then
    // this is a nearby shower =>  kCloseShower  
#ifdef DEBUG_T3_CLASSIFYEVENT 
    // printf("Event Accepted - Close Shower - less than 1 microsec time difference  \n");
#endif 
    fEventClass = TEyeEventHeader::kCloseShower;
  }
  else  {
    // All what is left is considered a shower candidate => kShowerCandidate
#ifdef DEBUG_T3_CLASSIFYEVENT 
    cout << "Event Accepted - Shower Candidate " << endl;
#endif 
    fEventClass = TEyeEventHeader::kShowerCandidate;
  }
  
  if ( ratio > 0.5 ) {
    // If more than 50% of the pixels have the same time then muon+noise
#ifdef DEBUG_T3_CLASSIFYEVENT 
      cout << "Event Rejected - Muon + Noise - " << endl;
#endif 
    fEventClass = TEyeEventHeader::kMuonAndNoise;
  }

}

// -------------------------------------------------------------------------

void EyeEventClassifier::Clear()
 {
  // clear variables 
  
  for ( int c=0; c<=Fd::kMI_NCOLS;c++ )  
    for ( int r=0; r<=Fd::kMI_NROWS; r++ )
      fPixelSelected[c][r] = kFALSE;
  
  // THJM: is this really necessary ?

  fEyeFadcData = NULL;   // clear fadc;
  fPixelData = NULL;     // clear pdata; 
  fEyePixelList = NULL;  // clear pixel_list;
  fTimeStamp = NULL;     // clear fTimeStamp;
  fHeader = NULL;        // clear header; 

  if ( fT3List ) {
    delete fT3List; 
    fT3List = NULL;
  }

  fEventClass = TEyeEventHeader::kUnClassified;

  for ( int i=0; i<7; i++ )  {
    fTimePattern[i] = 100;
    fNanoTime[i] = -1;
  }

#if (EYEEVENT_VERSION_CODE >= ModuleVersionCode(2,0,0))
  fCurrentAxis.fAzimuth   = 0.;
  fCurrentAxis.fElevation = 0.;
#endif // EYEVENT_VERSION_CODE
  
  // clear result variables
  
  fTime0 = fTimeSlope = fAzimuth0 = fAzimuthSlope = 0.;
  fAzimuthAtGround = 0.;
  fTimeAtGround = 0;
}

// -------------------------------------------------------------------------

void EyeEventClassifier::ClearMirrorData()
 {
  // clear variables 
  
  for ( int c=0; c<=Fd::kMI_NCOLS;c++ )  
    for ( int r=0; r<=Fd::kMI_NROWS; r++ )
      fPixelSelected[c][r] = kFALSE;
   
  if ( fPixelData ) { // clear pixel data;
    fPixelData = NULL;
  }

  if ( fTimeStamp ) { // clear fTimeStamp; 
    fTimeStamp = NULL;
  }
}

// -------------------------------------------------------------------------

void EyeEventClassifier::ConvertAzimuthTime()
 {
  // Only Los Leones is being considered for the moment (EA)  

  fAzimuthAtGround = fAzimuth0;
  fAzimuthAtGround /= 57.2958;  // 180. / pi

  // The difference between the extrapolated shower time at ground 
  // and the local time of the first T2 is added to the GPS nanotime.
  // The factor 1000 comes from the local time units of 1 microseconds.
  // We assume that the GPSNanotime corresponds to the first T2.     

  fTimeAtGround = (Int_t)fTime0;
}

// -------------------------------------------------------------------------

#ifdef USE_UBA_EXTENSION
void EyeEventClassifier::FirstProcess() {

  int nPixels = fT3List->GetT3Pixels();
  int nNoPulse = fT3List->GetNullPixels();
  //cout << "np: " << NoPulse << endl;

  double noPulseRatio = nNoPulse/(float)nPixels;

  if ( (nPixels - nNoPulse) < 4 || noPulseRatio >= 0.5 ) {
    fEventClass = TEyeEventHeader::kUBANoise;
  }

}
#endif // USE_UBA_EXTENSION

// -------------------------------------------------------------------------

void EyeEventClassifier::ConvertAzimuthTimePixel()
 {
  T3PixelSel *s;
  
  for ( int i=0; i<fT3List->GetT3Pixels(); i++ ) {

    if ( (s = fT3List->GetPixel(i)) == NULL ) break;

    int t1 = s->GetT1();
    int tp = s->GetTPulse();
    float az = s->GetAzimuth();
    int mirr = s->GetMirror(); // 1..6

#if 0
    cout << "Mirr= " << mirr << " az= " << az << " t1= " << t1 <<" tp= " << tp << endl;
#endif

    // Set azimuth 
#if (FDUTIL_VERSION_CODE >= ModuleVersionCode(3,6,0))
    s->SetAzimuth( FdGeometry::GetPixelAzimuth( fHeader->GetEyeNo(), mirr, 
                                                s->GetCol(), s->GetRow() ) );
#else
    s->SetAzimuth(   (az + 15 + ( mirr - Fd::kFD_FIRST_MIRROR ) * 30 
                   - AngletoEast[fHeader->GetEyeNo()-Fd::kFD_FIRST_EYE]));
#endif // FDUTIL_VERSION_CODE

    // The difference between the extrapolated shower time at ground 
    // and the local time of the first T2 is added to the GPS nanotime.
    // The factor 1000 comes from the local time units of 1 microseconds.
    // We assume that the GPSNanotime corresponds to the first T2.     
    // Now the Raw time stamp is the time of the NxPage !
    // The time of the first T2 is T_NxPage - NxPageDelay
    // but NxPageDelay is no longer accessible

#if 0
    cout << "fNanoTime[" << mirr << "]=    " << setw(9) << fNanoTime[mirr] << endl;
    cout << "fTimePattern[" << mirr << "]= " << setw(4) << fTimePattern[mirr] << endl;
    cout << "t1=              " << setw(4) << t1 << endl;
#endif

    if ( t1 >= 0 ) {
      // float deltaT1 = t1 - fTimePattern[mirr];
      float deltaT1 = t1 - Fd::kMI_NSLT_BINS;
      // s->SetT1( (fNanoTime[mirr] + (int)(deltaT1*1000))/1000);
      // s->SetT1( fNanoTime[mirr] + (int)(deltaT1*1000) );
      // ??? s->SetT1( fNanoTime[mirr] + (int)(deltaT1*1000) + fTimeOffset[mirr] );
      s->SetT1( fNanoTime[mirr] + (int)(deltaT1*1000) );
    }
    
    if ( tp >= 0 ) {
      // float deltaTp = tp/10. - fTimePattern[mirr];
      float deltaTp = tp/10. - Fd::kMI_NSLT_BINS;
      // s->SetTPulse(  fNanoTime[mirr] + (int)(deltaTp*1000) );
      // ??? s->SetTPulse(  fNanoTime[mirr] + (int)(deltaTp*1000) + fTimeOffset[mirr] );
      s->SetTPulse(  fNanoTime[mirr] + (int)(deltaTp*1000) );
    }
    
    t1 = s->GetT1();
    tp = s->GetTPulse();
    az = s->GetAzimuth();
    mirr = s->GetMirror();
    
#if 0
    cout << "New Mirr= " << mirr << " az= " << az << " t1= " << t1 <<" tp= " << tp << endl;
#endif
  }
}

// -------------------------------------------------------------------------

void EyeEventClassifier::DoClassification(TEyeEvent* event)
 {
#if (EYEEVENT_VERSION_CODE >= ModuleVersionCode(2,0,0))
  static int event_no = -1;
#endif // EYEEVENT_VERSION_CODE

  fClassificationTime.Start( kFALSE );
  
  // Perform the classification of the passed TMirrorEvent, i.e. do a 
  // partial analysis of the passed data in order to find out its basic
  // signature.

  // Inizialization of variables
  fSDPTheta = 0.;
  fSDPPhi   = 0.;
  fSDPPhiGround = 0.;
  fTotSignal = 0.;

  Clear();

  fHeader = event->GetEventHeader();

#ifdef DEBUG_T3
  FD_COUT << "EyeEventClassifier::DoClassification() - start ..." << endl;
  FD_COUT << "\t\teye no= " << fHeader->GetEyeNo() << endl;

  for ( unsigned int mi=Fd::kEYE_FIRST_MIRROR; 
                     mi<=Fd::kEYE_LAST_MIRROR; mi++ ) {
    if ( fHeader->IsMirrorDataPresent( mi ) )
      FD_COUT << "\t\t mi_no= " << mi << endl;
  }
#endif // DEBUG_T3
  
#if (EYEEVENT_VERSION_CODE >= ModuleVersionCode(2,0,0))
  if ( event_no > (int)fHeader->GetEventNo() ) event_no = -1;
  
  if ( event_no == -1 )
    event->GetGeometryData()->Show(FD_LOGSTREAM(ErrorLogger::eInfo));

  event_no = fHeader->GetEventNo();
#endif // EYEEVENT_VERSION_CODE
  
  if ( (fEyePixelList = event->GetPixelList()) == NULL ) {
    FD_CERR << "EyeEventClassifier::DoClassification() - "
            << "no pixelList !" << endl;
    
    fEventClass = TEyeEventHeader::kEmpty;
    fNEmpty++;

    fClassificationTime.Stop();
    
    return;
  }

  if ( (fEyeFadcData = event->GetFADCData()) == NULL ) {
    
    FD_CERR << "EyeEventClassifier::DoClassification() - "
            << "no FADC data !" << endl;

    fEventClass = TEyeEventHeader::kEmpty;
    fNEmpty++;

    fClassificationTime.Stop();
    
    return;
  }
  
  TEyePixelData* eye_pixeld;
  if ( (eye_pixeld = event->GetPixelData()) == NULL ) {
    
    FD_CERR << "EyeEventClassifier::DoClassification() - "
            << "no Eye Triggerdata  " << endl;
    
    fEventClass = TEyeEventHeader::kEmpty;
    fNEmpty++;

    fClassificationTime.Stop();
    
    return;
  }

  fT3List = new T3PixelList();
  fNT1 = 0;

  // --- loop over all telescopes

  for ( unsigned int i=0; i<Fd::kEYE_NMIRRORS; i++ ) {

    ClearMirrorData();
    
    unsigned int mirror_id = i + Fd::kEYE_FIRST_MIRROR;
    fPixelData = eye_pixeld->GetPixelData( mirror_id );
    if ( fPixelData != NULL ) {
#ifdef DEBUG_T3 
      cout << "Eye = " << fHeader->GetEyeNo() << " , Mirror = " << mirror_id 
           << " , n.  triggered pixels = " 
	   << eye_pixeld->GetNumberOfTriggeredPixels(mirror_id, 0,Fd::kMI_NSLT_BINS) 
	   << endl; 
#endif // DEBUG_T3 
      // this is raw, i.e. uncorrected
      fTimeStamp = event->GetEventHeader()->GetRawMirrorTimeStamp( mirror_id );

      // this is raw, too
      fNanoTime[mirror_id] = fTimeStamp->GetNanoSec();
      // this is corrected and mustn't be mixed with raw values !!!
      // => we have to calculate the offsets between the mirrors ourself
      //    on the other side, they are not needed, as the time scales are the 
      //    same for all telescopes of an Eye
      fTimeOffset[mirror_id] = event->GetEventHeader()->GetMirrorTimeOffset( mirror_id );

      // as all calibration is using uncorrected nsec scale we have to correct
      // it finally !
      fCurrentMirror = mirror_id;

#if (   (EYEEVENT_VERSION_CODE >= ModuleVersionCode(2,0,7)) \
     && (FDUTIL_VERSION_CODE >= ModuleVersionCode(4,1,0)))
      // Set the geometry information for the current mirror which
      // is to be used by the PixelSelection() method
      
      // this scheme is possible within the DAQ & Offline
      event->GetGeometryData()->GetAxisDirection( mirror_id, fCurrentAxis );
      
#ifdef DEBUG_T3 
      //cout << "mirror_id= " << mirror_id << " (az,el) = ("
      //     << fCurrentAxis.fAzimuth << "," << fCurrentAxis.fElevation
      //   << ")" << endl;
#endif // DEBUG_T3
#endif // EYEEVENT_VERSION_CODE && FDUTIL_VERSION_CODE
      
      // Select pixels to be included in class T3PixelSel. 
      // Use T2 patterns and multipatterns.
      PixelSelection();     
    }
  }

#ifdef DEBUG_T3      
  cout << "*********** INITIAL LIST OF T3 PIXELS *** " << fT3List->GetT3Pixels() << " pixels selected **********" << endl;        
  ShowShowerPixels();
#endif

  // Fast muon rejection: check if the event contains 
  // only 4 pixels with the same T1 => kIsMuon. Use only T1 info.
  RejectMuons();

  // Fast rejection of lightening. There are more than 9 T1 pixels 
  // not participating in T2 patterns, and it is a large event (more than 14
  // T1 pixels) 
  // taken out 11/05/04
  //  fNT2 = fT3List->GetT3Pixels();
  //  Fill histos
  //  NT2mT1->Fill(fNT1-fNT2);
  //  NT2vT1->Fill(fNT1,fNT2);
  //  if ( fNT1-fNT2 > 9 &&  fNT1 > 14  ) {
  // Cut lightning
  //  fEventClass = TEyeEventHeader::kLargeEvent;
  //}

  if ( fEventClass == TEyeEventHeader::kUnClassified ) {  
    // Calculate for each pixel the time corresponding to the 
    // maximum in the FADC pulse, TPulse.
    // Only pixels with maximum>2.5*sigma are considered 
    // (muons have typically large signals).
    AssignPixel();
    PulseFinding();
#ifdef USE_UBA_EXTENSION
    FirstProcess();
#endif // USE_UBA_EXTENSION

#ifdef DEBUG_T3      
    cout << "*********** LIST OF T3 PIXELS after Pulse Finding *** " 
         << fT3List->GetT3Pixels() << " pixels selected **********" << endl;        
    ShowShowerPixels();
#endif

    ConvertAzimuthTimePixel();
    // Muon rejection: FADC data are used to find the time of the pixel pulse. 
    // If the maximum time difference between 
    // the pixels pulse time is 200 ns =>   kLongMuon 
    // If all are at the same time apart from one  =>  kMuonAndNoise 
    // If the maximum time difference between the pixels 
    // pulse time is > 100 ns and < 1 microsecond  => kCloseShower
    //
    // All what is remaining is a shower candidate!!!!!! 
#ifdef USE_UBA_EXTENSION
    if ( fEventClass != TEyeEventHeader::kUBANoise )
      ClassifyEvent();
#else
    ClassifyEvent();
#endif // USE_UBA_EXTENSION
  }

  //  if ( fEventClass == TEyeEventHeader::kShowerCandidate ){

  // Uncommented to process kCloseShower (sa,pp,19/03/2004)
  if ( fEventClass == TEyeEventHeader::kShowerCandidate 
       || fEventClass == TEyeEventHeader::kCloseShower) {

    // Shower candidate was found, calculate the T3 info
    ProcessShower();
  }
      
  // finally increase the appropriate counting variable

  switch ( fEventClass ) {

    case TEyeEventHeader::kLargeEvent: 
      fNLargeEvent++;
       break;

    case TEyeEventHeader::kIsMuon: 
      fNMuon++;
       break;

    case TEyeEventHeader::kLongMuon:
      fNLongMuon++;
      break;

    case TEyeEventHeader::kCloseShower:
      fNCloseShower++;
      break;

    case TEyeEventHeader::kNoise:
      fNNoise++;
      break;

    case TEyeEventHeader::kMuonAndNoise:
       fNMuonAndNoise++;
       break;

    case TEyeEventHeader::kEmpty:
      fNEmpty++;
      break;

    case TEyeEventHeader::kShowerCandidate:
      fNShowerCandidate++;
      break;

    case TEyeEventHeader::kHorizontalShower:
       fNHorizontalShower++;
       break;

    default:
      fNUnclassified++;
      break;
  }

#ifdef DEBUG_T3
  cout << "*********** FINAL LIST OF T3 PIXELS *** " 
       << fT3List->GetT3Pixels() << " pixels selected **********" << endl;        
  ShowShowerPixels();
#endif

  //  Show( fPrintLevel );
 
  fClassificationTime.Stop();
  
  if ( fPrintLevel > 2 )
    printf( "DoClassification: <time>= %f ms\n", 
	    fClassificationTime.RealTime() * 1000. 
	    / (Double_t)fClassificationTime.Counter() );

#ifdef DEBUG_T3        
  event->GetEventHeader()->Show();
#endif

  Show( fPrintLevel, FdUtil::ErrorLogger::GetInstance().GetStream() );
  
  if ( fLogfile ) Show( fLogLevel, fLogfile );
}

// -------------------------------------------------------------------------

void EyeEventClassifier::FitTime()
 {
  //** Linear fit:  fT1 = fTimeSlope*fElevation + fTime0 , in local coordinates.
  //** NOTE: we fit the using the T1 time. We have also tried with Tpulse, but
  //**       the difference is only at the level of 1 microseconds. 
  
  double sx=0., sy=0., sx2=0., sxy=0., sy2=0.;
  int nn = fT3List->GetT3Pixels();
  int n = 0;

  if ( nn <= 3 ) {
    fEventClass = TEyeEventHeader::kNoise;
    return;
  }
  
  T3PixelSel *s;
  
  for ( int i=0; i<nn; i++ ) {
    
    // nn might be increased below !?
    if ( (s = fT3List->GetPixel(i)) == NULL ) break;
    
    float y = s->GetT1();
    float x = s->GetElevation();  
    
    if ( fEventClass == TEyeEventHeader::kHorizontalShower ) {
      x = s->GetAzimuth();
    } 
    
    if ( y > 0 ) {
      sx += x;
      sy += y;
      sx2 += x*x;
      sy2 += y*y;
      sxy += x*y;
      n++;
    }
  }

  double det = sx*sx/(double)n - sx2; 

  if ( fabs(det) > 0.1 ) {

    fTimeSlope = (sx*sy/(double)n - sxy)/det;
    fTime0 = (sy/(double)n) - (fTimeSlope*sx/(double)n);
  
    double chi2 = sy2 + (double)n*fTime0*fTime0 
                      + fTimeSlope*fTimeSlope*sx2 
		      + 2*fTimeSlope*fTime0*sx 
		      - 2*fTime0*sy 
		      - 2*fTimeSlope*sxy;

    //** Noise do not have the correct timing sequence, which is reflected 
    //** in a large value of the chi2. 
    if ( chi2/(double)(n-1) > 500000.) {
#ifdef DEBUG_T3_FITTIME      
      cout << "FitTime chi2/dof " << chi2/(double)(n-1) 
           << " > 500000 , Noise" << endl; 
#endif      
      fEventClass = TEyeEventHeader::kNoise;
      return;
    }
    
    //** NOTE: we add 1 microseconds to the extrapolated time, to take into 
    //** account the fact that T1 time is approx 1 microsecond before the 
    //** pulse time (due to the T1 running sum). Checked also by Tpulse.
    //    fTime0 += 1000.; 
  } else {
    //** If the time fit fails then we do not know what to do.
#ifdef DEBUG_T3_FITTIME      
      cout << "FitTime determinant " << det << " < 0.1 , Unclassified" << endl; 
#endif      
    fEventClass = TEyeEventHeader::kUnClassified;
  }
}

// -------------------------------------------------------------------------

void EyeEventClassifier::FitTrack()
 {
  // Do linear fit:  fAzimuth = fAzimuthSlope*fElevation + fAzimuth0 , 
  //                 in local coordinates.
  double sx=0., sy=0., sx2=0., sy2=0., sxy=0.;
  int n = fT3List->GetT3Pixels();
  T3PixelSel *s;

  for ( int i=0; i<n; i++ ) {
    if ( (s = fT3List->GetPixel(i)) == NULL ) break;
    float y = s->GetAzimuth();
    float x = s->GetElevation();
    sx += x;
    sy += y;
    sx2 += x*x;
    sy2 += y*y;
    sxy += x*y;
  }

  double det = sx*sx/(double)n - sx2; 
  if ( fabs(det) > 0.1 ) {
    fAzimuthSlope = (sx*sy/(double)n - sxy)/det;
    fAzimuth0 = (sy/(double)n) - (fAzimuthSlope*sx/(double)n);
    double chi2 = sy2 + (double)n*fAzimuth0*fAzimuth0 
                      + fAzimuthSlope*fAzimuthSlope*sx2 
		      + 2*fAzimuthSlope*fAzimuth0*sx 
		      - 2*fAzimuth0*sy 
		      - 2*fAzimuthSlope*sxy;
   
    // Scattered noise will give a bad fit, which is
    // reflected in a large value of the chi2.  
    if ( chi2/(double)(n-1) > 5. ) {
#ifdef DEBUG_T3_FITTRACK
      cout << " FITTRACK Chi2/dof = " << chi2/(double)(n-1) << " > 5 , Noise " << endl;
#endif      
      fEventClass = TEyeEventHeader::kNoise;
      return;
    }
  }
  else {
    // in case the determinant is close to zero
#ifdef DEBUG_T3_FITTRACK
    cout << " FITTRACK determinant = " << det << " < 0.1 , Horizontal Shower " << endl;
#endif      
    fEventClass = TEyeEventHeader::kHorizontalShower;
  }
    
}

// -------------------------------------------------------------------------

void EyeEventClassifier::PixelTrackReject()
 {
  // Reject the pixel with max contribution to chi2, 
  // until fit converges or list has less than 4 pixels 

#ifdef DEBUG_T3_FITTRACK
  FD_COUT << "PixelTrackReject() - event classified as Noise" << endl;
#endif      

  int imaxdelta2 = 0;
  float maxdelta2;
  
  while ( 1 ) {
   
    maxdelta2 = 0.;
    bool reject_pixel = false;
    int nn = fT3List->GetT3Pixels();
  
    fEventClass = TEyeEventHeader::kShowerCandidate; //?

    if ( nn <= 3 ) {
      fEventClass = TEyeEventHeader::kNoise;
      break;
    }
    
    T3PixelSel *s;
    
    for ( int i=0; i<nn; i++ ) {
      if ( (s = fT3List->GetPixel(i)) == NULL ) break;
      float y = s->GetAzimuth();
      float x = s->GetElevation();

      float delta2 = pow(y-(x*fAzimuthSlope+fAzimuth0),2);
      if ( delta2 > maxdelta2 ) {
	maxdelta2 = delta2;
	imaxdelta2 = i;
	reject_pixel = true;
      }
    } // for ( int i=0; i<nn; ...

    if ( reject_pixel )
      fT3List->DeleteT3Pixel( imaxdelta2 );
    else
      break;
    
    FitTrack();
    
    if ( fEventClass == TEyeEventHeader::kShowerCandidate) break;
    
  } // while ( 1 ) ...
}

// -------------------------------------------------------------------------

void EyeEventClassifier::OpenLogfile(const char* filename)
 {
  if ( !fLogfile )
    fLogfile = fopen( filename, "w" );

  fLogLevel = 1;  // set the default log level
}

// -------------------------------------------------------------------------

void EyeEventClassifier::PixelSelection()
 {
  // Algorithm to select pixels 
  bool trigpix[Fd::kMI_NPIXELS+1];

  for ( int pix=Fd::kFD_FIRST_PIXEL; pix<=Fd::kMI_NPIXELS; pix++ )
    trigpix[pix] = false;
  
  for ( int col=Fd::kMI_FIRST_COL; col<=Fd::kMI_NCOLS; col++ ) {    
    for ( int bin=0; bin<Fd::kMI_NSLT_BINS-1; bin++ ) { //Bug in Pixeldata?
      TMirrorPixelData::PixelDataWord data;
      data = fPixelData->GetPixelData(col,bin);
      int t = data->fWord & (0x03f800000);
      t = t >> 23;
      for ( int k=0; k<Fd::kMI_NROWS; k++ ) {
	if ( (data->fWord & (0x0001<<k)) != 0 )
	  trigpix[Fd::kMI_NROWS*(col-1)+(k+1)] = true;
      }
    }
  }
  
  int np = 0;
  
  for ( int pix=Fd::kFD_FIRST_PIXEL; pix<=Fd::kMI_NPIXELS; pix++ ) {
    if ( trigpix[pix] ) np++;
  }
  
  fNT1 = fNT1 + np;

  // Loop over T2 time (100 bins of 1 microsecond)
  
  for ( int time=0; time<Fd::kMI_NSLT_BINS; time++ ) {

    // Loop over columns for each T2 time bin 
    for ( int col=Fd::kMI_FIRST_COL; col<=Fd::kMI_NCOLS; col++ ) {
      
      TMirrorPixelData::PixelDataWord data;
      data = fPixelData->GetPixelData(col,time);
      
      // Check if a pattern is present in the column
      int pattern = data->fWord & (0x03f800000);
      pattern = pattern >> 23;
     
      // Set the time of the first pattern: needed to define the T3 GPS nanotime 
      // at the end of the classification 
      if ( pattern!=0 && time<fTimePattern[fCurrentMirror] )
	fTimePattern[fCurrentMirror] = time;

      if ( pattern==0 ) {
	// No pattern found. Continue the search.
	continue;
      }
      else if ( pattern<127 ) {
        //cout << " pattern " << pattern << endl;
	// A 4 pixels pattern found.
        // Loop over the 22 rows of the column to find pixels matching the pattern  
	for ( int k=0; k<Fd::kMI_NROWS; k++ ) {
	  // Find a pixel in this row which was on at the time of the pattern. This is
          // the first pixel of the pattern.
	  if ( ((data->fWord & (0x0001<<k)) != 0) ) {
            // Validate the starting pixel checking if the other expected pixels (given the pattern)
            // have also T1. If ok, fill T3PixelSel 
	    TestPattern(pattern,col,k+1, time);
	  }
	}
      }
      else if ( pattern== 127 ) {   // A multipattern found.
 
        // Loop over the 22 rows of the column to find pixels matching the pattern  
	for ( int k=0; k<Fd::kMI_NROWS; k++ ) {
	  // Find a pixel in this row which was on at the time of the multipattern. 
	  if ( ((data->fWord & (0x0001<<k)) != 0) ) {
            // Validate the starting pixel given the multipattern.  
            // If ok, fill T3PixelSel 
	    TestMultiPattern(col,k+1, time);
	  }
	}
      }  // multipattern
    } // column
  }  // time

  if ( fT3List->GetT3Pixels() < 4 ) {
    fEventClass = TEyeEventHeader::kEmpty;
  }
}

// -------------------------------------------------------------------------

void EyeEventClassifier::PixelTimeReject()
 {
  // Reject the pixel with max contribution to chi2, 
  // until fit converges or list has less than 3 pixels 

  int imaxdelta2 = 0;
  float maxdelta2;

  while ( 1 ) {
    
    maxdelta2 = 0.;
    bool reject_pixel = false;
    int nn = fT3List->GetT3Pixels();
  
    fEventClass = TEyeEventHeader::kShowerCandidate; //?

    if ( nn <= 3 ) {
      fEventClass = TEyeEventHeader::kNoise;
      break;
    }
    
    T3PixelSel *s;

    for ( int i=0; i<nn; i++ ) {
      if ( (s = fT3List->GetPixel(i)) == NULL ) break;
      float y = s->GetT1();
      float x = s->GetElevation();
      float delta2 = pow(y-(x*fTimeSlope+fTime0),2);
      if ( delta2 > maxdelta2 ) {
	maxdelta2 = delta2;
	imaxdelta2 = i;
        reject_pixel = true;
      }
    } // for ( int i=0; i<nn; ...)

    if ( reject_pixel )
      fT3List->DeleteT3Pixel( imaxdelta2 );
    else
      break;
    
    FitTime();
    
    if ( fEventClass == TEyeEventHeader::kShowerCandidate) break;
    
  } // while ( 1 ) ...
}

// -------------------------------------------------------------------------

void EyeEventClassifier::PrintStatistics(FILE *str)
 {
  int n =   fNMuon + fNNoise + fNEmpty + fNLongMuon + fNMuonAndNoise
          + fNCloseShower + fNShowerCandidate + fNUnclassified+ fNLargeEvent;  
  
  if ( n == 0 ) return;  // nothing classified
  
  fprintf( str, "************ T3 Statistics *************\n" );
  fprintf( str, "*        Large Event    %4d           *\n", fNLargeEvent ); 
  fprintf( str, "*              Muons    %4d           *\n", fNMuon ); 
  fprintf( str, "*              Noise    %4d           *\n", fNNoise );
  fprintf( str, "*              Empty    %4d           *\n", fNEmpty );
  fprintf( str, "*         Long Muons    %4d           *\n", fNLongMuon ); 
  fprintf( str, "*     Muon and noise    %4d           *\n", fNMuonAndNoise );
  fprintf( str, "*      Close showers    %4d           *\n", fNCloseShower );
  fprintf( str, "* Horizontal Showers    %4d           *\n", fNHorizontalShower );
  fprintf( str, "*  Shower candidates    %4d           *\n", fNShowerCandidate );
  fprintf( str, "*       Unclassified    %4d           *\n", fNUnclassified );
  fprintf( str, "*                                      *\n" );
  fprintf( str, "*              Total    %4d           *\n", n );
  fprintf( str, "*                                      *\n" );

  fprintf( str, "* <Real time> [msec]  %6.2f           *\n",
                  fClassificationTime.RealTime() * 1000. 
		/ (Double_t)fClassificationTime.Counter() );
  fprintf( str, "* <CPU time>  [msec]  %6.2f           *\n", 
                  fClassificationTime.CpuTime() * 1000 
		/ (Double_t)fClassificationTime.Counter() );

  fprintf( str, "****************************************\n" );

  fflush( str );
}

// -------------------------------------------------------------------------

void EyeEventClassifier::ProcessShower()
 {
  // Find shower azimuth at ground (local coordinates).
#ifdef DEBUG_T3_FITTRACK      
  cout << "*********** LIST OF T3 PIXELS before PixelTrack *** " << fT3List->GetT3Pixels() << " pixels selected **********" << endl;        
  ShowShowerPixels();
#endif

  FitTrack();

  // Reject pixels with position inconsistent with fit 
  if ( fEventClass == TEyeEventHeader::kNoise ) {
    PixelTrackReject();  
#ifdef DEBUG_T3_FITTRACK      
  cout << "*********** LIST OF T3 PIXELS AFTER PixelTrackReject *** " << fT3List->GetT3Pixels() << " pixels selected **********" << endl;        
  ShowShowerPixels();
#endif
    if ( fEventClass == TEyeEventHeader::kNoise ) return;
  }

  // No extrapolation if horizontal shower or chi2 noise
  //  if ( fEventClass == TEyeEventHeader::kHorizontalShower ) return;

  // Find shower time at ground (local time).
#ifdef DEBUG_T3_FITTIME      
  cout << "*********** FITTIME ********************** " << endl;        
#endif
  FitTime();

  if ( fEventClass == TEyeEventHeader::kUnClassified ) return;
  
  // Reject pixels with time inconsistent with fit 
  if ( fEventClass == TEyeEventHeader::kNoise ) {
    PixelTimeReject();  
    if ( fEventClass == TEyeEventHeader::kNoise ) return;
#ifdef DEBUG_T3_FITTIME      
    cout << "*********** LIST OF T3 PIXELS AFTER PixelTimeReject *** " << fT3List->GetT3Pixels() << " pixels selected **********" << endl;        
    ShowShowerPixels();
#endif
    ClassifyEvent();
    FitTrack();
  }

  if ( fEventClass == TEyeEventHeader::kHorizontalShower ) return;
  if ( fEventClass == TEyeEventHeader::kUnClassified ) return;
  if ( fEventClass == TEyeEventHeader::kNoise ) return;

  // Convert azimuth and time from local to CDAS reference system and time. 
  //cout << "convert  " << endl;  
  ConvertAzimuthTime();
  
  FindSDP();
  
  CalculateTotalSignal();
}

// -------------------------------------------------------------------------

void EyeEventClassifier::CalculateTotalSignal()
 {
  // Calculate total signal of the shower
  fTotSignal = 0.;
  T3PixelSel *s;
  
  for ( int i=0; i<fT3List->GetT3Pixels(); i++ ) {
    if ( (s = fT3List->GetPixel(i)) == NULL ) break;
    fTotSignal += s->GetIntegral();
  }
}

// -------------------------------------------------------------------------

void EyeEventClassifier::AssignPixel()
 {
  T3PixelSel *s;
  
  // Find pixel position in TEyePixelList in order to retrieve FADC data
  for ( int i=0; i<fT3List->GetT3Pixels(); i++ ) {

    if ( (s = fT3List->GetPixel(i)) == NULL ) break;
    
    int t3Mirror = s->GetMirror();
    int t3col = s->GetCol();
    int t3row = s->GetRow();

    for ( unsigned int j=0; j<fEyePixelList->GetNumPixels(); j++ ) {  

      Fd::FdPixelNumber fdpixel = fEyePixelList->GetPixel( j );
      int pixcol = Fd::GetColumnNo( Fd::GetPixelNumber( fdpixel ) );
      int pixrow =  Fd::GetRowNo( Fd::GetPixelNumber( fdpixel ) );
      // thjm: Mirror number in the range 1..24
      int pixmir =   Fd::GetEyeMirrorNo( fdpixel ) ;
      if ( pixmir == t3Mirror && pixcol == t3col && pixrow == t3row ) {
        s->SetfPixel(j);
        //	cout << " Found in pixel list " << j << endl;
        //	cout << " Mirror " << t3Mirror << " column " << t3col << " row " << t3row << endl;
        //	cout << " T3pixel fdpixel added to list " << i << endl;
      }
      else if ( pixmir != t3Mirror ) {
        //	cerr << "EyeEventClassifier::AssignPixel() - "
        //   << "could not add pixel at position " << j << endl
        //    << "pixmir= " << pixmir << " t3Mirror= " << t3Mirror << endl;
      }
    }
  }
}

// -------------------------------------------------------------------------

void EyeEventClassifier::PulseFinding()
 {
#ifdef USE_UBA_EXTENSION
  int no_pulse = 0;
#endif // USE_UBA_EXTENSION
  
  // Calculate the time corresponding to the maximum ADC in the pulse, Tpulse
  
  T3PixelSel *s;
  float value;
  
//  for ( int i=1; i<fT3List->GetT3Pixels()+1; i++ ) {
//     cout << "N. Pixels in T3List" << fT3List->GetT3Pixels() << endl;
  for ( int i=0; i<fT3List->GetT3Pixels(); i++ ) {
  
    double ped = 0., sig = 0.;
    
    if ( (s = fT3List->GetPixel(i)) == NULL ) break;
    
#ifdef DEBUG_T3_PULSE      
    cout << "Col " <<  s->GetMirror() 
         << "Col " <<  s->GetCol() 
         << " Row " << s->GetRow() 
	 << endl;
#endif   
    // --- TEyePixelList: 0 ... GetNumPixel()-1
    
    Fd::FdPixelNumber fdpixel = fEyePixelList->GetPixel( s->GetfPixel() );

    if ( fdpixel == NULL ) {
      cerr << "FdPixelNumber is NULL, returned by TEyePixelList(" 
           << s->GetfPixel() << ") " << i << endl;
      cerr << "TEyePixelList::GetNumPixels()= "
           << fEyePixelList->GetNumPixels() << endl;
      continue;
    }
    
    const TFADCData* fadc = fEyeFadcData->GetFADCData( fdpixel );    

    // well, this should not really happen
    
    if ( fadc == NULL ) {
      cerr << "TEyeFADCData::GetFADCData(" 
           << Fd::GetPixelNo( fdpixel ) << ") is NULL !" << endl;
      cerr << "\tloop index= " << i << endl;
      continue;
    }
    
#if (MIEVENT_VERSION_CODE >= ModuleVersionCode(4,0,0))
    // factor accounts for FADC bin width
    
    int factor =   fadc->GetBinWidth( TFADCData::kFADC_RESOLUTION_1000_BINS )
                 / fadc->GetBinWidth();
#else
    int factor = 1;
#endif // MIEVENT_VERSION_CODE

    TFADCData::FADCDataWord fadc_word_ptr = fadc->GetFADCTrace();
    
    // Calculate pedestal and its RMS, 200 bins for HEAT

    for ( int k=0; k<100*factor; k++ ) {
      value = FADCDataWordGetData( &fadc_word_ptr[k] );
      ped += (double)value;
      sig += (double)value*value;
    }
    ped /= (100. * factor);
    sig /= (100. * factor);
    
    sig = sqrt(sig-ped*ped);
    // cout << "Pedestal " << ped << ". Sigma " << sig << endl;
    
    // Define start and end time to find maximum    

    // tstart = T1 time - 1 microsecond
    // tend   = tstart + 3 microseconds
    // int tstart = s->GetT1()*10-10;
    // int tend = tstart + 30;
    
    // tstart = T1 time - 2 microsecond
    // tend   = tstart + 5 microseconds
#if (MIEVENT_VERSION_CODE >= ModuleVersionCode(4,0,0))
//    int tstart = max( s->GetT1()*10-20, 0 );
    int tstart = max( factor * (s->GetT1()*10 - 20), 0 );
//    int tend = min( (unsigned int)(tstart + 50), fadc->GetNFADCBins()-1 );
    int tend = min( (unsigned int)(tstart + factor * 50), fadc->GetNFADCBins()-1 );
#else
    int tstart = max( s->GetT1()*10-20, 0 );
    int tend = min( tstart + 50, Fd::kMI_NFADC_BINS-1 );
#endif // MIEVENT_VERSION_CODE
    
    float pulsemax = 0; 
    int tmax = 0;
    float pulseint = 0; 

    for ( int t=tstart+1; t<tend; t++ ) {      
      value = (  FADCDataWordGetData( &fadc_word_ptr[t-1] )
	       + FADCDataWordGetData( &fadc_word_ptr[t] )
	       + FADCDataWordGetData( &fadc_word_ptr[t+1] ) ) / 3.;

#ifdef DEBUG_T3_PULSE
      cout << "time value " << t << " " << value << endl;      
#endif

      pulseint = pulseint + value;
      //     value = FADCDataWordGetData( &fadc_word_ptr[t] );
      if ( value > pulsemax ) {
	pulsemax = value;
	tmax = t;
      }
    }
#ifdef DEBUG_T3_PULSE      
    cout << "tmax integ pede " << tmax << " " << pulseint << " " << ped << endl;
#endif

    pulseint = pulseint-ped*(float)(tend - tstart - 1);

#ifdef DEBUG_T3_PULSE      
    cout << "Pulse " << pulseint << endl;
#endif
    // Define threshold cut = pedestal + 2.5*rms      
    double cut = ped + 2.5*sig; 
    double max = (double)pulsemax;

    // Set Tpulse = tmax if above threshold, else Tpulse = -1
    if ( max > cut ){
      s->SetTPulse(int(tmax));
      s->SetPeakValue(pulsemax);
      s->SetIntegral(pulseint);
      float sigovsigm = 0.;
      if ( sig > 0. ) {
	sigovsigm = (pulsemax-ped)/sig;
      }
      s->SetSignOvSigm(sigovsigm);
    }
    else {
      s->SetTPulse(-1);
      s->SetIntegral(0.);
#ifdef USE_UBA_EXTENSION
      no_pulse++;
#endif // USE_UBA_EXTENSION
    }
  }   // for ( int i=0; i<fT3List->GetT3Pixels(); ...

#ifdef USE_UBA_EXTENSION
  fT3List->SetNullPixels(no_pulse);
#endif // USE_UBA_EXTENSION
}

// -------------------------------------------------------------------------

void EyeEventClassifier::RejectMuons()
 {
  // Reject muons: the event has only 4 pixels selected and all have the same 
  // T1 time
  if ( fT3List->GetT3Pixels() == 4 ) {
    int t[4];
    T3PixelSel *s;
    for ( int i=0; i<4; i++ ) {
      if ( (s = fT3List->GetPixel(i)) == NULL ) break;
      t[i]=s->GetT1();
      //           cout << "t[" << i <<"] = " << t[i] << endl; 
    }
    if ( t[0]==t[1] && t[0]==t[2] && t[0]==t[3] )
      fEventClass=TEyeEventHeader::kIsMuon;
  } // fT3List->GetT3Pixels() == 4
}

// -------------------------------------------------------------------------

void EyeEventClassifier::Show(unsigned int level, FILE *str)
 {
  assert( str != NULL );
  
  FdUtil::OstdioStream ostr( str );
  
  Show( level, ostr );
}

// -------------------------------------------------------------------------

void EyeEventClassifier::Show(unsigned int level, ostream& ostr)
 {
  ostringstream log_prefix;
  char str[120];

#if (defined (_FD_INF_STR_)) && (defined (_FD_LOGINFO_STR_)) 
  log_prefix << _FD_INF_STR_ << _FD_LOGINFO_STR_;
#endif // _FD_INF_STR_ && _FD_LOGINFO_STR_
  log_prefix << "EyeEventClassifier - ";

#ifdef DEBUG_T3
  FD_COUT << "Event " << setw(5) << fHeader->GetEventNo() 
          << " Shower Candidate " << setw(2) << fT3List->GetT3Pixels() 
	  << " pixels selected" << endl;

  printf("TSl %8.2f T0 %8.2f ASl %8.2f A0 %8.2f GAz %8.2f GTm  %10d\n" ,
	          fTimeSlope, fTime0, fAzimuthSlope, fAzimuth0, 
		  fAzimuthAtGround, fTimeAtGround );
#endif
 
  if ( level == 1 && fEventClass == TEyeEventHeader::kShowerCandidate ) {
  
    ostr << log_prefix.str()
         << "Event " << setw(5) << fHeader->GetEventNo() 
         << " Shower Candidate " << setw(2) << fT3List->GetT3Pixels() 
	 << " pixels selected" << endl;

    sprintf( str, "TSl %8.2f T0 %8.2f ASl %8.2f A0 %8.2f GAz %8.2f GTm %10d\n" ,
	          fTimeSlope, fTime0, fAzimuthSlope, fAzimuth0, 
		  fAzimuthAtGround, fTimeAtGround );
    ostr << log_prefix.str() << str << endl;

    //     fprintf( str, "******** Event %3d: Shower Candidate *******\n",
    //	          fHeader->GetEventNo());
    //     fprintf( str, "*  %2d pixels selected                      *\n", 
    //          fT3List->GetT3Pixels());
//     //    fprintf( str, "*  GPS Time    --> Second  = %10d    *\n", TimeConvert::GetGPSTimeFromUTCTime( fTimeStamp->GetSec() ));
//     //   fprintf( str, "*                  NSec    = %10d    *\n", 
//     //	          fTimeStamp->GetNanoSec());
  //    fprintf( str, "*  Time   Fit  --> Slope   = %8.2f      *\n", fTimeSlope);
  //   fprintf( str, "*                  Time0   = %8.2f      *\n", fTime0);
  //   fprintf( str, "*  Azimuth Fit --> Slope   = %8.2f      *\n", fAzimuthSlope);
  //    fprintf( str, "*                  Slope0  = %8.2f      *\n", fAzimuth0);
  //     fprintf( str, "*                                          *\n"); 
  //    fprintf( str, "*  Ground      --> Azimuth = %8.2f      *\n", fAzimuthAtGround);
  //     fprintf( str, "*                  Time    = %10d    *\n", fTimeAtGround);
  //    fprintf( str, "********************************************\n");
  }
  else if ( level == 1 && fEventClass == TEyeEventHeader::kCloseShower ) {

    ostr << log_prefix.str()
         << "Event " << setw(5) << fHeader->GetEventNo() 
         << " : Close Shower " << setw(2) << fT3List->GetT3Pixels() 
	 << " pixels selected" << endl;

    //   fprintf( str, "******** Event %3d: Close Shower *********\n",
    //	 fHeader->GetEventNo());
    //   fprintf( str, "*  %2d pixels selected                    *\n",
    //   fT3List->GetT3Pixels()); 
    //   //fprintf( str, "*  GPS Time   --> Second = %10d    *\n", TimeConvert::GetGPSTimeFromUTCTime( fTimeStamp->GetSec() ));
    //   //fprintf( str, "*                 NSec   = %10d    *\n", fTimeStamp->GetNanoSec() );
    //   fprintf( str, "******************************************\n");
  }

  if ( level >= 2 ) {

    sprintf( str, "T3 results: event %3d, sel pixels= %d - class= %s", 
	          fHeader->GetEventNo(),  
	          fT3List->GetT3Pixels(),
	          TEyeEventHeader::GetVerboseEventClass( fEventClass ) );
    ostr << log_prefix.str() << str << endl;
    
    if ( fEventClass == TEyeEventHeader::kShowerCandidate 
	 || fEventClass == TEyeEventHeader::kCloseShower ) {
      sprintf( str, "GAz %4.2f GTm %8d SDP-Az %4.2f SDP-Z %4.2f TotSig %.0f",
                    fAzimuthAtGround, fTimeAtGround, fSDPPhi, fSDPTheta, 
		    fTotSignal );  
      ostr << log_prefix.str() << str << endl;
    }

 
//     fprintf( str, ">>> Results for event %3d:\n", fHeader->GetEventNo() );
//     fprintf( str, "%d pixels selected\n", fT3List->GetT3Pixels()) ;
//     if ( level == 3 ) fT3List->Show(str);
//     fprintf( str,"%s\n", TEyeEventHeader::GetVerboseEventClass( fEventClass ) );
//     if ( fEventClass == TEyeEventHeader::kShowerCandidate 
//     || fEventClass == TEyeEventHeader::kCloseShower ) {
//       //    fprintf( str, "GPS Time  = %8d\nGPS NTime = %8d\n", TimeConvert::GetGPSTimeFromUTCTime( fTimeStamp->GetSec() ),fTimeStamp->GetNanoSec());
//         fprintf( str, "Azimuth  Ground = %4.2f\nNanoTime     = %8d\n", fAzimuthAtGround, fTimeAtGround );
//         fprintf( str, "SDP Normal Azimuth  = %4.2f  Zenith  = %4.2f\n", fSDPPhi, fSDPTheta );
//     }
//     if ( fEventClass == TEyeEventHeader::kCloseShower ) {
// //     fprintf( str, "GPS Time  = %8d\nGPS NTime = %8d\n", TimeConvert::GetGPSTimeFromUTCTime( fTimeStamp->GetSec() ), fTimeStamp->GetNanoSec() );
//     }
//     fprintf( str, "-------------------------------\n" );
  }
    
  ostr.flush();
}

// -------------------------------------------------------------------------

void EyeEventClassifier::ShowSettings(std::ostream& ostr)
 {
  ostr << "=== Current settings of class EyeEventClassifier ===" << endl;
  
#if (FDUTIL_VERSION_CODE < ModuleVersionCode(3,6,0))
  for ( unsigned int i=0; i<Fd::kFD_NEYES; i++ ) {
    ostringstream msg;
    msg << "\tAngletoEast[" << i << "]= " << AngletoEast[i];
    INFO(msg);
  }
#else
  unsigned int eye_id = FdUtil::GetEyeId();
  
  unsigned int n_mirrors = Fd::kEYE_NMIRRORS;
# if (FDUTIL_VERSION_CODE >= ModuleVersionCode(4,0,0))
  if ( eye_id == Fd::kFD_NEYES )
    n_mirrors = Fd::kHEAT_NMIRRORS;
  else
    n_mirrors = Fd::kEYE_NMIRRORS;
# endif // FDUTIL_VERSION_CODE

  // output telescope axises for all telescopes of current site
  for ( unsigned int mid=1; mid<=n_mirrors; mid++ ) {
    
# if (FDUTIL_VERSION_CODE >= ModuleVersionCode(4,1,0))
    double az, el;
# else
    float az, el;
# endif // FDUTIL_VERSION_CODE
    int abs_tel_id = Fd::GetFdMirrorNo( eye_id, mid );
    
    FdGeometry::GetTelescopeAxis( abs_tel_id, az, el );
    
    ostr << "\tAxis[" << abs_tel_id << "]= (" << az << "," << el << ")" << endl;
  }
#endif // FDUTIL_VERSION_CODE

  ostr << "====================================================" << endl;
}

// -------------------------------------------------------------------------

void EyeEventClassifier::ShowShowerPixels()
 {
  T3PixelSel *s;
  
  for ( int i=0; i< fT3List->GetT3Pixels(); i++ ) {
    if ( (s = fT3List->GetPixel(i)) == NULL ) break;
    int t3Mirror = s->GetMirror();
    int t3col = s->GetCol();
    int t3row = s->GetRow();
    
    printf("Mirror %3d , Column %3d , Row %3d , T1 %10d , TPulse %10d , Azimuth %8.2f  , Elevation %8.2f \n" ,t3Mirror,t3col,t3row,s->GetT1(),s->GetTPulse(),s->GetAzimuth(),s->GetElevation());
  }
}

// -------------------------------------------------------------------------

void EyeEventClassifier::SetLogLevel(UInt_t level)
 {
  // This has no effect if the log file has not been set
  
  if ( fLogfile )
    fLogLevel = level;
}

// -------------------------------------------------------------------------

void EyeEventClassifier::TestMultiPattern(int col, int row, int time)
 {
  // Find pixels consistent with a multipattern starting from pixel at (col,row) at time
  int tot=0, newtot=0;
  int k=row-1;
  
  // Collecting pixels in the window column+4, (row-4,row+4)
  int dcmax = 5; 
  int drmin = -4; 
  int drmax = 5;

  // Count how many pixels in the window were on at the multipattern time
  for ( int dc=0; dc<dcmax; dc++ ) {
    for ( int dr=drmin; dr<drmax; dr++ ) {
      if (((col+dc)<=Fd::kMI_NCOLS) && ((row+dr)>=1) && ((row+dr)<=Fd::kMI_NROWS) ) {
	TMirrorPixelData::PixelDataWord datap;
	datap = fPixelData->GetPixelData(col+dc,time);  
	int t0 = datap->fWord & (0x0001<<(k+dr));
	if ( t0 ) {
	  tot++;
	  if ( !fPixelSelected[col+dc][row+dr] ) newtot++;
	}
      }
    }
  }

  // The first time a multipattern is found starting from pixel at col,row
  // we test all 108 patterns and add all pixels consistent to patterns.
  // This test will be done again only if at least one new pixel is found. 
  // (Once a T2 is on, it will be on for several time slots on the same 
  // pattern or multipattern, due to the 20 microseconds T1 length) 
  
  if ( tot>=4 && newtot>0 ) {
    for ( int i=1; i<=108; i++ ) TestPattern(i,col,row,time); 
  }
}

// -------------------------------------------------------------------------

void EyeEventClassifier::TestPattern(int pattern, int col, int row, int time)
 {
  // Validate the starting pixel (col,row) checking if the other expected pixels (given the pattern)
  // were also on at the time of the pattern.

  int p[4];  // pixels numbers for this pattern
  short r[4]; // row number
  short c[4]; // col number
  int  t[4]; // test t1
	
  p[0] = Fd::kMI_NROWS*(col-1)+row; //p[0] is the origin of the pattern
  c[0] = ((p[0] - 1) / Fd::kMI_NROWS) + 1;
  r[0] = ((p[0] - 1) % Fd::kMI_NROWS) + 1;
	     
  // find candidate pixels of the pattern, given the starting pixel
  for ( int m=1; m<4; m++ ) {
    if ( (row%2)==0 ){
            p[m] = p[0] + SltPattern::fgDiffEven[pattern-1][m-1]; //abs pixel number
      // p[m] = p[0] + fgDiffEven[pattern-1][m-1]; //abs pixel number
    }
    else {
      p[m] = p[0] + SltPattern::fgDiffOdd[pattern-1][m-1]; // abs pixel number
      // p[m] = p[0] + fgDiffOdd[pattern-1][m-1]; // abs pixel number
    }
    c[m] = ((p[m] - 1) / Fd::kMI_NROWS ) + 1; // col for this pixel
    r[m] = ((p[m] - 1) % Fd::kMI_NROWS ) + 1; // row for this pixel
  
    //cout << " col " << c[m] << " row "<< r[m] << endl;
    // Check if they were on at the pattern time
    TMirrorPixelData::PixelDataWord datap;
    if ( c[m]<=Fd::kMI_NCOLS && c[m]>=1 && r[m]<=Fd::kMI_NROWS && r[m]>=1) {
      datap = fPixelData->GetPixelData(c[m],time); 
      t[m] = datap->fWord & (0x0001<<(r[m]-1)); //test t1
    } else t[m]=0;   
  }

  // If all were on at the pattern time, the pattern is validated
  if ( (t[1]!=0) && (t[2]!=0) &&(t[3]!=0) ) {
    for ( int m=0; m<4; m++ ) {
      // Check if pixels were already included in T3PixelList
      if ( !fPixelSelected[c[m]][r[m]]) {
	//! Now find T1 time. T1 is the time bin (before the T2 multipattern trigger)
        //! at which the pixel is first on.
	//!          ____________________
	//!         |                    |
	//!    _____|                    |________
        //!          T1<--      T2
	int t0=1,t1;
	for ( t1=time; t0!=0 && t1>0;) {
	  t1--;
	  TMirrorPixelData::PixelDataWord datap;
	  datap = fPixelData->GetPixelData(c[m],t1);
	  t0 = datap->fWord & (0x00000001<<(r[m]-1)); //test t1
	}
	
	// cout << "Adding pixel to list " << endl;
	// Pixel is added to the T3PixelList
#if (FDUTIL_VERSION_CODE >= ModuleVersionCode(4,1,0))
//# ifndef DPA
//	T3PixelSel *selpix = new T3PixelSel( FdUtil::GetEyeId(), 
//	                                     fCurrentMirror,
//	                                     c[m], r[m], t1+1 );
//# else
	T3PixelSel *selpix = new T3PixelSel( fCurrentAxis, 
	                                     fCurrentMirror,
	                                     c[m], r[m], t1+1 );
//# endif // DPA
#else
	T3PixelSel *selpix = new T3PixelSel( fCurrentMirror,
	                                     c[m], r[m], t1+1 );
#endif // FDUTIL_VERSION_CODE
	if ( !selpix ) {
	  FD_CERR << "EyeEventClassifier::TestPattern() - "
	          << "new T3PixelSel() returned NULL-ptr!" << endl;
	}
	else
          fT3List->AddT3Pixel(selpix);
	//cout << "done " << endl;
	fPixelSelected[c[m]][r[m]] = kTRUE;
      }
    }
  }
}

// -------------------------------------------------------------------------

void EyeEventClassifier::FindSDP()
 {
  T3PixelSel *s;
  int n = fT3List->GetT3Pixels();
  
  TVector3 meandir(0,0,0); // sum of all the hit pixel directions 
  
  for ( int i=0; i<n; i++ ) {  
    if ( (s = fT3List->GetPixel(i)) == NULL ) break;
    
    float y = s->GetAzimuth()*TMath::Pi()/180.;
    float x = (90.-s->GetElevation())*TMath::Pi()/180.;
    TVector3 dir(1,1,1);
    dir.SetTheta(x);
    dir.SetPhi(y);
    meandir+= dir;
  } // for i
  
  meandir.Unit();
  
  // the SDPnormal is the sum of the cross product of each
  // pixel with the sum of all the directions  
  
  TVector3 SDPnormal(0,0,0); 
  
  for ( int i=0; i<n; i++ ) {
  
    if ( (s = fT3List->GetPixel(i)) == NULL ) break;
    
    float y = s->GetAzimuth()*TMath::Pi()/180.;
    float x = (90.-s->GetElevation())*TMath::Pi()/180.;

    TVector3 dir(1,1,1);
    dir.SetTheta(x);
    dir.SetPhi(y);

    TVector3 dircross(0,0,0);
    dircross = dir.Cross(meandir);
     if ( dircross.Phi() < 0. ) {
       dircross.SetPhi(dircross.Phi() + TMath::Pi());
       dircross.SetTheta(TMath::Pi()-dircross.Theta());
    }
   
    SDPnormal += dircross;
  }
  
  SDPnormal.Unit();
  
  fSDPTheta = SDPnormal.Theta();
  fSDPPhi   = SDPnormal.Phi();

  fSDPPhiGround = -atan2(cos(fSDPPhi),sin(fSDPPhi));
  // better definition of azimuth at ground

  // Following line commented out because of possible ambiguities in the
  // definition of sdp that may lead to phi_g behind the eye
  // Commenting out keeps linear approximation (pp,19/03/2004)
  //  fAzimuthAtGround = -atan2(cos(fSDPPhi),sin(fSDPPhi));
#ifdef DEBUG_CLASSIFICATION      
  cout << "SDP Theta: "<<  fSDPTheta << " SDP Phi: "<< fSDPPhi << endl; 
  cout << "SDP Phi ground: "<<  fSDPPhiGround << endl; 
#endif    

}

// -------------------------------------------------------------------------
// -------------------------------------------------------------------------
