
//
// File   : HoughSel.cc
//
// Purpose: Implementation of the class HoughSel
//
// $Id: HoughSel.cc 11539 2010-09-30 14:31:54Z mathes $
//


#include <iostream>
#include <string>
#include <cassert>
#include <cmath>

/** @file HoughSel.cc
  * Implementation of the class HoughSel.
  * @author F. Salamida, INFN
  */


#include <TTree.h>
#include <TFile.h>
#include <TBranch.h> 
#include <TH1.h>
#include <TMath.h>   // TMath::Pi

#include <FdNumbering.hh>
#include <ErrorLogging.hh>
#if (FDUTIL_VERSION_CODE >= ModuleVersionCode(3,6,0))
# include <FdGeometry.hh>
#endif // FDUTIL_VERSION_CODE
#include <MiEvent.hh>

#include "CheckVersion.hh"

#if (FDUTIL_VERSION_CODE < ModuleVersionCode(3,6,0))
/* Contains the pixels azimuthal and elevation angle */
# include "CameraGeometry.hh"
#endif // FDUTIL_VERSION_CODE

#include "HoughSel.hh"

using namespace std;
using namespace FdUtil;

// -------------------------------------------------------------------------
// -------------------------------------------------------------------------
// 
// class HoughSel - implementation
//

HoughSel::HoughSel(const char *name,bool fill_tree)
  : fHoughFile(NULL), fHoughTree(NULL)
 {
  if ( fill_tree ) {
  
    assert( name != NULL && *name != 0 );
  
    // --- setup name for the output file (name.root -> name_hough.root)
  
    string filename( name );
    string tag( ".root" );
  
    if ( filename.find( tag ) != string::npos )
      filename.replace( filename.find( tag ), tag.size(), "_hough.root" );
    else {
      filename = "hough.root";
      FD_CWARN << "HoughSel::HoughSel() - "
               << "chosen generic name '" << filename 
	       << "' for Hough output file!" << endl;
    }
  
    fHoughFile = new TFile( filename.c_str(), "recreate" );
    fHoughTree = new TTree( "h", "Hough" );
  
    // --- setup branch names, variables must be class members
  
    fHoughTree->Branch( "mirror", &fMirror, "mirror/I" );
    fHoughTree->Branch( "npixels", &fNPixels, "npixels/I" );
    fHoughTree->Branch( "evtype", &fEventType, "evtype/I" );
    fHoughTree->Branch( "nT1", &nT1, "nT1/I" );
    fHoughTree->Branch( "nC1", &nC1, "nC1/I" );
    fHoughTree->Branch( "nPair", &nPair, "nPair/I" );
    fHoughTree->Branch( "nbin", &nbin, "nbin/I" );
    fHoughTree->Branch( "xh", xh, "xh[nC1]/F" );
    fHoughTree->Branch( "yh", yh, "yh[nC1]/F" );
    fHoughTree->Branch( "hsl", hsl, "hsl[nPair]/F" );
    fHoughTree->Branch( "hin", hin, "hin[nPair]/F" );
    fHoughTree->Branch( "hcmax", &hcmax, "hcmax/F" );
    fHoughTree->Branch( "hcwin", &hcwin, "hcwin/F" );
    fHoughTree->Branch( "hcbas", &hcbas, "hcbas/F" );
    fHoughTree->Branch( "hcrms", &hcrms, "hcrms/F" );
    fHoughTree->Branch( "hazim", &hazim, "hazim/F" );
    fHoughTree->Branch( "evtnum", &fEventNo, "evtnum/I" );
  } // if ( fill_tree ) ...
} 

// -------------------------------------------------------------------------

HoughSel::~HoughSel()
 {
  if ( fHoughFile ) {
    // Finish();
    delete fHoughFile;
  }
}

// -------------------------------------------------------------------------

void HoughSel::DoHoughSel(TMirrorBitArray& pixT1,
                          int npix, int mirr, int evnum, int type)
 {  
  if ( fHoughFile ) fHoughFile->cd();
  
  // for the TTree filling ...
  nT1        = npix;
  fEventType = type;
  fMirror    = mirr;
  fNPixels   = npix;
  fEventNo   = evnum;
  
  ///////////////////////////////////////////////////////////////////////////
  // Cluster T1 pixels
  ///////////////////////////////////////////////////////////////////////////

  Cluster(pixT1, nC1);

  ///////////////////////////////////////////////////////////////////////////
  // Translate T1 clusters (Origin = centre of gravity)
  ///////////////////////////////////////////////////////////////////////////

  Translate(xh, nC1);
  Translate(yh, nC1);

  ///////////////////////////////////////////////////////////////////////////
  // Loop over pairs: calculate slopes and intercepts (unused)
  ///////////////////////////////////////////////////////////////////////////

  nPair = 0;

  for ( int i=0; i<nC1; i++ ) { 
    for ( int j=i; j<nC1; j++ ) { 
      if (yh[i] != yh[j]) {
        hsl[nPair] = (xh[i] - xh[j]) / (yh[i] - yh[j]);
        hin[nPair] = xh[i] - hsl[nPair] * yh[i];
        // printf(" ... %d %d %d --> hsl %f hin %f \n",nPair,i,j,hsl[nPair],hin[nPair]);
        nPair++;
      }
    }
  }
  
  ///////////////////////////////////////////////////////////////////////////
  // Define number of bins in Slope histogram
  ///////////////////////////////////////////////////////////////////////////

  int m = (int)(1800. / float(nPair));
  if ( m > 10 ) m = 10;
  if ( m < 1 ) m = 1;
  nbin = 180 / m;

  ///////////////////////////////////////////////////////////////////////////
  // Fill Slope histogram
  ///////////////////////////////////////////////////////////////////////////

  char  hist[6];
  sprintf(hist,"ev%d",fEventNo);
  TH1F * hev  = new TH1F( hist, hist, nbin, -90, 90 );
  for ( int n=0; n<nPair; n++ ) { 
    float phi = atan (hsl[n]) * 180 / TMath::Pi();
    hev->Fill(phi);
  }

  hcmax = 0;
  int bmax = 0;
  for ( int k=0; k<nbin; k++ ) {
    if ( hev->GetBinContent(k + 1) > hcmax ) {
      hcmax = hev->GetBinContent(k + 1); 
      bmax = k + 1;
    }
  }

  ///////////////////////////////////////////////////////////////////////////
  // Find peak (hcmax), window content (hcwin), baseline (hcbas) and RMS (hcrms)
  ///////////////////////////////////////////////////////////////////////////

  float h1 = 0;
  float h2 = 0;
  float hw = 0;
  hazim    = 0;
  int	b1 = bmax - nbin/18;
  int	b2 = bmax + nbin/18;
  if ( b1 < 0 )	b1 = 0;
  if ( b2 > nbin ) b2 = nbin;
  for ( int k=1; k<nbin + 1; k++) {
    if (k >= b1 && k <= b2) {
      hw += hev->GetBinContent(k);
      hazim += hev->GetBinContent(k) * hev->GetBinCenter(k);
    }
    else {
      h1 += hev->GetBinContent(k);
      h2 += hev->GetBinContent(k) * hev->GetBinContent(k);
    }
  }
  hazim /= hw;
  int	nout = (nbin - b2 + b1 - 1);
  hcbas = h1 / nout;
  hcwin = hw / (b2 - b1 + 1);
  hcrms = sqrt((h2 - nout * hcbas * hcbas) / (nout - 1));

  ///////////////////////////////////////////////////////////////////////////
  // Discrimination parameter 
  ///////////////////////////////////////////////////////////////////////////

  float Fdis = 0.;
  if ( hcrms != 0 ) {
    Fdis = (hcwin - hcbas)/ hcrms;
  }
  else {
    Fdis = 9999.;
  }
  
  printf (" ++ nbin %3d bmax %3d hazim %7.3f hcmax %6.3f hcwin %6.3f hcbas %7.3f hcrms %7.3f ===> Fdis= %7.2f\n",
               nbin, bmax, hazim, hcmax, hcwin, hcbas, hcrms, Fdis); 

  if ( fHoughFile ) hev->Write();
  
  hev->Delete();
}

// -------------------------------------------------------------------------

int HoughSel::FindFLTrigger(TMirrorPixelData *pixel_data, int col, int row) 
 {
  bool trig = false;
  int  T1   = 0;
  
  // Loop over T2 time (100 bins of 1 microsecond)
  
  for ( unsigned int bin=0; bin<Fd::kMI_NSLT_BINS-1; bin++ ) { 
   
    TMirrorPixelData::PixelDataWord data;
    
    data = pixel_data->GetPixelData(col,bin);
    
    // Check if a pattern is present in the column
    int pattern = data->fWord & (0x03f800000);
    
    pattern = pattern >> 23;
    
    if ( ( data->fWord & (0x0001<<(row-1))) != 0) {
      if ( !trig ) T1 = bin;
      trig = true;
    }
  }
  
  return T1;
}

// -------------------------------------------------------------------------

void HoughSel::Cluster(TMirrorBitArray& pix, int &N)
 {
  int clust[440]; 
  
  for ( int r = 0; r < 22; r++ ) {
    int fill = 0;
    for ( int c = 19; c >= 0; c-- ) {
      if ( pix.Test(c+1,r+1) ) fill++;
      else fill = 0; 
      clust[22*c+r] = fill;
    }
  }

  N = 0;
      
  for (int r = 0; r < 22; r++) {
    float barx = 0;
    float bary = 0;
    int Nc = 0;
    int cs = 1;
    for (int c = 0; c < 20; c++) {
      Nc = clust[22*c+r];
      if ( pix.Test(c+1,r+1) && Nc > 0) {
    	if ( bary == 0 ) cs = Nc;
#if (FDUTIL_VERSION_CODE >= ModuleVersionCode(3,6,0))
# if ((FDUTIL_VERSION_CODE == ModuleVersionCode(3,6,0)) \
     || (FDUTIL_VERSION_CODE == ModuleVersionCode(4,0,0)))
    	 barx += FdGeometry::GetPixelAzimuth( c, r );
    	 bary += FdGeometry::GetPixelElevation( c, r );
# else
    	 barx += FdGeometry::GetPixelAzimuth0( c, r );
    	 bary += FdGeometry::GetPixelElevation0( c, r );
# endif // FDUTIL_VERSION_CODE == 3.6.0 || 4.0.0
#else
    	 barx += gPixelAzimuth[22*c+r];
    	 bary += gPixelElev[22*c+r];
#endif // FDUTIL_VERSION_CODE
    	 //printf(" +++%d r %d c %d x clust %d \n",N+1,r+1,c+1,Nc);
      }
      if ( Nc == 1 ) {
    	 xh[N] = barx / cs;
    	 yh[N] = bary / cs;
    	 //printf(" ...%d r %d c %d clust %d x %f y %f \n",N+1,r+1,c+1,cs,xh[N],yh[N]);
    	 N++;
    	 barx = 0;
    	 bary = 0;
      } 
    }
  }
}

// -------------------------------------------------------------------------

void HoughSel::Translate(float x[], int &N)
 { 
  float xm = 0.0;
  
  for (int n = 0; n < N; n++) xm += x[n]; 
  
  xm /= N;
  
  for (int n = 0; n < N; n++) x[n] -= xm; 
}

// -------------------------------------------------------------------------

void HoughSel::FillTree(int evtype)
 {
  fEventType = evtype;
  
  if ( fHoughTree ) {
    fHoughTree->Fill();
  }
  else {
    FD_CWARN << "HoughSel::FillTree(" << evtype << ") - "
             << "hough output file and tree are disabled!" << endl;
  }
}

// -------------------------------------------------------------------------

void HoughSel::Finish()
 {
  if ( fHoughFile ) {
  
    // Close Output tree
    fHoughFile->cd();
    fHoughTree->Write();

    fHoughFile->Close();
  }
}

// -------------------------------------------------------------------------
// -------------------------------------------------------------------------
