
//
// File   : T3PixelSel.hh
//
// Purpose: Declaration of the class T3PixelSel.
//
// $Id: T3PixelSel.hh 11539 2010-09-30 14:31:54Z mathes $
//

/** @file T3PixelSel.hh
  * Declaration of the class T3PixelSel.
  * @author P. Privitera & P. Facal, INFN Rome
  */

#ifndef _T3PixelSel_hh_
#define _T3PixelSel_hh_

#include <FdUtilVersion.hh>

// --- forward declaration(s)

#if (FDUTIL_VERSION_CODE >= ModuleVersionCode(4,1,0))
# include <FdGeometry.hh>
#endif // FDUTIL_VERSION_CODE

/** This class contains for each event the pixels selected by the T3 algorithm. These pixels will be used by EyeEventClassifier to 
  * It is used to classify the event.
  *
  * The class contains for each pixel the following informations:
  * @li fCol = pixel column
  * @li fRow = pixel row
  * @li fAzimut = pixel Azimutal angle
  * @li fElevation = pixel elevation angle
  * @li fT1 = time of first level trigger (in units of 1 microseconds)
  * @li fTPulse = time of the maximum ADC (in units of 100 ns) 
  *
  * NOTE: fAzimut is defined with respect to azimuth of the center of 
  *       the camera, positive anticlockwise.
  */
class T3PixelSel {
public:
#if (FDUTIL_VERSION_CODE >= ModuleVersionCode(4,1,0))
    /** Constructor for the class T3PixelSel. */
    T3PixelSel(FdUtil::FdGeometry::CameraPointing_t&,
               int mirror, int col, int row, int t1);
    
    /** Constructor for the class T3PixelSel. */
    T3PixelSel(FdUtil::FdGeometry&,
               int mirror, int col, int row, int t1);
#else
    /** Constructor for the class T3PixelSel. */
    T3PixelSel(int mirror, int col, int row, int t1);
#endif // FDUTIL_VERSION_CODE
    
#if (FDUTIL_VERSION_CODE >= ModuleVersionCode(3,6,1))
    T3PixelSel(int eye, int mirror, int col, int row, int t1);
#endif // FDUTIL_VERSION_CODE

    /** Destructor of the class T3PixelSel. */
    ~T3PixelSel() { }

    /** Get the Mirror of the current pixel. */
    int GetMirror() { return fMirror; } 

    /** Get the column index of the current pixel. */
    int GetCol() { return fCol; } 

    /** Get the row index of the current pixel. */
    int GetRow() { return fRow; } 

    /** Get the pixel number in the pixel list to recover FADC. */
    int GetfPixel() { return fPixel; }

    /** Get the time when the T1 happened. */
    int GetT1() { return fT1; } 

    /** Get the time of the pulse of the current pixel. */
    int GetTPulse() { return fTPulse; }

    /** Set the Signal over sigma of the pulse of the current pixel. */
    float GetSignOvSigm() { return fSignOvSigm; }

    /** Get the value of pedestal subtracted ADC corresponding to TPulse. */
    double GetPeakValue() { return fPeak; }

    /** Get the integral of the signal. */
    float GetIntegral() { return fIntegral; }
	 
    /** Return the azimuth angle of the pixel. */
    float GetAzimuth() { return fAzimuth;}

    /** Return the elevation angle of ... . */
    float GetElevation() { return fElevation; }

    /** Set the azimuth angle of the pixel. */
    void SetAzimuth(float az) { fAzimuth = az;}

    /** Set the time when the T1 happened. */
    void SetT1(int t) { fT1 = t; } 

    /** Set the time of the pulse of the current pixel. */
    void SetTPulse(int t) { fTPulse = t; }

    /** Set the integral of the signal. */
    void SetIntegral(float integral) { fIntegral = integral; }

    /** Set the Signal over sigma of the pulse of the current pixel. */
    void SetSignOvSigm(float a) { fSignOvSigm = a; }

    /** Set the value of pedestal subtracted ADC corresponding to TPulse. */	 	
    void SetPeakValue(double peak) {fPeak =peak;}	
	 
    /** Set the pixel n. in the pixel list to recover FADC. */
    void SetfPixel(int ipixel) { fPixel = ipixel; }

private:
    int             fMirror;
    int             fCol;
    int 	    fRow;
    int   	    fPixel;
    float 	    fAzimuth;
    float 	    fElevation;
    int 	    fT1;
    int 	    fTPulse;
    double          fIntegral;
    double          fPeak;	
    float           fSignOvSigm;
};

#endif // _T3PixelSel_hh_
