
//
// File   : AnalysedPixelData.cc
//
// Purpose: Implementation of the class TAnalysedPixelData
//
// $Id: AnalysedPixelData.cc 9725 2010-09-28 07:31:17Z mathes $
//


#include <iostream>


/** @file AnalysedPixelData.cc
  * Implementation of the class TAnalysedPixelData.
  * @author H.-J. Mathes, FzK
  */

#include <FdNumbering.hh>
#include <ErrorLogging.hh>
#include <MiEvent.hh>

#include "CheckVersion.hh"

#include "AnalysedPixelData.hh"
#include "PixelPattern.hh"

using namespace std;
using namespace FdUtil;

// -------------------------------------------------------------------------
// -------------------------------------------------------------------------
//
// class TAnalysedPixelData - implementation
//

unsigned int TAnalysedPixelData::fgPrintLevel = 0;

TAnalysedPixelData::TAnalysedPixelData(TMirrorPixelData* pixeld)
 : fFirstSltBin(0), fLastSltBin(Fd::kMI_NSLT_BINS-1), 
   fNMaxPattern(5), fNMultiPattern(0)
 {
  TMirrorPixelData::operator=( *pixeld );
  
  fBitArray = new TMirrorBitArray( pixeld );
  
  // set 'global' pointer to the pixel data ...
  
  fRawPixelData = (uint32_t *)GetPixelData();
  
  // clear all pattern lists
  
  fFoundPattern.clear();
  fValidPattern.clear();
  fInvalidPattern.clear();
}

// -------------------------------------------------------------------------

TAnalysedPixelData::~TAnalysedPixelData()
 {
  if ( fgPrintLevel > 1 )
    FD_COUT << "TAnalysedPixelData::~TAnalysedPixelData()" << endl;
  
  if ( fFoundPattern.size() ) {  // erase the 'found' pattern list
  
    list<PixelPattern*>::const_iterator iter = fFoundPattern.begin();
    list<PixelPattern*>::const_iterator end = fFoundPattern.end();
    
    while ( iter != end ) {
      delete *iter;
      iter++;
    }
    
//    fFoundPattern.erase( fFoundPattern.begin(), fFoundPattern.end() );
    fFoundPattern.clear();
  }
  
  if ( fValidPattern.size() ) {  // erase the 'fValid' pattern list
  
    vector<PixelPattern*>::const_iterator iter = fValidPattern.begin();
    vector<PixelPattern*>::const_iterator end = fValidPattern.end();
    
    while ( iter != end ) {
      delete *iter;
      iter++;
    }
    
    fValidPattern.clear();
  }
  
  if ( fInvalidPattern.size() ) {  // erase the 'fInvalid' pattern list
  
    list<PixelPattern*>::const_iterator iter = fInvalidPattern.begin();
    list<PixelPattern*>::const_iterator end = fInvalidPattern.end();
    
    while ( iter != end ) {
      delete *iter;
      iter++;
    }
    
    fInvalidPattern.clear();
  }
  
  delete fBitArray;
}

// -------------------------------------------------------------------------

void TAnalysedPixelData::Analyse()
 {
  FindPattern();
  
  if ( fgPrintLevel > 1 ) {
    FD_COUT << "FindPattern() - found " << fFoundPattern.size() 
            << " set pattern(s)" << endl;
  }
  
  // processing the patterns assigning pixels ...
  
  if ( fFoundPattern.size() ) {
  
    list<PixelPattern*>::const_iterator iter = fFoundPattern.begin();
    list<PixelPattern*>::const_iterator end = fFoundPattern.end();
    
    while ( iter != end ) {
      
      if ( (*iter)->GetPatternId() > PixelPattern::kNPattern ) {
        
#if 0
        // cannot process pattern class 0x7f for the moment
      
        if ( fgPrintLevel ) {
	  FD_CWARN << "Skipped processing of pattern 0x7f !" << endl;
	}

	// add to the invalid pattern list ...
	
	AddInvalidPattern( new PixelPattern( *(*iter) ) );

	fNMultiPattern++;
#else	
	AssignMultiPatternAK( *iter );
	
	bool valid = (*iter)->Validate( fBitArray );
	  
        if ( fgPrintLevel >= 1 )
	  FD_COUT << "AssignMultiPatternAK => valid= " << (int)valid << endl;
	  
	// add a copy of the valid pattern to the valid list ...
	// todo: avoid copying !
	  
	if ( valid ) {
	  AddValidPattern( new PixelPattern( *(*iter) ) );
	    
	  if ( fgPrintLevel > 1 )
	    (*iter)->Show( "TAnalysedPixelData::Analyse()" );
	}
	else {
	  FD_CERR << "TAnalysedPixelData::Analyse() - "
	          << "found invalid 0x7f pattern !" << endl;
	  
	  // add to the invalid pattern list ...
	  
	  AddInvalidPattern( new PixelPattern( *(*iter) ) );

	  fNMultiPattern++;
	}
#endif
        
	iter++;
        
	continue;
      }
      
      // should use also time criteria (is pattern at the time of the trigger)
      
      unsigned int pattern_id = (*iter)->GetPatternId();
      unsigned int start_col  = (*iter)->GetStartColumn();
      unsigned int pixel_in_col = fBitArray->GetNumberOfBitsInColumn( start_col );
      
      if ( pixel_in_col == 0 ) {
      
	FD_CERR << "TAnalysedPixelData::Analyse() - "
	        << "found invalid (no pixels in col) " 
	        << hex << (*iter)->GetPatternId()
	        << dec << "pattern !" << endl;
	
	// add to the illegal pattern list ...
	
	AddInvalidPattern( new PixelPattern( *(*iter) ) );
      }
      
      if ( fgPrintLevel ) {
        FD_COUT << "Pattern class 0x" << hex << pattern_id << dec
                << " has " << pixel_in_col 
	        << " set pixel(s) in column " << start_col << endl;
      }
      
      // find row, trivial if only one set bit in column
      
      for ( int row=PixelPattern::fgMinRow[pattern_id-1]; 
                row<=PixelPattern::fgMaxRow[pattern_id-1]; row++ ) {
      
        if ( fBitArray->Test( start_col, row ) ) {
          
	  AssignPixels( (*iter), row );
	      
	  // must use also time criteria
	  bool valid = (*iter)->Validate( fBitArray );
	  
          if ( fgPrintLevel >= 1 )
	    FD_COUT << "AssignPixels => valid= " << (int)valid << endl;
	  
	  // add a copy of the valid pattern to the valid list ...
	  // todo: avoid copying !
	  
	  if ( valid ) {
	    AddValidPattern( new PixelPattern( *(*iter) ) );
	    
	    if ( fgPrintLevel > 1 )
	      (*iter)->Show( "TAnalysedPixelData::Analyse()" );
	  }
	  
	  // don't need to search further, so break
	  
	  if ( pixel_in_col == 1 ) break;
	  
	  if ( valid ) break;
	}
      }  // row= ...
      
      iter++;
      
    }  // while ( iter != fFoundPattern.end() ) {

  }  // if ( fFoundPattern.size() )
  
  // assign now SLT times

  for ( unsigned int n=0; n<GetNValidPattern(); n++ ) {
  
    AssignPixelTimes( GetValidPattern( n ) );
  }
}

// -------------------------------------------------------------------------

void TAnalysedPixelData::AssignPixelTimes(PixelPattern *pattern)
 {
  int start_bin;
  PixelDataWord pixelword;
  unsigned int col, row;
	  
  for ( unsigned int n=0; n<pattern->GetNPixel(); n++ ) {
	  
    unsigned int len = 0;
    
    pattern->GetColumnRow( n, col, row );
	    
    unsigned int row_mask = 1 << ( row - Fd::kMI_FIRST_ROW);
	
    // find pulse start
    
    for ( start_bin = fFirstSltBin; start_bin<fLastSltBin+1; start_bin++ ) {

      pixelword = TMirrorPixelData::GetPixelData( col, start_bin );
      if ( PixelDataGetData( pixelword ) & row_mask ) break;
    }

    // find pulse end
    
    for ( int bin=start_bin; bin<fLastSltBin+1; bin++ ) {
    
      pixelword = TMirrorPixelData::GetPixelData( col, bin );
      if ( !(PixelDataGetData( pixelword ) & row_mask) ) break;
      len++;
    }
	  
    if ( fgPrintLevel ) {
      FD_COUT << "Pixel[" << n << "] starts at: " << start_bin * 10
	      << " and stops at: " << (start_bin + len) * 10 << endl;
    }
     
    pattern->SetSltTime( n, col, start_bin, start_bin + len );
  }
}

// -------------------------------------------------------------------------

void TAnalysedPixelData::AssignPixels(PixelPattern *pattern,unsigned int row)
 {
  unsigned int col = pattern->GetStartColumn();
  unsigned int id  = pattern->GetPatternId();
  
  pattern->Reset();
  
  if ( fgPrintLevel >= 1 ) {
    FD_COUT << "Assigning pixels for pattern 0x" << hex << id << dec
            << " starting @ (col,row)= (" << col << "," << row << ")" << endl;
  }
  
  unsigned int pcol = col;
  unsigned int prow = row;
  
  pattern->AddPixel( pcol, prow );
  
  if ( fgPrintLevel >= 2 )
    FD_COUT << "{ " << PixelPattern::fgPattern[id-1][0]
            << ", " << PixelPattern::fgPattern[id-1][1]
            << ", " << PixelPattern::fgPattern[id-1][2] << " }" << endl;
  
  for ( unsigned int i=1; i<PixelPattern::kNPixel; i++ ) {
    
    if ( fgPrintLevel >= 2 )
      FD_COUT << "fgRow[" << PixelPattern::fgPattern[id-1][i-1] << "]= " 
              << PixelPattern::fgRow[PixelPattern::fgPattern[id-1][i-1] - 1] 
	      << endl;
    
    if ( prow % 2 ) { // odd row number
      if ( fgPrintLevel >= 2 )
        FD_COUT << "fgColOdd[" << PixelPattern::fgPattern[id-1][i-1] << "]= "
                << PixelPattern::fgColOdd[PixelPattern::fgPattern[id-1][i-1] - 1] 
	        << endl;

      pcol += PixelPattern::fgColOdd[PixelPattern::fgPattern[id-1][i-1] - 1];
    }
    else {            // even row number
     if ( fgPrintLevel >= 2 )
        FD_COUT << "fgColEven[" << PixelPattern::fgPattern[id-1][i-1] << "]= "
                << PixelPattern::fgColEven[PixelPattern::fgPattern[id-1][i-1] - 1] 
	        << endl;

      pcol += PixelPattern::fgColEven[PixelPattern::fgPattern[id-1][i-1] - 1];
    }

    prow += PixelPattern::fgRow[PixelPattern::fgPattern[id-1][i-1] - 1];
    
    // col, row are starting from 1
    
    pattern->AddPixel( pcol, prow );
  }
}

// -------------------------------------------------------------------------

void TAnalysedPixelData::AssignMultiPatternAK(PixelPattern *pattern)
 {
  if ( pattern->GetPatternId() != 0x7f ) return;
  
  int start_bin = pattern->GetStartBin();
  int start_col = pattern->GetStartColumn();
  
  if ( fgPrintLevel ) {
    FD_COUT << "AssignMultiPatternAK(): bin= " <<  start_bin
            << ", col= " << start_col << endl;
  }
  
  TMirrorBitArray ba( this, start_bin, start_bin + 1 );
  
  int first_row;
  
  // find the first row which has a set bit, could be possible start pixel

  for ( first_row=1; first_row<=22; first_row++ ) {
    if ( ba.Test( start_col, first_row ) ) break;
  }
  
  static const int kNList = 4;
  int col_list[kNList];
  int row_list[kNList];

  // Attention: The indices of this method start from 0 !
  
  int n_pixel = fSltPattern.findPixelList( 
                    fRawPixelData /* uint32_t *mem */,
                    Fd::kMI_NSLT_BINS * Fd::kMI_NCOLS /* int n */,
		    start_bin * Fd::kMI_NCOLS + start_col - 1 /* int iPattern */, 
                    pattern->GetPatternId() /* int patternId */,
                    first_row - 1 /* int firstRow */, 
		    col_list /* int *colList */,
		    row_list /* int *rowList */,
		    kNList /* int nList = 4 */ );
  
  // should we try to find more pixel in the direction of an initial pattern !?

  if ( n_pixel > 0 ) {
  
    int np = MIN(kNList,n_pixel);
  
    for ( int i=0; i<np; i++ ) {
      col_list[i] += 1;
      row_list[i] += 1;
    }
    
    if ( fgPrintLevel ) {
      FD_COUT << "AssignMultiAK() - found " << n_pixel << " pixel(s) !" << endl;
    }
    
    for ( int i=0; i<np; i++ ) {
      if ( fgPrintLevel ) {
        FD_COUT << "Pixel[" << i << "] (col,row) = (" 
                << col_list[i] << "," << row_list[i] << ")" << endl;
      }
      
      pattern->AddPixel( col_list[i], row_list[i] );
    }
  }
}

// -------------------------------------------------------------------------

//void TAnalysedPixelData::AssignPixelsAK(PixelPattern* pattern)
// {
//  // todo: write interface to SltPattern::getPixelList()
//}

// -------------------------------------------------------------------------

void TAnalysedPixelData::FindPattern()
 {
  bool pattern_found[Fd::kMI_NCOLS];
  bool leave = false;
  
  for ( int i=0; i<Fd::kMI_NCOLS; i++ )
    pattern_found[i] = kFALSE;
  
  PixelDataWord pixelword;
  
  for ( unsigned int col=Fd::kMI_FIRST_COL; 
                     col<=Fd::kMI_LAST_COL && !leave; col++ ) {

    for ( int bin=fFirstSltBin; bin<fLastSltBin+1 && !leave; bin++ ) {
    
      pixelword = TMirrorPixelData::GetPixelData( col, bin );
      
      if ( pixelword != NULL ) {
      
#ifdef MIEVENT_VERSION_CODE
# if (MIEVENT_VERSION_CODE >= ModuleVersionCode(4,0,7))
        if ( PixelDataGetPatternClass( pixelword ) ) { // valid pattern found
# else
        if ( PixelDataGetClass( pixelword ) ) { // valid pattern found
# endif // MIEVENT_VERSION_CODE
#else
        if ( PixelDataGetClass( pixelword ) ) { // valid pattern found
#endif // MIEVENT_VERSION_CODE
	  
	  pattern_found[col-Fd::kMI_FIRST_COL] = true;

          if ( fgPrintLevel >= 2 ) {
	    FD_COUT << "col,bin: " << col << "," << bin << ": 0x" 
#ifdef MIEVENT_VERSION_CODE
# if (MIEVENT_VERSION_CODE >= ModuleVersionCode(4,0,7))
                 << hex << PixelDataGetPatternClass( pixelword ) << dec << endl;
# else
                 << hex << PixelDataGetClass( pixelword ) << dec << endl;
# endif // MIEVENT_VERSION_CODE
#else
                 << hex << PixelDataGetClass( pixelword ) << dec << endl;
#endif // MIEVENT_VERSION_CODE
          }
	  
          // and add the pattern to the list ...
	  
#ifdef MIEVENT_VERSION_CODE
# if (MIEVENT_VERSION_CODE >= ModuleVersionCode(4,0,7))
	  PixelPattern* pattern = new PixelPattern( PixelDataGetPatternClass( pixelword ) );
# else
	  PixelPattern* pattern = new PixelPattern( PixelDataGetClass( pixelword ) );
# endif // MIEVENT_VERSION_CODE
#else
	  PixelPattern* pattern = new PixelPattern( PixelDataGetClass( pixelword ) );
#endif // MIEVENT_VERSION_CODE
	  
	  pattern->SetStart( col, bin );
	  // GetTime( unsigned int i ) { return fTime[i]; }
	  
	  AddPattern( pattern );
	  
	  if ( GetNPattern() >= fNMaxPattern ) leave = true;
	  
          break;           // leave loop over time bins
	}
      }
    }
  }
}

// -------------------------------------------------------------------------

PixelPattern* TAnalysedPixelData::GetValidPattern(unsigned int n)
 {
  if ( !fValidPattern.size() ) return NULL;
  if ( n >= fValidPattern.size() ) return NULL;
  
  return fValidPattern[n];
}

// -------------------------------------------------------------------------

void TAnalysedPixelData::InvalidatePattern(unsigned int n)
 {
  PixelPattern* p;
  
  if ( ( p = GetValidPattern(n) ) != NULL ) {
    fValidPattern[n] = NULL;
    delete p;
  }
}

// -------------------------------------------------------------------------

void TAnalysedPixelData::PrintSettings()
 {
  FD_COUT << "*** TAnalysedPixelData - your present settings:" << endl;
  
  FD_COUT << "fFirstSltBin:  " << fFirstSltBin << endl;
  FD_COUT << "fLastSltBin:   " << fLastSltBin << endl;
  FD_COUT << "fNMaxPattern:  " << fNMaxPattern << endl;
  
  FD_COUT << endl;
}

// -------------------------------------------------------------------------
// -------------------------------------------------------------------------

