
//
// File   : FEFileReadout.cc
//
// Purpose: Implementation of the class FEFileReadout
//
// $Id: FEFileReadout.cc 9725 2010-09-28 07:31:17Z mathes $
//


#include <cstdio>
#include <cstdlib>
#include <iostream>

#include <sys/time.h>   // #define __USE_BSD  // macros timeradd(), timercmp() etc.
#include <unistd.h>
#include <assert.h>


/** @file FEFileReadout.hh
  * Implementation of the class FEFileReadout.
  * @author H.-J. Mathes, FzK
  */


#include <FdNumbering.hh>
#include <TimeConvert.hh>
#include <FdStdioStream.hh>
#include <ErrorLogging.hh>
#if (FDUTIL_VERSION_CODE >= ModuleVersionCode(4,1,0))
# include <InifileChain.hh>
#else
# include <ChainedInifile.hh>
#endif // FDUTIL_VERSION_CODE

#include <MiEvent.hh>
#include <MiEventFile.hh>
#include <MiRun.hh>

#include "CheckVersion.hh"

#ifdef MEASURE_TIME
# include "FEReadoutTimeProfile.hh"
#endif

#include "ReadoutConfig.hh"
#include "FEPageDesc.hh"
#include "ExtendedPixelList.hh"

#include "FEFileReadout.hh"

using namespace std;
using FdUtil::Fd;
using FdUtil::TimeConvert;

// --------------------------------------------------------------------------
// --------------------------------------------------------------------------
//
// class FEFileReadout - implementation
//

int FEFileReadout::fgPrintLevel = 0;

FEFileReadout::FEFileReadout(const char * /* ini_file */,FILE* log_file)
 : fEndOfFile(false), fEvent(new TMirrorEvent), 
   fEventFile(NULL), fEventLoaded(false), fEventPtr(NULL),
   fLogfile(log_file), 
   fRunHeader(NULL), fRunTrailer(NULL),
   fMode(kFromFile)
   //fFADCResolution(TFADCData::kFADC_DEFAULT_RESOLUTION), fNFADCBins(0)
 {
  VirtualReadout::fColumnList = 0x1fffff;  // all modules present
  
  // -- set the logging stream
  
  if ( !fLogfile ) fLogfile = stdout;
  
  fprintf( fLogfile, "Class FEFileReadout sets fColumnList= 0x%x\n", 
                     (unsigned int)fColumnList );

  // --- read the initial parameters from ini-file etc.
  
  Init();
}

// --------------------------------------------------------------------------

FEFileReadout::FEFileReadout(TMirrorEvent **event,
                             const char * /* ini_file */,FILE * log_file)
 : fEndOfFile(false), fEvent(NULL), 
   fEventFile(NULL), fEventLoaded(false), fEventPtr(event), 
   fLogfile(log_file), fMode(kFromEvent)
   //fFADCResolution(TFADCData::kFADC_DEFAULT_RESOLUTION), fNFADCBins(0)
 {
  assert( event != NULL );

  VirtualReadout::fColumnList = 0x1fffff;
  
  // -- set the logging stream
  
  if ( !fLogfile ) fLogfile = stdout;
  
  fprintf( fLogfile, "Class FEFileReadout sets fColumnList= 0x%x\n", 
                     (unsigned int)fColumnList );

  // --- read the initial parameters from ini-file etc.
  
  Init();
}

// --------------------------------------------------------------------------

FEFileReadout::~FEFileReadout()
 {
  if ( fMode == kFromFile ) {
    delete fEventFile;
    delete fEvent;
  }
}

// --------------------------------------------------------------------------

int FEFileReadout::AcquisitionInit()
 {
  if ( fgPrintLevel > 2 ) {
    FD_COUT << "FEFileReadout::AcquisitionInit()" << endl;
  }
  
  // already done in ctor !
#if 0
  // --- read the initial parameters from ini-file etc.
  
  Init();
#endif

  // --- print the parameters
  
  PrintSettings( FD_LOGSTREAM( FdUtil::ErrorLogger::eInfo ) );
  
  return 0;
}

// --------------------------------------------------------------------------

int FEFileReadout::AcquisitionStart()
 {
  if ( fgPrintLevel > 2 ) {
    FD_COUT << "FEFileReadout::AcquisitionStart()" << endl;
  }
  
  // --- re-read initial parameters (-> might be modified by shifters)
  
  Init();
  
  // --- print the parameters
  
  PrintSettings( FD_LOGSTREAM( FdUtil::ErrorLogger::eInfo ) );

  if ( fMode == kFromFile ) {
    
    // could be a symlink to a real data file
  
    FD_CINFO << "FEFileReadout::AcquisitionStart() - "
             << "trying to open " << fDataFilename << " for reading..."
	     << endl;
	     
    fEventFile = new TMirrorEventFile( fDataFilename.c_str() );

    if ( fEventFile ) {

      FD_CWARN << "FEFileReadout::AcquisitionStart() - "
               << "data file " << fDataFilename << " opened successfuly !"
	       << endl;

      fprintf( fLogfile, "TMirrorEventFile opened successfuly !\n" );
      
      // --- no try to read run header and trailer ...

      if ( fEventFile->ReadRunHeader( &fRunHeader ) != kTRUE ) {
        FD_CWARN << "FEFileReadout::AcquisitionStart() - "
                 << "Could not find run header in file !" << endl;
      }

      if ( fEventFile->ReadRunTrailer( &fRunTrailer ) != kTRUE ) {
        FD_CWARN << "FEFileReadout::AcquisitionStart() - "
                 << "Could not find run trailer in file !" << endl;
      }
    }
  
    fEndOfFile = false;
    fEventLoaded = false;
  }
  else {
    fEndOfFile = false;
    fEventLoaded = true;
  }
  
  return 0;
}

// --------------------------------------------------------------------------

int FEFileReadout::AcquisitionStop()
 {
  if ( fgPrintLevel > 2 ) {
    FD_COUT << "FEFileReadout::AcquisitionStop()" << endl;
  }
  
  if ( fMode == kFromFile ) {
  
    delete fEventFile; fEventFile = NULL;
    delete fRunHeader; fRunHeader = NULL;
    delete fRunTrailer; fRunTrailer = NULL;
  }
  
  return 0;
}

// --------------------------------------------------------------------------

void FEFileReadout::DeleteEvent(FEPageDesc * /* page_desc */)
 {
#ifdef MEASURE_TIME
  FEReadoutTimeProfile::GetInstance()->fDeleteEvent.Start(kFALSE);
#endif

  if ( fMode == kFromFile ) {
    fEventLoaded = false;
  }
  
#ifdef MEASURE_TIME
  FEReadoutTimeProfile::GetInstance()->fDeleteEvent.Stop();
#endif
}

// --------------------------------------------------------------------------

void FEFileReadout::FillRunHeader(TMirrorRunHeader* header)
 {
  if ( fEventFile ) {
    
    if ( fRunHeader != NULL ) {
    
      int option;
      fRunHeader->GetOption( TMirrorRunHeader::RUNH_OPT_FDHWLIB_VERSION, option );
      header->SetOption( TMirrorRunHeader::RUNH_OPT_FDHWLIB_VERSION, option );
      
      fRunHeader->GetOption( TMirrorRunHeader::RUNH_OPT_FLT_INTEGRATION_TIME, option );
      header->SetOption( TMirrorRunHeader::RUNH_OPT_FLT_INTEGRATION_TIME, option );

      fRunHeader->GetOption( TMirrorRunHeader::RUNH_OPT_FLT_STATISTIC_SAMPLES, option );
      header->SetOption( TMirrorRunHeader::RUNH_OPT_FLT_STATISTIC_SAMPLES, option );
      
      fRunHeader->GetOption( TMirrorRunHeader::RUNH_OPT_FLT_STATISTIC_OFFSET, option );
      header->SetOption( TMirrorRunHeader::RUNH_OPT_FLT_STATISTIC_OFFSET, option );
    }
  }
  else
    FD_CERR << "FEFileReadout::FillRunHeader() - "
            << "No file open !" << endl;
}

// --------------------------------------------------------------------------

void FEFileReadout::FillRunTrailer(TMirrorRunTrailer* trailer)
 {
  if ( fRunTrailer != NULL ) {

    *trailer = *fRunTrailer;
  }
}

// --------------------------------------------------------------------------

TMirrorEventHeader *
FEFileReadout::GetEventHeader(FEPageDesc *page_desc, 
                              TMirrorEventHeader* header)
 {
  TMirrorEventHeader * local_header = header;
  
#ifdef MEASURE_TIME
  FEReadoutTimeProfile::GetInstance()->fEventHeader.Start(kFALSE);
#endif

  struct timezone tz;
  struct timeval tv;
  
  gettimeofday( &tv, &tz );

  if ( fMode == kFromEvent ) {
    fEvent = *fEventPtr;
    assert( fEvent != NULL );
  }
  
  if ( !local_header )
    local_header = fEvent->GetEventHeader();
  else
    *local_header = *fEvent->GetEventHeader();
  
  local_header->SetNextPageTime( page_desc->fSecond );
  local_header->SetNextPageNanoTime( page_desc->fNanoSecond );
  
#ifdef MEASURE_TIME
  FEReadoutTimeProfile::GetInstance()->fEventHeader.Stop();
#endif

  return local_header;
}

// --------------------------------------------------------------------------

TFADCData *FEFileReadout::GetFADCData(FEPageDesc * /* page_desc */, 
                                      TFADCData *fadc)
 {
  assert( fadc != NULL );
  
  if ( fMode == kFromEvent ) {
    fEvent = *fEventPtr;
    assert( fEvent != NULL );
  }
  
  TFADCData *fadc_buffer = 
       fEvent->GetFADCData()->GetPixelFADCData( fadc->GetPixelNumber() );

  // --- copy the data and delete the buffer afterwards
  
  if ( fadc_buffer ) {
    
    TFADCData::EFADCTraceResolution res = fadc->GetTraceResolution();
    
    while ( fgPrintLevel ) {
      static bool printed = false;
      
      if ( printed ) break;
      
      FD_CDEBUG(1) << "FEFileReadout::GetFADCData() - "
                   << "default resolution: " << res 
                   << " " << TFADCData::GetNFADCBins( res ) 
	           << " bins" << endl;
      FD_CDEBUG(1) << "FEFileReadout::GetFADCData() - "
                   << "resolution in file: " << fadc_buffer->GetTraceResolution() 
                   << " " << TFADCData::GetNFADCBins( fadc_buffer->GetTraceResolution() ) 
	           << " bins" << endl;
      
      printed = true;
    }
    
    // set FADC resolution
    if ( res == TFADCData::kFADC_RESOLUTION_4000_BINS ) {
      
      FD_CWARN << "FEFileReadout::GetFADCData() - "
               << "FADC resolution of 4000 is not supported!" << endl;
      
      res = TFADCData::kFADC_RESOLUTION_2000_BINS;
      
      fadc->SetTraceResolution( res );
    }
    else
      fadc->SetTraceResolution( res );
    
    if ( res == fadc_buffer->GetTraceResolution() ) {
      *fadc = *fadc_buffer;
    }
    else {
      if ( res == TFADCData::kFADC_RESOLUTION_2000_BINS ) {
        TFADCData::ConvertToV4( fadc_buffer, fadc );
      }
      else {
#if (MIEVENT_VERSION_CODE >= ModuleVersionCode(4,0,4))
        TFADCData::ConvertToV3( fadc_buffer, fadc );
#else
        FD_CWARN << "FEFileReadout::GetFADCData() - "
	         << "cannot convert to V3, only half trace copied!" << endl;
      
        *fadc = *fadc_buffer;
	
	fadc->SetTraceEndBin( Fd::kMI_NFADC_BINS_LOW - 1 );
	fadc->SetTraceResolution( TFADCData::kFADC_RESOLUTION_1000_BINS );
#endif // MIEVENT_VERSION_CODE
      }
    }

    delete fadc_buffer;
  }
  else
    FD_CERR << "FEFileReadout::GetFADCData() - "
            << "non-existing pixel requested!" << endl;
  
  return fadc;
}

// --------------------------------------------------------------------------

TFADCData* FEFileReadout::GetFADCData(FEPageDesc *page_desc,
                                      FdUtil::Fd::PixelNumber pixel,
				      TFADCData* fadc)
 {
  TFADCData::FADCTraceInfoRec info;
  
  // make sure that the FADCTraceInfo is defined
  
  if ( fadc == NULL ) {
  
    // Note: It is assumed that TFADCData::fgDefaultResolution has been set
    //       properly before, which is done in the ctor of this class
    unsigned int n_fadc_bins = TFADCData::GetNFADCBins( TFADCData::GetDefaultResolution() );

    FADCTraceSetStartBin( &info, 0 );
    FADCTraceSetEndBin( &info, n_fadc_bins - 1 );
    FADCTraceSetResolution( &info, TFADCData::GetDefaultResolution() );
  }
  else
    info = *(fadc->GetTraceInfo());
  
  return GetFADCData( page_desc, pixel, &info, fadc );
}

// --------------------------------------------------------------------------

TFADCData* FEFileReadout::GetFADCData(FEPageDesc *page_desc,
                                      FdUtil::Fd::PixelNumber pixel,
		                      TFADCData::FADCTraceInfo info,
				      TFADCData *fadc)
 {
  if ( fadc != NULL ) {

    fadc->SetPixelNumber( pixel );
    fadc->SetTraceInfo( info );

    return GetFADCData( page_desc, fadc );
  }
  else {
    TFADCData * local_fadc = new TFADCData();

    if ( local_fadc ) {
    
      local_fadc->SetPixelNumber( pixel );
      local_fadc->SetTraceInfo( info );
      
      if ( GetFADCData( page_desc, local_fadc ) == NULL ) {

        FD_CERR << "FEFileReadout::GetFADCData() - "
                << "non-existing pixel requested!" << endl;

        delete local_fadc;
	local_fadc = NULL;
      }
    }
    
    return local_fadc;  // needs to be deleted afterwards
  }
}

// --------------------------------------------------------------------------

#if 0
void FEFileReadout::GetFADCData(FEPageDesc * /* page_desc */,
                                TMirrorPixelList* pixel_list,
                                TMirrorFADCData* fadc)
 {
//  FD_CWARN << "FEFileReadout::GetFADCData(FEPageDesc*,TMirrorPixelList*,TMirrorFADCData*) -"
//           << "is obsolete !" << endl;

#ifdef MEASURE_TIME
  FEReadoutTimeProfile::GetInstance()->fFADCReadout.Start(kFALSE);
#endif

  for ( unsigned int n_pixel=0; n_pixel<pixel_list->GetNumPixels(); n_pixel++ ) {
    
    Fd::PixelNumber pixel = pixel_list->GetPixel( n_pixel );

    TFADCData *src = fEvent->GetFADCData()->GetPixelFADCData( pixel );
    TFADCData *dest = fadc->GetPixelFADCData( pixel );
    
#ifdef MEASURE_TIME
    FEReadoutTimeProfile::GetInstance()->fTraceReadout.Start(kFALSE);
#endif

    assert( src != NULL );
    assert( dest != NULL );
    
    *dest = *src;
    
    ExtendedPixelList *ext_pixel_list 
               = dynamic_cast<ExtendedPixelList*>( pixel_list );
   
    if ( ext_pixel_list != NULL ) {    
      dest->SetTraceStartBin( ext_pixel_list->GetStartBin( n_pixel ) );
      dest->SetTraceEndBin( ext_pixel_list->GetEndBin( n_pixel ) );
    }

#ifdef MEASURE_TIME
      FEReadoutTimeProfile::GetInstance()->fTraceReadout.Stop();
#endif
    
    delete src;
    delete dest;
  }
  
#ifdef MEASURE_TIME
  FEReadoutTimeProfile::GetInstance()->fFADCReadout.Stop();
#endif
}
#endif

// --------------------------------------------------------------------------

TFADCData::PixelMonitorData 
FEFileReadout::GetMonitorData(FEPageDesc * /* page_desc */,
                              Fd::PixelNumber pixel, 
			      TFADCData::PixelMonitorData data)
 {
  TFADCData::PixelMonitorData local_data = data;
  
  if ( fMode == kFromEvent ) {
    fEvent = *fEventPtr;
    assert( fEvent != NULL );
  }
  
#ifdef MEASURE_TIME
  FEReadoutTimeProfile::GetInstance()->fMonitorReadout.Start(kFALSE);
#endif

  if ( !local_data )
    local_data = fEvent->GetFADCData()->GetMonitorData( pixel );
  else
    *local_data = *fEvent->GetFADCData()->GetMonitorData( pixel );

#ifdef MEASURE_TIME
  FEReadoutTimeProfile::GetInstance()->fMonitorReadout.Stop();
#endif

  return local_data;
}

// --------------------------------------------------------------------------

TMirrorPixelData *
FEFileReadout::GetPixelData(FEPageDesc * /* page_desc */, 
                            TMirrorPixelData *pixel_data)
 {
  TMirrorPixelData *local_pixel_data = pixel_data;
  
#ifdef MEASURE_TIME
  FEReadoutTimeProfile::GetInstance()->fPixelReadout.Start(kFALSE);
#endif
  
  if ( fMode == kFromEvent ) {
    fEvent = *fEventPtr;
    assert( fEvent != NULL );
  }
  
  if ( !local_pixel_data )
    local_pixel_data = fEvent->GetPixelData();
  else
    *local_pixel_data = *fEvent->GetPixelData();
  
  // --- temporary: clear the Ext Trigger bit

#warning Ext Trigger bit cleared here !
  
  TMirrorPixelData::PixelDataWord pixel_data_ptr = 
                                            local_pixel_data->GetPixelData();
  
  for ( unsigned int i=0; i<FdUtil::Fd::kMI_NSLT_BINS*FdUtil::Fd::kMI_NCOLS; i++ )
    PixelDataSetTrigger( &pixel_data_ptr[i], 0 );

#ifdef MEASURE_TIME
  FEReadoutTimeProfile::GetInstance()->fPixelReadout.Stop();
#endif
  
  return local_pixel_data;
}

// --------------------------------------------------------------------------

void FEFileReadout::Init()
 {
  using FdUtil::Inifile;
  
  // --- initial settings ...
  
  fDataFilename = "data.root";
  fDoLoop       = false;
  fReadDelay    = 0;
  fReadRandom   = false;
  fFileFormat   = kMirrorEvent;
  
  // --- read from ini-file
  
  FdUtil::Inifile *inifile;
  
  inifile = ReadoutConfig::GetInstance()->GetInifile();
  
  if ( inifile != NULL ) {
  
    const char *section = "FEFileReadout";
    
    if ( inifile->SpecifyGroup( section ) != Inifile::kSUCCESS ) {
      FD_CERR << "FEFileReadout::Init() - "
              << "section [" << section << "] not found !!!" << endl;
    }
    else {
      FD_CINFO << "\tReading section [" << section << "] ..." << endl;

      fDataFilename = inifile->GetFirstString( "name", "data.root" );
      fDoLoop = (inifile->GetFirstValue( "loop", 0 ) != 0);
      fReadDelay = inifile->GetFirstValue( "delay", 0 );
      fReadRandom = (inifile->GetFirstValue( "random", 0 ) != 0);
      
      string file_format = inifile->GetFirstString( "format", "mirror" );
      
      if ( file_format.find( "mirror" ) != string::npos ) {
        fFileFormat = kMirrorEvent;
      }
      else if ( file_format.find( "eye" ) != string::npos ) {
        //fFileFormat = kEyeEvent;
        FD_CWARN << "FEFileReadout::Init() - "
	         << "[FEFileReadout] format=eye is not supported, yet!"
		 << endl;
        FD_CWARN << "                        using format=mirror instead !"
		 << endl;
      }
      else {
        FD_CWARN << "FEFileReadout::Init() - "
	         << "[FEFileReadout] format=" << file_format << " is unknown!"
		 << endl;
        FD_CWARN << "                        using format=mirror instead !"
		 << endl;
      }
      
      int error;
      int printlevel = inifile->GetFirstValue( "printlevel", 0, &error );
  
      if ( !error ) SetPrintLevel( printlevel );
    }
  }
  else
    FD_CERR << "FEFileReadout::Init() - no Inifile handle!" << endl;

  // --- initialize (primitive) random generator
  
  srand(190463);  
}

// --------------------------------------------------------------------------

void FEFileReadout::PrintSettings(std::ostream& ostr)
 {
  ostr << endl;
  ostr << "=== FEFileReadout - settings === " << endl;
  ostr << endl;
  
  ostr << "\tfgPrintLevel=  " << fgPrintLevel << endl;
  
  ostr << "\tfDataFilename= " << fDataFilename << endl;
  ostr << "\tfFileFormat=   " << ((fFileFormat == kEyeEvent) ? "eye" : "mirror") << endl;
  ostr << "\tfDoLoop=       " << (fDoLoop ? "true" : "false") << endl;
  ostr << "\tfReadDelay=    " << fReadDelay << " msec" << endl;
  ostr << "\tfReadRandom=   " << (fReadRandom ? "true" : "false") << endl;
  
  ostr << endl;
  ostr << "==========================================" << endl;
}

// --------------------------------------------------------------------------

FEPageDesc* FEFileReadout::WaitForNextPage()
 {
#warning Different behaviour for fMode=kFromEvent must be implemented !
  if ( fEventFile == NULL ) {
    FD_CERR << "FEFileReadout::WaitForNextPage() - no TMirrorEventFile !" << endl;
    return NULL;
  }
  
  static int page_id = 1;
  static struct timeval tv_last = { 0, 0 };
  static struct timeval tv_delay = { fReadDelay/1000, (fReadDelay * 1000) % 1000000 };
  struct timezone tz;
  struct timeval tv_limit, tv_now;
    
  FEPageDesc *page_desc = NULL;
  
  if ( !fEventLoaded && !fEndOfFile ) {
    
    if ( fReadDelay != 0 ) {  // only when set
    
      if ( tv_last.tv_sec == 0 && tv_last.tv_usec == 0 ) {
        gettimeofday( &tv_last, &tz );
      }
      else {
        gettimeofday( &tv_now, &tz );
      
        timeradd( &tv_last, &tv_delay, &tv_limit );
	
        if ( timercmp( &tv_now, &tv_limit, < ) ) {
          return NULL;
	}
	
	tv_last = tv_now;
      }
    }
    
    if ( fReadRandom ) {
      unsigned int pos = rand() % fEventFile->GetNEvents();
      
      fEventFile->Seek( pos );
      fEventLoaded = fEventFile->ReadMirrorEvent( &fEvent );
    }
    else
      fEventLoaded = fEventFile->ReadMirrorEvent( &fEvent );
    
    if ( fEventLoaded ) {
    
      TMirrorEventHeader * header = fEvent->GetEventHeader();
      
      FD_CINFO << "FEFileReadout::WaitForNextPage() - loaded event # " 
               << header->GetEventNo() << " "
	       << TMirrorEventHeader::GetVerboseEventType(header->GetEventType())
	       << " " 
	       << TMirrorEventHeader::GetVerboseEventLabel(header->GetEventLabel())
	       << " src= 0x" << hex << header->GetTriggerSource() << dec
	       << endl;
	   
      page_desc = new FEPageDesc( page_id );
      
      page_id++;

      // this is just a simulation, doesn't matter really !

      if ( page_id > 64 /* FE_PAGES */ ) page_id = 1;
      
      // need the current time ...
      
      gettimeofday( &tv_now, &tz );
      
      page_desc->fSecond     = TimeConvert::GetGPSTime();
      page_desc->fNanoSecond = tv_now.tv_usec * 1000;
      
      page_desc->fPageId     = page_id;
      
      // TimeStampHigh are the seconds
      page_desc->fNxPgTimeStampHigh = header->GetTimeStampHigh();
      // TimeStampLow is a composite binary value
      page_desc->fNxPgTimeStampLow  = header->GetTimeStampLow();
      
      page_desc->fTriggerSource     = header->GetTriggerSource();
  
      page_desc->fTimeStamp =   (long long)page_desc->fSecond * 1000000
			      + page_desc->fNanoSecond / 1000;
      
      page_desc->fDeadTimeCounter = header->GetDeadTimeCounter();
      page_desc->fVetoTimeCounter = header->GetVetoTimeCounter();
      
      // --- correction for some very old data (with possible hw defect)
      
      if ( (int)page_desc->fTriggerSource == 0 ) {
        page_desc->fTriggerSource = TMirrorEventHeader::kUnknownTrigger;
	FD_CERR << "FEFileReadout::WaitForNextPage() - trigger src corrected !" 
	        << endl;
      }
    }
    else {
      // end-of-file encountered
      if ( fDoLoop ) {
        fEventFile->Rewind();
	
	FD_CWARN << "FEFileReadout::WaitForNextPage() - "
	         << "end of file encountered, rewinding..." << endl; 
      }
      else
      
      fEndOfFile = true;
    }
  }
  
  return page_desc;
}

// --------------------------------------------------------------------------
// --------------------------------------------------------------------------
