
//
// File   : MStcCutTlt.cc
// 
// Purpose: Implementation of the class MStcCutTlt
//
// $Id: MStcCutTlt.cc 11881 2011-01-13 09:00:27Z mathes $
//


#include <iostream>
#include <iomanip>

#include <cstdio>
#include <cstdlib>
#include <ctime>


/** @file MStcCutTlt.cc
  * Implementation of the class MStcCutTlt.hh
  * @author Adrian Schmidt, FzK, IPE
  */
  

#include <FdNumbering.hh>
#include <FdStdioStream.hh>
#include <ErrorLogging.hh>
#include <Inifile.hh>
#include <MiEvent.hh>

#include <PixelPattern.hh>

#include "CheckVersion.hh"

#include "StcShowerCore.hh"

#include "MStcCutTlt.hh"

using namespace std;
using FdUtil::Fd;

// --------------------------------------------------------------------------
// --------------------------------------------------------------------------
//
// class  MStcCutTlt - implementation
//

MStcCutTlt::MStcCutTlt(MiReadout::MStcCutTltParameters tltpar)
 : VTltProcessor("MStcCutTlt"), 
   fPixelData(NULL), 
   fPixelMatrix(new TMirrorBitArray), 
   fReadout(NULL),
   fFillNtuple(false)
 {
  // set default print level (VTltProcessor, standard = 0)
  //VTltProcessor::fgPrintLevel = 1;
    
  if ( fgPrintLevel > 1 )
    FD_CINFO << "MStcCutTlt::MStcCutTlt() -" << endl;

  srand(1);
  
  SetTltType( VTltProcessor::kMStc );
  
  MStcCutTlt::Clear();
  
  fNEventsAnalysed  = 0;
  fNEventsRejected  = 0;
  fNRejectedKept    = 0;
  
  fNEarlyPixelCut    = 0;
  fNWeakLightning1   = 0;
  fNWeakLightning2   = 0;
  fNStrongLightning1 = 0;
  fNStrongLightning2 = 0;
  fNSTC              = 0;
  
#ifdef MSTC_TIME_OUTPUT
  fNDecision         = 0;
  fSumTime           = 0;
  fNDecisionManyP    = 0;
  fSumTimeManyP      = 0;
#endif // MSTC_TIME_OUTPUT

  // --- check if parameters have been passed in !!!
  
  if ( tltpar )
    SetParameters( tltpar );
  else {
    FD_CWARN << "MStcCutTlt::MStcCutTlt() - pointer to TLT-Parameters is NULL, "
             << "using default values!" 
	     << endl;
    
    MiReadout::MStcCutTltParametersRec default_tltpar = MSTC_TLT_DEFAULT_PARAMETERS;
    
    SetParameters( &default_tltpar );
  }
     
  if ( fTltLogfile ) {
    FdUtil::OstdioStream ostr( fTltLogfile );
    PrintSettings( ostr);
  }
     
  if ( fgPrintLevel ) PrintSettings( FD_LOGSTREAM(FdUtil::ErrorLogger::eInfo) );
}
 
// ------------------------------------------------------------------------
 
MStcCutTlt::~MStcCutTlt()
 {
  if ( fgPrintLevel > 1 )
    FD_CINFO << "MStcCutTlt::~MStcCutTlt() -" << endl;

  delete fPixelMatrix;
  
  if ( fTltLogfile ) {
    FdUtil::OstdioStream ostr( fTltLogfile );

    PrintFinalResults( ostr );
  }
  
  if ( fgPrintLevel ) {
    PrintFinalResults( FD_LOGSTREAM(FdUtil::ErrorLogger::eInfo) );
  }
}

// ------------------------------------------------------------------------

void MStcCutTlt::AnalyseMultiplicity(
                      TMirrorPixelData::MultiplicityDataWord multiplicityArray)
 {
  if ( fgPrintLevel > 1 )
    FD_CINFO << "MStcCutTlt::AnalyseMultiplicity() -" << endl;
   
  fResult.fNPix   = fPixelData->GetNumberOfTriggeredPixels(0, FdUtil::Fd::kMI_NSLT_BINS);
  fResult.fIntAll = multiplicityArray[0];
  fResult.fNGap   = 1;
  fResult.fNStep  = 0;
  // loop over multiplicity data
  for (int t = 1; t < FdUtil::Fd::kMI_NMULT_BINS; t++) { 
    fResult.fIntAll += multiplicityArray[t];
    
    fResult.fNStep += abs(multiplicityArray[t] - multiplicityArray[t-1]);
    
    if ( (multiplicityArray[t-1] == 0) && (multiplicityArray[t] != 0) ) 
      fResult.fNGap++;
  }

	 
  if ( fgPrintLevel > 2 )
    FD_CINFO << "M-Parameters: n_pix= " << fResult.fNPix 
             << " n_early= " << fResult.fNEarly 
	     << " int_all= " << fResult.fIntAll 
	     << " n_gap= " << fResult.fNGap 
	     << " n_step= " << fResult.fNStep << endl;

  return;
}

// ------------------------------------------------------------------------

void MStcCutTlt::Clear()
 {
  if ( fgPrintLevel > 1 )
    FD_CINFO << "MStcCutTlt::Clear() -" << endl;
    
  fReadout = NULL;
  
  VTltProcessor::SetEventAccepted( false );
  
  fPixelData = NULL;
  fPixelMatrix->Clear();
  
  VTltProcessor::SetEventLabel( TMirrorEventHeader::kUnlabelled );

  // --- initialize all flag variables
   
  fResult.fAccepted = true;
  
  fResult.fNPix   = 0;
  fResult.fNEarly = 0;
  fResult.fNGap   = 0;
  fResult.fNStep  = 0;
  fResult.fIntAll = 0;
  fResult.fStcPar = 0;
}

// ------------------------------------------------------------------------

void MStcCutTlt::LogResults(const char *where)
 {
  if ( !fTltLogfile ) return;  // logfile not open
  
  static int line_no = 0;

  // --- write symbol explanation ...
  
  if ( line_no == 0 ) {
    
    fprintf( fTltLogfile, 
             "========== MStcCutTlt: Results evt.by evt. ==========\n" );
    
    fprintf( fTltLogfile, "\tNE   - n_early-cut\n" );
    fprintf( fTltLogfile, "\tWL1  - weak lightning cut1\n" );
    fprintf( fTltLogfile, "\tWL2  - weak lightning cut2\n" );
    fprintf( fTltLogfile, "\tSL1  - strong lightning cut1\n" );
    fprintf( fTltLogfile, "\tSL2  - strong lightning cut2\n" );
    fprintf( fTltLogfile, "\tSTC  - Space-Time-Correlation-cut\n" );
    fprintf( fTltLogfile, "\tERR  - Error\n" );
    
    fprintf( fTltLogfile, 
             "................................................................\n" );
  }
  
  // --- write header line each 20 lines ...
  
  if ( (line_no % 20) == 0 )
    fprintf( fTltLogfile, 
             "STC -  acc  rej\n" );
  
  char where_str[5];
  snprintf( where_str, sizeof(where_str), "%4s", where );
  
  fprintf( fTltLogfile,
           "%s: %3d  %3d\n",
	   where_str, 
	   (int) fResult.fAccepted,
	   (int) !fResult.fAccepted);
  
  line_no++;
  
  fflush( fTltLogfile );
}

// ------------------------------------------------------------------------
 
void MStcCutTlt::Decide(VMirrorEvent* readout) 
 {
#ifdef MSTC_TIME_OUTPUT
  clock_t startclock = clock();
#endif // MSTC_TIME_OUTPUT

  if ( fgPrintLevel > 1 )
    FD_CINFO << "MStcCutTlt::Decide() - starting ..."<< endl;
    
  if ( !readout ) {
    FD_CERR << "MStcCutTlt::Decide() - "
            << "error passing parameters!" 
            << endl;
    LogResults("ERR");
    VTltProcessor::SetEventAccepted( false );
    return;
  }
  
  fNEventsAnalysed++;
   
  fReadout = readout;
 
  if ( ( fPixelData = fReadout->GetPixelData( ) ) == NULL ) 
   {
    FD_CERR << "MStcCutTlt::Decide() - no PixelData present !" << endl;
    LogResults("ERR");
    VTltProcessor::SetEventAccepted( false );
    return;
  }
  
  fPixelMatrix->Set( fPixelData );      // Do I need this ?
   
  // --- Perform Trigger cuts:
  
  ostringstream tlt_str;
  
  DoTEarlyCut();
  
  if ( fResult.fAccepted ) DoMultiplicityCut();
  
  if ( fResult.fAccepted && (fResult.fNPix <= fCutPar.fStcNPixMax) ) DoStcCut();
    
  VTltProcessor::SetEventAccepted( fResult.fAccepted );
  
  if ( fResult.fAccepted ) {
    VTltProcessor::SetEventLabel( TMirrorEventHeader::kShowerCandidate );
    
    if ( fgPrintLevel )
      tlt_str << "TLT says: Event is shower candidate!  --  ";
  }
    
  if ( !fResult.fAccepted ) {
  
    if ( fgPrintLevel )
      tlt_str << "TLT says: Event is background! --  ";
      
    fNEventsRejected++;
    
    if ( fCutPar.fPercentageRejKept > 0.0 ) {
      bool keep = false;	// keep a fixed percentage of rejected events for monitoring purposes
      if ( (100.0 * (double)rand()/(double)(RAND_MAX + 1.0)) <= fCutPar.fPercentageRejKept) keep = true;
      if ( keep ) {
    	VTltProcessor::SetEventAccepted( true );
    	fNRejectedKept++;
    	fNEventsRejected--;
	
    	if ( fgPrintLevel ) tlt_str << "will be kept! -- ";
      }
    }
  }
    
  if ( fgPrintLevel )
    switch ( VTltProcessor::GetEventLabel() ) {
      case TMirrorEventHeader::kUnlabelled:
           FD_CINFO << tlt_str.str() << "EventLabel is: 'kUnlabelled' " << endl;
	   break;
      case TMirrorEventHeader::kAcceptedEvent:
           FD_CINFO << tlt_str.str() << "EventLabel is: 'kAcceptedEvent' " << endl;
	   break;
      case TMirrorEventHeader::kMuonEvent:
           FD_CINFO << tlt_str.str() << "EventLabel is: 'kMuonEvent' " << endl;
	   break;
      case TMirrorEventHeader::kShowerCandidate:
           FD_CINFO << tlt_str.str() << "EventLabel is: 'kShowerCandidate' " << endl;
	   break;
      case TMirrorEventHeader::kShowerEvent:
           FD_CINFO << tlt_str.str() << "EventLabel is: 'kShowerEvent,' " << endl;
	   break;
      case TMirrorEventHeader::kRandomEvent:
           FD_CINFO << tlt_str.str() << "EventLabel is: 'kRandomEvent' " << endl;
	   break;
      case TMirrorEventHeader::kManyPixelEvent:
           FD_CINFO << tlt_str.str() << "EventLabel is: 'kManyPixelEvent' " << endl;
	   break;
      case TMirrorEventHeader::kRejected:
           FD_CINFO << tlt_str.str() << "EventLabel is: 'kRejected' " << endl;
	   break;
      default: 
           FD_CWARN << tlt_str.str() << "No valid EventLabel! " << endl;
    }
    
#ifdef MSTC_TIME_OUTPUT
  if ( fNPix > fCutPar.fStcNPixMax ) {
    fNDecisionManyP++;
    fSumTimeManyP += (clock() - startclock)/1000;
  }
  
  fNDecision++;
  fSumTime += (clock() - startclock)/1000;
#endif // MSTC_TIME_OUTPUT

  // --- deal with the optional Ntuple
  
  if ( fFillNtuple ) {
    fResultTuple->Fill();
  }
}

// ------------------------------------------------------------------------
 
void MStcCutTlt::DoTEarlyCut()
 {
  if ( fgPrintLevel > 1 )
    FD_CINFO << "MStcCutTlt::DoTEarlyCut() -" <<  endl;
     
  fResult.fNEarly = fPixelData->GetNumberOfTriggeredPixels(0, fCutPar.fTimeForEarly);
   
  if ( fResult.fNEarly >= fCutPar.fNEarlyCutValue ) {
    fNEarlyPixelCut++;
    fResult.fAccepted = false;
    LogResults("NE");
    VTltProcessor::SetEventLabel( TMirrorEventHeader::kRandomEvent );
    if ( fgPrintLevel > 2 )
      FD_CINFO << "  M-cuts:  N_early-cut applied: n_early = " << fResult.fNEarly 
               << " >= " << fCutPar.fNEarlyCutValue << endl;
  }
}

// ------------------------------------------------------------------------

void MStcCutTlt::DoMultiplicityCut()
 {
  if ( fgPrintLevel > 1 )
    FD_CINFO << "MStcCutTlt::DoMultiplicityCut() -" << endl;

  TMirrorPixelData::MultiplicityDataWord multiplicityArray;

#ifndef MSTC_CALC_MULTIPLICITY
  // --- check if multiplicity data are present
  if ( fgPrintLevel > 2 )
    FD_CINFO << "MStcCutTlt::DoMultiplicityCut() - Using Hardware Multiplicity Info " << endl;

  multiplicityArray = fPixelData->GetMultiplicityData();

# if (FDUTIL_VERSION_CODE >= ModuleVersionCode(4,0,0))
  unsigned int n_mult_bins = Fd::kMI_NMULT_BINS;
# else
  unsigned int n_mult_bins = Fd::kMI_NFADC_BINS;
# endif // FDUTIL_VERSION_CODE

  int mult_sum = 0;
  for ( unsigned int m=0; m<n_mult_bins; m++ )
    mult_sum += (int)multiplicityArray[m];
    
  if ( mult_sum == 0 ) {
    FD_CWARN << "MStcCutTlt::DoMultiplicityCut() - "
             << "there is no multiplicity data available, "
  	     << "the TLT will fail here!" << endl;
  }
#endif // MSTC_CALC_MULTIPLICITY
 
#ifdef MSTC_CALC_MULTIPLICITY   // Don't use Hardware Multiplicity  
  if ( fgPrintLevel > 2 )
    FD_CINFO << "MStcCutTlt::DoMultiplicityCut() - Using Software-calculated Multiplicity Info " << endl;

# if (FDUTIL_VERSION_CODE >= ModuleVersionCode(4,0,0))
  multiplicityArray = new TMirrorPixelData::MultiplicityDataWordRec[FdUtil::Fd::kMI_NMULT_BINS];
# else
  multiplicityArray = new TMirrorPixelData::MultiplicityDataWordRec[FdUtil::Fd::kMI_NFADC_BINS];
# endif // FDUTIL_VERSION_CODE
  CreateMultiplicity( multiplicityArray );
#else // MSTC_CALC_MULTIPLICITY
  multiplicityArray = fPixelData->GetMultiplicityData();
#endif // MSTC_CALC_MULTIPLICITY
         
  AnalyseMultiplicity( multiplicityArray );

#ifdef MSTC_CALC_MULTIPLICITY  // Don't use Hardware Multiplicity  
  delete [] multiplicityArray;
#endif  // MSTC_CALC_MULTIPLICITY
         
  if ((Int_t)fResult.fIntAll <= fCutPar.fWeakLightningCutSlope * (Int_t)fResult.fNPix + fCutPar.fWeakLightningCutOffset ) {
    fResult.fAccepted = false;
    LogResults("WL1");
    fNWeakLightning1++;
    VTltProcessor::SetEventLabel( TMirrorEventHeader::kRandomEvent );
    if ( fgPrintLevel > 2 )
      FD_CINFO << "  M-cuts:  WeakLightningCut1 applied: int_all = " << fResult.fIntAll 
    	       << " <= " << fCutPar.fWeakLightningCutSlope << " * " << fResult.fNPix 
               << " + " << fCutPar.fWeakLightningCutOffset << endl;
    return;
  }
  
  if ( fResult.fNGap >= fCutPar.fWeakLightningCutValue ) {
    fResult.fAccepted = false;
    LogResults("WL2");
    fNWeakLightning2++;
    VTltProcessor::SetEventLabel( TMirrorEventHeader::kRandomEvent );
    if ( fgPrintLevel > 2 )
      FD_CINFO << "  M-cuts:  WeakLightningCut2 applied: n_gap = "<< fResult.fNGap 
               << " >= "<< fCutPar.fWeakLightningCutValue << endl;
    return;
  }
  
  if ((Int_t) fResult.fNStep >= fCutPar.fStrongLightningCutSlope * (Int_t)fResult.fNPix + fCutPar.fStrongLightningCutOffset) {
    fResult.fAccepted = false;
    LogResults("SL1");
    fNStrongLightning1++;
    VTltProcessor::SetEventLabel( TMirrorEventHeader::kManyPixelEvent );
    if ( fgPrintLevel > 2 )
      FD_CINFO << "  M-cuts:  StrongLightningCut1 applied: n_step = " << fResult.fNStep
               << " >= " << fCutPar.fStrongLightningCutSlope << " * "<< fResult.fNPix 
	       << " + "<< fCutPar.fStrongLightningCutOffset << endl;
    return;
  }
	 
  if ( fResult.fIntAll >= fCutPar.fStrongLightningCutValue ) {
    fResult.fAccepted = false;
    LogResults("SL2");
    fNStrongLightning2++;
    VTltProcessor::SetEventLabel( TMirrorEventHeader::kManyPixelEvent );
    if ( fgPrintLevel > 2 )
      FD_CINFO << "  M-cuts:  StrongLightningCut2 applied: int_all = "
               << fResult.fIntAll 
               << " >= "<< fCutPar.fStrongLightningCutValue << endl;
    return;
  }
}

// ------------------------------------------------------------------------

void MStcCutTlt::DoStcCut()
 {
  if ( fgPrintLevel > 1 )
    FD_CINFO << "MStcCutTlt::DoStcCut() -" << endl;

  // find shower core in data
  StcShowerCore *thisShowerCore = GetShowerCore();

  if ( !thisShowerCore ) {
    fResult.fAccepted = false;
    VTltProcessor::SetEventLabel( TMirrorEventHeader::kRandomEvent );
    LogResults("ERR");
    fNSTC++;
    return;
  }
       
  thisShowerCore->AssignTime(fReadout, fCutPar.fPeakBoxcar, 
                                       fCutPar.fPeakStart, fCutPar.fPeakStop);
  
  thisShowerCore->RemoveWrongTPixel( fCutPar.fSigRemoveWrongT );

  if ( thisShowerCore->GetNCorePix() < 4 ) {
    fResult.fAccepted = false;
    VTltProcessor::SetEventLabel( TMirrorEventHeader::kRandomEvent );
    LogResults("STC");
    fNSTC++;
    delete thisShowerCore;
    return;
  }
      
  thisShowerCore->CalcLambda();

  thisShowerCore->Sort();

  thisShowerCore->DoStcCheck();
  
  if (thisShowerCore->GetNCorePix() <= fCutPar.fNPixSpecialSTC) thisShowerCore->DoSpecialStcCheck();

  fResult.fStcPar = thisShowerCore->GetTimeCorrelation();
  
  if ( fResult.fStcPar < fCutPar.fStcCutValue ) {
    fResult.fAccepted = false;
    LogResults("STC");
    fNSTC++;
    if (thisShowerCore->IsMuon()) VTltProcessor::SetEventLabel( TMirrorEventHeader::kMuonEvent );
    else VTltProcessor::SetEventLabel( TMirrorEventHeader::kRandomEvent );
    
    if ( fgPrintLevel > 2 )
      FD_CINFO << "  STC-cut:  STC-cut applied: n_STC = "
               << fResult.fStcPar << " < " << fCutPar.fStcCutValue << endl;
  }
  
  delete thisShowerCore;
}

// ------------------------------------------------------------------------

void MStcCutTlt::Finish()
 {
  if ( fFillNtuple ) {
    fResultTuple->Write();
  }
}
 
// ------------------------------------------------------------------------

StcShowerCore* MStcCutTlt::GetShowerCore()
 {
  if ( fgPrintLevel > 1 )
    FD_CINFO << "MStcCutTlt::GetShowerCore() -" << endl;

  //-------------------------
  // Explanation: First dividing Camera Picture in continuous groups,
  // then looking for biggest group, then adding all pixels of biggest group 
  // to the new shower core
  //-------------------------
      
  //camArray holds pointer to group-number of each pixel,
  //camArray has 2 additional (NULL containing) locations in each direction
    //in order to make later neighbor-search easier
  unsigned char *camArray[Fd::kMI_NCOLS+4][Fd::kMI_NROWS+2];
  
  for ( int k=0; k<Fd::kMI_NCOLS+4; k++ )
    for ( int j=0; j<Fd::kMI_NROWS+2; j++ ) camArray[k][j] = NULL;
                                                  
  //pixelGroups holds the actual group-numbers
  unsigned char *pixelGroups = new unsigned char[fResult.fNPix+1];
  memset(pixelGroups, 0, (fResult.fNPix+1)*sizeof(unsigned char));
  
  fResult.fNGroups = 0;
      
  // Looping over all triggered Pixels
  for ( unsigned int row=Fd::kMI_FIRST_ROW; row<=Fd::kMI_LAST_ROW; row++ ) {
    for ( unsigned int col=Fd::kMI_FIRST_COL; col<=Fd::kMI_LAST_COL; col++ ) {
    
      if ( fPixelData->HasTriggered(col,row,0,Fd::kMI_NSLT_BINS) ) {
        //Define 9 Pixels left under current position to search for triggered Pixels
        unsigned int chkrow[9], chkcol[9];		    
        chkrow[0]= row+1;   chkcol[0] = col; 
        chkrow[1]= row;     chkcol[1] = col+row%2;
        chkrow[2]= row;     chkcol[2] = col+row%2+1;
        chkrow[3]= row;     chkcol[3] = col+row%2+2;
        chkrow[4]= row+1;   chkcol[4] = col-1;
        chkrow[5]= row;     chkcol[5] = col+row%2-1;
        chkrow[6]= row-1;   chkcol[6] = col;
        chkrow[7]= row-1;   chkcol[7] = col+1;
        chkrow[8]= row-1;   chkcol[8] = col+2;
       
        unsigned char thisGroup = 0;
             
        for ( unsigned int j = 0; j < 9; j++ ) {  // Searching the 9 Pixels unter current Position
          if ( camArray[chkcol[j]][chkrow[j]] ){  // If one of them has already a group
            if (thisGroup == 0) {		  // and if current pixel has no group number yet
              thisGroup = *camArray[chkcol[j]][chkrow[j]];   //current pixel gets the group number of the found pixel
              camArray[col+1][row+1] = camArray[chkcol[j]][chkrow[j]];  		   
           
            }
	    else if (*camArray[chkcol[j]][chkrow[j]] != thisGroup) {
        					  //if current pixel has already a group number and it differs from the 
        					  //searched pixel, both groups are merged
              for (int k = 0; k < fResult.fNGroups; k++)
                if (pixelGroups[k]== *camArray[chkcol[j]][chkrow[j]])
		   pixelGroups[k] = thisGroup;
            }
	  }
        }
	
        if ( thisGroup == 0 ) { 		    //if current pixel didn't get a group number, a new group is started
          fResult.fNGroups++;
          thisGroup = fResult.fNGroups;
          pixelGroups[fResult.fNGroups] = fResult.fNGroups;
          camArray[col+1][row+1] = &pixelGroups[fResult.fNGroups];
        }
      } // for (j=..)

    }  // for (col=..)
  }  // for (row=..)
  
  // nGroupEntries holds number of entries for each group
  unsigned int *nGroupEntries = new unsigned int[fResult.fNGroups+1];
  memset(nGroupEntries, 0, (fResult.fNGroups+1)*sizeof(unsigned int));
         
  // Looping over all pixels, to count nGroupEntries
  for ( unsigned int row=Fd::kMI_FIRST_ROW; row<=Fd::kMI_LAST_ROW; row++ ) {
    for ( unsigned int col=Fd::kMI_FIRST_COL; col<=Fd::kMI_LAST_COL; col++ ) {
      if (camArray[col+1][row+1]) {
	nGroupEntries[*camArray[col+1][row+1]]++;
//      FD_COUT << "Pixel in col " << col << " row " << row << " is group " << (int) *camArray[col+1][row+1] << endl;
      }
    } // for col=..
  } // for row=..
	    
  // Loop over all groups, looking for the biggest one         
  unsigned int biggestGroup = 0, nBiggestGroupEntries = 0;
  for ( int j = 1; j <= fResult.fNGroups; j++ )
    if (nGroupEntries[j] > nBiggestGroupEntries) {
      nBiggestGroupEntries = nGroupEntries[j];
      biggestGroup = j;
    }
	  
  delete[] nGroupEntries;
            
  if ( nBiggestGroupEntries < 4 ) {
    FD_CERR << "MStcCutTlt::GetShowerCore() - "
    	    << "not enough Pixels in Shower core !" << endl;
    delete [] pixelGroups;
  
    return NULL;
  }
          
  if ( fgPrintLevel > 2 )
    FD_CINFO << "MStcCutTlt::GetShowerCore() - "
    	     << "biggest group " << biggestGroup << " with " 
	     << nBiggestGroupEntries << " Entries " << endl;

  // Creating and filling a new StcShowerCore
  StcShowerCore *myShowerCore = new StcShowerCore(nBiggestGroupEntries);
	 
  for ( unsigned int row=Fd::kMI_FIRST_ROW; row<=Fd::kMI_LAST_ROW; row++ ) 
    for ( unsigned int col=Fd::kMI_FIRST_COL; col<=Fd::kMI_LAST_COL; col++ ) 
      if ( camArray[col+1][row+1] ) 
	if ( (unsigned int)*camArray[col+1][row+1] == biggestGroup )
	  myShowerCore->AddPixel(col, row);

  delete [] pixelGroups;
  
  return myShowerCore;
}
 
// ------------------------------------------------------------------------

void MStcCutTlt::Init()
 {
  if ( fFillNtuple ) {
    fResultTuple = new TTree( "MStcCutTlt", "Results for MStcCutTlt" );
  
    fResultTuple->Branch( "MStcCutTlt", &fResult, 
  	 "Accepted/I:NGroups/b:NPix/i:NEarly/i:NGap/i:NStep/i:IntAll/i:StcPar/i" );
  
    FD_CINFO << "MStcCutTlt::Init() - "
  	     << "Ntuple created !" << endl;
  }
}    

// ------------------------------------------------------------------------
 
void MStcCutTlt::PrintFinalResults(std::ostream& ostr)
 {
  ostr << endl;
  ostr << "================ MStcCutTlt: Final results ===============" << endl;

  ostr << "TLT Statistic: " << endl;
  ostr << "Number of events analysed:   " << fNEventsAnalysed << endl;
  ostr << "Number of events rejected:   " << fNEventsRejected << endl;
  ostr << "Number of rej. evts kept:    " << fNRejectedKept << endl;
  
  ostr << endl;
  ostr << "- rejected n_early-cut:            " << fNEarlyPixelCut << endl;
  ostr << "- rejected weak lightning cut1:    " << fNWeakLightning1 << endl;
  ostr << "- rejected weak lightning cut2:    " << fNWeakLightning2 << endl;
  ostr << "- rejected strong lightning cut1:  " << fNStrongLightning1 << endl;
  ostr << "- rejected strong lightning cut2:  " << fNStrongLightning2 << endl;
  ostr << "- rejected Space-Time-Corr. cut:   " << fNSTC << endl;
  
  
#ifdef MSTC_TIME_OUTPUT
    double m_nT=0;
    ostr << endl;
    if (fNDecision) m_nT = (double)fSumTime/ (double)fNDecision;
    ostr << "- Mean Time needed for decision: (over "<<fNDecision <<"): " << m_nT << " ms." << endl;
    if (fNDecisionManyP) m_nT = (double)fSumTimeManyP/ (double)fNDecisionManyP; else m_nT=0;
    ostr << "- Mean Time needed for more than 25 Pixels: (over "<<fNDecisionManyP <<"): " << m_nT << " ms." << endl;
    if (fNDecision-fNDecisionManyP) m_nT = (double)(fSumTime-fSumTimeManyP)/ (double)(fNDecision-fNDecisionManyP); else m_nT=0;
    ostr << "- Mean Time needed for less than 26 Pixels: (over "<<fNDecision-fNDecisionManyP <<"): " << m_nT << " ms." << endl;
#endif // MSTC_TIME_OUTPUT

  ostr << "================================================================" << endl;
  
  ostr.flush();
}

// ------------------------------------------------------------------------

void 
MStcCutTlt::CreateMultiplicity(TMirrorPixelData::MultiplicityDataWord multid)
 {
  if ( fgPrintLevel > 1 )
    FD_CINFO << "MStcCutTlt::CreateMultiplicity() -" << endl;

  assert( multid != NULL );
  
  unsigned int fadc_bin_increment = 1;
  unsigned int n_fadc_bins = TFADCData::GetNFADCBins(TFADCData::GetDefaultResolution());
  
  if ( Fd::kMI_NMULT_BINS != Fd::kMI_NFADC_BINS_DEFAULT ) {
    FD_CWARN << "MStcCutTlt::CreateMultiplicity() - "
             << "might be incorrect, " << endl
	     << "\t\tkMI_NMULT_BINS(" << Fd::kMI_NMULT_BINS 
	     << ") != kMI_NFADC_BINS_DEFAULT(" << Fd::kMI_NFADC_BINS_DEFAULT
	     << ") !" << endl;
    fadc_bin_increment = n_fadc_bins / Fd::kMI_NMULT_BINS;
    FD_CWARN << "MStcCutTlt::CreateMultiplicity() - "
             << "calculating multiplicity from every " << fadc_bin_increment
	     << ". bin!" << endl;
  }
  
  memset(multid, 0, Fd::kMI_NMULT_BINS*sizeof(TMirrorPixelData::MultiplicityDataWordRec)); 

  // Get the ADC Data
  TFADCData *fadc;
  Fd::PixelNumberRec pixel;
  unsigned int col, row;
  TFADCData::FADCDataWord data;
  
  // setup FADCTraceInfo to request a full trace readout
  TFADCData::FADCTraceInfoRec trace_info;
  FADCTraceSetStartBin( &trace_info, 0 );
  FADCTraceSetEndBin( &trace_info, n_fadc_bins - 1 );

  // --- loop over all Pixels

  for ( row=Fd::kMI_FIRST_ROW; row<=Fd::kMI_NROWS; row++ ) {
    for ( col=Fd::kMI_FIRST_COL; col<=Fd::kMI_NCOLS; col++ ) {
   
    // returns value in range 0 ... (Fd::kMI_NSLT_BINS - 1) or -1
    if ( fPixelData->GetFirstTriggeredTimeBin(col,row) > -1 ) {

      Fd::SetPixelNoFromColumnRow( &pixel, col, row );
      
      fadc = fReadout->GetFADCData( &pixel, &trace_info );
      
      if ( fadc ) {
        
	if ( fadc->GetNFADCBins() / Fd::kMI_NMULT_BINS != fadc_bin_increment ) {
           
	   fadc_bin_increment = fadc->GetNFADCBins() / Fd::kMI_NMULT_BINS;
	   
	   FD_CWARN << "MStcCutTlt::CreateMultiplicity() - "
	            << "'fadc_bin_increment' changed to "
		    << fadc_bin_increment << " !" << endl;
	}
		    
	data = fadc->GetFADCTrace();
        
        int t2 = 0; 
	for ( unsigned int t=fadc->GetTraceStartBin(); 
	                   t<fadc->GetTraceEndBin(); t+=fadc_bin_increment ) {
          
	  // Increase Multiplicity for each triggered Pixel
          // V3: maximum multiplicity of hardware is limited to 63!
          // V4: maximum multiplicity of hardware is limited to 440
#if (MIEVENT_VERSION_CODE >= ModuleVersionCode(4,0,7))
          if (    FADCDataWordGetPixelTrigger( &data[t])) 
#else
          if (    FADCDataWordPixelTrigger( &data[t])) 
#endif // MIEVENT_VERSION_CODE
            if (multid[t2] < TMirrorPixelData::kMaxMultiplicity) multid[t2]++;

          t2++;
	     
	} // end for over fadc
	
        delete fadc;
      } // if (fadc != NULL)
      
    } // if (GetFirstTriggerTimeBin() ..)
   
   } // for col=
  } // for row=
}

// ------------------------------------------------------------------------

void MStcCutTlt::PrintSettings(std::ostream& ostr)
 {
  ostr << endl
       << "=======   MStcCutTlt::Trigger Parameters  =========" 
       << endl << endl;

#ifdef MSTC_CALC_MULTIPLICITY   // Don't use Hardware Multiplicity  
  ostr << " Using software calculated Multiplicity! " << endl << endl;
#else
  ostr << " Using hardware Multiplicity! " << endl << endl;
#endif // MSTC_CALC_MULTIPLICITY

  ostr << " Max. Time to count as early Pixel:                          " 
       << fCutPar.fTimeForEarly << " musec" << endl;
  ostr << " Number of early Pixels for Rejection:                       " 
       << fCutPar.fNEarlyCutValue << endl;
  ostr << " Slope of int_all-n_pix cut:                                 " 
       << fCutPar.fWeakLightningCutSlope << endl;
  ostr << " Offset of int_all-n_pix cut:                                " 
       << fCutPar.fWeakLightningCutOffset << endl;
  ostr << " Number of Multiplicity-gaps for Rejection:                  " 
       << fCutPar.fWeakLightningCutValue << endl;
  ostr << " Slope of n_step-n_pix cut:                                  " 
       << fCutPar.fStrongLightningCutSlope  << endl;
  ostr << " Offset of n_step-n_pix cut:                                 " 
       << fCutPar.fStrongLightningCutOffset  << endl;
  ostr << " Number of int_all for Rejection:                            " 
       << fCutPar.fStrongLightningCutValue  << endl;
  ostr << " Maximum number of pixels to perform STC-Check:              " 
       << fCutPar.fStcNPixMax  << endl;
  ostr << " Boxcar width for ADC-Peak:                                  " 
       << fCutPar.fPeakBoxcar << endl;
  ostr << " Start time for ADC-Peak search before SLT-time:             " 
       << fCutPar.fPeakStart << " x 100ns"<< endl;
  ostr << " Stop time for ADC-Peak search after SLT-time:               " 
       << fCutPar.fPeakStop << " x 100ns"<< endl;
  ostr << " Needed significance to delete pixels with wrong time:       " 
       << fCutPar.fSigRemoveWrongT << endl;
  ostr << " max. number of Pixels in shower core for special STC-Check: " 
       << fCutPar.fNPixSpecialSTC << endl;
  ostr << " Cut Value for STC-Paramter:                                 " 
       << fCutPar.fStcCutValue << endl;
  ostr << " Fraction of rejected Events to keep:                        " 
       << fCutPar.fPercentageRejKept << " % " << endl << endl;
  ostr << "===================================================" << endl;
}
 
// --------------------------------------------------------------------------

bool MStcCutTlt::ReadParameters(FdUtil::Inifile* inifile,
                                MiReadout::MStcCutTltParameters par,
				std::string section)
 {
  // --- some checks ...
  
  if ( !inifile ) {
    FD_CERR << "MStcCutTlt::ReadParameters() - "
            << "Inifile '*inifile' = NULL!" << endl;
    return false;
  }
  
  if ( !par ) {
    FD_CERR << "MStcCutTlt::ReadParameters() - "
            << "TltParameters 'par' = NULL!" << endl;
    return false;
  }
  
  if ( section.empty() ) {
    FD_CERR << "MStcCutTlt::ReadParameters() - "
            << "string 'section' is empty!" << endl;
    return false;
  }
  
  // --- let's do it
  
  if ( inifile->SpecifyGroup( section.c_str() ) != FdUtil::Inifile::kSUCCESS ) {
    FD_CERR << "!!! MStcCutTlt::Section [" << section << "] not found !!!" << endl;
    return false;
  }
  
  FD_CINFO << "\tReading section [" << section << "] ..." << endl;

  MiReadout::MStcCutTltParametersRec def = MSTC_TLT_DEFAULT_PARAMETERS;
  
  par->fTimeForEarly = 
       inifile->GetFirstValue( "TimeForEarly", def.fTimeForEarly );
  par->fNEarlyCutValue = 
       inifile->GetFirstValue( "NEarlyCutValue", def.fNEarlyCutValue );
  
  // lightning cut parameters ...
  
  par->fWeakLightningCutSlope = 
       inifile->GetFirstValue( "WeakLightningCutSlope", def.fWeakLightningCutSlope );
  par->fWeakLightningCutOffset = 
       inifile->GetFirstValue( "WeakLightningCutOffset", def.fWeakLightningCutOffset );
  par->fWeakLightningCutValue = 
       inifile->GetFirstValue( "WeakLightningCutValue", def.fWeakLightningCutValue );
  par->fStrongLightningCutSlope = 
       inifile->GetFirstValue( "StrongLightningCutSlope", def.fStrongLightningCutSlope );
  par->fStrongLightningCutOffset = 
       inifile->GetFirstValue( "StrongLightningCutOffset", def.fStrongLightningCutOffset );
  par->fStrongLightningCutValue = 
       inifile->GetFirstValue( "StrongLightningCutValue", def.fStrongLightningCutValue );
  
  par->fStcNPixMax = inifile->GetFirstValue( "StcNPixMax", def.fStcNPixMax );
  par->fPeakBoxcar = inifile->GetFirstValue( "PeakBoxcar", def.fPeakBoxcar );
  par->fPeakStart = inifile->GetFirstValue( "PeakStart", def.fPeakStart );
  par->fPeakStop = inifile->GetFirstValue( "PeakStop", def.fPeakStop );
  par->fSigRemoveWrongT = 
       inifile->GetFirstFloat( "SigRemoveWrongT", def.fSigRemoveWrongT );
  par->fNPixSpecialSTC = 
       inifile->GetFirstValue( "NPixSpecialSTC", def.fNPixSpecialSTC );
  par->fStcCutValue = 
       inifile->GetFirstValue( "StcCutValue", def.fStcCutValue );
  par->fPercentageRejKept = 
       inifile->GetFirstFloat( "PercentageRejKept", def.fPercentageRejKept );

  return true;
}

// ------------------------------------------------------------------------

void MStcCutTlt::SetParameters(const MiReadout::MStcCutTltParameters tlt)
 {
  if ( fgPrintLevel > 1 )
    FD_CINFO << "MStcCutTlt::SetParameters() -" << endl;

  fCutPar = *tlt;
  
  if ( fgPrintLevel ) PrintSettings( cout );
}

// ------------------------------------------------------------------------
// ------------------------------------------------------------------------
