
//
// File   : MStcCutTlt.hh
//
// Purpose: Declaration of the class MStcCutTlt
//
// $Id: MStcCutTlt.hh 11881 2011-01-13 09:00:27Z mathes $
//

/** @file MStcCutTlt.hh
  * Declaration of class MStcCutTlt.
  * @author Adrian Schmidt, FzK IPE
  */

  
#ifndef _MStcCutTlt_hh_
#define _MStcCutTlt_hh_

#include <iostream>
#include <cstdio>

#include <TStopwatch.h>
#include <TTree.h>

#include <MiEvent.hh>

#include <StcShowerCore.hh>

namespace MiReadout {

/** This structure describes the parameters which are used to configure the
  * software based, new (mstc) 3rd level trigger. 
  */
typedef struct _MStcCutTltParametersRec {

  unsigned int   fTimeForEarly;        // SLT-Time of First Pixel Trigger to count as early pixel (in mu secs)
  unsigned int   fNEarlyCutValue;      // minimal number of early Pixels to be rejected

  int		 fWeakLightningCutSlope;       // Cut Parameter (slope) for int_all-n_pix cut	
  int		 fWeakLightningCutOffset;      // Cut Parameter (offset) for int_all-n_pix cut   
  unsigned int   fWeakLightningCutValue;       // Cut Parameter (value) for n_gap cut	
  int		 fStrongLightningCutSlope;     // Cut Parameter (slope) for n_step-n_pix cut   
  int		 fStrongLightningCutOffset;    // Cut Parameter (offset) for n_step-n_pix cut  
  unsigned int   fStrongLightningCutValue;     // Cut Parameter (value) for int_all cut   

  unsigned int   fStcNPixMax;		  // maximum number of pixels to perform STC-Check
  unsigned int   fPeakBoxcar;		  // Boxcar width for ADC-Peak search
  unsigned int   fPeakStart;		  // Start time for ADC-Peak search in units of 100ns before SLT-time
  unsigned int   fPeakStop;		  // End time for ADC-Peak search in units of 100ns after SLT-time
  double	 fSigRemoveWrongT;	  // Number of standard deviations for time difference to remove a pixel
  unsigned int   fNPixSpecialSTC;	  // max. number of Pixels in shower core for special STC-Check removing single pixels
  unsigned int   fStcCutValue;  	  // minimal Space Time Correlation Parameter for being accepted as shower
  double	 fPercentageRejKept;	  // Percentage of rejected events to be kept.

} *MStcCutTltParameters, MStcCutTltParametersRec;

}

/** The default settings for the MStcCutTltParametersRec struct. */

#define MSTC_TLT_DEFAULT_PARAMETERS \
 { \
  4,	 /* fTimeForEarly */  \
  6,	 /* fNEarlyCutValue */  \
  8,	 /* fWeakLightning1CutSlope */  \
  -100,  /* fWeakLightning1CutOffset */  \
  40,	 /* fWeakLightning2CutValue */  \
  2,	 /* fStrongLightning1CutSlope */  \
  60,	 /* fStrongLightning1CutOffset */  \
  10000, /* fStrongLightning2CutValue */  \
  25,	 /* fStcNPixMax */  \
  2,	 /* fPeakBoxcar */  \
  20,	 /* fPeakStart */  \
  20,	 /* fPeakStop */  \
  2.0,   /* fSigRemoveWrongT */  \
  6,	 /* fNPixSpecialSTC */  \
  3,	 /* fStcCutValue */  \
  1.0	 /* fPercentageRejKept */  \
}

// --- forward declaration(s)

class VirtualReadout;
class VMirrorEvent;

namespace FdUtil {
  class Inifile;
}

#include "VTltProcessor.hh"

/** This is a TLT decision class based on the FLTs multiplicity data
  * plus an analysis of the Space-Time-Correlation of the camera pixels
  */
class MStcCutTlt : public VTltProcessor {
public:
    /** Constructor for the class MStcCutTlt. */
    MStcCutTlt(MiReadout::MStcCutTltParameters);

    /** Destructor of the class MStcCutTlt. */
    ~MStcCutTlt();

    /** Clear internal data structures to be prepared for the next Event. */
    void Clear();

    /** Perform the TLT decision. */
    void Decide(VMirrorEvent*);

public:          

    /** Analyses the Multiplicity Trace and calculates the curve parameters needed for the M-cuts      */
    void AnalyseMultiplicity(TMirrorPixelData::MultiplicityDataWord);
    
    /** Creates the multiplicity out of the ADC Data if needed. 
      *
      * Note: Create Multiplicity only if Multiplicity is not 
      *       implemented in the FE Hardware!
      */
    void CreateMultiplicity(TMirrorPixelData::MultiplicityDataWord);
    
    /** Performs cut on number of early Pixels  */
    void DoTEarlyCut();
    
    /** Performs the M-cuts */
    void DoMultiplicityCut();
    
    /** Performs the STC-Check */
    void DoStcCut();
    
    /**  */
    void FillNtuple(bool fill= true)
     { fFillNtuple = fill; }

    /** Write (eventually) Ntuple to disk. */
    void Finish();

    /**  */
    StcShowerCore* GetShowerCore();
    
    /** Create (eventually) the Ntuple. */
    void Init();
    
    /** Write the result of the current decision step to the log file. A short
      * label which is used as a prefix of the log message could be passed.
      */
    void LogResults(const char *);

    /** Print the final statistics of this processor. */
    void PrintFinalResults(std::ostream& ostr=std::cout);    

    /** Print the current TLT settings. */
    void PrintSettings(std::ostream& ostr=std::cout);

    /** Read the M-cut and STC-cut Parameters from the passed 
      * FdUtil::Inifile.
      *
      * @return 'true' on success, else 'false'.
      */
    static bool ReadParameters(FdUtil::Inifile*,
                MiReadout::MStcCutTltParameters,
		std::string section="MStcCutTlt");
    
    /** Sets the M-cut and STC-cut Parameters. 
      *  
      * If this Fuction has not been performed, default Parameters are used    
      */  
    void SetParameters(const MiReadout::MStcCutTltParameters);

private:
    
    // these ones don't exist
    MStcCutTlt(const MStcCutTlt&);
    MStcCutTlt& operator=(const MStcCutTlt&);
    
    MiReadout::MStcCutTltParametersRec fCutPar;
    
    
    typedef struct _MStcCutTltResultRec {
    
      bool           fAccepted;		    // needed for LogResults()
      
      unsigned char  fNGroups;
    
      unsigned int   fNPix;                 // Number of triggered pixels in camera
      unsigned int   fNEarly;               // Number of early triggered pixels in camera
      unsigned int   fNGap;                 // Number of Gaps in Multiplicity Curve
      unsigned int   fNStep;                // Number and Height of Steps in M. Curve
      unsigned int   fIntAll;               // Integral of M. Curve
      unsigned int   fStcPar;               // Space Time Correlation Parameter 

    } MStcCutTltResult_t;
        
    TMirrorPixelData      *fPixelData;
    TMirrorBitArray       *fPixelMatrix;          // Do I need this?
        
    VMirrorEvent          *fReadout;
    
    // --- counting and flag variables

    bool                   fFillNtuple;
    
    unsigned int           fNEventsAnalysed;      // number of events anal.
    unsigned int           fNEventsRejected;      // number of events rejected
    unsigned int           fNRejectedKept;        // number of rejected events
                                                  // kept for monitoring purposes

    unsigned int           fNEarlyPixelCut;       // Number of by n_early - cut rejected events
    unsigned int           fNWeakLightning1;      // Number of by int_all-n_pix - cut rejected events
    unsigned int           fNWeakLightning2;      // Number of by n_gap - cut rejected events
    unsigned int           fNStrongLightning1;    // Number of by n_step-n_pix - cut rejected events
    unsigned int           fNStrongLightning2;    // Number of by int_all - cut rejected events
    unsigned int           fNSTC;                 // Number of by stc - cut rejected events
    
    // --- result variables
    
    MStcCutTltResult_t     fResult;
    
    TTree                 *fResultTuple;
   
#ifdef MSTC_TIME_OUTPUT
    // --- used for timing and profiling
    
    unsigned int           fNDecision;
    unsigned int           fSumTime;
    unsigned int           fNDecisionManyP;
    unsigned int           fSumTimeManyP;
#endif // MSTC_TIME_OUTPUT
};

#endif  // _MStcCutTlt_hh_
