
//
// File   : NeighborPixelReadout.cc
//
// Purpose: Implementation of the class NeighborPixelReadout
//
// $Id: NeighborPixelReadout.cc 9725 2010-09-28 07:31:17Z mathes $
//


#include <iostream>

/** @file NeighborPixelReadout.cc
  * Implementation of the class .
  * @author H.-J. Mathes, FzK
  */

#include <FdNumbering.hh>
#include <ErrorLogging.hh>
#include <MiEvent.hh>

#include "CheckVersion.hh"

#include "NeighborPixelReadout.hh"

using namespace std;
using FdUtil::Fd;
using namespace MiReadout;

// -------------------------------------------------------------------------
// -------------------------------------------------------------------------
// 
// class NeighborPixelReadout - implementation
// 

NeighborPixelReadout::NeighborPixelReadout(SltParameters slt_par)
 : SelectivePixelReadout(slt_par)
 {
  if ( SelectivePixelReadout::fgPrintLevel > 2 )
    FD_CINFO << "NeighborPixelReadout::NeighborPixelReadout() -" << endl;

  // do local modification(s)
  SetSltParameters(slt_par);
  
  SetName("NeighborPixelReadout");
}

// -------------------------------------------------------------------------

NeighborPixelReadout::~NeighborPixelReadout()
 {
  if ( SelectivePixelReadout::fgPrintLevel > 2 )
    FD_CINFO << "NeighborPixelReadout::~NeighborPixelReadout() -" << endl;
}

// --------------------------------------------------------------------------

void NeighborPixelReadout::PrintFinalResults(ostream &ostr)
 {
  ostr << endl;
  ostr << "=== NeighborPixelReadout: Final results === " << endl;

  // using variables from the base class
  
  ostr << "calls/(real/CPU) time:  " << setw(5) << fStopwatch.Counter() << " "
       << setprecision(4) << fStopwatch.RealTime() << " " 
       << setprecision(4) << fStopwatch.CpuTime() << endl;
  ostr << "<Time>/call(real/CPU): " << "     "
       << setprecision(4) 
       << fStopwatch.RealTime()/fStopwatch.Counter() 
       << " " << setprecision(4) 
       << fStopwatch.CpuTime()/fStopwatch.Counter() << endl;

  ostr << "SelectivePixelReadout Statistic: " << endl;
  ostr << "Number of events analysed:   " << fNEventsAnalysed << endl;
  ostr << "Number of events rejected:   " << fNEventsRejected << endl;
  
  ostr << "==========================================" << endl;
}

// -------------------------------------------------------------------------

void NeighborPixelReadout::Select(TMirrorPixelData* pixel_data,
                                  TMirrorEventHeader::EMiTriggerSource trigger_source,
				  unsigned int /* column_list */)
 {
  if ( SelectivePixelReadout::fgPrintLevel > 2 )
    FD_CINFO << "NeighborPixelReadout::Select() -" << endl;

  if ( !pixel_data ) {
    FD_CERR << "NeighborPixelReadout::Select() - no PixelData present !" << endl;
    return;
  }
  
#if 1
  // will the parameter 'fSltParameters.fMinDistance' delete single pixels
  // (via SelectivePixelReadout::DeleteSinglePixels()) ?
  
  SelectivePixelReadout::Select( pixel_data, trigger_source );
#else
  // alternative code copied and modified from SelectivePixelReadout::Select(),
  // requires some base class variables to be protected, but at present
  // the way modifying the SLTParameters seems to be more elegant to me
#endif

  if ( fgPrintLevel ) {
    unsigned int n_pixels = GetPixelList().GetNumPixels();
  
    FD_CINFO << "NeighborPixelReadout::Select() -"
	     << "SelectivePixelReadout::Select() selected " << n_pixels
 	     << " pixels for readout" << endl;
  }
  
  // --- add my own specific pixels to the readout list
  
  unsigned int col_min, col_max;
  TMirrorBitArray ba = GetPixels();
  
  // determine col_min and col_max values
  
  if ( trigger_source & TMirrorEventHeader::kLeftTrigger ) {
    col_min = Fd::kMI_FIRST_COL;
    col_max = col_min + fSltParameters.fNExtColumns - 1;
  }
  else if ( trigger_source & TMirrorEventHeader::kRightTrigger ) {
    col_max = Fd::kMI_LAST_COL;
    col_min = col_max - fSltParameters.fNExtColumns + 1;
  }
  
  // add all pixels in columns col_min ... col_max to the list
  
  if ( trigger_source & (TMirrorEventHeader::kLeftTrigger
                         | TMirrorEventHeader::kRightTrigger) ) {
    for ( unsigned int col=col_min; col<=col_max; col++ ) {
      for ( unsigned int row=Fd::kMI_FIRST_ROW; row<=Fd::kMI_LAST_ROW; row++ )
        ba.Set( col, row );
    }
    
    TMirrorPixelList pixel_list( &ba );
    
    VPixelSelector::Set( &pixel_list );
    
    // ba.Show();

    if ( fgPrintLevel ) {
      unsigned int n_pixels = GetPixelList().GetNumPixels();
  
      FD_CINFO << "NeighborPixelReadout::Select() -"
	       << "with 'column extension' " << n_pixels
 	       << " pixels are selected for readout" << endl;
    }
  }
  else {
    FD_CERR << "NeighborPixelReadout::Select() - "
            << "unhandled trigger_source 0x" << hex << trigger_source
	    << dec << endl;
  }
}

// -------------------------------------------------------------------------

void NeighborPixelReadout::SetSltParameters(MiReadout::SltParameters slt)
 {
  if ( !slt ) {
    FD_CERR << "NeighborPixelReadout::SetSltParameters() - "
            << "slt = NULL !" << endl;
    return;
  }
  
  MiReadout::SltParametersRec slt_par = *slt;
  
  // switch off SelectivePixelReadout::DeleteSinglePixels()
  slt_par.fMinDistance = Fd::kMI_NROWS;
  // we might also set a different fMaxEventSize
  // slt_par.fMaxEventSize = Fd::kMI_NPIXELS;
  // slt_par.fMaxEventSize = 10;
  
  SelectivePixelReadout::SetParameters(&slt_par);
}

// -------------------------------------------------------------------------
// -------------------------------------------------------------------------
