
//
// File   : PFPedro.cc
//
// Purpose: Implementation of the class PFPedro
//
// $Id: PFPedro.cc 12285 2011-03-31 10:35:52Z mathes $
//


#include <iostream>
#include <cmath>
#include <unistd.h>


/** @file PFPedro.cc
  * Implementation of the pulse finder class PFPedro.
  * @author P. Facal, INFN Rome, H.-J. Mathes, FzK
  */


#include <TNamed.h>

#include <FdNumbering.hh>
#include <ErrorLogging.hh>
#include <MiEvent.hh>

#include "CheckVersion.hh"

#include "VPulseFinder.hh"
#include "PFPedro.hh"

using namespace std;
using namespace FdUtil;

// -------------------------------------------------------------------------
// -------------------------------------------------------------------------
//
// class PFPedro - implementation
//

const float PFPedro::kTriggerThreshold = 3.4;

unsigned int PFPedro::fgPrintLevel = 0;

PFPedro::PFPedro()
 : fBinWidthFactor(1), fMinPulseLength(kMinPulseLength)
 {
  if ( fgPrintLevel >= 2 )
    FD_COUT << "PFPedro::PFPedro(): " << hex << this << dec << endl;
  
  SetName( "PFPedro" );
  
  Clear();
}

// -------------------------------------------------------------------------

PFPedro::~PFPedro()
 {
  if ( fgPrintLevel > 1 )
    FD_COUT << "~PFPedro" << hex << this << dec << endl;
}

// -------------------------------------------------------------------------

void PFPedro::CalculatePulseIntegral()
 {
  fNMaxPulse = 0;
  
  for ( int i=fNPulses; i>0; i-- ) {
    
    float trigthresh = 0.;
      
    if ( fgPrintLevel > 0 )
      FD_COUT << "Integral = " << fPulseIntegral[i-1] << endl;
      
    fPulseIntegral[i-1] = fPulseIntegral[i-1] - 
	                    fPedestal * (float)fPulseLength[i-1];
    // SetTriggerThreshold(float sigmaf)
    trigthresh = kTriggerThreshold * fSigmaPedestal * sqrt( (float)fPulseLength[i-1] );
      
    
    if ( fgPrintLevel >= 2 )
      printf( "trigthresh %f %f \n", trigthresh, fPulseIntegral[i-1] );
      
    // remove pulses below the trigger threshold 
    // and find largest pulse ...
      
    if ( fPulseIntegral[i-1] <= trigthresh ) {
      fNPulses            = fNPulses - 1; 
      fPulseIntegral[i-1] = fPulseIntegral[i];
      fPulseLength[i-1]   = fPulseLength[i];
      fPulseStart[i-1]    = fPulseStart[i];
      fNMaxPulse          = i - 1;
    }
  }
}

// -------------------------------------------------------------------------

void PFPedro::CalculateStatistic(TFADCData* fadc)
 {
  const int kCalculationStart = 50;
  const int kCalculationEnd = 30;
  
  int value;
  float pedestal      = 0.;
  float pedestal2     = 0.;
  float sigped        = 0.;

  // --- calculate pedestal and its variance (outside the pulses area)
  
  TFADCData::FADCDataWord fadc_word_ptr = fadc->GetFADCTrace();
  
  const unsigned int kLength = fBinWidthFactor*(kCalculationStart-kCalculationEnd);
  const unsigned int kStart = max((int)0, fPulseStart[fNMaxPulseIntegral] - kCalculationStart*int(fBinWidthFactor)); 

  for ( unsigned int k=kStart;
	k < kStart + kLength; k++ ) {

    value = FADCDataWordGetData( &fadc_word_ptr[k] );
    pedestal  += (double)value;
    pedestal2 += (double)(value*value);
  }
    
  pedestal  /= (float)kLength;
  pedestal2 /= (float)kLength;

  sigped    = sqrt( pedestal2 - pedestal*pedestal ); 
    
  fPedestal      = pedestal;
  fSigmaPedestal = sigped;
    
  if ( fgPrintLevel >= 1 )
    printf( "\t(%d). Pedestal = %.2f. Sigped = %.2f\n\n", 
            fNMaxPulse, fPedestal, fSigmaPedestal );
}
    
// -------------------------------------------------------------------------

void PFPedro::Clear(Option_t * /* option */)
 {
  fNPulses = 0;
  
  for ( int i=0; i<kMaxPulses; i++ ) {
    fPulseLength[i]   = 0;
    fPulseIntegral[i] = 0.;
    fPulseStart[i]    = 0;
  }
}

// -------------------------------------------------------------------------

void PFPedro::FindPulse(TFADCData* fadc)
 {
  Clear();
  
  // --- calculate bin width dependant factor @ start of calculation
  
  fBinWidthFactor = 1;

  if ( fadc->GetTraceResolution() == TFADCData::kFADC_RESOLUTION_2000_BINS )
    fBinWidthFactor = 2;
  else if ( fadc->GetTraceResolution() == TFADCData::kFADC_RESOLUTION_4000_BINS )
    fBinWidthFactor = 4;

  int minPulseLength        = fBinWidthFactor * fMinPulseLength;
  
  int npulseold             = -9999;

  fNMaxPulse                = 0;
  fNMaxPulseIntegral        = 0;
  fMaxPulseIntegral         = 0.;
  
  if ( fgPrintLevel >= 1 ) {
    Fd::PixelNumber pixel = fadc->GetPixelNumber();
    FD_COUT << "** Pixel: " << Fd::GetPixelNo( pixel ) << " (col,row) = ("
            << Fd::GetColumnNo( pixel ) << "," << Fd::GetRowNo( pixel ) << ")" 
	    << endl;
  }
  
  // this is possible from v1r3 on ...
  
  if ( fgPrintLevel >= 2 ) {
    FD_COUT << "Start: " << fadc->GetTraceStartBin() << endl;
    FD_COUT << "End:   " << fadc->GetTraceEndBin() << endl;
  }

  unsigned int first_bin = fadc->GetTraceStartBin();
  unsigned int last_bin = fadc->GetTraceEndBin();

  // loop over FADC trace and find maximum ADC value to start threshold

  TFADCData::FADCDataWord fadc_word_ptr = fadc->GetFADCTrace();
  
  unsigned int integral       = 0; 
  
  for ( unsigned int i=first_bin; i<=last_bin; i++ ) {    // loop over time slots
    integral += FADCDataWordGetData( &fadc_word_ptr[i] );
  }
  
  // threshold: initial value = average + 100
  
  int thresh = (int)integral/(last_bin - first_bin + 1) + 100;
    
  // --- going down with threshold and finding pulses
    
  while ( thresh >= 0 ) {
    
    if ( fgPrintLevel >= 3 )
      FD_COUT << "Threshold = " << thresh << "**" << endl; 
      
    int nPulses = 0;
    int istart  = 0;
    int iend    = 0;
    
    thresh      -= 1;  // SetThresholdDecrement()

    // --- loop over ADC time slots to find pulse start and end
    
    int value;
    
    for ( unsigned int i=first_bin; i<=last_bin; i++ ) {
      
      value = FADCDataWordGetData( &fadc_word_ptr[i] );
	
      // pulse start found
	
      if ( ( value >= thresh ) && ( istart == 0 ) ) {
	fPulseStart[nPulses] = i;
	istart = 1;
	
        if ( fgPrintLevel >= 3 )
          FD_COUT << "PulseStart[" << nPulses << "]: " 
	          << fPulseStart[nPulses] << endl;
      }     
	
      // pulse end found
	
      if ( ( value < thresh ) && ( iend == 0 ) && ( istart == 1 ) ) {
	  
	// calculate pulse length
	  
	fPulseLength[nPulses] = i - fPulseStart[nPulses];
	  
	// accept pulse if at least 'minPulseLength' timeslots width (filter)
	  
        if ( fPulseLength[nPulses] >= minPulseLength ) {
            
	  // Calculate now the integral
	    
	  fPulseIntegral[nPulses] = 0.; 

	  for ( unsigned int k=fPulseStart[nPulses]; k<i; k++ ) {
	    value = FADCDataWordGetData( &fadc_word_ptr[k] );
	    fPulseIntegral[nPulses] += (float)value;
          }
	  
	  // look for the largest pulse (largest integral)
	  
	  if ( fPulseIntegral[nPulses] >= fMaxPulseIntegral ) { 
	    fMaxPulseIntegral  = fPulseIntegral[nPulses];
	    fNMaxPulseIntegral = nPulses;
	  }

	  istart = 0;
	  iend = 0;
	    
	  nPulses = nPulses + 1;
	}
	
	// how do we terminate the loop properly ?
	
	if ( nPulses == 20 ) {
	  // FD_CERR << "Error in PFPedro: nPulses= " << nPulses << endl;
          istart = 1;
          iend   = 1;
        }
      }  
    }
      
    if ( ( npulseold > 1 ) && ( nPulses == 1 ) ) {
	
      fNPulses = nPulses;
      
      if ( fgPrintLevel >= 1 ) {
        FD_COUT << "\t===> Found "<< fNPulses << " pulse(s)\n" << endl;
	
        for ( int r=0; r<fNPulses; r++ ) 
	  FD_COUT << "\ttstart = " << fPulseStart[r] << ", Nbins = "
	          << fPulseLength[r] << endl;
      }
      
      thresh = -9999;  // reset
    }
    
    // --- execute this if more than one pulse found
     
    if ( nPulses > 1 ) {
	
      if ( npulseold == 1 ) 
	thresh = thresh + 2; 
	
      if ( npulseold == 0 ) {
	
	fNPulses = nPulses;
	  
        if ( fgPrintLevel >= 1 ) {
	  FD_COUT << "\t===> Found " << fNPulses << " pulse(s)" << endl;
	  
	  for ( int r=0; r<fNPulses; r++ ) 
	    FD_COUT << "\t Pulse " << r << ": tstart = " << fPulseStart[r] 
	            << ", Nbins = " << fPulseLength[r] << endl;
	}
	
	thresh = -9999;
      }
    }
      
    npulseold = nPulses;
  }	  
  
  // --- pedestal and variance might be determined by other means ...
  
  TFADCData::PixelMonitorData monitor = fadc->GetPixelMonitorData();
  
  if ( fgPrintLevel >= 2 ) {
    FD_COUT << "Pedestal in data: " 
            << PixelMonitorDataGetMean( monitor ) << endl;
    FD_COUT << "Variance in data: " 
            << PixelMonitorDataGetVariance( monitor ) << endl;
  }
  
  // if no pedestal/variance values or if not reasonable, calculate them !
  
  if ( fPedestal <= 0. || fSigmaPedestal <= 0. ) CalculateStatistic( fadc );
  
  CalculatePulseIntegral();
}

// -------------------------------------------------------------------------
// -------------------------------------------------------------------------
