
//
// File   : PrototypeTltProcessor.cc
//
// Purpose: Implementation of the class PrototypeTltProcessor
//
// $Id: PrototypeTltProcessor.cc 11881 2011-01-13 09:00:27Z mathes $
//


#include <iostream>
#include <iomanip>
#include <algorithm>

#include <cstdio>
#include <ctime>
#include <list>


/** @file PrototypeTltProcessor.cc
  * Implementation of the class PrototypeTltProcessor.
  * @author H.-J. Mathes, FzK
  */

#include <FdNumbering.hh>
#include <FdStdioStream.hh>
#include <ErrorLogging.hh>
#include <Inifile.hh>
#include <MiEvent.hh>

#include <AnalysedFadcData.hh>
#include <AnalysedPixelData.hh>
#include <PFPedro.hh>
#include <PixelPattern.hh>
#include <VMiEvent.hh>

#include "CheckVersion.hh"

#include "ReadoutConfig.hh"

#include "PrototypeTltProcessor.hh"

using namespace std;
using FdUtil::Fd;
using namespace MiReadout;

// --------------------------------------------------------------------------
// --------------------------------------------------------------------------
//
// class PrototypeTltProcessor - implementation
//

PrototypeTltProcessor::PrototypeTltProcessor(MiReadout::TltParameters par)
 : VTltProcessor("PrototypeTltProcessor"),
   fAnalysedPixelData(NULL), fPixelData(NULL), 
   fPixelMatrix(new TMirrorBitArray), fReadout(NULL)
 {
  // set some default parameters
  
  TltParametersRec tlt = DEFAULT_TLT_PARAMETERS;
  
  fTltParameters = tlt;
  
  SetTltType( VTltProcessor::kPrototype );
  
  // set verbose print level for some of the used classes
  
  TAnalysedPixelData::SetPrintLevel( 0 );
  TAnalysedFADCData::SetPrintLevel( 0 );

  // set the pulse finder to be used and (optional) its parameters
  
  VPulseFinder *finder = new PFPedro();  // is then lost ?
  
  if ( dynamic_cast<PFPedro*>( finder ) ) {
    dynamic_cast<PFPedro*>( finder )->SetMinPulseLength( 5 );
    
    PFPedro::SetPrintLevel( 0 );
  }
  
  TAnalysedFADCData::SetPulseFinder( finder );
//  finder->SetPrintLevel( 0 );
    
  fStopwatch.Reset();
  
  // --- check if parameters have been passed in
  
  if ( par ) {
    SetParameters( par );
  }
  
  // --- get the default (hardcoded) parameters for the TLT 
  //     and write them into the base class variable

  else {
    TltParametersRec tlt = DEFAULT_TLT_PARAMETERS;
    SetParameters( &tlt );

  // --- try to read the actual desired TLT parameters from ini-file ...
  
    SetParameters( ReadoutConfig::GetInstance()->GetTltParameters() );
  }
  
  // --- initialize all flag/counting variables
  
  PrototypeTltProcessor::Clear();

  fNEventsAnalysed  = 0;
  fNEventsRejected  = 0;
  fNRejectedKept    = 0;
  
  fNPatternAnalysed = 0;
  
  fNError           = 0;     // number of errors occurred
  fNLessPulses      = 0;     // not enough pulses
  fNMulti           = 0;     // multi pattern
  fNManyPixels      = 0;     // too many pixels
  fNNotProp         = 0;     // time-space not prop.
  fNPulseMismatch   = 0;     // insufficient number of p.
  fNRandomTime      = 0;     // random pattern
  fNShowerC         = 0;     // shower candiates
}

// --------------------------------------------------------------------------

PrototypeTltProcessor::~PrototypeTltProcessor()
 {
  // fAdcList.clear();
  
  list<TFADCData*>::iterator iter = fAdcList.begin();
  
  while ( iter != fAdcList.end() ) {
    delete *iter;
    iter = fAdcList.erase( iter );
  }
    
  delete fPixelMatrix;
  delete fAnalysedPixelData;
  
  if ( fTltLogfile ) {
    FdUtil::OstdioStream ostr( fTltLogfile );

    PrintFinalResults( ostr );
  }
  
  if ( fgPrintLevel || true ) {
    PrintFinalResults( cout );
  }
}

// --------------------------------------------------------------------------

void PrototypeTltProcessor::AssignAdcTime(PixelPattern *pattern, 
                                          unsigned int n, 
					  TAnalysedFADCData* afadc)
 {
  int p_index = 0;
  unsigned int diff, abs_diff = afadc->GetNFADCBins() * afadc->GetBinWidth();
  bool pulse_found = false;
  
  if ( !afadc->PulseFound() ) {
    pattern->SetAdcTime( n, 0 );
    return;
  }
  
  if ( fgPrintLevel > 1 ) {
    FD_CINFO << "Analysing pulse " << n << " with SLT time= " 
             << pattern->GetTime( n ) << endl;
  }
  
  // --- calculate concidence window in ns (1500ns = 15 * 100ns, 15 * 50ns * 2)
  
  unsigned int factor =   afadc->GetBinWidth( TFADCData::kFADC_RESOLUTION_1000_BINS )
                        / afadc->GetBinWidth();
  
  unsigned int coincidence_window =   fTltParameters.fCoincidenceWindow
	                            * afadc->GetBinWidth()
			            * factor;
  
  // --- assign best matching pulse time 
  
  for ( int j=0; j<afadc->GetNPulses(); j++ ) {
  
    unsigned int t_adcstart = afadc->GetPulseStart( j ) * afadc->GetBinWidth();
    
    if ( fgPrintLevel > 1 ) {
      FD_CINFO << "Testing ADC pulse " << j << " with time= " 
               << t_adcstart << endl;
    }
    
    // check if the pulse is near the SLT pulse
     
    diff = abs( (int)(t_adcstart - pattern->GetTime( n )) );
    
    if ( diff < coincidence_window ) {
	    
      pulse_found = true;

      if ( fgPrintLevel > 1 ) {
        FD_CINFO << "Pulse found(1), diff = " << diff << endl;
      }
      
      if ( diff < abs_diff ) {
        abs_diff = diff;
        p_index = j;
      }
    }
    
    // check if the pulse is at least within the FLT trigger window
    
    if ( !pulse_found ) {

      if (   (int)(pattern->GetEndTime( n ) - t_adcstart) > 0
           && (int)(t_adcstart - pattern->GetTime( n )) > 0 ) {
        
	pulse_found = true;
	
	diff = abs( (int)(t_adcstart - pattern->GetTime( n )) );

        if ( fgPrintLevel > 1 ) {
          FD_CINFO << "Pulse found(2), diff = " << diff 
	           << " " << (int)(pattern->GetEndTime(n) - t_adcstart)
	           << " " << (int)(t_adcstart - pattern->GetTime(n))
	           << endl;
	}
	
        if ( diff < abs_diff ) {
          abs_diff = diff;
          p_index = j;
        }
      }
    }
  } // for ( int j=0; j<afadc->GetNPulses(); j++ ) 
  
  unsigned int t_max = afadc->GetNFADCBins() * afadc->GetBinWidth();
  
  if ( pulse_found ) {
  
    unsigned int t_adcstart = afadc->GetPulseStart(p_index) * afadc->GetBinWidth();
    
    if ( t_adcstart < t_max )
      pattern->SetAdcTime( n, t_adcstart );
    else
      pattern->SetAdcTime( n, 0 );
  }
}

// --------------------------------------------------------------------------

void PrototypeTltProcessor::Clear()
 {
  if ( fgPrintLevel > 1 )
    FD_CINFO << "PrototypeTltProcessor::Clear() !" << endl;

  fReadout = NULL;
  
  VTltProcessor::SetEventAccepted( false );
  
  // fAdcList.clear();
  
  list<TFADCData*>::iterator iter = fAdcList.begin();
  
  while ( iter != fAdcList.end() ) {
    delete *iter;
    iter = fAdcList.erase( iter );
  }
  
  fPixelData = NULL;
  VTltProcessor::SetEventLabel( TMirrorEventHeader::kUnlabelled );
  
  fPixelMatrix->Clear();

  if ( fAnalysedPixelData ) delete fAnalysedPixelData;
  fAnalysedPixelData = NULL;
  
  // --- initialize all flag variables
  
  fBigEvent   = false;
  fLessPulses = false;
  fUnlabelled = true;
  fMulti      = false;
  fMuon       = false;
  fNotProp    = false;
  fShowerC    = false;
  fRandom     = false;
  
  fEventSize  = 0;
  fCut1Result = 0.;
  fCut2Result = 0.;
  fCut2Error  = 0.;
  fNumPattern = 0;
}

// --------------------------------------------------------------------------

void PrototypeTltProcessor::Decide(VMirrorEvent* mi_event)
 {
  if ( !mi_event ) {
    FD_CCRIT << "PrototypeTltProcessor::Decide() - "
             << "error passing parameters to method Decide() !" 
             << endl;
    VTltProcessor::SetEventAccepted( false );
    return;
  }
  
  fReadout = mi_event;
  
  // ToDo: get somehow the pixel data
  
  if ( ( fPixelData = fReadout->GetPixelData() ) == NULL ) {
    FD_CCRIT << "PrototypeTltProcessor::Decide() - "
             << "no PixelData present !" << endl;
    return;
  }
  
  // fPixelData->Dump();
    
  // --- perform level 1 decision
  
  fNEventsAnalysed++;
      
  VTltProcessor::SetEventLabel( TMirrorEventHeader::kAcceptedEvent );
  VTltProcessor::SetEventAccepted();
      
  fStopwatch.Start(kFALSE);  // *** START
      
   // write actually set pixels to my internal bit array
      
   fPixelMatrix->Set( fPixelData, fTltParameters.fFirstSltBin,
                                  fTltParameters.fLastSltBin );
   
   // fPixelMatrix->Dump("after setting...");
   
   DoLevel1Decision();

  fStopwatch.Stop();         // *** STOP


  LogResults("Tlt1");

  // --- perform level 2 decision

  if ( !fAdcList.empty()) {
      
    VTltProcessor::SetEventAccepted( true );
      
    fStopwatch.Start(kFALSE);
     DoLevel2Decision();
    fStopwatch.Stop();
      
    if ( fgPrintLevel ) {
    
      FD_COUT << "--- Level2Decision: The final results ---" << endl;
	
      FD_COUT << "Unlabelled: " << (int)fUnlabelled << endl;  // -> kAcceptedEvent
      FD_COUT << "LessPulses: " << (int)fLessPulses << endl;  // -> reject
      FD_COUT << "Multi:      " << (int)fMulti << endl;       // ->
      FD_COUT << "Random:     " << (int)fRandom << endl;      // -> kRandomEvent, reject
      FD_COUT << "NotProp:    " << (int)fNotProp << endl;     // -> kMuonEvent, reject
      FD_COUT << "BigEvent:   " << (int)fBigEvent << endl;    // -> reject
      FD_COUT << "ShowerC:    " << (int)fShowerC << endl;     // -> kShowerCandidate

#if 0
      static int ctr = 1;
	
      FD_COUT << "ctr= " << ctr << " " << (int)fUnlabelled
                         << " " << (int)fLessPulses
                         << " " << (int)fMulti
                         << " " << (int)fRandom
                         << " " << (int)fNotProp
                         << " " << (int)fBigEvent
                         << " " << (int)fShowerC << endl;
      ctr++;
#endif
    }  // if ( fgPrintLevel )
      
    // ToDo: put into separate method !

    if ( fShowerC ) {
      VTltProcessor::SetEventLabel( TMirrorEventHeader::kShowerCandidate );
      VTltProcessor::SetEventAccepted( true );
    }
    else if ( fMulti || fUnlabelled ) {
      VTltProcessor::SetEventLabel( TMirrorEventHeader::kAcceptedEvent );
      VTltProcessor::SetEventAccepted( true );
    }
    else if ( fNotProp || fMuon ) {
      VTltProcessor::SetEventLabel( TMirrorEventHeader::kMuonEvent );
      VTltProcessor::SetEventAccepted( false );
    }
    else if ( fLessPulses ) {
      VTltProcessor::SetEventLabel( TMirrorEventHeader::kUnlabelled );
      VTltProcessor::SetEventAccepted( false );
    }
    else if ( fBigEvent ) {
      VTltProcessor::SetEventLabel( TMirrorEventHeader::kManyPixelEvent );
      VTltProcessor::SetEventAccepted( false );
    }
    else { /* fRandom */
      VTltProcessor::SetEventLabel( TMirrorEventHeader::kRandomEvent );
      VTltProcessor::SetEventAccepted( false );
    }
      
    LogResults("Tlt2");
  } // if ( !fAdcList.empty())
  else {
    FD_CERR << "PrototypeTltProcessor::Decide() - "
            << "fAdcList.empty() !" << endl;
  }
    
  // --- finalize the things if event has been rejected
    
  if ( !VTltProcessor::IsEventAccepted() ) {
      
    fNEventsRejected++;

    // keep some of the rejected events for monitoring of the algorithm
      
    if (    !VTltProcessor::IsEventAccepted()
         && fTltParameters.fKeepRejectedFraction ) {
	
      bool keep = false;
	
      if ( fTltParameters.fKeepRejectedFraction == 100 ) keep = true;
	  
      if ( (unsigned int)(1 + (unsigned int)(100.0*rand()/(RAND_MAX+1.0) )) <=
	fTltParameters.fKeepRejectedFraction ) {
	keep = true;
      }
	
      if ( keep ) {
        VTltProcessor::SetEventAccepted( true );
	fNRejectedKept++;
        fNEventsRejected--;
	  
	if ( fgPrintLevel ) {
	  FD_CWARN << "Keeping rejected event !" << endl;
	}
      }
    }
      
    LogResults("Rej");
      
  }  // !VTltProcessor::GetEventAccepted()

}

// --------------------------------------------------------------------------

void PrototypeTltProcessor::DoLevel1Decision()
 {
  if ( fgPrintLevel > 1 )
    FD_CINFO << "PrototypeTltProcessor::DoLevel1Decision()" << endl;
  
  if ( fAnalysedPixelData ) delete fAnalysedPixelData;
  fAnalysedPixelData = new TAnalysedPixelData( fPixelData );
  
  // --- calculate number of set pixels, needs some time ...
  
  TMirrorBitArray ba( fPixelData );  // full SLT range
  
  // ba.Dump();
  
  fEventSize = ba.GetNumberOfBits();
  
  // FD_COUT << "fEventSize= " << fEventSize << endl;
    
  // --- check how many pixels are really set (if < 440)
  
  if ( fTltParameters.fMaxEventSize < Fd::kMI_NPIXELS ) {
  
    if ( fEventSize > fTltParameters.fMaxEventSize ) {
      VTltProcessor::SetEventAccepted( false );
      fBigEvent = true;
      fUnlabelled = false;
      fNManyPixels++;
      return;
    }
  }
  
  // ---- set parameters of the (future) analyser ...
  
  if ( fTltParameters.fEnableMultipleRejection )
    fAnalysedPixelData->SetMaxNumSearchedPattern(
                         fTltParameters.fNumPatternToAnalyse );
  else
    fAnalysedPixelData->SetMaxNumSearchedPattern( 1 );
  
  fAnalysedPixelData->SetFirstSltBin( fTltParameters.fFirstSltBin );
  fAnalysedPixelData->SetLastSltBin( fTltParameters.fLastSltBin );
  
  // fAnalysedPixelData->PrintSettings();
  
  fAnalysedPixelData->Analyse();

  if ( fgPrintLevel > 1 )
    FD_CINFO << "TLT Level1: found " << fAnalysedPixelData->GetNPattern()
             << " pattern(s) and " << fAnalysedPixelData->GetNValidPattern()
             << " valid pattern(s) !" << endl;
  
  // --- set some flags to indicate the results ...
  
  if ( fAnalysedPixelData->HasMultiPattern() ) {
    fMulti = true;
    fUnlabelled = false;
  }

  TFADCData *fadc;
  Fd::PixelNumberRec pixel;
  unsigned int col, row;
  
  // --- loop over all found pattern to set up FADC list...
  
  if ( ( fNumPattern = fAnalysedPixelData->GetNValidPattern() ) != 0 ) {
  
    for ( unsigned int n=0; n<fAnalysedPixelData->GetNValidPattern(); n++ ) {
    
      PixelPattern *pattern = fAnalysedPixelData->GetValidPattern( n );
    
      // can this really happen or is there a fault in the analyser ?
      
      if ( (int)pattern->GetNPixel() < 4 ) {
        FD_CERR << "PrototypeTltProcessor::DoLevel1Decision() - "
	        << "not enough pixels in PixelPattern !" << endl;
      }
  
      unsigned int n_fadc_bins = TFADCData::GetNFADCBins(TFADCData::GetDefaultResolution());
      TFADCData::FADCTraceInfoRec trace_info;
      
#warning Could set limited trace, here !
      FADCTraceSetStartBin( &trace_info, 0 );
      FADCTraceSetEndBin( &trace_info, n_fadc_bins - 1 );
      
      for ( unsigned int i=0; i<pattern->GetNPixel(); i++ ) {
    
        pattern->GetColumnRow( i, col, row );

        Fd::SetPixelNoFromColumnRow( &pixel, col, row );
	
	fadc = fReadout->GetFADCData( &pixel, &trace_info );
        fAdcList.push_back( fadc ); // put on req.list

        if ( fgPrintLevel > 2 )
          FD_CDEBUG(1) << "Level1: Pixel(" << col << "," 
	               << row << ") is set !" << endl;
      }
      
      // ToDo: make sure that each pixel is only once in the list
      
      fAdcList.unique();
    }
  }
}

// --------------------------------------------------------------------------

void PrototypeTltProcessor::DoLevel2Decision()
 {
  list<TAnalysedFADCData*> ana_list;
  
  if ( fgPrintLevel > 1 )
    FD_CINFO << "PrototypeTltProcessor::DoLevel2Decision()" << endl;

  // Step 1: analyse all FADC traces which are present in the list
  
  TFADCData *fadc;
  list<TFADCData*>::const_iterator iter = fAdcList.begin();
  list<TFADCData*>::const_iterator end = fAdcList.end();
  
  TAnalysedFADCData* afadc;
  
  while ( iter != end ) {
   
    fadc = *iter;
    
    // todo: should use allocation pool
    
    afadc = new TAnalysedFADCData( fadc );
    
    afadc->Analyse();
    
    ana_list.push_back( afadc );
    
    iter++;
  }
  
  unsigned int col0, row0;
  unsigned int col, row;

  list<TAnalysedFADCData*>::const_iterator aiter;
  list<TAnalysedFADCData*>::const_iterator aend = ana_list.end();
  
  // Step 2: Assign pattern and FADC pulse times
  
  for ( unsigned int n=0; n<fAnalysedPixelData->GetNValidPattern(); n++ ) {
  
    PixelPattern *pattern = fAnalysedPixelData->GetValidPattern( n );
  
    for ( unsigned int i=0; i<pattern->GetNPixel(); i++ ) {
  
      pattern->GetColumnRow( i, col0, row0 );
      
      aiter = ana_list.begin();

      while ( aiter != aend ) {
   
        afadc = *aiter;
        Fd::PixelNumber pixel = afadc->GetPixelNumber();
      
        col = Fd::GetColumnNo( pixel );
        row = Fd::GetRowNo( pixel );
      
        if ( col == col0 && row == row0 ) {
	  AssignAdcTime( pattern, i, afadc );
	  break;
        }

        aiter++;
      }
    }
  
    if ( fgPrintLevel > 1 ) {
      pattern->Show();
    }
  
    // STEP 3: qualify the pattern
  
    QualifyPattern( pattern );  fNPatternAnalysed++;
    
    if ( fTltParameters.fEnableMultipleRejection ) {
      if ( n == fTltParameters.fNumPatternToAnalyse ) 
        break;
    }
    else {
      break;    // leave loop after the first pattern analysed
    }
  }
  
  // STEP 4: clear the temporary list

  aiter = ana_list.begin();
  
  while ( aiter != aend ) {

    afadc = *aiter;
    delete afadc;
    
    aiter++;
  }
  
  ana_list.clear();
}

// --------------------------------------------------------------------------

void PrototypeTltProcessor::PrintSettings(std::ostream& ostr)
 {
  time_t t = time( NULL );
  struct tm *tm = gmtime( &t );
  
  ostr << endl;
  ostr << "*** " << GetName() << " - your present settings@UTC " 
       << asctime( tm );
  ostr << endl;
  
  ostr << "\tfCoincidenceWindow=       " 
       << setw(3) << fTltParameters.fCoincidenceWindow << " * 100ns" << endl;
  ostr << "\tfEnableMultipleRejection= " 
       << setw(3) << fTltParameters.fEnableMultipleRejection << endl;
  ostr << "\tfNumPatternToAnalyse=     " 
       << setw(3) << fTltParameters.fNumPatternToAnalyse << endl;
  ostr << "\tfKeepRejectedFraction=    " 
       << setw(3) << fTltParameters.fKeepRejectedFraction << " %" << endl;

  ostr << "\tfFirstSltBin=             " 
       << setw(3) << fTltParameters.fFirstSltBin << endl;
  ostr << "\tfLastSltBin=              " 
       << setw(3) << fTltParameters.fLastSltBin << endl;

  ostr << "\tfMaxEventSize=            " 
       << setw(3) << fTltParameters.fMaxEventSize << " pixels" << endl;
  
  ostr << "\tfCutParameter1 (100ns/p)= "
       << fTltParameters.fCutParameter1 << endl;
  ostr << "\tfCutParameter2 (PropP)=   "
       << fTltParameters.fCutParameter2 << endl;

  ostr << endl;
  ostr.flush();
}

// --------------------------------------------------------------------------

/** Calculate the number of 100 ns bins per pixel. 
  * The calculated value is averaged over the number of interpixel intervals 
  * and its absolute value is then returned.
  *
  * Note: This is a very rough approximation of the next (more detailed) 
  * algorithm, but it allows also to remove some random pattern.
  */
float GetAveraged100NSPerPixel(unsigned int n_pulses, int *t_pulse)
 {
  if ( !n_pulses ) {
    FD_CERR << "GetAveraged100NSPerPixel() - no pulses !" << endl;
    return 0.;
  }
  
  float ns_per_pixel;
  
  // find first pixel with timing info
  
  ns_per_pixel = t_pulse[1] - t_pulse[0];
  
  for ( unsigned int i=2; i<n_pulses; i++ )
    ns_per_pixel +=   (float)t_pulse[i] - (float)t_pulse[i-1];
    
  ns_per_pixel = fabs( ns_per_pixel ) / 100.;
    
  ns_per_pixel /= (float)n_pulses;

  return ns_per_pixel;
}

// --------------------------------------------------------------------------

/** Calculate the propagation parameters according to Andreas' method.
  *
  * Results:
  * @li p_par[0] contains the value dr/dt
  * @li p_par[1] contains the deviation (error) of the 2nd and 3rd pixel
  * with respect to the line given by pixel 1 and 4.
  *
  * If the numbers of pulse are not sufficient, all resulting parameters
  * are set to 0.0 .
  */
void GetPropagationParameters(unsigned int n_pulses, 
                              int *x_pos, int *y_pos, int *t_pulse, 
			      float *p_par)
 {
  if ( n_pulses == 4 ) {
  
    int dt = t_pulse[3] - t_pulse[0];
    
    // calculate the total (begin to end) dr/dt value
    
    int dx = x_pos[3] - x_pos[0];
    int dy = y_pos[3] - y_pos[0];

    float dr = (float)sqrt( (float)(dx * dx + dy * dy) );
    float prop = dt / 100. / dr;
    
//    if ( fgPrintLevel > 1 ) {
//      FD_COUT << "dx= " << dx << ", dy= " << dy << ", dr= " << dr
//           << ", dr/dt= " << prop << endl;
//    }
    
    float t_diff = 0.;
    int dx_act, dy_act;
    float dr_act;
    
    for ( unsigned int i=1; i<n_pulses; i++ ) {

      // calculate the intermediate dr/dt values
      
      dx_act = x_pos[i] - x_pos[0];
      dy_act = y_pos[i] - y_pos[0];
      
      dr_act = (float)sqrt( (float)(dx_act * dx_act + dy_act * dy_act) );
      
      // time difference to the expected value (must be small)
      
      t_diff = t_diff + fabs( t_pulse[i] - 100. * prop * dr_act - t_pulse[0] );
    }
    
    t_diff /= 100.;

    prop *= 2.;   // propagation [100ns/pixel]
    t_diff /= 2.; // mean error  [100ns]
    
    p_par[0] = prop;
    p_par[1] = t_diff;
  }
  else {
    p_par[0] = 0.;
    p_par[1] = 0.;  
  }  // n_pulses = 4
}

// --------------------------------------------------------------------------

void PrototypeTltProcessor::LogResults(const char *where)
 {
  if ( !fTltLogfile ) return;  // logfile not open
  
  static int line_no = 0;

  // --- write symbol explanation ...
  
  if ( line_no == 0 ) {
    
    fprintf( fTltLogfile, 
             "========== PrototypeTltProcessor: Results evt.by evt. ==========\n" );
    
    fprintf( fTltLogfile, "\tKEEP  - event accepted by Tlt1/2\n" );
    fprintf( fTltLogfile, "\tNP    - number of pattern found in this evt.\n" );
    fprintf( fTltLogfile, "\tUnl   - no EventLabel assigned\n" );
    fprintf( fTltLogfile, "\tLP    - not enough matching pulses (SLT-FLT)\n" );
    fprintf( fTltLogfile, "\tMult  - >=1 multi-pattern detected\n" );
    fprintf( fTltLogfile, "\tRand  - random time order detected\n" );
    fprintf( fTltLogfile, "\tNPro  - time-space order is not proportional\n" );
    fprintf( fTltLogfile, "\tBigE  - number of triggered pixel > cut\n" );
    fprintf( fTltLogfile, "\tShoC  - pattern is shower candidate (CP1/2)\n" );
    fprintf( fTltLogfile, "\tSize  - number of triggered pixels\n" );
    fprintf( fTltLogfile, "\tCP1   - result of cut parameter1 evaluation\n" );
    fprintf( fTltLogfile, "\tCP2   - result of cut parameter2 evaluation\n" );

    fprintf( fTltLogfile, 
             "................................................................\n" );
  }
  
  // --- write header line each 20 lines ...
  
  if ( (line_no % 20) == 0 )
    fprintf( fTltLogfile, 
             "Tlt - KEEP  NP  Unl  LP  Mult Rand NPro BigE ShoC Size  CP1  CP2  CP2err\n" );
  
  char where_str[5];
  snprintf( where_str, sizeof(where_str), "%4s", where );
  
  fprintf( fTltLogfile,
           "%s: %3d  %3d  %3d  %3d  %3d  %3d  %3d  %3d  %3d  %3d %4.2f %4.2f %4.2f\n",
	   where_str, 
	   (int)VTltProcessor::IsEventAccepted(),
	   fNumPattern, 
           (int)fUnlabelled, (int)fLessPulses, (int)fMulti, (int)fRandom,
	   (int)fNotProp, (int)fBigEvent, (int)fShowerC,
	   fEventSize, fCut1Result, fCut2Result, fCut2Error );
  
  line_no++;
  
  fflush( fTltLogfile );
}

// --------------------------------------------------------------------------

void PrototypeTltProcessor::PrintFinalResults(std::ostream& ostr)
 {
  ostr << endl;
  ostr << "============= PrototypeTltProcessor: Final results =============" << endl;

  ostr << "calls/(real/CPU) time:  " << setw(5) << fStopwatch.Counter() << " "
       << setprecision(4) << fStopwatch.RealTime() << " " 
       << setprecision(4) << fStopwatch.CpuTime() << endl;
  ostr << "<Time>/call(real/CPU): " << "      "
       << setprecision(4) 
       << fStopwatch.RealTime()/fStopwatch.Counter() 
       << " " << setprecision(4) 
       << fStopwatch.CpuTime()/fStopwatch.Counter() << endl;

  ostr << "TLT Statistic: " << endl;
  ostr << "Number of events analysed:   " << fNEventsAnalysed << endl;
  ostr << "Number of events rejected:   " << fNEventsRejected << endl;
  ostr << "Number of rej. evts kept:    " << fNRejectedKept << endl;
  
  ostr << endl;
  ostr << "Number of pattern analysed:  " << fNPatternAnalysed << endl;
  ostr << "- no pulses found:           " << fNError << endl;
  ostr << "- not enough pulses:         " << fNLessPulses << endl;
  ostr << "- too many pixels:           " << fNManyPixels << endl;
  ostr << "- not proportional times:    " << fNNotProp << endl;
  ostr << "- number of pulses mismatch: " << fNPulseMismatch << endl;
  ostr << "- random time order:         " << fNRandomTime << endl;
  ostr << "- shower candidates:         " << fNShowerC << endl;
  
  ostr << "================================================================" << endl;
  
  ostr.flush();
}

// --------------------------------------------------------------------------

void PrototypeTltProcessor::QualifyPattern(PixelPattern *pattern)
 {
  // STEP 0: write everything to local variables
  //         and count the number of detected pulses
  
  int t_pulse[4];
  int x_pos[4], y_pos[4];  // position in artificial coordinates
  
  unsigned int n_pulses = 0;
  
  for ( unsigned int i=0; i<pattern->GetNPixel(); i++ ) {
    if ( pattern->GetAdcTime( i ) ) {
      x_pos[n_pulses] = pattern->GetX( pattern->GetColumn( i ), 
                                       pattern->GetRow( i ) );
      y_pos[n_pulses] = pattern->GetY( pattern->GetColumn( i ), 
                                       pattern->GetRow( i ) );
      t_pulse[n_pulses] = pattern->GetAdcTime( i );
      n_pulses++;
    }
  }
  
#if 0
  // STEP 1: check for easy detectable muons (using pulse gradient)
  
  // ... must be done !
  
  int n_muon = 0;
  
  
  if ( n_muon == 4 ) {
    fNMuon4Ev++;
  }
  else if ( n_muon == 3 ) {
    fNMuon3Ev++;
  }
#endif

  // STEP 2: check the number of pulses found, try a first classification
  
  if ( n_pulses < 4 )
    fNPulseMismatch++;

  if ( n_pulses == 0 ) {  // no pulses found -> reject
    if ( fgPrintLevel ) FD_CWARN << "No pulses found !" << endl;
    fUnlabelled = false;
    fLessPulses = true;
    fNError++;
    return;
  }

  if ( n_pulses == 1 ) { // only one pulse found -> reject
    fUnlabelled = false;
    fLessPulses = true;
    return;
  };
  
  if ( n_pulses == 2 ) { // two pulse found -> reject ?
  }
  
  // STEP 3: check random triggered pixel  -> no selection power ???
#if 0  
  bool random = false;

  for ( unsigned int i=1; i<pattern->GetNPixel(); i++ ) {
  }
#endif
  
  // STEP 4: check time order  -> no selection power ???
#if 0
  int n_time_error = 0;
  
  for ( unsigned int i=1; i<n_pulses; i++ ) {
    if ( abs( t_pulse[i] - t_pulse[i-1] ) > 5 ) {
      n_time_error++;
    }
  }
  
  // Note:
  // In the original code of Andreas here tPulse is replaced by a tStart
  // which is obviously the time of the pulse maximum
  // tPulse is then (tStart + xMin)/2 where xMin is the end time of the pulse
  
  if ( n_time_error > 1 ) {
    fNRandomTime++;
    // fUnlabelled = false;
    // fRandom = true;
  }
#endif
  
  // STEP 5: check time propagation of fine patterns (3-4 pulses)

  if ( n_pulses >= 3 && fUnlabelled ) {
    
    float ns_per_pixel = GetAveraged100NSPerPixel( n_pulses, t_pulse );
    
    if ( fgPrintLevel > 1 ) {
      FD_CINFO << "<100ns/pixel> = " << ns_per_pixel << endl;
    }
      
    if ( ns_per_pixel < fTltParameters.fCutParameter1 ) {
      fNotProp = true;
    }
    
    if ( ns_per_pixel >= fTltParameters.fCutParameter1 ) {
      fShowerC = true;
    }

    // this is for the logging ...
    
    if ( ns_per_pixel > fCut1Result ) fCut1Result = ns_per_pixel;
  }

  if ( n_pulses >= 3 ) {
//  if ( n_pulses >= 4 ) {
    
    float p_par[2];
    
    GetPropagationParameters( n_pulses, x_pos, y_pos, t_pulse, p_par );
    
    if ( fgPrintLevel > 1 ) {
      FD_CINFO << "dr/dt= " << p_par[0] << ", ddt= " << p_par[1] << endl;
    }
          
    if (    ( fabs( p_par[0] ) < fTltParameters.fCutParameter2 ) 
         && ( p_par[1] < 1. ) ) {
      fNotProp = true;
    }
    
    // possible showers ...
    
    if (    ( fabs( p_par[0] ) >= fTltParameters.fCutParameter2 )
         && ( p_par[1] < 1. ) ) {
      fShowerC = true;
    }
    
    // this is for the logging ...
    
    if ( fabs( p_par[0] ) > fCut2Result ) {
      fCut2Result = fabs( p_par[0] );
      fCut2Error = p_par[1];
    }
    
  }  // n_pulses >= 3
  
    
  // STEP 6: now do some final counting ...
  
  if ( fNotProp ) {
    fUnlabelled = false;
    fNNotProp++;
  }
  
  if ( fShowerC ) {
    fUnlabelled = false;
    fShowerC = true;
    fNShowerC++;
  }
}

// --------------------------------------------------------------------------

bool PrototypeTltProcessor::ReadParameters(FdUtil::Inifile* inifile,
                                           MiReadout::TltParameters par,
					   std::string section)
 {
  // --- some checks ...
  
  if ( !inifile ) {
    FD_CERR << "PrototypeTltProcessor::ReadParameters() - "
            << "Inifile '*inifile' = NULL!" << endl;
    return false;
  }
  
  if ( !par ) {
    FD_CERR << "PrototypeTltProcessor::ReadParameters() - "
            << "TltParameters 'par' = NULL!" << endl;
    return false;
  }
  
  if ( section.empty() ) {
    FD_CERR << "PrototypeTltProcessor::ReadParameters() - "
            << "string 'section' is empty!" << endl;
    return false;
  }
  
  // --- let's do it
  
  if ( inifile->SpecifyGroup( section.c_str() ) != FdUtil::Inifile::kSUCCESS ) {
    FD_CERR << "!!! PrototypeTltProcessor::Section [" << section << "] not found !!!" << endl;
    return false;
  }
  
  FD_CINFO << "\tReading section [" << section << "] ..." << endl;

  par->fCoincidenceWindow = inifile->GetFirstValue( "coincidence", 15 );
  par->fKeepRejectedFraction = inifile->GetFirstValue( "keepreject", 100 );
  par->fFirstSltBin = inifile->GetFirstValue( "firstsltbin", 0 );
  par->fLastSltBin = inifile->GetFirstValue( "lastsltbin", 99 );
  par->fMaxEventSize = inifile->GetFirstValue( "maxpixel", 70 );
  par->fCutParameter1 = inifile->GetFirstFloat( "cutpar1", 0.7 );
  par->fCutParameter2 = inifile->GetFirstFloat( "cutpar2", 0.9 );
  
  return true;
}

// --------------------------------------------------------------------------

void PrototypeTltProcessor::SetParameters(const TltParameters tlt)
 {
  assert( tlt != NULL );
  
  fTltParameters = *tlt;
  
  if ( !fTltLogfile ) return;
  
  FdUtil::OstdioStream ostr( fTltLogfile );

  PrintSettings( ostr );

  fprintf( fTltLogfile, "\n" );
  
  fflush( fTltLogfile );
}

// --------------------------------------------------------------------------
// --------------------------------------------------------------------------
