
//
// File   : PrototypeTltProcessor.hh
//
// Purpose: Declaration of class PrototypeTltProcessor
//
// $Id: PrototypeTltProcessor.hh 11881 2011-01-13 09:00:27Z mathes $
//


/** @file PrototypeTltProcessor.hh
  * Declaration of class PrototypeTltProcessor.
  * @author H.-J. Mathes, Fzk
  */

#ifndef _PrototypeTltProcessor_hh_ 
#define _PrototypeTltProcessor_hh_ 


#include <iostream>
#include <list>
#include <string>
#include <cstdio>

#include <MiEvent.hh>
#include <MiReadout.hh>

#include <TStopwatch.h>

// --- forward declaration(s)

class VMirrorEvent;
class TAnalysedFADCData;
class TAnalysedPixelData;
class PixelPattern;
class FEReadout;

namespace FdUtil {
  class Inifile;
}

#include "VTltProcessor.hh"

/** This class implements a combined TLT algorithm of A. Kopmann and 
  * H.-J. Mathes which is/was used for the AUGER EA (prototype) phase..
  *
  * This algorithm is controlled by a set of parameters contained in the struct 
  * TltParametersRec.
  *
  * General:
  * @li the decision making is based upon two steps
  * @li the first step takes only the SLT data into account, the parameters
  *     fFirstSltBin and fLastSltBin set the limits for the search in the SLT
  *     data
  * @li the second step refines the analysis of the found pattern and uses also
        the ADC pulses for its decision making
  *
  * Strategy:
  * @li if the number of triggered pixels within the search time interval
  *     exceeds the parameter fMaxEventSize the event is rejected
  * @li up to fNumPatternToAnalyse pattern are searched for. If the flag
  *     fEnableMultipleRejection is set to 'true' (the default) this search
  *     will consider also 0x7f pattern
  * @li a list of pixels whose FADC data is to be read out is compiled from the
  *     pixels contained in the found pattern
  * @li after the readout of the requested FADC data, the FADC traces are
  *     analysed for pulses
  * @li the pulse which matches best the time of the trigger in SLT is chosen 
  *     to be the time of the pulse and stored together with the other 
  *     information about the current pattern
  * @li depending on the number of pulses found in the FADC data and other
  *     features, a classification of all patterns is done.
  * @li each specific feature of an analysed pattern sets a specific flag. 
  *     Afterwards it is checked, if there any signature has been found which
  *     is a hint for a shower candidate
  *
  * Decision features (method QualifyPattern()):
  * <br> Note: All decision steps are commented in the source code !!!<br>
  * @li if only 0 or 1 pulses are found, the pattern is rejected
  * @li if the pulse times are too different, they might be random pulses
  * @li if the value '<time per pixel>' is too small, the pattern is rejected
  * @li the pattern is rejected if its propagation parameters (or their errors)
  *     are outside the allowed range
  *
  * Further notes:
  * @li pulse finding might be improved (both in speed and in reliability) by
  *     using another pulse finder or by limiting the range where the pulse is
  *     searched for.
  */
class PrototypeTltProcessor : public VTltProcessor {
public:
    /** Constructor for the class PrototypeTltProcessor. 
      *
      * The following things are done:
      * @li the internal working variables/objects are created
      * @li the appropiate pulse finder is chosen and parametrized
      * @li all counting and flag variables are initialized
      * @li if parameters are passed, they are used for initializing, else
      *     the processor is parametrized with the built-in default values
      */
    PrototypeTltProcessor(MiReadout::TltParameters tlt=NULL);

    /** Destructor of the class PrototypeTltProcessor.
      */
    ~PrototypeTltProcessor();

    /** Clear internal data structures to be prepared for the next Event.
      *
      * @li flag variables are reset
      * @li buffer variables are re-initialized
      */
    void Clear();

    /** Perform the TLT decision. 
      *
      * This decision is made in two steps:
      * @li first a decision is tried based on the passed TMirrorPixelData
      * @li if the first step accepts the event, it might request a list of
      *     FADC traces in orer to improve the decision.
      */
    void Decide(VMirrorEvent*);

    /** Print the current parameters settings for this processor to
      * the ostream passed (by default this is cout).
      */
    void PrintSettings(std::ostream& ostr=std::cout);
    
    /** Read the TltParameters from the passed FdUtil::Inifile.
      *
      * @return 'true' on success, else 'false'.
      */
    static bool ReadParameters(FdUtil::Inifile*,
                MiReadout::TltParameters,std::string section="TltReadout");
    
    /** Set the parameters for processing and deciding of this processor.
      */
    virtual void SetParameters(const MiReadout::TltParameters);
    
    /** Set the level of verbosity for all instances of this class.
      *
      * Note: Operates on the base class !
      */
    static void SetPrintLevel(int level)
     { fgPrintLevel = level; }

protected:

    /** Assign the bext matching time of the FADC trace to the selected pixel 
      * of the pattern.
      */
    void AssignAdcTime(PixelPattern*,unsigned int,TAnalysedFADCData*);
    
    /** Do the first part of the TLT decision based on the SLT data only. Here
      * we migt decide that the second part is not needed, i.e. the event could
      * be rejected. In the case the event could not rejected, a list of 
      * required ADC traces is compiled which is then requested from the
      hardware (by the steering class).
      */
    void DoLevel1Decision();

    /** Do the second part of the TLT decision based on the SLT data and the 
      * FADC data of selected pixels.
      */
    void DoLevel2Decision();
    
    /** Write the result of the current decision step to the log file. A short
      * label which is used as a prefix of the log message could be passed.
      */
    void LogResults(const char *);

    /** Print the final statistics of this processor.
      */
    void PrintFinalResults(std::ostream& ostr=std::cout);
    
    /** Analyse the specified pattern using the PixelData and the ADC data
      * of all pixels contained in this pattern.
      */
    void QualifyPattern(PixelPattern*);

    /** Set the coincidence window for the FLT-SLT matching.
      */
    void SetCoincidenceWindow(int window);
    
    /** The parameters of this TLT algorithm. */
    MiReadout::TltParametersRec fTltParameters;

private:

    PrototypeTltProcessor(const PrototypeTltProcessor&);
    PrototypeTltProcessor& operator=(const PrototypeTltProcessor&);
    
    static const int       kNFadcObjects = 4;
    
    std::list<TFADCData*>  fAdcList;
    
    TAnalysedPixelData    *fAnalysedPixelData;

    TMirrorPixelData      *fPixelData;
    TMirrorBitArray       *fPixelMatrix;
        
    TStopwatch             fStopwatch;
    
    VMirrorEvent          *fReadout;
    
    // --- counting and flag variables
    
    bool                   fBigEvent;             // a too big event (size)
    bool                   fLessPulses;           // not enough pulses
    bool                   fMulti;
    bool                   fMuon;                 // pulse form says 'muon'
    bool                   fNotProp;
    bool                   fRandom;
    bool                   fShowerC;
    bool                   fUnlabelled;
    
    unsigned int           fNEventsAnalysed;      // number of events anal.
    unsigned int           fNEventsRejected;      // number of events rejected
    unsigned int           fNRejectedKept;        // number of rejected events
                                                  // kept for monitoring purposes

    unsigned int           fNPatternAnalysed;     // number of pattern analysed
    
    unsigned int           fNError;               // number of errors occurred
    unsigned int           fNLessPulses;          // not enough pulses
    unsigned int           fNManyPixels;          // too many pixels
    unsigned int           fNMulti;               // multi pattern
    unsigned int           fNNotProp;             // time-space not prop.
    unsigned int           fNPulseMismatch;       // insufficeint number of p.
    unsigned int           fNRandomTime;          // random pattern
    unsigned int           fNShowerC;             // shower candiates
    
                                                  // analysis results
    unsigned int           fEventSize;            // number of pixels set
    float                  fCut1Result;           // keeps largest 100ns/Pixel
    float                  fCut2Result;           // keeps dr/dt
    float                  fCut2Error;            // keeps error of dr/dt
    unsigned int           fNumPattern;           // number of pattern in event
};

#endif  // _PrototypeTltProcessor_hh_ 
