
//
// File   : ReadoutConfig.cc
//
// Purpose: Implementation of the class ReadoutConfig
//
// $Id: ReadoutConfig.cc 9725 2010-09-28 07:31:17Z mathes $
//


#include <iostream>
#include <iomanip>
#include <cstring>
#include <ctype.h>
#include <string>
#include <cassert>


/** @file ReadoutConfig.cc
  * Implementation of the class ReadoutConfig
  * @author H.-J. Mathes, FzK
  */


#include <FdNaming.hh>
#include <ErrorLogging.hh>
#if (FDUTIL_VERSION_CODE >= ModuleVersionCode(4,1,0))
# include <InifileChain.hh>
#else
# include <ChainedInifile.hh>
#endif // FDUTIL_VERSION_CODE

#include <PrototypeTltProcessor.hh>

#include "MiReadout.hh"
#include "ReadoutConfig.hh"

using namespace std;
using namespace MiReadout;

using FdUtil::Fd;
using FdUtil::ErrorLogger;

// -------------------------------------------------------------------------
// -------------------------------------------------------------------------
//
// class ReadoutConfig - implementation
//

ReadoutConfig *ReadoutConfig::fgReadoutConfig = NULL;

ReadoutConfig::ReadoutConfig(const char* ini_name, const char* /* profile */)
 : fInifile(NULL), fInifileName(""), fProfileName("")
 {
  // --- set (compile time) default parameters
  
  SetDefaultConfiguration();

  // --- try to open an ini-file
  
#if 0
  fInifileName = GetCompleteInifileName( ini_name, fProfileName.c_str() );
#endif

  // --- open the configuration for reading ...
  
  OpenConfiguration( ini_name /* fInifileName.c_str() */, 'R' );
  
  // --- now read the various sections of the inifile
  
  LoadConfiguration();
}

// -------------------------------------------------------------------------

ReadoutConfig* ReadoutConfig::GetInstance(const char *config_name,
                                          const char *profile)
 {
  if ( fgReadoutConfig == NULL )
    fgReadoutConfig = new ReadoutConfig( config_name, profile );
  
  return fgReadoutConfig;
}

// -------------------------------------------------------------------------

ReadoutConfig::~ReadoutConfig()
 {
  CloseConfiguration();
  
  delete fInifile;
}

// -------------------------------------------------------------------------

void ReadoutConfig::CloseConfiguration()
 {
#warning ReadoutConfig::CloseConfiguration() not implemented !
  FD_CWARN << "ReadoutConfig::CloseConfiguration() not implemented !" << endl;
}

// -------------------------------------------------------------------------

void ReadoutConfig::DumpFltParameters(const MiReadout::FltParametersRec& par, 
                                      const char *where, ostream& ostr)
 {
  ostr << endl;
  ostr << "===== DumpFltParameters " 
       << ((where != NULL && *where != 0) ? where : "") << "=====" << endl;
  ostr << endl;
  
  ostr << "\tfStatisticEnabled=  " << (int)par.fStatisticEnabled << endl;
  ostr << "\tfStatisticOffset=   " << par.fStatisticOffset << endl;
  ostr << "\tfStatisticNSamples= " << par.fStatisticNSamples << endl;
  ostr << "\tfIntegrationTime=   " << par.fIntegrationTime << endl;
  ostr << "\tfCoincidenceTime=   " << par.fCoincidenceTime 
                                   << ", not read from ini-file !" << endl;
  ostr << "\tfNFADCBins=         " << par.fNFADCBins << endl;
  
  ostr << endl;
  ostr.flush();
}

// -------------------------------------------------------------------------

void ReadoutConfig::DumpGPSClockConfig(const MiReadout::GPSClockConfigRec& gps, 
                                       const char *where, ostream& ostr)
 {
  ostr << endl;
  ostr << "===== DumpGPSClockConfig " 
       << ((where != NULL && *where != 0) ? where : "") << " =====" << endl;
  ostr << endl;
  
  ostr << "\tfStrobeMode=  " 
       << ((gps.fStrobeMode == kGPSStrobeGPS) ? "gps" : "local") << endl;
  ostr << "\tfTimeScale=   " 
       << ((gps.fTimeScale == kGPSTimeScaleGPS) ? "GPS" : "UTC") << endl;
  
  ostr << endl;
  ostr.flush();
}

// -------------------------------------------------------------------------

void ReadoutConfig::DumpNeighborTriggerParameters(const char *where, 
                                                  ostream& ostr)
 {
  //DumpSltParameters( fSltParameters["NeighborTriggered"], "NeighborTriggered" );
  //DumpTltParameters( fTltParameters["NeighborTriggered"], "NeighborTriggered" );
  
  ostringstream where_str;
  
  if ( where != NULL && *where != 0 )
    where_str << "'" << where << "'";

  ostr << endl;
  ostr << "===== DumpNeighborParameters " << where_str.str() << " =====" << endl;
  ostr << endl;
  
  SltParametersRec& slt = fSltParameters["NeighborTriggered"];
  TltParametersRec& tlt = fTltParameters["NeighborTriggered"];

  ostr << "\tfMaxEventSize=            " 
       << setw(4) << tlt.fMaxEventSize << " pixels" << endl;
  ostr << "\tfKeepRejectedFraction=    " 
       << setw(4) << tlt.fKeepRejectedFraction << " %" << endl;
  
  ostr << endl;
  ostr.flush();
}

// -------------------------------------------------------------------------

void ReadoutConfig::DumpSltParameters(const MiReadout::SltParametersRec& par, 
                                      const char *where, ostream& ostr)
 {
  ostringstream where_str;
  
  if ( where != NULL && *where != 0 )
    where_str << "'" << where << "'";

  ostr << endl;
  ostr << "===== DumpSltParameters " << where_str.str() << " =====" << endl;
  ostr << endl;
  
  ostr << "\tfVersion=       " << par.fVersion << endl;
  ostr << "\tfStartBin=      " << par.fStartBin << endl;
  ostr << "\tfEndBin=        " << par.fEndBin << endl;
  ostr << "\tfDoFullReadout= " << (int)par.fDoFullReadout << endl;
  ostr << "\tfDoTlt=         " << (int)par.fDoTlt << endl;
  ostr << "\tfNExtColumns=   " << par.fNExtColumns << endl;
  ostr << "\tfNExtRows=      " << par.fNExtRows << endl;
  ostr << "\tfMinDistance=   " << par.fMinDistance << endl;
  ostr << "\tfFirstSltBin=   " << par.fFirstSltBin << endl;
  ostr << "\tfLastSltBin=    " << par.fLastSltBin << endl;
//  ostr << "\tfMinEventSize=  " << par.fMinEventSize << endl;
  ostr << "\tfMaxEventSize=  " << par.fMaxEventSize << endl;
  
  ostr << endl;
  ostr.flush();
}

// -------------------------------------------------------------------------

void ReadoutConfig::DumpTltParameters(const MiReadout::TltParametersRec& par, 
                                      const char *where, ostream& ostr)
 {
  ostringstream where_str;
  
  if ( where != NULL && *where != 0 )
    where_str << "'" << where << "'";

  ostr << endl;
  ostr << "===== DumpTltParameters " << where_str.str() << " =====" << endl;
  ostr << endl;
  
  ostr << "\tfCoincidenceWindow=       " 
       << setw(4) << par.fCoincidenceWindow << " * 100ns" << endl;
  ostr << "\tfEnableMultipleRejection= " 
       << setw(4) << par.fEnableMultipleRejection << endl;
  ostr << "\tfNumPatternToAnalyse=     " 
       << setw(4) << par.fNumPatternToAnalyse << endl;
  ostr << "\tfKeepRejectedFraction=    " 
       << setw(4) << par.fKeepRejectedFraction << " %" << endl;

  ostr << "\tfFirstSltBin=             " 
       << setw(4) << par.fFirstSltBin << endl;
  ostr << "\tfLastSltBin=              " 
       << setw(4) << par.fLastSltBin << endl;

  ostr << "\tfMaxEventSize=            " 
       << setw(4) << par.fMaxEventSize << " pixels" << endl;
  
  ostr << "\tfCutParameter1 (100ns/p)=  "
       << par.fCutParameter1 << endl;
  ostr << "\tfCutParameter2 (PropP)=    "
       << par.fCutParameter2 << endl;

  ostr << endl;
  ostr.flush();
}

// -------------------------------------------------------------------------

#if 0
string ReadoutConfig::GetCompleteInifileName(const char *ini_name, 
                                             const char *profile)
 {
  // cut the suffix, if existing (from the last point to the end)
  string suffix;

  const char *last_dot;
  if ( strstr( ini_name, "." ) != NULL ) {
    for ( int i=strlen(ini_name) - 1; i>=0 ; i-- ) {
      if ( ini_name[i] == '.' ) {
        last_dot = &ini_name[i];
	break;
      }
    }
    suffix = last_dot;
  }
  
  if ( suffix.length() <= 1 ) suffix = ".ini";
  
  cout << "Suffix= " << suffix << endl;
  
  string base;
  
  for ( const char *ch=ini_name; ch<last_dot; ch++ )
    base += *ch;

  cout << "Base= " << base << endl;
  
  string name;
  
  if ( profile && *profile != 0 ) {
    name = base;
    name += "#";
    name += profile;
    name += suffix;
  }
  else
    name = ini_name;
    
  return name;
}
#else
# warning Tbd: Profiles are not supported !
#endif

// -------------------------------------------------------------------------

MiReadout::SltParameters 
ReadoutConfig::GetSltParameters(const std::string& config)
 {
  // assume that 'Default' config is already in the map
  if ( config.length() == 0 ) {
    return &fSltParameters["Default"];
  }
  // try to read it from the ini-file
  else if ( fSltParameters.find(config) == fSltParameters.end() ) {
  
    SltParametersRec slt_par = DEFAULT_SLT_PARAMETERS;
    
    ReadSectionFullReadout( slt_par, config );
    ReadSectionSelectiveReadout( slt_par, config );
    
    fSltParameters[config] = slt_par;
  }
  
  return &fSltParameters[config];
}

// -------------------------------------------------------------------------

MiReadout::TltParameters 
ReadoutConfig::GetTltParameters(const std::string& config)
 {
  // assume that 'Default' config is already in the map
  if ( config.length() == 0 ) {
    return &fTltParameters["Default"];
  }
  // try to read it from the ini-file
  else if ( fTltParameters.find(config) == fTltParameters.end() ) {
    TltParametersRec tlt_par = DEFAULT_TLT_PARAMETERS;
    
    ReadSectionTltReadout( tlt_par, config );
    
    fTltParameters[config] = tlt_par;
  }
  
  return &fTltParameters[config];
}

// -------------------------------------------------------------------------

// assume that configuration is already open
void ReadoutConfig::LoadConfiguration()
 {
  if ( !fInifile || (fInifile->GetStatus() == FdUtil::Inifile::kFAIL) ) {
    FD_CERR << "ReadoutConfig::LoadConfiguration() - "
            << "Could not open/read any ini-file !" << endl
	    << "  --> using default values !" << endl;
    return;
  }
  
  ReadSectionGeneral();    // general section
  
  // --- now read the selected sections from the inifile
  
  ReadSectionFullReadout( fSltParameters["Default"] );      // ReadoutParameters (I)
  ReadSectionSelectiveReadout( fSltParameters["Default"] ); // ReadoutParameters (I)
  
  ReadSectionTltReadout( fTltParameters["Default"] );       // TLT parameters

  fSltParameters["NeighborTriggered"] = fSltParameters["Default"];   // use SLT defaults
  fTltParameters["NeighborTriggered"] = fTltParameters["Default"];   // use TLT defaults
  
  if ( ReadSectionNeighbor( fTltParameters["NeighborTriggered"] ) ) {
    //ReadSectionFullReadout( fSltParameters["NeighborTriggered"], "NeighborTriggered" );
    ReadSectionSelectiveReadout( fSltParameters["NeighborTriggered"], "NeighborTriggered" );
  }
  else
    ReadSectionNeighbor( fTltParameters["NeighborTriggered"], "Neighbor" ); // old section ...
  
  DumpNeighborTriggerParameters();
  
  ReadSectionMStcCutTlt();
  
  ReadSectionGPSClock( fGpsClockConfig );     // hardware setup: GPS clock init

  ReadSectionFltStatistic( fFltParameters );  // hardware setup: FLT statistic
}

// -------------------------------------------------------------------------

void ReadoutConfig::OpenConfiguration(const string& name, const char option)
 {
  string conf_name(name);
  
  if ( conf_name.length() == 0 ) conf_name = "readout.ini";
  
  if ( toupper( option ) == 'W' ) {
    FD_CINFO << "ReadoutConfig::OpenConfiguration() - open for writing !" << endl;
  }
  else {
    FD_CINFO << "ReadoutConfig::OpenConfiguration() - open for reading !" << endl;
    
#if (FDUTIL_VERSION_CODE >= ModuleVersionCode(4,1,0))
    fInifile = new FdUtil::InifileChain( name.c_str() );
#else
    fInifile = new FdUtil::ChainedInifile( name.c_str() );
#endif // FDUTIL_VERSION_CODE
    
#ifndef DPA
    // chain with FDcommon.ini if not DPA version
    
    FdUtil::FdNaming& naming = FdUtil::FdNaming::GetInstance();
    
    // only one file at Eye level !
    string common_ini = naming.GetEyeConfigDirectory();
    common_ini += "FDcommon.ini";
    
    fInifile->AddInifile( common_ini.c_str() );
#endif // DPA
  }
}

// -------------------------------------------------------------------------

bool ReadoutConfig::ReadSectionFltStatistic(MiReadout::FltParametersRec& par)
 {
  if ( fInifile->SpecifyGroup( "FltStatistic" ) != FdUtil::Inifile::kSUCCESS ) {
  
    FD_CERR << "!!! ReadoutConfig::Section [FltStatistic] not found !!!" << endl;
    
    return false;
  }
  
  MiReadout::FltParametersRec default_fltpar = DEFAULT_FLT_PARAMETERS;
  
  FD_CINFO << "\tReading section [FltStatistic] ..." << endl;

  par.fStatisticEnabled  = 
      fInifile->GetFirstValue( "enable", default_fltpar.fStatisticEnabled );
  par.fStatisticOffset   = 
      fInifile->GetFirstValue( "offset", default_fltpar.fStatisticOffset );
  par.fStatisticNSamples = 
      fInifile->GetFirstValue( "samples", default_fltpar.fStatisticNSamples );
  par.fIntegrationTime   = 
      fInifile->GetFirstValue( "integration", default_fltpar.fIntegrationTime );
  par.fNFADCBins         =
      fInifile->GetFirstValue( "fadcbins", default_fltpar.fNFADCBins );
  
  // perform a rough check of the NFADC bins parameter
  if (    par.fNFADCBins != Fd::kMI_NFADC_BINS_LOW
       && par.fNFADCBins != Fd::kMI_NFADC_BINS_MED
       /* && par.fNFADCBins != Fd::kMI_NFADC_BINS_HIGH */ ) {
    
    FD_CERR << "ReadoutConfig::ReadSectionFltStatistic() - "
            << "the parameter 'fadcbins', value= "
	    << par.fNFADCBins << " has none of the allowed values!" << endl;
  
    if ( par.fNFADCBins <= Fd::kMI_NFADC_BINS_LOW )
      par.fNFADCBins = Fd::kMI_NFADC_BINS_LOW;
    else if ( par.fNFADCBins <= Fd::kMI_NFADC_BINS_MED )
      par.fNFADCBins = Fd::kMI_NFADC_BINS_MED;
//    else if ( par.fNFADCBins <= Fd::kMI_NFADC_BINS_HIGH )
//      par.fNFADCBins = Fd::kMI_NFADC_BINS_HIGH;
    else 
      par.fNFADCBins = Fd::kMI_NFADC_BINS_MED;
  
    FD_CWARN << "ReadoutConfig::ReadSectionFltStatistic() - "
             << "set parameter 'fadcbins' to " << par.fNFADCBins << endl;
  }
    
  DumpFltParameters( par, "", cout );
  
  return true;
}

// -------------------------------------------------------------------------

bool ReadoutConfig::ReadSectionFullReadout(MiReadout::SltParametersRec& par,
                                           const std::string& section)
 {
  assert( section.length() != 0 );
  
  if ( fInifile->SpecifyGroup( section.c_str() ) != FdUtil::Inifile::kSUCCESS ) {
    FD_CERR << "!!! ReadoutConfig::Section [" << section 
            << "] not found !!!" << endl;
    
    return false;
  }
  
  SltParametersRec default_sltpar = DEFAULT_SLT_PARAMETERS;
  
  FD_CINFO << "\tReading section [" << section << "] ..." << endl;
  
  par.fDoFullReadout = fInifile->GetFirstValue( "full", default_sltpar.fDoFullReadout );
  par.fDoTlt         = fInifile->GetFirstValue( "tlt", default_sltpar.fDoTlt );
  
  par.fStartBin = fInifile->GetFirstValue( "firstbin", default_sltpar.fStartBin );
  par.fEndBin   = fInifile->GetFirstValue( "lastbin", default_sltpar.fEndBin );

  return true;
}

// -------------------------------------------------------------------------

bool ReadoutConfig::ReadSectionGeneral()
 {
  const char * section = "General";
  
  if ( fInifile->SpecifyGroup( section ) != FdUtil::Inifile::kSUCCESS ) {
    FD_CERR << "!!! ReadoutConfig::Section [" << section 
            << "] not found !!!" << endl;

    return false;
  }
  
  FD_CINFO << "\tReading section [" << section << "] ..." << endl;
  
  fLoggingLevel = (ErrorLogger::ESeverityLevel)fInifile->GetFirstValue( "loglevel", 0 );
  
  ErrorLogger& theLogger = ErrorLogger::GetInstance();
  
  theLogger.SetLoggingLevel( fLoggingLevel );
  
  SltParameters slt_par = &fSltParameters["Default"];
  
  slt_par->fDoFullReadout = fInifile->GetFirstValue( "full", slt_par->fDoFullReadout );
  slt_par->fDoTlt = fInifile->GetFirstValue( "tlt", slt_par->fDoTlt );

  return true;
}

// -------------------------------------------------------------------------

bool ReadoutConfig::ReadSectionGPSClock(MiReadout::GPSClockConfigRec& par)
 {
  const char *section = "GPSClock";
  
  if ( fInifile->SpecifyGroup( section ) != FdUtil::Inifile::kSUCCESS ) {

    FD_CERR << "!!! ReadoutConfig::Section [" << section 
            << "] not found !!!" << endl;
    
    par.fStrobeMode = kGPSStrobeLocal;
    par.fTimeScale  = kGPSTimeScaleGPS;

    DumpGPSClockConfig( par );
    
    return false;
  }
  
  FD_CINFO << "\tReading section [" << section << "] ..." << endl;

  char *strobe = strdup( fInifile->GetFirstString( "strobe", "gps" ) );
  assert( strobe != NULL );

  for ( unsigned int i=0; i<strlen( strobe ); i++ )
    strobe[i] = tolower( strobe[i] );
  
  if ( strstr( strobe, "gps" ) )
    par.fStrobeMode = kGPSStrobeGPS;
  else if ( strstr( strobe, "local" ) )
    par.fStrobeMode = kGPSStrobeLocal;
  else
    par.fStrobeMode = kGPSStrobeLocal;
  
  free( strobe );
  
  char *set = strdup( fInifile->GetFirstString( "set", "gps" ) );
  assert( set != NULL );

  for ( unsigned int i=0; i<strlen( set ); i++ )
    set[i] = tolower( set[i] );
  
  if ( strstr( set, "gps" ) )
    par.fTimeScale = kGPSTimeScaleGPS;
  else if ( strstr( set, "utc" ) )
    par.fTimeScale = kGPSTimeScaleUTC;
  else if ( strstr( set, "unix" ) )
    par.fTimeScale = kGPSTimeScaleGPS;
  else
    par.fTimeScale = kGPSTimeScaleUnix;

  free( set );
  
  DumpGPSClockConfig( par );
  
  return true;
}

// -------------------------------------------------------------------------

bool ReadoutConfig::ReadSectionMStcCutTlt()
 {
  if ( !MStcCutTlt::ReadParameters( fInifile, 
				    &fMStcCutTltPar, 
				    "MStcCutTlt" ) ) {
    FD_CERR << "ReadoutConfig::ReadSectionMStcCutTlt() - failed !" << endl;
    
    return false;
  }
  
  return true;
}

// -------------------------------------------------------------------------

bool ReadoutConfig::ReadSectionNeighbor(MiReadout::TltParametersRec& tlt,
                                        const std::string& section)
 {
  assert( section.length() != 0 );
  
  if ( !PrototypeTltProcessor::ReadParameters( fInifile, 
                                               &tlt, section ) ) {
    FD_CERR << "ReadoutConfig::ReadSectionNeighbor() - failed !" << endl;
    
    return false;
  }
  
  //ReadSectionFullReadout( slt );
  //ReadSectionSelectiveReadout( slt );

  return true;
}

// -------------------------------------------------------------------------

bool ReadoutConfig::ReadSectionSelectiveReadout(MiReadout::SltParametersRec& par,
                                                const std::string& section)
 {
  assert( section.length() != 0 );
  
  if (    fInifile->SpecifyGroup( section.c_str() )
       != FdUtil::Inifile::kSUCCESS ) {
    FD_CERR << "!!! ReadoutConfig::Section [" << section 
            << "] not found !!!" << endl;
    
    return false;
  }
  
  SltParametersRec default_sltpar = DEFAULT_SLT_PARAMETERS;

  FD_CINFO << "\tReading section [" << section << "] ..." << endl;
  
  par.fStartBin = fInifile->GetFirstValue( "firstbin", default_sltpar.fStartBin );
  par.fEndBin   = fInifile->GetFirstValue( "lastbin", default_sltpar.fEndBin );

  par.fMinDistance  = fInifile->GetFirstValue( "mindist", default_sltpar.fMinDistance );
  par.fNExtColumns  = fInifile->GetFirstValue( "colext", default_sltpar.fNExtColumns );
  par.fNExtRows     = fInifile->GetFirstValue( "rowext", default_sltpar.fNExtRows );
  par.fFirstSltBin  = fInifile->GetFirstValue( "firstsltbin", default_sltpar.fFirstSltBin );
  par.fLastSltBin   = fInifile->GetFirstValue( "lastsltbin", default_sltpar.fLastSltBin );
  par.fMaxEventSize = fInifile->GetFirstValue( "maxpixel", default_sltpar.fMaxEventSize );

  return true;
}

// -------------------------------------------------------------------------

bool ReadoutConfig::ReadSectionTltReadout(MiReadout::TltParametersRec& par,
                                          const std::string& section)
 {
  assert( section.length() != 0 );
  
  if ( !PrototypeTltProcessor::ReadParameters( fInifile, &par, section ) ) {
    FD_CERR << "ReadoutConfig::ReadSectionTltReadout() - failed !" << endl;
    
    return false;
  }
  
  return true;
}

// -------------------------------------------------------------------------

// assume that the configuration is already opened
void ReadoutConfig::SaveConfiguration()
 {
#warning ReadoutConfig::SaveConfiguration() not implemented !
  FD_CWARN << "ReadoutConfig::SaveConfiguration() not implemented !" << endl;
}

// -------------------------------------------------------------------------

void ReadoutConfig::SetDefaultConfiguration()
 {
  FltParametersRec flt_par = DEFAULT_FLT_PARAMETERS;
  fFltParameters = flt_par;
  
  SltParametersRec slt_par = DEFAULT_SLT_PARAMETERS;
  fSltParameters["Default"] = slt_par;

  TltParametersRec tlt_par = DEFAULT_TLT_PARAMETERS;
  fTltParameters["Default"] = tlt_par;
  
  GPSClockConfigRec clock_par = DEFAULT_GPSCLOCK_CONFIG;
  fGpsClockConfig = clock_par;
  
  MStcCutTltParametersRec mstc_par = MSTC_TLT_DEFAULT_PARAMETERS;
  fMStcCutTltPar = mstc_par;
  
  tlt_par.fMaxEventSize = 50;  // DEFAULT_NEIGHBOR_PARAMETERS;
  tlt_par.fKeepRejectedFraction = 0;

  fTltParameters["NeighborTriggered"] = tlt_par;
}

// -------------------------------------------------------------------------

void ReadoutConfig::SetSltParameters(MiReadout::SltParameters slt,
                                     const std::string& config)
 {
  if ( config.length() == 0 )
    fSltParameters["Default"] = *slt;
  else
    fSltParameters[config] = *slt;
}

// -------------------------------------------------------------------------

void ReadoutConfig::SetTltParameters(MiReadout::TltParameters slt,
                                     const std::string& config)
 {
  if ( config.length() == 0 )
    fTltParameters["Default"] = *slt;
  else
    fTltParameters[config] = *slt;
}

// -------------------------------------------------------------------------
// -------------------------------------------------------------------------
