
//
// File   : SelectivePixelReadout.cc
//
// Purpose: Implementation file for class SelectivePixelReadout
//
// $Id: SelectivePixelReadout.cc 9725 2010-09-28 07:31:17Z mathes $
//


#include <iostream>
#include <iomanip>
#include <cstdio>
#include <ctime>


/** @file SelectivePixelReadout.cc
  * Implementation of class SelectivePixelReadout.
  * @author H.-J. Mathes, Fzk
  */

#include <FdNumbering.hh>
#include <FdStdioStream.hh>
#include <ErrorLogging.hh>
#include <MiEvent.hh>

#include "CheckVersion.hh"

#include "SelectivePixelReadout.hh"

using namespace std;
using FdUtil::Fd;
using namespace MiReadout;

// --------------------------------------------------------------------------
// --------------------------------------------------------------------------
//
// class SelectivePixelReadout - implementation
//

int SelectivePixelReadout::fgPrintLevel = 0;

SelectivePixelReadout::SelectivePixelReadout(SltParameters slt_par)
 : VPixelSelector("SelectivePixelReadout"),
   fLogfile(NULL), 
   fBitArray(new TMirrorBitArray), 
   fPixelData(NULL), 
   fPixelList(new TMirrorPixelList),
   fNEventsAnalysed(0), fNEventsRejected(0)
 {
  if ( fgPrintLevel > 2 )
    FD_CINFO << "SelectivePixelReadout::SelectivePixelReadout() -" << endl;

  fStopwatch.Reset();

  // --- set the initial parameters
  
  SltParametersRec slt_default = DEFAULT_SLT_PARAMETERS;
  
  if ( slt_par == NULL ) 
    fSltParameters = slt_default;
  else
    fSltParameters = *slt_par;
}

// --------------------------------------------------------------------------

SelectivePixelReadout::~SelectivePixelReadout()
 {
  if ( fgPrintLevel > 2 )
    FD_CINFO << "SelectivePixelReadout::~SelectivePixelReadout()" << endl;

  delete fBitArray;
  delete fPixelList;

  if ( fLogfile ) {
    FdUtil::OstdioStream ostr( fLogfile );

    PrintFinalResults( ostr );
  }
  
  if ( fgPrintLevel ) {
    PrintFinalResults( cout );
  }
}
	
// --------------------------------------------------------------------------

void SelectivePixelReadout::Clear()
 {
  if ( fgPrintLevel > 2 )
    FD_CINFO << "SelectivePixelReadout::Clear()" << endl;

  fBitArray->Clear();

  fPixelData = NULL;
}

// --------------------------------------------------------------------------

void SelectivePixelReadout::Select(TMirrorPixelData* pixel_data,
                                   TMirrorEventHeader::EMiTriggerSource /* trigger_source */,
				   unsigned int column_list)
 {
  if ( fgPrintLevel > 2 )
    FD_CINFO << "SelectivePixelReadout::Select()" << endl;

  fNEventsAnalysed++;
  
  if ( !pixel_data ) {
    FD_CERR << "SelectivePixelReadout::Select() - no PixelData present !" << endl;
    return;
  }
  
  fPixelData = pixel_data;
  
  fStopwatch.Start(kFALSE);
  
  DeleteSinglePixels( fPixelData );
  
  // --- project pixel data contents into bit array (for specified bin range)
  
  fBitArray->Clear();

  // in case of laser shots, pixels which trigger after fLastSltBin are not
  // readout --> select until the end of the time range
  //fBitArray->Set( fPixelData, fSltParameters.fFirstSltBin,
  //                            fSltParameters.fLastSltBin + 1 );
  fBitArray->Set( fPixelData, fSltParameters.fFirstSltBin,
                              Fd::kMI_NSLT_BINS );

  if ( fSltParameters.fMaxEventSize < Fd::kMI_NPIXELS ) {
  
    if (    fBitArray->GetNumberOfBits() > fSltParameters.fMaxEventSize 
       /*  || fBitArray->GetNumberOfBits() < fSltParameters.fMinEventSize */ ) {
      
      if ( fgPrintLevel) {
        if ( fBitArray->GetNumberOfBits() > fSltParameters.fMaxEventSize )
          FD_CINFO << "SelectivePixelReadout::Select() - too many pixels !" << endl;
      //  else if ( fBitArray->GetNumberOfBits() < fSltParameters.fMinEventSize )
      //    FD_CINFO << "SelectivePixelReadout::Select() - not enough pixels !" << endl;
      }
      
      // does it make sense to set this ?
      
      VPixelSelector::Set( fBitArray );
      
      // set an empty pixel list
      
      TMirrorPixelList pixel_list;
      VPixelSelector::Set( &pixel_list );
  
      fNEventsRejected++;

      fStopwatch.Stop();
    
      return;
    }
  }
  else {  // fSltParameters.fMaxEventSize == Fd::kMI_NPIXELS
  }

  // --- extend the readout area by the specified number of cols/rows
  
  fBitArray->Extend( fSltParameters.fNExtColumns, fSltParameters.fNExtRows );
  
  // remove columns without modules from the pixel list
  
  for ( unsigned int col=Fd::kMI_FIRST_COL; col<=Fd::kMI_LAST_COL; col++ ) {
    if ( (column_list & (1 << (col - Fd::kMI_FIRST_COL))) == 0 )
      fBitArray->Clear( col );
  }
  
  VPixelSelector::Set( fBitArray );
  
  fPixelList->Set( fBitArray );
  VPixelSelector::Set( fPixelList );
  
  // thjm 2010-03-15: I don't understand why this is here ...
  
  if ( fBitArray->GetNumberOfBits() == 0 ) {
    fNEventsRejected++;
  }

  fStopwatch.Stop();
}

// --------------------------------------------------------------------------

void SelectivePixelReadout::DeleteSinglePixels(TMirrorPixelData* pixeld)
 {
  if ( fSltParameters.fMinDistance >= Fd::kMI_NROWS ) return;
  
  int min_dist = fSltParameters.fMinDistance;
  
  TMirrorBitArray* bit_arr;
  
  if ( ( bit_arr = new TMirrorBitArray( pixeld ) ) != NULL ) {
    
    unsigned int rmin, rmax, cmin, cmax;
    TMirrorPixelData::PixelDataWord pixelword;
      
    for ( unsigned int col=Fd::kMI_FIRST_COL; col<=Fd::kMI_LAST_COL; col++ ) {
      for ( unsigned int row=Fd::kMI_FIRST_ROW; row<=Fd::kMI_LAST_ROW; row++ ) {
          
	if ( bit_arr->Test( col, row ) ) {
          
	  if ( min_dist < Fd::kMI_NROWS ) {
	    
	    bool single = true;
	    
	    cmin = ( (col - min_dist) > Fd::kMI_LAST_COL ) ? (int)Fd::kMI_FIRST_COL : col - min_dist;
	    cmax = ( (col + min_dist) > Fd::kMI_LAST_COL ) ? (int)Fd::kMI_LAST_COL : col + min_dist;
	    rmin = ( (row - min_dist) > Fd::kMI_LAST_ROW ) ? (int)Fd::kMI_FIRST_ROW : row - min_dist;
	    rmax = ( (row + min_dist) > Fd::kMI_LAST_ROW ) ? (int)Fd::kMI_LAST_ROW : row + min_dist;
            
	    for ( unsigned int c=cmin; c<=cmax; c++ ) {
	      for ( unsigned int r=rmin; r<=rmax; r++ )
	        if ( c != col || r != row )
	          if ( bit_arr->Test( c, r ) ) { single = false; break; } 
	    }

	    if ( single ) {
	      bit_arr->Clear( col, row );
            
	      int mask = 1 << (row - Fd::kMI_FIRST_ROW);
	      
	      for ( int bin=0; bin<Fd::kMI_NSLT_BINS; bin++ ) {
                pixelword = pixeld->GetPixelData( col, bin );
		
		PixelDataSetData( pixelword, PixelDataGetData( pixelword ) & ~mask );
              }
	    }
	  }
	}
      }
    }
    
    delete bit_arr;
  }
  else {
    FD_CERR << "SelectivePixelReadout::DeleteSinglePixels() - "
            << "could not allocate memory !" << endl;
  }
}

// --------------------------------------------------------------------------

void SelectivePixelReadout::PrintFinalResults(ostream &ostr)
 {
  ostr << endl;
  ostr << "=== SelectivePixelReadout '"
       << GetName() << "': Final results === " << endl;

  ostr << "calls/(real/CPU) time:  " << setw(5) << fStopwatch.Counter() << " "
       << setprecision(4) << fStopwatch.RealTime() << " " 
       << setprecision(4) << fStopwatch.CpuTime() << endl;
  ostr << "<Time>/call(real/CPU): " << "     "
       << setprecision(4) 
       << fStopwatch.RealTime()/fStopwatch.Counter() 
       << " " << setprecision(4) 
       << fStopwatch.CpuTime()/fStopwatch.Counter() << endl;

  ostr << "SelectivePixelReadout Statistic: " << endl;
  ostr << "Number of events analysed:   " << fNEventsAnalysed << endl;
  ostr << "Number of events rejected:   " << fNEventsRejected << endl;
  
  ostr << "==========================================" << endl;
}

// --------------------------------------------------------------------------

void SelectivePixelReadout::PrintSettings(ostream& ostr)
 {
  time_t t = time( NULL );
  struct tm *tm = gmtime( &t );
  
  ostr << endl;
  ostr << "*** SelectivePixelReadout - your present settings@UTC " << asctime( tm );
  ostr << endl;
  
  ostr << "\tfNExtColumns=  " << fSltParameters.fNExtColumns << endl;
  ostr << "\tfNExtRows=     " << fSltParameters.fNExtRows << endl;
  ostr << "\tfMinDistance=  " << fSltParameters.fMinDistance << endl;
  ostr << "\tfFirstSltBin=  " << fSltParameters.fFirstSltBin << endl;
  ostr << "\tfLastSltBin=   " << fSltParameters.fLastSltBin << endl;
  ostr << "\tfMaxEventSize= " << fSltParameters.fMaxEventSize << endl;
  
  ostr << endl;
}

// --------------------------------------------------------------------------
// --------------------------------------------------------------------------
