
//
// File   : StcShowerCore.cc
//
// Purpose: Implementation of the class StcShowerCore
//
// $Id: StcShowerCore.cc 9725 2010-09-28 07:31:17Z mathes $
//


/** @file StcShowerCore.cc
  * Implementation of class StcShowerCore.
  * @author Adrian Schmidt, FzK IPE
  */



#include <iostream>
#include <iomanip>
#include <cstdio>
#include <cstdlib>
#include <ctime>
#include <algorithm>

#include <TMath.h>
#include <TNamed.h>

#include <FdNumbering.hh>
#include <FdStdioStream.hh>
#include <ErrorLogging.hh>
#include <PixelPattern.hh>
#include <MiEvent.hh>

#include "CheckVersion.hh"
#include "ReadoutConfig.hh"

#include "StcShowerCore.hh"

using namespace std;
using FdUtil::Fd;
using namespace MiReadout;

// --------------------------------------------------------------------------
// --------------------------------------------------------------------------
//
// class  StcShowerCore - implementation
//

StcShowerCore::StcShowerCore(unsigned int nMaxCorePix) 
 {
  if ( VTltProcessor::GetPrintLevel() > 1 )
    FD_CINFO << "StcShowerCore::StcShowerCore(nMaxCorePix="
             << nMaxCorePix << ") -" << endl;
  
  fNMaxCorePix = nMaxCorePix;
  fNCorePix = 0;
  
  fNUp = 0;
  fNDown = 0;
  fNEqual = 0;
  
  fMeanX = 0;
  fMeanY = 0;
  fFittedAlpha = 0;
  
  fMeanT = fTruncMeanT = 0;
  fDevT = 0;
  
  fCore = new CoreRec[fNMaxCorePix];
  fCoreSortIndex = new size_t[fNMaxCorePix];
  
  fHasSTCPar = false;
  fHasTime = false;
  fHasLambda = false;
  fIsTTrimmed = false;
  fIsSorted = false;
  fIsSpecialChecked = false;
  
  if ( VTltProcessor::GetPrintLevel() > 2 )
    FD_CINFO << " Shower core with " << fNMaxCorePix 
             << " pixels constructed! " << endl;
}
     
// --------------------------------------------------------------------------

StcShowerCore::~StcShowerCore()
 {
  delete [] fCoreSortIndex;
  delete [] fCore;
}

// --------------------------------------------------------------------------
     
void StcShowerCore::AddPixel(unsigned int col,unsigned int row)
 {
  if ( VTltProcessor::GetPrintLevel() > 2 )
     FD_CINFO << "StcShowerCore::AddPixel(col=" << col 
              << ",row=" << row << ") -" << endl;
  
  if ( fNCorePix >= fNMaxCorePix ) {
    FD_CERR << "StcShowerCore::AddPixel() - "
    	    << "ShowerCore List full !" << endl;
    return;
  } 
  
  fCore[fNCorePix].fCol = col;
  fCore[fNCorePix].fRow = row;
  fNCorePix++;
}

// --------------------------------------------------------------------------

void StcShowerCore::AssignTime(VMirrorEvent* readout,
                               unsigned int boxcarLength, 
                               unsigned int tStart,unsigned int tStop)
 {
  if ( VTltProcessor::GetPrintLevel() > 1 )
    FD_CINFO << "StcShowerCore::AssignTime() -" << endl;
      
  TMirrorPixelData *miPd = readout->GetPixelData();

  unsigned int n_fadc_bins = TFADCData::GetNFADCBins(TFADCData::GetDefaultResolution());
  unsigned int fadc_2_slt_factor = 1;
  
  // looping over all pixels in shower core
  for ( unsigned int j = 0; j<fNCorePix; j++ ) {
    
    CoreRecRef core = fCore[j];
    
    // Get FADC Data
    Fd::PixelNumberRec pixel;
    Fd::SetPixelNoFromColumnRow( &pixel, core.fCol, core.fRow);
    
    TFADCData::FADCTraceInfoRec trace_info;
    FADCTraceSetStartBin( &trace_info, 0 );
    FADCTraceSetEndBin( &trace_info, n_fadc_bins - 1 );

    TFADCData *fadc = readout->GetFADCData( &pixel, &trace_info );
    if ( fadc ) {
    
      // adopt this factor for FADC types currently in use
      fadc_2_slt_factor = fadc->GetNFADCBins() / 10 / Fd::kMI_NSLT_BINS;
      
      // Setting up time interval for peak-search
      // -> gives values in range 0 ... (Fd::kMI_NSLT_BINS - 1) or -1
      unsigned int nFirst = miPd->GetFirstTriggeredTimeBin(core.fCol, core.fRow, 10, 100);
      unsigned int tMin = nFirst - tStart;
      unsigned int tMax = nFirst + tStop;
    
      tMin *= fadc_2_slt_factor;
      tMax *= fadc_2_slt_factor;
      
      if ( tMin < boxcarLength ) tMin = boxcarLength;
      if ( tMax > n_fadc_bins ) tMax = n_fadc_bins;
    
      TFADCData::FADCDataWord data = fadc->GetFADCTrace();
      int boxcar  = 0, maxBoxcar = -100000;
      unsigned int tMaxBoxcar = 0;
      
      if ( VTltProcessor::GetPrintLevel() > 2 )
	FD_CINFO << "StcShowerCore::AssignTime() - pix " << core.fCol 
                 << "," << core.fRow << ": nFirst= " << nFirst
		 << " tMin= " << tMin
	         << " tMax= " << tMax << endl;
    
      // searching for maximum boxcar
      for ( unsigned int t=tMin; t<tMax; t++ ) {

    	boxcar += FADCDataWordGetData(&data[t]);
    	boxcar -= FADCDataWordGetData(&data[t - boxcarLength]);
    	
    	if ( boxcar > maxBoxcar ) {
    	  maxBoxcar = boxcar;
    	  tMaxBoxcar = t;
    	}
      }

      core.fTime = tMaxBoxcar; 
    
      if ( VTltProcessor::GetPrintLevel() > 2 )
        FD_CINFO << "StcShowerCore::AssignTime() - fTime(" << core.fCol 
	         << "," << core.fRow << ")= " << core.fTime << endl;
      
      delete fadc;
    
    } else {
    
      FD_CERR << "StcShowerCore::AssignTime() - "
              << "no FADC Data of Pixel in Shower core found!" << endl;
      
      core.fTime = 0;
    }
  }
  
  fHasTime = true;
  
  if ( VTltProcessor::GetPrintLevel() > 3 ) {
    FD_CINFO << "StcShowerCore::AssignTime() - calculated Pixel Times for " 
             << fNCorePix  << " Pixels: " << endl;
    
    for ( UInt_t j = 0; j < fNCorePix; j++ ) 
      FD_LOGSTREAM(FdUtil::ErrorLogger::eInfo) << "  " <<  fCore[j].fTime;
    
    FD_LOGSTREAM(FdUtil::ErrorLogger::eInfo) << endl;
  }
}

// --------------------------------------------------------------------------

void StcShowerCore::RemoveWrongTPixel(double significance)
 {
  if ( VTltProcessor::GetPrintLevel() > 1 )
    FD_CINFO << "StcShowerCore::RemoveWrongTPixel() -" << endl;
      
  if ( !fHasTime ) {
    FD_CERR << "StcShowerCore::RemoveWrongTPixel() - "
            << " Error: Shower Core has no time information! "
	    << "Run StcShowerCore::AssignTime() first!" << endl;
    fNCorePix = 0; 
    return; 
  }
  
  //remove pixels with t == 0
  unsigned int k = 0;
  for (unsigned int j = 0; j < fNCorePix; j++) {
    if ( fCore[j].fTime != 0 ) {
      fCore[k].fTime = fCore[j].fTime;
      fCore[k].fCol = fCore[j].fCol;
      fCore[k].fRow = fCore[j].fRow;
      k++;
    }
  }

  fNCorePix = k;

  if ( fNCorePix == 0 ) {
    FD_CERR << "StcShowerCore::RemoveWrongTPixel() - "
            << "all Pixels have t == 0 !" << endl;
    return;
  }
 
  
  // iteratively removing all pixels where t differs more than significance*deviation from mean value
  bool trimmed;
  do {
    trimmed = false;
     
    if ( fNCorePix < 4 ) {
      FD_CERR << "StcShowerCore::RemoveWrongTPixel() - "
              << "not enough Pixels left in Shower core !" << endl;
      return;
    }
 
    CalcMeanT();  // -> fMeanT, fDevT, fTruncMeanT
    
    k = 0;  

    for (unsigned int j = 0; j < fNCorePix; j++) {
      if (fabs(fMeanT - (double)fCore[j].fTime) <= significance*fDevT) {
        fCore[k].fTime = fCore[j].fTime;
        fCore[k].fCol = fCore[j].fCol;
        fCore[k].fRow = fCore[j].fRow;
        k++;
      }
      else {
        if ( VTltProcessor::GetPrintLevel() > 2 )
	  FD_COUT << "StcShowerCore::RemoveWrongTPixel() - "
	           << "pix " << fCore[j].fCol << "," << fCore[j].fRow 
		   << " t= " << fCore[j].fTime << " removed" << endl;
        trimmed = true;
      }
    }  // for ( j=..)
    
    fNCorePix = k;
    
  } while ( trimmed );

  fIsTTrimmed = true;
}

// --------------------------------------------------------------------------

void StcShowerCore::CalcLambda()
 {
  if ( VTltProcessor::GetPrintLevel() > 1 )
    FD_CINFO << "StcShowerCore::CalcLambda() -" << endl;
 
  CalcDirection();
  
  for ( UInt_t j = 0; j < fNCorePix; j++ ) 
    fCore[j].fLambda = ((Double_t) fCore[j].fCol + 0.5*(Double_t)(fCore[j].fRow % 2) - fMeanX)*sin(fFittedAlpha) + ((Double_t) fCore[j].fRow - fMeanY) *cos(fFittedAlpha);
  
  fHasLambda = true;
}

// --------------------------------------------------------------------------

void StcShowerCore::Sort() 
 {
  if (VTltProcessor::GetPrintLevel() > 1)
    FD_CINFO << "StcShowerCore::Sort() -" << endl;
  
  if ( !fHasLambda ) {
    FD_CERR << "StcShowerCore::Sort() - "
            << "shower core has no Lambda values! "
	    << "Run StcShowerCore::CalcLambda() first!" << endl;
    return;
  }
  
  for ( unsigned int i=0; i<fNCorePix; i++ )
    fCoreSortIndex[i] = i;

  std::sort(fCoreSortIndex, fCoreSortIndex + fNCorePix, CompareCoreAsc_t( fCore ) );

#if 0
  // false = sort in increasing order
  TMath::Sort( fNCorePix, fLambda, fCoreSortIndex, false );

template <typename Element, typename Index> 
void TMath::Sort(Long64_t n, const Element* a, Index* index, Bool_t down)
{
   // Sort the n1 elements of the Short_t array a.
   // In output the array index contains the indices of the sorted array.
   // If down is false sort in increasing order (default is decreasing order).

   // NOTE that the array index must be created with a length >= n
   // before calling this function.

   for(Long64_t i = 0; i < n; i++) { index[i] = i; }
   if ( down )
      std::sort(index, index + n, CompareDesc<const Element*>(a) );
   else
      std::sort(index, index + n, CompareAsc<const Element*>(a) );
}

template<typename T> 
struct CompareAsc { 

   CompareAsc(T d) : fData(d) {}

   bool operator()(int i1, int i2) { 
      return *(fData + i1) < *(fData + i2);
   }

   T fData; 
};
#endif

  fIsSorted = true;
}
     
// --------------------------------------------------------------------------
   
void StcShowerCore::DoStcCheck()
 {
  if (VTltProcessor::GetPrintLevel() > 1)
    FD_CINFO << "StcShowerCore::DoStcCheck() -" << endl;
	 
  if ( !fIsSorted ) {
    FD_CERR << "StcShowerCore::DoStcCheck() - "
            << "Shower core is unsortet! "
	    << "Run StcShowerCore::Sort() first!" << endl;
    return;
  }
   
  fNUp = 0;
  fNDown = 0;
  fNEqual = 0;

  for ( unsigned int j = 0; j < fNCorePix-1; j++ ) {
    if (fCore[fCoreSortIndex[j]].fTime<fCore[fCoreSortIndex[j+1]].fTime)
      fNUp++;
    else if (fCore[fCoreSortIndex[j]].fTime>fCore[fCoreSortIndex[j+1]].fTime)
      fNDown++;
    else
      fNEqual++;
  }   
  
  fHasSTCPar = true;
}

// --------------------------------------------------------------------------
     
void StcShowerCore::DoSpecialStcCheck()
 {
  if ( VTltProcessor::GetPrintLevel() > 1 )
    FD_CINFO << "StcShowerCore::DoSpecialStcCheck() -" << endl;
 
  if ( !fHasSTCPar ) {
    FD_CERR << "StcShowerCore::DoSpecialStcCheck() - "
            << "Shower was not STC-checked yet! Run StcShowerCore::DoStcCheck() first!" << endl;
    return;
  }
  
  // Try to leave out the Pixel number j each cylce and 
  // test wether The Time Correlation becomes better.
  for ( UInt_t j = 0; j < fNCorePix; j++ ) {
    UInt_t nUp =0, nDown = 0, nEqual = 0;   
  
    UInt_t lastT = fCore[fCoreSortIndex[0]].fTime;
    if (j == 0) lastT = fCore[fCoreSortIndex[1]].fTime;
    
    for (UInt_t k = 1; k < fNCorePix; k++) {
      if (k == j || (j == 0 && k == 1)) continue;
      UInt_t thisT = fCore[fCoreSortIndex[k]].fTime;
      
      if (lastT < thisT)
        nUp++;
      else if (lastT > thisT)
        nDown++;
      else
        nEqual++;
      
      lastT =  thisT;
    }
    
    if (abs((Int_t) nUp - (Int_t) nDown) > abs((Int_t) fNUp - (Int_t) fNDown)) {
      fNUp = nUp;
      fNDown = nDown;
      fNEqual = nEqual;
      fIsSpecialChecked = true;
    }
  }
  
  if ( fIsSpecialChecked ) fNCorePix--;
}

// --------------------------------------------------------------------------
     
unsigned int StcShowerCore::GetTimeCorrelation()
 {
  if ( VTltProcessor::GetPrintLevel() > 2 )
    FD_CINFO << "StcShowerCore::GetTimeCorrelation() - "
              << "n_corepix= " << fNCorePix 
	      << " n_up= " << fNUp 
	      << " n_down= " << fNDown <<  endl;
   
  if ( fHasSTCPar )
    return abs((Int_t)fNUp - (Int_t)fNDown);
  else {
    FD_CERR << "StcShowerCore::GetTimeCorrelation() - "
            << "Shower was not STC-checked yet! Run StcShowerCore::DoStcCheck() first!" << endl;
    return 0;
  }
}
     
// --------------------------------------------------------------------------
       
void StcShowerCore::CalcMeanT()
 {
  if ( VTltProcessor::GetPrintLevel() > 2 )
    FD_CINFO << "StcShowerCore::CalcMeanT() -" << endl;
 
  int sumT = 0, sumT2 = 0;
  unsigned int tmin = 9999, tmax = 0;
  
  for ( unsigned int j = 0; j < fNCorePix; j++) {
    if ( fCore[j].fTime < tmin ) tmin = fCore[j].fTime;
    if ( tmax < fCore[j].fTime ) tmax = fCore[j].fTime;
    sumT += fCore[j].fTime;
    sumT2 += fCore[j].fTime*fCore[j].fTime;
  }
  
  fMeanT = sumT / (double)fNCorePix;
  fDevT = sqrt((double)(fNCorePix*sumT2 - sumT*sumT)/(Double_t)(fNCorePix*(fNCorePix-1)));

  if ( fNCorePix >= 4 ) {
    fTruncMeanT = (fMeanT * (double)fNCorePix - tmin - tmax) / 
                  (double)(fNCorePix - 2);
  }
  
  if ( VTltProcessor::GetPrintLevel() > 3 ) {
    FD_CINFO << "StcShowerCore::CalcMeanT() - " 
             << "tmin= " << tmin << " tmax= " << tmax << endl;
    FD_CINFO << "StcShowerCore::CalcMeanT() - " 
             << "fMeanT= " << fMeanT 
             << " fDevT= " << fDevT 
	     << " fTruncMeanT= " << fTruncMeanT << endl;
  }
}

// --------------------------------------------------------------------------

void StcShowerCore::CalcDirection()
 {
  if ( VTltProcessor::GetPrintLevel() > 2 )
    FD_CINFO << "StcShowerCore::CalcDirection() -" << endl;
 
  // summate all x and y values of shower core
  Double_t sx=0, sx2=0, sy=0, sy2=0, sxy=0;
  for (UInt_t j = 0; j < fNCorePix; j++) {
    sy += (Double_t) fCore[j].fRow;
    sx += (Double_t) fCore[j].fCol+ .5*(Double_t)(fCore[j].fRow%2);
    sy2 += (Double_t) (fCore[j].fRow*fCore[j].fRow);
    sx2 += ((Double_t) fCore[j].fCol+ .5*(Double_t)(fCore[j].fRow%2))*((Double_t) fCore[j].fCol+ .5*(Double_t)(fCore[j].fRow%2));
    sxy += (Double_t) fCore[j].fRow * ((Double_t) fCore[j].fCol+ .5*(Double_t)(fCore[j].fRow%2));
  }
  
  fMeanX = sx/(Double_t) fNCorePix;
  fMeanY = sy/(Double_t) fNCorePix;
  
  Double_t denominator = (Double_t) fNCorePix *sy2 - sy*sy - (Double_t) fNCorePix *sx2 + sx*sx;
  Double_t numerator = 2. * ((Double_t) fNCorePix * sxy - sx*sy);
  
  if ( denominator == 0. ) {
    if ( numerator > 0. )
      fFittedAlpha = 0.7853;
    else if (numerator < 0.)
      fFittedAlpha = -0.7853;
    else
      fFittedAlpha  = 0.;
  }
  else fFittedAlpha = 0.5*atan(numerator/denominator);

  if (denominator < 0.) fFittedAlpha += 1.5707;
  if (fFittedAlpha > 1.5707) fFittedAlpha -= 3.1415;
}
     
// --------------------------------------------------------------------------
   
bool StcShowerCore::IsMuon()
 {
  if ( !fHasSTCPar ) {
    FD_CERR << "StcShowerCore::IsMuon() - "
            << "shower was not STC-checked yet! "
	    << "Run StcShowerCore::DoStcCheck() first!" << endl;
    return false;
  }
  
  if ( fNEqual > (fNCorePix*2)/3) {
    if ( VTltProcessor::GetPrintLevel() > 2 )
      FD_CINFO << "StcShowerCore::IsMuon() - "
               << "Event is camera muon hit" <<  endl;
    return true;
  }
  else
    return false;
}    

// --------------------------------------------------------------------------
// --------------------------------------------------------------------------
