
//
// File   : StcShowerCore.hh
//
// Purpose: Declaration of the class StcShowerCore
//
// $Id: StcShowerCore.hh 9725 2010-09-28 07:31:17Z mathes $
//


/** @file StcShowerCore.hh
  * Declaration of class StcShowerCore.
  * @author Adrian Schmidt, FzK IPE
  */


#ifndef _StcShowerCore_hh_
#define _StcShowerCore_hh_

#include <RVersion.h>

// --- forward declaration(s)

class TFADCData;
class VirtualReadout;
class VMirrorEvent;


/** This is a class belonging to the Space-TimeCorrelation-TLT cut.
  *
  * It provides all necessary STC analysis functions.
  */
class StcShowerCore {
public:
    /** Constructor for the class StcShowerCore. 
      *
      * Place for nMaxCorePix CoreRec structs is allocated internally.
      */
    StcShowerCore(unsigned int);
    
    /** Destructor for the class StcShowerCore. */
    ~StcShowerCore();
    
    /** Add a Pixel to the Shower Core. */
    void AddPixel(unsigned int,unsigned int);
    
    /** Calculate the lambda values for all shower core pixels. */
    void CalcLambda();
    
    /** Sort ShowerCore according to lambda values. */
    void Sort();
    
    /** Assign Pixel Time with simple Pulsefinder. */
    void AssignTime(VMirrorEvent*, UInt_t, UInt_t, UInt_t);

    /** Cut out Pixels where time is some standard deviations away 
      * from mean time. */
    void RemoveWrongTPixel(Double_t);
   
    /** Calculate Number of up-down steps for all Showe core pixels. */
    void DoStcCheck();
    
    /** Test whether the Space Time Correlation becomes better, when 
      * leaving one Pixel out of the up-down step calculation. */
    void DoSpecialStcCheck();
    
    /** Returns the absolute value up n_up - n_down, i.e. the final 
      * STC-Parameter. */
    UInt_t GetTimeCorrelation();
    
    /** Returns the absolute value up n_up - n_down, i.e. the final 
      * STC-Parameter. */
    UInt_t GetNCorePix()
     { return fNCorePix; }
    
    /** Return if shower core was produced by camera muon hit */
    bool IsMuon();
      
protected:
    
    /**  */
    void CalcMeanT();

    /**  */
    void CalcDirection();

    /** struct which concentrates the former 'fCore...' attributes. */
    typedef struct _CoreRec {
    
      // initialize the struct, efficient ?
      _CoreRec() : fCol(0), fRow(0), fTime(0), fLambda(0.0) { }
      
      unsigned int  fCol;     //<<< column info
      unsigned int  fRow;     //<<< row info
      unsigned int  fTime;    //<<< time info
      double        fLambda;  //<<< 

    } CoreRec, &CoreRecRef;

private:

    /** A 'compare object for comparing CoreRec structs used in std::sort(). */
    typedef struct _CompareCoreAscRec { 

      _CompareCoreAscRec(StcShowerCore::CoreRec* d) : fData(d) {}

      bool operator()(int i1, int i2) { 
        return fData[i1].fLambda < fData[i2].fLambda;
      }

      StcShowerCore::CoreRec* fData;
      
    } CompareCoreAsc_t;
    
    /** A result struct for this class. */
    typedef struct _StcShowerCoreResultRec {
    
      unsigned int   fNCorePix;
      
    } StcShowerCoreResult_t;
    
    unsigned int   fNCorePix;
    unsigned int   fNMaxCorePix;
    
    unsigned int   fNUp;
    unsigned int   fNDown;
    unsigned int   fNEqual;

    double 	   fMeanX;
    double 	   fMeanY;
    double 	   fFittedAlpha;
    
    double 	   fMeanT;
    double 	   fDevT;
    double         fTruncMeanT;
    
    CoreRec       *fCore;

    size_t	  *fCoreSortIndex;
    
    bool	   fHasSTCPar;
    bool	   fHasTime;
    bool	   fHasLambda;
    bool	   fIsTTrimmed;
    bool	   fIsSorted;
    bool	   fIsSpecialChecked;
};

#endif //  _StcShowerCore_hh_


