
//
// File   : ErrorLogger.cc
//
// Purpose: Implementation of the class ErrorLogger
//
// $Id: ErrorLogger.cc 8758 2010-09-27 12:16:24Z mathes $
//


#include <string>
#include <iostream>
#include <iomanip>
#include <sstream>
#include <cassert>

#if 0
#ifdef __GNUC__
// class stdiostream is available for gcc 3.x
# if __GNUC__ < 3
#  include <stdiostream.h>
# else
#  include <ostream>
#  include <ext/stdio_filebuf.h>
# endif // __GNUC__ < 3
#else
# error This software was never compiled with compilers other than gcc !
#endif // __GNUC__
#endif


/** @file ErrorLogger.cc
  * Implementation of the class ErrorLogger.
  * @author L. Nellen, UNAM Mexico, S. Argir`o, INFN Torino, H.-J. Mathes, FzK
  */


static const char CvsId[] = 
"$Id: ErrorLogger.cc 8758 2010-09-27 12:16:24Z mathes $";

#include "TimeConvert.hh"

#include "ErrorLoggerStream.hh"
#include "ErrorLogging.hh"
#include "ErrorLogger.hh"


using namespace std;
using namespace FdUtil;

// -------------------------------------------------------------------------
// -------------------------------------------------------------------------
// 
// class ErrorLogger - implementation
// 

std::ostream *ErrorLogger::fgNullStream = new ofstream("/dev/null");

ErrorLogger::ErrorLogger()
 : fOStream(&std::clog), // default logging to std::clog
   fLoggingOption(0), 
#ifdef DPA
   fLoggingLevel(ErrorLogger::eInfo),
#else
   fLoggingLevel(ErrorLogger::eDebug),
#endif // DPA
   fVerbosity(ErrorLogger::eTerse)
 {
#warning ErrorLoggerStream is no longer connected to the ErrorLogger !!!
  // fOStream = &ErrorLoggerStream::GetInstance();
}

// -------------------------------------------------------------------------

std::ostream& ErrorLogger::GetStream(ESeverityLevel level)
 {
  if ( level < fLoggingLevel )
    return *fgNullStream;
  else {
    if ( fOStream )
      return *fOStream;
    else
      return std::clog;
  }
}

// -------------------------------------------------------------------------

std::string 
ErrorLogger::GetVerboseSeverityShort(ErrorLogger::ESeverityLevel severity)
 {
  switch ( severity ) {
    case eDebug:    return "DBG";
    case eInfo:     return "INF";
    case eWarning:  return "WRN";
    case eError:    return "ERR";
    case eCritical: return "CRI";
    case eSevere:   return "SEV";
    case eFatal:    return "FAT";
  }
  
  return "???";
}

// -------------------------------------------------------------------------

std::string 
ErrorLogger::GetVerboseSeverity(ErrorLogger::ESeverityLevel severity)
 {
  switch ( severity ) {
    case eDebug:    return "Debug";
    case eInfo:     return "Info";
    case eWarning:  return "Warning";
    case eError:    return "Error";
    case eCritical: return "Critical";
    case eSevere:   return "Severe";
    case eFatal:    return "Fatal";
  }
  
  return "Unknown";
}

// -------------------------------------------------------------------------

void ErrorLogger::Log(ErrorLogger::ESeverityLevel severity, 
		      std::string functionName, 
		      std::string fileName,
		      int lineNumber,
		      std::string message,
		      ErrorLogger::EVerbosityLevel verbosity) const
 {
  if ( severity < fLoggingLevel ) return;
  
  ostream *myOStream;

  if ( fOStream )
    myOStream = fOStream;
  else
    myOStream = &std::clog;

  if ( verbosity == ErrorLogger::eDefaultVerbosity ) 
    verbosity = fVerbosity;

  ostringstream messageHeader;
  
  if ( fLoggingOption & (int)ErrorLogger::eShowSeverity )
    messageHeader << GetVerboseSeverityShort( severity ) << ":";

  if ( fLoggingOption & (int)ErrorLogger::eShowGPSTime )
    messageHeader << std::setw(9) << FdUtil::TimeConvert::GetGPSTime() << ": ";
  
  if ( verbosity == ErrorLogger::eVerbose ) {
    messageHeader << fileName << ":";
    messageHeader << lineNumber << ": ";
  }
  else {
    if ( fLoggingOption & (int)ErrorLogger::eShowFilename )
      messageHeader << fileName << ":";
    if ( fLoggingOption & (int)ErrorLogger::eShowLineNumber )
      messageHeader << lineNumber << ": ";
  }
  
  if ( fLoggingOption & (int)ErrorLogger::eShowFuncName ) {
//    if ( fLoggingOption & (int)ErrorLogger::eShowClassName )
//      messageHeader << fClassName << "::";
    messageHeader << functionName << "() - ";
  }
  
  string messageHeaderSpace;
  messageHeaderSpace.resize(messageHeader.str().size(), ' ');
  
  assert( myOStream != NULL );

#ifdef __GNUC__
# if __GNUC__ < 3
// --- this is for gcc 2.95.3, the construction with dynamic_cast<>() 
//     doesn't work if object is not already an ErrorLoggerStream
  if ( myOStream == &ErrorLoggerStream::GetInstance() ) {
# else // __GNUC__ < 3
// this is for gcc 3.x
  if ( dynamic_cast<FdUtil::ErrorLoggerStream *>( myOStream ) != NULL ) {
# endif // __GNUC__ < 3
#else // __GNUC__
# error This software was never compiled with compilers other than gcc !
  if ( NON_AUGER_PLATFORM ) {
#endif // __GNUC__
   
    *myOStream << messageHeader.str() << message;
    
    dynamic_cast<ErrorLoggerStream *>( myOStream )->endm();
  }
  else
    *myOStream << messageHeader.str() << message << std::endl;
}

// -------------------------------------------------------------------------

void ErrorLogger::SetFile(const char* filename)
 {
  assert( filename != NULL && *filename != 0 );
  
  if ( fOStream )
    fOStream->flush();

  if ( fOfStream.is_open() )
    fOfStream.close();
  
  fOfStream.open( filename );
  
  if ( fOfStream.is_open() )
    fOStream = &fOfStream;
  else
    fOStream = &std::clog;
}

// -------------------------------------------------------------------------

#if 0
void ErrorLogger::SetFile(FILE *file)
 {
  assert( file != NULL );
  
#ifdef __GNUC__
# if __GNUC__ < 3
  ostdiostream *ostr = new ostream( file );
# else
  typedef __gnu_cxx::stdio_filebuf<char> ostdiostream;
  ostdiostream logfile( file, ios_base::out );
  
  ostream *ostr = new ostream( &logfile );
# endif // __GNUC__ < 3
  SetStream( ostr );
#endif // __GNUC__
}
#endif

// -------------------------------------------------------------------------

void ErrorLogger::SetLoggingLevel(ESeverityLevel level,bool quiet)
 {
  if ( !quiet) {
    FD_CWARN << "ErrorLogger::SetLoggingLevel(" << (int)level << "="
              << GetVerboseSeverity( level ) << " ) - ..." 
             << endl;
  }
    
  fLoggingLevel = level;
}

// -------------------------------------------------------------------------

void ErrorLogger::SetOption(ELoggingOption option, bool enable)
 {
  if ( enable ) 
    fLoggingOption |= (int)option;
  else
    fLoggingOption &= ~(int)option;
    
  if ( (fLoggingOption & eShowGPSTime) && (fLoggingOption & eShowUnixTime) )
    SetOption( eShowUnixTime, false );
}

// -------------------------------------------------------------------------

void ErrorLogger::SetVerbosity(EVerbosityLevel verbosity)
 {
  if ( verbosity != eDefaultVerbosity )
    fVerbosity = verbosity;
    
  if ( verbosity == eVerbose )
    SetOption( eShowFuncName );
}

// -------------------------------------------------------------------------
// -------------------------------------------------------------------------
