
//
// File:    FdConfig.cc
//
// Purpose: Implementation of the class FdConfig
//
// $Id: FdConfig.cc 11901 2011-01-19 09:09:07Z mathes $
//


#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <unistd.h>
#include <netdb.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <cerrno>

#include <iostream>
#include <string>

/** @file FdConfig.cc
  * Implementation of class FdConfig
  * @author H.-J. Mathes, FzK
  */
  
#include "FdNumbering.hh"
#include "ErrorLogging.hh"

#include "FdConfig.hh"

using namespace std;
using namespace FdUtil;

// -------------------------------------------------------------------------
// -------------------------------------------------------------------------
//
// class FdConfig - implementation
//

int FdConfig::fgPrintLevel = 0;

FdConfig::FdConfig()
 : fDefaultEyeId(Fd::kFD_FIRST_EYE), fDefaultMirrorId(Fd::kEYE_FIRST_MIRROR), 
   fIsEye(false), fIsMirror(false)
 {
  // --- look for default values defined by the environment
  
  if ( getenv( "FD_EYE_ID" ) != NULL ) {
    fDefaultEyeId = atoi( getenv( "FD_EYE_ID" ) );
    FD_CINFO << "FdConfig::FdConfig() - getenv(FD_EYE_ID)= " 
             << fDefaultEyeId  << endl;
  }
  
  if ( getenv( "FD_MIRROR_ID" ) != NULL ) {
    fDefaultMirrorId = atoi( getenv( "FD_MIRROR_ID" ) );
    FD_CINFO << "FdConfig::FdConfig() - getenv(FD_MIRROR_ID)= " 
             << fDefaultMirrorId  << endl;
  }
  
  // --- set to the default values
  
  fEyeId = fDefaultEyeId;
  fMirrorId = fDefaultMirrorId;
  
  char hostname[80];  
  bool eye_id_set = false;
  int stat = gethostname( hostname, sizeof( hostname ) );
  
  if ( stat < 0 ) {
    FD_CERR << "FdConfig::FdConfig() - gethostname() returned error, errno= "
            << errno << ", " << strerror(errno) << endl;
    strcpy( hostname, "localhost" );
  }
  
  if ( fgPrintLevel ) {
    FD_CDEBUG(1) << "FdConfig::FdConfig() - gethostname()= " 
                 << hostname << endl;
  }
  
  for ( unsigned int i=0; i<strlen( hostname ); i++ )
    hostname[i] = tolower( hostname[i] );
  
  // --- try to determine Mirror and Eye id from the hostname
  
  DetermineMirrorId( hostname );
  // DetermineEyeId( hostname );
  
  // --- try to determine Mirror and Eye id from /etc/hosts
  
  struct hostent* hostent = gethostbyname( hostname );
  
  if ( hostent != NULL ) {
    
    for ( unsigned int i=0; i<strlen( hostent->h_name ); i++ )
      fFQHostname += tolower( hostent->h_name[i] );
      
    if ( fgPrintLevel )
      FD_CDEBUG(1) << "FdConfig::FdConfig() - fFQHostname= " 
                   << fFQHostname.c_str() << endl;
    
    // analyse the fully qualified host name

    if ( !fIsMirror ) {
      DetermineMirrorId( fFQHostname.c_str() );
    }
    
    DetermineEyeId( fFQHostname.c_str() );
    eye_id_set = true;
  }
  else {
    eye_id_set = false;
    FD_CERR << "FdConfig::FdConfig() - gethostbyname() "
            << "could not do name lookup for '"
            << hostname << "' !" << endl;
    FD_CERR << "FdConfig::FdConfig() - gethostbyname() "
            << "h_errno= " << h_errno << endl;
  }
  
  // --- determine EyePC from hostname
  
  if (  !fIsMirror && !fIsEye ) {
    
    if ( strstr( hostname, "eyepc" ) != NULL ) fIsEye = true;
      
    // --- for old setup at IPE
      
    if ( strstr( hostname, "hpepc29" ) != NULL ) {
      fIsEye = true;
      fEyeId = 1; // use the default
      eye_id_set = true;
      if ( fgPrintLevel )
	FD_CDEBUG(1) << "FdConfig::FdConfig() - "
	             << "match with test EyePC 'hpepc29' was detected !" << endl;
    }
    
    // --- for new setup at IPE
      
    if ( strstr( hostname, "ipeauger1" ) != NULL ) {
      fIsEye = true;
      fEyeId = 1; // use the default
      eye_id_set = true;
      if ( fgPrintLevel )
	FD_CDEBUG(1) << "FdConfig::FdConfig() - "
	             << "match with test EyePC 'ipeauger1' was detected !" << endl;
    }
  }

  // --- determine Eye id from /etc/networks

  if ( !eye_id_set ) {
  
    setnetent( 1 );  // open /etc/networks
  
    struct netent *netent = getnetent();
    if ( netent != NULL ) {
      while ( netent != NULL ) {
        if ( fgPrintLevel )
          FD_CDEBUG(1) << "FdConfig::FdConfig() - getnetent()= " 
	               << netent->n_name << endl;
	DetermineEyeId( netent->n_name );
        netent = getnetent();
      }
    }
    else
      FD_CERR << "FdConfig::FdConfig() - "
              << "could not read net entry from /etc/networks !" << endl;
  
    endnetent();   // close /etc/networks
  }  // eye_id_set
  
  // --- set the composite mirror number
  
  Fd::SetEyeNo( &fMirrorNumber, GetEyeId() );
  Fd::SetMirrorNo( &fMirrorNumber, GetMirrorId() );
  
  // --- some (crucial) final check
  
  if ( fIsMirror && fIsEye ) {
    FD_CFATAL << "FdConfig::FdConfig() - "
              << "application cannot run both on EyePC and MirrorPC !" << endl;
    throw;
  }
}

// -------------------------------------------------------------------------

void FdConfig::DetermineEyeId(const char *name)
 {
  const char *fullq_hostname;
  
  if ( name != NULL && *name != 0 )
    fullq_hostname = name;
  else
    fullq_hostname = fFQHostname.c_str();

  // --- analyse the network name (Mirror LAN)

  char eye_name[20];
  
  for ( int i=0; i<7; i++ ) {
  
    sprintf( eye_name, "eye%d-auger", i+1 );
    
    if ( strstr( fFQHostname.c_str(), eye_name ) != NULL ) {
      fEyeId = i + 1;
      if ( fgPrintLevel )
	FD_CINFO << "FdConfig::DetermineEyeId() - "
	         << "match with Mirror LAN '" << eye_name << "' detected !" 
		 << endl;
      if ( !fIsMirror ) fIsEye = true;
      break;
    }
  }
  
  // --- workaround for EyeIDs > kFD_LAST_EYE, for example on ipeauger1
  
  if ( fEyeId > Fd::kFD_LAST_EYE ) {
    FD_CWARN << "FdConfig::DetermineEyeId() - "
             << "fEyeId= " << fEyeId << " detected ! Set to " 
	     << fDefaultEyeId /* Fd::kFD_FIRST_EYE */ << " !!!" << endl;
    fEyeId = fDefaultEyeId /* Fd::kFD_FIRST_EYE */;
  }
  
  // --- analyse the network name (Eye LAN)
    
  // Note: might also result in:
  //       hostname --fqdn --> EyePC.eye1-auger.org
  
  if (    fEyeId == fDefaultEyeId /* Fd::kFD_FIRST_EYE */ 
       && !fIsMirror && !fIsEye ) {
    if ( strstr( fFQHostname.c_str(), "fd-leones" ) != NULL ) {
      fEyeId = 1;
      fIsEye = true;
    }
    if ( strstr( fFQHostname.c_str(), "fd-morados" ) != NULL ) {
      fEyeId = 2;
      fIsEye = true;
    }
    if ( strstr( fFQHostname.c_str(), "fd-amarilla" ) != NULL ) {
      fEyeId = 3;
      fIsEye = true;
    }
    if ( strstr( fFQHostname.c_str(), "fd-coihueco" ) != NULL ) {
      fEyeId = 4;
      fIsEye = true;
    }
    if ( strstr( fFQHostname.c_str(), "fd-heat" ) != NULL ) {
      fEyeId = 5;
      fIsEye = true;
    }
    if ( strstr( fFQHostname.c_str(), "fd-karlsruhe" ) != NULL ) {
      fEyeId = 7;
      fIsEye = true;
    }
    
    if ( fIsEye && fEyeId > Fd::kFD_NEYES ) {
      if ( fgPrintLevel )
	FD_CWARN << "FdConfig::DetermineEyeId() - "
	         << "found EyeID= " << fEyeId << " mapped to default EyeID= "
	         << fDefaultEyeId << endl;
      fEyeId = fDefaultEyeId;
    }
  }
}

// -------------------------------------------------------------------------

void FdConfig::DetermineMirrorId(const char *instr)
 {
  string lc_hostname;
  for ( unsigned int i=0; i<strlen( instr ); i++ )
    lc_hostname += tolower( instr[i] );

  // analyse the host name

  char mirror_name[20];
  for ( int i=0; i<6; i++ ) {
    sprintf( mirror_name, "mirrorpc%d", i+1 );
    if ( strstr( lc_hostname.c_str(), mirror_name ) != NULL ) {
      fMirrorId = i + 1;
      fIsMirror = true;
      if ( fgPrintLevel )
	FD_CDEBUG(1) << "Match with Mirror name '" << mirror_name 
	             << "' detected !" << endl;
      break;
    }
  }
}

// -------------------------------------------------------------------------

const char *FdUtil::GetEyeName(bool nospaces)
 {
  return GetEyeName( GetEyeId(), nospaces );
}

// -------------------------------------------------------------------------

const char *FdUtil::GetEyeName(unsigned int eye_id, bool nospaces)
 {
  static char eye_name[40];
  
  switch ( eye_id ) {

    case FdConfig::kID_LosLeones: 
         if ( nospaces ) 
           strcpy( eye_name, "LosLeones" );
         else
	   strcpy( eye_name, "Los Leones" );
	 break;

    case FdConfig::kID_LosMorados:
         if ( nospaces ) 
           strcpy( eye_name, "LosMorados" );
         else
	   strcpy( eye_name, "Los Morados" );
	 break;

    case FdConfig::kID_LomaAmarilla: 
         if ( nospaces ) 
           strcpy( eye_name, "LomaAmarilla" );
         else
           strcpy( eye_name, "Loma Amarilla" );
	 break;

    case FdConfig::kID_Coihueco:
         strcpy( eye_name, "Coihueco" );
	 break;

    case FdConfig::kID_HEAT:
         strcpy( eye_name, "Heat" );
	 break;

    case FdConfig::kID_Karlsruhe:
         strcpy( eye_name, "Karlsruhe" );
	 break;

    default: if ( nospaces ) // nowadays this is called Altue
              strcpy( eye_name, "LosAmigosSupermercados" );
            else
	      strcpy( eye_name, "Los Amigos Supermercados" );
	    break;
  }
  
  return eye_name;
}

// -------------------------------------------------------------------------

const char *FdConfig::GetEyeNameById(unsigned int eye_id,bool nospaces)
 { return FdUtil::GetEyeName( eye_id, nospaces ); }

// -------------------------------------------------------------------------
// -------------------------------------------------------------------------
