
//
// File:    FdGeometry.cc
//
// Purpose: Implementation of the class FdGeometry
//
// $Id: FdGeometry.cc 11528 2010-09-29 02:50:34Z mathes $
//


#include <iostream>
#include <algorithm>
#include <stdexcept>
#include <cstdio>
#include <cstring>
#include <cstdlib>
#include <cmath>
#include <cerrno>
#include <ctype.h>


/** @file FdGeometry.cc
  * Implementation of class FdGeometry.
  * @author H.-J. Mathes, FzK
  */
 
#include <ErrorLogging.hh>
#include <InifileChain.hh>
 
#include "FdGeometry.hh"

#undef DEBUG_MAP

using namespace std;
using FdUtil::FdGeometry;
using FdUtil::Fd;

// -------------------------------------------------------------------------
// -------------------------------------------------------------------------
//
// class FdGeometry - implementation
//

#ifdef DEBUG_MAP
// counts the number of existing objects of type FdGeometry
static int fgObjectsCreated = 0;
#endif // DEBUG_MAP

// flag to indicate that working tables have been initialized
bool FdGeometry::fgClassInitialized = false;

// level of verbosity
unsigned int FdGeometry::fgPrintLevel = 1;

// not yet initialized ...
FdGeometry* FdGeometry::fgGeometry[Fd::kFD_NMIRRORS + 1];

// map axis -> FdGeometry*
std::map<FdGeometry::CameraPointing_t,FdGeometry*,FdGeometry::AxisCompare> FdGeometry::fgGeometryMap;

// -------------------------------------------------------------------------

FdGeometry::FdGeometry(unsigned int abs_telescope_id)
 : fTelescopeId(abs_telescope_id)
 {
#ifdef DEBUG_MAP
  FD_COUT << "FdGeometry::FdGeometry(unsigned int) - obj(" << fgObjectsCreated++ <<  ")= " 
          << hex << this << dec << " created." << endl;
#endif // DEBUG_MAP

  if ( fgPrintLevel > 1 )
    FD_CDEBUG(1) << "FdGeometry::FdGeometry() - fTelescopeId= " 
                 << fTelescopeId << endl;

  // --- initalize static class structures properly
  
  InitializeClass();
  
  // --- parameter checking ...
 
  if ( abs_telescope_id > Fd::kFD_NMIRRORS ) {
    FD_CFATAL << "FdGeometry::FdGeometry(abs_tel_id= " << abs_telescope_id 
              << ") - wrong telescope id!" << endl;
    throw std::logic_error("Illegal absolute telescope id!");
  }
  
  // --- resize vector for later filling ...
  
  fPixelAzimuth.resize( Fd::kMI_NPIXELS );
  fPixelElevation.resize( Fd::kMI_NPIXELS );
  
  // --- special case: abs_telescope_id == 0 --> access static table !?
  
  if ( abs_telescope_id == 0 ) {
    for ( unsigned int nPix=0; nPix<Fd::kMI_NPIXELS; nPix++ ) {
      fPixelAzimuth[nPix]   = fgDefaultPixelAzimuth[nPix];
      fPixelElevation[nPix] = fgDefaultPixelElevation[nPix];
    }
  }
  else {
    GetTelescopeAxis( fTelescopeId, fAxis.fAzimuth, fAxis.fElevation );
    
    CalculateAzElevTable();
  }
  
  // --- add this object to the map
  
  if ( !fgGeometryMap.insert( make_pair(fAxis, this) ).second ) {
    FD_CWARN << "FdGeometry::FdGeometry(unsigned int) - (az,el) = (" 
             << fAxis.fAzimuth << "," << fAxis.fElevation 
	     << ") is already in map!" << endl;
  }
#ifdef DEBUG_MAP
  else
    FD_COUT << "FdGeometry::FdGeometry(unsigned int) - obj= " 
            << hex << this << dec << " (az,el) = ("
	    <<  fAxis.fAzimuth << "," << fAxis.fElevation
	    << ") added to map." << endl;
#endif // DEBUG_MAP
}

// -------------------------------------------------------------------------

FdGeometry::FdGeometry(CameraPointing_t& axis)
 : fTelescopeId(99), fAxis(axis)
 {
#ifdef DEBUG_MAP
  FD_COUT << "FdGeometry::FdGeometry(CameraPointing_t&) - obj(" << fgObjectsCreated++ <<  ")= " 
  	  << hex << this << dec << " created." << endl;
#endif // DEBUG_MAP

  if ( fgPrintLevel > 1 )
    FD_CDEBUG(1) << "FdGeometry::FdGeometry(CameraPointing_t&) - (az,el) = (" 
                 << axis.fAzimuth << "," << axis.fElevation << ")" << endl;

  // --- initalize static class structures properly
  
  InitializeClass();
  
  // --- resize vector for later filling ...
  
  fPixelAzimuth.resize( Fd::kMI_NPIXELS );
  fPixelElevation.resize( Fd::kMI_NPIXELS );

  CalculateAzElevTable();

  // --- add this object to the map
  
  if ( !fgGeometryMap.insert( make_pair(fAxis, this) ).second ) {
    FD_CWARN << "FdGeometry::FdGeometry() - (az,el) = (" 
             << fAxis.fAzimuth << "," << fAxis.fElevation 
	     << ") is already in map!" << endl;
  }
#ifdef DEBUG_MAP
  else
    FD_COUT << "FdGeometry::FdGeometry() - obj= " 
            << hex << this << dec << " (az,el) = ("
	    <<  fAxis.fAzimuth << "," << fAxis.fElevation
	    << ") added to map." << endl;
#endif // DEBUG_MAP
}

// -------------------------------------------------------------------------

FdGeometry::~FdGeometry()
 {
  if ( fgPrintLevel > 1 ) {
    FD_CDEBUG(1) << "FdGeometry::~FdGeometry() - fTelescopeId= " 
                 << fTelescopeId << endl;
  
    FD_CDEBUG(1) << "\t\t(ax,el) = (" << fAxis.fAzimuth << ","
                 << fAxis.fElevation << ")" << endl;
  }
  
  // --- remove this object from the map, locate via (az,el)
  
  map<CameraPointing_t,FdGeometry*>::iterator found;
  
  found = fgGeometryMap.find( fAxis );

#ifdef DEBUG_MAP
  if ( found == fgGeometryMap.end() ) {
    FD_COUT << "FdGeometry::~FdGeometry() - map could not 'find' obj (az,el) = ("
	    <<  fAxis.fAzimuth << "," << fAxis.fElevation
	    << ") in map." << endl;
  }
#endif // DEBUG_MAP
  
  // --- if not found, try to locate it via its value_type
  
  if ( found == fgGeometryMap.end() ) {
    
    map<CameraPointing_t,FdGeometry*>::iterator iter = fgGeometryMap.begin();
    
    while ( iter != fgGeometryMap.end() ) {
      if ( iter->second == this ) {
        found = iter;
	break;
      }
      iter++;
    }
  }
  
  if ( found != fgGeometryMap.end() ) {

    if ( fgPrintLevel > 1 )
      FD_CDEBUG(1) << "FdGeometry::~FdGeometry() - erasing (ax,el) = ("
                   << fAxis.fAzimuth << ","
		   << fAxis.fElevation << ") from map!" << endl;
#ifdef DEBUG_MAP
    FD_COUT << "FdGeometry::~FdGeometry() - removing obj(" << --fgObjectsCreated <<  ")= "
                 << hex << this << dec << " (az,el) = ("
	    <<  fAxis.fAzimuth << "," << fAxis.fElevation
	    << ") from map." << endl;
#endif // DEBUG_MAP

    fgGeometryMap.erase( found );
  }
  else {
    FD_CWARN << "FdGeometry::~FdGeometry() - (ax,el) = ("
                   << fAxis.fAzimuth << ","
		   << fAxis.fElevation << ") not found in map!" << endl;
  }

#ifdef DEBUG_MAP
    FD_COUT << "FdGeometry::~FdGeometry() - deleting obj= "
            << hex << this << dec << endl;
#endif // DEBUG_MAP
}

// -------------------------------------------------------------------------

const FdGeometry& FdGeometry::GetInstance(unsigned int abs_tel_id)
 {
  // --- initalize static class structures properly
  
  InitializeClass();
  
  // --- parameter checking ...
  
  if ( abs_tel_id > Fd::kFD_NMIRRORS ) {
    FD_CCRIT << "FdGeometry::GetInstance(" << abs_tel_id << ") - "
             << "telescope number range exceeded!" << endl;
    throw std::range_error( "FD telescope number, range exceeded");
  }
  
  //FD_COUT << "FdGeometry::GetInstance(" << abs_tel_id << ") - "
  //        << fgGeometry[abs_tel_id] << endl;

  if ( fgGeometry[abs_tel_id] == NULL )
    fgGeometry[abs_tel_id] = new FdGeometry( abs_tel_id );

#ifdef DEBUG_MAP
  FD_COUT << "FdGeometry::GetInstance(int) - returns obj= "
          << hex << fgGeometry[abs_tel_id] << dec << " (az,el) = ("
	    << fgGeometry[abs_tel_id]->fAxis.fAzimuth << "," 
	    << fgGeometry[abs_tel_id]->fAxis.fElevation
	    << ")" << endl;
#endif // DEBUG_MAP

  return *fgGeometry[abs_tel_id];
}

// -------------------------------------------------------------------------

const FdGeometry& FdGeometry::GetInstance(unsigned int eye_no,
                                          unsigned int mirror_no)
 {
  // --- initalize static class structures properly
  
  InitializeClass();
  
  // --- parameter checking ...

  if ( eye_no < Fd::kFD_FIRST_EYE || eye_no > Fd::kFD_LAST_EYE ) {
    FD_CCRIT << "FdGeometry::GetInstance(" << eye_no << ","
             << mirror_no << ") - eye number outside valid range!" << endl;
    throw std::range_error( "Eye number exceeds valid range" );
  }
  
  if ( mirror_no < Fd::kEYE_FIRST_MIRROR || mirror_no > Fd::kEYE_LAST_MIRROR ) {
    FD_CCRIT << "FdGeometry::GetInstance(" << eye_no << ","
             << mirror_no << ") - mirror number outside valid range!" << endl;
    throw std::range_error( "Mirror number exceeds valid range" );
  }
  
  Fd::MirrorNumberRec mirrorNum;
  
  Fd::SetEyeNo( &mirrorNum, eye_no );
  Fd::SetMirrorNo( &mirrorNum, mirror_no );
  
#if 0
  return GetInstance( &mirrorNum );
#else
  return GetInstance( Fd::GetFdMirrorNo( &mirrorNum ) );
#endif
}

// -------------------------------------------------------------------------

#if 0
const FdGeometry& FdGeometry::GetInstance(Fd::MirrorNumber mirrorNum)
 {
  unsigned int abs_tel_id = Fd::GetFdMirrorNo( mirrorNum );
  
  return GetInstance( abs_tel_id );
}
#endif

// -------------------------------------------------------------------------

const FdGeometry& FdGeometry::GetInstance(CameraPointing_t& axis)
 {
  // --- initalize static class structures properly
  
  InitializeClass();
  
#ifdef DEBUG_MAP
  static bool map_displayed = false;
  
  if ( !map_displayed ) {
    
    FD_COUT << "FdGeometry::GetInstance(CameraPointing_t&) - "
            << "showing map, size= " << fgGeometryMap.size() << endl;
  
    map<CameraPointing_t,FdGeometry*>::iterator iter = fgGeometryMap.begin();
    
    while ( iter != fgGeometryMap.end() ) {
      FD_COUT << "FdGeometry::GetInstance(CameraPointing_t&) - obj= "
              << hex << iter->second << dec << "(az,el) = ("
              << iter->first.fAzimuth << "," 
	      << iter->first.fElevation << ")" << endl;
      iter++;
    }
    
    map_displayed = true;
  }
#endif // DEBUG_MAP

  map<CameraPointing_t,FdGeometry*>::iterator found;
  
  found = fgGeometryMap.find( axis );

#ifdef DEBUG_MAP
  if ( found == fgGeometryMap.end() ) { 
    FD_COUT << "FdGeometry::GetInstance(CameraPointing_t&) - "
            << "cannot 'find' (az,el) = ("
	    << axis.fAzimuth << "," << axis.fElevation
	    << ")" << endl;
  }
#endif // DEBUG_MAP

  //found = find( fgGeometryMap.begin(), fgGeometryMap.end(), axis );
  //found = find_if( fgGeometryMap.begin(), fgGeometryMap.end(), 
  //                 bind2nd( fgGeometryMap.key_comp(), axis ) );
  
  if ( found == fgGeometryMap.end() ) {  // !found -> create a new one
  
    FdGeometry * geo = new FdGeometry( axis );
    
#ifdef DEBUG_MAP
  FD_COUT << "FdGeometry::GetInstance(CameraPointing_t&) - created obj= "
          << hex << geo << dec << " (az,el) = ("
	  << geo->fAxis.fAzimuth << "," << geo->fAxis.fElevation
	  << ")" << endl;
#endif // DEBUG_MAP

    return *geo;
  }
  
#ifdef DEBUG_MAP
  FD_COUT << "FdGeometry::GetInstance(CameraPointing_t&)) - returns obj= "
          << hex << (found->second) << dec << " (az,el) = ("
	  <<  found->second->fAxis.fAzimuth << "," 
	  << found->second->fAxis.fElevation
	  << ") from map." << endl;
#endif // DEBUG_MAP

  return *(found->second);
}

// -------------------------------------------------------------------------

void FdGeometry::CalculateAzElevTable()
 {
  if ( !fgClassInitialized ) InitializeClass();

  const double deg = acos(0.)/90.;
  
  // axisPhi & axisTheta are in degrees
  double axisPhi = fAxis.fAzimuth, axisTheta = fAxis.fElevation;
  
#ifdef DESIGN_GEOMETRY
  // 1 ... kFD_NEYES
  unsigned int eye_id = (fTelescopeId - Fd::kFD_FIRST_MIRROR) / Fd::kEYE_NMIRRORS + Fd::kFD_FIRST_EYE;
#endif // DESIGN_GEOMETRY

  if ( fgPrintLevel > 1 )
    FD_CDEBUG(1) << "FdGeometry::CalculateAzElevTable() - " 
                 << "fTelescopeId= " << fTelescopeId << " (az,el) = (" 
		 << axisPhi << "," << axisTheta << ")" << endl;

  for ( unsigned int col=Fd::kMI_FIRST_COL; col<=Fd::kMI_NCOLS; col++ ) {
  
    for ( unsigned int row=Fd::kMI_FIRST_ROW; row<=Fd::kMI_NROWS; row++ ) {
    
      double a1 = ((double)row - 11.5) * sqrt(3.0) * 0.75 * deg;
      double b2 = (10.0 - col) * 1.5 * deg + 0.75 * deg * (double)((row + 1) % 2);

      double elev = asin( sin(a1 + (double)axisTheta * deg - 0.2165 * deg) * cos( b2 ) );
      double az = asin( sin( b2 ) / cos( elev ) ) + (double)axisPhi * deg;

      unsigned int iPix =   Fd::kMI_NROWS * (col - Fd::kMI_FIRST_COL)
                          + row - Fd::kMI_FIRST_ROW;
    
      fPixelAzimuth[iPix]   = az / deg;
      
#ifdef DESIGN_GEOMETRY
      if ( eye_id <= 4 )
        fPixelAzimuth[iPix] += fgEyeGeometry[eye_id - Fd::kFD_FIRST_EYE].fAngleToEast;
      else {
        switch ( fTelescopeId ) {
	  case 25: az += fgCameraGeometry[25].fAxis.fAzimuth;
	           break;
	  case 26: az += fgCameraGeometry[26].fAxis.fAzimuth;
	           break;
	  case 27: az += fgCameraGeometry[27].fAxis.fAzimuth;
	           break;
          default: FD_CERR << "FdGeometry::CalculateAzElevTable() - "
	                   << "the telescope id exceeds the maximum id "
			   << Fd::kFD_NMIRRORS << endl;
	}
      }
#endif // DESIGN_GEOMETRY

      fPixelElevation[iPix] = elev / deg;
    }
  }
}

// -------------------------------------------------------------------------

void FdGeometry::ExitHandler()
 {
#ifdef DEBUG_MAP
  if ( 1 ) {
#else
  if ( fgPrintLevel > 1 ) {
#endif // DEBUG_MAP
    FD_CINFO << "FdGeometry::ExitHandler() - map size= " 
             << fgGeometryMap.size() << endl;
  }

  while ( !fgGeometryMap.empty() ) {
#ifdef DEBUG_MAP
    FD_COUT << "FdGeometry::ExitHandler() - going to delete obj= "
            << hex << fgGeometryMap.begin()->second << dec << " (az,el) = ("
	    << fgGeometryMap.begin()->second->fAxis.fAzimuth << "," 
	    << fgGeometryMap.begin()->second->fAxis.fElevation
	    << ")" << endl;
#endif // DEBUG_MAP
    delete fgGeometryMap.begin()->second;
#ifdef DEBUG_MAP
    FD_COUT << "FdGeometry::ExitHandler() - map size= " 
            << fgGeometryMap.size() << endl;
#endif // DEBUG_MAP
  }
}

// -------------------------------------------------------------------------

double FdGeometry::GetPixelAzimuth(unsigned int col,unsigned int row) const
 {
  if ( col < Fd::kMI_FIRST_COL || col > Fd::kMI_LAST_COL )
    throw std::range_error( "pixel column number invalid" );
  if ( row < Fd::kMI_FIRST_ROW || row > Fd::kMI_LAST_ROW )
    throw std::range_error( "pixel row number invalid" );

  unsigned int idx = Fd::kMI_NROWS * (col - Fd::kMI_FIRST_COL)
                     + row - Fd::kMI_FIRST_ROW;

  return fPixelAzimuth[idx];
}

// -------------------------------------------------------------------------

double FdGeometry::GetPixelAzimuth(const Fd::PixelNumber pixel)
 {
  return GetPixelAzimuth( Fd::GetEyeNo(pixel), Fd::GetMirrorNo(pixel),
                          Fd::GetColumnNo(pixel), Fd::GetRowNo(pixel) );
}

// -------------------------------------------------------------------------

double FdGeometry::GetPixelAzimuth(unsigned int eye,unsigned int mirror,
                                  unsigned int col,unsigned int row)
 {
  if ( eye < Fd::kFD_FIRST_EYE || eye > Fd::kFD_LAST_EYE )
    throw std::range_error( "Eye number invalid" );
  if ( mirror < Fd::kEYE_FIRST_MIRROR || mirror > Fd::kEYE_LAST_MIRROR )
    throw std::range_error( "Mirror number invalid" );
#if 1
  if ( col < Fd::kMI_FIRST_COL || col > Fd::kMI_LAST_COL )
    throw std::range_error( "pixel column number invalid" );
  if ( row < Fd::kMI_FIRST_ROW || row > Fd::kMI_LAST_ROW )
    throw std::range_error( "pixel row number invalid" );
#endif

  //unsigned int idx = Fd::kMI_NROWS * (col - Fd::kMI_FIRST_COL)
  //                   + row - Fd::kMI_FIRST_ROW;

#if 0
  if ( idx >= Fd::kMI_LAST_PIXEL )
    throw std::range_error("maximum pixel number exceeded");
#endif

#if 1
  const FdGeometry& geom = FdGeometry::GetInstance( eye, mirror );
  
  //return fgGeometry[].fPixelAzimuth[idx];

  double azimuth = geom.GetPixelAzimuth( col, row );  
#else
  double azimuth =   fgDefaultPixelAzimuth[idx] 
                   + 15.0
		   + 30.0 * ( mirror - Fd::kEYE_FIRST_MIRROR ) 
		   - fgEyeGeometry[eye - Fd::kFD_FIRST_EYE].fAngleToEast;
#endif

  return azimuth;
}

// -------------------------------------------------------------------------

double FdGeometry::GetPixelAzimuth0(unsigned int col,unsigned int row)
 {
#if 1
  if ( col < Fd::kMI_FIRST_COL || col > Fd::kMI_LAST_COL )
    throw std::range_error( "pixel column number invalid" );
  if ( row < Fd::kMI_FIRST_ROW || row > Fd::kMI_LAST_ROW )
    throw std::range_error( "pixel row number invalid" );
#endif

  unsigned int idx = Fd::kMI_NROWS * (col - Fd::kMI_FIRST_COL)
                     + row - Fd::kMI_FIRST_ROW;

#if 0
  if ( idx >= Fd::kMI_LAST_PIXEL )
    throw std::range_error("maximum pixel number exceeded");
#endif

  return fgDefaultPixelAzimuth[idx];
}

// -------------------------------------------------------------------------

double FdGeometry::GetPixelElevation(unsigned int col,unsigned int row) const
 {
  if ( col < Fd::kMI_FIRST_COL || col > Fd::kMI_LAST_COL )
    throw std::range_error( "pixel column number invalid" );
  if ( row < Fd::kMI_FIRST_ROW || row > Fd::kMI_LAST_ROW )
    throw std::range_error( "pixel row number invalid" );

  unsigned int idx = Fd::kMI_NROWS * (col - Fd::kMI_FIRST_COL)
                     + row - Fd::kMI_FIRST_ROW;

  return fPixelElevation[idx];
}

// -------------------------------------------------------------------------

double FdGeometry::GetPixelElevation(const Fd::PixelNumber pixel)
 {
  return GetPixelElevation( Fd::GetEyeNo(pixel), Fd::GetMirrorNo(pixel),
                            Fd::GetColumnNo(pixel), Fd::GetRowNo(pixel) );
}

// -------------------------------------------------------------------------

double FdGeometry::GetPixelElevation(unsigned int eye,unsigned int mirror,
                                    unsigned int col,unsigned int row)
 {
  if ( eye < Fd::kFD_FIRST_EYE || eye > Fd::kFD_LAST_EYE )
    throw std::range_error( "Eye number invalid" );
  if ( mirror < Fd::kEYE_FIRST_MIRROR || mirror > Fd::kEYE_LAST_MIRROR )
    throw std::range_error( "Mirror number invalid" );
#if 1
  if ( col < Fd::kMI_FIRST_COL || col > Fd::kMI_LAST_COL )
    throw std::range_error( "pixel column number invalid" );
  if ( row < Fd::kMI_FIRST_ROW || row > Fd::kMI_LAST_ROW )
    throw std::range_error( "pixel row number invalid" );
#endif

  //unsigned int idx = Fd::kMI_NROWS * (col - Fd::kMI_FIRST_COL)
  //                   + row - Fd::kMI_FIRST_ROW;

#if 0
  if ( idx >= Fd::kMI_LAST_PIXEL )
    throw std::range_error("maximum pixel number exceeded");
#endif

  const FdGeometry& geom = FdGeometry::GetInstance( eye, mirror );
  
  //return fgDefaultPixelElevation[idx];
  
  double elevation = geom.GetPixelElevation( col, row );

  return elevation;
}

// -------------------------------------------------------------------------

double FdGeometry::GetPixelElevation0(unsigned int col,unsigned int row)
 {
#if 1
  if ( col < Fd::kMI_FIRST_COL || col > Fd::kMI_LAST_COL )
    throw std::range_error( "pixel column number invalid" );
  if ( row < Fd::kMI_FIRST_ROW || row > Fd::kMI_LAST_ROW )
    throw std::range_error( "pixel row number invalid" );
#endif

  unsigned int idx = Fd::kMI_NROWS * (col - Fd::kMI_FIRST_COL)
                     + row - Fd::kMI_FIRST_ROW;

#if 0
  if ( idx >= Fd::kMI_LAST_PIXEL )
    throw std::range_error("maximum pixel number exceeded");
#endif

  return fgDefaultPixelElevation[idx];
}

// -------------------------------------------------------------------------

void FdGeometry::GetPixelPointing(const Fd::PixelNumber pixel,
                                  double& azimuth,double& elevation)
 {
  GetPixelPointing( Fd::GetEyeNo(pixel), Fd::GetMirrorNo(pixel),
                    Fd::GetColumnNo(pixel), Fd::GetRowNo(pixel),
                    azimuth, elevation );
}

// -------------------------------------------------------------------------

void FdGeometry::GetPixelPointing(unsigned int eye,unsigned int mirror,
                                  unsigned int col,unsigned int row,
				  double& azimuth,double& elevation)
 {
  azimuth = GetPixelAzimuth( eye, mirror, col, row );
  elevation = GetPixelElevation( eye, mirror, col, row );
}

// -------------------------------------------------------------------------

void FdGeometry::GetTelescopeAxis(unsigned int abs_telescope_id,
                                  double& az,double& elev)
 {
  if ( !fgClassInitialized ) InitializeClass();

  // --- check passed absolute telescope id
  
  if ( abs_telescope_id > Fd::kFD_NMIRRORS ) {
    FD_CFATAL << "FdGeometry::GetTelescopeAxis(abs_tel_id= " << abs_telescope_id 
              << ") - wrong telescope id!" << endl;
    throw std::logic_error("Illegal absolute telescope id!");
  }

  // special case: abs_telescope_id == 0 is here also handled
  
  az   = fgCameraGeometry[abs_telescope_id].fAxis.fAzimuth;
  elev = fgCameraGeometry[abs_telescope_id].fAxis.fElevation;
}

// -------------------------------------------------------------------------

void FdGeometry::GetTelescopeAxis4HEAT(unsigned int heat_telescope_id,
                                       double& az,double& el,
                                       ETiltPosition tilt_position)
 {
  // --- check passed HEAT telescope id
  
  if (    heat_telescope_id < Fd::kEYE_FIRST_MIRROR
       || heat_telescope_id > Fd::kHEAT_NMIRRORS ) {
    FD_CFATAL << "FdGeometry::GetTelescopeAxis4HEAT(heat_tel_id= " 
              << heat_telescope_id 
              << ") - wrong HEAT telescope id!" << endl;
    throw std::logic_error("Illegal HEAT telescope id!");
  }
  
  unsigned int idx;
  
  switch ( tilt_position ) {
    case kPositionDown:
         idx = 0;
         break;
    case kPositionUp:
         idx = 1;
         break;
    case kPositionUnknown:
    default:
         idx = 2;
  }
  
  az = fgHeatCameraPointing[heat_telescope_id - Fd::kEYE_FIRST_MIRROR][idx].fAzimuth;
  el = fgHeatCameraPointing[heat_telescope_id - Fd::kEYE_FIRST_MIRROR][idx].fElevation;
}

// -------------------------------------------------------------------------

double FdGeometry::GetTelescopeAzimuth(unsigned int abs_telescope_id)
 {
  double az, el;
  
  GetTelescopeAxis( abs_telescope_id, az, el );
  
  return az;
}

// -------------------------------------------------------------------------

double FdGeometry::GetTelescopeElevation(unsigned int abs_telescope_id)
 {
  double az, el;
  
  GetTelescopeAxis( abs_telescope_id, az, el );
  
  return el;
}

// -------------------------------------------------------------------------

void FdGeometry::InitializeClass()
 {
  if ( fgPrintLevel > 1 )
    FD_COUT << "FdGeometry::InitializeClass() - fgClassInitialized= " 
            << (int)fgClassInitialized << endl;
  
  if ( fgClassInitialized ) return;
  
  // --- initialize vector[telesecope id] -> FdGeometry
  
  if ( fgPrintLevel )
    FD_COUT << "FdGeometry::InitializeClass( - initializing ..." << endl;
  
  for ( int tel=0; tel<Fd::kFD_NMIRRORS + 1; tel++ )
    fgGeometry[tel] = NULL;

  // --- check for and install exit handler

  if ( atexit( ExitHandler ) < 0 ) {
    FD_CERR << "FdGeometry::InitializeClass() - "
  	    << "trying installing 'atexit' handler with error, "
  	    << errno << " - " << strerror( errno ) << endl;
  }
  else if ( fgPrintLevel ) {
    FD_COUT << "FdGeometry::InitializeClass() - "
  	     << "'atexit' handler successfully installed !" << endl;
  }
  
  InitializeTables();
  
  fgClassInitialized = true;
  
  // --- populate vector[telesecope id] -> FdGeometry
  
  // Note: If this part is omitted, in some applications we might have a small
  //       memory leak as objects added to the map via the method
  //       GetInstance(FdCameraGeometry&) have no valid telescope id but might 
  //       have the same pointing as any other axis object
  //       --> it is probably a good choice to use the telescope id also
  //           for comparism !?

  for ( int abs_tel_id=0; abs_tel_id<Fd::kFD_NMIRRORS+1; abs_tel_id++ )
    fgGeometry[abs_tel_id] = new FdGeometry( abs_tel_id );
}

// -------------------------------------------------------------------------

void FdGeometry::InitializeTables()
 {
  if ( fgPrintLevel > 1 )
    FD_COUT << "FdGeometry::InitializeTables() -" << endl;
  
  for ( unsigned int tel_id=0; tel_id<=Fd::kFD_NMIRRORS; tel_id++ ) {

    switch ( tel_id ) {
    
      case 25: /* use initially the default values for HEAT */
      case 26:
      case 27:
         fgCameraGeometry[tel_id].fAxis = fgHeatCameraPointing[tel_id-25][2];
         break;

      default:  /* 0..24 */
         fgCameraGeometry[tel_id].fAxis = fgDefaultCameraPointing[tel_id];
    }
  }

  // --- debug: dump values for HEAT

  if ( fgPrintLevel > 1 ) {
  
    FD_COUT << "FdGeometry::InitializeTables() - dump for HEAT" << endl;
    
    for ( int i=0; i<3; i++ ) {
      
      FD_COUT << "HEAT configuration " << i+1 << ":" << endl;
      
      for ( int j=0; j<3; j++ )
        FD_COUT << "HEAT[" << j+1 << "] (az,el) = ("
                << fgHeatCameraPointing[j][i].fAzimuth << "," 
	        << fgHeatCameraPointing[j][i].fElevation << ")" << endl;
    }
    
  } // fgPrintLevel
}

// -------------------------------------------------------------------------

bool FdGeometry::ReadCameraPointing(FdUtil::InifileChain& inifile)
 {
  if ( !fgClassInitialized ) InitializeClass();
  
  if ( fgPrintLevel > 1 )
    FD_CDEBUG(1) << "FdGeometry::ReadCameraPointing() -" << endl;

  const char *section = "Geometry";
  
  if ( inifile.SpecifyGroup( section ) != Inifile::kSUCCESS ) {
    FD_CWARN << "FdGeometry::ReadCameraPointing() - "
             << "Section [" << section << "] not found !!!" << endl;
    return false;
  }
  
  char tag[10];
  int error = 0;
  double az, el;
  
  for ( unsigned int mi_no=Fd::kFD_FIRST_MIRROR;
                     mi_no<=Fd::kFD_LAST_MIRROR; mi_no++ ) {
    
    // get hard-coded default first as fallback
    
    GetTelescopeAxis( mi_no, az, el );
    
    bool auto_mode = false;
    
    sprintf( tag, "axis%u", mi_no );
    
    string mode = inifile.GetFirstString( tag, "" );
    if ( !mode.empty() ) {
      for ( size_t i=0; i<mode.length(); i++ )
        mode[i] = tolower( mode[i] );
    
      if ( mode == "auto" ) auto_mode = true;
    }
    
    if ( auto_mode ) // look for fallback defaults
      az = inifile.GetNextFloat( az, &error );
    else
      az = inifile.GetFirstFloat( tag, az, &error );
    //if ( error ) continue;
    
    el = inifile.GetNextFloat( el, &error );
    //if ( error ) continue;
    
    if ( auto_mode ) {
      GetAxisFunction_t get_axis_func = fgCameraGeometry[mi_no].fGetAxisFunc;
    
      if ( get_axis_func != NULL )
        get_axis_func( mi_no, az, el );
      else {
        FD_CWARN << "FdGeometry::ReadCameraPointing() - "
               << "No function for reading axis pointing of telescope " 
	       << mi_no << " was provided, " 
	       << "using fallback constants instead!"
	       << endl;
      }
    }
    
    SetTelescopeAxis( mi_no, az, el, auto_mode );
  }
  
  ReadCameraPointing4HEAT( inifile );
  
  return true;
}

// -------------------------------------------------------------------------

bool FdGeometry::ReadCameraPointing4HEAT(FdUtil::InifileChain& inifile)
 {
  if ( fgPrintLevel > 1 )
    FD_CDEBUG(1) << "FdGeometry::ReadCameraPointing4HEAT() -" << endl;

  const char *section = "Geometry";
  
  if ( inifile.SpecifyGroup( section ) != Inifile::kSUCCESS ) {
    FD_CWARN << "FdGeometry::ReadCameraPointing4HEAT() - "
             << "Section [" << section << "] not found !!!" << endl;
    return false;
  }
  
  char tag[10];
  int error = 0;
  double az, el;
  
  for ( unsigned int mi_no=Fd::kEYE_FIRST_MIRROR;
                     mi_no<=Fd::kHEAT_NMIRRORS; mi_no++ ) {
    
    sprintf( tag, "heat%u", mi_no );
    
    // try to read values for down position...
    az = inifile.GetFirstFloat( tag, 0., &error );
    if ( !error )
      el = inifile.GetNextFloat( 0., &error );
      
    if ( !error ) {
      fgHeatCameraPointing[mi_no - Fd::kEYE_FIRST_MIRROR][0].fAzimuth = az;
      fgHeatCameraPointing[mi_no - Fd::kEYE_FIRST_MIRROR][0].fElevation = el;
    }
    
    // try to read values for up position...
    if ( !error )
      az = inifile.GetNextFloat( 0., &error );
    if ( !error )
      el = inifile.GetNextFloat( 0., &error );
      
    if ( !error ) {
      fgHeatCameraPointing[mi_no - Fd::kEYE_FIRST_MIRROR][1].fAzimuth = az;
      fgHeatCameraPointing[mi_no - Fd::kEYE_FIRST_MIRROR][1].fElevation = el;
    }
    
    // try to read values for default position...
    if ( !error )
      az = inifile.GetNextFloat( 0., &error );
    if ( !error )
      el = inifile.GetNextFloat( 0., &error );
      
    if ( !error ) {
      fgHeatCameraPointing[mi_no - Fd::kEYE_FIRST_MIRROR][2].fAzimuth = az;
      fgHeatCameraPointing[mi_no - Fd::kEYE_FIRST_MIRROR][2].fElevation = el;
      
      // store default position also in fgCameraGeometry
      SetTelescopeAxis( 24 + mi_no, az, el, 
                        fgCameraGeometry[24 + mi_no].fAutomatic );
    }
  }
  
  // --- debug: dump values for HEAT

  if ( fgPrintLevel > 1 ) {
  
    FD_COUT << "FdGeometry::ReadCameraPointing4HEAT() - dump for HEAT" << endl;
    
    for ( int i=0; i<3; i++ ) {
      
      FD_COUT << "HEAT configuration " << i+1 << ":" << endl;
      
      for ( int j=0; j<3; j++ )
        FD_COUT << "HEAT[" << j+1 << "] (az,el) = ("
                << fgHeatCameraPointing[j][i].fAzimuth << "," 
	        << fgHeatCameraPointing[j][i].fElevation << ")" << endl;
    }
    
  } // fgPrintLevel

  return true;
}

// -------------------------------------------------------------------------

void FdGeometry::SetGetAxisFunction(unsigned int abs_tel_id,
                                    GetAxisFunction_t get_axis_func)
 {
  // --- check passed absolute telescope id
  
  if ( abs_tel_id == 0 || abs_tel_id > Fd::kFD_NMIRRORS ) {
    FD_CFATAL << "FdGeometry::SetGetAxisFunction(abs_tel_id= " << abs_tel_id 
              << ") - wrong telescope id!" << endl;
    throw std::logic_error("Illegal absolute telescope id!");
  }
  
  fgCameraGeometry[abs_tel_id].fGetAxisFunc = get_axis_func;  

  if ( !get_axis_func )
    FD_CWARN << "FdGeometry::SetGetAxisFunction(abs_tel_id= " 
             << abs_tel_id << " function pointer deleted!" << endl;
}

// -------------------------------------------------------------------------

void FdGeometry::SetTelescopeAxis(unsigned int abs_tel_id)
 {
  if ( !fgClassInitialized ) InitializeClass();

  if ( fgPrintLevel > 1 )
    FD_CDEBUG(1) << "FdGeometry::SetTelescopeAxis(abs_tel_id= " 
                 << abs_tel_id << ") -" << endl;

  // --- check passed absolute telescope id
  
  if ( abs_tel_id == 0 || abs_tel_id > Fd::kFD_NMIRRORS ) {
    FD_CFATAL << "FdGeometry::SetTelescopeAxis(abs_tel_id= " << abs_tel_id 
              << ") - wrong telescope id!" << endl;
    throw std::logic_error("Illegal absolute telescope id!");
  }
  
  double az, el;
  
  // get hard-coded defaults first as fallback
  
  //GetTelescopeAxis( abs_tel_id, az, el );
  if ( abs_tel_id <= 24 ) {
    az = fgDefaultCameraPointing[abs_tel_id].fAzimuth;
    el = fgDefaultCameraPointing[abs_tel_id].fElevation;
  }
  else {
    az = fgHeatCameraPointing[abs_tel_id - 25][2].fAzimuth;
    el = fgHeatCameraPointing[abs_tel_id - 25][2].fElevation;
  }
  
  bool auto_mode = fgCameraGeometry[abs_tel_id].fAutomatic;
  
  if ( auto_mode ) {
    GetAxisFunction_t get_axis_func = fgCameraGeometry[abs_tel_id].fGetAxisFunc;
  
    if ( get_axis_func != NULL )
      get_axis_func( abs_tel_id, az, el );
    else {
      FD_CWARN << "FdGeometry::SetTelescopeAxis() - "
  	     << "No function for reading axis pointing of telescope " 
	     << abs_tel_id << " was provided, "
             << "using fallback constants instead!"
             << endl;
    }
  }
    
  SetTelescopeAxis( abs_tel_id, az, el, auto_mode );
}

// -------------------------------------------------------------------------

void FdGeometry::SetTelescopeAxis(unsigned int abs_telescope_id,
                                  double az,double el,bool auto_mode)
 {  
  if ( !fgClassInitialized ) InitializeClass();

  if ( fgPrintLevel > 1 )
    FD_CDEBUG(1) << "FdGeometry::SetTelescopeAxis() - "
                 << "abs_tel_id= " << abs_telescope_id << " (az,el)= "
                 << "(" << az << "," << el << ") auto= " << auto_mode 
		 << endl;
    
  // --- check passed absolute telescope id
  
  if ( abs_telescope_id == 0 || abs_telescope_id > Fd::kFD_NMIRRORS ) {
    FD_CFATAL << "FdGeometry::SetTelescopeAxis(abs_tel_id= " << abs_telescope_id 
              << ") - wrong telescope id!" << endl;
    throw std::logic_error("Illegal absolute telescope id!");
  }
  
  // --- check if anything was changed
  
  double curr_az = fgCameraGeometry[abs_telescope_id].fAxis.fAzimuth;
  double curr_el = fgCameraGeometry[abs_telescope_id].fAxis.fElevation;
  
  fgCameraGeometry[abs_telescope_id].fAutomatic = auto_mode;
  
  bool changed = false;
  
  if ( (fabs(curr_az - az) > 0.001) || (fabs(curr_el - el) > 0.001) )
    changed = true;

  if ( !changed ) return;
  
  FD_CWARN << "FdGeometry::SetTelescopeAxis(" << abs_telescope_id << ") - "
           << "replaced axis (az,el) = (" 
	   << curr_az << "," 
	   << fgCameraGeometry[abs_telescope_id].fAxis.fElevation 
	   << ") by (" << az << "," << el << ") !" << endl;

  if ( auto_mode )
    FD_CWARN << "FdGeometry::SetTelescopeAxis(" << abs_telescope_id << ") - "
             << "mode is set to 'auto' !" << endl;
  
  fgCameraGeometry[abs_telescope_id].fAxis.fAzimuth = az;
  fgCameraGeometry[abs_telescope_id].fAxis.fElevation = el;
  fgCameraGeometry[abs_telescope_id].fAutomatic = auto_mode;
  
  // this will retrigger the calculation of all pixel pointings, too
  
  if ( fgGeometry[abs_telescope_id] != NULL ) {
    delete fgGeometry[abs_telescope_id];
    fgGeometry[abs_telescope_id] = new FdGeometry( abs_telescope_id );
  }
}

// -------------------------------------------------------------------------
// ---              Table with FD Eye geometry constants                 ---
// -------------------------------------------------------------------------

// see also: Offline/trunk/Framework/FDetector/FTelescopeList.xml
//
// 'fAngleToEast' values are update with those one from ArgenSig 2005
//

const FdGeometry::EyeGeometryRec FdGeometry::fgEyeGeometry[] = {

  // fLongitude, fLatitude, fGPSAltitude, fFloorAltitude, fAngleToEast
  
  { -69.44968, -35.49576, 1420.48, 1415.00,  -29.995 },
  { -69.01220, -35.29198, 1422.50, 1416.44,   59.992 },
  { -69.21085, -34.93592, 1482.71, 1476.69, -171.865 },
  { -69.59963, -35.11414, 1718.56, 1712.30, -116.697 },
  // backwall based calculation gives only design value for HE2 telescope
  { -69.59803, -35.11345, 1711.60, 1709.00,   41.757 }
};

// -------------------------------------------------------------------------
// ---            Table with camera Eye geometry constants               ---
// -------------------------------------------------------------------------

// convenient macro to be used further down

#define GEO_INIT(_az,_el) FdGeometry::_CameraPointingRec(_az,_el)

// this table is loaded initially and might be modified

FdGeometry::CameraGeometryRec FdGeometry::fgCameraGeometry[] = {

  { GEO_INIT( 0.00, 0.00 ), NULL, false },   // design value for a telescope looking to east, [0]

  { GEO_INIT( 0.00, 0.00 ), NULL, false },   // Los Leones, [1]
  { GEO_INIT( 0.00, 0.00 ), NULL, false },
  { GEO_INIT( 0.00, 0.00 ), NULL, false },
  { GEO_INIT( 0.00, 0.00 ), NULL, false },
  { GEO_INIT( 0.00, 0.00 ), NULL, false },
  { GEO_INIT( 0.00, 0.00 ), NULL, false },
  { GEO_INIT( 0.00, 0.00 ), NULL, false },   // Los Morados [7]
  { GEO_INIT( 0.00, 0.00 ), NULL, false },
  { GEO_INIT( 0.00, 0.00 ), NULL, false },
  { GEO_INIT( 0.00, 0.00 ), NULL, false },
  { GEO_INIT( 0.00, 0.00 ), NULL, false },
  { GEO_INIT( 0.00, 0.00 ), NULL, false },
  { GEO_INIT( 0.00, 0.00 ), NULL, false },   // Loma Amarilla [13]
  { GEO_INIT( 0.00, 0.00 ), NULL, false },
  { GEO_INIT( 0.00, 0.00 ), NULL, false },
  { GEO_INIT( 0.00, 0.00 ), NULL, false },
  { GEO_INIT( 0.00, 0.00 ), NULL, false },
  { GEO_INIT( 0.00, 0.00 ), NULL, false },
  { GEO_INIT( 0.00, 0.00 ), NULL, false },   // Coihueco [19]
  { GEO_INIT( 0.00, 0.00 ), NULL, false },
  { GEO_INIT( 0.00, 0.00 ), NULL, false },
  { GEO_INIT( 0.00, 0.00 ), NULL, false },
  { GEO_INIT( 0.00, 0.00 ), NULL, false },
  { GEO_INIT( 0.00, 0.00 ), NULL, false },

  { GEO_INIT( 0.00, 0.00 ), NULL, false },   // HEAT [25]
  { GEO_INIT( 0.00, 0.00 ), NULL, false },
  { GEO_INIT( 0.00, 0.00 ), NULL, false }

};

//
// I am using here the values found in the Offline code for axis and elevation.
// The axis values which are then stored are calculated using the backwall
// angle. For the standard sites it is taken from the ArgenSig report, for HEAT
// the value from Offline is taken
//
// See also: Offline/trunk/Framework/FDetector/FTelescopeList.xml
//
// Note 1: Use the observatory mirror number (i.e. abs. telescope number) 
//         to address this table!
//
// Note 2: A second set of values is selected by setting -DDESIGN_GEOMETRY
//
// Note 3: The backwall angles used here (ArgenSig) can be converted into the
//         Offline value by subtracting 260 degrees ?.
//

const FdGeometry::CameraPointingRec FdGeometry::fgDefaultCameraPointing[] = {

  GEO_INIT( 0.00, 16.00) ,   // design value for a telescope looking to east, [0]
#ifndef DESIGN_GEOMETRY
  GEO_INIT( (  15.13 -  29.995 ), 16.04 ),   // Los Leones, [1]
  GEO_INIT( (  45.00 -  29.995 ), 16.06 ),
  GEO_INIT( (  74.94 -  29.995 ), 15.96 ),
  GEO_INIT( ( 104.96 -  29.995 ), 16.14 ),
  GEO_INIT( ( 134.93 -  29.995 ), 15.96 ),
  GEO_INIT( ( 164.90 -  29.995 ), 15.85 ),
  GEO_INIT( (  14.94 +  59.992 ), 15.97 ),   // Los Morados [7]
  GEO_INIT( (  44.80 +  59.992 ), 15.81 ),
  GEO_INIT( (  74.79 +  59.992 ), 15.82 ),
  GEO_INIT( ( 104.99 +  59.992 ), 15.75 ),
  GEO_INIT( ( 135.02 +  59.992 ), 15.84 ),
  GEO_INIT( ( 164.92 +  59.992 ), 15.89 ),
  GEO_INIT( (  15.00 - 171.865 ), 16.00 ),   // Loma Amarilla [13]
  GEO_INIT( (  45.00 - 171.865 ), 16.00 ),
  GEO_INIT( (  75.00 - 171.865 ), 16.00 ),
  GEO_INIT( ( 105.00 - 171.865 ), 16.00 ),
  GEO_INIT( ( 135.00 - 171.865 ), 16.00 ),
  GEO_INIT( ( 165.00 - 171.865 ), 16.00 ),
  GEO_INIT( (  15.13 - 116.697 ), 15.94 ),   // Coihueco [19]
  GEO_INIT( (  45.06 - 116.697 ), 16.19 ),
  GEO_INIT( (  74.91 - 116.697 ), 16.21 ),
  GEO_INIT( ( 104.92 - 116.697 ), 16.26 ),
  GEO_INIT( ( 134.96 - 116.697 ), 16.17 ),
  GEO_INIT( ( 164.78 - 116.697 ), 16.23 )
#else // DESIGN_GEOMETRY         // these are the design values
  GEO_INIT(  15.00, 16.00 ),	    // Los Leones, [1]
  GEO_INIT(  45.00, 16.00 ),
  GEO_INIT(  75.00, 16.00 ),
  GEO_INIT( 105.00, 16.00 ),
  GEO_INIT( 135.00, 16.00 ),
  GEO_INIT( 165.00, 16.00 ),
  GEO_INIT(  15.00, 16.00 ),	    // Los Morados [7]
  GEO_INIT(  45.00, 16.00 ),
  GEO_INIT(  75.00, 16.00 ),
  GEO_INIT( 105.00, 16.00 ),
  GEO_INIT( 135.00, 16.00 ),
  GEO_INIT( 165.00, 16.00 ),
  GEO_INIT(  15.00, 16.00 ),	    // Loma Amarilla [13]
  GEO_INIT(  45.00, 16.00 ),
  GEO_INIT(  75.00, 16.00 ),
  GEO_INIT( 105.00, 16.00 ),
  GEO_INIT( 135.00, 16.00 ),
  GEO_INIT( 165.00, 16.00 ),
  GEO_INIT(  15.00, 16.00 ),	    // Coihueco [19]
  GEO_INIT(  45.00, 16.00 ),
  GEO_INIT(  75.00, 16.00 ),
  GEO_INIT( 105.00, 16.00 ),
  GEO_INIT( 135.00, 16.00 ),
  GEO_INIT( 165.00, 16.00 )
#endif // DESIGN_GEOMETRY
};

FdGeometry::CameraPointingRec FdGeometry::fgHeatCameraPointing[][3] = {

  { GEO_INIT(  44.94 - 87.0, 16.00 ),	     // HEAT [25], preliminary from S.Mueller (down)
    GEO_INIT(  44.94 - 87.0, 45.00 ),	     // (upward)
    GEO_INIT(  44.94 - 87.0, 45.00 ) },      // default, if position is unknown

  { GEO_INIT(  90.24 - 87.0, 16.00 ),	     // HEAT [26]
    GEO_INIT(  90.24 - 87.0, 45.00 ),
    GEO_INIT(  90.24 - 87.0, 45.00 ) },

  { GEO_INIT( 135.08 - 87.0, 16.00 ),	     // HEAT [27]
    GEO_INIT( 135.08 - 87.0, 45.00 ),
    GEO_INIT( 135.08 - 87.0, 45.00 ) }
};

// -------------------------------------------------------------------------
// ---         tables for default azimuth and elevation values           ---
// -------------------------------------------------------------------------

// The pixel directions might be calculated according to GAP 2000-032,
//
//  M.Giller et al., Angular directions of pixels in the FD camera
//

const float FdGeometry::fgDefaultPixelAzimuth[] = {

      13.51, 14.27, 13.54, 14.33, 13.61, 14.41, 13.70, 14.52, 13.82, 14.66,
      13.97, 14.83, 14.15, 15.04, 14.36, 15.28, 14.61, 15.56, 14.90, 15.89,
      15.23, 16.26, 12.01, 12.77, 12.04, 12.82, 12.10, 12.89, 12.18, 12.99,
      12.28, 13.12, 12.42, 13.27, 12.58, 13.46, 12.77, 13.68, 13.00, 13.94,
      13.25, 14.23, 13.55, 14.56, 10.51, 11.27, 10.54, 11.31, 10.58, 11.38,
      10.66, 11.46, 10.75, 11.58, 10.87, 11.72, 11.01, 11.88, 11.18, 12.08,
      11.38, 12.30, 11.61, 12.57, 11.87, 12.86,  9.01,  9.77,  9.03,  9.80,
       9.07,  9.86,  9.13,  9.94,  9.22, 10.04,  9.32, 10.16,  9.44, 10.30,
       9.59, 10.47,  9.76, 10.67,  9.95, 10.90, 10.18, 11.16,  7.51,  8.26,
       7.53,  8.30,  7.56,  8.34,  7.61,  8.41,  7.68,  8.49,  7.77,  8.60,
       7.87,  8.72,  7.99,  8.87,  8.14,  9.03,  8.30,  9.23,  8.49,  9.45,
       6.00,  6.76,  6.02,  6.79,  6.05,  6.83,  6.09,  6.88,  6.15,  6.95,
       6.21,  7.03,  6.30,  7.14,  6.40,  7.26,  6.51,  7.40,  6.64,  7.56,
       6.80,  7.74,  4.50,  5.26,  4.52,  5.28,  4.54,  5.31,  4.57,  5.35,
       4.61,  5.41,  4.66,  5.47,  4.72,  5.55,  4.80,  5.65,  4.88,  5.75,
       4.98,  5.88,  5.10,  6.02,  3.00,  3.76,  3.01,  3.77,  3.02,  3.79,
       3.05,  3.82,  3.07,  3.86,  3.11,  3.91,  3.15,  3.97,  3.20,  4.03,
       3.26,  4.11,  3.32,  4.20,  3.40,  4.30,  1.50,  2.25,  1.51,  2.26,
       1.51,  2.28,  1.52,  2.29,  1.54,  2.32,  1.55,  2.35,  1.57,  2.38,
       1.60,  2.42,  1.63,  2.47,  1.66,  2.52,  1.70,  2.58,  0.00,  0.75,
       0.00,  0.75,  0.00,  0.76,  0.00,  0.76,  0.00,  0.77,  0.00,  0.78,
       0.00,  0.79,  0.00,  0.81,  0.00,  0.82,  0.00,  0.84,  0.00,  0.86,
      -1.50, -0.75, -1.51, -0.75, -1.51, -0.76, -1.52, -0.76, -1.54, -0.77,
      -1.55, -0.78, -1.57, -0.79, -1.60, -0.81, -1.63, -0.82, -1.66, -0.84,
      -1.70, -0.86, -3.00, -2.25, -3.01, -2.26, -3.02, -2.28, -3.05, -2.29,
      -3.07, -2.32, -3.11, -2.35, -3.15, -2.38, -3.20, -2.42, -3.26, -2.47,
      -3.32, -2.52, -3.40, -2.58, -4.50, -3.76, -4.52, -3.77, -4.54, -3.79,
      -4.57, -3.82, -4.61, -3.86, -4.66, -3.91, -4.72, -3.97, -4.80, -4.03,
      -4.88, -4.11, -4.98, -4.20, -5.10, -4.30, -6.00, -5.26, -6.02, -5.28,
      -6.05, -5.31, -6.09, -5.35, -6.15, -5.41, -6.21, -5.47, -6.30, -5.55,
      -6.40, -5.65, -6.51, -5.75, -6.64, -5.88, -6.80, -6.02, -7.51, -6.76,
      -7.53, -6.79, -7.56, -6.83, -7.61, -6.88, -7.68, -6.95, -7.77, -7.03,
      -7.87, -7.14, -7.99, -7.26, -8.14, -7.40, -8.30, -7.56, -8.49, -7.74,
      -9.01, -8.26, -9.03, -8.30, -9.07, -8.34, -9.13, -8.41, -9.22, -8.49,
      -9.32, -8.60, -9.44, -8.72, -9.59, -8.87, -9.76, -9.03, -9.95, -9.23,
     -10.18, -9.45,-10.51, -9.77,-10.54, -9.80,-10.58, -9.86,-10.66, -9.94,
     -10.75,-10.04,-10.87,-10.16,-11.01,-10.30,-11.18,-10.47,-11.38,-10.67,
     -11.61,-10.90,-11.87,-11.16,-12.01,-11.27,-12.04,-11.31,-12.10,-11.38,
     -12.18,-11.46,-12.28,-11.58,-12.42,-11.72,-12.58,-11.88,-12.77,-12.08,
     -13.00,-12.30,-13.25,-12.57,-13.55,-12.86,-13.51,-12.77,-13.54,-12.82,
     -13.61,-12.89,-13.70,-12.99,-13.82,-13.12,-13.97,-13.27,-14.15,-13.46,
     -14.36,-13.68,-14.61,-13.94,-14.90,-14.23,-15.23,-14.56,-15.01,-14.27,
     -15.05,-14.33,-15.12,-14.41,-15.22,-14.52,-15.35,-14.66,-15.51,-14.83,
     -15.71,-15.04,-15.95,-15.28,-16.22,-15.56,-16.54,-15.89,-16.90,-16.26
};

const float FdGeometry::fgDefaultPixelElevation[] = {

       2.08,  3.34,  4.61,  5.85,  7.14,  8.37,  9.66, 10.89, 12.18, 13.40,
      14.71, 15.91, 17.23, 18.42, 19.74, 20.93, 22.26, 23.44, 24.77, 25.94,
      27.28, 28.43,  2.10,  3.36,  4.64,  5.89,  7.18,  8.42,  9.72, 10.96,
      12.26, 13.49, 14.80, 16.02, 17.33, 18.54, 19.87, 21.07, 22.40, 23.59,
      24.93, 26.11, 27.46, 28.63,  2.11,  3.38,  4.66,  5.92,  7.22,  8.47,
       9.77, 11.02, 12.32, 13.56, 14.87, 16.11, 17.43, 18.65, 19.97, 21.19,
      22.52, 23.73, 25.07, 26.27, 27.61, 28.80,  2.12,  3.39,  4.68,  5.95,
       7.25,  8.51,  9.81, 11.07, 12.38, 13.63, 14.94, 16.19, 17.51, 18.75,
      20.07, 21.30, 22.63, 23.86, 25.19, 26.41, 27.75, 28.96,  2.13,  3.41,
       4.70,  5.98,  7.28,  8.55,  9.85, 11.12, 12.43, 13.69, 15.00, 16.26,
      17.58, 18.83, 20.15, 21.39, 22.72, 23.96, 25.29, 26.53, 27.86, 29.09,
       2.13,  3.42,  4.72,  6.00,  7.30,  8.58,  9.88, 11.16, 12.47, 13.74,
      15.05, 16.32, 17.63, 18.89, 20.21, 21.47, 22.80, 24.05, 25.38, 26.62,
      27.96, 29.20,  2.14,  3.43,  4.73,  6.02,  7.32,  8.60,  9.91, 11.19,
      12.50, 13.78, 15.09, 16.36, 17.68, 18.95, 20.26, 21.53, 22.85, 24.12,
      25.44, 26.70, 28.03, 29.29,  2.14,  3.44,  4.74,  6.03,  7.33,  8.62,
       9.92, 11.21, 12.52, 13.80, 15.11, 16.40, 17.71, 18.99, 20.30, 21.58,
      22.89, 24.17, 25.49, 26.76, 28.08, 29.35,  2.14,  3.44,  4.74,  6.04,
       7.34,  8.63,  9.93, 11.23, 12.53, 13.82, 15.13, 16.42, 17.73, 19.02,
      20.32, 21.61, 22.92, 24.21, 25.52, 26.80, 28.11, 29.40,  2.14,  3.44,
       4.74,  6.04,  7.34,  8.64,  9.94, 11.24, 12.54, 13.83, 15.13, 16.43,
      17.73, 19.03, 20.33, 21.63, 22.93, 24.22, 25.53, 26.82, 28.12, 29.42,
       2.14,  3.44,  4.74,  6.04,  7.34,  8.64,  9.93, 11.24, 12.53, 13.83,
      15.13, 16.43, 17.73, 19.03, 20.32, 21.63, 22.92, 24.22, 25.52, 26.82,
      28.11, 29.42,  2.14,  3.44,  4.74,  6.04,  7.33,  8.63,  9.92, 11.23,
      12.52, 13.82, 15.11, 16.42, 17.71, 19.02, 20.30, 21.61, 22.89, 24.21,
      25.49, 26.80, 28.08, 29.40,  2.14,  3.44,  4.73,  6.03,  7.32,  8.62,
       9.91, 11.21, 12.50, 13.80, 15.09, 16.40, 17.68, 18.99, 20.26, 21.58,
      22.85, 24.17, 25.44, 26.76, 28.03, 29.35,  2.13,  3.43,  4.72,  6.02,
       7.30,  8.60,  9.88, 11.19, 12.47, 13.78, 15.05, 16.36, 17.63, 18.95,
      20.21, 21.53, 22.80, 24.12, 25.38, 26.70, 27.96, 29.29,  2.13,  3.42,
       4.70,  6.00,  7.28,  8.58,  9.85, 11.16, 12.43, 13.74, 15.00, 16.32,
      17.58, 18.89, 20.15, 21.47, 22.72, 24.05, 25.29, 26.62, 27.86, 29.20,
       2.12,  3.41,  4.68,  5.98,  7.25,  8.55,  9.81, 11.12, 12.38, 13.69,
      14.94, 16.26, 17.51, 18.83, 20.07, 21.39, 22.63, 23.96, 25.19, 26.53,
      27.75, 29.09,  2.11,  3.39,  4.66,  5.95,  7.22,  8.51,  9.77, 11.07,
      12.32, 13.63, 14.87, 16.19, 17.43, 18.75, 19.97, 21.30, 22.52, 23.86,
      25.07, 26.41, 27.61, 28.96,  2.10,  3.38,  4.64,  5.92,  7.18,  8.47,
       9.72, 11.02, 12.26, 13.56, 14.80, 16.11, 17.33, 18.65, 19.87, 21.19,
      22.40, 23.73, 24.93, 26.27, 27.46, 28.80,  2.08,  3.36,  4.61,  5.89,
       7.14,  8.42,  9.66, 10.96, 12.18, 13.49, 14.71, 16.02, 17.23, 18.54,
      19.74, 21.07, 22.26, 23.59, 24.77, 26.11, 27.28, 28.63,  2.07,  3.34,
       4.58,  5.85,  7.09,  8.37,  9.60, 10.89, 12.10, 13.40, 14.61, 15.91,
      17.11, 18.42, 19.61, 20.93, 22.10, 23.44, 24.60, 25.94, 27.09, 28.43
};

// -------------------------------------------------------------------------
// -------------------------------------------------------------------------
