
//
// File:    FdGeometry.hh
//
// Purpose: Declaration of the class FdGeometry
//
// $Id: FdGeometry.hh 11528 2010-09-29 02:50:34Z mathes $
//


/** @file FdGeometry.hh
  * Declaration of the class FdGeometry.
  * @author H.-J. Mathes, FzK
  */
  
#ifndef _FdUtil_FdGeometry_hh_
#define _FdUtil_FdGeometry_hh_

#include <cmath>
#include <vector>
#include <memory>
#include <map>
#include <stdexcept>   // std::logic_error()

#ifndef NO_VERSION
# include <FdUtilVersion.hh>
#endif // NO_VERSION

#include <FdNumbering.hh>

// --- forward declaration(s)

namespace FdUtil {
  class InifileChain;
}

namespace FdUtil {

/** enum which represents the different tilt states. */
typedef enum {

  kPositionDown = -1,
  kPositionUnknown = 0,
  kPositionDefault = kPositionUnknown,
  kPositionUp = 1
  
} ETiltPosition;

/** This class acts as a container to store certain FD geometry data.
  *
    These data are:
    @li the pointing of the camera pixels (azimuth & elevation), in degree
    @li the position and orientation of the FD eyes
   
    The information provided by this class, namely the pointing of tte axis 
    of the FD cameras can be overwritten by the settings within the section
    [Geometry] on an ini-file. This is the format of this section:
   
    @verbatim
    
    [Geometry]
    ;axis25 = -42.06 16.00     ; axis pointing for abs. telescope 25 (HEAT 1)
    axis26 =  auto 3.24 16.00  ; axis pointing for abs. telescope 26 (HEAT 2),
                               ; 'auto' with fallback is selected
    ;axis27 = 48.08 16.00      ; axis pointing for abs. telescope 27 (HEAT 3)
    axis27 = auto              ; axis pointing for abs. telescope 27 (HEAT 3),
                               ; 'auto' mode is selected
    
    @endverbatim
   
    One can either specify the azimuth and elevation angles value or the 
    option 'auto' which means that the class tries to enquire the axis 
    direction by some other means. If this is not posisble, then the 
    hard-coded values are used as fallback value. These fallback values might 
    be also specified after the 'auto' option. For the 24 'standard' telescopes
    it makes no sense to specify the 'auto' option.
    
    To set the values for the HEAT telescopes, additional entries might be
    specified in the section [Geometry]:
    
    @endverbatim

    ; special axis declaration for tiltable telescopes @ HEAT, down, up & default
    heat1 = -42.06 16.00 -42.06 45.00 -42.06 45.00
    heat2 =   3.24 16.00   3.24 45.00   3.24 45.00
    heat3 =  48.08 16.00  48.08 45.00  48.08 45.00
    
    @verbatim
   
    When the method GetInstance() of this class is called for the first
    time for a particular telescope, the FdGeometry objects for this telescope
    is constructed using the hard-coded axis pointing values. 
    The the class tries to find out via the ini-file FDCommoni.ini, if any
    of those predefined values are changed by the settings in the ini-file.
    In that case the corresponding FdGeometry object is deleted and a new 
    one according to the new axis specification is constructed.
    
    In case the 'auto' option is provided, the class tries during its
    initialisation to access the axis pointing values via a user provided
    function. If this function is not available, or getting the axis parameters
    failed for any reason (user provided function return value != 0) the
    object is initialized with the hard-coded fallback values.
    
    Note: All angles are measured in degree unless otherwise noted. The azimuth
          angles count counterclockwise, with the east direction is 0.
  *
  */
class FdGeometry {

public:

    /** A structure to describe the axis (pointing) of a FD camera. */
    typedef struct _CameraPointingRec {
    
      _CameraPointingRec()
       { fAzimuth = fElevation = 0.0; fHashValue = 0; }

      _CameraPointingRec(double az,double el)
       : fAzimuth(az), fElevation(el) { fHashValue = 0; }
      
      unsigned int fHashValue;     //<<< for sorting structs of this type
      
      double       fAzimuth;       //<<< azimuth of the camera axis
      double       fElevation;     //<<< elevation of the camera axis
    
      void CalculateHash() {

        double normalized_azimuth = fAzimuth;
	double normalized_elevation = fElevation;
	
	// normalise azimuth and elevation angles
	while ( normalized_azimuth < 0.) normalized_azimuth += 360.;
	while ( normalized_azimuth >= 360. ) normalized_azimuth -= 360.;
	
	if ( normalized_elevation < 0 )
	  throw std::logic_error("Negative elevation angle is not supportet!");
	while ( normalized_elevation >= 90. ) normalized_elevation -= 90.;
	
	fHashValue =   (unsigned int)(normalized_azimuth * 100.) * 10000 
	             + (unsigned int)(normalized_elevation * 1000.);
      }
      
    } CameraPointingRec, CameraPointing_t, *CameraPointing;
    
protected:

    /** Constructor of the class FdGeometry for passed abs. telescope number. 
      *
      * The az,el table for the pixel pointings of this telescope is 
      * initialized using the axis direction stored in the table 
      * fgCameraGeometry[] (which might be overwritten by SetTelescopeAxis()).
      */
    FdGeometry(unsigned int);
    
    /** Constructor of the class FdGeometry for passed axis direction. */
    FdGeometry(CameraPointing_t&);
    
    /** Destructor of the class FdGeometry. */
    virtual ~FdGeometry();
    
    /** The store for the FdGeometry objects. */
    static FdGeometry* fgGeometry[];

public:

    /** Get an instance of the FdGeometry for the passed absolute telescope 
      * number. 
      *
      * At the first call to any GetInstance() method the table fgGeometry[]
      * is initialized with NULLs. Then only the entry for the specified
      * telescope is calculated and filled.
      * 
      * An exception is thrown if an illegal telescope number is passed.
      */
    static const FdGeometry& GetInstance(unsigned int);
    
    /** Get an instance of the FdGeometry for the passed Eye and Mirror number. */
    static const FdGeometry& GetInstance(unsigned int,unsigned int);
    
#if 0
    /** Get an instance of the FdGeometry for the passed MirrorNumber. */
    static const FdGeometry& GetInstance(FdUtil::Fd::MirrorNumber);
#endif
    
    /** Get an instance of the FdGeometry for the passed axis. */
    static const FdGeometry& GetInstance(CameraPointing_t&);
    
    /** Get the azimuth value in degree for the pixel specified by its 
      * column and row number.
      *
      * The orientation is counter-clock wise starting from the east direction.
      */
    double GetPixelAzimuth(unsigned int,unsigned int) const;
    
    /** Get the azimuth value in degree for the pixel specified by its 
      * Fd::PixelNumber. 
      *
      * The orientation is counter-clock wise starting from the east direction.
      */
    static double GetPixelAzimuth(const Fd::PixelNumber);
    
    /** Get the azimuth value in degree for the pixel specified by its 
      * (eye_id,mirror_id,column,row) tuple. 
      *
      * The orientation is counter-clock wise starting from the east direction.
      */
    static double GetPixelAzimuth(unsigned int,unsigned int,
                                 unsigned int,unsigned int);
    
    /** Get the default value for the pixel azimuth angle in degree (i.e. not 
      * transformed into Eye coordinates).
      */
    static double GetPixelAzimuth0(unsigned int,unsigned int);
    
    /** Get the elevation angle value in degree for the pixel specified by its 
      * column and row number. 
      */
    double GetPixelElevation(unsigned int,unsigned int) const;
    
    /** Get the elevation angle value in degree for the pixel specified by its 
      * Fd::PixelNumber.
      */
    static double GetPixelElevation(const Fd::PixelNumber);
    
    /** Get the elevation angle value in degree for the pixel specified by its 
      * (eye_id,mirror_id,column,row) tuple. 
      */
    static double GetPixelElevation(unsigned int,unsigned int,
                                   unsigned int,unsigned int);
    
    /** Get the default value for the pixel elevation angle in degree (i.e. not 
      * transformed into Eye coordinates).
      */
    static double GetPixelElevation0(unsigned int,unsigned int);
    
    /** Get the pixel pointing (azimuth,elevation) (in degree) for the pixel 
      * specified by its Fd::PixelNumber.
      */
    static void GetPixelPointing(const Fd::PixelNumber,double& az,double& el);
    
    /** Get the pixel pointing (azimuth,elevation) (in degree) for the pixel 
      * specified by its (eye_id,mirror_id,column,row) tuple.
      */
    static void GetPixelPointing(unsigned int,unsigned int,
                                 unsigned int,unsigned int,double& az,double& el);

    /** Get the axis parameters for a HEAT telescope depending on the
      * current tilt status.
      */
    static void GetTelescopeAxis4HEAT(unsigned int,double& az,double& el,
                                     ETiltPosition = kPositionDefault);
    
    /** Get the axis parameters for a HEAT telescope depending on the
      * current tilt status.
      */
    static void GetTelescopeAxis4HEAT(unsigned int tel_id,CameraPointingRec& axis,
                                     ETiltPosition tilt = kPositionDefault)
     { GetTelescopeAxis4HEAT( tel_id, axis.fAzimuth, axis.fElevation, tilt ); }
    
    /** Get the direction of the Axis in degree. */
    static void GetTelescopeAxis(unsigned int,double& az,double& el);
    
    /** Get the direction of the Axis in degree. */
    static void GetTelescopeAxis(unsigned int abs_telescope_id,
                                 CameraPointingRec& axis)
     { GetTelescopeAxis( abs_telescope_id, axis.fAzimuth, axis.fElevation ); }

    /** Get the azimuth angle for the specified telescope (abs. telescope id). */
    static double GetTelescopeAzimuth(unsigned int);
    
    /** Get the elevation angle for the specified telescope (abs. telescope id). */
    static double GetTelescopeElevation(unsigned int);
    
    /** Read the camera pointing (telescope axis) from ini-file. */
    static bool ReadCameraPointing(FdUtil::InifileChain&);
    //static void ReadCameraPointing(const std::string&);
    
    /** Set the print level (= level of verbosity) for all
      * instances of this class.
      */
    static void SetPrintLevel(unsigned int level=1)
     { fgPrintLevel = level; }

    /** (Re-)do the telescope axis and geometry calculation. */
    static void SetTelescopeAxis(unsigned int);
    
    /** A typedef for a function which might be called to get the
      * axis pointing direction for a specific telescope.
      *
      * The function must return 0 on success, != 0 else.
      */
    typedef int (*GetAxisFunction_t)(unsigned int,double&,double&);

    /** Set a function to get the telescope axis pointing from elsewhere. */
    static void SetGetAxisFunction(unsigned int,GetAxisFunction_t);

protected:

    /** A structure describing the geometry parameters of an Eye building. */
    typedef struct _EyeGeometryRec {
    
      double   fLongitude;     //<<< longitude, west = negative
      double   fLatitude;      //<<< latitude, south = negative
      double   fGPSAltitude;   //<<< GPS antenna WGS84 height above see level
      double   fFloorAltitude; //<<< floors WGS84 height above see level

      double   fAngleToEast;   //<<< orientation of backwall, for Evb
      
    } EyeGeometryRec, *EyeGeometry;

    /** A structure describing the geometry parameters of FD camera. */
    typedef struct _CameraGeometryRec {
    
      CameraPointingRec    fAxis;         //<<< camera axis
      
      GetAxisFunction_t    fGetAxisFunc;  //<<< pointer to function to get pointing
      
      bool   fAutomatic;          //<<< can be determined automatically
      
    } CameraGeometryRec, *CameraGeometry;
    
    /** Calculate the azimuth, elevation table for a given telescope. */
    void CalculateAzElevTable();
    
    /** Initialize the class static data structures. */
    static void InitializeClass();

    /** Initialize the working tables. */
    static void InitializeTables();
    
    /** Read the (extended) camera pointing (telescope axis) for HEAT from ini-file. */
    static bool ReadCameraPointing4HEAT(FdUtil::InifileChain&);
    
    /** Set the direction of the Axis in degree. 
      *
      * This might re-trigger the calculation of the pixel directions
      * of the specified telescope.
      */
    static void SetTelescopeAxis(unsigned int,
                                 double az,double elev,bool auto_mode=false);

    unsigned int       fTelescopeId;     //<<< abs. telescope id, [1..]
    
    CameraPointing_t   fAxis;            //<<< axis pointing
    
    std::vector<float> fPixelAzimuth;    //<<< table of azimuth values
    std::vector<float> fPixelElevation;  //<<< table of elevation values
     
private:
    // these ones don't exist
    FdGeometry();
    FdGeometry(const FdGeometry&);
    FdGeometry& operator=(const FdGeometry&);

    /** less<key_type>()-operator for CameraPointing_t */
    struct AxisCompare {
      // operator() does all the magic
      bool operator()(const CameraPointing_t& a1, const CameraPointing_t& a2) const {

        // sorry for that nasty cast but I know what I am doing... (thjm)
	if ( a1.fHashValue == 0 ) ((CameraPointing_t&)a1).CalculateHash();
        if ( a2.fHashValue == 0 ) ((CameraPointing_t&)a2).CalculateHash();
	
	return (a1.fHashValue < a2.fHashValue);
      }
    };

    /**  */
    static void ExitHandler();

    /** Map between axis pointing and FdGeometry (table) object. */
    static std::map<CameraPointing_t,FdGeometry*,AxisCompare> fgGeometryMap;
    
    /** Flag to indicate that the static class structures have been initialized. */
    static bool fgClassInitialized;
    
    /** Level of verbosity */
    static unsigned int fgPrintLevel;
    
    /** Table of the 440 default pixel azimuths. */
    static const float fgDefaultPixelAzimuth[];
    /** Table of the 440 default pixel elevation angles. */
    static const float fgDefaultPixelElevation[];
    
    /** Table of default FD camera geaometry constants. */
    static CameraGeometryRec fgCameraGeometry[];

    /** Table of FD Eye geometry constants. */
    static const EyeGeometryRec fgEyeGeometry[];
    
    /** Table of default FD camera geometry constants. */
    static const CameraPointingRec fgDefaultCameraPointing[];
    
    /** Extended table of HEAT camera geometry constants. */
    static CameraPointingRec fgHeatCameraPointing[][3];
};

}  // namespace FdUtil

#endif  // _FdUtil_FdGeometry_hh_
