
//
// File:    FdNaming.cc
//
// Purpose: Implementation of the class FdNaming
//
// $Id: FdNaming.cc 8758 2010-09-27 12:16:24Z mathes $
//



#include <iostream>
#include <iomanip>
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <string>
#include <sstream>
#include <climits>
#include <cassert>

#include <pwd.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <ctype.h>
#include <unistd.h>


/** @file FdNaming.cc
  * Implementation of the class FdNaming.
  * @author H.-J. Mathes, FzK
  */

#include "FdConfig.hh"
#include "FdNumbering.hh"
#include "TimeConvert.hh"
#include "FdNaming.hh"

using namespace std;
using namespace FdUtil;

// -------------------------------------------------------------------------
// -------------------------------------------------------------------------
//
// class FdNaming - implementation
//

unsigned int FdNaming::fgPrintLevel = 0;

FdNaming::FdNaming()
 {
  FdConfig& config = FdConfig::GetInstance();
  
  fEyeId = config.GetEyeId();
  fMirrorId = config.GetMirrorId();
  
  fIsAugerSystem = config.IsEyePC() || config.IsMirrorPC();
  
#warning Tbd: fIsAugerTestSystem has to be set ! 
  // might use the true Eye Id for that !?
  fIsAugerTestSystem = false;
  
#ifndef __DARWIN__
  struct passwd *passwd = getpwuid( geteuid() );
  
  fIsAugerUser = ( strcmp( passwd->pw_name, "auger" ) == 0 ) ? true : false;
#else
  fIsAugerUser = false;
#endif // __DARWIN__

  SetOption( kDefaultOptions );
}

// -------------------------------------------------------------------------

FdNaming::~FdNaming()
 {
}

// -------------------------------------------------------------------------

void FdNaming::CreateDirectory(const std::string directory)
 {
  string mkdir_str = "mkdir -p ";
  
  mkdir_str += directory;
  
  system( mkdir_str.c_str() );
}

// -------------------------------------------------------------------------

const std::string FdNaming::GetBaseDirectory()
 {
  string base_dir;

  if ( fIsAugerSystem && fIsAugerUser ) {

    // --- construct Eye dependant name
  
    string eye_name = "FD-";
  
    eye_name += FdUtil::GetEyeName( fEyeId, true );
 
    // --- try variant: /home/auger/FD-LosLeones (only for user 'auger')
  
    base_dir = "/home/auger/";
    base_dir += eye_name;
    
    // check if directory exists, accounts for a possible installation bug
    
    struct stat st_stat;
    if ( stat( base_dir.c_str(), &st_stat ) == 0 ) {
      if ( S_ISDIR( st_stat.st_mode ) ) {
        return base_dir;
      }
      else {
        cerr << "FdNaming::GetBaseDirectory() - " 
	     << base_dir << " is not a directory !" << endl;

        return string("");
      }
    }
  
#if 0
  // --- try variant: /export/mirror_home
  
  struct passwd *passwd = getpwuid( geteuid() );
  
  base_dir = "/export/mirror_home";
  base_dir += passwd->pw_name;
  
  struct stat st_stat;
  if ( stat( base_dir.c_str(), &st_stat ) == 0 ) {   // directory exists
    if ( S_ISDIR( st_stat.st_mode ) ) {
      free( fDirectory );
      fDirectory = strdup( base_dir.c_str() );
	
      return fDirectory;
    }
    else
      cerr << "FdFileStructure::GetBaseDirectory() - " 
           << base_dir.c_str() << " is not a directory !" << endl;
  }
#endif

  }
  else if (   (!fIsAugerSystem && fIsAugerUser)
           || (fIsAugerSystem && !fIsAugerUser)
	   || (!fIsAugerSystem && !fIsAugerUser) ) {
   
    // --- try variant: $HOME (for all other users or when FD-LosLeones is away)
  
    assert( getenv( "HOME" ) != NULL );
    
    base_dir = getenv( "HOME" );
  }

  
  return base_dir;
}

// -------------------------------------------------------------------------

const std::string FdNaming::GetHomeDirectory()
 {
  assert( getenv( "HOME" ) != NULL );
    
  return getenv( "HOME" );
}

// -------------------------------------------------------------------------

const std::string FdNaming::GetEyeHomeDirectory() throw ( std::domain_error )
 {
  if ( !fIsAugerSystem ) 
    throw std::domain_error( "not running on AUGER DAQ host" );

  FdConfig& config = FdConfig::GetInstance();

  if ( config.IsEyePC() )
    return GetHomeDirectory();
  else {
    struct passwd *passwd = getpwuid( geteuid() );
  
    assert( passwd->pw_name != NULL );
    
    ostringstream dir;
    dir << "/home/eyePC/" << passwd->pw_name;
    
    return dir.str();
  }
}

// -------------------------------------------------------------------------

const std::string FdNaming::GetMirrorHomeDirectory() throw ( std::domain_error )
 {
  if ( !fIsAugerSystem ) 
    throw std::domain_error( "not running on AUGER DAQ host" );

  FdConfig& config = FdConfig::GetInstance();

  if ( config.IsMirrorPC() )
    return GetHomeDirectory();
  else {
    struct passwd *passwd = getpwuid( geteuid() );
  
    assert( passwd->pw_name != NULL );
    
    ostringstream dir;
    dir << "/export/mirror_home/" << passwd->pw_name;
    
    return dir.str();
  }
}

// -------------------------------------------------------------------------

const std::string 
FdNaming::GetAuxDataFilename(const std::string prefix,
                             const std::string suffix,
			     unsigned int mirror_id)
 {
  ostringstream aux_filename;
  
  aux_filename.fill('0');
  
  aux_filename << prefix;
  
  if ( mirror_id ) 
    aux_filename << "_" << "m" << setw(2) << GetObservatoryMirrorId( mirror_id );

  if ( fUseIsoDate || fUseIsoTime ) aux_filename << "_" << GetISODateTime();
  
  aux_filename << "." << suffix;
   
  return aux_filename.str();
}

// -------------------------------------------------------------------------

const std::string
FdNaming::GetDataFilename(const std::string prefix,
                          unsigned int run_id, 
			  const std::string suffix,
			  unsigned int mirror_id)
 {
  ostringstream data_filename;
  
  data_filename.fill('0');
  
  data_filename << prefix << "_";
  
  if ( mirror_id )
    data_filename << "m" << setw(2) << GetObservatoryMirrorId( mirror_id );
  else
    data_filename << "e" << GetEyeId();
    
  if ( fUseRunId ) data_filename << "_" << setw(6) << run_id;
  
  if ( fUseIsoDate || fUseIsoTime ) data_filename << "_" << GetISODateTime();
  
  data_filename << "." << suffix;

  //cout << "FdNaming::GetDataFilename() - '" 
  //     << data_filename.str() << "'" << endl;

  return data_filename.str();
}

// -------------------------------------------------------------------------

const std::string FdNaming::GetDataDirectory(const std::string sysname,
                                             const std::string subdir)
 {
  string directory = GetBaseDirectory();
  
  directory += "/";
  
  directory += sysname;
  
  directory += "/";
  directory += GetDirectoryFromDate();

  directory += subdir;
  directory += "/";
  
  if ( fCreateDirectory ) CreateDirectory( directory );
  
  return directory;
}

// -------------------------------------------------------------------------

static const std::string gConfigDirectory = "general";

const std::string FdNaming::GetConfigDirectory()
 {
  if ( !fIsAugerSystem ) {
    // throw std::domain_error( "not running on AUGER DAQ host" );
    return GetEyeConfigDirectory();
  }

  FdConfig& config = FdConfig::GetInstance();
  
  if ( config.IsEyePC() )
    return GetEyeConfigDirectory();
  else
    return GetMirrorConfigDirectory( GetObservatoryMirrorId( fMirrorId ) );
}
  
// -------------------------------------------------------------------------

const std::string FdNaming::GetEyeConfigDirectory()
 {
  ostringstream dir;
  
  if ( !fIsAugerSystem ) { 
    // throw std::domain_error( "not running on AUGER DAQ host" );
    dir << GetHomeDirectory()         // --> /home/auger, /home/$USER
        << "/" << gConfigDirectory
        << "/" << GetEyeDirectoryName() << "/";
  
    return dir.str();
  }

  FdConfig& config = FdConfig::GetInstance();
  
  if ( config.IsEyePC() )
    dir << GetHomeDirectory();    // --> /home/auger, /home/$USER
  else // config.IsMirrorPC()
    dir << GetEyeHomeDirectory(); // --> /home/eyePC/auger, /home/$USER

  dir << "/" << gConfigDirectory << "/"
      << GetEyeDirectoryName() << "/";
  
  return dir.str();
}

// -------------------------------------------------------------------------

const std::string FdNaming::GetMirrorConfigDirectory(unsigned int mirror_id)
		                           throw ( std::invalid_argument )
 {
  if ( !fIsAugerSystem ) {
    // throw std::domain_error( "not running on AUGER DAQ host" );
    return GetEyeConfigDirectory();
  }

  if ( mirror_id > Fd::kFD_NMIRRORS )
    throw std::invalid_argument( "passed Mirror id not valid" );

  // check if mirror_id fits to the current building
  
  int priv_mirror_id = mirror_id;
  
  if ( mirror_id > Fd::kEYE_LAST_MIRROR )
    priv_mirror_id = mirror_id % Fd::kEYE_NMIRRORS;
  
  FdConfig& config = FdConfig::GetInstance();
  
  ostringstream dir;

  if ( config.IsEyePC() ) {
    dir << GetMirrorHomeDirectory(); // --> /export/mirror_home/auger, /home/$USER
    dir << "/" << gConfigDirectory << "/eyepc/";
  }
  else { // config.IsMirrorPC()
    dir << GetHomeDirectory();       // --> /home/eyePC/auger, /home/$USER
    dir << "/" << gConfigDirectory << "/" << "mirrorpc" << priv_mirror_id << "/";
  }
   
  return dir.str();
}

// -------------------------------------------------------------------------

const std::string FdNaming::GetEyeDirectoryName()
 {
  string sysname;
  
  if ( fIsAugerSystem ) {
    sysname = "eyepc";
  }
  else {
    char hostname[80];
    gethostname( hostname, sizeof( hostname ) );
    
    for ( unsigned int i=0; i<strlen( hostname ); i++ )
      sysname += tolower( hostname[i] );
  }

  return sysname;
}

// -------------------------------------------------------------------------

const std::string FdNaming::GetMirrorDirectoryName(unsigned int mirror_id)
 {
  ostringstream directory;
  
  // ---> `hostname`
  if ( !fIsAugerSystem ) return GetEyeDirectoryName();
  
  if ( mirror_id == 0 ) {
  // ---> eyepc
    if ( IsEyePC() ) return GetEyeDirectoryName();
  // ---> mirrorpc#, where #=fMirrorId
    else return GetMirrorDirectoryName( fMirrorId );
  }

  // do some checks
  if (   mirror_id < Fd::kEYE_FIRST_MIRROR 
      || mirror_id > Fd::kEYE_LAST_MIRROR ) {
    cerr << "FdNaming::GetMirrorDirectoryName(" << mirror_id << ") - "
         << "passed Mirror id is illegal !" << endl;
   exit( EXIT_FAILURE );
  }
  
  directory << "mirrorpc" << mirror_id;
  
  return directory.str();
}

// -------------------------------------------------------------------------

const std::string FdNaming::GetSystemDirectoryName()
 {
  string sysname;
  
  if ( fIsAugerSystem ) {
    if ( IsMirrorPC() )
      sysname = GetMirrorDirectoryName( GetMirrorId() );
    else if ( IsEyePC() )
      sysname = GetEyeDirectoryName();
    else {  // just a stupid check, should never happen !
      cerr << "FdNaming::GetDataDirectory() - "
           << "cannot run both on MirrorPC and EyePC !!!" << endl;
      exit( EXIT_FAILURE );
    }
  }
  else
    sysname = GetEyeDirectoryName();

  return sysname;
}

// -------------------------------------------------------------------------

const std::string FdNaming::GetDirectoryFromDate()
 {
  time_t ltime = time( NULL );
  struct tm* tm = localtime( &ltime );   
  char date_str[20];
  
  // Using astronomers conventions: days start at 12:00 (MINUS 12 hours in regards to real time)

  if ( tm->tm_hour < 12 ) {
    ltime = (time_t) ((unsigned int) ltime - 12*3600 );
    struct tm* tm = localtime( &ltime );

    sprintf( date_str, "%04d/%02d/%02d/", 
                       1900 + tm->tm_year, tm->tm_mon + 1, tm->tm_mday );
		            
  }
  else {
    sprintf( date_str, "%04d/%02d/%02d/", 
                       1900 + tm->tm_year, tm->tm_mon + 1, tm->tm_mday );
  }

  return string( date_str );
}

// -------------------------------------------------------------------------

const std::string FdNaming::GetISODate(struct tm *tm)
 {
  if ( !fUseIsoDate ) return string("");
  
  if ( !tm ) {
    time_t now = time( NULL );
    tm = gmtime( &now );
  }
  
  return TimeConvert::GetISODate( *tm );
}

// -------------------------------------------------------------------------

const std::string FdNaming::GetISODateTime(struct tm *tm)
 {
  string date_time_str;
  if ( !tm ) {
    time_t now = time( NULL );
    tm = gmtime( &now );
  }
  
  if ( fUseIsoDate ) {
    date_time_str += GetISODate( tm );
    if ( fUseIsoTime ) date_time_str += ".";
  }
  if ( fUseIsoTime )
    date_time_str += GetISOTime( tm );
  
  return date_time_str;
}

// -------------------------------------------------------------------------

const std::string FdNaming::GetISOTime(struct tm *tm)
 {
  if ( !fUseIsoTime ) return string("");
  
  if ( !tm ) {
    time_t now = time( NULL );
    tm = gmtime( &now );
  }

  return TimeConvert::GetISOTime( *tm, false );
}

// -------------------------------------------------------------------------

unsigned int FdNaming::GetObservatoryMirrorId(unsigned int mirror_id)
 {
  // check the bounds
  if ( mirror_id < Fd::kEYE_FIRST_MIRROR || mirror_id > Fd::kFD_NMIRRORS ) {
    cerr << "FdNaming::GetObservatoryMirrorId(" << mirror_id << ") - "
         << "passed Mirror id is illegal !" << endl;
    // exit( EXIT_TERMINATE );
    return 0;
  }
  
  if ( mirror_id > Fd::kEYE_LAST_MIRROR )  // assume that it is already OK
    return mirror_id;
  else {
    return   (fEyeId - Fd::kFD_FIRST_EYE) * Fd::kEYE_NMIRRORS
           + mirror_id;
  }
}

// -------------------------------------------------------------------------

void FdNaming::Test(ostream& ostr)
 {
  ostr << endl;
  ostr << "FdNaming::Test() ..." << endl;

  ostr << "fEyeId= " << fEyeId << " fMirrorId= " << fMirrorId << endl;
  ostr << "fIsAugerSystem= " << (int)fIsAugerSystem
       << " fIsAugerUser= " << (int)fIsAugerUser << endl;
  
  ostr << "GetBaseDirectory:   " << this->GetBaseDirectory() << endl;
  
  ostr << "GetISODate:         " << this->GetISODate() << endl;
  ostr << "GetISOTime(UTC):    " << this->GetISOTime() << endl;
  ostr << "GetISODateTime:     " << this->GetISODateTime() << endl;

  ostr << endl;
}

// -------------------------------------------------------------------------

void FdNaming::SetOption(ENamingOption option, bool enabled)
 {
  switch ( option ) {

    case kDefaultOptions:
         SetOption( kUseIsoDate, true );
	 SetOption( kUseIsoTime, true );
	 SetOption( kUseRunId, true );
	 SetOption( kCreateDirectory, true );
         break;

    case kUseIsoDate:
         fUseIsoDate = enabled;
	 break;

    case kUseIsoTime:
         fUseIsoTime = enabled;
         break;

    case kUseRunId:
         fUseRunId = enabled;
         break;

    case kCreateDirectory:
         fCreateDirectory = enabled;
         break;

    default:
      cerr << "FdNaming::SetOption() - illegal option specified !" << endl;
  }
}

// -------------------------------------------------------------------------
// -------------------------------------------------------------------------
