
//
// File:    Inifile.cc
//
// Purpose: Reading Windows(TM) like *.ini files 
//          (derived from inifile.hxx, Rev. 1.5)
//
// $Id: Inifile.cc 8758 2010-09-27 12:16:24Z mathes $
//


#include <cstring>
#include <cstdlib>
#include <cstdio>
#include <cassert>

#include <iostream>
#include <string>
#include <sstream>

/** @file Inifile.cc
  * Reading Windows(TM) like *.ini files.
  * @author Th. Wiegert, FzK
  */

#include "ErrorLogging.hh"
#include "FdNaming.hh"
#include "FdStdioStream.hh"

#include "Inifile.hh"

using namespace std;
using namespace FdUtil;

// -------------------------------------------------------------------------
// -------------------------------------------------------------------------
//
// class Inifile - implementation
//

/** Definition of error code for internal error. */
#define INIFILE_BAD_VALUE  127
/** Definition of line buffer size. */
#define LINE_BUFFER_LENGTH (INIFILE_LINE_BUFFER_SIZE - 1)

unsigned int Inifile::fgPrintLevel = 1;

// -------------------------------------------------------------------------

Inifile::Inifile(const char* filename)
 : fInifile(NULL)
 {
  fGrPos    = 0;
  fGroup    = NULL;
  fPtr      = NULL;
  fLineptr  = NULL;
  fStatus   = kFAIL;
  
  assert( filename != NULL && *filename != 0 );
  
  if ( fgPrintLevel > 1 ) {
    FD_CINFO << "Inifile::Inifile('" << filename << "') -" << endl;
  }
  
  fFilename = Locate( filename );
  
  if ( !fFilename.empty() ) {
    
    FD_CINFO << "Inifile::Inifile() - found " << fFilename << endl;
    
    Open( fFilename.c_str() );
  }

  if ( fInifile == NULL ) {
    fStatus = kFAIL;
    return;
  }

  fStatus = kSUCCESS;
}

// -------------------------------------------------------------------------

Inifile::Inifile(const char* filename, std::FILE *logfile,
                 const char *secondDir, const char *thirdDir)
 : fInifile(NULL)
 {
  fGrPos    = 0;
  fGroup    = NULL;
  fPtr      = NULL;
  fLineptr  = NULL;
  fStatus   = kFAIL;
  
  assert( logfile != NULL );
  FdUtil::OstdioStream logstream( logfile );

  assert( filename != NULL && *filename != 0 );
  
  if ( fgPrintLevel > 1 ) {
    FD_CINFO << "Inifile::Inifile('" << filename 
             << "','" << logfile
	     << "','" << secondDir
	     << "','" << thirdDir << "'" << endl;
  }
  
  // ---> 1. check the file name only, case: absolute directory name
  
  fFilename = TestFile ( "", filename );
  
  if ( fFilename.empty() )
    logstream << "Inifile::Inifile() - " 
              << "could not open " << filename << endl;
  
  // ---> 2. local directory

  if ( fFilename.empty() ) {
    fFilename = TestFile( "./", filename );
  
    if ( fFilename.empty() )
      logstream << "Inifile::Inifile() - " 
                << "could not open " << "./" << filename << endl;
  }
  
  // ---> 3. secondDir
  
  if ( fFilename.empty() && secondDir ) {

    fFilename = TestFile( secondDir, filename );
    
    if ( fFilename.empty() )
      logstream << "Inifile::Inifile() - " 
                << "could not open " << secondDir << "/" << filename << endl;
  }
  
  // ---> 4. thirdDir
  
  if ( fFilename.empty() && thirdDir ) {

    fFilename = TestFile( thirdDir, filename );
    
    if ( fFilename.empty() )
      logstream << "Inifile::Inifile() - " 
                << "could not open " << thirdDir << "/" << filename << endl;
  }
  
  if ( !fFilename.empty() ) {
    Open( fFilename.c_str() );

    if ( fInifile )
      logstream << "Inifile::Inifile() - " 
                << "reading from " << fFilename << " ..." << endl;
  }
  
  if ( fInifile == NULL ) {
    fStatus = kFAIL;
    logstream << "Inifile::Inifile() - " 
              << "could not open any " << filename << " !!!" << endl;

    return;
  }
  
  fStatus = kSUCCESS;
}

// -------------------------------------------------------------------------

Inifile::Inifile(const Inifile& rhs)
 : fInifile(NULL)
 {
  fGrPos    = 0;
  fGroup    = NULL;
  fPtr      = NULL;
  fLineptr  = NULL;
  fStatus   = kFAIL;
  
  if ( fgPrintLevel > 1 ) {
    FD_CINFO << "Inifile::Inifile(const Inifile&) - file= " 
             << rhs.GetFilename() << endl;
  }
  
  fFilename = Locate( rhs.GetFilename() );
  
  if ( !fFilename.empty() ) {
    
    FD_CINFO << "Inifile::Inifile(const Inifile&) - found " << fFilename << endl;
    
    Open( fFilename.c_str() );
  }

  if ( fInifile == NULL ) {
    fStatus = kFAIL;
    return;
  }

  fStatus = kSUCCESS;
}

// -------------------------------------------------------------------------

Inifile::~Inifile()
 {
  Close();
  
  if ( fGroup != NULL ) delete[] fGroup;
}

// -------------------------------------------------------------------------

void Inifile::Close()
 {
  if ( fgPrintLevel > 1 ) {
    FD_CINFO << "Inifile::Close() - fInifile= " 
             << hex << fInifile << dec << endl;
  }
  
  if ( fStatus == kSUCCESS && fInifile != NULL ) {  // 2.8.00 ak
    fclose( fInifile );
  }
  
  fInifile = NULL;
}

// -------------------------------------------------------------------------

char Inifile::Digit2Value(char digit)
 {
  if ((digit >= '0') && (digit <= '9'))
    return (digit - '0');

  if ((digit >= 'A') && (digit <= 'F'))
    return (digit - 'A' + 10);

  if ((digit >= 'a') && (digit <= 'f'))
    return (digit - 'a' + 10);

  return INIFILE_BAD_VALUE;
}

// -------------------------------------------------------------------------

char Inifile::Digits2Value(char** src, char* end, int base)
 {
  char digval;
  int  value = 0, maxdigits = 3;
 
  if (base == 16) maxdigits = 2;
  if ((base == 16) || (base == 8)) (*src)++;
  if ((base == 8) && (**src == 'x')) return Digits2Value (src, end, 16);

  while ((*src < end) && maxdigits && ((digval = Digit2Value(**src)) < base)) {
    
    value *= base;
    value += digval;
    (*src)++;
    maxdigits--;
  }
  
  (*src)--;
 
  return ((char)(value % 256));
}

// -------------------------------------------------------------------------

std::string Inifile::Locate(const char* filename)
 {
  assert( filename != NULL && *filename != 0 );
  
  string directory_str;
  string filename_str;
  
  // ---> 0. check for filename(s) with either complete or incomplete path spec
  
  string filename_str2 = filename;
  
  if ( filename_str2.rfind( "/" ) != string::npos ) {
  
    directory_str = filename_str2.substr( 0, filename_str2.rfind( "/" ) );
    filename_str2 = filename_str2.replace( 0, filename_str2.rfind( "/" ) + 1, "" );
    
    filename_str = TestFile( directory_str, filename_str2 );
    
#if 0
    cout << "filename_str2= " << filename_str2 << endl;
    cout << "directory_str= " << directory_str << endl;
    
    cout << "filename_str= " << filename_str << endl;
#endif

    if ( !filename_str.empty() ) return filename_str;
  }
  
  // ---> 1. local directory
  
  directory_str = "./";
  
  filename_str = TestFile( directory_str, filename );
  if ( !filename_str.empty() ) return filename_str;
  
  FD_CWARN << "Inifile::Locate() - "
           << "could not find " << directory_str << filename << endl;
  
  // ---> 2. FdNaming::GetConfigDirectory()
  
  FdNaming& naming = FdNaming::GetInstance();
  directory_str = naming.GetConfigDirectory();
  
  filename_str = TestFile( directory_str, filename );
  if ( !filename_str.empty() ) return filename_str;

  FD_CWARN << "Inifile::Locate() - "
           << "could not find " << directory_str << filename << endl;

  // ---> 3. users home directory

  directory_str = getenv( "HOME" );
  directory_str += "/";
  
  filename_str = TestFile( directory_str, filename );
  if ( !filename_str.empty() ) return filename_str;
  
  FD_CWARN << "Inifile::Locate() - " 
           << "could not find " << directory_str << filename << endl;
  
  // ---> 4. /opt/auger/etc

  directory_str = "/opt/auger/etc/";
  
  filename_str = TestFile( directory_str, filename );
  if ( !filename_str.empty() ) return filename_str;
  
  FD_CWARN << "Inifile::Locate() - " 
           << "could not find " << directory_str << filename << endl;

  // ---> 5. $FD_ROOT/etc

  if ( getenv( "FD_ROOT" ) != NULL ) {
  
    directory_str = getenv( "FD_ROOT" );
    directory_str += "/etc/";
    
    filename_str = TestFile( directory_str, filename );
    if ( !filename_str.empty() ) return filename_str;

    FD_CWARN << "Inifile::Locate() - " 
             << "could not find " << directory_str << filename << endl;
  }
  else
    FD_CWARN << "Inifile::Locate() - " 
             << "could not find $FD_ROOT/etc/" << filename << endl;

  // ---> 6. $FDDAS_CONFIG_DIR
  
  if ( getenv( "FDDAS_CONFIG_DIR" ) != NULL ) {
  
    directory_str = getenv( "FDDAS_CONFIG_DIR" );
    directory_str += "/config/";
    
    filename_str = TestFile( directory_str, filename );
    if ( !filename_str.empty() ) return filename_str;

    FD_CWARN << "Inifile::Locate() - " 
             << "could not find " << directory_str << filename << endl;
  }
  else
    FD_CWARN << "Inifile::Locate() - " 
             << "could not find $FDDAS_CONFIG_DIR/" << filename << endl;

  // ---> 7. $FDDAS_BASE_DIR
  
  if ( getenv( "FDDAS_BASE_DIR" ) != NULL ) {
  
    directory_str = getenv( "FDDAS_BASE_DIR" );
    directory_str += "/config/";

    filename_str = TestFile( directory_str, filename );
    if ( !filename_str.empty() ) return filename_str;

    FD_CWARN << "Inifile::Locate() - " 
             << "could not find " << directory_str << filename << endl;
  }
  else
    FD_CWARN << "Inifile::Locate() - " 
  	     << "could not find $FDDAS_BASE_DIR/" << filename << endl;

  // ---> 8. $INIFILE_PATH
  
  if ( getenv( "INIFILE_PATH" ) != NULL ) {

    directory_str = getenv( "INIFILE_PATH" );
    directory_str += "/";
    
    filename_str = TestFile( directory_str, filename );
    if ( !filename_str.empty() ) return filename_str;

    FD_CWARN << "Inifile::Locate() - " 
             << "could not find " << directory_str << filename << endl;
  }
  else
    FD_CWARN << "Inifile::Locate() - " 
  	     << "could not find $INIFILE_PATH/" << filename << endl;

  if ( filename_str.empty() ) {
    FD_CERR << "Inifile::Locate() - " 
            << "could not find " << filename << " in any directory !!!" 
	    << endl;
  }
  
  return filename_str;
}

// -------------------------------------------------------------------------

char Inifile::GetNextCharacter(char** src, char* end)
 {
  (*src)++; 
  if ( *src == end ) return ((char)0);
  if ( (**src == '\\') || (**src == '#') ) return (**src);
  if ( **src == 'n' ) return ('\n');

  if ( **src == '0' ) return Digits2Value (src, end, 8);

  if ( (**src > '0') && (**src <= '9') )
    return Digits2Value( src, end, 10 );

  if ( **src == 'x' ) return Digits2Value( src, end, 16 );

  return (char)0;
}

// -------------------------------------------------------------------------

char* Inifile::GetLine()
 {
  int bufsize = sizeof( fLinebuf ) - 1; 

  char *ptr;
  char *src, *dest, *end;

  ptr = fgets( fLinebuf, bufsize, fInifile );
  if ( !ptr ) return NULL;
 
  char *cp = fLinebuf;
  
  while ( *cp ) {
    // convert '\t' into blank
    if ( *cp == '\t' ) *cp = ' ';
    // convert '\r' into blank (might be a DOS/Windows file)
    // if ( *cp == '\r' ) *cp = ' ';
    cp++;
  }

  dest = src = ptr; 
  end = ptr + strlen( ptr );

  while ( *src /* && *dest */ ) { // commented part not tested (thjm, 2.11.03)
  
    if ( *src == '\\' ) {
      char next;
      next = GetNextCharacter( &src, end );

      if ( next == (char)0 ) {
        bufsize -= (dest - ptr);
        *(ptr = src = dest) = 0;
        if ( fgets ( dest, bufsize, fInifile ) == NULL ) return fLinebuf;
        dest--; src--;
        end = ptr + strlen (ptr);
      }
      else *dest = next;
    }
    else {
      if ( *src == '#' ) *dest = 0;  /* for comments */
      if ( *src == ';' ) *dest = 0;  /* thjm, 02.11.03, for comments */
      else *dest = *src;
    }
    
    dest++; src++;
  } 

  *dest = (char)0;
  
  return fLinebuf;
}

// -------------------------------------------------------------------------

double Inifile::GetFirstFloat(const char* _entry, int* error)
 {
  if ( error ) *error = kSUCCESS;
  
  if ( fStatus == kFAIL ) {
    if ( error ) *error = kFAIL;
    return 0;
  }
  
  if ( error ) *error = kFAIL;
  fPtr = (char *)GetFirstString( _entry );
  if ( fPtr == NULL ) {
    if ( error ) *error = kFAIL;
    return 0;
  }
 
  char* end;
  double ret;
  ret = strtod( fPtr, &end ); 
  if ( ( error ) && (end != fPtr) ) *error = kSUCCESS;
  
  return ret;
}

// -------------------------------------------------------------------------

double Inifile::GetFirstFloat(const char* entry, double def_value, int* error)
 {
  double value = 0.0;
  int _error;
  
  if ( error ) *error = kSUCCESS;

  if ( GetStatus() == kFAIL )
    _error = kFAIL;
  else
    value = GetFirstFloat( entry, &_error );

  if ( _error == kFAIL ) {
    value = def_value;
    if ( error ) *error = _error;
  }

  return value;
}

// -------------------------------------------------------------------------

const char* Inifile::GetFirstString(const char* _entry )
 {
  if ( fStatus == kFAIL ) {  // 2.8.00 ak
   return NULL;
  }
  
  if ( ( fGroup == NULL ) || ( *fGroup == (char)0 ) ) {
    perror ("group has to be defined!!");
    return NULL;
  }
     
  fseek ( fInifile, fGrPos, SEEK_SET);
  fLineptr = GetLine(); 

  while ( fLineptr && (*fLineptr != '[') ) {
  
    fPtr = strtok( fLineptr, "\n= " );  
    if ( (fPtr != NULL) && !strcmp(_entry, fPtr) )
      return ( fPtr = strtok(NULL, "\n= ") );
    
    fLineptr = GetLine();
  }
  
  return NULL; 
}

// -------------------------------------------------------------------------

const char* Inifile::GetFirstString(const char *entry, const char* def_str, int *error)
 {
  const char* ptr;
  if ( error ) *error = kSUCCESS;

  if ( GetStatus() == kFAIL )
    ptr = NULL;
  else
    ptr = Inifile::GetFirstString( entry );

  if ( ptr == NULL ) {
    if ( error ) *error= kFAIL;
    ptr = def_str;
  } else
    if ( error ) *error = kSUCCESS;

  return ptr;
}

// -------------------------------------------------------------------------

int Inifile::GetFirstValue(const char* _entry, int* error)
 {
  if ( error ) *error = kSUCCESS;
  
  if ( fStatus == kFAIL ) {  // 2.8.00 ak
    if ( error ) *error = kFAIL;
    return 0;
  }
  
  if ( error ) *error = kFAIL;
  fPtr = (char *)GetFirstString( _entry );
  if ( fPtr == NULL ) {
    if ( error ) *error = kFAIL;
    return 0;
  }
 
  char* end;
  int ret;
  ret = strtol( fPtr, &end, 0 ); 
  if ( ( error ) && (end != fPtr) ) *error = kSUCCESS;
  
  return ret;
}

// -------------------------------------------------------------------------

int Inifile::GetFirstValue(const char *entry, int def_value, int *error)
 {
  int value = 0;
  int _error;
  
  if ( error ) *error = kSUCCESS;

  if ( GetStatus() == kFAIL )
    _error = kFAIL;
  else
    value = GetFirstValue( entry, &_error );

  if ( _error == kFAIL ) {
    value = def_value;
    if ( error ) *error = _error;
  }

  return value;
}

// -------------------------------------------------------------------------

double Inifile::GetNextFloat(int *error)
 {
  const char* ptr;
  if ( error ) *error = 0;
  
  if ( fStatus == kFAIL ) {  // 2.8.00 ak
    if ( error ) *error = kFAIL;
    return 0;
  }
  
  if ( error ) *error = kFAIL;
  ptr = GetNextString();
  if ( ptr == NULL ) {
    if ( error ) *error = kFAIL;
    return 0;
  }
 
  char* end;
  double ret;
  ret = strtod( ptr, &end );
  if ( ( error ) && (end != ptr) ) *error = kSUCCESS;
  
  return ret;
}

// -------------------------------------------------------------------------

double Inifile::GetNextFloat(double def_value, int *error)
 {
  float value = 0.0;
  int _error;
  
  if ( error ) *error = 0;

  if ( GetStatus() == kFAIL )
    _error = kFAIL;
  else
    value = GetNextFloat( &_error );

  if ( _error == kFAIL ) {
    value = def_value;
    if ( error ) *error = _error;
  }

  return value;
}

// -------------------------------------------------------------------------

const char * Inifile::GetNextString(int *error)
 {
  if ( fStatus == kFAIL ) {  // 2.8.00 ak
    if ( error ) *error = kFAIL;
    return NULL;
  }
  
  if ( (fGroup == NULL) || (*fGroup == (char)0) ) {
    perror( "group has to be defined!!" );
    if ( error ) *error = kFAIL;
    return NULL;
  }
  
  if ( !fPtr ) {
    if ( error ) *error = kFAIL;
    return NULL;  
  }
  
  fPtr = strtok( NULL, "\n= " );
  
  return fPtr;
}

// -------------------------------------------------------------------------

const char * Inifile::GetNextString(const char* def_string, int* error)
 {
  const char *_string = NULL;
  int _error;
  
  if ( error ) *error = kSUCCESS;

  if ( GetStatus() == kFAIL )
    _error = kFAIL;
  else
    _string = GetNextString( &_error );
  
  if ( _error == kFAIL ) {
    _string = def_string;
    if ( error ) *error = _error;
  }
  else if ( _string == NULL )
    _string = def_string;

  return _string;
}

// -------------------------------------------------------------------------

int Inifile::GetNextValue(int* error)
 {
  const char * ptr;
  if ( error ) *error = 0;
  
  if ( fStatus == kFAIL ) {  // 2.8.00 ak
    if ( error ) *error = kFAIL;
    return 0;
  }
  
  if ( error ) *error = kFAIL;
  ptr = GetNextString();
  if ( ptr == NULL ) {
    if ( error ) *error = kFAIL;
    return 0;
  }
 
  char * end;
  int ret;
  ret = strtol( ptr, &end, 0 );
  if ( ( error ) && (end != ptr) ) *error = kSUCCESS;
  
  return ret;
}

// -------------------------------------------------------------------------

int Inifile::GetNextValue(int def_value, int* error)
 {
  int value = 0;
  int _error;
  
  if ( error ) *error = 0;

  if ( GetStatus() == kFAIL )
    _error = kFAIL;
  else
    value = GetNextValue( &_error );

  if ( _error == kFAIL ) {
    value = def_value;
    if ( error ) *error = _error;
  }

  return value;
}

// -------------------------------------------------------------------------

int Inifile::Open(const char *filename)
 {
  if ( fgPrintLevel > 1 ) {
    FD_CINFO << "Inifile::Open('" << filename << "') -" << endl;
  }

  FILE * inifile;
  
  if ( (inifile = fopen( filename, "r" ) ) != NULL ) {
    
    if ( fgPrintLevel > 1 ) {
      FD_CINFO << "Inifile::Open() - fInifile= " 
               << hex << fInifile << dec << endl;
    }
    
    fInifile = inifile;
    
    if ( fgPrintLevel > 1 ) {
      FD_CINFO << "Inifile::Open() - " 
    	       << "reading from " << filename << " ..." << endl;
    }
    
    fStatus = Inifile::kSUCCESS;

    return fStatus;
  }
  
  fInifile = NULL;
  fStatus = Inifile::kFAIL;
  
  return fStatus;
}

// -------------------------------------------------------------------------

int Inifile::Open()
 {
  if ( fFilename.empty() ) return Inifile::kFAIL;
  
  return Open( fFilename.c_str() );
}

// -------------------------------------------------------------------------

void Inifile::Reset()
 {
  fseek( fInifile, 0, SEEK_SET );
  fLineptr = fLinebuf;
}

// -------------------------------------------------------------------------

int Inifile::SpecifyGroup(const char *_string)
 {
  fGrPos = 0;

  if ( fStatus == kFAIL ) {  // 2.8.00 ak
   return kFAIL;
  }
  
  if ( fGroup != NULL ) {
    delete[] fGroup;
    fGroup = NULL;
  }

  fGroup = new char[strlen(_string) + 3];
  if ( fGroup == NULL ) {
    FD_CSEV << "no memory for [" << _string << "]\n";
    return kFAIL;
  }

  Reset(); 
  sprintf( fGroup, "[%s]", _string ); 
 
  do {
    
    if ( !(fLineptr = GetLine()) ) return kFAIL;

    fPtr = strtok( fLineptr, "\n" ); 
    if ( fPtr != NULL )
      fPtr = strstr( fPtr, fGroup );

    if ( fPtr != NULL ) {
    
      while ( *fPtr && (*fPtr == ' ') ) fPtr++;

//      if ( !strcmp( fPtr, fGroup) ) {
      if ( !strncmp( fPtr, fGroup, strlen(fGroup) ) ) {
      
        fGrPos = ftell( fInifile );
        if ( !(fLineptr = GetLine()) )
          return kFAIL;

        return kSUCCESS;
      }
    }
    
  } while ( fLineptr );
  
  return kFAIL;
}

// -------------------------------------------------------------------------

std::string Inifile::TestFile(const std::string directory, 
                              const std::string filename)
 {
#if 0
  cout << "Inifile::TestFile(directory='" << directory << "',filename='"
       << filename << "') -" << endl;
#endif
  
  string name;

  // ToDo:
  // - check for empty strings (?)
  // - different check if filename is absolut
  
  if ( directory.empty() ) {
    if ( filename.empty() ) return "";
    
    if (    filename.rfind( "./" ) == string::npos 
         || filename.rfind( "./" ) == string::npos) return "";

    name = filename;
  }
  else {
    name = directory;
    if ( name[name.length()-1] != '/' ) name += "/";
    name += filename;
  }

#if 0
  cout << "Inifile::TestFile() - name= " << name << endl;
#endif
  
  FILE * file = fopen( name.c_str(), "r" );
  if ( file != NULL ) {
    fclose( file );
    return name;
  }
  
  return "";
}

// -------------------------------------------------------------------------
// -------------------------------------------------------------------------
