
//
// File   : Singleton.hh
//
// Purpose: Declaration of the template Singleton
//
// $Id: Singleton.hh 8758 2010-09-27 12:16:24Z mathes $
//


#ifndef _FdUtil_Singleton_hh_
#define _FdUtil_Singleton_hh_

/** @file Singleton.hh
  * Declaration of the Singleton template.
  * @author H.-J. Mathes, FzK
  */

#ifndef NO_VERSION
# include <FdUtilVersion.hh>
#endif // NO_VERSION

namespace FdUtil {

  /**
   * \class Singleton Singleton.hh src/Singleton.hh
   *
   * \brief Curiously Recurring Template Pattern (CRTP) for Meyers singleton
   *
   * The singleton class is implemented as follows
   * \code
   * #include <Singleton.hh>
   *
   * class SomeClass : public FdUtil::Singleton<SomeClass> {
   *   ...
   * private:
   *   SomeClass() { }
   *   ~SomeClass() { }
   *   friend class FdUtil::Singleton<SomeClass>;
   * };
   * \endcode
   * Singleton automatically prevents copying of the derived class.
   *
   * \author Darko Veberic
   * \date 9 Aug 2006
   * \version $Id: Singleton.hh 8758 2010-09-27 12:16:24Z mathes $
   */

  template<typename T_DERIVED>
  class Singleton {

  public:
    static T_DERIVED& GetInstance()
    // causes problems with rootcint in ROOT 5.18
    // fix provided by P.Canal, P.Lebrun
#ifdef __MAKECINT__
       ;
#else
    {
      static T_DERIVED instance;
      return instance;
    }
#endif

  protected:
    // derived class can call ctor and dtor
    Singleton() { }
    ~Singleton() { }

  private:
    // no one should do copies
    Singleton(const Singleton&);
    Singleton& operator=(const Singleton&);

  };

} // namespace FdUtil


#endif // _FdUtil_Singleton_hh_

