
// 
// File   : TIniFile.hh
// 
// Purpose: Implementation of the class TIniFile
//
// Author : Jens Berger, berger@ikf.uni-frankfurt.de
// 
// $Id: TIniFile.cc 8758 2010-09-27 12:16:24Z mathes $
//


#include <iostream>


/** @file TIniFile.cc
  * Implementation of the class TIniFile.
  * @author Jens Berger, berger@ikf.uni-frankfurt.de
  */

#include <ErrorLogging.hh>

#include "TIniFile.hh"

using namespace std;
using namespace FdRoot;

// -------------------------------------------------------------------------

// Constructor, that opens or creates an ini-file
TIniFile::TIniFile(const TString& filename)
 : fLastErrorCode(kSuccess)
 {
  // open file
  fFileName = filename;

  fstream fin(fFileName,ios::in); 
    
  char *fs= new char[kMaxInifileSize];
  
  if ( fin.good() ) {
    FD_COUT << "TIniFile::TIniFile() - "
            << "ini-file " <<  fFileName << " is now open for reading."  
	    << endl;
    fin.get( fs, kMaxInifileSize, EOF );
    fFStr = new TString(fs);
    delete [] fs;
    fin.close();
  }
  else {
    FD_CERR << "TInifile::TInifile() - "
            << "cannot open ini-file " <<  fFileName << " for reading." 
	    << endl;
    FD_CWARN << "Ini-file " << fFileName << " will be created." << endl;
    // File created with blanks
    fFStr = new TString("");  
  }
}

// -------------------------------------------------------------------------

// Destructor, writes the fFStr->Data to file, closes file and cleans up
TIniFile::~TIniFile()
 {
  fstream fout(fFileName,ios::out|ios::trunc); 
  if ( fout.good() ) {
    FD_CINFO << "TIniFile::~TIniFile() - "
             << "opening ini-file " << fFileName 
	     << " for writing init parameters." << endl;
    fout << fFStr->Data();
    fout.close();  
  }
  else
    FD_CERR << "TIniFile::~TIniFile() - "
            << "Cannot create ini-file " << fFileName 
	    << " for writing init parameters  :-(" << endl;
    
  if ( fFStr ) delete fFStr;
}

// -------------------------------------------------------------------------

// Get section from filestring
TIniFile::ETInifileStatusCode 
TIniFile::GetSection(const TString& Section, int& spos, int& sn, int& pos, int& n)
 {
  TString BrackSectStr(Section);
  BrackSectStr.Prepend("[");
  BrackSectStr.Append("]");

  int BrackSectStrPos = 0;

  BrackSectStrPos = fFStr->Index( BrackSectStr );
 
  // Not able to find section
  if ( BrackSectStrPos < 0 ) {
    spos = 0; sn = 0; pos = 0; n = 0;
    return kSectionNotFound;
  }
  
  spos = BrackSectStrPos;
  pos = BrackSectStrPos + (BrackSectStr.Length());
  int SectStopPos = fFStr->Index("[",pos)-1;
  int WhSpaces = fFStr->Index("\n\n",pos);
  if ( WhSpaces < SectStopPos ) SectStopPos=WhSpaces;
  if ( SectStopPos < 0 ) SectStopPos = fFStr->Length()-1;  
  n = ( SectStopPos-pos)+1;
  sn = ( SectStopPos-spos)+1;

  return kSuccess;
}

// -------------------------------------------------------------------------

TIniFile::ETInifileStatusCode 
TIniFile::GetSection(const TString& Section, int& spos, int& sn)
 {
  int pos = 0; int n = 0;
  
  return GetSection( Section, spos, sn, pos, n );
}

// -------------------------------------------------------------------------

TIniFile::ETInifileStatusCode 
TIniFile::GetVarAreaInSection(const TString& Section, int& pos, int& n)
 {
  int spos = 0; int sn = 0;
  
  return GetSection( Section, spos, sn, pos, n );
}

// -------------------------------------------------------------------------

// Get variable index from section in filestring
TIniFile::ETInifileStatusCode 
TIniFile::GetVarInSect(const TString& Section, const TString& Variable, int& lpos, int& ln ,int& pos, int& n)
 {
  int SectPos = 0;
  int SectLen = 0;
  if ( GetVarAreaInSection( Section, SectPos, SectLen ) < 0 ) {
    lpos = 0; ln = 0; pos = 0; n = 0;
    return kSectionNotFound;
  }
 
  TString VarStr(Variable);
  VarStr.Append("=");
  int VarStartPos = fFStr->Index( VarStr, SectPos );
  
  // Can't find variable in section
  if ( VarStartPos<0 || SectLen<1 ) {
    lpos = 0; ln = 0; pos = 0; n = 0;
    return kVariableNotFound;
  }
 
  // Variable is in another section
  if ( VarStartPos>(SectPos+SectLen-1) ) {
    lpos = 0; ln = 0; pos = 0; n = 0;
    return kWrongSection;
  }
  
  lpos = VarStartPos;
  pos = VarStartPos + VarStr.Length();
  int newline = fFStr->Index("\n",pos);
  ln = (newline-VarStartPos)+1;
  n = ((newline-1)-pos)+1;
  
  return kSuccess;
}

// -------------------------------------------------------------------------

TIniFile::ETInifileStatusCode 
TIniFile::GetVarDataInSect(const TString& Section, const TString& Variable, int& pos, int& n)
 {
  int lpos = 0; int ln = 0;
  
  return GetVarInSect( Section, Variable, lpos, ln, pos, n );
}

// -------------------------------------------------------------------------

TIniFile::ETInifileStatusCode 
TIniFile::GetVarLineInSect(const TString& Section, const TString& Variable, int& lpos, int& ln)
{
 int pos = 0; int n = 0;
 
 return GetVarInSect( Section, Variable, lpos, ln, pos, n );
}

// -------------------------------------------------------------------------

// remove section with all its variables
TIniFile::ETInifileStatusCode 
TIniFile::DeleteSection(const TString& Section)
 {
  int pos = 0; int n = 0;

  if ( (fLastErrorCode = GetSection( Section, pos, n )) < 0 )
    return fLastErrorCode;
  
  fFStr->Remove( pos, n+1 );
  
  return kSuccess;
}

// -------------------------------------------------------------------------

// remove Variable in Section
TIniFile::ETInifileStatusCode
TIniFile::DeleteVarInSect(const TString& Section, const TString& Variable)
 {
  int lpos = 0;
  int ln = 0;

  if ( (fLastErrorCode = GetVarLineInSect( Section, Variable, lpos, ln )) < 0 )
    return fLastErrorCode;

  fFStr->Remove( lpos, ln );
  
  return kSuccess;
}

// -------------------------------------------------------------------------

// read methods
// 1. TStrings
void TIniFile::Read(const TString& Section, const TString& Variable, 
                    const TString& Default, TString& Result)
 {
  int VarPos = 0;
  int VarLen = 0;
  
  fLastErrorCode = GetVarDataInSect(Section,Variable,VarPos,VarLen);
  if ( fLastErrorCode == kSectionNotFound ) {
    FD_CERR << "TIniFile::Read() - "
            << "not able to find section '" << Section 
            << "' in ini-file '" << fFileName 
	    << "' using default " << Default << endl;
    Result = Default;
    return;
  }
  else if ( fLastErrorCode == kVariableNotFound ) {
    FD_CERR << "TIniFile::Read() - "
            << "not able to find variable '" << Variable
	    << "' in section '" << Section
	    << "' in ini-file '" << fFileName 
	    << "' using default " << Default << endl;
    Result = Default;
    return;
  }
  else if ( fLastErrorCode == kWrongSection ) {
    FD_CERR << "TIniFile::Read() - "
         << "not able to find variable '" << Variable
	 << "' in section '" << Section 
	 << "' it's in another section !"
	 << " In ini-file '" << fFileName 
	 << "' using default " << Default << endl;
    Result = Default;
    return;
  }

  // Extract substring
  TString s = fFStr->Copy();
  s.Replace( 0, VarPos, "", 0 );
  Result = s.Replace( VarLen, s.Length()-(VarLen), "", 0 );
}

// -------------------------------------------------------------------------

// 2. integers, booleans
void TIniFile::Read(const TString& Section, 
                    const TString& Variable, const int Default, int& Result)
 {
  TString ResultStr;
  char *DefaultText = new char[20];
 
  sprintf( DefaultText, "%d", Default ); 
  Read( Section, Variable, DefaultText, ResultStr );
  Result = atoi( ResultStr );
 
  delete [] DefaultText;
}

// -------------------------------------------------------------------------

// 4. floats
void TIniFile::Read(const TString& Section, 
                    const TString& Variable, const float Default, float& Result)
 {
  TString ResultStr;
  char *DefaultText = new char[30];
  
  sprintf( DefaultText, "%f", Default ); 
  Read( Section, Variable, DefaultText, ResultStr );
  Result = atof( ResultStr );
 
  delete [] DefaultText;
}

// -------------------------------------------------------------------------

// 5. doubles
void TIniFile::Read(const TString& Section, 
                    const TString& Variable, const double Default, double& Result)
 {
  TString ResultStr;
  char *DefaultText = new char[40];
  sprintf( DefaultText, "%f", Default ); 
  Read( Section, Variable, DefaultText, ResultStr );
  Result = atof( ResultStr );
 
  delete [] DefaultText;
}

// -------------------------------------------------------------------------

// write methods
// 1. TStrings
void TIniFile::Write(const TString& Section, 
                     const TString& Variable, const TString& Value)
 {
  int pos = 0;
  int n = 0;
  TString VarStr( Variable );
  VarStr.Append( "=" );
 
  fLastErrorCode = GetVarDataInSect( Section, Variable, pos, n );
 
  if ( fLastErrorCode == kSectionNotFound ) {
    // First parameter in section
    
    TString SectionStr( Section );
    
    TString EndStr1= fFStr->Copy();
   
    EndStr1.Replace( 0, EndStr1.Length()-2, "", 0 );
    EndStr1.Replace( 2, EndStr1.Length()-(2), "", 0 );
    TString EndStr2 = fFStr->Copy();
    EndStr2.Replace( 0, EndStr2.Length()-1, "", 0 );
    EndStr2.Replace( 1, EndStr2.Length()-(1), "", 0 );

    
    if ( !strcmp( EndStr1, "\n\n" ) ) {
      SectionStr.Prepend( "[" );
    }
    else if ( strcmp( EndStr1, "\n\n" ) && !strcmp( EndStr2, "\n" ) ) {
      SectionStr.Prepend( "\n[" );
    }
    else {
      SectionStr.Prepend( "\n\n[" );
    }
      
    SectionStr.Append( "]\n" );
    
    // Adding first parameter to SectionStr
    SectionStr.Append( VarStr );
    SectionStr.Append( Value );
    SectionStr.Append( "\n" );
    
    fFStr->Append( SectionStr );
    
    return;
  }
  else if ( fLastErrorCode == kVariableNotFound ) {
    // Following parameters
    
    int SectPos = 0;
    int SectN = 0;
    
    GetSection( Section, SectPos, SectN );
    TString VarLineStr( VarStr );
    VarLineStr.Append( Value );
    VarLineStr.Append( "\n" );
    
    fFStr->Replace( SectPos+SectN, 0, VarLineStr, VarLineStr.Length() );
    return;
  }
  else if ( fLastErrorCode == kWrongSection ) {
    FD_CERR << "TIniFile::Write() - "
            << "err=kWrongSection, not able to find variable '" << Variable
	    << "' in section '" << Section 
	    << " it's in another section !" << " In ini-file '"
	    << fFileName << "' nothing is written" << endl;
    return;
  }
  
  // File exists
  fFStr->Replace( pos, n, Value, Value.Length() );
  
  return;
}

// -------------------------------------------------------------------------

// 2. integers, booleans
void TIniFile::Write(const TString& Section, 
                     const TString& Variable, const int Value)
 {
  char *ValueText = new char[20];
  
  sprintf( ValueText, "%d", Value ); 
  Write( Section, Variable, ValueText );

  delete [] ValueText;
}

// -------------------------------------------------------------------------

// 4. floats
void TIniFile::Write(const TString& Section, 
                     const TString& Variable, const float Value)
 {
  char *ValueText = new char[30];

  sprintf( ValueText, "%f", Value ); 
  Write( Section, Variable, ValueText );

  delete [] ValueText;
}

// -------------------------------------------------------------------------

// 5. double
void TIniFile::Write(const TString& Section, 
                     const TString& Variable, const double Value)
 {
  char *ValueText = new char[30];

  sprintf( ValueText, "%f", Value ); 
  Write( Section, Variable, ValueText );

  delete [] ValueText;
}

// -------------------------------------------------------------------------
// -------------------------------------------------------------------------
