
//
// File   : TSocketAcceptor.cc
//
// Purpose: Implementation of the class TSocketAcceptor
//
// $Id: TSocketAcceptor.cc 11950 2011-01-25 15:17:30Z mathes $
//


/** @file TSocketAcceptor.cc
  * Implementation of the class TSocketAcceptor.
  * @author H.-J. Mathes, FzK
  */
  

#include <iostream>
#include <cstdio>
#include <cstdlib>
#include <cassert>


#include <TList.h>
#include <TMessage.h>
#include <TMonitor.h>
#include <TIterator.h>
#include <TServerSocket.h>
#include <TSocket.h>

#include <ErrorLogging.hh>

#include "TSocketAcceptor.hh"

using namespace std;
using namespace FdRoot;

// -------------------------------------------------------------------------
// -------------------------------------------------------------------------
//
// class TSocketAcceptor - implementation
//

int TSocketAcceptor::fgPrintLevel = 0;

TSocketAcceptor::TSocketAcceptor()
 : fCurrentSocket(0), fMaxClients(1), fMonitor(0), 
   fNumClients(0), fServerSocket(0), fStatus(0)
 {
  fMonitor = new TMonitor;
}

// -------------------------------------------------------------------------

TSocketAcceptor::~TSocketAcceptor()
 {
  delete fMonitor;
  delete fServerSocket;
}

// -------------------------------------------------------------------------

void TSocketAcceptor::_Accept()
 {
  fCurrentSocket = fServerSocket->Accept();
    
  if ( !fCurrentSocket ) return;
    
  if ( fNumClients == fMaxClients ) {  // no further clients allowed

    delete fCurrentSocket;
    fCurrentSocket = NULL;

    return;
  }
  
  if ( ( fStatus = fCurrentSocket->GetErrorCode() ) ) {

    FD_CERR << "TSocketAcceptor::_Accept() - "
            << "error " << fCurrentSocket->GetErrorCode() 
            << " on accepted socket !" << endl;

    delete fCurrentSocket;
    fCurrentSocket = NULL;

    return;
  }
  
  OpenConnection();
}

// -------------------------------------------------------------------------

void TSocketAcceptor::Close()
 {
  if ( fServerSocket ) {
    fServerSocket->Close();
    delete fServerSocket;
    
    fServerSocket = NULL;
  }
}

// -------------------------------------------------------------------------

void TSocketAcceptor::OpenConnection()
 {
  if ( fgPrintLevel > 1 )
    FD_COUT << "TSocketAcceptor::OpenConnection() - "
            << "client connected !" << endl;
  
  fMonitor->Add( fCurrentSocket );
  fNumClients++;
    
//  if ( fNumClients == fMaxClients ) {
//    fMonitor->Remove( fServerSocket );  // no further clients allowed
//  }
  
  Connect();
}

// -------------------------------------------------------------------------

void TSocketAcceptor::CloseConnection()
 {
  if ( fgPrintLevel > 1 )
    FD_COUT << "TSocketAcceptor::CloseConnection() - "
            << "client has disconnected !" << endl;

  if ( fgPrintLevel ) {
    FD_COUT << "TSocketAcceptor::CloseConnection() - "
            << "received " << (float)fCurrentSocket->GetBytesRecv()/1024. 
            << " kBytes" << endl;
  }
  
  fMonitor->Remove( fCurrentSocket );
	  
  fNumClients--;
  if ( fNumClients < fMaxClients )  // add eventually server socket to the TMonitor
    fMonitor->Add( fServerSocket );
  
  Disconnect();
  
  delete fCurrentSocket;
  fCurrentSocket = NULL;
}

// -------------------------------------------------------------------------

bool TSocketAcceptor::HandleString(char *message)
 {
  if ( fgPrintLevel > 1 ) {
    FD_COUT << "TSocketAcceptor::HandleString() - "
            << "received kMESS_STRING" << endl;
    FD_COUT << message << endl;
  }
  
  return true;
}

// -------------------------------------------------------------------------

bool TSocketAcceptor::HandleTObject(TObject* obj)
 {
  FD_CERR << "TSocketAcceptor::HandleTObject() - "
          << "Message '" << obj->Class_Name() << "' not handled !" << endl;
  
  delete obj;
  
  return true;
}

// -------------------------------------------------------------------------

bool TSocketAcceptor::Open(unsigned short port)
 {
  Close();   // close eventually existing server port
  
  fServerSocket = new TServerSocket( port, kTRUE, 0 );
  
  if ( fServerSocket ) {
    if ( ( fStatus = fServerSocket->GetErrorCode() ) ) {
      FD_CERR << "TSocketAcceptor::Open() - "
              << "TServerSocket::GetErrorCode()= " 
              << fServerSocket->GetErrorCode() << endl;
      return false;
    }

    fMonitor->Add( fServerSocket );
    
    return true;
  }
  else
    return false;
}

// -------------------------------------------------------------------------

void TSocketAcceptor::Run(Long_t millisec)
 {
  fStatus = 0;   // reset status and error flag
  
  if ( ( fCurrentSocket = fMonitor->Select( millisec ) ) == (TSocket *)-1 ) {
    return;  // TMonitor::Select() timed out
  }
  
  if ( fCurrentSocket == fServerSocket ) {
    _Accept();
    fCurrentSocket = NULL;
    return;
  }
  
  if ( fCurrentSocket->IsValid() ) {
  
    TMessage *mess;
    
    int msg_length = fCurrentSocket->Recv( mess );

    if ( msg_length > 0 ) {  // received some message 
     
      char str_buffer[512];
      bool mess_handled = false;
     
      mess_handled = HandleMessage( mess );
      
      if ( !mess_handled ) { // call the standard handlers
        
	switch ( mess->What() ) {
	
	  case kMESS_OBJECT:
	       mess_handled = HandleTObject( mess->ReadObject( mess->GetClass() ) );
	       break;
	 
	  case kMESS_STRING:
	       mess->ReadString( str_buffer, sizeof(str_buffer) );
	       mess_handled = HandleString( str_buffer );
	       break;

	  default:
	       FD_CERR << "TSocketAcceptor::Run(" << millisec << ") - "
	               << "unknown message type: " << mess->What() << endl;
	       delete mess->ReadObject( mess->GetClass() );
        }
	
	FD_CERR << "TSocketAcceptor::Run(" << millisec << ") - "
	        << "message handled: " << (int)mess_handled << endl;
      }
      
      delete mess;
      
    } 
    else if ( msg_length == 0 ) { // msg_length == 0 => disconnection
      CloseConnection();
    }
    else {  // msg_length < 0 = error => disconnection
      FD_CERR << "TSocketAcceptor::Run(" << millisec << ") - "
              << "TSocket::Recv()= " << msg_length << endl;
      CloseConnection();
    }
  }
  else {  // fCurrentSocket->IsValid()
    FD_CERR << "TSocketAcceptor::Run(" << millisec << ") - "
            << "socket to client not Valid !" << endl;

    delete fCurrentSocket;
    fCurrentSocket = NULL;
  }
}

// -------------------------------------------------------------------------

int TSocketAcceptor::Send(TObject* obj)
 {
  if ( !fCurrentSocket || !fCurrentSocket->IsValid() ) return -1;
  
  return fCurrentSocket->SendObject( obj, kMESS_OBJECT );
}
    
// -------------------------------------------------------------------------

int TSocketAcceptor::Send(TMessage* mess)
 {
  if ( !fCurrentSocket || !fCurrentSocket->IsValid() ) return -1;
  
  return fCurrentSocket->Send( *mess );
}

// -------------------------------------------------------------------------

int TSocketAcceptor::Send(const char* str)
 {
  if ( !fCurrentSocket || !fCurrentSocket->IsValid() ) return -1;
  
  return fCurrentSocket->Send( str, kMESS_STRING );
}

// -------------------------------------------------------------------------

void TSocketAcceptor::SetSocketOption(ESockOptions opt, int val)
 {
  // --- loop over list of all active sockets stored in the TMonitor
  
  TIterator *iter = fMonitor->GetListOfActives()->MakeIterator();
  
  TSocket * sock = (TSocket *)iter->Next();
  
  do {
    if ( sock == NULL ) continue;  // empty list
    
    if ( !sock->IsValid() ) {
      FD_CERR << "TSocketAcceptor::SetSocketOption() - "
              << "socket is invalid !" << endl;
      continue;
    }
    
#warning Tbd: do we need to set the options also for the prototype socket ?
    if ( sock == fServerSocket ) continue;
    
    int status = sock->SetOption( opt, val );
    
    if ( status ) {
      FD_CERR << "TSocketAcceptor::SetSocketOption() - "
              << "status= " << status << endl;
    }
  }
  while ( iter->Next() != NULL );

  // --- loop over list of all deactive sockets stored in the TMonitor
  
  iter = fMonitor->GetListOfDeActives()->MakeIterator();
  
  sock = (TSocket *)iter->Next();
  
  do {
    if ( sock == NULL ) continue;  // empty list
    
    if ( !sock->IsValid() ) {
      FD_CERR << "TSocketAcceptor::SetSocketOption() - "
              << "socket is invalid !" << endl;
      continue;
    }
    
#warning Tbd: do we need to set the options also for the prototype socket ?
    if ( sock == fServerSocket ) continue;
    
    int status = sock->SetOption( opt, val );
    
    if ( status ) {
      FD_CERR << "TSocketAcceptor::SetSocketOption() - "
              << "status= " << status << endl;
    }
  }
  while ( iter->Next() != NULL );

  map<ESockOptions,int>::iterator found = fOptionMap.find( opt );
  
  if ( found != fOptionMap.end() ) {
    fOptionMap.erase( found );
  }
  
  fOptionMap.insert( pair<ESockOptions,int>( opt, val ) );
}

// -------------------------------------------------------------------------

void TSocketAcceptor::SetStoredSocketOptions(TSocket *sock)
 {
  assert( sock != NULL );
  
  if ( !fOptionMap.empty() ) {
  
    map<ESockOptions,int>::iterator iter = fOptionMap.begin();

    while ( iter != fOptionMap.end() ) {
      
      int status = sock->SetOption( iter->first, iter->second );
      
      if ( status ) {
  	FD_CERR << "TSocketAcceptor::SetStoredSocketOptions() - "
  		<< "status= " << status << endl;
      }
      
      iter++;
    }
  }
}

// -------------------------------------------------------------------------
// -------------------------------------------------------------------------
