
//
// File   : TSocketAcceptor.hh
//
// Purpose: Declaration of the class TSocketAcceptor
//
// $Id: TSocketAcceptor.hh 11950 2011-01-25 15:17:30Z mathes $
//


/** @file TSocketAcceptor.hh
  * Declaration of the class TSocketAcceptor.
  * @author H.-J. Mathes, FzK
  */
  

#ifndef _FdRoot_TSocketAcceptor_hh_
#define _FdRoot_TSocketAcceptor_hh_

#include <iostream>
#include <map>

#include <Rtypes.h>
#include <TSocket.h>                  // ESockOptions

// --- forward declaration(s)

class TMessage;
class TMonitor;
class TObject;
class TServerSocket;

namespace FdRoot {

/** This class implements a socket server for ROOT sockets which might exchange
  * objects using a high level protocol.
  */
class TSocketAcceptor {
public:
    /** Constructor of class TSocketAcceptor.
      *
      *  All internal variables are initialized and a new TMonitor 
      * (socket monitor class) object is created.
      */
    TSocketAcceptor();

    /** Destructor of class TSocketAcceptor.
      *
      * The server socket and the TMonitor object are deleted.
      */
    virtual ~TSocketAcceptor();

    /** Close an existing server socket. 
      *
      * Already existing and accepted client connections are not affected 
       * by this call.
      */
    virtual void Close();
    
    /** Callback: called when a client has connected. */
    virtual void Connect() {}

    /** Callback: called when a client has disconnected. */
    virtual void Disconnect() {}

    /** Get the socket handle (TSocket) of the socket which was just active
      * (either accept or message received). 
      *
      * The return value must be checked to be not NULL !
      */
    virtual TSocket* GetCurrentSocket()
     { return fCurrentSocket; }

    /** Get the number of clients currently connected. */
    virtual int GetNumClients()
     { return fNumClients; }
    
    /** Get the status of the servers last operation. */
    virtual int GetStatus()
     { return fStatus; }

    /** Callback: called to handle all (user defined) messages types.
      *
      * Has to return kTRUE if message could be handled, kFALSE otherwise.
      */
    virtual bool HandleMessage(TMessage*)
     { std::cerr << "TSocketAcceptor::HandleMessage(): "
            << "Message not handled !" << std::endl; return false; }

    /** Create a server socket on the specified port. 
      *
      * If a server socket was already created before, it will be closed.
      * This method returns kTRUE if the operation was successful, 
      * kFALSE otherwise.
      */
    virtual bool Open(unsigned short port = 4000);
  
    /** Run the server which might wait a specified time for a socket message.
      */
    virtual void Run(Long_t millisec=20);
    
    /** Send an object via the socket from where a message was recently
      * received.
      *
      * The return value of this method is the return value of 
      * the underlying TSocket::Send*(), i.e.:
      * @li -1 in case of error.
      * @li -4 in case of kNoBlock and errno == EWOULDBLOCK
      * @li -5 (new) in case of invalid or non-existing socket
      */
    virtual int Send(TObject*);
    
    /** Send an object already wrapped into a TMessage via the socket from 
      * where a message was recently received.
      */
    virtual int Send(TMessage*);
    
    /** Send a string via the socket from where a message was recently received. */
    virtual int Send(const char*);
    
    /** Send a string via the socket from where ... */
    virtual int Send(const std::string& message)
     { return Send( message.c_str() ); }

    /** Set the maximum number of clients the server can handle. */
    void SetMaxClients(unsigned int max_clients)
     { fMaxClients = max_clients; }
     
    /** Set the verbose print level of this class. */
    static void SetPrintLevel(int level=1)
     { fgPrintLevel = level; }

    /** Set options for the TSocket. 
      *
      * @see ROOT::TSocket.h for the possible ESockOptions.
      */
    void SetSocketOption(ESockOptions opt, int val);
    
    /** Set the stored socket options for a newly created TSocket. */
    void SetStoredSocketOptions(TSocket*);
    
protected:

    /** 'accept()' method of the socket server.
      *
      * After the accept() has been done, the return code of the operation as
      * well the newly created socket to the client are checked for errors. In
      * case of an error the socket to the cleint is deleted, otherwise the
      * _Connect() method is called.
      */
    void _Accept();

    /** Internal method of the TSocketAcceptor class.
      * 
      * This method is called when a client has successfully connected to the
      * server. The client socket is then added to the TMonitor object for 
      * further watching and if the maximum allowed number of clients is 
      * already reached, the server socket is deactivated (= removed from the 
      * list sockets watched by the TMonitor. Then the Connect() method is 
      * called giving derived classes the possibility to add own specific 
      * 'connect' specific actions.
      */
    void OpenConnection();

    /** Internal method of the TSocketAcceptor class.
      *
      * This method is called when a client is normally or abnormally
      * disconnected. The socket where the disconnection occurred is removed 
      * from the list of watched sockets of the TMonitor. Then the Disconnect() 
      * method is called giving derived classes the possibility to add own 
      * 'disconnect' specific actions.
      */
    void CloseConnection();

    /** Default handler for string messages. */
    bool HandleString(char *);

    /** Default handler for object messages. */
    bool HandleTObject(TObject*);

private:
    TSocket        *fCurrentSocket;   // the socket of the current activity or
                                      // message
    int             fMaxClients;      // maximum number of clients allowed
    TMonitor       *fMonitor;         // the socket monitor object
    int             fNumClients;      // number of clients currently connected
    TServerSocket  *fServerSocket;    // the servers prototype socket
    int             fStatus;          // status of the last operation

    static int      fgPrintLevel;      // the verbose print level of this class
    
    std::map<ESockOptions,int> fOptionMap;
};

} // namespace FdRoot

#endif  // _FdRoot_TSocketAcceptor_hh_
