
//
// File   : TSocketConnector.cc
//
// Purpose: Implementation of the class TSocketConnector
//
// $Id: TSocketConnector.cc 11950 2011-01-25 15:17:30Z mathes $
//


/** @file TSocketConnector.cc
  * Implementation of the class TSocketConnector.
  * @author H.-J. Mathes, FzK
  */
  

#include <iostream>
#include <cstdio>
#include <cstdlib>
#include <cassert>
#include <cerrno>

#include <unistd.h>
#include <fcntl.h>        // fcntl()
#include <sys/time.h>     // struct timeval
#include <sys/types.h>
#include <sys/select.h>   // select()
#include <sys/socket.h>   // socket(), connect()
#include <netinet/in.h>   // htons()
#include <arpa/inet.h>    // struct sockaddr_in
#include <netdb.h>        // gethostbyname()


#include <TMessage.h>
#include <TMonitor.h>
#include <TServerSocket.h>
#include <TSocket.h>

#include <ErrorLogging.hh>

#include "TSocketConnector.hh"

using namespace std;
using namespace FdRoot;

// -------------------------------------------------------------------------
// -------------------------------------------------------------------------
//
// class TSocketConnector - implementation
//

int TSocketConnector::fgPrintLevel = 0;

const char TSocketConnector::kDefaultServerName[] = "127.0.0.1";

TSocketConnector::TSocketConnector()
 : fDoReconnect(false),fIsConnected(false),
   fMonitor(NULL), fServerName(kDefaultServerName), 
   fServerPort(kDefaultServerPort), fSocket(NULL), fStatus(0)
 {
  fMonitor = new TMonitor();
}

// -------------------------------------------------------------------------

TSocketConnector::~TSocketConnector()
 {
  Close();
  
  delete fMonitor;
}

// -------------------------------------------------------------------------

void TSocketConnector::Close()
 {
#if 1
  if ( fSocket ) {
    fSocket->Close();
    
    fMonitor->Remove( fSocket );
    
    fIsConnected = false;
  
    delete fSocket;
    fSocket = NULL;
  }
#else
  CloseConnection();
#endif
}

// -------------------------------------------------------------------------

void TSocketConnector::CloseConnection()
 {
  if ( fgPrintLevel > 1 )
    FD_COUT << "TSocketConnector::CloseConnection() - "
            << "client has disconnected from server !" << endl;
  
  if ( fSocket ) fMonitor->Remove( fSocket );

  fIsConnected = false;
  
  Disconnect();  // call method of derived class
  
  delete fSocket;
  fSocket = NULL;
}

// -------------------------------------------------------------------------

TSocket *
TSocketConnector::CreateSocketWithTimeout(const char *server,
                                          unsigned short port, 
					  unsigned int timeout)
 {
  if ( !timeout ) {
    FD_CWARN << "TSocketConnector::CreateSocketWithTimeout("
             << timeout << ") - no timeout specified, "
	     << "using 'standard' TSocket ctor!" << endl;
    return new TSocket( server, port, 0 );
  }
  
  struct sockaddr_in addr;
  unsigned long inaddr;
  const char *host = server;
  
  // code from Sock/src/socklib.c to resolve the host name
  
  struct hostent *hp, host_info;
  
  bzero ( (char *)&addr, sizeof( addr ) );

#warning Tbd: replace inet_addr() by inet_aton() for cleaner error interface!
  if ( (inaddr = inet_addr(host) ) != INADDR_NONE ) {
    bcopy( (char *)&inaddr, (char *)&addr.sin_addr, sizeof( inaddr ) );
    host_info.h_name = NULL;
  }
  else {
   
    if ( ( hp = gethostbyname(host) ) == NULL )  {
      FD_CERR << "TSocketConnector::CreateSocketWithTimeout() - "
    	      << "gethostbyname: host name error: " 
	      << host << " " << strerror(errno) << endl;
      return NULL;
    }
     
    host_info = *hp;
    bcopy( hp->h_addr, (char *)&addr.sin_addr, hp->h_length );
  }
  
  addr.sin_family = AF_INET; 
  addr.sin_port = htons( port ); 
  //addr.sin_addr.s_addr = inet_addr( fServerName.c_str() ); 

  // code from: http://developerweb.net/forum/showthread.php?p=13486

  int res, soc; 
  long arg; 
  fd_set myset; 
  struct timeval tv; 
  int valopt; 
  socklen_t lon; 

  // Create socket 
  soc = socket( AF_INET, SOCK_STREAM, 0 ); 
  
  if ( soc < 0 ) { 
    FD_CERR << "TSocketConnector::CreateSocketWithTimeout() - "
            << "error creating socket (" 
	    << errno << " " << strerror(errno) << ")" << endl; 
    return NULL; 
  } 

  // Set non-blocking 
  if ( (arg = fcntl( soc, F_GETFL, NULL ) ) < 0) { 
    FD_CERR << "TSocketConnector::CreateSocketWithTimeout() - "
            << "error fcntl(..., F_GETFL) (" << strerror(errno) << ")" << endl; 
    return NULL; 
  }
  
  arg |= O_NONBLOCK; 
  
  if ( fcntl( soc, F_SETFL, arg) < 0 ) { 
    FD_CERR << "TSocketConnector::CreateSocketWithTimeout() - "
            << "error fcntl(..., F_SETFL) (" << strerror(errno) << ")" << endl; 
    return NULL; 
  }
  
  // Trying to connect with timeout 
  res = connect( soc, (struct sockaddr *)&addr, sizeof(addr) ); 
  
  if ( res < 0 ) { 
    
    if ( errno == EINPROGRESS ) { 
      
      FD_CWARN << "TSocketConnector::CreateSocketWithTimeout() - "
               << "EINPROGRESS in connect() - selecting ..." << endl; 
      
      do { 
        tv.tv_sec = timeout;
        tv.tv_usec = 0;
        FD_ZERO( &myset );
        FD_SET( soc, &myset );
	
        res = select( soc+1, NULL, &myset, NULL, &tv );
        if ( res < 0 && errno != EINTR ) { 
          FD_CERR << "TSocketConnector::CreateSocketWithTimeout() - "
                  << "error connecting " << errno << " - " 
		  << strerror(errno) << endl; 
          return NULL; 
        }
        else if ( res > 0 ) {
          // Socket selected for write
          lon = sizeof(int);
          if ( getsockopt( soc, SOL_SOCKET, SO_ERROR, (void*)(&valopt), &lon ) < 0 ) { 
            FD_CERR << "TSocketConnector::CreateSocketWithTimeout() - "
                    << "error in getsockopt() " << errno << " - " 
	  	    << strerror(errno) << endl; 
            return NULL; 
          } 
        
	  // Check the value returned... 
          if ( valopt ) { 
            FD_CERR << "TSocketConnector::CreateSocketWithTimeout() - "
                    << "error in delayed connect() " << valopt 
		    << " - " << strerror(valopt) << endl; 
            return NULL; 
          } 
          break; 
        } 
        else { 
          FD_CERR << "TSocketConnector::CreateSocketWithTimeout() - "
                  << "timeout in select() - cancelling!" << endl; 
          return NULL; 
        } 
      } while ( 1 ); 
    } 
    else { 
      FD_CERR << "TSocketConnector::CreateSocketWithTimeout() - "
              << "error connecting " << errno << " - "
	      << strerror(errno) << endl; 
      return NULL; 
    } 
  }
  
  // Set to blocking mode again... 
  if ( ( arg = fcntl( soc, F_GETFL, NULL ) ) < 0 ) { 
    FD_CERR << "TSocketConnector::CreateSocketWithTimeout() - "
            << "error fcntl(..., F_GETFL) (" 
	    << strerror(errno) << ")" << endl; 
    return NULL; 
  } 
  
  arg &= ~O_NONBLOCK; 
  
  if ( fcntl ( soc, F_SETFL, arg ) < 0 ) { 
    FD_CERR << "TSocketConnector::CreateSocketWithTimeout() - "
            << "error fcntl(..., F_SETFL) (" << strerror(errno) << ")" << endl; 
    return NULL; 
  } 

  return new TSocket( soc );
}

// -------------------------------------------------------------------------

bool TSocketConnector::HandleMessage(TMessage *)
 {
  FD_CERR << "TSocketConnector::HandleMessage(): "
          << "Message not handled !" << endl; 
  
  return false;
}

// -------------------------------------------------------------------------

bool TSocketConnector::HandleString(const char *text)
 {
  if ( fgPrintLevel > 1 ) {
    FD_COUT << "TSocketConnector::HandleString() - "
            << "received kMESS_STRING" << endl;
    FD_COUT << text << endl;
  }
  
  return true;
}

// -------------------------------------------------------------------------

bool TSocketConnector::HandleTObject(TObject *obj)
 {
  FD_CERR << "TSocketConnector::HandleTObject() - "
          << "Message '" << obj->Class_Name() << "' not handled !" << endl;

  delete obj;
  
  return true;
}

// -------------------------------------------------------------------------

bool TSocketConnector::Open(unsigned int timeout)
 {
  Close();   // close eventually existing connection
  
  assert( !fServerName.empty() );
  assert( fServerPort > 0 );
  
  fSocket = CreateSocketWithTimeout( fServerName.c_str(), fServerPort, timeout );
  
  if ( fSocket ) {
  
    if ( ( fStatus = fSocket->GetErrorCode() ) != 0 ) {
      FD_CERR << "TSocketConnector::Open() - TSocket::GetErrorCode()= " 
              << fStatus << endl;
      return false;
    }

    SetStoredSocketOptions( fSocket );
    
    fMonitor->Add( fSocket );
    
    return true;
  }
  else
    return false;
}

// -------------------------------------------------------------------------

bool TSocketConnector::Open(const char *server,unsigned short port,
                                               unsigned int timeout)
 {
  fServerName = server;
  fServerPort = port;
  
  return Open( timeout );
}

// -------------------------------------------------------------------------

void TSocketConnector::OpenConnection()
 {
  if ( fgPrintLevel > 1 )
    FD_COUT << "TSocketConnector::OpenConnection() - "
            << "client connected !" << endl;
  
  fIsConnected = true;
  
  fMonitor->Add( fSocket );
  
  Connect();  // call method of derived class
}

// -------------------------------------------------------------------------

void TSocketConnector::Run(Long_t millisec)
 {
  if ( !fSocket || !fSocket->IsValid() ) {
    fStatus = -1;
    return;
  }
  
  fStatus = 0;   // reset status and error flag
  
  TSocket *sock;
  
  if ( ( sock = fMonitor->Select( millisec ) ) == (TSocket *)-1 ) {
    return;  // TMonitor::Select() timed out
  }
  
  if ( sock != fSocket ) {
    FD_CERR << "TSocketConnector::Run(" << millisec << ") - "
            << "received unexpected response !" << endl;
    return;
  }
  
  if ( fSocket->IsValid() ) {
  
    TMessage *mess;
    int msg_length = fSocket->Recv( mess );

    if ( msg_length > 0 ) {  // received some message 
     
      char str_buffer[512];
      bool message_handled = false;
     
      message_handled = HandleMessage( mess );
      
      if ( !message_handled ) { // call the standard handlers
        
	switch ( mess->What() )  {
	
	  case kMESS_OBJECT:
	       message_handled = HandleTObject( mess->ReadObject( mess->GetClass() ) );
	       break;
	 
	  case kMESS_STRING:
	       mess->ReadString( str_buffer, sizeof(str_buffer) );
	       message_handled = HandleString( str_buffer );
	       break;

	  default:
	       FD_CERR << "TSocketConnector::Run(" << millisec << ") - "
	               << "unknown message type: " << mess->What() << endl;
	       
	       delete mess->ReadObject( mess->GetClass() );
        }
	
	if ( fgPrintLevel )
	  FD_COUT << "TSocketConnector::Run(" << millisec << ") - "
	          << "message handled: " << (int)message_handled << endl;
      }
      
      delete mess;
      
    } 
    else if ( msg_length == 0 ) { // msg_length == 0 => disconnection
      CloseConnection();
    }
    else {  // msg_length < 0 = error => disconnection
      FD_CERR << "TsocketConnector::Run(" << millisec << ") - "
              << "TSocket::Recv()= " << msg_length << endl;
      CloseConnection();
    }
  }
  else {  // fCurrentSocket->IsValid()
    FD_CERR << "TsocketConnector::Run(" << millisec << ") - "
            << "socket to server not Valid !" << endl;

    delete fSocket;
    fSocket = NULL;
  }
}

// -------------------------------------------------------------------------

int TSocketConnector::Send(TObject* obj,int mtype)
 {
  if ( !fSocket || !fSocket->IsValid() ) return -1;
  
  return fSocket->SendObject( obj, mtype );
}

// -------------------------------------------------------------------------

int TSocketConnector::Send(TMessage* mess)
 {
  if ( !fSocket || !fSocket->IsValid() ) return -1;
  
  return fSocket->Send( *mess );
}

// -------------------------------------------------------------------------

int TSocketConnector::Send(const char* str)
 {
  if ( !fSocket || !fSocket->IsValid() ) return -1;
  
  return fSocket->Send( str, kMESS_STRING );
}
   
// -------------------------------------------------------------------------

void TSocketConnector::SetSocketOption(ESockOptions opt,int val)
 {
  if ( fSocket != NULL && fSocket->IsValid() ) {
    
    int status = fSocket->SetOption( opt, val );
    
    if ( status ) {
      FD_CERR << "TSocketConnector::SetSocketOption() - "
              << "status= " << status << endl;
    }
  }
  else
    FD_CERR << "TSocketConnector::SetSocketOption() - "
            << "fSocket either invalid or NULL !" << endl;

  map<ESockOptions,int>::iterator found = fOptionMap.find( opt );
  
  if ( found != fOptionMap.end() ) {
    fOptionMap.erase( found );
  }
  
  fOptionMap.insert( pair<ESockOptions,int>( opt, val ) );
}

// -------------------------------------------------------------------------

void TSocketConnector::SetStoredSocketOptions(TSocket *sock)
 {
  assert( sock != NULL );
  
  if ( !fOptionMap.empty() ) {
  
    map<ESockOptions,int>::iterator iter = fOptionMap.begin();

    while ( iter != fOptionMap.end() ) {
      
      int status = sock->SetOption( iter->first, iter->second );
      
      if ( status ) {
  	FD_CERR << "TSocketConnector::SetStoredSocketOptions() - "
  		<< "status= " << status << endl;
      }
      
      iter++;
    }
  }
}

// -------------------------------------------------------------------------
// -------------------------------------------------------------------------
