
//
// File   : TTimeStamp.cc
//
// Purpose: Implementation of the class TTimeStamp
//
// Author :
//
// $Id: TTimeStamp.cc 8758 2010-09-27 12:16:24Z mathes $
//


#include <iostream>

/** @file TTimeStamp.cc
  * Implementation of the class TTimeStamp.
  * @author R. Hatcher, FNAL
  */
  
#include <TString.h>

#include <TimeConvert.hh>
#include <ErrorLogging.hh>

#include "TTimeStamp.hh"

using namespace std;
using namespace FdRoot;

/** Constant: how many nano secs has a second. */
const int nsPerSec = 1000000000;

// -------------------------------------------------------------------------
// -------------------------------------------------------------------------

/** Insert a TTimpestamp into std::ostream */
ostream& operator<<(std::ostream& os, const FdRoot::TTimeStamp& vldts) 
 {
  if ( os.good() ) {
    if ( os.tie() ) os.tie()->flush(); // instead of opfx
    os << vldts.AsString("c");
  }
  // instead of os.osfx()
  if ( os.flags() & ios::unitbuf ) os.flush();
  
  return os;
}

// -------------------------------------------------------------------------
// -------------------------------------------------------------------------
//
// class TTimeStamp - implementation
//

TTimeStamp::TTimeStamp()
 {
  Set();
}

// -------------------------------------------------------------------------

TTimeStamp::TTimeStamp(const timespec_t *ts)
 { 
  if ( ts == NULL ) {
    FD_CWARN << "TTimeStamp::TTimeStamp(const timespec_t*) - "
             << "trying to initialize from NULL-ptr!" << endl;
    Set();
  }
  else {
    fSec = ts->tv_sec;
    fNanoSec = ts->tv_nsec;
    NormalizeNanoSec();
  }
}

// -------------------------------------------------------------------------

TTimeStamp::TTimeStamp(const time_t *t,const int nsec)
 {
  if ( t == NULL ) {
    FD_CWARN << "TTimeStamp::TTimeStamp(const time_t*,...) - "
             << "trying to initialize from NULL-ptr!" << endl;
    fSec = 0;
  }
  else
    fSec = (*t);
  
  fNanoSec = nsec;
  NormalizeNanoSec();
}

// -------------------------------------------------------------------------

TTimeStamp::TTimeStamp(const unsigned int year, const unsigned int month,
                       const unsigned int day,  const unsigned int hour,
                       const unsigned int min,  const unsigned int sec,
                       const unsigned int nsec, 
                       const bool isUTC, const int secOffset)
 {
  Set( year, month, day, hour, min, sec, nsec, isUTC, secOffset );
}

// -------------------------------------------------------------------------

TTimeStamp::TTimeStamp(const unsigned int date, const unsigned int time,
                       const unsigned int nsec, 
                       const bool isUTC, const int secOffset)
 {
  Set( date, time, nsec, isUTC, secOffset );
}

// -------------------------------------------------------------------------

FdRoot::TTimeStamp& TTimeStamp::operator=(const FdRoot::TTimeStamp &rhs)
 { 
  if ( this != &rhs ) {
    fSec = rhs.fSec; 
    fNanoSec = rhs.fNanoSec;
  }
  return *this;
}

// -------------------------------------------------------------------------

const char * TTimeStamp::AsString(Option_t *option) const
 {
  static char formatted[64];  // strftime fields substituted
  static char formatted2[64]; // nanosec field substituted

  TString opt = option;
  opt.ToLower();

  if ( opt.Contains("2") ) {
    // return string formatted as integer {sec,nsec}
    sprintf(formatted,"{%d,%d}",fSec,fNanoSec);
    return formatted;
  }

#ifdef linux
  // under linux %z is the hour offset and %Z is the timezone name
  const char *RFC822   = "%a, %d %b %Y %H:%M:%S %z (%Z) +#9ld nsec";
  const char *ISO8601  = "%Y-%m-%d %H:%M:%S.#9.9ld%z";
  const char *ISO8601Z = "%Y-%m-%d %H:%M:%S.#9.9ldZ";
#else
  // otherwise only %Z is guarenteed to be defind
  const char *RFC822   = "%a, %d %b %Y %H:%M:%S %Z +#9ld nsec";
  const char *ISO8601  = "%Y-%m-%d %H:%M:%S.#9.9ld%Z";
  const char *ISO8601Z = "%Y-%m-%d %H:%M:%S.#9.9ldZ";
#endif
  const char *SQL = "%Y-%m-%d %H:%M:%S";

  bool asLocal = opt.Contains("l");
  bool asSQL   = opt.Contains("s");
  
  if ( asSQL ) asLocal = false;

  const char *format = RFC822;
  if ( opt.Contains("c") ) {
    format = ISO8601;
    if ( !asLocal ) format = ISO8601Z;
  }      
  if ( asSQL ) format = SQL;

  struct tm *ptm;
  time_t seconds = fSec;   // deal with possible mismatch of types 
                           // of fSec and the time_t required by functions

  // get the components into a tm struct
  ptm = (asLocal) ? localtime(&seconds) : gmtime(&seconds);

  // format all but the nsec field
  // size_t length = 
  strftime( formatted, sizeof(formatted), format, ptm );

  if ( asSQL ) return formatted;

  // hack in the nsec part
  char* ptr = strrchr( formatted, '#' ); 
  if ( ptr ) *ptr = '%';    // substitute % for #
  sprintf( formatted2, formatted, fNanoSec );

  return formatted2;
}

// -------------------------------------------------------------------------

void TTimeStamp::Copy(FdRoot::TTimeStamp &vldts)
 {
  vldts.fSec     = fSec;
  vldts.fNanoSec = fNanoSec;
}

// -------------------------------------------------------------------------

int TTimeStamp::GetDate(const bool inUTC, const int secOffset,
                          int *year, int *month, int *day) const
 {
  time_t atime = fSec + secOffset;
  struct tm *ptm = (inUTC) ? gmtime(&atime) : localtime(&atime);

  if ( year )  *year  = ptm->tm_year + 1900;
  if ( month ) *month = ptm->tm_mon + 1;
  if ( day )   *day   = ptm->tm_mday;

  return (1900+ptm->tm_year)*10000 + (1+ptm->tm_mon)*100 + ptm->tm_mday;
}

// -------------------------------------------------------------------------

int TTimeStamp::GetTime(const bool inUTC, const int secOffset,
                            int *hour, int *min, int *sec) const
 {
  time_t atime = fSec + secOffset;
  struct tm *ptm = (inUTC) ? gmtime(&atime) : localtime(&atime);

  if ( hour ) *hour = ptm->tm_hour;
  if ( min )  *min  = ptm->tm_min;
  if ( sec )  *sec  = ptm->tm_sec;

  return ptm->tm_hour*10000 + ptm->tm_min*100 + ptm->tm_sec;
}

// -------------------------------------------------------------------------

int TTimeStamp::GetZoneOffset() const
 {
#ifdef __DARWIN__
  return 0;  // no implementation provided
#else // __DARWIN__
  // ?? should tzset (_tzset) be called?
# ifndef WIN32
  tzset();
  return  timezone;   /* unix has extern long int */
# else
  _tzset();
  return _timezone;   /* Win32 prepends "_" */
# endif // WIN32
#endif // __DARWIN__
}

// -------------------------------------------------------------------------

void TTimeStamp::Print(ostream& str, Option_t *option) const
 {
  str << "Date/Time = " << AsString(option) << endl;
}

// -------------------------------------------------------------------------

void TTimeStamp::Set()
 {
  time_t now;
  time( &now );
  fSec     = now;

  static int fake_ns = 0;
  fNanoSec = fake_ns++;
}

// -------------------------------------------------------------------------

void TTimeStamp::Set(int year, int month, int day,
                     int hour, int min, int sec, 
                     int nsec, bool isUTC, int secOffset)
 {
  // deal with special formats of year
  if ( year <= 37 )                year += 2000;
  if ( year >= 70 && year <= 137 ) year += 1900;
  
  // tm.tm_year is years since 1900
  if ( year >= 1900 )              year -= 1900;

  struct tm tmstruct;
  tmstruct.tm_year  = year;    // years since 1900
  tmstruct.tm_mon   = month-1; // months since Jan [0,11]
  tmstruct.tm_mday  = day;     // day of the month [1,31]
  tmstruct.tm_hour  = hour;    // hours since midnight [0,23]
  tmstruct.tm_min   = min;     // minutes after the hour [0,59]
  tmstruct.tm_sec   = sec + secOffset;     // seconds after the minute [0,59]
  tmstruct.tm_isdst = -1;     // let "mktime" determine DST setting

  const time_t bad_time_t = (time_t) -1;
  // convert tm struct to time_t
  // if values are given in UTC then no standard routine exists
  // and we'll have to use our homegrown routine
  // if values are given in local time then use "mktime"
  // which also normalizes the tm struct as a byproduct
  time_t utc_sec = (isUTC) ? FromUTC(&tmstruct) : mktime(&tmstruct);
 
  //   TTimeStamp::Dump_tm_struct(tmstruct);

  if ( utc_sec == bad_time_t ) {
    FD_CERR << "TTimeStamp::Set() mktime returned -1" << endl;
  }

  fSec  = utc_sec;
  fNanoSec = nsec;

  NormalizeNanoSec();
}

// -------------------------------------------------------------------------

void TTimeStamp::Set(int date, int time, int nsec,
                     bool isUTC, int secOffset)
 {
  int year  = date/10000;
  int month = (date-year*10000)/100;
  int day   = date%100;

  // protect against odd attempts at time offsets
  const int oneday = 240000;
  while ( time < 0 ) {
    time += oneday;
    day  -= 1;
  }
  while ( time > oneday ) {
    time -= oneday;
    day  += 1;
  }
  
  int hour  = time/10000;
  int min   = (time-hour*10000)/100;
  int sec   = time%100;

  Set( year, month, day, hour, min, sec, nsec, isUTC, secOffset );
}

// -------------------------------------------------------------------------

void TTimeStamp::NormalizeNanoSec()
 {
  // deal with negative values
  while ( fNanoSec < 0 ) {
    fNanoSec += nsPerSec;
    fSec -= 1;
  }
  // deal with values inf fNanoSec greater than one sec
  while ( fNanoSec >= nsPerSec ) {
    fNanoSec -= nsPerSec;
    fSec += 1;
  }
}

// -------------------------------------------------------------------------

timespec_t TTimeStamp::FromGPS(unsigned int sec,unsigned int nano)
 {
  timespec_t t;
  
  t.tv_sec  = FdUtil::TimeConvert::GetUnixTimeFromGPSTime( sec );
  t.tv_nsec = nano;
  
  // cout << "timespec_t= " << t.tv_sec << "," << t.tv_nsec << endl;
  
  return t;
}    

// -------------------------------------------------------------------------

time_t TTimeStamp::FromUTC(tm_t* tmstruct) 
 {
  const int days[]     = { 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 };
  const int daysLeap[] = { 31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 };

  int year = tmstruct->tm_year + 1900;
  bool isleap = FdRoot::TTimeStamp::IsLeapYear(year);

  const int *daysInMonth = days;
  if ( isleap ) daysInMonth = daysLeap;

  // fill in tmstruct->tm_yday

  int &ref_tm_mon = tmstruct->tm_mon;
  int &ref_tm_mday = tmstruct->tm_mday;
  
  // count days in months past
  tmstruct->tm_yday = 0;
  for ( int imonth = 0; imonth < ref_tm_mon; imonth++ ) {
    tmstruct->tm_yday += daysInMonth[imonth];
  }
  tmstruct->tm_yday += ref_tm_mday - 1;  // day [1-31] but yday [0-365]

  // adjust if day in this month is more than the month has
  while ( ref_tm_mday > daysInMonth[ref_tm_mon] ) {
    ref_tm_mday -= daysInMonth[ref_tm_mon];
    ref_tm_mon++;
  }

  // *should* calculate tm_wday (0-6) here ...

  // UTC is never DST
  tmstruct->tm_isdst = 0;

  // Calculate seconds since the Epoch based on formula in
  // POSIX  IEEEE Std 1003.1b-1993 pg 22

  int utc_sec = tmstruct->tm_sec +
                tmstruct->tm_min*60 +
                tmstruct->tm_hour*3600 +
                tmstruct->tm_yday*86400 +
                (tmstruct->tm_year-70)*31536000 +
                ((tmstruct->tm_year-69)/4)*86400;

  return utc_sec;
}

// -------------------------------------------------------------------------

bool TTimeStamp::IsLeapYear(int year)
 {
  if ( year%4 != 0 ) {
    return false;
  }
  else {
    if (year%400 == 0) {
      return true;
    }
    else { 
      if (year%100 == 0) {
        return false;
      }
      else {
        return true;
      }
    }
  }
}

// -------------------------------------------------------------------------

void TTimeStamp::Dump(const tm_t& tmstruct /* ,ostream& str */) 
 {
   // tmstruct.tm_year = year;    // years since 1900
   // tmstruct.tm_mon  = month-1; // months since Jan [0,11]
   // tmstruct.tm_mday = day;     // day of the month [1,31]
   // tmstruct.tm_hour = hour;    // hours since midnight [0,23]
   // tmstruct.tm_min  = min;     // minutes after the hour [0,59]
   // tmstruct.tm_sec  = sec;     // seconds after the minute [0,59]
   // tmstruct.tm_wday            // day of week [0,6]
   // tmstruct.tm_yday            // days in year [0,365]
   // tmstruct.tm_isdst           // DST [-1/0/1]  (unknown,false,true)

  printf(" tm { year %4d, mon   %2d, day   %2d,\n",
          tmstruct.tm_year,
          tmstruct.tm_mon,
          tmstruct.tm_mday);
  printf("      hour   %2d, min   %2d, sec   %2d,\n",
          tmstruct.tm_hour,
          tmstruct.tm_min,
          tmstruct.tm_sec);
  printf("      wday   %2d, yday %3d, isdst %2d",
          tmstruct.tm_wday,
          tmstruct.tm_yday,
          tmstruct.tm_isdst);

#ifdef linux
//#ifdef __GNUC__
// special GCC extras
  printf(",\n      tm_gmtoff %7ld,  tm_zone \"%s\"",
#ifdef __USE_BSD
          tmstruct.tm_gmtoff,tmstruct.tm_zone);
#else
          tmstruct.__tm_gmtoff,tmstruct.__tm_zone);
#endif
#endif
  printf("}\n");
}

/** ROOT class implementation macro. */
ClassImp(FdRoot::TTimeStamp)

// -------------------------------------------------------------------------
// -------------------------------------------------------------------------
