
//
// File   : TTimeStamp.hh
//
// Purpose: Declaration of the class TTimeStamp
//
// Author : R. Hatcher, FNAL
//
// $Id: TTimeStamp.hh 8758 2010-09-27 12:16:24Z mathes $
//


/** @file TTimeStamp.hh
  * Declaration of the class TTimeStamp.
  * @author R. Hatcher, FNAL
  */
  
#ifndef _FdRoot_TTimeStamp_hh_
#define _FdRoot_TTimeStamp_hh_


#include <ctime>
#include <iostream>

#include <TObject.h>
#include <TimeConvert.hh>

#ifdef __CINT__
/** Explicit definition of struct timespec because "rootcint" won't look in 
  * appropriate <time.h>.
  *
  * <b>time_t</b> appears to be defined as "typedef long time_t;" in the 
  * CINT version  of <time.h>.  This isn't required by the standard:<br>
  * To be compatible w/ std functions it must be at least 32-bits long,
  * but it might be longer to avoid the year-2037 cutoff.
  *
  * Note: The types 'long int' and 'time_t' (as used here by ROOT::TTimeStamp)
  *       are not portable. To keep the 32 bits we have to ... ?
  */
struct timespec 
 {
  long int   tv_sec;   /* seconds */
  long int   tv_nsec;  /* nanoseconds */
};

/** Redefinition of struct tm for CINT.
  */
struct tm
{
  int tm_sec;                   /* Seconds.     [0-60] (1 leap second) */
  int tm_min;                   /* Minutes.     [0-59] */
  int tm_hour;                  /* Hours.       [0-23] */
  int tm_mday;                  /* Day.         [1-31] */
  int tm_mon;                   /* Month.       [0-11] */
  int tm_year;                  /* Year - 1900.  */
  int tm_wday;                  /* Day of week. [0-6] */
  int tm_yday;                  /* Days in year.[0-365] */
  int tm_isdst;                 /* DST.         [-1/0/1]*/
};
#endif  // __CINT__

// define this typedef so that CINT doesn't barf at using these
// for function return values
/** POSIX.1b structure for a time value.  
    This is like a `struct timeval' but
    has nanoseconds instead of microseconds.
  */
typedef struct timespec timespec_t;
/** Struct used by other time functions. */
typedef struct tm       tm_t;

// --- forward declaration(s)

#if defined(R__ANSISTREAM)
#include <iosfwd>
using namespace std;
#else
class ostream;
#endif

namespace FdRoot {
  class TTimeStamp;
}

/** This makes a TTimeStamp object streamable to cout and other ostreams.
  */
ostream& operator<<(ostream& os, const FdRoot::TTimeStamp& vldts);

namespace FdRoot {

/** The TTimeStamp encapsulates the unit datum out of the GPS system.
  *
  * This extends (and isolates) struct timespec:
  * \code
  *   struct timespec {
  *     long int     tv_sec;   // seconds
  *     long int     tv_nsec;  // nanoseconds
  *   }
  * \endcode
  *   time_t seconds is relative to Jan 1, 1970 00:00:00 UTC
  *
  * Due to ROOT/CINT limitations TTimeStamp does not explicitly
  * hold a timespec struct; attempting to do so means the Streamer
  * must be hand written.  Instead we have chosen to simply contain
  * similar fields within the private area of this class.
  *
  * <b>NOTE:</b> the use of time_t (and its default implementation as a 32 int)
  *    implies overflow conditions occurs somewhere around
  *    Jan 18, 19:14:07, 2038.<br>
  *    If this experiment is still going when it becomes significant
  *    someone will have to deal with it.
  *
  * The original implementation of Robert Hatcher can be found at:
  * http://www-numi.fnal.gov:8875/fnal_minos/computing/srt_public_context/Validity/VldTimeStamp.h
  * http://www-numi.fnal.gov:8875/fnal_minos/computing/srt_public_context/Validity/VldTimeStamp.cxx
  *
  * Changes:  
  *
  * @li 2002/01/23 (thjm) -- adopted for AUGER, coding style, documentation
  * @li 2000/12/20 (R. Hatcher) -- convert from TDatime to struct timespec
  * @li 2000/04/19 (R. Hatcher) -- initial version
  */
class TTimeStamp {

    friend bool operator==(const FdRoot::TTimeStamp &lhs, 
			   const FdRoot::TTimeStamp &rhs);
    friend bool operator!=(const FdRoot::TTimeStamp &lhs,
			   const FdRoot::TTimeStamp &rhs);
    friend bool operator< (const FdRoot::TTimeStamp &lhs,
			   const FdRoot::TTimeStamp &rhs);
    friend bool operator<=(const FdRoot::TTimeStamp &lhs, 
			   const FdRoot::TTimeStamp &rhs);
    friend bool operator> (const FdRoot::TTimeStamp &lhs, 
			   const FdRoot::TTimeStamp &rhs);
    friend bool operator>=(const FdRoot::TTimeStamp &lhs, 
			   const FdRoot::TTimeStamp &rhs);
    
public:

    /** Constructor for the class TTimeStamp which build "now" 
      * with nsec field = 0, i.e. create a TTimeStamp and set it to the 
      * current time (as best as possible).
      */
    TTimeStamp();

    /** Copy constructor of the class TTimeStamp. */
    TTimeStamp(const FdRoot::TTimeStamp &source)
     { fSec = source.fSec; fNanoSec = source.fNanoSec; }
    
    /** Constructor for the class TTimeStamp which constructs from the
      * timespec struct
      */
    TTimeStamp(const timespec_t &ts)
     { fSec = ts.tv_sec; fNanoSec = ts.tv_nsec; NormalizeNanoSec(); }

    /** Constructor for the class TTimeStamp which constructs from the
      * timespec struct
      */
    TTimeStamp(const timespec_t*);

    /** Constructor for the class TTimeStamp which constructs from
      * time_t and separate nsec.
      */
    TTimeStamp(const time_t &t,const int nsec)
     { fSec = t; fNanoSec = nsec; NormalizeNanoSec(); }

    /** Constructor for the class TTimeStamp which constructs from
      * time_t and separate nsec.
      */
    TTimeStamp(const time_t *t,const int nsec);

    /** Constructor for the class TTimeStamp which constructs
      * from bits and pieces.
      *
      * Create a TTimeStamp and set it to the specified year, month,
      * day, time, hour, minute, second and nanosec.
      * If !isUTC then it is assumed to be the standard local time zone
      *
      * If local time is PST then one can use
      * \code
      *  TTimeStamp(year,month,day,hour,min,sec,nsec,kFALSE,0);
      * \endcode
      * or
      * \code
      *  tnt secOffset = 8*60*60;
      *  TTimeStamp(year,month,day,hour,min,sec,nsec,kTRUE,8*60*60);
      * \endcode
      */
    TTimeStamp(const unsigned int year, const unsigned int month,
               const unsigned int day,  const unsigned int hour,
               const unsigned int min,  const unsigned int sec,
               const unsigned int nsec=0, 
               const bool isUTC=true, const int secOffset=0);

    /** Constructor for the class TTimeStamp for compatiblity with TDatime.
      *
      * Create a TTimeStamp and set it to the specified date, time, nanosec.
      * If !isUTC then it is assumed to be the standard local time zone.
      */
    TTimeStamp(const unsigned int date, const unsigned int time, 
               const unsigned int nsec, 
               const bool isUTC=kTRUE, const int secOffset=0);

    /** Destructor of the class TTimeStamp. */
    virtual ~TTimeStamp() { }

    /** operator=() (assigment) for the class TTimeStamp. */
    FdRoot::TTimeStamp& operator=(const FdRoot::TTimeStamp&);

    /** operator+=() for the class TTimeStamp. */
    FdRoot::TTimeStamp& operator+=(const FdRoot::TTimeStamp &rhs)
     {
      fNanoSec += rhs.fNanoSec; NormalizeNanoSec();
      fSec += rhs.fSec;
      
      return *this;
    }
    
    /** operator+() for the class TTimeStamp. */
    FdRoot::TTimeStamp operator+(const FdRoot::TTimeStamp &rhs)
     {
      FdRoot::TTimeStamp result = *this;
      
      result += rhs;
      
      return result;
    }
    
    /** operator-=() for the class TTimeStamp. */
    FdRoot::TTimeStamp& operator-=(const FdRoot::TTimeStamp &rhs)
     {
      if ( fNanoSec < rhs.fNanoSec ) {
        fNanoSec += 1000000000 - rhs.fNanoSec;
	fSec--;
      }
      else
        fNanoSec -= rhs.fNanoSec;
      
      fSec -= rhs.fSec;
      
      return *this;
    }
    
    /** operator-() for the class TTimeStamp. */
    FdRoot::TTimeStamp operator-(const FdRoot::TTimeStamp &rhs)
     {
      FdRoot::TTimeStamp result = *this;
      
      result -= rhs;
      
      return result;
    }
    
   // --- various getters ...

    /** Get the time stored in the TTimeStamp object in struct timespec
      * representation (system time scale).
      */
    timespec_t GetTimeSpec() const
      { timespec_t value = { fSec, fNanoSec }; return value; }

    /** Get the seconds part of the time stored in the TTimeStamp object.
      *
      * The returned time is in GPS representation.
      */
    time_t GetGPSSec() const
     { return FdUtil::TimeConvert::GetGPSTimeFromUnixTime( fSec );}

    /** Get the seconds part of the time stored in the TTimeStamp object.
      *
      * The returned time is in GPS representation.
      */
    time_t GetSec() const
     { return GetGPSSec();}
    
    /** Get the seconds part of the time stored in the TTimeStamp object.
      *
      * The returned time is in Unix representation.
      */
    time_t GetUnixSec() const
     { return fSec;}

#if 0
    /** Get the seconds part of the time stored in the TTimeStamp object.
      *
      * The returned time is in UTC representation.
      */
    time_t GetUTCSec() const
     { return FdUtil::TimeConvert::GetUTCTimeFromUnixTime( fSec );}
#endif

    /** Get the nano seconds part of the time stored in the TTimeStamp object.
      */
    int GetNanoSec() const
     { return fNanoSec; }

    /** Return the date & time as a string.
      *
      * Result is pointer to a statically allocated string.
      * User should copy this into their own buffer before calling this method 
      * again (i.e. method is non-reentrant).
      *
      * Possible options:
      * @li option "l" returns it in local zone format
      * (can be applied to default or compact format).
      *
      * The default format is RFC822 compliant:
      * \code
      *  "Mon, 02 Jan 2001 18:11:12 +0000 (GMT) +999999999 nsec"
      *  "Mon, 02 Jan 2001 10:11:12 -0800 (PST) +999999999 nsec"
      * \endcode
      *
      * @li option "c" compact is (almost) ISO 8601 compliant:
      * \code
      *  "2001-01-02 18:11:12.9999999999Z"
      *  "2001-01-02 10:11:12.9999999999-0800"  if PST
      * \endcode
      *      * uses "-" as date separator as specified in ISO 8601
      *      * uses "." rather than preferred "," for decimal separator
      *      * -HHMM is the difference between local and UTC (if behind, + if ahead).
      *    The "-HHMM" is replaced with "Z" if given as UTC.
      *    To be strictly conforming it should use "T" instead of the 
      *    blank separating the date and time.
      *
      * @li option "2" returns as {sec,nsec} integers
      *
      * @li option "s" returns "2001-01-02 18:11:12" with an implied UTC
      *     overrides "l" option.
      */
    const char *AsString(const Option_t *option="") const;

    /** Copy *this to the passed TTimeStamp object reference. */
    void Copy(FdRoot::TTimeStamp &vldts);

    /** Return date in form of 19971224 (i.e. 24/12/1997).
      *
      * If non-zero pointers are supplied for year,month,day,
      * fill those as well.
      */
    int  GetDate(const bool inUTC=true, 
                 const int secOffset=0,
                 int* year=0, int* month=0, int* day=0) const;
    /** Return time in form of 123623 (i.e. 12:36:23).
      *
      * If non-zero pointers supplied for hour,min,sec, fill those as well.
      */
    int GetTime(const bool inUTC=true,
                const int secOffset=0,
                int* hour=0, int* min=0, int* sec=0) const;

    /** Return local (current) time zone offset from UTC.
      *
      * This is the difference in seconds between UTC and local standard time.
      */
    int GetZoneOffset() const;

    /** Print date and time to the specified std::ostream. */
    void Print(ostream& str=cout, const Option_t *option="") const;

    // --- Utility functions
    
    /** Convert GPS time scale into Unix time timespec_t type. */
    static timespec_t FromGPS(unsigned int,unsigned int);

    /** Equivalent of standard routine "mktime" but using the assumption 
      * that tm struct is filled with UTC, not local time.
      *
      * @li this version *ISN'T* configured to handle every possible
      * weirdness of out-of-range values in the case of normalizing
      * the tm struct.
      *
      * @li this version *DOESN'T* correctly handle values that can't be
      * fit into a time_t (ie. beyond year 2038-01-18 19:14:07, or
      * before the start of Epoch.
      */
    static time_t FromUTC(tm_t* tmstruct);
    
    /** Return kTRUE, if the given year is a leap year.
      *
      * The calendar year is 365 days long, unless the year is exactly 
      * divisible by 4, in which case an extra day is added to February to 
      * make the year 366 days long. If the year is the last year of a century, 
      * eg. 1700, 1800, 1900, 2000, then it is only a leap year if it is 
      * exactly divisible by 400. <br>
      * Therefore, 1900 wasn't a leap year but 2000 will be. The reason for
      * these rules is to bring the average length of the calendar year into 
      * line with the length of the Earth's orbit around the Sun, so that the
      * seasons always occur during the same months each year.
      */
    static bool IsLeapYear(int year);

    /** Print out the "tm" structure. */
    static void Dump(const tm_t& tmstruct /*,ostream& str=cout */);

protected:

    /** Set Date/Time to current time as reported by the system.
      *
      * No accounting for nanoseconds with std ANSI functions, so ns part is
      * faked so that subsequent calls simply add 1 to it. This ensures that 
      * calls within the same second come back distinct (and sortable).
      */
    void Set();

    /** Set Date/Time from components.
      *
      * month & day both use normal 1..12 and 1..31 counting.
      * hours, min, sec run from 0 to 23, 59, 59 respectively.
      * secOffset provides method for adjusting for alternative timezones.
      *
      * \code
      * "year"  |    0    1 ... 37 | 38...69   |   70 .. 100  101 ..  137 
      * true    | 2000 2001   2037 | undefined | 1970   2000 2001 .. 2037
      *
      * "year"  | 138...1969 | 1970 .. 2037 | ...
      * true    | undefined  | 1970 .. 2037 | undefined
      * \endcode
      */
    void Set(int year, int month, int day, int hour, 
             int min, int sec, int nsec, bool isUTC, int secOffset);

    /** Set date/time from integers of the form [yy]YYMMDD and HHMMSS,
      * assume UTC (UTC) components.
      *
      * \code
      *  MM: 01=January .. 12=December
      *  DD: 01 .. 31
      *  
      *  HH: 00=midnight .. 23
      *  MM: 00 .. 59
      *  SS: 00 .. 69
      *  \endcode
      * 
      * Constraints:
      * @li date must be in format 980418 or 19980418 or
      *                       1001127 or 20001127  (ie. year 100 = 2000)
      * @li time must be in format 224512 (second precision)
      * @li date must be >= 700101
      */
    void Set(int date, int time, int nsec,
             bool isUTC, int secOffset);

    /** Ensure that the NanoSec field is in range [0,99999999]. */
    void NormalizeNanoSec();

    // --- Data members:
    // similar fields to struct timespec
    // use ROOT versions to "know" that they are platform consistent
    // 32-bit integers to avoid IO confusion.
    
    /** The internal representation of the seconds (in UTC). */
    Int_t  fSec;

    /** The internal representation of the nano seconds [0..999999999]. */
    Int_t  fNanoSec;

    ClassDef(TTimeStamp,1)
};

/** operator==() for TTimeStamp objects.
  */
inline bool operator==(const FdRoot::TTimeStamp &lhs, 
		       const FdRoot::TTimeStamp &rhs)
   { return lhs.fSec  == rhs.fSec && 
            lhs.fNanoSec == rhs.fNanoSec; }

/** operator!=() for TTimeStamp objects.
  */
inline bool operator!=(const FdRoot::TTimeStamp &lhs, 
		       const FdRoot::TTimeStamp &rhs)
   { return lhs.fSec  != rhs.fSec ||
            lhs.fNanoSec != rhs.fNanoSec; }

/** operator<() for TTimeStamp objects.
  */
inline bool operator<(const FdRoot::TTimeStamp &lhs, 
		      const FdRoot::TTimeStamp &rhs)
   { return lhs.fSec  < rhs.fSec ||
             ( lhs.fSec  == rhs.fSec &&
               lhs.fNanoSec <  rhs.fNanoSec   ); }

/** operator<=() for TTimeStamp objects.
  */
inline bool operator<=(const FdRoot::TTimeStamp &lhs, 
		       const FdRoot::TTimeStamp &rhs)
   { return lhs.fSec  < rhs.fSec ||
             ( lhs.fSec  == rhs.fSec &&
               lhs.fNanoSec <= rhs.fNanoSec   ); }

/** operator>() for TTimeStamp objects.
  */
inline bool operator>(const FdRoot::TTimeStamp &lhs, 
		      const FdRoot::TTimeStamp &rhs)
   { return lhs.fSec  > rhs.fSec ||
             ( lhs.fSec  == rhs.fSec &&
               lhs.fNanoSec >  rhs.fNanoSec   ); }

/** operator>=() for TTimeStamp objects.
  */
inline bool operator>=(const FdRoot::TTimeStamp &lhs, 
		       const FdRoot::TTimeStamp &rhs)
   { return lhs.fSec  > rhs.fSec ||
             ( lhs.fSec  == rhs.fSec &&
               lhs.fNanoSec >= rhs.fNanoSec   ); }

}  // namespace FdRoot

#endif // _FdRoot_TTimeStamp_hh_
