
//
// File   : TimeConvert.cc
//
// Purpose: Implementation of time conversion functions.
//
// $Id: TimeConvert.cc 8758 2010-09-27 12:16:24Z mathes $
//


#include <iostream>
#include <iomanip>
#include <string>
#include <sstream>
#include <ctime>
#include <list>


/** @file TimeConvert.cc
  * Implementation of the time conversion functions.
  * @author H.-J. Mathes, FzK
  */

#include "CheckVersion.hh"

#include "ErrorLogging.hh"
#include "InifileChain.hh"

#include "TimeConvert.hh"

using namespace std;
using FdUtil::TimeConvert;
using FdUtil::Inifile;

// some constants: seconds per day
//
const int kSecPerDay = 24*3600;

//    
// Initial offset between GPS and UTC times.
// (1 January 1970 to 6 January 1980 = 10 years, 5 days, 2 leap-days)
//
const unsigned long int kUTCGPSOffset0 = (10*365 + 7)*kSecPerDay;

//
// Table contains pairs of (GPS second when leap occurred, correction to apply)
//
// from PAO Offline, class TimeStamp (T.Paul, D.Veberic)
//
const unsigned int kNumLeaps = 15;
const unsigned int kLeapSeconds[kNumLeaps][2] = {
  //
  // (GPS epoch + years + leap days + Jan-Jul)*kSecPerDay + leapSeconds
  //                                           |
  { (361 +       0 +   0 + 181)*kSecPerDay +  0, +1  },    // 1 JUL 1981
  { (361 +     365 +   0 + 181)*kSecPerDay +  1, +2  },    // 1 JUL 1982
  { (361 +   2*365 +   0 + 181)*kSecPerDay +  2, +3  },    // 1 JUL 1983
  { (361 +   4*365 +   1 + 181)*kSecPerDay +  3, +4  },    // 1 JUL 1985
  { (361 +   7*365 +   1      )*kSecPerDay +  4, +5  },    // 1 JAN 1988
  { (361 +   9*365 +   2      )*kSecPerDay +  5, +6  },    // 1 JAN 1990
  { (361 +  10*365 +   2      )*kSecPerDay +  6, +7  },    // 1 JAN 1991
  { (361 +  11*365 +   3 + 181)*kSecPerDay +  7, +8  },    // 1 JUL 1992
  { (361 +  12*365 +   3 + 181)*kSecPerDay +  8, +9  },    // 1 JUL 1993
  { (361 +  13*365 +   3 + 181)*kSecPerDay +  9, +10 },    // 1 JUL 1994
  { (361 +  15*365 +   3      )*kSecPerDay + 10, +11 },    // 1 JAN 1996
  { (361 +  16*365 +   4 + 181)*kSecPerDay + 11, +12 },    // 1 JUL 1997
  { (361 +  18*365 +   4      )*kSecPerDay + 12, +13 },    // 1 JAN 1999
  // DV: 2000 IS a leap year since it is divisible by 400,
  //     ie. leap years here are 2000 and 2004 -> leap days = 6
  { (361 +  25*365 +   6      )*kSecPerDay + 13, +14 },    // 1 JAN 2006
  { (361 +  28*365 +   7      )*kSecPerDay + 14, +15 }     // 1 JAN 2009

};

//
// 'dynamic' table of GPS seconds of leap boundaries & associated leap seconds
//
vector<TimeConvert::LeapSecondEntryRec> TimeConvert::fLeapSecondTable;

// -------------------------------------------------------------------------

unsigned int TimeConvert::GetGPSTimeFromUnixTime(unsigned int unix_time)
 {   // unix_time - 315964800 + 13
  unsigned int gps_time = unix_time - kUTCGPSOffset0;
  
  gps_time += GetLeapSeconds( gps_time );
  
  if ( IsLeapBoundary( gps_time + 1 ) ) gps_time++;
  
  return gps_time;
}

// -------------------------------------------------------------------------

// ISO 8601 date format
std::string TimeConvert::GetISODate(const struct tm& tm)
 {
  char date_str[20];
  
  // ISO C++
  strftime( date_str, sizeof(date_str) - 1, "%Y-%m-%d", &tm );
  // C99 standard
  //strftime( date_str, sizeof(date_str) - 1, "%F", &tm );
  
  return date_str;
}

// -------------------------------------------------------------------------

std::string TimeConvert::GetISOTime(const struct tm& tm,bool seconds)
 {
  char time_str[20];
  
//  sprintf( time_str, "%02d-%02d-%02d", 
//           tm.tm_hour, tm.tm_min, tm.tm_sec );
  if ( seconds )
    strftime( time_str, sizeof( time_str ) - 1, "%H-%M-%S", &tm );
  else
    strftime( time_str, sizeof( time_str ) - 1, "%H-%M", &tm );
  
  return time_str;
}

// -------------------------------------------------------------------------

unsigned int TimeConvert::GetLeapSeconds(unsigned int gps_sec)
 {
  // we will access the leap second table -> make sure that it is initialized
  InitLeapSecondTable();

  for (int i=fLeapSecondTable.size()-1; i>=0; --i )
    if ( gps_sec >= fLeapSecondTable[i].fGPSTime )
      return fLeapSecondTable[i].fNLeapSeconds;
  
  return 0;
}

// -------------------------------------------------------------------------

std::string TimeConvert::GetTimeStringFromGPSTime(unsigned int gps_time,
                                                  const std::string format)
 {
  time_t unix_time_t = GetUnixTimeFromGPSTime( gps_time );
  
  // broken-time representation in UTC
  struct tm *tm = gmtime( &unix_time_t );
  
  if ( IsLeapBoundary( gps_time ) )
    tm->tm_sec = 60;
  
  return GetFormattedFromTm( tm, format );
}

// -------------------------------------------------------------------------

std::string TimeConvert::GetTimeStringFromUnixTime(unsigned int unix_time,
                                                   const std::string format)
 {
  time_t t_unix_time = unix_time;
  
  // broken-time representation in UTC
  struct tm *tm = gmtime( &t_unix_time );
  
  if ( IsLeapBoundary( GetGPSTimeFromUnixTime( unix_time ) ) )
    tm->tm_sec = 60;
  
  return GetFormattedFromTm( tm, format );
}

// -------------------------------------------------------------------------

std::string TimeConvert::GetFormattedFromTm(const struct tm *tm_struct,
                                            const std::string format)
 {
#if 0
  // needed for PAO (+Offline) month naming conventions
  static const char *month_abbr[] = { "JAN", "FEB", "MAR", "APR", "MAY", "JUN",
                                      "JUL", "AUG", "SEP", "OCT", "NOV", "DEC",
                                      "#!#" };
#endif

  const int tmmon = tm_struct->tm_mon;
  int month;
  if ( 0 <= tmmon && tmmon <= 11 )
    month = tmmon;
  else {
    FD_CWARN << "TimeConvert::GetFormattedFromTm() - "
             << "invalid month number in time structure" << endl;
    month = 12;
  }
  
#if 1
  // my implementation
  char time_date_str[40];
  
  if ( format.empty() )
    strftime( time_date_str, sizeof(time_date_str) - 1,
              "%a %b %d %H:%M:%S %Y", tm_struct );
  else
    strftime( time_date_str, sizeof(time_date_str) - 1,
              format.c_str(), tm_struct );

  return time_date_str;
#else
  // AUGER offline implementation
  stringstream time_date_str;

  time_date_str << setw(2) << setfill('0') << tm_struct->tm_hour << ":"
                << setw(2) << setfill('0') << tm_struct->tm_min  << ":"
                << setw(2) << setfill('0') << tm_struct->tm_sec  << " "
                << setw(2) << setfill('0') << tm_struct->tm_mday << " "
		<<                            month_abbr[month]  << " "
		<<                            tm_struct->tm_year + 1900;
  
  return time_date_str.str();
#endif
}

// -------------------------------------------------------------------------

unsigned int TimeConvert::GetUnixTimeFromGPSTime(unsigned int gps_time)
 {
  return gps_time + kUTCGPSOffset0 - GetLeapSeconds( gps_time );
}

// -------------------------------------------------------------------------

bool TimeConvert::IsLeapBoundary(unsigned int sec)
 {
  // we will access the leap second table -> make sure that it is initialized
  InitLeapSecondTable();
  
  for ( int i = fLeapSecondTable.size()-1; i >= 0; --i ) {
    const unsigned int leapSec = fLeapSecondTable[i].fGPSTime;
    if ( sec > leapSec )
      return false;
    if ( sec == leapSec )
      return true;
  }
  
  return false;
}

// -------------------------------------------------------------------------

void TimeConvert::InitLeapSecondTable()
 {
  static bool is_initialized = false;
  
  if ( is_initialized ) return;
  
  LeapSecondEntryRec entry;

  for (unsigned int i=0; i<kNumLeaps; i++ ) {
    entry.fGPSTime      = kLeapSeconds[i][0];
    entry.fNLeapSeconds = kLeapSeconds[i][1];
    
    fLeapSecondTable.push_back( entry );
  }
  
  is_initialized = true;
}

// -------------------------------------------------------------------------

void TimeConvert::PrintLeapSecondTable(std::ostream& ostr)
 {
  // we will access the leap second table -> make sure that it is initialized
  InitLeapSecondTable();
  
  ostr << "=== TimeConvert::PrintLeapSecondTable() ===" << endl;
  
  //      "123456789012345678901234567890"
  ostr << "    UTC      GPS(UTC)   nLeap" << endl;
  
  for (unsigned int i=0; i<fLeapSecondTable.size(); i++ ) {
    ostr << setw(10) << GetUnixTimeFromGPSTime( fLeapSecondTable[i].fGPSTime )
         << setw(12) << fLeapSecondTable[i].fGPSTime
	 << setw(6) << fLeapSecondTable[i].fNLeapSeconds
	 << endl;

  }

  ostr << "===========================================" << endl;
}

// -------------------------------------------------------------------------

bool TimeConvert::LeapSecondEntryLess(const LeapSecondEntryRec& l,
                                      const LeapSecondEntryRec& r)
 {
  return (l.fGPSTime < r.fGPSTime);
}

// -------------------------------------------------------------------------

void TimeConvert::ReadLeapSecondTable(string filename)
 {
  if ( filename.empty() ) {
    FD_CFATAL << "TimeConvert::ReadLeapSecondTable() - "
              << "no file name was specified!!!" << endl;
    return;
  }

  InifileChain inifile( filename.c_str() );

  if ( inifile.GetStatus() != Inifile::kSUCCESS ) {
    FD_CERR << "TimeConvert::ReadLeapSecondTable() - "
            << "could not read specified ini-file '" 
	    << filename << "'!!!" << endl;
    return;
  }

  // I hoped for an implicit type conversion, it didn't work (probably good)
#if 0
  InifileChain ch_inifile( inifile );
  
  ReadLeapSecondTable( ch_inifile );
#else
  ReadLeapSecondTable( inifile );
#endif
}
  
// -------------------------------------------------------------------------

void TimeConvert::ReadLeapSecondTable(FdUtil::InifileChain& inifile)
 {
  // we will access the leap second table -> make sure that it is initialized
  InitLeapSecondTable();
  
  const char *section = "LeapSeconds";
  
  if ( inifile.SpecifyGroup( section ) != Inifile::kSUCCESS ) {
    FD_CWARN << "TimeConvert::ReadLeapSecondTable() - "
             << "Section [" << section << "] not found !!!" << endl;
    return;
  }
  
  unsigned int n = 0;
  int error = 0;
  unsigned int leap_seconds;
  const char *utc_date_str;
  time_t t_leap;
  list<LeapSecondEntryRec> leapsec_list;
  
  char tag[10];
  do {
    sprintf( tag, "leap%u", n );
    
    utc_date_str = inifile.GetFirstString( tag, "", &error );
    if ( error ) break;
    
    leap_seconds = inifile.GetNextValue( &error );
    if ( error ) break;
    
    struct tm leap_tm;
    
    leap_tm.tm_sec = 0;
    leap_tm.tm_min = 0;
    leap_tm.tm_hour = 0;
    
    if ( strptime( utc_date_str, "%Y-%m-%d", &leap_tm ) == NULL ) {
      FD_CERR << "TimeConvert::ReadLeapSecondTable() - "
              << "strptime() for entry '" << tag 
	      << "' returned NULL, skipped!!!" << endl;
      n++;
      continue;
    }
    
    //FD_COUT << asctime( &leap_tm );
    
    // see 'man timegm' for more details ... (mktime() converts in local time)
    t_leap = timegm( &leap_tm );
    
    // tm_mon - The number of months since January, in the range 0 to 11
    if (    ((leap_tm.tm_mon != 0) && (leap_tm.tm_mon != 6))
         || (leap_tm.tm_mday != 1) || (t_leap == (time_t)-1) ) {
      FD_CERR << "TimeConvert::ReadLeapSecondTable() - "
              << "entry '" << tag << "' contains illegal date, skipped!!!"
	      << endl;
      n++;
      continue;
    }
    
    //FD_COUT << utc_date_str << " -> " << t_leap << ", " << leap_seconds << endl;
    
    LeapSecondEntryRec entry;
    
    entry.fGPSTime = t_leap; // to be replaced later
    entry.fNLeapSeconds = leap_seconds;
    
    leapsec_list.push_back( entry );
    
    n++;
    
  } while ( !error );
  
  // --- process the list of GPSTime/NLeapSeconds entries
  
  // last entry of the table
  unsigned int last_utc_time = 
                        GetUnixTimeFromGPSTime( kLeapSeconds[kNumLeaps-1][0]);
  
  //FD_COUT << "last_utc_time: " << last_utc_time << endl;
  
  // sort the list 
  leapsec_list.sort( LeapSecondEntryLess );
  
  // check and remove already known entries
  
  list<LeapSecondEntryRec>::iterator iter = leapsec_list.begin();
  
  while ( iter != leapsec_list.end() ) {
    
    unsigned int utc_time = iter->fGPSTime - 1;
    unsigned int gps_time = GetGPSTimeFromUnixTime( utc_time );
    
    if ( utc_time > last_utc_time ) {
      iter++;
      continue;
    }
    
    //FD_COUT << "UTC: " << utc_time << " GPS: " << gps_time
    //	    << " IsLeapBoundary: " << IsLeapBoundary( gps_time ) << endl;
        
    if ( IsLeapBoundary( gps_time ) ) { // already in list
      iter = leapsec_list.erase( iter );
      continue;
    }
    
    FD_CERR << "TimeConvert::ReadLeapSecondTable() - " 
            << "an entry (UTC=" << utc_time 
            << ",GPS=" << gps_time << ") conflicts with present table!!!"
            << endl;
	    
    iter++;
  }
  
  if ( leapsec_list.empty() ) {
    FD_CWARN << "TimeConvert::ReadLeapSecondTable() - " 
             << "no usable entries found in leap second table!!!"
	     << endl;
    return;
  }
  
  FD_CINFO << "TimeConvert::ReadLeapSecondTable() - " 
           << "there " << ((leapsec_list.size() == 1) ? "is" : "are")
	   << " " << leapsec_list.size() << " "
	   << ((leapsec_list.size() == 1) ? "entry" : "entries")
	   << " to be added to the leap second table"
	   << endl;

  iter = leapsec_list.begin();
  
  while ( iter != leapsec_list.end() ) {
    
    // set the GPS time of the previous second where th eleap occurs
    iter->fGPSTime = GetGPSTimeFromUnixTime( iter->fGPSTime );
    
    fLeapSecondTable.push_back( *iter );
    
    iter++;
  }
}

// -------------------------------------------------------------------------
// -------------------------------------------------------------------------
