
//
// File   : EyeProcFramework.cc
//
// Purpose: Implementation of class EyeProcessingFramework.
//
// $Id: EyeProcFramework.cc 8757 2010-09-27 12:16:09Z mathes $
//


#include <iostream>
#include <iomanip>

#include <TList.h>
#include <TStopwatch.h>

/** @file EyeProcFramework.cc
  * Implementation of class EyeProcessingFramework.
  * @author H.-J. Mathes, FzK
  */
  
#include <EyeEvent.hh>
#include <EyeEventFile.hh>


#include "EyeDataProc.hh"
#include "EyeProcFramework.hh"

// -------------------------------------------------------------------------
// -------------------------------------------------------------------------
//
// class EyeProcessingFramework - implementation
//

bool EyeProcessingFramework::fgTerminate = false;

EyeProcessingFramework::EyeProcessingFramework()
 : fEvent(0), fHistoFile(0), fInputFile(0), fOutEvent(0), fOutputFile(0), 
   fOutputFileOption(kAuto), fProcessorList(new TList), fStatus(0)
 {
  fEvent = new TEyeEvent();
  fOutEvent = new TEyeEvent();
  
  fNEventsProcessed = 0;
}

// -------------------------------------------------------------------------

EyeProcessingFramework::~EyeProcessingFramework()
 {
  SetOutputFileOption( kAuto );  // close file in any case
  
  Close();
  
  delete fEvent;
  delete fOutEvent;
  
  // delet all processors ...
  
  TIterator* iter = fProcessorList->MakeIterator();
  EyeDataProcessor* proc;
    
  while ( ( proc = (EyeDataProcessor*)iter->Next() ) != NULL ) {
    delete proc;
  }
  
  // print final profiling statistics
  
  cout.setf( ios::fixed, ios::floatfield );
  cout << setprecision(2);

  cout << "Time spent for loading data:    Real= "
       << fLoadingTime.RealTime() << " Cpu= " 
       << fLoadingTime.CpuTime() << endl;
  cout << "Time spent for saving data:     Real= "
       << fSavingTime.RealTime() << " Cpu= "
       << fSavingTime.CpuTime() << endl;
  cout << "Time spent for processing data: Real= "
       << fProcessingTime.RealTime() << " Cpu= "
       << fProcessingTime.CpuTime() << endl;
  cout << "Processed " << fNEventsProcessed << " events, "
       << "time/event= " << setprecision(3)
       << fProcessingTime.CpuTime()/(Float_t)fNEventsProcessed
       << " seconds" << endl;
}

// -------------------------------------------------------------------------

void EyeProcessingFramework::AddProcessor(EyeDataProcessor* processor)
 {
  if ( processor )
    fProcessorList->Add( processor );
}

// -------------------------------------------------------------------------

void EyeProcessingFramework::CloseOutputFile()
 {
  fSavingTime.Start(kFALSE);
   fOutputFile->Close();
  fSavingTime.Stop();

  delete fOutputFile;
  fOutputFile = NULL;
}

// -------------------------------------------------------------------------

void EyeProcessingFramework::Close()
 {
  if ( fInputFile ) {
    TEyeRunHeader* header = NULL;
    
    fLoadingTime.Start(kFALSE);
     fInputFile->ReadRunHeader( &header );
    fLoadingTime.Stop();

    if ( fOutputFile ) {
      fSavingTime.Start(kFALSE);
       fOutputFile->WriteRunHeader( header );
      fSavingTime.Stop();
    }
    
    TEyeRunTrailer* trailer = NULL;
      
    fLoadingTime.Start(kFALSE);
     fInputFile->ReadRunTrailer( &trailer );
    fLoadingTime.Stop();

    if ( fOutputFile ) {
      fSavingTime.Start(kFALSE);
       fOutputFile->WriteRunTrailer( trailer );
      fSavingTime.Stop();
    }
    
    // forward the header and trailer data also to the processing modules
    
    TIterator* iter = fProcessorList->MakeIterator();
    EyeDataProcessor* proc;
    
    while ( ( proc = (EyeDataProcessor*)iter->Next() ) != NULL ) {

      fProcessingTime.Start(kFALSE);
      
       proc->ProcessRunHeader( header );
       proc->ProcessRunTrailer( trailer );
       
      fProcessingTime.Stop();
      
      // read error status of the last processor
      
      SetStatus( proc->GetStatus() );
      
      // break if error is signalled by one processor
      
      if ( GetStatus() ) break;
    }
   
    delete header;
    delete trailer;
  }
    
  if ( fOutputFile ) {
    if ( fOutputFileOption == kAuto ) {
      fSavingTime.Start(kFALSE);
       CloseOutputFile();
      fSavingTime.Stop();
    }
  }

  if ( fInputFile ) {
    fInputFile->Close();
    delete fInputFile;
    fInputFile = NULL;
  }
  
  if ( fHistoFile ) {
    fHistoFile->Close();
    delete fHistoFile;
    fHistoFile = NULL;
  }
}

// -------------------------------------------------------------------------

int EyeProcessingFramework::Run()
 {
  int error = 0;
  bool event_status = false;
  
  SetStatus( 0 );
  
  if ( !fgTerminate ) {
  
    fLoadingTime.Start(kFALSE);
     
     event_status = ( fEvent && 
                         fInputFile->ReadEvent( &fEvent ) != kFALSE );
    fLoadingTime.Stop();
  }
  
  if ( event_status && !error && !fgTerminate ) {
  
    bool do_save = true;
  
    TIterator* iter = fProcessorList->MakeIterator();
    EyeDataProcessor* proc;
    
    while ( ( proc = (EyeDataProcessor*)iter->Next() ) != NULL ) {

      fProcessingTime.Start(kFALSE);
      
       do_save = proc->ProcessEvent( fEvent );
       fNEventsProcessed++;
       
      fProcessingTime.Stop();
      
      // read error status of the last processor
      
      SetStatus( proc->GetStatus() );
      
      // break if error is signalled by one processor
      
      if ( GetStatus() ) break;
      
      // break processing if one processor signals that
      
      if ( !do_save ) break;
    }
    
    // all processing steps are finished, eventually save the data
    
    if ( do_save && fOutputFile ) {
      
      *fOutEvent = *fEvent;
      
      fSavingTime.Start(kFALSE);
       
       if ( fOutputFile->WriteEvent( &fOutEvent ) == kFALSE ) {
         cout << "Error when writing to output file !" << endl;
         error = 1;
       }
       
      fSavingTime.Stop();
    }
  
    error = GetStatus();
  }
  else if ( fgTerminate ) {  // termination flag set by user
    
    cout << "Program terminated by the user !" << endl;
    
    Close();
    error = 1;
  }
  else { // end of input file encountered
    
    cout << "End of input file !" << endl;
    
    Close();
    error = 2;
  }
  
  return error;
}

// -------------------------------------------------------------------------

void EyeProcessingFramework::SetInputFile(const char* name)
 {
  if ( fInputFile ) {
    fInputFile->Close();
    delete fInputFile;
  }
  
  fInputFile = new TEyeEventFile( name );
}

// -------------------------------------------------------------------------

void EyeProcessingFramework::SetOutputFile(const char* name)
 {
  if ( fOutputFile ) {
    fOutputFile->Close();
    delete fOutputFile;
  }
  
  fOutputFile = new TEyeEventFile( name, "recreate" );
}

// -------------------------------------------------------------------------
// -------------------------------------------------------------------------
