
//
// File   : EyeProcFramework.hh
//
// Purpose: Declaration of class EyeProcessingFramework.
//
// $Id: EyeProcFramework.hh 8757 2010-09-27 12:16:09Z mathes $
//



#ifndef _EyeProcFramework_hh_
#define _EyeProcFramework_hh_


#include <climits>

/** @file EyeProcFramework.hh
  * Declaration of class EyeProcessingFramework.
  * @author H.-J. Mathes, FzK
  */
 
#include <TList.h>
#include <TStopwatch.h>

// ----- forward declarations

class TEyeEventFile;
class EyeDataProcessor;

/** Framework for EyerrorEventFile based data processing.
  * To this framework processors of type '' could be added which are then 
  * applied sequentially on the data.
  *
  * Changes:
  * @li 2001/07/30 (thjm) added profiling capability.
  * @li 2001/08/14 (thjm) added some counting variables
  *
  * @todo
  *  @li add parameter 'first & last event for processing'
  */
class EyeProcessingFramework {
public:
    /** Options for the output file and the histogram file. 
      */
    typedef enum {
      /** File closed when input file is closed. */
      kAuto,
      /** File close is initiated by the user. */
      kUser
    } EFileOption;

public:
    /** Constructor of the class EyeProcessingFramework. Sufficient buffer
      * space for input and output event is allocated.
      */
    EyeProcessingFramework();
    
    /** Destructor of the class EyeProcessingFramework. All temporary created
      * objects are deleted. A final statistic about the time consumption of
      * the program is printed.
      */
    ~EyeProcessingFramework();
	
    /** */ 
    void AddProcessor(EyeDataProcessor*);

    /** Ask the program to close the output data file. This method is only 
      * needed when the option 'kUser' has been chosen.
      */
    void CloseOutputFile();

    /** Return application internal error status.
      */
    int GetStatus() { return fStatus; }
    
    /** Run the framework, i.e. open the input, output and histogram file
      * and begin to read events from the input file. For each event the 
      * method(s) of the added EyeDataProcessor objects are called.
      */
    int Run();

    /** Set the limits (first, last event to process) for the current run.
      * The specified numbers are event numbers and not the position of the
      * event in the file.
      */
    void SetEventLimits(UInt_t first, UInt_t last = UINT_MAX);
    
    /** Set the name of the optional histogramming file (not yet implemented !).
      */
    void SetHistoFile();
    
    /** Set the name for the input data file.
      */
    void SetInputFile(const char*);
    
    /** Set the name of the output data file.
      */
    void SetOutputFile(const char*);
    
    /** Set the options for the output file. If this option is set to
      * 'kAuto' then the output file is closed when the input files end is 
      * reached. With the option 'kUser' the user decides when to close the 
      * output data file.
      */
    void SetOutputFileOption(EFileOption e)
     { fOutputFileOption = e; }
     
    /** Set the termination flag which means that the the last event is
      * eventually saved and all files are closed rather than reading the 
      * next event from the input file.
      */
    static void SetTerminate()
     { fgTerminate = true; }

protected:
    
    /** */
    void Close();
    
    /** Set the internal error status of this application. */
    void SetStatus(int status) { fStatus = status; }

private:
    TEyeEvent           *fEvent;             // the EyerrorEvent currently worked on
    TFile               *fHistoFile;         // handle to the histogram file
    TEyeEventFile       *fInputFile;         // handle to the input file
    TEyeEvent           *fOutEvent;          // handle to the event to be stored
    TEyeEventFile       *fOutputFile;        // handle to the output file
    EFileOption          fOutputFileOption;  // option of output file handling
    TList               *fProcessorList;     // list of event processors
    int                  fStatus;            // applications internal error st.

    unsigned int         fNEventsProcessed;  // number of events processed
    unsigned int         fFirstEvent;        // first event for processing
    unsigned int         fLastEvent;         // last event for processing

    // for time profiling purposes
    TStopwatch           fLoadingTime;
    TStopwatch           fProcessingTime;
    TStopwatch           fSavingTime;
    
    static bool          fgTerminate;         // flag to terminate soon
};

#endif // _EyeProcFramework_hh_
