
//
// File   : T3TestApp.cc
//
// Purpose: Implementation of the class T3TestApplication
//
// $Id: T3TestApp.cc 11967 2011-01-27 16:12:07Z mathes $
//


#include <iostream>
#include <climits>
#include <unistd.h>   // getopt()
#include <stdint.h>   // uint32_t

/** @file T3TestApp.cc
  * Implementation of the class T3TestApplication.
  * @author H.-J. Mathes, FzK
  */


#include <TTimeStamp.hh>
#include <ErrorLogging.hh>

#include "CheckVersion.hh"

#include "EyeEventClassifier.hh"

#ifdef HOUGH_TEST
# include "HoughSel.hh"
#endif // HOUGH_TEST
#include "T3TestApp.hh"
#ifdef BLOB_TEST
# include "BlobSel.hh"
#endif // BLOB_TEST

using namespace std;
using FdRoot::TTimeStamp;

// -------------------------------------------------------------------------
// -------------------------------------------------------------------------
// 
// class T3TestApplication - implementation
// 

T3TestApplication::T3TestApplication(int argc,char** argv)
 : fArgc(argc), fArgv(argv), fStatus(0),
   fLogFilename("t3test.log"),
   fMiInputFile(NULL), fEyeInputFile(NULL), 
   fOutputFile(NULL), fRejOutputFile(NULL)
 {
  fNEventLimit    = UINT_MAX;  // nearly unlimited
  
  fReadMiEvent    = false;

  if ( fArgc <= 1 )
    Usage();

  Parse();
}

// -------------------------------------------------------------------------

T3TestApplication::~T3TestApplication()
 {
  delete fMiInputFile;
  delete fEyeInputFile;
  delete fOutputFile;
  delete fRejOutputFile;
}

// -------------------------------------------------------------------------

#ifdef BLOB_TEST
void T3TestApplication::DoBlobSel(TEyeEvent* /* event */, 
                                  EyeEventClassifier* /* evclass */, 
                                  BlobSel* /* blob */)
 {
}
#endif // BLOB_TEST

// -------------------------------------------------------------------------

#ifdef HOUGH_TEST
void T3TestApplication::DoHoughSel(TEyeEvent *event, 
                                   HoughSel *hough)
 {
  TEyeEventHeader *header = event->GetEventHeader();
  TEyePixelList *pixel_list = event->GetPixelList();
  
  TMirrorBitArray pixT1[6];
  unsigned int nT1[6] = { 0 };
   
  // --- determine total number of triggered pixels
  
  unsigned int n_trigpix = 0;
  
  for ( unsigned int i=1; i<=6; i++ ) {
  
    TMirrorPixelData *pixel_data = event->GetPixelData()->GetPixelData(i);
    
    if ( pixel_data != NULL ) {
      TMirrorBitArray ba(pixel_data);
      n_trigpix += ba.GetNumberOfBits();
    }
  }
  
  cout << "n_trigpix --> " << n_trigpix << endl;
  
  // number of pixels in PixelList
  unsigned int n_pixel = pixel_list->GetNumPixels();
    
//  cout << "PixelList->GetNumPixels() --> " << n_pixel << endl;

  for ( unsigned int i=0; i<n_pixel; i++) {
     
    Fd::FdPixelNumber pixx = pixel_list->GetPixel(i);
  
    unsigned int mir_t = Fd::GetEyeMirrorNo( pixx );
    unsigned int row_t = Fd::GetRowNo( Fd::GetPixelNumber( pixx ) );
    unsigned int col_t = Fd::GetColumnNo( Fd::GetPixelNumber( pixx ) );
    unsigned int T1 = 0;
    
    if ( row_t < 23 ) {

      TMirrorPixelData *pixel_data = event->GetPixelData()->GetPixelData(mir_t-1);
      
      T1 = hough->FindFLTrigger( pixel_data, col_t, row_t );
      
      if ( T1 > 0 ) { 
    	pixT1[mir_t-1].Set( col_t, row_t );
    	nT1[mir_t-1]++;
      }    
    }
    
  }
  
  for ( int mirr=Fd::kEYE_FIRST_MIRROR; mirr<=Fd::kEYE_LAST_MIRROR; mirr++ ) {
    
    if ( nT1[mirr-Fd::kEYE_FIRST_MIRROR] == 0 ) continue;
    
    //cout << "Hough T1_" << mirr <<  endl;
    
    hough->DoHoughSel( pixT1[mirr-Fd::kEYE_FIRST_MIRROR], 
                       nT1[mirr-Fd::kEYE_FIRST_MIRROR], 
                       mirr, header->GetEventNo(), 
		       EyeEventClassifier::GetInstance().GetEventClass() );
    
    hough->FillTree( EyeEventClassifier::GetInstance().GetEventClass() );
  }  
}
#endif // HOUGH_TEST

// -------------------------------------------------------------------------

int T3TestApplication::Finish()
 {
  // --- close all files
  
  if ( fMiInputFile ) {
    fMiInputFile->Close();
    delete fMiInputFile;
    fMiInputFile = NULL;
  }
  
  if ( fEyeInputFile ) {
    fEyeInputFile->Close();
    delete fEyeInputFile;
    fEyeInputFile = NULL;
  }
  
  if ( fOutputFile ) {
    fOutputFile->Close();
    delete fOutputFile;
    fOutputFile = NULL;
  }
  
  if ( fRejOutputFile ) {
    fRejOutputFile->Close();
    delete fRejOutputFile;
    fRejOutputFile = NULL;
  }
  
  // --- final cleanup
  
  //delete fEventClassifier;
  
  //fclose( fLogFile );
  
  return 0;
}

// -------------------------------------------------------------------------

int T3TestApplication::Init()
 {
  if ( fInputFilename.empty() ) {
    
    FD_CERR << fArgv[0] << ": no input file name provided!" << endl;
    
    fStatus = 1;
  }
  else {
    if ( fReadMiEvent )
      fMiInputFile = new TMirrorEventFile( fInputFilename.c_str() );
    else
      fEyeInputFile = new TEyeEventFile( fInputFilename.c_str() );
    
    if ( !fReadMiEvent && !fEyeInputFile ) {
      FD_CERR << fArgv[0] << ": cannot open input file!" << endl;
      fStatus = 1;
    }

    if ( fReadMiEvent && !fMiInputFile ) {
      FD_CERR << fArgv[0] << ": cannot open input file!" << endl;
      fStatus = 1;
    }
  }
  
  if ( !fOutputFilename.empty() ) {
  
    fOutputFile = new TEyeEventFile( fOutputFilename.c_str(), "recreate" );
    
    if ( !fOutputFile ) {
      
      FD_CERR << fArgv[0] << ": cannot open output file(accepted events) !" << endl;
      
      fStatus = 1;
    }
  }
  
  if ( !fRejOutputFilename.empty() ) {
  
    fRejOutputFile = new TEyeEventFile( fRejOutputFilename.c_str(), "recreate" );
    
    if ( !fRejOutputFile ) {
      
      FD_CERR << fArgv[0] << ": cannot open output file (rejected events) !" << endl;
      
      fStatus = 1;
    }
  }
  
  // --- exit if unsuccessful so far ...
  
  if ( fStatus ) return fStatus;
  
  // --- display run header/trailer information ...

  if ( fReadMiEvent ) { // reading TMirrorEvent format ?
  
    TMirrorRunHeader *header = NULL;
    TMirrorRunTrailer *trailer = NULL;
  
    if ( fMiInputFile->ReadRunHeader( &header ) && header != NULL ) {
      header->Show();
      //outfile->WriteRunHeader( &header );
    }
    else
      FD_CERR << fArgv[0] << ": no run header found !" << endl;

    if ( fMiInputFile->ReadRunTrailer( &trailer ) && trailer != NULL ) {
      trailer->Show();
      //outfile->WriteRunTrailer( &trailer );
    }
    else
      FD_CERR << fArgv[0] << ": no run trailer found !" << endl;

    delete header;
    delete trailer;
  }
  else {  // reading TEyeEvent format

    TEyeRunHeader *header = NULL;
    TEyeRunTrailer *trailer = NULL;
  
    if ( fEyeInputFile->ReadRunHeader( &header ) && header != NULL ) {
      
      header->Show();
      
      if ( fOutputFile ) 
        fOutputFile->WriteRunHeader( header );
      
      if ( fRejOutputFile ) 
        fRejOutputFile->WriteRunHeader( header );
    }
    else
      FD_CERR << fArgv[0] << ": no run header found !" << endl;

    if ( fEyeInputFile->ReadRunTrailer( &trailer ) && trailer != NULL ) {
      
      trailer->Show();
      
      if ( fOutputFile ) 
        fOutputFile->WriteRunTrailer( trailer );
      
      if ( fRejOutputFile ) 
        fRejOutputFile->WriteRunTrailer( trailer );
    }
    else
      FD_CERR << fArgv[0] << ": no run trailer found !" << endl;

    delete header;
    delete trailer;
  }
  
  // --- create the event classifier (T3)
  
  try {
    EyeEventClassifier::GetInstance().SetLogLevel( 2 );
  
    if ( !fLogFilename.empty() )
      EyeEventClassifier::GetInstance().OpenLogfile( fLogFilename.c_str() );
  }
  catch (...) {
    FD_CERR << fArgv[0] 
            << " caught exception when trying to access the EyeEventClassifier!"
	    << endl;
    fStatus = 1;
  }
  
  return fStatus;
}

// -------------------------------------------------------------------------

void T3TestApplication::Parse()
 {
  int getopt_status;
  
  do {
    
    getopt_status = getopt( fArgc, fArgv, "hi:l:mn:o:r:?" );
    
    if ( getopt_status != EOF ) {
    
      switch ( getopt_status ) {

	case 'i': fInputFilename = optarg;
                  break;

        case 'l': fLogFilename = optarg;
	          break;

        case 'm': fReadMiEvent = true;
	          break;

        case 'n': fNEventLimit = atoi( optarg );
	          break;

	case 'o': fOutputFilename = optarg;
                  break;

	case 'r': fRejOutputFilename = optarg;
                  break;

	case 'h':
	case '?': Usage();
	          break;

	default: printf( "Encountered unknown option: %d,%c\n",
	         getopt_status, getopt_status );
      }
    }
  
  } while ( getopt_status != EOF );
}

// -------------------------------------------------------------------------

int T3TestApplication::Run()
 {
  if ( fStatus ) return fStatus; // don't run
  
  if ( !fMiInputFile && !fEyeInputFile ) {
    FD_CERR << fArgv[0] << ": no valid input data file was provided!" << endl;
    
    return 1;
  }
  
  // cout << "File contains " << file->GetNEntries() << " events !" << endl;
  
#ifdef HOUGH_TEST
  HoughSel * hough = new HoughSel( argv[1] );
#endif // HOUGH_TEST

  TEyeEvent * eye_event = NULL;
  TMirrorEvent * mi_event = NULL;
  
  // --- main event loop ...
  
  while ( 1 ) {

    if ( fReadMiEvent ) {
    
      if ( !fMiInputFile->ReadMirrorEvent( &mi_event ) ) break;
      
      unsigned int ev_no = mi_event->GetEventHeader()->GetEventNo();
      
      if ( !eye_event )
        eye_event = new TEyeEvent();
      else
        eye_event->Clear();
	
      eye_event->AddEvent( mi_event );
      eye_event->GetEventHeader()->SetEventNo( ev_no );
    }
    else
      if ( !fEyeInputFile->ReadEvent( &eye_event ) ) break;

#if 0
    // --- show some info about the event (EventHeader)
    
    eye_event->GetEventHeader()->Show();
#else
    FD_COUT << "Event number " << eye_event->GetEventHeader()->GetEventNo() << endl;
#endif

    // --- do t3 classification 
    
    FD_COUT << "DO T3 CLASSIFICATION " << endl;
    
    try {
      EyeEventClassifier::GetInstance().DoClassification( eye_event );
      
      TEyeEventHeader * header = eye_event->GetEventHeader();
      TEyeEventHeader::EEventClass event_class = EyeEventClassifier::GetInstance().GetEventClass();
      
      header->SetEventClass( event_class );
      
      bool accepted = (   event_class == TEyeEventHeader::kShowerCandidate 
                       || event_class == TEyeEventHeader::kIsShower 
                       || event_class == TEyeEventHeader::kCloseShower
		       || event_class == TEyeEventHeader::kHorizontalShower );

      if ( fOutputFile && accepted )
        fOutputFile->WriteEvent( &eye_event );
      
      if ( fRejOutputFile && !accepted )
        fRejOutputFile->WriteEvent( &eye_event );
    }
    catch ( std::exception& ex ) {
      FD_COUT << "EyeEventClassifier::DoClassification() - caught exception " 
              << ex.what() << endl;
    }
    
    FD_COUT << "EXIT FROM T3 CLASSIFICATION " << endl;

#ifdef HOUGH_TEST
    // --- do Hough classification
    
    cout << "DO HOUGH CLASSIFICATION " << endl;

    DoHoughSel( event, evclass, hough);
    
    cout << "EXIT FROM HOUGH CLASSIFICATION " << endl;
#endif // HOUGH_TEST

#if 1
    // --- display the time(s) needed for the classification
#endif
#if 0
    // --- display the times of each individual Mirror

    for ( unsigned int i=Fd::kEYE_FIRST_MIRROR; i<=Fd::kEYE_LAST_MIRROR; i++ ) {
      TTimeStamp *tstamp = event->GetEventHeader()->GetRawMirrorTime( i );

      if ( tstamp != NULL ) {
        cout << "Time (sec,nsec) for Mirror " << i << " :" << endl;
        cout << TimeConvert::GetGPSTimeFromUTCTime( tstamp->GetSec() )
             << "," << tstamp->GetNanoSec() << endl;
      } 
    }
#endif

#if 0
    // --- show info about the event
    
    eye_event->Show();
#endif

#if 0
    // --- dump the T3 pixel list
    
    T3PixelList *plist = evclass->GetT3PixelList();
    
    if ( plist ) {
      plist->Dump( stdout );
    }
#endif

#if 1
    // --- apply GPS time correction to T3 time
    
    if (    EyeEventClassifier::GetInstance().GetEventClass() == TEyeEventHeader::kShowerCandidate 
         || EyeEventClassifier::GetInstance().GetEventClass() == TEyeEventHeader::kIsShower ) {
      
      TEyeEventHeader *header = eye_event->GetEventHeader();
      
      int time_at_ground = EyeEventClassifier::GetInstance().GetTimeAtGround();
      
      uint32_t sec_out, nsec_out;
      
      header->DoTimeCorrection( header->GetTimeStamp()->GetGPSSec(),
                                time_at_ground, &sec_out, &nsec_out );

      TTimeStamp t3_time(sec_out, (time_t)nsec_out );
      
      cout << "T3 time: " << setw(10) << t3_time.GetGPSSec() << " [sec] "
           << setw(10) << t3_time.GetNanoSec() << " [nsec] (corrected)"
	   << endl;
    }
#endif

  } // while(1)
  
  delete mi_event;
  delete eye_event;
  
#ifdef HOUGH_TEST
  //f->Write();
  hough->Finish();
#endif // HOUGH_TEST

  return fStatus;
}

// -------------------------------------------------------------------------

void T3TestApplication::Usage()
 {
  cout << "Usage: " << fArgv[0] << " -i <input_file> [-m] [-n <event_limit>]"
       << endl
       << "\t\t[-l <log_file>] [-o <output_file>]"
       << endl
       << "\t\t[-r <reject_output_file>] [-h|-?]" 
       << endl;
  cout << endl;
  cout << "\t-i  name of input file" << endl;
  cout << "\t-l  specify file for logging output, default= " 
       << fLogFilename << endl;
  cout << "\t-m  use TMirrorEvent format for input file" << endl;
  cout << "\t-n  specify number of events to analyze" << endl;
  cout << "\t-o  optional name of output file, accepted events" << endl;
  cout << "\t-r  optional name of output file, rejected events" << endl;
  cout << "\t-h | -?  display this help page" << endl;
  cout << endl;
}

// -------------------------------------------------------------------------
// -------------------------------------------------------------------------
