
//
// File   : TltTestApp.cc
//
// Purpose: Implementation of the class TltTestApplication
//
// $Id: TltTestApp.cc 9725 2010-09-28 07:31:17Z mathes $
//


#include <iostream>
#include <climits>
#include <unistd.h>   // getopt()

/** @file TltTestApp.cc
  * Implementation of the class TltTestApplication.
  * @author H.-J. Mathes, FzK
  */


#include <FdNumbering.hh>
#include <FEPageDesc.hh>
#include <FEMiEvent.hh>

#include <FEFileReadout.hh>
#include <PrototypeTltProcessor.hh>
#include <MStcCutTlt.hh>

#include "CheckVersion.hh"

#include "TltTestApp.hh"

using namespace std;
// using namespace MiReadout;
using FdUtil::Fd;

// -------------------------------------------------------------------------
// -------------------------------------------------------------------------
// 
// class TltTestApplication - implementation
// 

TltTestApplication::TltTestApplication(int argc,char** argv)
 : fArgc(argc), fArgv(argv), fStatus(0), 
   fInputFile(NULL), fParameterFilename(""), 
   fEyeOutputFile(NULL), fMiOutputFile(NULL), 
   fEyeRejectOutputFile(NULL), fMiRejectOutputFile(NULL), 
   fLogFile(NULL), 
   fTltProcessor(NULL)
 {
  fNEventLimit    = UINT_MAX;  // nearly unlimited
  fUseEyeEvent    = false;
  fFillNtuple     = false;
  
  fOutputFilename = "tlttest.root";
  
  fUseMTLT        = false;
  
  fVerbosity      = 0;
  
  if ( fArgc <= 1 )
    Usage();

  Parse();
} 

// -------------------------------------------------------------------------

TltTestApplication::~TltTestApplication()
 {
  delete fInputFile;
  delete fMiOutputFile;
  delete fEyeOutputFile;
  delete fMiRejectOutputFile;
  delete fEyeRejectOutputFile;
}

// -------------------------------------------------------------------------

int TltTestApplication::Finish()
 {
  // --- call TltProcessors Finish() method
  
  fTltProcessor->Finish();
  
  // --- close all files
  
  if ( fInputFile ) {
    fInputFile->Close();
    delete fInputFile;
    fInputFile = NULL;
  }
  
  if ( fMiOutputFile ) {
    
    //fMiOutputFile->Write();
    fMiOutputFile->Close();
    
    delete fMiOutputFile;
    fMiOutputFile = NULL;
  }
  
  if ( fEyeOutputFile ) {
  
    //fEyeOutputFile->Write();
    fEyeOutputFile->Close();
    
    delete fEyeOutputFile;
    fEyeOutputFile = NULL;
  }
  
  // --- final cleanup
  
  delete fTltProcessor;
  
  fclose( fLogFile );
  
  return 0;
}

// -------------------------------------------------------------------------

int TltTestApplication::Init()
 {
  MiReadout::MStcCutTltParametersRec mtlt_par = MSTC_TLT_DEFAULT_PARAMETERS;
  MiReadout::TltParametersRec ptlt_par = DEFAULT_TLT_PARAMETERS;

  // -- input filename and file ?
  
  if ( fInputFilename.empty() ) {
    
    FD_CERR << fArgv[0] << ": no input file name provided!" << endl;
    
    fStatus = 1;
  }
  else {
    fInputFile = new TMirrorEventFile( fInputFilename.c_str() );
    
    if ( !fInputFile ) {
      FD_CERR << fArgv[0] << ": cannot open input file!" << endl;
      fStatus = 1;
    }
    
    fRunNumber = 0;
  }

  if ( fStatus ) return fStatus;
    
  // --- parameter file
  
  if ( !fParameterFilename.empty() ) {

    ReadoutConfig *config = ReadoutConfig::GetInstance( fParameterFilename );
    
    if ( fUseMTLT ) {
    
      mtlt_par = *config->GetMStcCutTltParameters();
    
      if ( dynamic_cast<MStcCutTlt*>( fTltProcessor ) != NULL )
        dynamic_cast<MStcCutTlt*>( fTltProcessor )->SetParameters( &mtlt_par );
    }
    else {
    
      ptlt_par = *config->GetTltParameters();
      
      if ( dynamic_cast<PrototypeTltProcessor*>( fTltProcessor ) != NULL )
        dynamic_cast<PrototypeTltProcessor*>( fTltProcessor )->SetParameters( &ptlt_par );
    }
 
  }
  
  // --- output file event format
  
  if ( fUseEyeEvent ) {
    fEyeOutputFile = new TEyeEventFile( fOutputFilename.c_str(), "recreate" );
    
    if ( !fEyeOutputFile ) fStatus = 1;
    
    if ( fRejectOutputFilename.size() != 0 ) {
      fEyeRejectOutputFile = new TEyeEventFile( fRejectOutputFilename.c_str(), "recreate" );
      
      if ( !fEyeRejectOutputFile ) fStatus = 1;
    }
  }
  else {
    fMiOutputFile = new TMirrorEventFile( fOutputFilename.c_str(), "recreate" );
    
    if ( !fMiOutputFile ) fStatus = 1;
    
    if ( fRejectOutputFilename.size() != 0 ) {
      fMiRejectOutputFile = new TMirrorEventFile( fRejectOutputFilename.c_str(), "recreate" );
      
      if ( !fMiRejectOutputFile ) fStatus = 1;
    }
  }
  
  // --- deal with the log-file
  
  if ( fLogFilename.empty() )
    fLogFilename = "tlttest.log";
  
  fLogFile = fopen( fLogFilename.c_str(), "w" );
  
  if ( !fLogFile ) fStatus = 1;
  
  // --- exit if unsuccessful so far ...
  
  if ( fStatus ) return fStatus;
  
  // --- display run header/trailer information ...
  
  TMirrorRunHeader *header = NULL;
  TMirrorRunTrailer *trailer = NULL;
  
  if ( fInputFile->ReadRunHeader( &header ) && header != NULL ) {
    header->Show();
    //outfile->WriteRunHeader( &header );
  }
  else
    FD_CERR << fArgv[0] << ": no run header found !" << endl;

  if ( fInputFile->ReadRunTrailer( &trailer ) && trailer != NULL ) {
    trailer->Show();
    //outfile->WriteRunTrailer( &trailer );
  }
  else
    FD_CERR << fArgv[0] << ": no run trailer found !" << endl;

  delete header;
  delete trailer;

  // --- create the selected TLT processor (algorithm)
  
  if ( fUseMTLT ) {
    fTltProcessor = new MStcCutTlt(&mtlt_par);
    
    if ( fFillNtuple )
      dynamic_cast<MStcCutTlt*>(fTltProcessor)->FillNtuple( true );
  }
  else
    fTltProcessor = new PrototypeTltProcessor();

  if ( !fTltProcessor )
    fStatus = 1;
  else {
    // cd() to a writable directory/file
    if ( fUseEyeEvent )
      fEyeOutputFile->cd();
    else
      fMiOutputFile->cd();
    
    fTltProcessor->Init();
    
    fTltProcessor->SetPrintLevel( GetVerbosityLevel() );
    fTltProcessor->SetLogfile( fLogFile );
  }
  
  return fStatus;
}

// -------------------------------------------------------------------------

void TltTestApplication::Parse()
 {
  int getopt_status;
  
  do {
    
    getopt_status = getopt( fArgc, fArgv, "aehi:l:mn:o:p:r:v:?" );
    
    if ( getopt_status != EOF ) {
    
      switch ( getopt_status ) {
      
        case 'a': fFillNtuple = true;
	          break;
	
        case 'e': fUseEyeEvent = true;
	          break;

	case 'i': fInputFilename = optarg;
                  break;

        case 'l': fLogFilename = optarg;
	          break;

        case 'm': fUseMTLT = true;
	          break;

        case 'n': fNEventLimit = atoi( optarg );
	          break;

	case 'o': fOutputFilename = optarg;
                  break;

	case 'p': fParameterFilename = optarg;
                  break;

	case 'r': fRejectOutputFilename = optarg;
                  break;

        case 'v': fVerbosity = atoi( optarg );
	          break;

	case 'h':
	case '?': Usage();
	          break;

	default: FD_CERR << "Encountered unknown option: " 
	                 << (int)getopt_status
	                 << "," << (char)getopt_status << endl;
      }
    }
  
  } while ( getopt_status != EOF );
}

// -------------------------------------------------------------------------

int TltTestApplication::Run()
 {
  if ( fStatus ) return fStatus; // don't run
  
  // here only needed as a dummy argument to the TltProcessor
  FEPageDesc *page_desc = new FEPageDesc();
  
  unsigned int event_no = 0;
  
  // --- event buffer(s)

  TMirrorEvent *inevent = new TMirrorEvent();
  TMirrorEvent *mi_outevent = new TMirrorEvent();
  
  TEyeEvent *eye_outevent = new TEyeEvent();
  
  // --- handles to the readout means/method
  
  FEFileReadout *readout = new FEFileReadout( &inevent );
  FEMirrorEvent *mi_event = new FEMirrorEvent( readout, page_desc );
  
  VTltProcessor * tlt = GetTltProcessor();
  
  // --- main event loop ...

  while ( inevent && fInputFile->ReadMirrorEvent( &inevent ) != kFALSE 
                  && fNEventLimit-- ) {
    
    event_no++;
    
    FD_COUT << "(> " << fArgv[0] << ": TLT event= " << event_no 
            << " event on file= " << inevent->GetEventHeader()->GetEventNo()
	    << endl;

    fprintf( fLogFile, "(> TLT event= %u event on file= %d\n",
                     event_no, inevent->GetEventHeader()->GetEventNo() );

    // the RUN number might change in concatenated test files
    fRunNumber = inevent->GetEventHeader()->GetRunNo();

    // inevent->GetEventHeader()->Dump();

    // inevent->GetPixelData()->Dump();
    
#if 0
    FD_COUT << "PixelList size: " << inevent->GetPixelList()->GetNumPixels() << endl;
    //TMirrorBitArray ba( inevent->GetPixelData(), 15, 35 );
    TMirrorBitArray ba;
//    ba.Set( inevent->GetPixelData(), 15, 35 );
    ba.Set( inevent->GetPixelData(), 15, 99 );
    ba.Dump("after Set()");
    ba.Extend( 1, 1 );
    ba.Dump("after Extend()");
#endif

    // --- optional: generate multiplicity data

    if ( fUseMTLT ) {
      TMirrorPixelData *pdata = inevent->GetPixelData();
    
      // pointer to array (owned by TMirrorPixelData) of mult data words
      TMirrorPixelData::MultiplicityDataWord mult = pdata->GetMultiplicityData();

#if (FDUTIL_VERSION_CODE >= ModuleVersionCode(4,0,0))
      unsigned int n_mult_bins = Fd::kMI_NMULT_BINS;
#else
      unsigned int n_mult_bins = Fd::kMI_NFADC_BINS;
#endif // FDUTIL_VERSION_CODE

      int mult_sum = 0;
      for ( unsigned int m=0; m<n_mult_bins; m++ )
        mult_sum += (int)mult[m];
      
      if ( mult_sum == 0 ) {
        FD_CWARN << "TltTestApplication: there is no multiplicity data available, "
                 << "I am going to calculate it from the data!" << endl;

        TMirrorFADCData *mi_fadc = inevent->GetFADCData();
        TMirrorPixelList *plist = inevent->GetPixelList();
      
        TMirrorPixelData::CreateMultiplicityData( mi_fadc, plist, mult );
      }

# if 0  // control output
      mult_sum = 0;
      for ( unsigned int m=0; m<n_mult_bins; m++ )
        mult_sum += (int)mult[m];
      
      FD_COUT << "mult_sum= " << mult_sum << endl;
# endif
    } // if ( fUseMTLT ) ...
        
    tlt->Clear();
    
    tlt->Decide( mi_event );
    
    FD_COUT << fArgv[0] << ": Accept= " 
            << (int)tlt->IsEventAccepted() 
	    << " <)" << endl;
    
    if ( tlt->IsEventAccepted() ) {
      
      if ( fUseEyeEvent ) {
        eye_outevent->Clear();
	eye_outevent->AddEvent( inevent );
	
	eye_outevent->GetEventHeader()->SetRunNo( fRunNumber );
	eye_outevent->GetEventHeader()->SetEventNo( 
	                      inevent->GetEventHeader()->GetEventNo() );
	
	fEyeOutputFile->WriteEvent( &eye_outevent );
      }
      else {
        *mi_outevent = *inevent;
    
        fMiOutputFile->WriteMirrorEvent( &mi_outevent );
      }
    } // is accepted
    else if ( fRejectOutputFilename.size() ){ 

      if ( fUseEyeEvent ) {
	eye_outevent->Clear();
	eye_outevent->AddEvent( inevent );
	
	eye_outevent->GetEventHeader()->SetRunNo( fRunNumber );
	eye_outevent->GetEventHeader()->SetEventNo( 
	                      inevent->GetEventHeader()->GetEventNo() );

	fEyeRejectOutputFile->WriteEvent( &eye_outevent );
      }
      else {
        *mi_outevent = *inevent;
      
        fMiRejectOutputFile->WriteMirrorEvent( &mi_outevent );
      }
    }
  }
  
  // --- final cleanup after main processing loop
  
  delete page_desc;
  delete readout;

  delete inevent;
  delete mi_outevent;
  delete eye_outevent;

  return fStatus;
}

// -------------------------------------------------------------------------

void TltTestApplication::ShowSettings(std::ostream& ostr)
 {
  ostr << "--- TltTestApplication::ShowSettings() --- " << endl;
}

// -------------------------------------------------------------------------

void TltTestApplication::Usage()
 {
  cout << "Usage: " << fArgv[0] << " -i <input_file> [-o <output_file>] [-e]"
       << endl
       << "\t\t[-n <event_limit>] [-m] [-l <log_file>] [-p <parameter_file>]" 
       << endl
       << "\t\t[-v <verb_level>] [-r <reject_output_file>] [-a] [-h|-?]"
       << endl;
  cout << endl;
  cout << "\t-i  name of input file" << endl;
  cout << "\t-p  name of optional ini-file (TLT parameters)" << endl;
  cout << "\t-o  optional name of output file, default= 'tlttest.root'" << endl;
  cout << "\t-a  write analysis results into ntuple" << endl;
  cout << "\t-e  use TEyeEvent format for output file" << endl;
  cout << "\t-n  specify number of events to analyze" << endl;
  cout << "\t-m  use mTLT instead of pTLT" << endl;
  cout << "\t-r  optional name of output file for rejected events" << endl;
  cout << "\t     the same file format is used as selected for the output" << endl;
  cout << "\t-l  specify file for logging output, default= 'tlttest.log'" << endl;
  cout << "\t-v  specify the verbosity level, default= 0 " << endl;
  cout << "\t-h | -?  display this help page" << endl;
  cout << endl;
}

// -------------------------------------------------------------------------
// -------------------------------------------------------------------------
