
//
// File   : eyeread.cc
//
// Purpose: Simple test program to read events in TEyeEvent format
//
// $Id: eyeread.cc 8757 2010-09-27 12:16:09Z mathes $
//


#include <iostream>
#include <iomanip>
#include <string>
#include <sstream>
#include <cstdlib>

#include <boost/program_options.hpp>


/** @file eyeread.cc
  * Simple test read program for Eye event format.
  * @author H.J. Mathes, FzK
  */

#include <TClass.h>
#include <TStreamerInfo.h>

#include <FdNumbering.hh>
#include <ErrorLogging.hh>

#include <MiEvent.hh>
#include <EyeEvent.hh>
#include <EyeEventFile.hh>

using namespace std;
using FdUtil::ErrorLogger;
using FdUtil::Fd;

namespace po = boost::program_options;

// -------------------------------------------------------------------------

static void ShowFdPixelNumberInfo(Fd::FdPixelNumber pixel)
 {
  cout << "FdPixel: " << setw(5) << Fd::GetPixelNo( pixel )
       << " = Mirror: " << setw(2) << Fd::GetMirrorNo( pixel )
       << "  Eye: " << Fd::GetEyeNo( pixel ) 
       << endl;
}

// -------------------------------------------------------------------------

static void ShowPixelNumberInfo(Fd::PixelNumber pixel)
 {
  cout << "Pixel: " << setw(3) << Fd::GetPixelNo( pixel )
       << " =  Col.: " << setw(2) << Fd::GetColumnNo( pixel )
       << "   Row: " << setw(2) << Fd::GetRowNo( pixel ) 
       << ",  Mod.: " << setw(2) << Fd::GetModuleNo( pixel )
       << "   Ch.: " << setw(2) << ",  " << Fd::GetChannelNo( pixel )
       << "Mirror: " << setw(2) << Fd::GetMirrorNo( pixel) 
       << "  Eye: " << Fd::GetEyeNo( pixel ) << endl;
}

// -------------------------------------------------------------------------

int main(int argc, char **argv)
 {
  // --- setup ErrorLogger properly
  
  ErrorLogger &logger = ErrorLogger::GetInstance();
  
  //logger.SetLoggingLevel( ErrorLogger::eInfo );  // see below
  logger.SetStream( &std::cout );
  
  // --- setup Boost option parser etc.
  
  ostringstream generic_opt_str;
  
  generic_opt_str << "Usage: " << argv[0] << " [options] <input-file>"
                  << endl << "Allowed options";
	  
  // Declare the supported options.
  po::options_description generic_opts( generic_opt_str.str() );
  
  generic_opts.add_options()
    ("help,h", ": produce help message")
    ("input-file,i", po::value<string>(), ": input file (optional switch)")
    ;
  
  // to allow parsing of "unnamed" arguments
  po::positional_options_description positional_opts;
  positional_opts.add("input-file", 1);
  
  // event dump specific options
  po::options_description display_opts("Display options");
  display_opts.add_options()
    ("logging-level", po::value<int>()->default_value(0), ": set logging level")
    ("eyeevent-printlevel", po::value<int>()->default_value(0), ": set logging level for TEyeEvent")
    ("mievent-printlevel", po::value<int>()->default_value(0), ": set logging level for TMirrorEvent")
    ("root-debug-level", po::value<int>()->default_value(0), ": set ROOT debug level")
    ("show-streamerinfo", po::value<bool>()->default_value(false), ": show TStreamerInfo of class TEyeEvent")
    ("show-file-statistics", po::value<bool>()->default_value(true), ": show some statistics about TEyeEventFile")
    ("show-file-streamerinfo", po::value<bool>()->default_value(false), ": show TStreamerInfo of all objects found in filet")
    ("show-runheader", po::value<bool>()->default_value(true), ": display TEyeRunHeader")
    ("show-runtrailer", po::value<bool>()->default_value(true), ": display TEyeRunTrailer")
    ("show-eventheader", po::value<bool>()->default_value(true), ": display TEyeEventHeader")
    ("show-pixellist", po::value<bool>()->default_value(false), ": display TEyePixelList")
    ("show-pixellist-extended", po::value<bool>()->default_value(false), ": display TEyePixelList (extended form)")
    ("show-pixeldata", po::value<bool>()->default_value(false), ": display TEyePixelData")
    ("show-fadcdata", po::value<bool>()->default_value(false), ": display TEyeFADCData")
    ("show-geometrydata", po::value<bool>()->default_value(true), ": display TEyeGeometryData")
    ("show-t3data", po::value<bool>()->default_value(true), ": display TEyeT3Data")
    ;

  // concatenate all these options
  generic_opts.add(display_opts);
  
  po::variables_map var_map;
  
  po::store(po::command_line_parser(argc, argv).options(generic_opts).
                    positional(positional_opts).run(), var_map );
  po::notify(var_map);   
  
  // issue 'usage' ...
  if (var_map.count("help") || (argc < 2)) {
    cout << generic_opts << endl;
    exit( EXIT_FAILURE );
  }

  // --- set the print level of the TEye* and TMirror* classes ...
  
  int eye_event_printlevel = var_map["eyeevent-printlevel"].as<int>();
  
  TEyeEventFile::SetPrintLevel( eye_event_printlevel );
  TEyeEvent::SetPrintLevel( eye_event_printlevel );
  TMirrorEvent::SetPrintLevel( 0 );
 
  // set logging level of the ErrorLogger
  logger.SetLoggingLevel( (ErrorLogger::ESeverityLevel)var_map["logging-level"].as<int>() );
  
  // turn on ROOT debugging mode
  gDebug = var_map["root-debug-level"].as<int>();

  // Note: Both ways work as ROOT will perform new/delete during reading
  //       from the TTree but finally you must do a delete !
  TEyeEvent* event = /* NULL */ new TEyeEvent();

#if 0
  event->Show();
#endif

  TEyeEventFile *file;
  
  // --- display StreamerInfo of some classes (classes in memory)

  if ( var_map["show-streamerinfo"].as<bool>() )
    TEyeEvent::Class()->GetStreamerInfo()->ls();
  
  if ( argc > 1 ) // filename was provided
    file = new TEyeEventFile( argv[1] );
  else
    file = new TEyeEventFile( "EyeEvent2.root" );
  
  if ( file ) {
    
    if ( var_map["show-file-statistics"].as<bool>() )
      file->Show();

    // see ROOT TStreamerInfo class documentation
    if ( var_map["show-file-streamerinfo"].as<bool>() )
      file->ShowStreamerInfo();
  }
  
  // --- test reading of header/trailer objects ...
  
  TEyeRunHeader *header = NULL;
  if ( file->ReadRunHeader( &header ) == kTRUE ) {
    if ( var_map["show-runheader"].as<bool>() )
      header->Show();
    delete header;
  }
  else
    FD_CERR << "No TEyeRunHeader found !" << endl;
  
  TEyeRunTrailer *trailer = NULL;
  if ( file->ReadRunTrailer( &trailer ) == kTRUE ) {
    if ( var_map["show-runtrailer"].as<bool>() )
      trailer->Show();
    delete  trailer;
  }
  else
    FD_CERR << "No TEyeRunTrailer found !" << endl;
  
  while ( 1 ) {
  
    gDebug = var_map["root-debug-level"].as<int>();
    
    if ( (file->ReadEvent( &event ) != kTRUE) ) break;
  
    // --- access to TEyeEventHeader
    
    if ( var_map["show-eventheader"].as<bool>() ) {
      
      TEyeEventHeader *header = event->GetEventHeader();
      
      header->Show();
    }
    
    if ( var_map["show-pixellist"].as<bool>() ) {
    
      TEyePixelList* eye_pixel_list = event->GetPixelList();
      
      eye_pixel_list->Show();
    }
    
#if 1
    // --- access to TEyePixelList

    if ( var_map["show-pixellist-extended"].as<bool>() ) {
    
      TEyePixelList* eye_pixel_list = event->GetPixelList();

      // iterator-like access to the pixel list ...
  
      for ( unsigned int i=0; i<eye_pixel_list->GetNumPixels(); i++ ) {
        
	Fd::FdPixelNumber fdpixel = eye_pixel_list->GetPixel( i );
	
	// skip this for virtual pixels
	if ( Fd::GetRowNo( Fd::GetPixelNumber( fdpixel ) ) > Fd::kMI_NROWS ) continue;
        
	int first_trigger = event->GetPixelData()->GetFirstTriggeredTimeBin(fdpixel);
	
	ostringstream ft;
	ft << event->GetPixelData()->GetFirstTriggeredTimeBin(fdpixel);
	
	FD_COUT << "Found pixel: " << Fd::GetPixelNo( fdpixel ) << ", "
	        << ((first_trigger != TEyePixelData::kNoTrigger) ? 
		                      "first triggered bin= " : "not triggered !")
	        << ((first_trigger != TEyePixelData::kNoTrigger) ? ft.str() : "")
	        << endl;
      }
    }
#endif

#if 0
#ifndef DPA
    // --- array-like access to the pixel list ... (not in DPA version)

    if ( var_map["show-pixellist-extended"].as<bool>() ) {
    
      TEyePixelList* eye_pixel_list = event->GetPixelList();
      
      //eye_pixel_list->Show();
  
      for ( unsigned int i=0; i<Fd::kEYE_NEXTPIXELS; i++ ) {
      
        Fd::FdPixelNumber fdpixel = (*eye_pixel_list)[i + Fd::kEYE_FIRST_PIXEL];

	if ( !fdpixel ) continue;
	
        cout << "Found pixel: " << Fd::GetPixelNo( fdpixel ) << ","
                                << Fd::GetEyePixelNo( fdpixel ) << "," 
		                << Fd::GetMirrorPixelNo( fdpixel ) << endl;
      }
    }
#endif // DPA
#endif

#if 0
    // --- access to the TEyePixelData ...
    
    if ( var_map["show-pixeldata"].as<bool>() ) {
          
      TEyePixelData *eye_pixel_data = event->GetPixelData();
    
      eye_pixel_data->Show();
    }
#endif

    // --- access to the TEyePixelData ...
    
    if ( var_map["show-pixeldata"].as<bool>() ) {
    
      TEyeEventHeader *header = event->GetEventHeader();
      TEyePixelData *eye_pixel_data = event->GetPixelData();
      TMirrorPixelData *mi_pixel_data;
    
      for ( unsigned int m=Fd::kEYE_FIRST_MIRROR;
                         m<=Fd::kEYE_LAST_MIRROR; m++ ) {
      
        if ( !header->IsMirrorDataPresent( m ) ) continue;
      
        mi_pixel_data = eye_pixel_data->GetPixelData( m );
      
        if ( !mi_pixel_data ) {
          FD_CERR << "Error: no TMirrorPixelData for Mirror " 
	          << m << " !" << endl;
          continue;
        }
      
        mi_pixel_data->Show();
      }
    }

#if 0
    // --- access all FADC data using the TIterator class
  
    if ( var_map["show-fadcdata"].as<bool>() ) {
    
      TEyeFADCData* eye_fadc_data = event->GetFADCData();
      TIterator* iter = eye_fadc_data->GetIterator();
    
      TFADCData* fadc;
  
      iter->Reset();
  
      while ( ( fadc = (TFADCData*)((*iter)()) ) != NULL ) {
      
        if ( fadc != NULL ) {
          cout << "Retrieved FADC data "
               << "PixNo,EyePixNo = " << Fd::GetPixelNo( fadc->GetPixelNumber() ) 
               << "," << Fd::GetEyePixelNo( fadc->GetPixelNumber() ) << endl;
        }
      }
    }
#endif

#if 1
    // --- access all FADC data using the TEyePixelList
    
    if ( var_map["show-fadcdata"].as<bool>() ) {
    
      TEyePixelList* eye_pixel_list = event->GetPixelList();
      TEyeFADCData* eye_fadc_data = event->GetFADCData();

      // access to all pixels FADC data traces
  
      for ( unsigned int i=0; i<eye_pixel_list->GetNumPixels(); i++ ) {
  
        Fd::FdPixelNumber fdpixel = eye_pixel_list->GetPixel( i );
        const TFADCData* fadc = eye_fadc_data->GetFADCData( fdpixel );
    
        cout << "Pixel(no,Eye,Mi): " 
             << Fd::GetPixelNo( fdpixel ) << ","
             << Fd::GetEyePixelNo( fdpixel ) << "," 
	     << Fd::GetMirrorPixelNo( fdpixel ) << endl;
    
        if ( fadc != NULL ) {
	
          cout << "Retrieved FADC data " << Fd::GetPixelNo( fadc->GetPixelNumber() ) 
               << "," << Fd::GetEyePixelNo( fadc->GetPixelNumber() ) << endl;
    
          // check of data integrity only with specially created test data
#if 0
    
          unsigned int pixel_no = Fd::GetMirrorPixelNo( fdpixel );

          TFADCData::FADCDataWord data = fadc->GetFADCTrace();

#if (MIEVENT_VERSION_CODE < ModuleVersionCode(4,0,0))
          unsigned int n_fadc_bins = Fd::kMI_NFADC_BINS;
#else
          unsigned int n_fadc_bins = fadc->GetNFADCBins();
#endif // MIEVENT_VERSION_CODE
	
          for ( unsigned int j=0; j<n_fadc_bins; j++ )
            if ( FADCDataWordGetData( &data[j] ) != ( j + pixel_no ) ) {
	      cout << "Data mismatch, pixel "
	           << Fd::GetPixelNo( fadc->GetPixelNumber() )
		   << " data= " << FADCDataWordGetData( &data[j] ) << endl;
              break;
            }
#endif
        }
      }
    }
#endif
  
#if 0
    event->Show();
#endif    

    if ( var_map["show-t3data"].as<bool>() ) {
    
      event->GetT3Data()->Show();
    }
    
    if ( var_map["show-geometrydata"].as<bool>() ) {
      
      event->GetGeometryData()->Show();
    }
    
  } // while ( ... )
  
  delete file;
  delete event;
  
  exit( EXIT_SUCCESS );
}

// -------------------------------------------------------------------------
// -------------------------------------------------------------------------
