
//
// File   : eyeselect.cc
//
// Purpose: Program to create a new event file with the selected events.
//
// $Id: eyeselect.cc 8757 2010-09-27 12:16:09Z mathes $
//


#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <unistd.h>   // getopt()

#include <iostream>
#include <algorithm>
#include <list>

/** @file eyeselect.cc
  * Program to create a new event file with selected events.
  * @author H.-J. Mathes, FzK
  */

#include <EyeEvent.hh>
#include <EyeEventFile.hh>


#include "EyeProcFramework.hh"
#include "EyeDataProc.hh"

using namespace std;

// -------------------------------------------------------------------------
// -------------------------------------------------------------------------
//
// class SelectEvents - declaration
//

class SelectEvents : public EyeDataProcessor {
public:
    /** Default constructor of the class SelectEVents. */
    SelectEvents();
    
    /**  */
    ~SelectEvents() {}

    /** ProcessEvent() method of the SelectEvents class. */
    bool ProcessEvent(TEyeEvent*);
    
    /** Read the list of events to select. */
    void ReadEventList(FILE*);
    
    /** Set the print output level. */
    void SetPrintLevel(unsigned int print) { fPrintLevel = print; }
 
private:
    std::list<unsigned int>   fEventsToBeSelected;
    unsigned int              fPrintLevel; // to steer additional debug output
};

// -------------------------------------------------------------------------
// -------------------------------------------------------------------------
//
// class SelectEvents - implementation
//

SelectEvents::SelectEvents()
 : EyeDataProcessor("SelectEvents"), fPrintLevel(0)
 { }

// -------------------------------------------------------------------------

bool SelectEvents::ProcessEvent(TEyeEvent* event)
 {
  bool store = false;
  
  if ( fEventsToBeSelected.size() == 0 ) {
    SetStatus( 1 );
    cerr << "Last event is reached or number of events to process is 0 !\n";
  }
  
  // --- event processing ...

  unsigned int event_no = event->GetEventHeader()->GetEventNo();
  
  std::list<unsigned int>::iterator found;
  
  if ( (found = find( fEventsToBeSelected.begin(),
                      fEventsToBeSelected.end(), 
		      event_no )) != fEventsToBeSelected.end() ) {

    if ( fPrintLevel >= 2 ) {
      cout << "Selected event " << *found 
           << " !" << endl;
      cout << "Actual list size: " << fEventsToBeSelected.size() << endl;
    }

    fEventsToBeSelected.erase( found );

    store = true;
    
    if ( fEventsToBeSelected.empty() ) {
       cout << "SelectEvents::ProcessEvent() - last event in list reached !"
            << endl;
       cout << "                             -> trying to terminate properly ..." << endl;
       EyeProcessingFramework::SetTerminate();
    }
  }
    
  if ( fPrintLevel >= 1 )
    cout << "Evt: " << event->GetEventHeader()->GetEventNo() 
         << " Store: " << (Int_t)store << endl;
  
  if ( store ) event->GetFADCData()->Show();
  
  return store;
}

// -------------------------------------------------------------------------
  
void SelectEvents::ReadEventList(FILE *file)
 {
  if ( !file ) {
    cerr << "No file/input stream is open to read event ids from ..." << endl;
    SetStatus( 1 );
    return;
  }
  
  char line[80];
  
  while ( !feof( file ) ) {

    fgets( line, sizeof( line ), file );
    
    if ( !feof( file ) ) {
      unsigned int event_id;
      sscanf( line, "%u", &event_id );
    
      fEventsToBeSelected.push_back( event_id );
    }
  }
  
  // fEventsToBeSelected.unique();  // each event is only allowed once
  fEventsToBeSelected.sort();
  
  if ( fPrintLevel >= 1 ) {
    cout << "Entered " << fEventsToBeSelected.size() 
         << " events to be selected." << endl;
    cout << "First event to be selected: " 
         << fEventsToBeSelected.front() << endl;
  }
}

// -------------------------------------------------------------------------
// -------------------------------------------------------------------------

static void Usage(const char *cmd_name)
 {
  cout << "Usage: " << cmd_name << " <input_file> <output_file> "
       << "[-f <select_file>] [-p <log_level>]" << endl;
  cout << "\t-f <select_file>\t: specify name of select-file, "
       << "otherwise we read from stdin" << endl;
  cout << "\t-p \t\t\t: optional print level" << endl;
  cout << "\t-? \t\t\t: print this information" << endl;
}

// -------------------------------------------------------------------------

/** main() function of the event selection program.
  *
  * This program takes its input parameters from the command line. The list
  * of event numbers which should be writen to the output file are taken
  * either from a file (option -f) or by stdin redirection.
  */
int main(int argc, char** argv)
 {
  if ( argc < 3 ) {
    Usage( argv[0] );
    exit( EXIT_FAILURE );
  }
  
  EyeProcessingFramework *framework = new EyeProcessingFramework();
  
  if ( argc >= 2 )
    framework->SetInputFile( argv[1] );    

  if ( argc >= 3 )
    framework->SetOutputFile( argv[2] );
    
  FILE *input = NULL;
  
  SelectEvents* eye_select = new SelectEvents();
  
  framework->AddProcessor( eye_select );
  
  // --- evaluate cmd line parameters via getopt() ...
  
  int print_level = 1;
  
  int getopt_status;
  do 
   {
    getopt_status = getopt ( argc, argv, "f:p:?" );
    if ( getopt_status != EOF )
     {
      switch ( getopt_status )
       {
	case 'f': char infilename[200];
	          strcpy ( infilename, optarg );
		  input = fopen( infilename, "r" );
                  break;
	
	case 'p': sscanf( optarg, "%d", &print_level );
	          break;
		  
	case '?': Usage( argv[0] );
		  exit( EXIT_FAILURE );
	          break;

	default: printf ( "Encountered unknown option: %d,%c\n",
	         getopt_status, getopt_status );
      }
    }
  }
  while ( getopt_status != EOF );

  eye_select->SetPrintLevel( print_level );
  
  if ( !input ) input = stdin;
  
  eye_select->ReadEventList( input );
  
  // ... done !
  
  int status;
  
  do {
    status = framework->Run();
    
  } while ( !status );

  cout << argv[0] << ": main loop terminated with status= " << status << endl;
    
  delete framework;
  
  exit( EXIT_SUCCESS );
}

// -------------------------------------------------------------------------
// -------------------------------------------------------------------------
