#include "MdFile.h"
#include "MdEvent.h"
#include "TFile.h"
#include "TObjArray.h"
#include "TKey.h"

#include <iostream>
#include <sstream>
using std::cout;
using std::cerr;
using std::endl;

namespace md {

File::File(): fKeyIterator(0) , fCycle ( 0 ) {};

File::File(const File & theFile){
  fFiles         = theFile.fFiles;
  fFilesIterator = theFile.fFilesIterator;
  fKeyIterator   = theFile.fKeyIterator;
  fCycle         = theFile.fCycle;
};

File::~File() 
{
////  // delete all open files
////  for (FileListIterator it = fFiles.begin();
////       it != fFiles.end(); it++)
////    delete *it;
  fFiles.clear();
  delete fKeyIterator;
};

File::File(int argc, char **argv) 
{
  fKeyIterator = 0;
  fCycle = 0;
  for (int i = 0; i < argc; i++)
    this->Open(argv[i], File::eRead);
};

File::File( std::vector<std::string> v) 
{
  fKeyIterator = 0;
  fCycle = 0;
  for (unsigned int i = 0; i < v.size(); i++)
    this->Open(v[i].c_str(), File::eRead);
};

File::File(const char *filename, Mode mode) {
  fKeyIterator = 0;
  fCycle = 0;
  this->Open(filename, mode);
};

const char * File::GetFileName(){
   if( fFiles.size()==1)
      return (*(fFiles.begin())).get()->GetName();      
   else
     return "unknown";
}

int File::GetBytesWritten(){
   if( fFiles.size()==1)
      return (*(fFiles.begin())).get()->GetFileBytesWritten();      
   else
     return -1;
}

File::Status File::Open(const char *filename, Mode mode) {
  // translate mode for ROOT
  std::string rootmode;
  switch (mode) {
  case eRead:
    rootmode = "read";
    break;
  case eWrite:
    rootmode = "recreate";
    break;
  case eAppend:
    rootmode = "update";
    break;

  default:
    cerr << "File:: Unknown mode " << mode << endl;
    return eFailure;
  }

  TFile *file = TFile::Open(filename, rootmode.c_str());

  if (!file)
    { 
      cerr << "Error in: " << filename << endl;
      return eFailure;
    }

  fFiles.push_back( counted_ptr<TFile>(file) );

  // set our pointers to beginning
  fKeyIterator = new TIter(file->GetListOfKeys());
  fFilesIterator = fFiles.begin();
  (*fFilesIterator)->cd();
  return eSuccess;
};

unsigned int File::NumberOfEvents() {
  unsigned int ret=0;
  for (unsigned int i=0;i<fFiles.size();i++) {
    ret+=fFiles[i]->GetNkeys();
  }
  return ret;
}

File::Status File::Open(const char *filenames) {

  std::stringstream files;
  files << filenames;
  std::string buffer;

  // separate the file names assuming they are separated by space
  // create the TFiles and push them in the list
  while (files >> buffer) {
    TFile *file = TFile::Open(buffer.c_str(), "read");
    if ( !file || file->IsZombie() )
     { 
      cerr << "Error in: " << buffer << endl;
      return eFailure;
     }
    fFiles.push_back(counted_ptr<TFile>(file));
  }

  // set pointers to beginning
  fFilesIterator = fFiles.begin();
  (*fFilesIterator)->cd();
  fKeyIterator = new TIter((*fFilesIterator)->GetListOfKeys());
  return eSuccess;
};


File::Status File::ReadObj(TKey * key, Event & event) {
  (*fFilesIterator)->cd();
  //cout << "key->GetName() " << key->GetName() << endl;
  //cout << "key->GetCycle() " << key->GetCycle() << endl;
  //cout << "Cycle " << fCycle << endl;

  // read the object with that key
  Event *tmp = dynamic_cast < Event * >(key->ReadObj());
  if (!tmp)
    return eFailure;

  event = *tmp;
  event.SetId( fCycle );//manage the event ID globally

  return eSuccess;
};

File::Status File::ReadNext(Event & event) {

  // try to grab the next key
  TKey *key = dynamic_cast < TKey * >( (*fKeyIterator)() );
  fCycle++;

  // if it is empty, go to next file
  if (!key) {
    delete fKeyIterator;
    fKeyIterator = 0;
    fFilesIterator++;
    if (fFilesIterator == fFiles.end())
      return eEOF;
    fKeyIterator = new TIter((*fFilesIterator)->GetListOfKeys());
    key = dynamic_cast < TKey * >( (*fKeyIterator)() );
  }
  // read the object with that key
  return ReadObj(key, event);
};

File::Status File::Read(Event & event, Key & key) 
{
  const char *name = key.fTriggerAlgo.c_str();

  FileListIterator file = fFiles.begin();

  while (file != fFiles.end()) {

    // randomly access the key
    TKey *rootkey = (*file)->FindKey(name);
    // if not availble, try next file
    if (!rootkey) {
      file++;
      continue;
    }
    // read the object with that key
    return ReadObj(rootkey, event);
  }

  return eEOF;
};




File::Status File::Write(Event & event, bool verbose) 
{
  // always write in the first file of the list
  // If it does not exist, then cannot write
  //if (!fFiles[0]){
  if (!fFiles.size()){
     cerr << "Error trying to write in root file"<< endl;
    return eFailure;
  }

  //TFile *file = fFiles[0];
  TFile *file = (*(fFiles.begin())).get();
  //cout << "Recovering root file: " << file->GetName() << endl;
  file->cd();

  // generate the key for the event
  Key key(event);

  if (verbose)
    //cout << "Writing event with key: " << key.fTriggerAlgo << endl;

  //cout << "Cycle " << fCycle << endl;
  event.SetId( fCycle++ );//Control eventID when writing besides any other previous value
  event.Write(key.fTriggerAlgo.c_str());

  return eSuccess;

};

void File::Close() {
  // close all files
  for (FileListIterator it = fFiles.begin();
       it != fFiles.end(); it++)
    (*it)->Close();

  fFiles.clear();
};

}//end namespace
