#include <AtmCorr.h>
#include <Atmosphere.h>

using namespace AtmCorrNS;

AtmCorr::AtmCorr(int DetectorType)
{
  if (DetectorType<0 || DetectorType>2) 
    {
      printf("Incorrect Detector type \n");
      exit(1);
    }
  fDetectorType=DetectorType;
};


//----------------------------------------------------------------------------------------------------------------------------
//             Type I atmospheric corection:  
//                       Multiplying factor to convert signal for a shower with (r,rhoGround) to the reference case defined as:
//                       Density at ground 1.043e-3 g/cm^3 

double AtmCorr::GetAlphaRho(double r,int icomp)  // Alpha rho was obtained from simulations with hground=1452.e2 for Water Cherenkov          
{
  double r_km=r/1.e5;
  double alphaRho;
  if (icomp==1)         alphaRho=0.072*pow(r_km,0.47);
  else if (icomp==2)    alphaRho=0.037*pow(r_km,0.82);
  else if (icomp==3)    alphaRho=0.035*pow(r_km,0.96);
  return alphaRho;
}

double AtmCorr::GetCorrTypeI(double r, int icomp, double rhoGround, double hGround)
{
  AtmosphereNS::Atmosphere fatm;  
  const double rhoGroundRef=fatm.GetDensity(hGround,-1); 

  //Moliere correction
  if (icomp>0 && icomp<4)
    {
      double alphaRho=GetAlphaRho(r,icomp);
      double fcorr= (1.-alphaRho*(rhoGround-rhoGroundRef)*1.e3/0.03);
      return 1./fcorr;    
    } 
  else
    return 1.;
}


//----------------------------------------------------------------------------------------------------------------------------
//             Type II atmospheric corection:  
//                       Multiplying factor to convert signal for a shower with theta to the reference zenith angle 0 [deg]

double AtmCorr::GetCorrTypeII( double r, double theta, int icomp, double hground )
{
  AtmosphereNS::Atmosphere fatm;  
  const double rhoAboveGroundRef=fatm.GetDensityAboveGround(hground,0.,2,-1); //theta=0 

  if (icomp>0 && icomp<4)
    {
      double r_km=r/1.e5;
      double alphaRho=GetAlphaRho(r,icomp);
      double rhoAboveGround=fatm.GetDensityAboveGround(hground,theta,2,-1);
      
      double fcorr= (1.-alphaRho*(rhoAboveGround-rhoAboveGroundRef)*1.e3/0.03);
      return 1./fcorr;    
    } 
  else
    return 1.;
}

//----------------------------------------------------------------------------------------------------------------------------
//             Type III atmospheric corection + Other possible effects  
//                       Multiplying factor to convert signal for a shower with psi to the reference psi angle 90 [deg]

double AtmCorr::GetParCorrTypeIII( double r,  int icomp, int ipar )
{
  if (ipar<0 || ipar>1) return 0.;

  const double f_mod1_par_WCD[4][3]= { { 0.000, 0.177, -0.120 } , { 0.000, -0.534, 0.028 } , { 0.000, -0.063, -0.041 } , { 0.068, 0.163,  0.000 }};
  const double f_mod2_par_WCD[4][3]= { { 0.300, 0.150, 0.000 } , { 0.300, 0.150, 0.000 } , { 0.300, 0.150, 0.000 } , { 2.600, -0.600, 0.000 }};

  const double f_mod1_par_Scin[4][3]= { { 0.000, 0.177,-0.120 } , { -0.164, -0.365, 0.000 } , {  0.000, -0.063, -0.041 } , { -0.224, 0.401, 0.000 }};
  const double f_mod2_par_Scin[4][3]= { { 0.300, 0.150, 0.000 } , { 0.130, 0.210, 0.000 } ,   { 0.300, 0.150, 0.000 } , { 3.500, -0.500, 0.000 }};


  const double *par;
  if (ipar==0 && fDetectorType==0 ) par= f_mod1_par_WCD[icomp];
  if (ipar==1 && fDetectorType==0 ) par= f_mod2_par_WCD[icomp];
  if (ipar==0 && fDetectorType==1 ) par= f_mod1_par_Scin[icomp];
  if (ipar==1 && fDetectorType==1 ) par= f_mod2_par_Scin[icomp];

  return par[0] + par[1]*r/1.e5 + par[2]*r*r/1.e10;
  
}

double AtmCorr::GetCorrTypeIII( double r, double theta, double psi, int icomp )
{

  //f_mod1/f_mod2
  double f_mod1=GetParCorrTypeIII(r,icomp,0);
  double f_mod2=GetParCorrTypeIII(r,icomp,1);

  //----
  double f_mod;
  if (icomp==3)  f_mod=f_mod1*pow(theta/M_PI*2.,f_mod2);
  else           f_mod=f_mod1*(1.-(cos(theta)+sin(theta)*f_mod2) );
  double fcorr=(1.+cos(psi)*f_mod);
  
  return 1./fcorr;
}


