#include "GeomAsym.h"

#include <Math/PdfFuncMathCore.h>
#include <Math/ProbFuncMathCore.h>
#include <Math/QuantFuncMathCore.h>

using namespace GeomAsymNS;

GeomAsym::GeomAsym(int DetectorType)
{
  if (DetectorType<0 || DetectorType>1 )
    {
      printf("GeomAsym: Wrong Detector Type %d \n",DetectorType);
      exit(1);
    }
  fDetectorType=DetectorType;
}

//------------------------------------------------------------------
//----Gamma function ( GSL implementation/ ROOT implementation/...)
//-------------------------------------------------------------------
double GeomAsym::gamma_cdf(double x, double alpha, double beta)
{
  return ROOT::Math::gamma_cdf(x, alpha, beta, 0.); 
}


double GeomAsym::gamma_quantile(double q, double alpha, double beta)
{
  return ROOT::Math::gamma_quantile(q, alpha, beta); 
}

//------------------------------------------------------------------
//----Parameterization of mean Pz
//-------------------------------------------------------------------

double GeomAsym::GetPzMeanPar(double r, int icomp,int ipar)
{
  if (ipar<0  || ipar  > nPar) return 0.;

  const double *par;
  if (fDetectorType==0) par=parPzMean_WCD[ipar][icomp];
  else                  par=parPzMean_TopScin[ipar][icomp];

  double r_km=r/1.e5;
  double val0km=par[0];
  double val1km=par[1];
  double B=par[2];
  
  double C=(val1km-val0km)/(exp(-B)-1.0);
  double A=val0km-C;
  
  double val=A+C*exp(-B*r_km);  
  if (val>=0.999) val=0.999;
  return val;
}



double GeomAsym::GetPzMean(double DX, double r, int icomp)  // DX (g/cm^2)    r (cm)
{

  if (icomp<0 || icomp > 3) return 0.;

  double valDX0=GetPzMeanPar( r, icomp, 0);
  double valDX1=GetPzMeanPar( r, icomp, 1);
  if (valDX1<valDX0) return (valDX1+valDX0)/2.;

  double DX0=DX_Eval[icomp][0];
  double DX1=DX_Eval[icomp][1];

  double A=1.-valDX0;
  double B=-log((1.-valDX1)/(1.-valDX0));
  double pz_mean=1.-A*exp(-(DX-DX0)/(DX1-DX0)*B);


  // <pz> is calculated in pz=(0.,1.)
  if (pz_mean < 0.000   ) return 0.000;
  if (pz_mean > 0.999   ) return 0.999;
  return pz_mean;
}


//------------------------------------------------------------------
//----Parameterization of RMS Pz
//-------------------------------------------------------------------

double GeomAsym::GetPzRMSPar(double r, int icomp, int ipar)
{
  if (ipar<0  || ipar  > nPar) return 0.;
  
  const double *par;
  if (fDetectorType==0) par=parPzRMS_WCD[ipar][icomp];
  else                  par=parPzRMS_TopScin[ipar][icomp];

  double r_km=r/1.e5;
  double val0km=par[0];
  double val1km=par[1];
  double B=par[2];
  
  double C=(val1km-val0km)/(exp(-B)-1.0);
  double A=val0km-C;

  double val= A+C*exp(-B*r_km);  
  if (val<=0.001) val=0.001;  
  return val;
}


double GeomAsym::GetPzRMS(double DX, double r, int icomp)  // DX (g/cm^2)    r (cm)
{
  if (icomp<0 || icomp > 3) return 0.;

  double valDX0=GetPzRMSPar( r, icomp, 0);
  double valDX1=GetPzRMSPar( r, icomp, 1);

  double DX0=DX_Eval[icomp][0];
  double DX1=DX_Eval[icomp][1];

  double A=valDX0;
  double B=-log(valDX1/valDX0);
  double pz_rms=A*exp(-(DX-DX0)/(DX1-DX0)*B);

  // RMS pz is calculated in pz=(0.,1.)    
  if (pz_rms <  0.001) return 0.001;
  if (pz_rms >  1.000) return 1.000;
  return pz_rms;
}


//--------------------------------------
//  Amod x Tmod (cos(theta_p),pz,r)
//--------------------------------------

double GeomAsym::GetTmodAmod_Par_i(double r, int itheta, int icomp, int ipar)
{
  double r_km=r/1.e5;
  //-----
  const double *par;
  if ( ipar==0 )
    {
      if      (fDetectorType==0)  par=parR0_WCD[icomp][itheta];
      else if (fDetectorType==1)  par=parR0_TopScin[icomp][itheta];
    }
  else if ( ipar==1 )
    {
      if      (fDetectorType==0)  par=parR1_WCD[icomp][itheta];
      else if (fDetectorType==1)  par=parR1_TopScin[icomp][itheta];
    }
  else if ( ipar==2 )
    {
      if      (fDetectorType==0)  par=parR2_WCD[icomp][itheta];
      else if (fDetectorType==1)  par=parR2_TopScin[icomp][itheta];
    }

  //-----

  double val=0.;
  if ( ipar==2 && icomp>0 && fDetectorType==0)
    val=par[0];
  else
    {
      double y0=par[0];
      double y1=par[1];
      double B=par[2];
  
      double C=(y1-y0)/(exp(-B)-1.0);
      double A=y0-C;
      
      val=A+C*exp(-B*r_km);
    
    }
  return val;    
}

double GeomAsym::GetTmodAmod(double ctheta_p, double *par, int icomp )
{
  double cthetaEval[2]={0.3,0.6};
  
  double x=1.-ctheta_p;
  double x0=1.-cthetaEval[0];
  double x1=1.-cthetaEval[1];

  double y0=par[0];
  double val;

  if (icomp==0)
    {
      val=1.+x*(y0-1.)/x0;
      if (val<0.) val=0.;      
    }
  else
    {   
      double y1=par[1];
      
      double b= ( (y1-1.)/x1 -(y0-1.)/x0  )/(x1-x0);
      double a= (y0-1.-x0*x0*b)/x0;
      
      val=1.+x*a+x*x*b;
      if (val<0.) val=0.;
    }
  return val;
}

double GeomAsym::GetTmodAmod(double ctheta_p, double pz, double r,  int icomp)
{
  if (pz>1.000) pz=1.000;
  if (ctheta_p < 0.) return 0.; 
  
  double par[2];
  for (int itheta=0;itheta<2;itheta++)
    {
      if (icomp==0 && itheta==1) par[itheta]=0.;

      double x=(PzEval[icomp][0]-pz)/(PzEval[icomp][0]-PzEval[icomp][1]);

      double p0    =GetTmodAmod_Par_i(r,itheta,icomp,0);
      double p1    =GetTmodAmod_Par_i(r,itheta,icomp,1);
      double p2    =GetTmodAmod_Par_i(r,itheta,icomp,2);
      if (p2<0.01) { p0=(p0+p1)/2.; p1=p0; p2=0.01;} //cte

      double B=p2;
      double C=(p1-p0)/(exp(-B)-1.0);
      double A=p0-C;

      par[itheta]=A+C*exp(-B*x);
    }
  return GetTmodAmod(ctheta_p,par,icomp);

}

//------------------------------------------------------------------------------------------------------------
// Utilities to calculate the truncated <pz> and the truncated dS0dpz integral
//------------------------------------------------------------------------------------------------------------

double GeomAsym::GetARad(double r, int icomp)
{
  double r_km=r/1.e5;
  double ARadPar_TopScin[4][3]= {
    {7.75627e-01 , 9.26886e-01, 3.72025e+00 },    
    {3.64786e-01 , 7.79736e-01, 1.53002e+00 },    
    {2.35261e-01 , 5.58190e-01, 9.11762e-01 },    
    {3.96769e-01 , 8.16377e-01, 1.98344e+00 } };

  double ARadPar_WCD[4][3]= { 
    {7.77336e-01 , 9.37276e-01, 5.45685e+00 },    
    {6.64014e-01 , 9.21383e-01, 5.64988e+00 },    
    {3.91995e-01 , 7.98898e-01, 2.12950e+00 },    
    {6.39475e-01 , 9.12194e-01, 5.97321e+00 } };


  double *par;
  if      (fDetectorType==0)  par=ARadPar_WCD[icomp];
  else if (fDetectorType==1)  par=ARadPar_TopScin[icomp];

  double y0=par[0];
  double y1=par[1];
  double B=par[2];

  double C=(y1-y0)/(exp(-B)-1.0);
  double A=y0-C;
  
  double val=A+C*exp(-B*r_km);
  return val;  
}

double GeomAsym::GetCosTheta_p(double pz, double r, double theta, double psi, int icomp)
{
  if (pz>1. || pz<-1.) return 0.;
  double ARad=GetARad(r,icomp);
  double cosTheta_p=pz*cos(theta)+ARad*sqrt(1.-pz*pz)*sin(theta)*cos(psi);
  if (cosTheta_p<-1.) return -1.;
  if (cosTheta_p> 1.) return  1.;
  return cosTheta_p;
}

double GeomAsym::GetPzCut(double r, double theta, double psi, int icomp)
{
  double cut=ctheta_pCut[fDetectorType];
  if (fabs(psi-M_PI/2.)<1.e-5*180./M_PI ) return cut/cos(theta);

  double ARad=GetARad(r,icomp);
  double alpha=sin(theta)*cos(psi)*ARad;
  double a=cos(theta)*cos(theta)+alpha*alpha;
  double b=-2.*cut*cos(theta);
  double c=(cut*cut-alpha*alpha);
  
  double d=b*b-4.*a*c;
  if (d<0.) d=0.;
  double pz_cut;
  if ( cos(psi)>0.) pz_cut=(-b-sqrt(d) )/2./a;
  else              pz_cut=(-b+sqrt(d) )/2./a;

  if (pz_cut<-1.) return -1.;
  if (pz_cut> 1.) return  1.;
  return pz_cut;
}

//------------------------------------------------------------------------------------------------------------
// Utilities to calculate the truncated <pz> and the truncated dS0dpz integral
//------------------------------------------------------------------------------------------------------------

double GeomAsym::GetPzMeanRange(double alpha, double beta, double pz0, double pz1)
{
  double B=gamma_cdf(1.-pz0, alpha, beta)-gamma_cdf(1.-pz1, alpha, beta);
  double A=gamma_cdf(1.-pz0, alpha+1, beta)-gamma_cdf(1.-pz1, alpha+1, beta);
  double pz=1.-alpha*beta*A/B;
  return pz;
}

double GeomAsym::GetShadow(double DX, double r, double theta, double psi, int icomp) // DX (g/cm^2)    r (cm)  theta/psi (rad)
{
  // <pz>, Variance pz
  double pz_mean=GetPzMean(DX,r,icomp);
  double pz_variance=pow(GetPzRMS(DX,r,icomp),2.);

  // Parameters of Gamma distribution
  //    Gamma dist:  x=(0,inf)   mean=alpha*beta  variance=alpha*beta^2  | but mean/RMS estimated x=(0,1)  differences are negligible 
  double beta=pz_variance/(1.-pz_mean);
  double alpha=(1.-pz_mean)*(1.-pz_mean)/pz_variance;
  double I0=gamma_cdf(2, alpha, beta); 

  //The cut in the p_z distribution
  double pz_cut=GetPzCut(r,theta,psi,icomp);
  
  //The shadowing factor
  double fshadow=gamma_cdf(1.-pz_cut, alpha, beta)/I0; 
 
  return fshadow;  
}


//------------------------------------------------------------------------------------------------------------
// Multiplying factor to convert the signal in a spherical tank (S0) to the signal in an Auger tank (S)
//           Corrections due to shadowing effects/Tmod/Amod
//------------------------------------------------------------------------------------------------------------
double GeomAsym::GetGeomAsym(double DX, double r, double theta, double psi, int icomp ) // DX (g/cm^2)    r (cm)  theta/psi (rad)
{
  int nSteps=4;
  return GetGeomAsym(DX,r,theta,psi,icomp,nSteps);
}


double GeomAsym::GetGeomAsym(double DX, double r, double theta, double psi, int icomp, int nSteps) // DX (g/cm^2)    r (cm)  theta/psi (rad)
{
 
  const double pzMin=-1.,pzMax=1.;

  // <pz>, Variance pz
  double pz_mean=GetPzMean(DX,r,icomp);
  double pz_variance=pow(GetPzRMS(DX,r,icomp),2.);
 
  // Parameters of Gamma distribution
  //    Gamma dist:  x=(0,inf)   mean=alpha*beta  variance=alpha*beta^2  | but mean/RMS estimated in x=(0,1)  differences are negligible 
  double beta=pz_variance/(1.-pz_mean);
  double alpha=(1.-pz_mean)*(1.-pz_mean)/pz_variance;
  double I0=gamma_cdf(2, alpha, beta); 

  //The cut in the p_z distribution
  double pz_cut=GetPzCut(r,theta,psi,icomp);
  //
  double cdf[2],pz[2];
  pz[0]=pz_cut;
  cdf[0]=gamma_cdf(1.-pz[0], alpha, beta)/I0;   

   
  double cdfStep=cdf[0]/double(nSteps);
  //printf("nSteps=%d cdfStep=%lf <pz>=%e RMS pz %e  pz_cut %e\n",nSteps,cdfStep,pz_mean,sqrt(pz_variance),pz_cut);
  
  double val=0.;
  for (int istep=0;istep<nSteps;istep++)
    {
      bool stop=false;

      if ( pz[0]>0.999 )
	{
	  pz[1]=1.;
	  cdf[1]=0.;
	  stop=true;
	}
      else 
	{
	  cdf[1]=cdf[0]-cdfStep;
	  pz[1]=1.-gamma_quantile(cdf[1]*I0,alpha,beta);   
	  if (pz[1]>0.999) { pz[1]=1.; cdf[1]=0.; stop=true; } //Fixes bug in gamma_quantile
	}

      double pzBin=GetPzMeanRange(alpha,beta,pz[0],pz[1]);
      double ctheta_p=GetCosTheta_p( pzBin, r, theta, psi,icomp);
      double AmodTmod=GetTmodAmod( ctheta_p,  pzBin ,  r,  icomp );
      val+= ( (cdf[0]-cdf[1])*AmodTmod );

      //printf("pz=(%e,%e) (%5.4lf -> %5.4lf) cdf=(%e,%e) AmodTmod=%e ctheta_p=%e | Bin integral=%e \n",pz[0],pz[1],
      //	     pzBin,pz[1]-pz[0],cdf[0],cdf[1],AmodTmod,ctheta_p,(cdf[0]-cdf[1])*AmodTmod );
      
      if ( stop ) break;
      cdf[0]=cdf[1];
      pz[0]=pz[1];
    }
  //printf("Integral %lf \n\n\n",val);
  
  return val;
}


